/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 2002, Emmanuel CURIS
   Cration : le 14 avril 1998

   Module de cration d'un modle structural
   Gestion des chemins (au sens FEFF : ensemble ordonn des atomes parcouru par l'lectron)

   Un chemin est un ensemble ordonn d'atomes, par lesquels passe l'lectron.
   Il commence et fini toujours pas l'absorbeur -> il n'est pas inclus au chemin.
   Un atome peut apparatre plusieurs fois dans le modle, mais pas deux fois de suite.

   Note : il n'est pas vident qu'un tel modle soit le meilleur pour du XANES...
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "graphes.h"

#include "atomes.h"
#include "chemins.h"
#include "couches.h"
#include "modeles.h"
#include "science.h" /* Au cas o M_PI n'est pas dfini */

#include "atm_chemins.txt"

/* ~~~~~~~~ Variables locales  ce module ~~~~~~~~ */

struct infos_affichage
{
  Couche *couche;
  Chemin *chemin;
  unsigned int num_chemin;
  unsigned int num_variable;
  unsigned int prem_atome;
  unsigned int atm_choisi;
};

/* Pour la bote de dialogue */
#define LGMAX_TXT_NOM_ATOME     23
#define LGMAX_TXT_NOM_SIGNAL    48
#define LGMAX_TXT_NOM_AMPLITUDE 48
#define LGMAX_TXT_NOM_PHASE     48

#define NOMBRE_ATOMES_VISIBLES  (PC_NOM_DERN_ATM - PC_NOM_ATOME_1 + 1)

/* ~~~~~~~~ Fonctions locales  ce module ~~~~~~~~ */

#ifdef ancien
static void    affiche_chemin   (struct infos_affichage *infos);
static void    enregistre_chemin(struct infos_affichage *infos);
static void    aff_lst_atomes   (struct infos_affichage *infos);

static void    prepare_modif_chemin(void);
static boolean GER_modifier_chemin(unsigned int objet, int x, int y);
#endif

/* ~~~~~~~~ Ralisation pratique du module ~~~~~~~ */

/* ~-~-~-~- Cration, effacement,... des chemins - */

/* -------- Cration d'un nouveau chemin, pour un absorbeur donn -----------\
 |   Renvoie l'adresse du chemin cr (NULL en cas d'erreur)                 |
 |                                                                           |
 | nbr_atomes = le nombre d'atomes au total dans le chemin, absorbeur exclus |
 \-------------------------------------------------------------------------- */

Chemin *creer_chemin(unsigned int nbr_atomes)
{
  Chemin *chemin;

  if (nbr_atomes == 0) return NULL; /* Au moins 1 atome dans le chemin ! */

  chemin = (Chemin *) malloc(sizeof(Chemin));
  if ( NULL == chemin ) {
    (void) fprintf(stderr, "creer_chemin : mmoire insuffisante pour rserver le chemin\n");
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/chemins.c", "creer_chemin", "chemin");
    return NULL;
  }

  /* Initialisation du chemin */
  chemin->nbr_atomes = nbr_atomes;
  chemin->atomes = (Atome **) calloc(nbr_atomes, sizeof(Atome *));
  if ( NULL == chemin->atomes ) {
    (void) fprintf(stderr, "creer_chemin : mmoire insuffisante pour rserver la table des atomes\n");
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/chemins.c", "creer_chemin", "chemin->atomes");
    free(chemin);
    return NULL;
  }
  chemin->exf_calcule = NULL;
  chemin->amplitude = NULL;
  chemin->phase = NULL;
  chemin->longueur_reelle = 0.0; /* Somme des distances */

  chemin->multiplicite = 1.0;
  chemin->pseudo_longueur = 0.0;
  chemin->dw_global = 0.0;

  chemin->nbr_equivalents = 0;
  chemin->equivalents = NULL;

  return chemin; /* On l'a finalement cr */
}

/* -------- Effacer un chemin, avec vrifications ventuelles ---\
 |   Renvoie FAUX en cas de problme, VRAI si effacement russi  |
 |                                                               |
 | chemin    = le chemin  effacer                               |
 | forcer    = si VRAI, on ne demande pas de confirmation        |
 | atomes    = si VRAI, on efface les atomes du chemin           |
 | graphes   = si VRAI, on efface les graphes du chemin          |
 |    pour les 2 derniers, si QUESTION on demande                |
 \-------------------------------------------------------------- */

boolean CHEMIN_Effacer(Couche *couche, Chemin *chemin, boolean forcer, boolean atomes, int graphes)
{
  unsigned int i;

  if ( NULL == chemin ) return VRAI;

  /* Vrifications */
  if (forcer == FAUX) {
    int reponse;

    reponse = BI_ERR_alerte( NULL, &Q_EffacerChemin, VRAI );
    if ( 2 == reponse )
      return FAUX;
  }

  /* On supprime tout lien ventuel avec la couche d'ajustement  laquelle il appartient */
  if ( couche != NULL ) {
    if (couche->chemin == chemin) { /* On perd les chemins quivalents avec ! */
      couche->chemin = NULL;
    }
  }

  /* On efface les chemins quivalents */
  for ( i = 0 ; i < chemin->nbr_equivalents ; i++ ) {
    int code_effacement = EFF_CHEMIN_SANS_GRAPHES;

    (void) fprintf(stderr, "%s,%s l.%u : Le chemin %s a %u quivalents, effacement du %i\n",
		   __FILE__, __FUNCTION__, __LINE__, chemin->nom, chemin->nbr_equivalents, i );

    if ( ( EFF_CHEMIN_TOUS_GRAPHES == graphes ) || ( EFF_CHEMIN_GRAPHES_EQU == graphes ) )
      code_effacement = EFF_CHEMIN_TOUS_GRAPHES;
    if ( FAUX == CHEMIN_Effacer(NULL, chemin->equivalents[i], VRAI, atomes, code_effacement ) ) {
      (void) fprintf(stderr, "CHEMIN_Effacer : problme en effaant les chemins quivalents [%s]\n", chemin->nom);
      return FAUX;
    }
  }
  free( chemin->equivalents ); chemin->equivalents = NULL;
  chemin->nbr_equivalents = 0;

  /* Effacement des graphes, des atomes  faire */
  if (VRAI == atomes) { /* On efface les atomes associs */
    OPTION_A_FAIRE( "Effacer les atomes impliqus dans le chemin" );
  }

  if (EFF_CHEMIN_TOUS_GRAPHES == graphes) {
    Graphe *tmp_exafs, *tmp_amplitude, *tmp_phase;

    /* On s'arrange pour viter des appels rcursifs */
    tmp_exafs = chemin->exf_calcule; chemin->exf_calcule = NULL;
    tmp_amplitude = chemin->amplitude; chemin->amplitude = NULL;
    tmp_phase = chemin->phase; chemin->phase = NULL;
    GRF_efface(tmp_exafs, VRAI, FAUX); /* On n'efface pas si protg ! */
    GRF_efface(tmp_amplitude, VRAI, FAUX);
    GRF_efface(tmp_phase, VRAI, FAUX);
  }

  /* Plus besoin de rien */
  free(chemin->atomes);
  free(chemin);

  return VRAI;
}

/* -------- Nommer un chemin, en contrlant la longueur -------- */

void nommer_chemin(Chemin *chemin, char *nom)
{
  if (chemin == NULL) return;
  if (nom == NULL) return;

  strncpy(chemin->nom, nom, LG_MAX_NOM_CHEMIN);
  *(chemin->nom + LG_MAX_NOM_CHEMIN) = 0; /* Pour tre sr d'avoir un zro final */
}

/* -------- Ajouter un (ou des) chemins quivalents  un chemin -------- */

int CHEMIN_AjouterEquivalent(Chemin *reference, Chemin *equivalent)
{
  unsigned int nbr_equivalents, idx_depart;
  unsigned int i;

  if ( (NULL == reference) || (NULL == equivalent) ) return -1;
  if ( reference == equivalent ) {
    (void) fprintf(stderr, "CHEMIN_AjouterEquivalent : un chemin est forcment quivalent  lui-mme !\n");
    return +1;
  }

  nbr_equivalents = 1 + equivalent->nbr_equivalents;
  if (NULL == reference->equivalents) {
    reference->equivalents = (Chemin **) malloc( sizeof(Chemin *) * nbr_equivalents );
    if ( NULL == reference->equivalents ) {
      (void) fprintf(stderr, "CHEMIN_AjouterEquivalent : impossible de crer la table des quivalents !\n");
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/chemins.c", "CHEMIN_AjouterEquivalent", "reference->equivalents");
      return -2;
    }

    idx_depart = 0;
  } else {
    Chemin **nv_equivalents;

    nv_equivalents = (Chemin **) realloc( reference->equivalents,
					  sizeof(Chemin *) * (nbr_equivalents + reference->nbr_equivalents) );
    if (NULL == nv_equivalents) {
      (void) fprintf(stderr, "CHEMIN_AjouterEquivalent : impossible de redimensionner la table des quivalents !\n");
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "Atomes/chemins.c", "CHEMIN_AjouterEquivalent", "reference->equivalents");
      return -2;
    }
    reference->equivalents = nv_equivalents;

    idx_depart = reference->nbr_equivalents;
  }

  /* On remplit la table avec les nouveaux chemins quivalents */
  reference->equivalents[idx_depart++] = equivalent;
  for (i = 0 ; i < equivalent->nbr_equivalents ; i++) {
    reference->equivalents[i + idx_depart] = equivalent->equivalents[i];
  }
  reference->nbr_equivalents += nbr_equivalents;

  /* Plus d'quivalents pour le chemin insr */
  free(equivalent->equivalents);
  equivalent->equivalents = NULL;
  equivalent->nbr_equivalents = 0;

  /* Tout s'est bien pass */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Modifier les paramtres d'un chemin ~-~-~-~-~-~-~-~- */

#ifdef ancien

int modifier_chemins(Absorbeur *absorbeur, Chemin *chemin)
{
  int retour;
  boolean etat;
  struct infos_affichage infos;

  if (absorbeur == NULL) return FAUX;

  if ( arbres[PARAM_CHEMINS].drapeaux.prepare == FAUX )
    prepare_modif_chemin();

/*   infos.absorbeur = absorbeur; */
/*  etat = (absorbeur->nbr_chemins > 0);*/
  OBJ_active(PARAM_CHEMINS, PC_EFF_CHEMIN, etat);
  OBJ_active(PARAM_CHEMINS, PC_TT_EFFACER, etat);
  OBJ_active(PARAM_CHEMINS, PC_FOND_CHEMIN, etat);
  OBJ_active(PARAM_CHEMINS, PC_NUM_CHEMIN, etat);
#ifdef sans_modele
  if (absorbeur->nbr_chemins > 0) /* Il y a des chemins : on commence au premier */
  {
    if (chemin == NULL) {
      infos.num_chemin = 1;
      chemin = absorbeur->chemins; /* On dmarre au premier, par dfaut */
    } else {
      Chemin *chm;

      for (infos.num_chemin = 1, chm = absorbeur->chemins ; chm != chemin ; chm = chm->suivant, infos.num_chemin++) {
	if (chm == NULL) {
#ifdef _DEVELOPPE_
	  fprintf(stderr, "modifier_chemins : le chemin %s n'est pas dans l'absorbeur %p\n", chemin->nom, absorbeur);
#endif
	  return FAUX;
	}
      }
    }

    (void) gere_ascenseur(PARAM_CHEMINS, PC_CHEMIN_AVANT, PC_FOND_CHEMIN, PC_NUM_CHEMIN, PC_CHEMIN_APRES, ASC_HORIZONTAL,
			  1, absorbeur->nbr_chemins, 1, infos.num_chemin, ROOT, 0, 0);
    infos.chemin = chemin;
  }
  else
#endif
  {
    unsigned int larg_asc;

    OBJ_position(PARAM_CHEMINS, PC_NUM_CHEMIN, 0, 0, NULL, NULL);
    OBJ_taille(PARAM_CHEMINS, PC_FOND_CHEMIN, -1, -1, &larg_asc, NULL);
    OBJ_taille(PARAM_CHEMINS, PC_NUM_CHEMIN, larg_asc, -1, NULL, NULL);
    infos.num_chemin = 0;
    infos.chemin = NULL;
  }

/*  met_valeur(PARAM_CHEMINS, PC_NBR_CHEMINS, absorbeur->nbr_chemins); */
  OBJ_MettreEntier_U( PARAM_CHEMINS, PC_NUM_CHEMIN, infos.num_chemin );
/*  OBJ_active(PARAM_CHEMINS, PC_DERN_CHEMIN, absorbeur->nbr_chemins > 1);*/

  /* Paramtres du chemin, maintenant... */
  affiche_chemin(&infos);

  arbres[PARAM_CHEMINS].pointeur = &infos;

  do {
    retour = ext_formdo(PARAM_CHEMINS, "Modifier les chemins");
    if (retour == PC_TT_EFFACER) {
    } else if (retour == PC_EFF_CHEMIN) {
    }
  } while (retour != PC_FINI);

  enregistre_chemin(&infos);

  return VRAI; /* Modification faite */
}

/* ~~~~~~~~ Prparation de l'interface graphique pour modifier les paramtres d'un chemin ~~~~~~~~ */

static void prepare_modif_chemin(void)
{
  int i;

  OBJ_deplace(PARAM_CHEMINS, PC_CHEMIN_AVANT, +1, 0, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_CHEMINS, PC_FOND_CHEMIN, +2, 0, DEPLACE_PIXELS);
  OBJ_agrandit(PARAM_CHEMINS, PC_FOND_CHEMIN, -2, 0);
  OBJ_deplace(PARAM_CHEMINS, PC_DERN_CHEMIN, +1, 0, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_CHEMINS, PC_NOM_CHEMIN, 0, -1, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_CHEMINS, PC_DEBYEWALLER, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_CHEMINS, PC_R, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_CHEMINS, PC_GAMMA, 0, +1, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_CHEMINS, PC_VAL_INITIALE, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_CHEMINS, PC_PROPOSER, 0, +1, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_CHEMINS, PC_CONSTANTE, 0, -1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_CHEMINS, PC_VAL_MAX, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_CHEMINS, PC_VAL_PAS, 0, +2, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_CHEMINS, PC_VAL_RELATION, 0, +1, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_CHEMINS, PC_TT_AJOUTER, 0, +1, DEPLACE_PIXELS);

  OBJ_agrandit(PARAM_CHEMINS, PC_FND_LST_ATMS, 0, NOMBRE_ATOMES_VISIBLES - 1);
  for (i = 0 ; i < NOMBRE_ATOMES_VISIBLES ; i++)
    OBJ_deplace(PARAM_CHEMINS, PC_NOM_ATOME_1 + i, 0, i, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_CHEMINS, PC_ATM_AVANT, -1, 0, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_CHEMINS, PC_FOND_ASC_ATM, -1, +1, DEPLACE_PIXELS);
  OBJ_agrandit(PARAM_CHEMINS, PC_FOND_ASC_ATM, 0, NOMBRE_ATOMES_VISIBLES - 2);
  OBJ_deplace(PARAM_CHEMINS, PC_ATM_SUIVANT, -1, NOMBRE_ATOMES_VISIBLES - 1, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_CHEMINS, PC_AMPLITUDE, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_CHEMINS, PC_PHASE, 0, +2, DEPLACE_PIXELS);

  arbres[PARAM_CHEMINS].drapeaux.prepare = VRAI;
  arbres[PARAM_CHEMINS].fonction.arbre = GER_modifier_chemin;
}

/* ~~~~~~~~ Gestion de l'interface graphique pour modifier les paramtres d'un chemin ~~~~~~~~ */

static boolean GER_modifier_chemin(unsigned int objet, int x, int y)
{
  struct infos_affichage *infos;

  infos = (struct infos_affichage *) arbres[PARAM_CHEMINS].pointeur;

  if ( (objet >= PC_NOM_ATOME_1) && (objet <= PC_NOM_DERN_ATM) ) {
    infos->atm_choisi = infos->prem_atome + (objet - PC_NOM_ATOME_1);
    aff_lst_atomes(infos);
  }

  switch (objet)
  {
  case PC_PREM_CHEMIN:
    enregistre_chemin(infos);
    infos->num_chemin = 1;
/*    (void) gere_ascenseur(PARAM_CHEMINS, PC_CHEMIN_AVANT, PC_FOND_CHEMIN, PC_NUM_CHEMIN, PC_CHEMIN_APRES, ASC_HORIZONTAL,
			  1, infos->absorbeur->nbr_chemins, 1, infos->num_chemin, ROOT, 0, 0);*/
    affiche_chemin(infos);
    OBJ_active(PARAM_CHEMINS, PC_CHEMIN_AVANT, FORCE_NON);
    OBJ_active(PARAM_CHEMINS, PC_CHEMIN_APRES, FORCE_OUI);
    return VRAI;

  case PC_DERN_CHEMIN:
    enregistre_chemin(infos);
/*    infos->num_chemin = infos->absorbeur->nbr_chemins;
    (void) gere_ascenseur(PARAM_CHEMINS, PC_CHEMIN_AVANT, PC_FOND_CHEMIN, PC_NUM_CHEMIN, PC_CHEMIN_APRES, ASC_HORIZONTAL,
			  1, infos->absorbeur->nbr_chemins, 1, infos->num_chemin, ROOT, 0, 0);*/
    affiche_chemin(infos);
    OBJ_active(PARAM_CHEMINS, PC_CHEMIN_AVANT, FORCE_OUI);
    OBJ_active(PARAM_CHEMINS, PC_CHEMIN_APRES, FORCE_NON);
    return VRAI;

  case PC_CHEMIN_AVANT:
  case PC_FOND_CHEMIN:
  case PC_NUM_CHEMIN:
  case PC_CHEMIN_APRES:
    {
/*      int nv_chemin = gere_ascenseur(PARAM_CHEMINS, PC_CHEMIN_AVANT, PC_FOND_CHEMIN, PC_NUM_CHEMIN, PC_CHEMIN_APRES, ASC_HORIZONTAL,
				     1, infos->absorbeur->nbr_chemins, 1, infos->num_chemin, objet, x, y);
      if (nv_chemin != infos->num_chemin) {
	enregistre_chemin(infos);
	infos->num_chemin = nv_chemin;
	affiche_chemin(infos);
      }*/
    }
    break;

  case PC_UTILISER:
/*    if (infos->chemin->couche != NULL) {
      infos->chemin->couche->utilisee = OBJ_choisit(PARAM_CHEMINS, PC_UTILISER, ECHANGE);
      if (infos->chemin->couche == couche_editee) {
	OBJ_choisit(PARAM_COUCHES, COUCHE_UTILISER, couche_editee->utilisee);
      }
    }
*/
    break;
  }

  return VRAI;
}

/* Afficher tout ce qui se rapporte  un chemin */

static void affiche_chemin(struct infos_affichage *infos)
{
  unsigned int i;
  Chemin *chemin;

#ifdef sans_modele
  for (i = 1, chemin = infos->absorbeur->chemins ; i < infos->num_chemin ; i++, chemin = chemin->suivant) {
    if (chemin == NULL) {
#ifdef _DEVELOPPE_
      fprintf(stderr, "affiche_chemin : Impossible d'atteindre le chemin %u\n", i);
#endif
      break;
    }
  }
#endif

  infos->chemin = chemin;
  if (chemin != NULL) {
    OBJ_active(PARAM_CHEMINS, PC_PREM_CHEMIN, infos->num_chemin > 1);
/*    OBJ_active(PARAM_CHEMINS, PC_DERN_CHEMIN, infos->num_chemin < infos->absorbeur->nbr_chemins);*/

    infos->prem_atome = 0;
    infos->atm_choisi = 0;
    met_chaine(PARAM_CHEMINS, PC_NOM_CHEMIN, 0, chemin->nom, TXT_COPIER);
    if (chemin->exf_calcule) {
      met_chaine(PARAM_CHEMINS, PC_SIGNAL, LGMAX_TXT_NOM_SIGNAL, chemin->exf_calcule->nom, TXT_COPIER);
    } else {
      met_chaine(PARAM_CHEMINS, PC_SIGNAL, LGMAX_TXT_NOM_SIGNAL, "Pas de signal EXAFS associ", TXT_COPIER);
    }
    if (chemin->amplitude) {
      met_chaine(PARAM_CHEMINS, PC_AMPLITUDE, LGMAX_TXT_NOM_AMPLITUDE, chemin->amplitude->nom, TXT_COPIER);
    } else {
      met_chaine(PARAM_CHEMINS, PC_AMPLITUDE, LGMAX_TXT_NOM_AMPLITUDE, "Pas d'amplitude associe", TXT_COPIER);
    }
    if (chemin->phase) {
      met_chaine(PARAM_CHEMINS, PC_PHASE, LGMAX_TXT_NOM_PHASE, chemin->phase->nom, TXT_COPIER);
    } else {
      met_chaine(PARAM_CHEMINS, PC_PHASE, LGMAX_TXT_NOM_PHASE, "Pas de dphasage associ", TXT_COPIER);
    }

/*    OBJ_active(PARAM_CHEMINS, PC_UTILISER, chemin->couche != NULL);*/
/*    if (chemin->couche != NULL)
      OBJ_choisit(PARAM_CHEMINS, PC_UTILISER, chemin->couche->utilisee);
    else
      OBJ_choisit(PARAM_CHEMINS, PC_UTILISER, FORCE_NON);
*/

    aff_lst_atomes(infos);
  }
}

/* Enregistrer les informations pour un chemin : valeurs des variables, nom,... */

static void enregistre_chemin(struct infos_affichage *infos)
{
  nommer_chemin(infos->chemin, lit_chaine(PARAM_CHEMINS, PC_NOM_CHEMIN, 0, NULL, TXT_COPIER));
}

/* Afficher la liste des atomes constituant un chemin */

static void aff_lst_atomes(struct infos_affichage *infos)
{
  Chemin *chemin;
  unsigned int i;

  chemin = infos->chemin;

  ARB_ascenseur(PARAM_CHEMINS, PC_ATM_AVANT, PC_FOND_ASC_ATM, PC_BOUT_ASC_ATM, PC_ATM_SUIVANT, ASC_VERTICAL,
		chemin->nbr_atomes, NOMBRE_ATOMES_VISIBLES, infos->prem_atome, ROOT, -1, -1, METTRE);
  for (i = 0 ; i < infos->prem_atome ; i++); /* On se place sur le premier atome affich */
  for ( ; i < NOMBRE_ATOMES_VISIBLES ; i++) { /* On affiche */
    int j;

    j = PC_NOM_ATOME_1 + i;
    if (i >= chemin->nbr_atomes) {
      OBJ_active(PARAM_CHEMINS, j, FORCE_NON);
      OBJ_choisit(PARAM_CHEMINS, j, FORCE_NON);
      met_chaine(PARAM_CHEMINS, j, LGMAX_TXT_NOM_ATOME, "", TXT_COPIER);
    } else {
      OBJ_active(PARAM_CHEMINS, j, FORCE_OUI);
      OBJ_choisit(PARAM_CHEMINS, j, i == infos->atm_choisi);
      if (i == infos->atm_choisi) {
	Atome *atome = chemin->atomes[i];

	OBJ_MettreCoordonnees( PARAM_CHEMINS, PC_X_ATM, PC_Y_ATM, PC_Z_ATM, &(atome->position), &(atome->err_position) );
	/* Distance  l'absorbeur - qui est, TOUJOURS, le dernier de la liste (car chemins ferms) */
	OBJ_MettreNombre( PARAM_CHEMINS, PC_DIST_ATM, ATM_distance(atome, chemin->atomes[chemin->nbr_atomes - 1]) );
	met_chaine(PARAM_CHEMINS, PC_TYPE_ATM, 0, ATM_donnees[atome->num_atomique - 1].symbole, TXT_COPIER);
      }
      met_chaine(PARAM_CHEMINS, j, LGMAX_TXT_NOM_ATOME, ATM_donnees[chemin->atomes[i]->num_atomique - 1].nom, TXT_COPIER);
    }
  }
}

#endif
