/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996-2003 Emmanuel CURIS
   Cration : le 5 novembre 2003

   Import de structures 3D (formats ASCII gnraliss, en gnral)
   Interface utilisateur : choix du format, des options,...
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "atomes.h"		/* Pour Molecule */
#include "cristallo.h"
#include "exafs.h"
#include "fichiers.h"
#include "interface.h"
#include "pdb.h"     /* Pour le paramtrage PDB */
#include "science.h" /* Pour M_PI */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

struct infos_arbre {
  int format_choisi;
  struct systeme_cristallin *systemes; /* Table des rseaux dfinis */
  unsigned long nbr_systemes; /* Nombre de systmes */
  int systeme_choisi;
  int groupe_choisi;
  boolean opt_appellees;
};

#define LGMAX_NOM_FORMAT 30
#define LGMAX_NOM_RESEAU 17
#define LGMAX_NOM_GROUPE 17

#include "atm_import.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int  GER_ImportASCII(unsigned int objet, int x, int y);
static void    prepare_arbre_ImportASCII(void);

static void ActiverFractionnaires(int sens);
static void ActiverParametres_F  (int sens);
static void ActiverParametres_U  (int sens);
static void ActiverSymetries     (int sens);
static void DesactiverFormatImpose(int sens);
static void InstallerFormat      (int format_choisi);

static int CRIS_LireInfosFormat(struct infos_lecture *infos);

static int PDB_LireOptions(PDB_infos *infos_pdb);

static const char *TXT_SuffixeFichier(int format_choisi);
static const char *TXT_TitreSelecteur(int format_choisi);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~-~ Choix du format, options gnrales ~-~-~-~-~-~-~-~-~ */

int MOL_ImportASCII(void)
{
  char **liste_noms;
  int retour;
  struct infos_arbre infos = { FORMAT3D_ASCII_XYZ, NULL, 0, 0, 0, FAUX };
  struct infos_lecture lecture = { 0U, 0UL, 0UL, '#', 0,
				   0U, 0U, 0U, 0U, 0U, 0U, 0U,
				   0U, 0U, 0U, 0U, 0U, 0U, 0UL, 0UL,
				   { NULL, 'T', 'P', NULL, 0, NULL }, 0U,
				   NULL, NULL, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL };
  unsigned int nbr_fichiers, i;
  Molecule *molecule_lue = NULL;

  /* On prpare les infos associes  l'arbre */
  /* 1. Lecture des systmes et groupes disponibles */
  infos.systemes = CRIS_CreerListeSystemes( PREFIXE_REP "/Lase/Cristallo/", &(infos.nbr_systemes) );
  if ( NULL == infos.systemes ) { /* Pas de gnration de symtries possibles */
    OBJ_active ( IMPORT_CARINE, I3D_GENERE_SYM, FORCE_NON);
    OBJ_choisit( IMPORT_CARINE, I3D_GENERE_SYM, FORCE_NON);
    ActiverSymetries( FAUX );
  } else {
    OBJ_active( IMPORT_CARINE, I3D_GENERE_SYM, FORCE_OUI );
    /* On installe d'abord le systme : met  jour les champs ditables ; quand on dsactive tout ensuite au cas o, pas de surprise ! */
    OBJ_ChoisirSystemeCr( IMPORT_CARINE, I3D_BRAVAIS, LGMAX_NOM_RESEAU, 0, METTRE, infos.systemes, infos.nbr_systemes,
			  I3D_VALEUR_A, I3D_VALEUR_B, I3D_VALEUR_C, I3D_VAL_ALPHA, I3D_VAL_BETA, I3D_VAL_GAMMA );
    /* On active ce qu'il faut pour les coordonnes fractionnaires */
    ActiverFractionnaires( OBJ_choisit( IMPORT_CARINE, I3D_FRACTION, QUESTION ) );
  }

  /* 2. Installer le format choisi */
  InstallerFormat( infos.format_choisi );

  if ( FAUX == arbres[ IMPORT_CARINE ].drapeaux.prepare ) {
    prepare_arbre_ImportASCII();
  }

  /* On associe les infos  l'arbre */
  ARB_MettreComplement( IMPORT_CARINE, &infos );

  /* On appelle l'arbre */
  retour = ARB_Bloquant( IMPORT_CARINE, txt_titre_arbre_import, principale, 0 );

  /* On dissocie les informations de l'arbre */
  ARB_MettreComplement( IMPORT_CARINE, NULL );

  /* Si on annule, on efface tout et on ne fait rien ! */
  if ( I3D_ANNULER == retour ) {
    CRIS_EffacerListeSystemes( infos.systemes, infos.nbr_systemes ); /* Plus besoin de cette liste des systmes */
    return +1;
  }

  /* Si on demande les symtries cristallines */
  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_GENERE_SYM, QUESTION) ) {
    /* On mmorise le groupe d'espace  utiliser, pour pouvoir effacer les autres */
    lecture.systeme = infos.systemes[ infos.systeme_choisi ];
    infos.systemes[ infos.systeme_choisi ].nom = NULL; /* On ne le librera pas en l'effaant ! */
    infos.systemes[ infos.systeme_choisi ].dossier = NULL;
    infos.systemes[ infos.systeme_choisi ].nbr_groupes = 0;
    infos.systemes[ infos.systeme_choisi ].groupes = NULL; /* Mme chose */

    lecture.groupe_choisi = infos.groupe_choisi;
  }

  /* On efface tous les systmes inutiles */
  CRIS_EffacerListeSystemes( infos.systemes, infos.nbr_systemes ); /* Plus besoin de cette liste des systmes */
  infos.systemes = NULL;

  /* On rcupre les informations dans l'arbre */
  retour = CRIS_LireInfosFormat( &lecture );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Incohrence dans les infos du format ! [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -1;
  }

  /* On demande la liste des fichiers */
  liste_noms = CH_selecteur( TXT_SuffixeFichier( infos.format_choisi ), TXT_TitreSelecteur( infos.format_choisi ), 0, NULL, &nbr_fichiers, fenetre_onglets );
  if ( NULL == liste_noms ) {
    CRIS_EffacerListeSystemes( &(lecture.systeme), 0 ); /* Plus besoin de ce systme non plus */
    return -2;
  }

  /* Pour chacun de ces fichiers, on charge */
  switch ( infos.format_choisi )
  {
  case FORMAT3D_ASCII_XYZ:
  case FORMAT3D_ASCII_CARINE: /* C'est un driv de l'ASCII pur ! */
    for ( i = 0 ; i < nbr_fichiers ; i++ ) {
      retour = MOL_Import_XYZ( liste_noms[ i ], infos.format_choisi, &lecture );
      if ( lecture.molecule != NULL ) {
	molecule_lue = lecture.molecule;
      }
    }
    break;

  case FORMAT3D_ASCII_PDB:
    for ( i = 0 ; i < nbr_fichiers ; i++ ) {
      PDB_infos infos_pdb;
      if ( FAUX == infos.opt_appellees ) {
	int retour;

	retour = ARB_Bloquant(  IMPORT_PDB, txt_titre_arbre_options_PDB, principale, 0 );
	if ( PDB_ANNULER == retour ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Chargement annul pour %s\n",
			 __FILE__, __FUNCTION__, __LINE__, liste_noms[ i ]);
	  continue;
	}
      }
      PDB_LireOptions( &infos_pdb );
      retour = MOL_Import_PDB( liste_noms[ i ], &infos_pdb );
      if ( infos_pdb.premiere_chaine != NULL ) {
	if ( infos_pdb.premiere_chaine->chaine != NULL ) {
	  molecule_lue = infos_pdb.premiere_chaine->chaine;
	}
      }
    }
    break;

  case FORMAT3D_ASCII_CCDB:
    for ( i = 0 ; i < nbr_fichiers ; i++ ) {
      retour = MOL_Import_CCDC( liste_noms[ i ], &lecture );
      if ( lecture.molecule != NULL ) {
	molecule_lue = lecture.molecule;
      }
    }
    break;

  case FORMAT3D_ASCII_CIF:
    OPTION_A_FAIRE( "Import au format CIF" );
    retour = 2478;
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Format inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, infos.format_choisi);
    retour = -1975;
  }

  /* Plus besoin du systme cristallin */
  CRIS_EffacerListeSystemes( &(lecture.systeme), 0 );

  /* Plus besoin de la maille */
  CRIS_EffacerMaille( lecture.maille ); lecture.maille = NULL;

  /* On libre la table des noms */
  /* On peut librer la liste des noms */
  for ( i = 0 ; i < nbr_fichiers ; i++ ) {
    free( liste_noms[ i ] ); liste_noms[ i ] = NULL;
  }
  free( liste_noms ); liste_noms = NULL;

  /* On va  la dernire molcule lue */
  MOL_aller( molecule_lue );

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le chargement des fichiers [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return -2;
  }

  /* Chargement russi */
  return 0;
}

/* -------- Analyser la bote de dialogue pour paramtrer le format de chargement -------- */

static int CRIS_LireInfosFormat(struct infos_lecture *infos)
{
  infos->drapeaux = 0;

  /* Combien de colonnes ? */
  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_AUTO_NBRCOL, QUESTION) ) {
    infos->nbr_colonnes = 0;
  } else {
    infos->nbr_colonnes = OBJ_LireEntier_U( IMPORT_CARINE, I3D_NBR_COLONNES );
  }

  /* Sparateur de colonnes [0 = espace au sens large] */
  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_SEP_TAB, QUESTION) ) {
    infos->separateur = 0;
  } else {
    infos->separateur = OBJ_LireCaractere( IMPORT_CARINE, I3D_SEP_AUTRE );
  }

  /* Commentaire */
  infos->commentaire = OBJ_LireCaractere( IMPORT_CARINE, I3D_COMMENTAIRE );

  /* Colonnes pour les informations essentielles */
  infos->col_x = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_X );
  infos->col_y = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_Y );
  infos->col_z = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_Z );

  infos->col_Z = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_SYMBOLE );

  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_AV_ETIQUETTE, QUESTION) ) {
    infos->col_txt = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_ETIQUET );
  } else {
    infos->col_txt = 0;
  }

  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_AVEC_RAYON, QUESTION) ) {
    infos->col_rayon = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_RAYON );
  } else {
    infos->col_rayon = 0;
  }

  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_AVEC_CHARGE, QUESTION) ) {
    infos->col_charge = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_CHARGE );
  } else {
    infos->col_charge = 0;
  }

  /* Doit-on prendre en compte les incertitudes indiques ? */
  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_GARDER_ERR, QUESTION) ) {
    infos->drapeaux |= TROISD_LIRE_ERREURS;
  }

  /* Nom propos par l'utilisateur */
  infos->nom = lit_chaine( IMPORT_CARINE, I3D_NOM_COMPOSE, 0, NULL, TXT_REMPLACE );

  /* Utilise-t-on des coordonnes fractionnaires ? */
  if ( FAUX == OBJ_choisit( IMPORT_CARINE, I3D_FRACTION, QUESTION) ) {
    /* NON => les autres informations ne sont pas utiles pour des coordonnes absolues ! */
    return 0;
  }
  infos->drapeaux |= TROISD_COORD_FRAC;

  /* On aura donc besoin de la maille */
  infos->maille = CRIS_CreerMaille();
  if ( NULL == infos->maille ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de prparer la maille / coordonnes fractionnaires\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Les paramtres de maille sont-ils inclus dans le fichier ? */
  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_P_DANS_FICH, QUESTION ) ) {
    /* Quelle ligne ? */
    infos->ligne_maille = OBJ_LireEntier_U( IMPORT_CARINE, I3D_LIGNE_PARAMS );

    /* Quelles colonnes ? */
    infos->col_a = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_A );
    infos->col_b = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_B );
    infos->col_c = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_C );

    infos->col_alpha = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_ALPHA );
    infos->col_beta  = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_BETA  );
    infos->col_gamma = OBJ_LireEntier_U( IMPORT_CARINE, I3D_COL_GAMMA );
  } else {
    infos->ligne_maille = 0;

    /* Valeurs des paramtres cristallins */
    infos->maille->a = OBJ_LireNombre( IMPORT_CARINE, I3D_VALEUR_A, NULL );
    infos->maille->b = OBJ_LireNombre( IMPORT_CARINE, I3D_VALEUR_B, NULL );
    infos->maille->c = OBJ_LireNombre( IMPORT_CARINE, I3D_VALEUR_C, NULL );

    infos->maille->alpha = OBJ_LireNombre( IMPORT_CARINE, I3D_VAL_ALPHA, NULL ); /* Angles donns en degrs, utiliss en radians */
    infos->maille->beta  = OBJ_LireNombre( IMPORT_CARINE, I3D_VAL_BETA , NULL ); /* --> ils devront tre convertis plus tard ! */
    infos->maille->gamma = OBJ_LireNombre( IMPORT_CARINE, I3D_VAL_GAMMA, NULL ); /* Pas tout de suite, car utilis en degrs dans les contrles */
  }

  /* Doit-on crer plusieurs mailles ? */
   if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_MULTIMAILLE, QUESTION) ) {
    infos->drapeaux |= TROISD_MULTIMAILLES;

    /* Limites des index de mailles */
    infos->i_min = OBJ_LireEntier_S( IMPORT_CARINE, I3D_I_MIN );
    infos->i_max = OBJ_LireEntier_S( IMPORT_CARINE, I3D_I_MAX );

    infos->j_min = OBJ_LireEntier_S( IMPORT_CARINE, I3D_J_MIN );
    infos->j_max = OBJ_LireEntier_S( IMPORT_CARINE, I3D_J_MAX );

    infos->k_min = OBJ_LireEntier_S( IMPORT_CARINE, I3D_K_MIN );
    infos->k_max = OBJ_LireEntier_S( IMPORT_CARINE, I3D_K_MAX );

    /* Un fragment par maille ? */
    if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_FRG_MAILLE, QUESTION) ) {
      infos->drapeaux |= TROISD_FRG_MAILLE;
    }
  }

  /* Doit-on crer les symtriques ? */
  if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_GENERE_SYM, QUESTION) ) {
    infos->drapeaux |= TROISD_SYMETRIES;

    /* Utilise-t-on les infos du fichier ? */
    if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_SYS_INCLUS, QUESTION ) ) {
      /* Quelle ligne ? */
      infos->ligne_groupe = OBJ_LireEntier_U( IMPORT_CARINE, I3D_LIGNE_SYS );

      /* Plus besoin du systme cristallin, puisque sera dtermin  partir du fichier */
      CRIS_EffacerListeSystemes( &(infos->systeme), 0 );
    } else {
      int retour;

      infos->ligne_groupe = 0;

      /* On lit les oprations de symtrie du groupe choisi */
      retour = CRIS_LireOperationsGroupes( infos->systeme.groupes + infos->groupe_choisi,
					   PREFIXE_REP "/Lase/Cristallo/", infos->systeme.dossier );
      if ( retour < 0 ) { /* Erreur grave */
	(void) fprintf(stderr, "%s,%sl.%u : ERREUR dans les oprations du groupe [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
	/* On annule donc l'utilisation des symtries */
	OPTION_A_FAIRE( "Message d'erreur, demander  l'utilisateur ce qu'il faut faire !" );
	infos->drapeaux &= ~ TROISD_SYMETRIES;
      }

      /* Si paramtres utilisateurs, on utilise le systme pour contrler les paramtres cristallins */
      if ( infos->maille->a > 0.0 ) {
	retour = CRIS_AdapteParametres( &(infos->systeme), &(infos->maille->a), &(infos->maille->b), &(infos->maille->c),
					&(infos->maille->alpha), &(infos->maille->beta), &(infos->maille->gamma) );
	if ( retour < 0 ) { /* Erreur grave */
	  (void) fprintf(stderr, "%s,%sl.%u : ERREUR dans les paramtres cristallins [code %d]\n",
			 __FILE__, __FUNCTION__, __LINE__, retour);
	  /* On annule donc l'utilisation des symtries */
	  OPTION_A_FAIRE( "Message d'erreur, demander  l'utilisateur ce qu'il faut faire !" );
	  infos->drapeaux &= ~ TROISD_SYMETRIES;
	}
      }
    }
  } else {
    infos->groupe_choisi = 0;
    infos->systeme.groupes = NULL;
  }

  /* On adapte les angles en radians */
  infos->maille->alpha *= M_PI / 180.0;
  infos->maille->beta  *= M_PI / 180.0;
  infos->maille->gamma *= M_PI / 180.0;

  /* La lecture s'est bien passe */
  return 0;
}

/* -------- Activer les options pour les coordonnes fractionnaires -------- */

static void ActiverFractionnaires(int sens)
{
  /* Coordonnes fractionnaires -> demander les paramtres de maille ! (fichier ou utilisateur) */
  OBJ_active(IMPORT_CARINE, I3D_P_DANS_FICH , sens);
  OBJ_active(IMPORT_CARINE, I3D_PARAM_MANUEL, sens);

  if ( FAUX == sens ) { /* Coordonnes absolues -> pas de symtries gnrables, paramtres de maille inutiles */
    OBJ_active( IMPORT_CARINE, I3D_GENERE_SYM, FORCE_NON );
    ActiverSymetries( FAUX );

    ActiverParametres_F( FAUX ); /* Aucun besoin des paramtres de maille ! */
    ActiverParametres_U( FAUX );
  } else {
    OBJ_active( IMPORT_CARINE, I3D_GENERE_SYM, FORCE_OUI );
    ActiverSymetries ( OBJ_choisit( IMPORT_CARINE, I3D_GENERE_SYM, QUESTION ) );
    if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_PARAM_MANUEL, QUESTION ) ) {
      ActiverParametres_F( FAUX ); /* Les paramtres sont donns par l'utilisateur */
      ActiverParametres_U( VRAI );
    } else {
      ActiverParametres_F( VRAI ); /* Les paramtres sont donns dans le fichier */
      ActiverParametres_U( FAUX );
    }
  }
}

static void ActiverParametres_F(int sens)
{
  /* Si sens = VRAI : les paramtres dans le fichier sont activs */
  OBJ_active( IMPORT_CARINE, I3D_LIGNE_PARAMS, sens );
  OBJ_active( IMPORT_CARINE, I3D_COL_A, sens );
  OBJ_active( IMPORT_CARINE, I3D_COL_B, sens );
  OBJ_active( IMPORT_CARINE, I3D_COL_C, sens );
  OBJ_active( IMPORT_CARINE, I3D_COL_ALPHA, sens );
  OBJ_active( IMPORT_CARINE, I3D_COL_BETA, sens );
  OBJ_active( IMPORT_CARINE, I3D_COL_GAMMA, sens );
}

static void ActiverParametres_U(int sens)
{
  /* Si sens = VRAI : les paramtres utilisateurs sont activs */
  OBJ_active( IMPORT_CARINE, I3D_VALEUR_A, sens );
  OBJ_active( IMPORT_CARINE, I3D_VALEUR_B, sens );
  OBJ_active( IMPORT_CARINE, I3D_VALEUR_C, sens );
  OBJ_active( IMPORT_CARINE, I3D_VAL_ALPHA, sens );
  OBJ_active( IMPORT_CARINE, I3D_VAL_BETA , sens );
  OBJ_active( IMPORT_CARINE, I3D_VAL_GAMMA, sens );
}

static void ActiverSymetries(int sens)
{
  /* Le groupe est-il dans le fichier ou donn par l'utilisateur ? */
  OBJ_active(IMPORT_CARINE, I3D_SYS_INCLUS, sens);
  OBJ_active(IMPORT_CARINE, I3D_SYS_MANUEL, sens);

  if ( FAUX == sens ) {
    /* On ne veut pas de symtries */
    OBJ_active( IMPORT_CARINE, I3D_BRAVAIS, FORCE_NON );
    OBJ_active( IMPORT_CARINE, I3D_GRP_ESPACE, FORCE_NON );

    OBJ_active( IMPORT_CARINE, I3D_LIGNE_SYS, FORCE_NON );
  } else {
    boolean grp_utilisateur;

    /* VRAI si le groupe est donn par l'utilisateur */
    grp_utilisateur = OBJ_choisit( IMPORT_CARINE, I3D_SYS_MANUEL, QUESTION );

    /* groupe choisi & son groupe d'espace */
    OBJ_active( IMPORT_CARINE, I3D_BRAVAIS, grp_utilisateur );
    OBJ_active( IMPORT_CARINE, I3D_GRP_ESPACE, grp_utilisateur );

    /* Ligne dcrivant le groupe dans le fichier */
    OBJ_active( IMPORT_CARINE, I3D_LIGNE_SYS, !grp_utilisateur );
  }
}

/* -------- Si format prdfini impos, on coupe tout ! -------- */

#define ForcerDesactivation(objet) OBJ_choisit( IMPORT_CARINE, objet, FORCE_OUI ); GER_ImportASCII( objet, 0, 0 )

static void DesactiverFormatImpose(boolean format_impose)
{
  if ( VRAI == format_impose ) {
    /* On coupe : on commence par dsactiver ce que l'on peut... */
    ForcerDesactivation( I3D_FRACTION );
    ForcerDesactivation( I3D_MULTIMAILLE );
    ForcerDesactivation( I3D_AVEC_RAYON );
    ForcerDesactivation( I3D_AVEC_CHARGE );
    ForcerDesactivation( I3D_AV_ETIQUETTE );
  }

  /* On active ou non ces options en fonction du sens
     format_impose = VRAI : on doit dsactiver...
  */
  OBJ_active( IMPORT_CARINE, I3D_FRACTION, ! format_impose );

  OBJ_active( IMPORT_CARINE, I3D_EN_TETE    , ! format_impose );
  OBJ_active( IMPORT_CARINE, I3D_COMMENTAIRE, ! format_impose );

  OBJ_active( IMPORT_CARINE, I3D_NBR_COLONNES, ! format_impose );
  OBJ_active( IMPORT_CARINE, I3D_AUTO_NBRCOL , ! format_impose );
  OBJ_active( IMPORT_CARINE, I3D_SEP_TAB     , ! format_impose );
  OBJ_active( IMPORT_CARINE, I3D_SEP_AUTRE   , ! format_impose );

  OBJ_active( IMPORT_CARINE, I3D_COL_X, ! format_impose );
  OBJ_active( IMPORT_CARINE, I3D_COL_Y, ! format_impose );
  OBJ_active( IMPORT_CARINE, I3D_COL_Z, ! format_impose );
  OBJ_active( IMPORT_CARINE, I3D_COL_SYMBOLE, ! format_impose );

  OBJ_active( IMPORT_CARINE, I3D_GARDER_ERR, ! format_impose );
}

/* -------- Prparer pour un format particulier -------- */

static void InstallerFormat(int format_choisi)
{
  met_chaine( IMPORT_CARINE, I3D_FORMAT, LGMAX_NOM_FORMAT, formats_3d[ format_choisi ], TXT_COPIER );

  switch ( format_choisi )
  {
  case FORMAT3D_ASCII_XYZ:
    OBJ_active( IMPORT_CARINE, I3D_OPTIONS, FORCE_NON );
    DesactiverFormatImpose( FAUX );

    /* Colonnes utiles */
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_X, 2 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_Y, 3 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_Z, 4 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_SYMBOLE, 1 );
    break;

  case FORMAT3D_ASCII_CARINE:
    OBJ_active( IMPORT_CARINE, I3D_OPTIONS, FORCE_NON );
    DesactiverFormatImpose( FAUX );

    /* Les paramtres sont dans le fichier */
    GER_ImportASCII( I3D_P_DANS_FICH, 0, 0 );

    /* Colonnes utiles */
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_X, 3 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_Y, 4 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_Z, 5 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_SYMBOLE, 1 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_CHARGE, 2 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_RAYON, 6 ); /* Mais tout le temps 1 : peu utile... */
    break;

  case FORMAT3D_ASCII_PDB:
    OBJ_active( IMPORT_CARINE, I3D_OPTIONS, FORCE_OUI );
    DesactiverFormatImpose( VRAI );
    break;

  case FORMAT3D_ASCII_CCDB:
    OBJ_active( IMPORT_CARINE, I3D_OPTIONS, FORCE_NON );
    DesactiverFormatImpose( FAUX );

    /* Colonnes utiles */
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_X, 2 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_Y, 3 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_Z, 4 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_SYMBOLE, 1 );
    break;

  case FORMAT3D_ASCII_CIF:
    DesactiverFormatImpose( VRAI );

    /* Colonnes utiles */
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_X, 3 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_Y, 4 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_Z, 5 );
    OBJ_MettreEntier_U( IMPORT_CARINE, I3D_COL_SYMBOLE, 2 );
    break;
  }
}

/* -------- Gestion de la bote de dialogue -------- */

static int GER_ImportASCII(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  boolean etat;
  struct infos_arbre *infos;
  int nv_choix;

  infos = ( struct infos_arbre * ) ARB_LireComplement( IMPORT_CARINE );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%sl.%u : pas d'informations associes  l'arbre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_ERREUR;
  }

  switch ( objet )
  {
  case I3D_P_DANS_FICH: /* Paramtres de maille dans le fichier */
    OBJ_choisit(IMPORT_CARINE, objet, FORCE_OUI);
    ActiverParametres_F( VRAI ); /* On active tous les champs de coordonnes fichier */
    ActiverParametres_U( FAUX ); /* On dsactive tous les champs de coordonnes utilisateur */
    return BI_EVENEMENT_TRAITE;

  case I3D_PARAM_MANUEL: /* Paramtres de maille entrs manuellement */
    OBJ_choisit(IMPORT_CARINE, objet, FORCE_OUI);
    ActiverParametres_F( FAUX ); /* On dsactive tous les champs de coordonnes fichier */
    ActiverParametres_U( VRAI ); /* On active tous les champs de coordonnes utilisateur */
    /* Mais on tient compte des contraintes du systme cristallin... */
    if ( VRAI == OBJ_choisit(IMPORT_CARINE, I3D_GENERE_SYM, QUESTION ) ) {
      OBJ_ChoisirSystemeCr( IMPORT_CARINE, I3D_BRAVAIS, LGMAX_NOM_RESEAU, 0, METTRE, infos->systemes, infos->nbr_systemes,
			    I3D_VALEUR_A, I3D_VALEUR_B, I3D_VALEUR_C, I3D_VAL_ALPHA, I3D_VAL_BETA, I3D_VAL_GAMMA );
      OBJ_ChoisirGroupeEsp( IMPORT_CARINE, I3D_GRP_ESPACE, LGMAX_NOM_GROUPE, infos->groupe_choisi, METTRE, infos->systemes + infos->systeme_choisi );
    }
    return BI_EVENEMENT_TRAITE;

#ifdef _DEVELOPPE_
  case I3D_SYS_INCLUS: /* Groupe d'espace inclus dans le fichier */
    OBJ_choisit( IMPORT_CARINE, objet, FORCE_OUI );
    OBJ_active( IMPORT_CARINE, I3D_LIGNE_SYS, FORCE_OUI );
    OBJ_active( IMPORT_CARINE, I3D_BRAVAIS, FORCE_NON );
    OBJ_active( IMPORT_CARINE, I3D_GRP_ESPACE, FORCE_NON );
    return BI_EVENEMENT_TRAITE;
#endif

  case I3D_SYS_MANUEL:
    OBJ_choisit( IMPORT_CARINE, objet, FORCE_OUI );
    OBJ_active( IMPORT_CARINE, I3D_LIGNE_SYS, FORCE_NON );
    OBJ_active( IMPORT_CARINE, I3D_BRAVAIS, FORCE_OUI );
    OBJ_active( IMPORT_CARINE, I3D_GRP_ESPACE, FORCE_OUI );
    return BI_EVENEMENT_TRAITE;

  case I3D_BRAVAIS: /* Choix du rseau de Bravais */
    nv_choix = OBJ_ChoisirSystemeCr( IMPORT_CARINE, I3D_BRAVAIS, LGMAX_NOM_RESEAU, infos->systeme_choisi, QUESTION, infos->systemes, infos->nbr_systemes,
				     I3D_VALEUR_A, I3D_VALEUR_B, I3D_VALEUR_C, I3D_VAL_ALPHA, I3D_VAL_BETA, I3D_VAL_GAMMA );
    if ( (nv_choix >= 0) && (nv_choix != (int) infos->systeme_choisi) ) {
      infos->systeme_choisi = (unsigned int) nv_choix;
      OBJ_active( IMPORT_CARINE, I3D_GRP_ESPACE, VRAI ); /* A pu tre dsactiv si aucun groupe pour le prcdent choisi */
      /* On installe le premier groupe de la liste */
      OBJ_ChoisirGroupeEsp( IMPORT_CARINE, I3D_GRP_ESPACE, LGMAX_NOM_GROUPE, 0, METTRE, infos->systemes + infos->systeme_choisi );
      /* On demande le groupe d'espace associ, s'il y en a */
      infos->groupe_choisi = OBJ_ChoisirGroupeEsp( IMPORT_CARINE, I3D_GRP_ESPACE, LGMAX_NOM_GROUPE, 0, QUESTION, infos->systemes + infos->systeme_choisi );
    }
    return BI_EVENEMENT_TRAITE;

  case I3D_GRP_ESPACE: /* Choix du groupe d'espace */
    infos->groupe_choisi = OBJ_ChoisirGroupeEsp( IMPORT_CARINE, I3D_GRP_ESPACE, LGMAX_NOM_GROUPE, infos->groupe_choisi, QUESTION, infos->systemes + infos->systeme_choisi );
    return BI_EVENEMENT_TRAITE;

  case I3D_FORMAT: /* Choix du format de fichier */
    nv_choix = SS_menu_texte( formats_3d, NOMBRE_FORMATS_3D, "Formats de fichier 3D", IMPORT_CARINE, I3D_FORMAT );
    if ( nv_choix != infos->format_choisi ) {
      infos->format_choisi = nv_choix;
      InstallerFormat( nv_choix );
    }
    return BI_EVENEMENT_TRAITE;

  case I3D_MULTIMAILLE:
    etat = OBJ_choisit( IMPORT_CARINE, objet, ECHANGE );
    OBJ_active( IMPORT_CARINE, I3D_I_MIN, etat );
    OBJ_active( IMPORT_CARINE, I3D_I_MAX, etat );
    OBJ_active( IMPORT_CARINE, I3D_J_MIN, etat );
    OBJ_active( IMPORT_CARINE, I3D_J_MAX, etat );
    OBJ_active( IMPORT_CARINE, I3D_K_MIN, etat );
    OBJ_active( IMPORT_CARINE, I3D_K_MAX, etat );
    return BI_EVENEMENT_TRAITE;

  case I3D_AUTO_NBRCOL:
    etat = OBJ_choisit( IMPORT_CARINE, objet, ECHANGE );
    OBJ_active( IMPORT_CARINE, I3D_NBR_COLONNES, !etat );
    return BI_EVENEMENT_TRAITE;

  case I3D_SEP_TAB:
    etat = OBJ_choisit( IMPORT_CARINE, objet, ECHANGE );
    OBJ_active( IMPORT_CARINE, I3D_SEP_AUTRE, !etat );
    return BI_EVENEMENT_TRAITE;

  case I3D_AVEC_RAYON:
    etat = OBJ_choisit( IMPORT_CARINE, objet, ECHANGE );
    OBJ_active( IMPORT_CARINE, I3D_COL_RAYON, etat );
    return BI_EVENEMENT_TRAITE;

  case I3D_AVEC_CHARGE:
    etat = OBJ_choisit( IMPORT_CARINE, objet, ECHANGE );
    OBJ_active( IMPORT_CARINE, I3D_COL_CHARGE, etat );
    return BI_EVENEMENT_TRAITE;

  case I3D_AV_ETIQUETTE:
    etat = OBJ_choisit( IMPORT_CARINE, objet, ECHANGE );
    OBJ_active( IMPORT_CARINE, I3D_COL_ETIQUET, etat );
    return BI_EVENEMENT_TRAITE;

  case I3D_OPTIONS:
    OBJ_choisit( IMPORT_CARINE, objet, FORCE_OUI );
    switch ( infos->format_choisi )
    {
    case FORMAT3D_ASCII_PDB:
      /* On appelle les options spciales PDB */
      ARB_Bloquant( IMPORT_PDB, txt_titre_arbre_options_PDB, NULL, IMPORT_CARINE );
      infos->opt_appellees = VRAI;
      break;
    }
    OBJ_choisit( IMPORT_CARINE, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case I3D_FRACTION:
    etat = OBJ_choisit( IMPORT_CARINE, objet, ECHANGE );
    ActiverFractionnaires( etat );
    return BI_EVENEMENT_TRAITE;

  case I3D_GENERE_SYM: /* Appel que si en fractionnaire */
    etat = OBJ_choisit( IMPORT_CARINE, objet, ECHANGE );
    ActiverSymetries( etat );
    if ( VRAI == etat ) { /* On installe donc le systme cristallin actif */
      OBJ_ChoisirSystemeCr( IMPORT_CARINE, I3D_BRAVAIS, LGMAX_NOM_RESEAU, 0, METTRE, infos->systemes, infos->nbr_systemes,
			    I3D_VALEUR_A, I3D_VALEUR_B, I3D_VALEUR_C, I3D_VAL_ALPHA, I3D_VAL_BETA, I3D_VAL_GAMMA );
      OBJ_ChoisirGroupeEsp( IMPORT_CARINE, I3D_GRP_ESPACE, LGMAX_NOM_GROUPE, infos->groupe_choisi, METTRE, infos->systemes + infos->systeme_choisi );
      if ( FAUX == OBJ_choisit( IMPORT_CARINE, I3D_SYS_MANUEL, QUESTION ) ) { /* Si on n'est pas en mode manuel */
	ActiverParametres_U( FAUX ); /* On dsactive les champs ventuellement activs par le systme cristallin */
      }
    } else {
      if ( VRAI == OBJ_choisit( IMPORT_CARINE, I3D_SYS_MANUEL, QUESTION ) ) {
	ActiverParametres_U( VRAI ); /* On active tous les champs de coordonnes */
      }
    }
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Prparation de la bote de dialogue -------- */

static void prepare_arbre_ImportASCII(void)
{
  if ( FAUX == arbres[ IMPORT_CARINE ].drapeaux.init ) {
    int retour;

    OPTION_A_FAIRE( "Documentation, import de structure 3D depuis un fichier ASCII" );
    retour = ARB_Initialiser( IMPORT_CARINE, NULL, GER_ImportASCII, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
  }

  /* Les titres des blocs */
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_PARAMS, 0, 500, DEPLACE_RELATIF );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_SYMETRIE, 0, 500, DEPLACE_RELATIF );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_MAILLES, 0, 500, DEPLACE_RELATIF );

  /* On amliore le placement des objets */
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_IGNORER , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_EN_TETE     , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_LIGNES  , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_FICHIERA, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_NBR_COLONNES, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_AUTO_NBRCOL , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_SEP     , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_SEP_TAB     , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_SEP_AUTRE   , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( IMPORT_CARINE, I3D_AVEC_RAYON  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_R       , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_RAYON   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_AVEC_CHARGE , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_CHARGE  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_CHARGE  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_AV_ETIQUETTE, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_ETIQUET , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_ETIQUET , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( IMPORT_CARINE, I3D_GENERE_SYM , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_GENERER, 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( IMPORT_CARINE, I3D_TXT_COL_A   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_A       , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_COL_B   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_B       , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_COL_C   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_C       , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_COL_ALPH, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_ALPHA   , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_COL_BETA, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_BETA    , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_COL_GAMM, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_COL_GAMMA   , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( IMPORT_CARINE, I3D_TXT_B       , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_VALEUR_B    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_VAL_BETA, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_VAL_BETA    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_C       , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_VALEUR_C    , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_VAL_GAMM, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_VAL_GAMMA   , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( IMPORT_CARINE, I3D_GRP_ESPACE, 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( IMPORT_CARINE, I3D_TXT_MIN, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_I_MIN  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_J_MIN  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_K_MIN  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_TXT_MAX, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_I_MAX  , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_J_MAX  , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( IMPORT_CARINE, I3D_K_MAX  , 0, +2, DEPLACE_PIXELS );

  arbres[IMPORT_CARINE].drapeaux.prepare = VRAI;
}

/* ~-~-~-~-~-~-~-~-~ Slecteur de fichiers ~-~-~-~-~-~-~-~-~ */

/* -------- Donne le suffixe des fichiers "classiques" de ce format -------- */

static const char *TXT_SuffixeFichier(int format_choisi)
{
  switch ( format_choisi )
  {
  case FORMAT3D_ASCII_XYZ:
    return "*.xyz,*.asc";

  case FORMAT3D_ASCII_CARINE:
    return "*.ace,*.acr";

  case FORMAT3D_ASCII_PDB:
    return "*.pdb";

  case FORMAT3D_ASCII_CCDB:
    return "*.cor";

  case FORMAT3D_ASCII_CIF:
    return "*.cif";
  }

  /* Format inconnu : tout est autoris */
  return "*";
}

/* -------- Titre  utiliser pour le slecteur de fichiers -------- */

static const char *TXT_TitreSelecteur(int format_choisi)
{
  switch ( format_choisi )
  {
  case FORMAT3D_ASCII_XYZ:
    return "Fichier ASCII 3D format libre";

  case FORMAT3D_ASCII_CARINE:
    return "Fichier ASCII Carine v. 3";

  case FORMAT3D_ASCII_PDB:
    return "Fichier PDB";

  case FORMAT3D_ASCII_CCDB:
    return "Fichier Cambridge";

  case FORMAT3D_ASCII_CIF:
    return "Fichier CIF";
  }

  /* Format inconnu */
  return NULL;
}

/* ~-~-~-~-~-~-~-~-~ Options spcfiques ~-~-~-~-~-~-~-~-~ */

/* -------- Lire & initialiser les options pour le chargement d'un fichier PDB -------- */

static int PDB_LireOptions(PDB_infos *infos_pdb)
{
  /* Base du nom : celui indiqu dans la bote principale */
  infos_pdb->base_nom = lit_chaine( IMPORT_CARINE, I3D_NOM_COMPOSE, 0, NULL, TXT_REMPLACE );

  /* Options de chargement */
  infos_pdb->drapeaux = 0;
  if ( VRAI == OBJ_choisit( IMPORT_PDB, PDB_EAU, QUESTION) ) {
    infos_pdb->drapeaux |= PDB_OPTIONS_CHARGER_EAU;
  }
  if ( VRAI == OBJ_choisit( IMPORT_PDB, PDB_H, QUESTION) ) {
    infos_pdb->drapeaux |= PDB_OPTIONS_CHARGER_H;
  }
  if ( VRAI == OBJ_choisit( IMPORT_PDB, PDB_1_CHAINE, QUESTION) ) {
    infos_pdb->drapeaux |= PDB_OPTIONS_UNE_CHAINE;
    infos_pdb->ref_type = OBJ_LireCaractere( IMPORT_PDB, PDB_CHAINE );
  } else {
    infos_pdb->ref_type = 0;
  }
  if ( VRAI == OBJ_choisit( IMPORT_PDB, PDB_1_MOLECULE, QUESTION) ) {
    infos_pdb->drapeaux |= PDB_OPTIONS_UNE_MOLECULE;
  }

  /* Aucune chane prpare */
  infos_pdb->premiere_chaine = NULL;
  infos_pdb->premier_groupe = NULL;

  /* On est au dbut partout */
  infos_pdb->type = 0;

  infos_pdb->ref_modele = 0;
  infos_pdb->num_modele = 0;

  infos_pdb->num_aa = 0;

  infos_pdb->code_retour = 0; /* 0 = PDB_LIGNE_OK */

  memset( infos_pdb->code, 0, 5 ) ;

  /* Initialisation russie */
  return 0;
}
