/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996-2003, Emmanuel Curis

   Interface graphique : routines générales indépendantes du système

   Gestion des fenêtres : création
   Remarque : comme c'est assez bas niveau, très peu de routines sont indépendantes du système !
*/

#include <stdio.h>
#include <stdlib.h>

#ifdef _UNIX_
# define _LIBRAIRIE_X_ /* Pour avoir accès aux fonctions internes */
#elif defined(_WINDOWS_)
# define _LIBRAIRIE_W_
#endif

#include "bib_interface.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

static const BI_Boite_Alerte ERR_ImpossibleOuvrirFenetreOpenGL = {
  ERR_ICONE_STOP,
  "Incident Open-GL",
  "Impossible d'ouvrir une fen"e_CIRC"tre Open-GL. Consultez les messages sur stderr/stdout pour comprendre ! On passe en mode sans Open-GL...",
  1, 1, 1,
  { "Tant pis", NULL, NULL, NULL, NULL }
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int      FEN_ChainerFenetre ( Fenetre *fenetre );
static Fenetre *FEN_ReserverFenetre( int type, int attributs );

/* ~-~-~-~-~-~-~-~- Gestion des structures ~-~-~-~-~-~-~-~- */

/* -------- Réserver la structure, l'initialiser -------- */

Fenetre *FEN_ReserverFenetre( int type, int attributs )
{
  Fenetre *fenetre;
  int retour;

  /* On réserve la place en mémoire */
  fenetre = (Fenetre *) calloc( 1, sizeof(Fenetre) ); /* Place de la structure */
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : mémoire insuffisante [demandée : %lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, (unsigned long) sizeof(Fenetre) );
    return NULL;
  }

  /* Initialisation : paramètres communs à toutes les versions
     Remarque : calloc met à 0 -> il reste peu de choses à initialiser !
  */
  fenetre->suivante = NULL;
  fenetre->type = type; /* Type */
  fenetre->attributs = attributs;

  /* Pas de titre ni d'infos */
  fenetre->titre = NULL;
  fenetre->infos = NULL;

  /* Position & dimensions (externes) initiales */
  fenetre->larg = BI_larg_defaut;
  fenetre->haut = BI_haut_defaut;

  /* Souris initialement au dehors */
  fenetre->vx_x = -1;
  fenetre->vx_y = -1;

  /* Pas d'arbre, de sous-fenêtres,... */
  fenetre->arbre = NULL;
  fenetre->ss_fenetres = NULL;

  /* Pas de routines de gestion */
  fenetre->ascenseur = ( void (*)(Fenetre *, int, int, int) ) NULL;
  fenetre->clavier   = ( int  (*)(Fenetre *, unsigned int, int) ) NULL;
  fenetre->clic      = ( int  (*)(Fenetre *, int, int, int, int, int) ) NULL;
  fenetre->message   = ( int  (*)(Fenetre *, Message *) ) NULL;
  fenetre->rect      = ( void (*)(Fenetre *, int, int, int, int) ) NULL;
  fenetre->redessin  = FEN_RedessinDefaut;
  fenetre->temps     = ( int  (*)(Fenetre *) ) NULL;

  /* Pas d'infos utilisateur */
  fenetre->info_1 = NULL;
  fenetre->info_2 = NULL;

  /* Initialisation dépendant du système graphique [zone de travail essentiellement] */
  retour = FEN_InitialiserFenetre( fenetre );
  if ( retour != 0 ) { /* Initialisation a échoué */
    (void) fprintf( stderr, "%s,%s l.%u : ERREUR dans FEN_InitialiserFenetre() [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour );
    FEN_LibererFenetre( fenetre );
  }

  /* Initialisation réussie, fenêtre créée & chaînée */
  return fenetre;
}

/* -------- Le chaînage -------- */

static int FEN_ChainerFenetre( Fenetre *fenetre )
{
  ENTREE_FONCTION( __FUNCTION__, "fenetre=%p", (void *) fenetre );

  if ( NULL == fenetre ) {
    SORTIE_FONCTION(  __FUNCTION__, "%d [fenetre=NULL]", -1 );
    return -1;
  }

  if ( NULL == BI_premiere_fenetre ) {
    /* Première fenêtre du chaînage */
    BI_premiere_fenetre = fenetre;
  } else {
    Fenetre *derniere_fenetre;

    /* On place la nouvelle fenêtre en fin de chaîne */
    derniere_fenetre = BI_premiere_fenetre;
    while ( derniere_fenetre->suivante != NULL )
      derniere_fenetre = derniere_fenetre->suivante;
    derniere_fenetre->suivante = fenetre;
  }

  /* Tout s'est bien passé */
  SORTIE_FONCTION(  __FUNCTION__, "%d [OK]", 0 );
  return 0;
}

/* -------- Création "complète" d'une nouvelle fenêtre -------- */

Fenetre *FEN_NouvelleFenetre( int type, int attributs, Fenetre *mere )
{
  Fenetre *fenetre;
  int retour;

  ENTREE_FONCTION( __FUNCTION__, "type=%d attributs=%d mere=%p", type, attributs, (void *) mere );

  /* On réserve la structure de la fenêtre */
  fenetre = FEN_ReserverFenetre( type, attributs );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "FEN_NouvelleFenetre : ERREUR dans FEN_ReserverFenetre()\n");
    SORTIE_FONCTION( __FUNCTION__, "%d [fenetre=NULL]", NULL );
    return NULL;
  }

  /* On la crée pour l'interface graphique utilisée */
  retour = FEN_CreerFenetre( fenetre, mere );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FEN_NouvelleFenetre : ERREUR dans FEN_CreerFenetre() [code %d]\n", retour);
    FEN_LibererFenetre( fenetre );
    SORTIE_FONCTION( __FUNCTION__, "%d [Erreur dans FEN_CreerFenetre(), code %d]", NULL, retour );
    return NULL;
  }

  /* Si c'est une fenêtre racine : on la place dans le chaînage des fenêtres */
  if ( NULL == mere ) {
    FEN_ChainerFenetre( fenetre );
  }

  /* On a fini, tout s'est bien passé */
  SORTIE_FONCTION( __FUNCTION__, "%p [OK]", (void *) fenetre );
  return fenetre;
}

/* -------- Création d'une fenêtre, cas le plus général ----------------\
 | Renvoie un pointeur sur la structure décrivant la fenêtre            |
 |                                                                      |
 | attributs : ceux que la fenêtres doit avoir (comme wind_create)      |
 | infos     : adresse de la chaîne d'informations                      |
 | titre     : adresse de la chaîne de titre                            |
 | redessin  : adresse de la fonction à appeler pour le redessin        |
 | ascenseur : adresse de la fonction de gestion des ascenseurs         |
 | clavier   : adresse de la fonction de gestion du clavier             |
 | clic      : adresse de la fonction de gestion du clic souris         |
 | message   : adresse de la fonction de gestion des messages           |
 | rect      : adresse de la fonction de gestion de MU_RECT1 (pointeur) |
 | temps     : adresse de la fonction de gestion du timer               |
 \--------------------------------------------------------------------- */

Fenetre *FEN_CreerFenetreSimple(int attributs, const char *titre, const char *infos,
			       void (*redessin) (Fenetre *, int, int, unsigned int, unsigned int),
			       void (*ascenseur)(Fenetre *, int, int, int),
			       int  (*clavier)  (Fenetre *, unsigned int, int),
			       int  (*clic)     (Fenetre *, int, int, int, int, int),
			       int  (*message)  (Fenetre *, Message *),
			       void (*rect)     (Fenetre *, int , int, int, int),
			       int  (*temps)    (Fenetre *) )
{
  Fenetre *fenetre;

  /* On prépare la fenêtre */
  fenetre = FEN_NouvelleFenetre( 0, attributs, NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FEN_NouvelleFenetre()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On met le titre */
  FEN_titre( fenetre, titre );
  if ( ( attributs & INFO ) && ( infos != NULL ) ) { /* Il y a une barre d'infos */
    FEN_infos(fenetre, infos);
  }

  /* On associe les routines de gestion des événements */
  FEN_AssocierEvenements( fenetre, redessin, ascenseur, clavier, clic, message, rect, temps );

  /* La fenêtre a été créée correctement, plus qu'à l'ouvrir (à la charge de l'appelant) */
  return fenetre;
}

/* -------- Fenêtre contenant juste un arbre GEM ------------ */

Fenetre *FEN_OuvrirArbre(Fenetre *mere, int attributs, unsigned int numero_arbre, const char *titre, int x, int y)
{
  Fenetre *fenetre;
  OBJECT *adresse = NULL;
  unsigned int larg, haut;

  ENTREE_FONCTION( __FUNCTION__, "mere=%p attributs=%d numero_arbre=%u titre=%s",
		   (void *) mere, attributs, numero_arbre, titre );

  /* Le numéro d'arbre est-il valide ? */
  if ( numero_arbre >= BI_nbr_arbres ) {
    (void) fprintf(stderr, "%s,%s l.%u : numéro d'arbre incorrect [%u / max = %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_arbre, BI_nbr_arbres - 1 );
    SORTIE_FONCTION( __FUNCTION__, "%p [Numéro d'arbre incorrect]", NULL );
    return NULL;
  }

  /* Une fenêtre contient-elle déjà l'arbre considéré ? */
  for ( fenetre = BI_premiere_fenetre ; fenetre ; fenetre = fenetre->suivante ) {
    /* Un simple contrôle du numéro de l'arbre n'est pas suffisant,
       car numéro = 0 pour une fenêtre sans arbre, or l'arbre 0 n'est pas forcément un menu ! */
    if ( FAUX == FEN_ContientArbre( fenetre->type ) ) continue; /* La fenêtre ne peut pas être associée à un arbre ! */

    if ( NULL == fenetre->arbre ) {
      /* Pas d'arbre associé => fenêtre autre
         Arrive en particulier si arbre = 0, or 0 = aussi valeur de numero_arbre si pas d'arbre...
           Exemple : sur fenêtre avec des onglets... */
#ifdef _DEVELOPPE_
      if ( fenetre->numero_arbre != 0 ) {
	(void) fprintf( stderr, "%s,%s l.%u : PROBLEME dans la fenêtre %s : associée l'arbre %u mais arbre NULL !\n",
			__FILE__, __FUNCTION__, __LINE__, fenetre->titre, fenetre->numero_arbre );
	BI_AfficherFenetre( fenetre );
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      }
#endif
      continue;
    }

    if ( fenetre->numero_arbre == numero_arbre ) { /* Fenêtre pour cet arbre déjà existante */
      DANS_FONCTION( "La fenêtre %p contient déjà l'arbre", (void *) fenetre );
#ifdef _DEVELOPPE_
      if ( fenetre->arbre != arbres + numero_arbre ) {
	(void) fprintf(stderr, "%s,%s l.%u : PROBLEME dans l'attribution arbre [%u : %p] / fenêtre [%s : %s déjà associé à %p : %u]\n",
		       __FILE__, __FUNCTION__, __LINE__, numero_arbre, (void *) (arbres + numero_arbre), fenetre->titre, titre, (void *) fenetre->arbre, (unsigned int) (fenetre->arbre - arbres));
	BI_AfficherFenetre( fenetre );
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      }
#endif
      
      if ( FAUX == fenetre->drapeaux.ouverte ) {
	int retour;

	/* Si déjà existante, ne fait rien -> évite des tests dépendant du système (type fenetre->handle == None) */
	retour = FEN_Reouvrir(fenetre);
	if ( retour <= 0 ) { /* Problème */
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FEN_Reouvrir(), code %d\n",
			 __FILE__, __FUNCTION__, __LINE__, retour);
	  SORTIE_FONCTION( __FUNCTION__, "%p [Erreur dans FEN_Reouvrir(), code %d]", NULL, retour );
	  return NULL;
	}
      }

      /* On met à jour le titre */
      FEN_titre( fenetre , titre );

      /* On passe la fenêtre au premier plan */
      FEN_premierplan( fenetre );
      SORTIE_FONCTION( __FUNCTION__, "%p [Fenêtre trouvée]", fenetre );
      return fenetre;
    }
  }

  /* Ici, on doit donc créer la fenêtre */
  DANS_FONCTION( "La fenêtre doit être créée" );

  /* Si on est là, c'est que la création est complètement nouvelle */
  fenetre = FEN_NouvelleFenetre( BI_FENETRE_ARBRE, attributs, mere );
  if ( fenetre == NULL ) { /* Fenêtre n'a pas pu être créée */
    (void) fprintf(stderr, "%s,%s l.%u : erreur dans FEN_NouvelleFenetre()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    SORTIE_FONCTION( __FUNCTION__, "%p [Erreur dans FEN_NouvelleFenetre()]", NULL );
    return NULL;
  }

  /* On met le titre */
  FEN_titre( fenetre, titre );

  /* On assigne l'arbre à la fenêtre */
  adresse = arbres[ numero_arbre ].adresse;
  fenetre->numero_arbre  = numero_arbre;
  fenetre->arbre         = arbres + numero_arbre;

  /* On détermine les dimensions de la fenêtre */
  larg = OB_WIDTH(adresse, ROOT);
  haut = OB_HEIGHT(adresse, ROOT);
  larg += fenetre->larg - fenetre->w_larg; /* Correction pour les éléments de fenêtre */
  haut += fenetre->haut - fenetre->w_haut;

  /* On les impose pour la fenêtre */
  if ( NULL == mere ) {
    /* Fenêtre directement gérée par le gestionnaire de fenêtres */
    FEN_ImposerTaille( fenetre, x, y, larg, haut );
  } else {
    /* Fenêtre interne */
    FEN_dimensions( fenetre, x, y, larg, haut );
  }

  /* On place l'arbre dedans */
  OB_X(adresse, ROOT) = fenetre->w_x;
  OB_Y(adresse, ROOT) = fenetre->w_y;

  /* On crée les fenêtres pour les boutons */
  OBJ_creer_fenetres( fenetre->arbre, fenetre->handle, ROOT, ROOT ); /* On crée tous les objets */

  /* On ouvre finalement la fenêtre */
  FEN_Ouvrir( fenetre );

  SORTIE_FONCTION( __FUNCTION__, "%p [Fenêtre créée]", fenetre );
  return fenetre;
}

/* -------- Fenêtre ne contenant qu'une image OpenGL -------- */

Fenetre *FEN_CreerFenetreGL(int attributs, const char *titre,
			    void (*redessin) (Fenetre *, int, int, unsigned int, unsigned int),
			    void (*ascenseur)(Fenetre *, int, int, int),
			    int  (*clavier)  (Fenetre *, unsigned int, int),
			    int  (*clic)     (Fenetre *, int, int, int, int, int),
			    int  (*message)  (Fenetre *, Message *message),
			    void (*rect)     (Fenetre *, int x, int y, int vx_x, int vx_y),
			    int  (*temps)    (Fenetre *) )
{
  Fenetre *fenetre;

  /* On prépare la fenêtre */
  fenetre = FEN_NouvelleFenetre( BI_FENETRE_GL, attributs, NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans FEN_NouvelleFenetre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On met le titre */
  FEN_titre( fenetre, titre );

  /* On associe les routines de gestion des événements */
  FEN_AssocierEvenements( fenetre, redessin, ascenseur, clavier, clic, message, rect, temps );

  return fenetre;
}

/* -------- Fenêtre mixte, arbre GEM + partie utilisateur -------- */

Fenetre *FEN_CreerAvecArbre(int attributs, const char *titre, const char *infos,
			    unsigned int numero_arbre, int position_arbre, int (*init_arbre)(unsigned int),
			    void (*redessin)(Fenetre *, int, int, unsigned int, unsigned int),
			    int  (*clavier) (Fenetre *, unsigned int, int),
			    int  (*clic)    (Fenetre *, int, int, int, int, int),
			    int  (*message) (Fenetre *, Message *message),
			    void (*rect)    (Fenetre *, int, int, int, int),
			    int  (*temps)   (Fenetre *) )
{
  Fenetre *fenetre;
  OBJECT *adresse_arbre __attribute__((__unused__));

  ENTREE_FONCTION( __FUNCTION__, "attributs=%d titre=%s infos=%s numero_arbre=%d position_arbre=%d init_arbre=%p...",
		   attributs, titre, infos, numero_arbre, position_arbre, (void *) init_arbre );

  /* On contrôle si l'arbre a déjà une fenêtre associée */
  fenetre = ARB_QuelleFenetre( arbres + numero_arbre, VRAI );
  if ( fenetre != NULL ) {
    /* La fenêtre existe déjà ! */
    if ( FAUX == fenetre->drapeaux.ouverte ) {
      /* Elle est fermée : on l'ouvre de nouveau */
      FEN_Reouvrir( fenetre );
    }

    /* On la fait passer au premier plan */
    FEN_premierplan(fenetre);

    /* On renvoie cette fenêtre */
    SORTIE_FONCTION( __FUNCTION__, "%p [Arbre déjà ouvert]", fenetre );
    return fenetre;
  }

  /* --- Si on est là : on doit créer la fenêtre --- */

  /* a. On prépare l'arbre, si nécessaire */
  if ( FAUX == arbres[ numero_arbre ].drapeaux.init ) {
    int retour;

    if ( init_arbre != ( int (*)(unsigned int) ) NULL ) {
      retour = init_arbre( numero_arbre );
    } else {
      AVERTISSEMENT( "Arbre %u sans procédure d'initialisation !", numero_arbre );
      retour = ARB_Initialiser( numero_arbre, NULL,  ( int (*)(unsigned int, int, int) ) NULL, ( int (*)(unsigned int, int, int *) ) NULL );
    }

    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Initialisation de l'arbre a échoué [arbre %u]\n",
		     __FILE__, __FUNCTION__, __LINE__, numero_arbre);
      SORTIE_FONCTION( __FUNCTION__, "%p [initialisation échouée, code %d]", NULL, retour );
      return NULL;
    }
  }

  /* b. On prépare la fenêtre [si barre d'infos, ascenseurs : sera dans la partie utilisateur] */
  fenetre = FEN_NouvelleFenetre( BI_FENETRE_SIMPLE_ARBRE, attributs & ~(INFO|ASC_VERTICAL|ASC_HORIZONTAL), NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la fenêtre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    SORTIE_FONCTION( __FUNCTION__, "%p [fenêtre racine non créée]", NULL );
    return NULL;
  }

  /* c. Création des deux sous-fenêtres */
  fenetre->ss_fenetres = (Fenetre **) calloc( 2, sizeof( Fenetre * ) );
  if ( NULL == fenetre->ss_fenetres ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de réserver les sous-fenêtres\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FEN_Detruire( fenetre );
    SORTIE_FONCTION( __FUNCTION__, "%p [table des sous-fenêtre non réservée]", NULL );
    return NULL;
  }

  /* d.  en 0, la sous-fenêtre contenant l'arbre
         en 1, la sous-fenêtre utilisateur
  */
  fenetre->nbr_ss_fenetres = 2;
  fenetre->ss_fenetres[ 0 ] = FEN_NouvelleFenetre( BI_FENETRE_ARBRE , 0, fenetre );
  fenetre->ss_fenetres[ 1 ] = FEN_NouvelleFenetre( BI_FENETRE_SIMPLE, attributs & (INFO | ASC_VERTICAL | ASC_HORIZONTAL), fenetre );
  if ( ( NULL == fenetre->ss_fenetres[ 0 ] ) || ( NULL == fenetre->ss_fenetres[ 1 ] ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer l'une des deux sous-fenêtres\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FEN_Detruire( fenetre );
    SORTIE_FONCTION( __FUNCTION__, "%p [l'une des sous-fenêtre n'a pas pu être créée]", NULL );
    return NULL;
  }

  /* On met le titre */
  FEN_titre( fenetre, titre );
  if ( ( attributs & INFO ) && ( infos != NULL ) ) { /* Il y a une barre d'infos [dans la partie utilisateur !] */
    FEN_infos( fenetre->ss_fenetres[ 1 ], infos );
  }

  /* e. On assigne l'arbre à la bonne sous-fenêtre */
  adresse_arbre = arbres[ numero_arbre ].adresse;
  fenetre->ss_fenetres[ 0 ]->arbre = arbres + numero_arbre;
  fenetre->ss_fenetres[ 0 ]->numero_arbre = numero_arbre;
  fenetre->ss_fenetres[ 0 ]->position = position_arbre;

  /* f. On crée les objets de l'arbre */
  OBJ_creer_fenetres( fenetre->ss_fenetres[ 0 ]->arbre, fenetre->ss_fenetres[ 0 ]->handle, ROOT, ROOT ); /* On crée tous les objets */

  /* g. On s'occupe des tailles de ces fenêtres, en fonction du mode */
  FEN_AdapterArbre( fenetre, fenetre->ss_fenetres[ 0 ] );
  FEN_AdapterSousFenetre( fenetre, fenetre->ss_fenetres[ 0 ], fenetre->ss_fenetres[ 1 ] );

  /* h. On associe les routines de gestion des événements */
  FEN_AssocierEvenements( fenetre, NULL, NULL, clavier, NULL, message, NULL, NULL );
  FEN_AssocierEvenements( fenetre->ss_fenetres[ 1 ], redessin, NULL, clavier, clic, NULL, rect, temps );

  /* Terminé ! */
  SORTIE_FONCTION( __FUNCTION__, "%p [OK]", fenetre );
  return fenetre;
}

/* -------- Fenêtre mixte, Open GL + partie utilisateur -------- */

Fenetre *FEN_CreerAvecGL(int attributs, const char *titre, int position_gl,
			 void (*redessin)   (Fenetre *, int, int, unsigned int, unsigned int),
			 void (*redessin_gl)(Fenetre *, int, int, unsigned int, unsigned int),
			 void (*ascenseur)  (Fenetre *, int, int, int),
			 int  (*clavier)    (Fenetre *, unsigned int, int),
			 int  (*clic)       (Fenetre *, int, int, int, int, int),
			 int  (*message)    (Fenetre *, Message *message),
			 void (*rect)       (Fenetre *, int, int, int, int),
			 int  (*temps)      (Fenetre *) )
{
  Fenetre *fenetre;

  /* a. On prépare la fenêtre [si ascenseurs : seront dans la partie utilisateur] */
  fenetre = FEN_NouvelleFenetre( BI_FENETRE_GL_SIMPLE, attributs & ~(ASC_VERTICAL|ASC_HORIZONTAL), NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la fenêtre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* b. Création des deux sous-fenêtres */
  fenetre->ss_fenetres = (Fenetre **) calloc( 2, sizeof( Fenetre * ) );
  if ( NULL == fenetre->ss_fenetres ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de réserver les sous-fenêtres\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FEN_Detruire( fenetre );
    return NULL;
  }

  /* c.  en 0, la sous-fenêtre Open GL
         en 1, la sous-fenêtre utilisateur
  */
  fenetre->nbr_ss_fenetres = 2;
  fenetre->ss_fenetres[ 0 ] = FEN_NouvelleFenetre( BI_FENETRE_GL , 0, fenetre );
  fenetre->ss_fenetres[ 1 ] = FEN_NouvelleFenetre( BI_FENETRE_SIMPLE, attributs & (ASC_VERTICAL | ASC_HORIZONTAL), fenetre );
  if ( ( NULL == fenetre->ss_fenetres[ 0 ] ) || ( NULL == fenetre->ss_fenetres[ 1 ] ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer l'une des deux sous-fenêtres\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FEN_Detruire( fenetre );
    return NULL;
  }

  /* d. On met le titre */
  FEN_titre( fenetre, titre );

  fenetre->ss_fenetres[ 0 ]->position = position_gl;

  /* e. On s'occupe des tailles de ces fenêtres, en fonction du mode */
  FEN_AdapterOpenGL     ( fenetre, fenetre->ss_fenetres[ 0 ], 500 );
  FEN_AdapterSousFenetre( fenetre, fenetre->ss_fenetres[ 0 ], fenetre->ss_fenetres[ 1 ] );

  printf( "%s : contenant = %ux%u, GL = X=%u Y=%u %ux%u, Util = X=%u Y=%u %ux%u\n",
	  __FUNCTION__, fenetre->w_larg, fenetre->w_haut, 
	  fenetre->ss_fenetres[ 0 ]->x, fenetre->ss_fenetres[ 0 ]->y, fenetre->ss_fenetres[ 0 ]->larg, fenetre->ss_fenetres[ 0 ]->haut,
	  fenetre->ss_fenetres[ 1 ]->x, fenetre->ss_fenetres[ 1 ]->y, fenetre->ss_fenetres[ 1 ]->larg, fenetre->ss_fenetres[ 1 ]->haut );

  /* f. On associe les routines de gestion des événements */
  FEN_AssocierEvenements( fenetre, NULL, NULL, clavier, NULL, message, NULL, NULL );
  FEN_AssocierEvenements( fenetre->ss_fenetres[ 0 ], redessin_gl, NULL, clavier, NULL, NULL, NULL, NULL );
  FEN_AssocierEvenements( fenetre->ss_fenetres[ 1 ], redessin, ascenseur, clavier, clic, NULL, rect, temps );

  /* Terminé ! */
  return fenetre;
}

/* -------- Fenêtre triple : Open GL, arbre & région utilisateur -------- */

Fenetre *FEN_CreerFenetreTriple(int attributs, const char *titre,
				unsigned int numero_arbre, int position_arbre, int (*init_arbre)(unsigned int),
				int position_gl,
				void (*redessin)   (Fenetre *, int, int, unsigned int, unsigned int),
				void (*redessin_gl)(Fenetre *, int, int, unsigned int, unsigned int),
				void (*ascenseur)  (Fenetre *, int, int, int),
				int  (*clavier)    (Fenetre *, unsigned int, int),
				int  (*clic)       (Fenetre *, int, int, int, int, int),
				int  (*message)    (Fenetre *, Message *message),
				void (*rect)       (Fenetre *, int, int, int, int),
				int  (*temps)      (Fenetre *) )
{
  Fenetre *fenetre;
  OBJECT *adresse_arbre __attribute__((__unused__));

  /* On contrôle si l'arbre a déjà une fenêtre associée */
  fenetre = ARB_QuelleFenetre( arbres + numero_arbre, VRAI );
  if ( fenetre != NULL ) {
    /* La fenêtre existe déjà ! */
    if ( FAUX == fenetre->drapeaux.ouverte ) {
      /* Elle est fermée : on l'ouvre de nouveau */
      FEN_Reouvrir( fenetre );
    }

    /* On la fait passer au premier plan */
    FEN_premierplan(fenetre);

    /* On renvoie cette fenêtre */
    return fenetre;
  }

  /* --- Si on est là : on doit créer la fenêtre --- */

  /* a. On prépare l'arbre, si nécessaire */
  if ( FAUX == arbres[ numero_arbre ].drapeaux.init ) {
    int retour;

    if ( init_arbre != ( int (*)(unsigned int) ) NULL ) {
      retour = init_arbre( numero_arbre );
    } else {
      AVERTISSEMENT( "Arbre %u sans procédure d'initialisation !", numero_arbre );
      retour = ARB_Initialiser( numero_arbre, NULL,  ( int (*)(unsigned int, int, int) ) NULL, ( int (*)(unsigned int, int, int *) ) NULL );
    }

    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Initialisation de l'arbre a échoué [arbre %u]\n",
		     __FILE__, __FUNCTION__, __LINE__, numero_arbre);
      return NULL;
    }
  }

  /* b. On prépare la fenêtre */
  fenetre = FEN_NouvelleFenetre( BI_FENETRE_GL_ARBRE_SIMPLE, attributs, NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer la fenêtre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* c. Création des trois sous-fenêtres */
  fenetre->ss_fenetres = (Fenetre **) calloc( 3, sizeof( Fenetre * ) );
  if ( NULL == fenetre->ss_fenetres ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de réserver les sous-fenêtres\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FEN_Detruire( fenetre );
    return NULL;
  }

  /* d.  en 0, la sous-fenêtre contenant l'arbre
         en 1, la sous-fenêtre Open GL
	 en 2, la sous-fenêtre utilisateur
  */
  fenetre->nbr_ss_fenetres = 3;
  fenetre->ss_fenetres[ 0 ] = FEN_NouvelleFenetre( BI_FENETRE_ARBRE , 0, fenetre );
  fenetre->ss_fenetres[ 1 ] = FEN_NouvelleFenetre( BI_FENETRE_GL    , 0, fenetre );
  if ( NULL == fenetre->ss_fenetres[ 1 ] ) {
    /* Peut arriver si OpenGL compilable mais pas installé, ou manque une extension
       Cas sur la machine de Samba : extension GLX non installée ! [le 4 juillet 2010] */
    (void) BI_ERR_alerte( NULL, &ERR_ImpossibleOuvrirFenetreOpenGL, VRAI );

    /* On passe en mode sans Open_GL */
    BI_OpenGL_actif = FAUX;

    /* Et on crée une sous-fenêtre fictive */
    fenetre->ss_fenetres[ 1 ] = FEN_NouvelleFenetre( BI_FENETRE_SIMPLE    , 0, fenetre );    
  }
  fenetre->ss_fenetres[ 2 ] = FEN_NouvelleFenetre( BI_FENETRE_SIMPLE, 0, fenetre );
  if ( ( NULL == fenetre->ss_fenetres[ 0 ] ) || ( NULL == fenetre->ss_fenetres[ 1 ] ) || ( NULL == fenetre->ss_fenetres[ 2 ] ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de créer l'une des trois sous-fenêtres\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FEN_Detruire( fenetre );
    return NULL;
  }

  /* On met le titre */
  FEN_titre( fenetre, titre );

  /* e. On assigne l'arbre à la bonne sous-fenêtre */
  adresse_arbre = arbres[ numero_arbre ].adresse;
  fenetre->ss_fenetres[ 0 ]->arbre = arbres + numero_arbre;
  fenetre->ss_fenetres[ 0 ]->numero_arbre = numero_arbre;

  /* Position relative des fenêtres */
  fenetre->ss_fenetres[ 0 ]->position = position_arbre;
  fenetre->ss_fenetres[ 1 ]->position = position_gl;
  switch ( position_arbre )
  {
  case BI_FEN_ARBRE_COIN_HG:
    if ( BI_FEN_ARBRE_DROITE == position_gl ) {
      fenetre->ss_fenetres[ 2 ]->position = BI_FEN_ARBRE_COIN_BG;
    } else if ( BI_FEN_ARBRE_COIN_HD == position_gl ) {
      fenetre->ss_fenetres[ 2 ]->position = BI_FEN_ARBRE_BAS;
    } else if ( BI_FEN_ARBRE_COIN_BG == position_gl ) {
      fenetre->ss_fenetres[ 2 ]->position = BI_FEN_ARBRE_DROITE;
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : Combinaison inconnue, arbre=%d, GL=%d ==> utilisateur = ??\n",
		     __FILE__, __FUNCTION__, __LINE__, position_arbre, position_gl);
    }
    break;

  default:
    OPTION_A_FAIRE("Arbre en position [...]" );
  }

  /* f. On crée les objets de l'arbre */
  OBJ_creer_fenetres( fenetre->ss_fenetres[ 0 ]->arbre, fenetre->ss_fenetres[ 0 ]->handle, ROOT, ROOT ); /* On crée tous les objets */

  /* g. On s'occupe des tailles de ces fenêtres, en fonction du mode */
  FEN_AdapterArbre( fenetre, fenetre->ss_fenetres[ 0 ] );
  FEN_AdapterSousFenetre( fenetre, fenetre->ss_fenetres[ 0 ], fenetre->ss_fenetres[ 1 ] );
  FEN_AdapterSousFenetre( fenetre, fenetre->ss_fenetres[ 0 ], fenetre->ss_fenetres[ 2 ] );

  /* h. On associe les routines de gestion des événements */
  FEN_AssocierEvenements( fenetre, NULL, ascenseur, clavier, NULL, message, NULL, NULL );
  FEN_AssocierEvenements( fenetre->ss_fenetres[ 1 ], redessin_gl, NULL, clavier, NULL, NULL, NULL, NULL );
  FEN_AssocierEvenements( fenetre->ss_fenetres[ 2 ], redessin, NULL, clavier, clic, NULL, rect, temps );

  /* Terminé ! */
  return fenetre;
}

/* -------- Fenêtre contenant une série de sous-fenêtres --------
 *  Renvoie la fenêtre globale contenant cela
 *
 * titre        = le titre de cette fenêtre globale [les sous-fenêtres n'ont pas de titre]
 * nbr_fenetres = le nombre de sous-fenêtres
 * attributs    = attributs de chacune de ces sous-fenêtres
 * types        = types de chacune de ces sous-fenêtres
 */

Fenetre *FEN_CreerFenetreMultiple(const char *titre, unsigned int nbr_fenetres, int attributs[], int types[])
{
  Fenetre *fenetre, **ss_fenetres;
  unsigned int i;
  int retour = 0;

  if ( 0 == nbr_fenetres ) {
    (void) fprintf(stderr, "FEN_CreerFenetreMultiple : Aucune sous-fenêtre !\n");
    return NULL;
  }

  fenetre = FEN_NouvelleFenetre( BI_FENETRE_MULTIPLE, NAME|CLOSER|FULLER|MOVER|SIZER|HOTCLOSE|SMALLER, NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "FEN_CreerFenetreMultiple : ERREUR dans FEN_NouvelleFenetre\n");
    return NULL;
  }

  /* On adapte au type souhaité */
  /* On prépare pour les sous-fenêtres */
  ss_fenetres = (Fenetre **) calloc( nbr_fenetres, sizeof( Fenetre * ) );
  if ( NULL == ss_fenetres ) {
    (void) fprintf(stderr, "FEN_CreerFenetreMultiple : Impossible de réserver les sous-fenêtres\n");
    FEN_Detruire( fenetre );
    return NULL;
  }

  /* On crée les sous-fenêtres */
  for ( i = 0 ; i < nbr_fenetres ; i++ ) {
    ss_fenetres[ i ] = FEN_NouvelleFenetre( types[ i ], attributs[ i ], fenetre );
    if ( NULL == ss_fenetres[ i ] ) {
      (void) fprintf(stderr, "FEN_CreerFenetreMultiple : ATTENTION, impossible de créer la %u-ème sous-fenêtre\n", i + 1 );
    }
#ifdef _DEVELOPPE_
    FEN_titre( ss_fenetres[ i ], "Sous-fenêtre...");
#endif
  }

  /* On asocie les sous-fenêtres à la fenêtre */
  fenetre->nbr_ss_fenetres = nbr_fenetres;
  fenetre->ss_fenetres     = ss_fenetres;

  /* On répartie correctement les sous-fenêtres dans la fenêtre */
  retour = FEN_OptimiserDimensionsInternes( fenetre );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "FEN_CreerFenetreMultiple : ATTENTION, l'initialisation des dimensions internes a échoué\n");
  }

  /* On met le titre */
  FEN_titre( fenetre, titre );

  /* Tout s'est bien passé */
  return fenetre;
}

/* ~-~-~-~-~-~-~-~- Aides au développement ~-~-~-~-~-~-~-~- */

#ifdef _DEVELOPPE_
/* -------- Vérifie que la fenêtre indiquée appartient bien au chaînage - revoie FAUX sinon */

boolean BI_VerifierChainageFenetre(Fenetre *fenetre)
{
  Fenetre *fen;
  fen = BI_premiere_fenetre;

  while (fen != fenetre) {
    if (fen == NULL) {
      /* On affiche tout le contenu. Raison : la fenêtre n'existe plus -> son utilisation
	 risque de faire une erreur X difficile à localiser. Avec un peu de chance, la mémoire de
	 la structure associée est déjà libérée -> en lisant, on peut avoir un seg. fault. plus
	 facile à utiliser car visible par le débogueur... */

      fprintf(stderr, "\n **************** ERREUR GRAVE ********************************\n");
      fprintf(stderr, " *                                                            *\n");
      fprintf(stderr, " * Fenêtre %p non trouvée dans le chaînage !\n", (void *) fenetre);
      BI_AfficherFenetre( fenetre );
      fprintf(stderr, " *                                                            *\n");
      fprintf(stderr, " **************************************************************\n");

      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return FAUX;
    }
    fen = fen->suivante;
  }

  return VRAI;
}
#endif
