/* LASE - Logiciel d'analyse des spectres EXAFS
   (C) Emmanuel CURIS, 1998 - Création du fichier le 14 juin 1998

   Sauvegarde au format postscript
   L'unité de base est le 1/100 de mm = taille d'un point donc 
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "bin_utils.h" /* Pour BIN_InclureFichier() */

#include "exafs.h"

#include "fichiers.h"		/* Pour export_ps() */
#include "ps.h"			/* Inclus figures.h, donc tracer.h */
#include "serie.h"

#include "export_ps.txt"

/* ***** Les fonctions locales *************************************************************** */

static void PS_unique (FILE *fichier, struct infos_ps *infos, Graphe *graphe);

static void PS_InitRectangle(struct infos_ps *infos, struct rect_ps *rectangle);

static double PS_haut_texte(FILE *fichier, const char *nom_fonte, double haut); /* En points = 1/72 p */
static void   PS_sort_chaine(FILE *fichier, const char *chaine);

static void prepare_arbre_PS(void);

extern InfosRegion2D *P_QuelleRegion(void); /* Dans Interface/principale.c, juste pour garder zone_dessin en interne à ce dernier fichier ! */

/* ***** Les variables locales *************************************************************** */

#ifndef REPERTOIRE_ENTETES
# define REPERTOIRE_ENTETES "/usr/local/share/Lase"
#endif

static char *chemin_export_PS = NULL; /* Dossier de sauvegarde des PS */
static FILE *fichier;

#define NOMBRE_DE_COULEURS 16
#define NOMBRE_DE_MARQUES   7

static const char *marqueurs[ NOMBRE_DE_MARQUES ] = /* Noms des procédures de dessin des marqueurs */
{
  "Point", "Plus", "Asterisque", "Carre", "Croix", "Losange", "Cercle"
};

#define PS_marqueur(j) ( (j >= 0) && (j < 7) ) ? marqueurs[j] : "Point"; /* Point par défaut */

static const char *couleurs[ NOMBRE_DE_COULEURS ] = /* Noms des procédures de choix d'une couleur
						       16 couleurs prédéfinies ~ VDI */
{
  "Blanc", "Noir", "Rouge", "Vert", "Bleu", "Cyan", "Jaune", "Magenta",
  "Gris", "Gris2", "Rouge2", "Vert2", "Bleu2", "Cyan2", "Jaune2", "Magenta2"
};

#define PS_couleur(j) if ( (j >= 0) && (j < 16) ) \
                        fprintf(fichier, "%s\n", couleurs[j]); \
                      else \
			fprintf(fichier, "Noir\n"); /* Noir couleur par défaut */

/* ~-~-~-~-~-~-~-~-~ Interfaces d'export pour divers objets LASE ~-~-~-~-~-~-~-~-~ */

/* -------- Graphe --------------------------------\
 | Renvoie 0 si tout va bien, != 0 en cas d'erreur |
 |                                                 |
 | graphe : le graphe à exporter (NULL == courant) |
 \------------------------------------------------ */

int EXP_PS_GrapheUnique(struct graphe *graphe)
{
  char *nom_fichier;
  const char *masque = "*.ps";
  int retour;
  struct infos_ps infos;
  FILE *fichier;
  BI_InfosSelecteur infos_selecteur = {
    NULL,
    BI_SELECTEUR_FILTRE_UTILISATEUR | BI_SELECTEUR_AJOUTER_PREDEFINI | BI_SELECTEUR_AVEC_SAUVEGARDE | BI_SELECTEUR_MEMORISER_CHEMIN,
    NULL,
    1, NULL,
    0, NULL,
    (void (*)(unsigned int, unsigned int, const char *) ) NULL
  };

  if ( NULL == graphe ) {
    if ( NULL == graphe_courant ) {
      (void) fprintf(stderr, "Patrick, tu triches, il n'est pas encore l'heure d'aller fumer\n");
      return -1;
    }
    graphe = graphe_courant;
  }

  /* On appelle la boîte de dialogue */
  retour = PS_Parametrage( &infos );
  if ( retour != 0 ) {
    if ( retour < 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans PS_Parametrage() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }
    return +1; /* On annule, en général */
  }

  /* Préparation du sélecteur de fichiers */
  infos_selecteur.titre  = (char *) txt_titre_selecteur;
  infos_selecteur.filtres = &masque;
  infos_selecteur.chemin = chemin_export_PS;

  nom_fichier = BI_SelecteurSauvegarde( &infos_selecteur, NULL, principale );
  /* On mémorise les chemins */
  chemin_export_PS = infos_selecteur.chemin;
 
  if ( nom_fichier == NULL ) { /* Erreur à la sauvegarde */
    return +2;
  }

  /* Ouverture du fichier */
  fichier = PS_OuvrirFichier( nom_fichier, graphe->nom , &infos );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans PS_OuvrirFichier() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    free( nom_fichier ); nom_fichier = NULL; /* Plus besoin du nom de fichier */
    return -10;
  }
  free( nom_fichier ); nom_fichier = NULL; /* Plus besoin du nom de fichier */

  /* Export du graphe */
  PS_unique( fichier, &infos, graphe );

  /* Log dans l'historique */
  HIST_Titre( 1, "Figure PostScript", "Export" );
  HIST_Texte( 1, 0, "Graphe export"e_AIGU" : %s\n", graphe->nom );

  /* Terminé */
  (void) fclose( fichier );
  return 0;
}

/* -------- Figure --------------------------------\
 | Renvoie 0 si tout va bien, != 0 en cas d'erreur |
 \------------------------------------------------ */

int EXP_PS_Figure(Figure *figure)
{
  char *nom_fichier;
  const char *masque = "*.ps";
  int retour;
  struct infos_ps infos;
  Bloc_Figure *bloc;
  BI_InfosSelecteur infos_selecteur = {
    NULL,
    BI_SELECTEUR_FILTRE_UTILISATEUR | BI_SELECTEUR_AJOUTER_PREDEFINI | BI_SELECTEUR_AVEC_SAUVEGARDE | BI_SELECTEUR_MEMORISER_CHEMIN,
    NULL,
    1, NULL,
    0, NULL,
    (void (*)(unsigned int, unsigned int, const char *) ) NULL
  };

  /* La figure existe-t-elle ? */
  if ( NULL == figure ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas de figure actuellement !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* On initialise taille de papier [=figure] & marges [=0] */
  if ( OBJ_choisit(  EXPORT_PS, EPS_TOURNEE, QUESTION ) == VRAI ) {
    OBJ_MettreNombre( EXPORT_PS, EPS_LARGEUR, figure->hauteur, 0.0 );
    OBJ_MettreNombre( EXPORT_PS, EPS_HAUTEUR, figure->largeur, 0.0 );
  } else {
    OBJ_MettreNombre( EXPORT_PS, EPS_LARGEUR, figure->largeur, 0.0 );
    OBJ_MettreNombre( EXPORT_PS, EPS_HAUTEUR, figure->hauteur, 0.0 );
  }

  OBJ_MettreNombre( EXPORT_PS, EPS_MARGE_HAUTE , 0.0, 0.0 );
  OBJ_MettreNombre( EXPORT_PS, EPS_MARGE_BASSE , 0.0, 0.0 );
  OBJ_MettreNombre( EXPORT_PS, EPS_MARGE_GAUCHE, 0.0, 0.0 );
  OBJ_MettreNombre( EXPORT_PS, EPS_MARGE_DROITE, 0.0, 0.0 );

  /* Appel de la boîte de dialogue */
  retour = PS_Parametrage( &infos );
  if ( retour != 0 ) {
    if ( retour < 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans PS_Parametrage() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }
    return retour;
  }

  /* Préparation du sélecteur de fichiers */
  infos_selecteur.titre  = (char *) txt_titre_selecteur;
  infos_selecteur.filtres = &masque;
  infos_selecteur.chemin = chemin_export_PS;

  nom_fichier = BI_SelecteurSauvegarde( &infos_selecteur, NULL, principale );
  /* On mémorise les chemins */
  chemin_export_PS = infos_selecteur.chemin;

  if (nom_fichier == NULL) /* Erreur à la sauvegarde */
    return -1;

  /* Ouverture du fichier */
  fichier = PS_OuvrirFichier( nom_fichier, "Figure utilisateur", &infos );
  if ( NULL == fichier ) {
    (void) SYS_ERREUR(ERREUR_ECRITURE, nom_fichier, NULL, NULL);
    free( nom_fichier ); nom_fichier = NULL;
    return - 2;
  }
  free( nom_fichier ); nom_fichier = NULL; /* Plus besoin du nom de fichier */

  /* ---- Pour déboguage : on trace le pourtour de la figure ---- */
#ifdef _DEVELOPPE_
  (void) fprintf( fichier, "%% Contour de la figure, rouge" );
  (void) PS_Rectangle( fichier, FAUX, 1.0, ROUGE, -1, infos.marges.gauche, infos.marges.bas, figure->largeur * 100.0, figure->hauteur * 100.0 );
#endif

  /* ---- On va maintenant dessiner chaque bloc, un par un ---- */
  for ( bloc = figure->premier_bloc ; bloc != NULL ; bloc = bloc->suivant ) {
    (void) fprintf(fichier, "%% Nouveau cadre dessiné\n");
    retour = PS_BlocFigure( fichier, &infos, figure, bloc );
    if ( retour < 0 ) break; /* Erreur grave */
  }
  
  (void) fprintf(fichier, "showpage\n");

  /* Log dans l'historique */
  HIST_Titre( 1, "Figure PostScript", "Export" );

  /* Terminé */
  (void) fclose( fichier );
  return 0;
}

/* -------- Export de la superposition automatique --- */

int EXP_PS_Superposition(struct infos_superpose *infos_superposition)
{
  char *nom_fichier;
  const char *masque = "*.ps";
  int retour;
  struct infos_ps infos;
  struct rect_ps rectangle; /* Qui contient la figure */
  Graphe *graphe;
  FILE *fichier;
  BI_InfosSelecteur infos_selecteur = {
    NULL,
    BI_SELECTEUR_FILTRE_UTILISATEUR | BI_SELECTEUR_AJOUTER_PREDEFINI | BI_SELECTEUR_AVEC_SAUVEGARDE | BI_SELECTEUR_MEMORISER_CHEMIN,
    NULL,
    1, NULL,
    0, NULL,
    (void (*)(unsigned int, unsigned int, const char *) ) NULL
  };

  if ( NULL == infos_superposition ) {
    (void) fprintf(stderr, "Patrick, tu triches, il n'est pas encore l'heure d'aller fumer\n");
    return -1;
  }
  if ( NULL == serie_courante ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucune série en mémoire !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* On appelle la boîte de dialogue */
  retour = PS_Parametrage( &infos );
  if ( retour != 0 ) {
    if ( retour < 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans PS_Parametrage() [code %d]\n",
		     __FILE__, __FUNCTION__, __LINE__, retour);
    }
    return +1; /* On annule, en général */
  }

  /* Préparation du sélecteur de fichiers */
  infos_selecteur.titre  = (char *) txt_titre_selecteur;
  infos_selecteur.filtres = &masque;
  infos_selecteur.chemin = chemin_export_PS;

  nom_fichier = BI_SelecteurSauvegarde( &infos_selecteur, NULL, principale );
  /* On mémorise les chemins */
  chemin_export_PS = infos_selecteur.chemin;

  if ( nom_fichier == NULL ) { /* Erreur à la sauvegarde */
    return +2;
  }

  /* Ouverture du fichier */
  fichier = PS_OuvrirFichier( nom_fichier, "Superposition automatique" , &infos );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans PS_OuvrirFichier() [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    free( nom_fichier ); nom_fichier = NULL; /* Plus besoin du nom de fichier */
    return -10;
  }
  free( nom_fichier ); nom_fichier = NULL; /* Plus besoin du nom de fichier */

  /* Préparation du rectangle */
  PS_InitRectangle( &infos, &rectangle );
  rectangle.cadre = infos_superposition->dessin;
  rectangle.cadre.axe_X.proprietes &= (~AXE_ECRAN); /* L'axe ne sera pas fait à l'écran ! */
  rectangle.cadre.axe_Y.proprietes &= (~AXE_ECRAN); /* L'axe ne sera pas fait à l'écran ! */

  rectangle.cadre.axe_X.debord_gauche =   0.0;
  rectangle.cadre.axe_X.debord_droite = 600.0;
  rectangle.cadre.axe_Y.debord_gauche =   0.0;
  rectangle.cadre.axe_Y.debord_droite = 600.0;

  /* Cadre autour de la superposition */
  PS_Cadre( fichier, &infos, &rectangle );

  /* Dessin des graphes retenus */
  for ( graphe = serie_courante->premier_graphe ; graphe != NULL ; graphe = graphe->suivant ) {
    if ( VRAI == TRC_GrapheConvient( infos_superposition, graphe ) ) {
      int erreurs = 0, retour __attribute__((__unused__));

      if ( ( VRAI == graphe->drapeaux.erreur_x ) && ( graphe->erreur_x != NULL ) ) {
	erreurs |= AXE_X;
      }
      if ( ( VRAI == graphe->drapeaux.erreur_y ) && ( graphe->erreur_y != NULL ) ) {
	erreurs |= AXE_Y;
      }
      retour = PS_Graphe( fichier, &infos, &rectangle, graphe, erreurs, -1, -1, -1 );
    }
  }

  /* Log dans l'historique */
  HIST_Titre( 1, "Figure PostScript", "Export" );

  /* Terminé */
  (void) fprintf( fichier, "showpage\n" ); /* On a fini avec ce dessin... Sinon, imprime pas ! */
  (void) fclose( fichier );
  return 0;
}

/* Préparer un rectangle contenant, en fonction des marges */

static void PS_InitRectangle(struct infos_ps *infos, struct rect_ps *rectangle)
{
  /* Dimensions en fonction des marges */
  rectangle->x_BG = infos->larg_dessin * 5.0; /* 95% pour le graphe, 5% pour la légende en Y */
  if ( rectangle->x_BG > 2000 ) {
    rectangle->x_BG = 2000; /* Mais pas plus de 2 cm de bord à gauche */
  }
  rectangle->y_BG = infos->haut_dessin * 2.5; /* 97.5% pour le graphe, 2.5% pour la légende en X */
  if ( rectangle->y_BG > 1000 ) {
    rectangle->y_BG = 1000; /* Mais pas plus de 1 cm de bord en bas */
  }
  rectangle->larg = infos->larg_dessin * 100.0 - rectangle->x_BG;
  rectangle->haut = infos->haut_dessin * 100.0 - rectangle->y_BG;

  /* On affiche un cadre, par défaut */
  rectangle->cadre.proprietes = CADRE2D_AVEC_CADRE;
}

/* -------- Ouverture d'un fichier PostScript -------- */

FILE *PS_OuvrirFichier(const char *nom, const char *titre, struct infos_ps *infos)
{
  FILE *fichier;
  int retour = 0;
  const char *nom_fonte;

  fichier = fopen( nom, "w" );
  if ( NULL == fichier ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'ouvrir le fichier %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
    return NULL;
  }

  /* Mention dans l'historique */
  HIST_Titre( 1, "Export d'une figure PostScript", "E/S" );
  HIST_Texte( 1, 0, "Fichier : %s\n", nom );

  /* ---- On s'occupe de l'en-tête ---- */
  (void) fprintf( fichier, "%%!PS-Adobe EPSF-3.0\n" );
  if ( NULL == titre ) {
    (void) fprintf( fichier, "%%%%Title: (non-indiqué)\n" );
  } else {
    (void) fprintf( fichier, "%%%%Title: %s\n", titre );
  }
  (void) fprintf(fichier, "%%%%Creator: LASE v. %.2f - (C) 1998-2003, Emmanuel CURIS\n", VERSION / 10.0);
  (void) fprintf(fichier, "%%%%Pages: (atend)\n");
  (void) fprintf(fichier, "%%%%DocumentFonts: Helvetica\n");
  if ( infos->proprietes & PS_PAGE_TOURNEE ) { /* On demande de tourner la page */
    (void) fprintf(fichier, "%%%%Orientation: Landscape\n");
    (void) fprintf(fichier, "%%%%BoundingBox: 0 0 %f %f\n",
		   infos->haut_page * 100.0 * 2.8346457e-2, infos->larg_page * 100.0 * 2.8346457e-2 );
  } else {
    (void) fprintf(fichier, "%%%%Orientation: Portrait\n");
    (void) fprintf(fichier, "%%%%BoundingBox: 0 0 %f %f\n",
		   infos->larg_page * 100.0 * 2.8346457e-2, infos->haut_page * 100.0 * 2.8346457e-2 );
  }
  (void) fprintf(fichier, "%%%%EndComments\n%%%%BeginProlog\n");

  /* Comment sont codés les accents & caractères spéciaux ? */
#ifdef _TOS_
  retour = BIN_InclureFichier( fichier, REPERTOIRE_ENTETES"/codage_tos.ps" );
#endif
#if defined(_UNIX_) || defined(_WINDOWS_) /* Les deux sont en latin-1, plus ou moins... */
  if ( infos->proprietes & PS_NIVEAU_2 ) {
    (void) fprintf( fichier, "/Codage ISOLatin1Encoding def\n" );
    retour = 0;
  } else {
    retour = BIN_InclureFichier( fichier, REPERTOIRE_ENTETES"/codage_latin1.ps" );
  }
#endif
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans la sauvegarde du codage des caractères spéciaux\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) fclose( fichier );
    return NULL;
  }

  /* Les commandes courantes (couleurs, marques,...) */
  retour = BIN_InclureFichier( fichier, REPERTOIRE_ENTETES"/prologue.ps" );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans la sauvegarde du prologue\n",
		   __FILE__, __FUNCTION__, __LINE__);
    (void) fclose( fichier );
    return NULL;
  }

  /* ---- Définition des codages de fontes. Ici, codage pour la fonte Helvetica ---- */
  if ( NULL == infos->nom_fonte )
    nom_fonte = "Helvetica";
  else
    nom_fonte = infos->nom_fonte;

  (void) fprintf( fichier, "/%s findfont\n", nom_fonte ); /* On récupère la fonte */
  (void) fprintf(fichier, "dup length dict begin {1 index /FID ne {def} {pop pop} ifelse} forall /Encoding Codage def currentdict end /%s exch definefont\n", nom_fonte );
  /* Codage par défaut - inspiré (fortement) du code PS sauvé par Netscape */
  (void) PS_haut_texte( fichier, nom_fonte, 12.0 ); /* Taille 12 par défaut */

  /* On passe en 1/100 mm, pour rester dans 1 int */
  (void) fprintf( fichier, "2.8346457e-2 2.8346457e-2 scale\n" );

  /* --- On positionne l'origine, en tenant compte des marges ---
     ===> (0, 0) correspond au coin BG du rectangle utilisable */
  if ( infos->proprietes & PS_PAGE_TOURNEE ) { /* On demande de tourner la page */
    (void) fprintf( fichier, "90 rotate\n" ); /* Axe des x sur le grand côté */
    (void) fprintf( fichier, "0 %f translate\n", - infos->larg_page * 100.0 ); /* Origine : coin inférieur droit */
    (void) fprintf( fichier, "%f %f translate\n", infos->marges.haut * 100.0, infos->marges.gauche * 100.0 );
  } else {
    (void) fprintf( fichier, "%f %f translate\n", infos->marges.gauche * 100.0, infos->marges.haut * 100.0);
  }

#ifdef _DEVELOPPE_
  /* Pour contrôles éventuels : on trace un rectangle rouge autour de la région accessible */
  (void) fprintf( fichier, "%%Zone de dessin sur la feuille [pour contrôles]\n" );
  (void) fprintf( fichier, "%%Rouge 0.0 0.0 moveto 0.0 %f lineto %f %f lineto %f 0 lineto closepath stroke\n",
		  infos->haut_dessin * 100.0, infos->larg_dessin * 100.0, infos->haut_dessin * 100.0, infos->larg_dessin * 100.0 );
#endif

  /* Épaisseur de ligne */
  PS_EpaisLigne( fichier, infos->epaisseur );

  /* Hauteur de marques */
  PS_TailleMarque( fichier, infos->haut_marque );

  /* Ouverture réussie */
  return fichier;
}

/* ------ Définir la couleur ------ */

inline int PS_Couleur(FILE *fichier, unsigned int couleur)
{
  if ( couleur < NOMBRE_DE_COULEURS ) {
    (void) fprintf( fichier, "%s\n", couleurs[ couleur ] );
  } else { /* Noir par défaut */
    (void) fprintf( fichier, "Noir\n" );
  }

  return 0;
}

/* ------ Définir l'épaisseur de tracé des lignes ------ */

inline int PS_EpaisLigne(FILE *fichier, double epaisseur)
{
  /* Épaisseur donnée en mm, mais on travaille en 1/100 mm en interne dans le fichier */
  (void)  fprintf(fichier, "%f setlinewidth\n", epaisseur * 100.0 );
  return 0;
}

/* ------ Définir le style de marque ------ */

inline const char *PS_Marque(unsigned int marque)
{
  /* On définit la bonne marque */
  if ( marque < NOMBRE_DE_MARQUES ) {
    return marqueurs[ marque ];
  }

  return "Cercle";
}

/* -------- Définie la taille des marques */

inline int PS_TailleMarque(FILE *fichier, double haut_marque)
{
  (void) fprintf( fichier, "/haut_marque %g def\n", haut_marque * 100.0 );
  return 0;
}

/* -------- Définir le style de ligne -------- */

inline int PS_StyleLigne(FILE *fichier, unsigned int style)
{
  switch ( style )
  {
  case LDASHED:
    (void) fprintf( fichier, "[200 200] 0 setdash\n" );
    break;

  case DOTTED:
    (void) fprintf( fichier, "[100 300] 0 setdash\n" );
    break;

  case DASHDOT:
    (void) fprintf( fichier, "[400 150 100 150] 0 setdash\n" );
    break;

  case DASH:
    (void) fprintf( fichier, "[400 400] 0 setdash\n" );
    break;

  case DASHDOTDOT:
    (void) fprintf( fichier, "[250 150 100 100 50 150] 0 setdash\n" );
    break;

  default:
    (void) fprintf( fichier, "[] 0 setdash\n" ); /* Pas de pointillé */
  }
  return 0;
}

/* ~-~-~-~-~-~-~-~-~ Primitives de dessin ~-~-~-~-~-~-~-~-~ */

/* -------- Tracer un rectangle -------- */

int PS_Rectangle(FILE *fichier, boolean niveau2, double epaisseur_bord, int couleur_bord, int couleur_fond,
		 double x, double y, double largeur, double hauteur)
{
  if ( NULL == fichier ) return -1;

  /* a. Tracé du fond, en à-plat */
  if ( couleur_fond >= 0 ) {
    PS_Couleur( fichier, couleur_fond );

    if ( niveau2 != FAUX ) { /* Pas == VRAI, car avec masque de bit, peut être autre chose que 1 */
    } else {
      (void) fprintf( fichier, "%g %g moveto %g 0 rlineto 0 %g rlineto %g 0 rlineto closepath fill\n",
		      x, y, largeur, hauteur, -largeur );
    }
  }

  /* b. Tracé du bord */
  if ( couleur_bord >= 0 ) {
    PS_EpaisLigne( fichier, epaisseur_bord );
    PS_Couleur( fichier, couleur_bord );
    if ( VRAI == niveau2 ) {
    } else {
      (void) fprintf( fichier, "%g %g moveto %g 0 rlineto 0 %g rlineto %g 0 rlineto closepath stroke\n",
		      x, y, largeur, hauteur, -largeur );
    }
  }

  /* Tracé réussi */
  return 0;
}

/* -------- Tracer un texte -------- */

int PS_Texte(FILE *fichier, double x, double y, const char *chaine, int align_hor, int align_ver, double angle, int couleur)
{
  /* Contrôles */
  if ( NULL == fichier ) return -1;

  /* On passe l'angle dans 0-360° : angle indiqué en degrés */
  while (angle < 0.0)
    angle += 360.0;
  while (angle >= 360.0)
    angle -= 360.0;

  /* On met la couleur correcte */
  (void) PS_Couleur( fichier, couleur );

  /* On se place au point indiqué, tout ce qui suit sera relatif à ce point */
  (void) fprintf( fichier, "%g %g moveto", x, y );

  if (angle != 0.0) { /* Si rotation du texte */
    (void) fprintf( fichier, " gsave %g rotate", angle );
  }

  /* On écrit la chaîne, avec conversions PS si besoin est */
  PS_sort_chaine(fichier, chaine);

  /* Position réelle de départ en X */
  if ( BI_TEXTE_CENTRER_H == align_hor ) { /* On centre */
    (void) fprintf(fichier, " dup stringwidth pop"); /* Pile : Texte Dx */
    /* Dy donné par stringwidth correspond au décalage de ligne de base après écriture : nul pour du texte latin... */
    (void) fprintf(fichier, " -2 div"); /* Pile : -Dx/2 */
  } else if ( BI_TEXTE_FER_A_DROITE == align_hor ) { /* On aligne à droite */
    (void) fprintf(fichier, " dup stringwidth pop -1 mul"); /* Pile : -Dx */
  } else {
    (void) fprintf(fichier, " 0"); /* Pile : 0 */
  }

  /* Position réelle de départ en Y */
  if ( BI_TEXTE_CENTRER_V == align_ver ) { /* Demande à être au milieu */
    (void) fprintf(fichier, " hauteur_chiffre -2 div");
  } else if ( BI_TEXTE_FER_EN_HAUT == align_ver ) {
    (void) fprintf(fichier, " hauteur_texte -1 mul"); /* Pile : -haut */
  } else if ( BI_TEXTE_FER_EN_BAS == align_ver ) {
    (void) fprintf(fichier, " currentfont /FontBBox get 1 get"); /* Pile : bas */
  } else {
    (void) fprintf(fichier, " 0"); /* Pile 0 */
  }

  /* On se place au point de départ correct */
  (void) fprintf(fichier, " rmoveto"); /* Pile : texte */

  /* Affiche la chaîne */
  (void) fprintf(fichier, " show");

  if (angle != 0.0) { /* On remet comme avant : on a fini la rotation */
    (void) fprintf(fichier, " grestore");
  }
  (void) fprintf(fichier, "\n");

  return 0;
}

static void PS_unique(FILE *fichier, struct infos_ps *infos, struct graphe *graphe)
{
  int erreurs = 0, retour;
  struct rect_ps rectangle;
  InfosRegion2D *region;

  /* Informations sur le rectangle qui contient ce graphe unique */
  PS_InitRectangle( infos, &rectangle );

  /* On récupère la région 2D de la fenêtre principale */
  region = P_QuelleRegion();

  /* Et on la met dans le cadre PS, en adaptant les propriétés des axes */
  rectangle.cadre = *region;
  rectangle.cadre.axe_X.proprietes &= ~AXE_ECRAN; /* L'axe ne sera pas fait à l'écran ! */
  rectangle.cadre.axe_X.debord_gauche =   0.0;
  rectangle.cadre.axe_X.debord_droite = 600.0;

  rectangle.cadre.axe_Y.proprietes &= ~AXE_ECRAN; /* L'axe ne sera pas fait à l'écran ! */
  rectangle.cadre.axe_Y.debord_gauche =   0.0;
  rectangle.cadre.axe_Y.debord_droite = 600.0;

  /* Tracé du cadre grâce à ce rectangle */
  retour = PS_Cadre( fichier, infos, &rectangle );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le tracé du cadre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* ----- On trace le graphe associé ------------- */
  if ( ( VRAI == graphe->drapeaux.erreur_x ) && ( graphe->erreur_x != NULL ) ) {
    erreurs |= AXE_X;
  }
  if ( ( VRAI == graphe->drapeaux.erreur_y ) && ( graphe->erreur_y != NULL ) ) {
    erreurs |= AXE_Y;
  }
  retour = PS_Graphe( fichier, infos, &rectangle, graphe, erreurs, -1, -1, -1 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Sortie de PS_Graphe avec une erreur [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  /* ----- On met le titre du graphe, quand même ! */
  PS_Texte( fichier, rectangle.x_BG + rectangle.larg / 2.0, rectangle.y_BG + rectangle.haut + 200.0,
	    graphe->nom, BI_TEXTE_CENTRER_H, 0, 0.0, NOIR );

  (void) fprintf(fichier, "showpage\n"); /* On a fini avec ce dessin... Sinon, imprime pas ! */
}

#ifdef ancien

static void PS_GRD_hor_logarithme(int x_deb, int x_fin, int y,
				  double x_min, double x_max, int precision, int ht_tics)
{
  int p_10, p; /* Puissance de 10 étudiée */
  double echelle;
  int points[4];

  if ( x_min > x_max ) {
    double x = x_min; x_min = x_max; x_max = x;
  }
  if ( (x_min == x_max) || (x_fin == x_deb) ) {
    fprintf(stderr, "PS_GRD_hor_logarithme : bornes égales, impossible de graduer !\n");
    return;
  }
  if ( (x_min <= 0.0) || (x_max <= 0.0) ) {
    fprintf(stderr, "PS_GRD_hor_logarithme : bornes négatives, impossible de prendre le log. !\n");
    return;
  }
  if ( precision <= 0 ) {
    precision = (int) ( (log(x_max) - log(x_min)) / log(10.0) );
  }

  fprintf(fichier, "%% Graduation axe horizontal - logarithmique\n");

  /* Calcul de l'échelle */
  echelle = ( (double) (x_fin - x_deb)) / (log(x_max) - log(x_min)) ;

  /* Traçage des taquets */
  points[1] = y;
  points[3] = y - ht_tics;

  p_10 = (int) (log( x_max ) / log(10.0));
  p = 0;
  do {
    double x, facteur;
    int i;

    facteur = pow(10.0, (double) p_10);

    /* Unités : tiret plus grand */
    if ( (facteur >= x_min) && (facteur <= x_max) ) {
      points[0] = x_deb + (log(facteur) - log(x_min)) * echelle;
      points[2] = points[0];
      points[3] = y - 2*ht_tics;

      fprintf(fichier, "%d %d moveto %d %d lineto\n", points[0], points[1], points[2], points[3]);
    }

    points[3] = y - ht_tics;
    for (i = 2 ; i <= 9 ; i++) {

      x = facteur * (double) i;

      if ( (x >= x_min) && (x <= x_max) ) {
	points[0] = x_deb + (log(x) - log(x_min)) * echelle;
	points[2] = points[0];

	fprintf(fichier, "%d %d moveto %d %d lineto\n", points[0], points[1], points[2], points[3]);
      }
    }
    p++;
    p_10--;

  } while ( p <= precision );
  fprintf(fichier, "stroke\n");
}

/* -------- Graduation logarithmique de l'axe vertical -------- */

static void PS_GRD_ver_logarithme(int y_deb, int y_fin, int x,
				  double y_min, double y_max, int precision, int ht_tics)
{
  int p_10, p; /* Puissance de 10 étudiée */
  double echelle;
  double points[4];

  if ( y_min > y_max ) {
    double y = y_min; y_min = y_max; y_max = y;
  }
  if ( (y_min == y_max) || (y_fin == y_deb) ) {
    fprintf(stderr, "PS_GRD_ver_logarithme : bornes égales, impossible de graduer !\n");
    return;
  }
  if ( (y_min <= 0.0) || (y_max <= 0.0) ) {
    fprintf(stderr, "PS_GRD_ver_logarithme : bornes négatives, impossible de prendre le log. !\n");
    return;
  }
  if ( precision <= 0 ) {
    precision = (int) ( (log(y_max) - log(y_min)) / log(10.0) );
  }

  fprintf(fichier, "%% Graduation axe vertical - logarithmique\n");

  /* Calcul de l'échelle */
  echelle = ( (double) (y_fin - y_deb)) / (log(y_max) - log(y_min)) ;

  /* Traçage des taquets */
  points[0] = x;

  p_10 = (int) (log( y_max ) / log(10.0));
  p = 0;
  do {
    double y, facteur;
    int i;

    facteur = pow(10.0, (double) p_10);

    /* Unités : tiret plus grand */
    if ( (facteur >= y_min) && (facteur <= y_max) ) {
      points[1] = y_deb + (log(facteur) - log(y_min)) * echelle;
      points[3] = points[1];
      points[2] = x + 2*ht_tics;

      fprintf(fichier, "%f %f moveto %f %f lineto\n", points[0], points[1], points[2], points[3]);
    }

    points[2] = x + ht_tics;
    for (i = 2 ; i <= 9 ; i++) {

      y = facteur * (double) i;

      if ( (y >= y_min) && (y <= y_max) ) {
	points[1] = y_deb + (log(y) - log(y_min)) * echelle;
	points[3] = points[1];

	fprintf(fichier, "%f %f moveto %f %f lineto\n", points[0], points[1], points[2], points[3]);
      }
    }
    p++;
    p_10--;

  } while ( p <= precision );
  fprintf(fichier, "stroke\n");
}

static void PS_GRD_txt_ver_log(int y_deb, int y_fin, int x,
			       double y_min, double y_max, int precision, double taille_texte)
{
  double haut_caractere;
  int y_precedent = -2000;
  double echelle;
  int p;
  double p_10;

  if ( (y_deb == y_fin) || (y_max == y_min) ) {
    (void) fprintf(stderr, "PS_GRD_txt_ver_log : bornes égales, graduation impossible !\n");
    return;
  }
  if (y_max < y_min) {
    double s = y_max; y_max = y_min; y_min = s;
  }
  if (y_min < 0.0) {
    (void) fprintf(stderr, "PS_GRD_txt_ver_log : borne négative [%f, %f], graduation impossible !\n", y_min, y_max);
    return;
  }

  fprintf(fichier, "%% Valeurs des graduations retenues (axe des Y)\n");

  /* On définit la taille du texte */
  haut_caractere = PS_haut_texte(fichier, "Helvetica", taille_texte);
  haut_caractere *= 2.0; /* 1 ligne blanche d'écart entre deux textes, au minimum */

  echelle = ((double) (y_fin - y_deb)) / ( log(y_max) - log(y_min) );
  if ( precision <= 0 ) {
    precision = (int) ( (log(y_max) - log(y_min)) / log(10.0) );
  }

  /* Écriture des textes */
  p_10 = (int) (log( y_max ) / log(10.0)) - precision;
  p = 0;
  do {
    double facteur;
    int i;

    facteur = pow(10.0, p_10);

    for (i = 1 ; i <= 9 ; i++) {
      double y;

      y = facteur * (double) i;
      if ( (y >= y_min) && (y <= y_max) ) {
	int y_texte;

	y_texte = y_deb + (log(y) - log(y_min)) * echelle;
	if ( abs(y_texte - y_precedent) >= haut_caractere ) { /* Pas de chevauchement */
	  char valeur[50];

	  (void) TXT_CNV_D_vTexte(y, -1.0, 3, 2, 2, valeur, NULL);
	  PS_gtext(fichier, x - 100, y_texte, valeur, ALI_RIGHT, ALI_HALF, 0.0); /* 1 mm d'écart entre l'axe et la valeur */
	  y_precedent = y_texte;
	}
      }
    }
    p++;
    p_10++;

  } while ( p <= precision );
}

/* -~-~-~-~-~-~-~-~ Tracé d'une grille (fond du dessin) -~-~-~-~-~-~-~-~ */

/* -------- Grille axe horizontal, en logarithme -------- */

static void PS_grille_hor_log(int x_deb, int x_fin, int y, int haut,
			      double x_min, double x_max, int precision, int couleur, int style)
{
  int p_10, p; /* Puissance de 10 étudiée */
  double echelle;
  int points[4];

  if ( (x_min == x_max) || (x_fin == x_deb) ) {
    (void) fprintf(stderr, "PS_grille_hor_log : bornes égales, impossible de graduer !\n");
    return;
  }
  if ( (x_min <= 0.0) || (x_max <= 0.0) ) {
    (void) fprintf(stderr, "PS_grille_hor_log : bornes négatives, impossible de prendre le log. !\n");
    return;
  }
  if ( precision <= 0 ) {
    precision = (int) ( (log(x_max) - log(x_min)) / log(10.0) );
  }

  /* Calcul de l'échelle */
  (void) fprintf(fichier, "%% Grille axe horizontal, logarithmique\n");
  echelle = ( (double) (x_fin - x_deb)) / (log(x_max) - log(x_min)) ;

  points[1] = y;
  points[3] = y + haut;

  p_10 = (int) (log( x_max ) / log(10.0));
  p = 0;
  if (style < 7) { /* On trace avec des lignes */
    PS_couleur(couleur);
    PS_StyleLigne(fichier, style);

    do {
      double x, facteur;
      int i;

      facteur = pow(10.0, (double) p_10);
      for (i = 1 ; i <= 9 ; i++) {
	x = facteur * (double) i;

	if ( (x >= x_min) && (x <= x_max) ) {
	  points[0] = x_deb + (log(x) - log(x_min)) * echelle;
	  points[2] = points[0];
	  
	  (void) fprintf(fichier, "%d %d moveto %d %d lineto stroke\n", points[0], points[1], points[2], points[3]);
	}
      }
      p++;
      p_10--;
    } while ( p <= precision );
  } else {
    fprintf(stderr, "PS_grille_hor_log : tracer en polymark, À FAIRE\n");
  }
}

/* -------- Grille axe vertical, en logarithme -------- */

static void PS_grille_ver_log(int y_deb, int y_fin, int x, int larg,
			      double y_min, double y_max, int precision, int couleur, int style)
{
  int p_10, p; /* Puissance de 10 étudiée */
  double echelle;
  int points[4];

  if ( (y_min == y_max) || (y_fin == y_deb) ) {
    fprintf(stderr, "grille_ver_log : bornes égales, impossible de graduer !\n");
    return;
  }
  if ( (y_min <= 0.0) || (y_max <= 0.0) ) {
    fprintf(stderr, "grille_ver_log : bornes négatives, impossible de prendre le log. !\n");
    return;
  }
  if ( precision <= 0 ) {
    precision = (int) ( (log(y_max) - log(y_min)) / log(10.0) );
  }

  /* Calcul de l'échelle */
  (void) fprintf(fichier, "%% Grille axe vertical, logarithmique\n");
  echelle = ( (double) (y_fin - y_deb)) / (log(y_max) - log(y_min)) ;

  points[0] = x;
  points[2] = x + larg;

  p_10 = (int) (log( y_max ) / log(10.0));
  p = 0;
  if (style < 7) { /* On trace avec des lignes */
    PS_couleur(couleur);
    PS_StyleLigne(fichier, style);

    do {
      double y, facteur;
      int i;

      facteur = pow(10.0, (double) p_10);
      for (i = 1 ; i <= 9 ; i++) {
	y = facteur * (double) i;

	if ( (y >= y_min) && (y <= y_max) ) {
	  points[1] = y_deb + (log(y) - log(y_min)) * echelle;
	  points[3] = points[1];

	  (void) fprintf(fichier, "%d %d moveto %d %d lineto stroke\n", x, points[1], x + larg, points[3]);
	}
      }
      p++;
      p_10--;
    } while ( p <= precision );
  } else {
    fprintf(stderr, "PS_grille_ver_log : tracer en polymark, À FAIRE\n");
  }
}
#endif

/* -~-~-~-~-~-~-~-~ Routines de dessins fondamentales -~-~-~-~-~-~-~-~ */

/* -------- Définition de la taille du texte -------------------------\
 | Renvoie la hauteur en 1/100 mm du texte [unité utilisé pour le PS] |
 |                                                                    |
 | fichier    : le fichier PS en cours d'écriture                     |
 | nom_fonte  : la fonte PS à utiliser                                |
 | haut_texte : taille à utiliser, en points                          |
 \------------------------------------------------------------------- */

static double PS_haut_texte(FILE *fichier, const char *nom_fonte, double haut_texte)
{
  double h_reelle;

  h_reelle = haut_texte * (2540.0 / 72.0) ; /* hauteur réelle en 1/100 mm [1 point = 1/72 pouce, 1 pouce = 2,54 cm = 25,4 mm] */
  fprintf(fichier, "/%s findfont %f scalefont setfont\n", nom_fonte, h_reelle); /* La fonte est au début haute de 1 unité = 1/100 mm */
  fprintf(fichier, "/hauteur_texte %f def\n", h_reelle);
  /* Remarque : comment ne pas afficher le 0 qui sert à calculer la taille des chiffres ? */
  fprintf(fichier, "gsave Blanc 0 0 moveto /hauteur_chiffre (0) false charpath flattenpath pathbbox exch pop exch sub exch pop def stroke grestore\n");
  return h_reelle;
}

/* -------- Sortie d'une chaîne pour Postscript -------- */

static void PS_sort_chaine(FILE *fichier, const char *chaine)
{
  unsigned long lg_chaine, i;
  const char *car;

  if (fichier == NULL) return;
  if (chaine == NULL) return;

  fprintf(fichier, " (");
  lg_chaine = strlen(chaine);
  for (i = 0, car = chaine ; i < lg_chaine; i++) {
    switch ( *car )
    {
    case '\\':
      (void) fprintf(fichier, "\\\\");
      car++;
      break;

    case '(': /* Marque de début de chaîne en PS */
      (void) fprintf(fichier, "\\(");
      car++;
      break;

    case ')': /* Marque de fin de chaîne en PS */
      (void) fprintf(fichier, "\\)");
      car++;
      break;

    default:
      fputc(*car, fichier);
      car++;
    }
  }

  (void) fprintf(fichier, ") ");
}

/* ~-~-~-~-~-~-~-~-~ Interface graphique de paramétrage ~-~-~-~-~-~-~-~-~ */

/* -------- Appel -------- */

int PS_Parametrage(struct infos_ps *infos)
{
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas de place pour mémoriser les infos.\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Préparation éventuelle de la boîte de dialogue */
  if ( FAUX == arbres[ EXPORT_PS ].drapeaux.prepare ) {
    prepare_arbre_PS();
  }

  /* Appel de la boîte de dialogue */
  if ( ARB_Bloquant( EXPORT_PS, txt_titre_dialogue, principale, 0 ) == EPS_ANNULER )
    return +5; /* On annule l'impression */

  /* Lecture des options */
  infos->proprietes = 0;
  if ( VRAI == OBJ_choisit( EXPORT_PS, EPS_TOURNEE, QUESTION ) ) {
    infos->proprietes |= PS_PAGE_TOURNEE;
  }
  if ( VRAI == OBJ_choisit( EXPORT_PS, EPS_NIVEAU2, QUESTION ) ) {
    infos->proprietes |= PS_NIVEAU_2;
  }

  /* Lecture des dimensions */
  infos->larg_page = OBJ_LireNombre( EXPORT_PS, EPS_LARGEUR, NULL ); /* Page, en mm */
  infos->haut_page = OBJ_LireNombre( EXPORT_PS, EPS_HAUTEUR, NULL );

  infos->marges.haut   = OBJ_LireNombre( EXPORT_PS, EPS_MARGE_HAUTE, NULL ); /* Valeurs des marges, en mm */
  infos->marges.bas    = OBJ_LireNombre( EXPORT_PS, EPS_MARGE_BASSE, NULL );
  infos->marges.gauche = OBJ_LireNombre( EXPORT_PS, EPS_MARGE_GAUCHE, NULL );
  infos->marges.droite = OBJ_LireNombre( EXPORT_PS, EPS_MARGE_DROITE, NULL );

  /* On en déduit la taille de la zone de dessin (en mm) */
  if ( infos->proprietes & PS_PAGE_TOURNEE ) {
    infos->larg_dessin = infos->haut_page - (infos->marges.bas    + infos->marges.haut  );
    infos->haut_dessin = infos->larg_page - (infos->marges.droite + infos->marges.gauche);
  } else {
    infos->larg_dessin = infos->larg_page - (infos->marges.droite + infos->marges.gauche);
    infos->haut_dessin = infos->haut_page - (infos->marges.bas    + infos->marges.haut  );
  }

  /* Lecture des tailles des objets */
  infos->haut_marque = OBJ_LireNombre( EXPORT_PS, EPS_HAUT_MARK, NULL ); /* Hauteur de marque, en mm */
  infos->epaisseur   = OBJ_LireNombre( EXPORT_PS, EPS_EPAISSEUR , NULL ); /* Épaisseur de ligne, en mm */
  infos->haut_texte  = OBJ_LireNombre( EXPORT_PS, EPS_HAUT_TEXTE, NULL ); /* Hauteur de texte, en pt */

  /* Police de caractères */
  OPTION_A_FAIRE( "Choix de la police de caractères" );
  infos->nom_fonte = NULL;

  /* Tout va bien */
  return 0;
}

/* ~~~~~~~~ Préparation de l'interface graphique ~~~~~~~~ */

static void prepare_arbre_PS(void)
{
  int retour;

  /* Initialisation */
  retour = ARB_Initialiser( EXPORT_PS, "menu_fichiers.html#export_ps", ( int (*)(unsigned int, int , int) ) NULL, ( int (*)(unsigned int, int, int *) ) NULL );
  if ( retour != 0 ) return;

  /* Traductions éventuelles */
#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_PAGE   , texte_dimensions_page );
  OBJ_TraduireTexte( EXPORT_PS, EPS_TOURNE     , texte_tournee );
  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_LARGEUR, texte_largeur);
  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_HAUTEUR, texte_hauteur);

  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_MARGE_H, texte_marge_haute);
  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_MARGE_B, texte_basse );
  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_MARGE_D, texte_droite );
  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_MARGE_G, texte_gauche );

  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_EPAIS   , texte_epaisseur );
  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_HAUT_MK , texte_hauteur_marque );
  OBJ_TraduireTexte( EXPORT_PS, EPS_TXT_HAUT_TXT, texte_hauteur_texte );

  OBJ_TraduireTexte( EXPORT_PS, EPS_NIVEAU2, texte_niveau2 );
  OBJ_TraduireTexte( EXPORT_PS, EPS_FONTE  , texte_fonte );

  OBJ_TraduireTexte( EXPORT_PS, EPS_OK     , texte_ok );
  OBJ_TraduireTexte( EXPORT_PS, EPS_ANNULER, texte_annuler );
#endif

  OBJ_deplace( EXPORT_PS, EPS_TXT_HAUTEUR , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_HAUTEUR     , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_TXT_MARGE_D , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_MARGE_DROITE, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_TXT_MARGE_G , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_MARGE_GAUCHE, 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( EXPORT_PS, EPS_TXT_HAUT_MK , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_HAUT_MARK   , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_TXT_HAUT_TXT, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_HAUT_TEXTE  , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_TXT_FONTE   , 0, +4, DEPLACE_PIXELS );
  OBJ_deplace( EXPORT_PS, EPS_FONTE       , 0, +4, DEPLACE_PIXELS );

#ifndef _DEVELOPPE_
  OBJ_active(EXPORT_PS, EPS_FONTE, FORCE_NON);
#endif

  arbres[EXPORT_PS].drapeaux.prepare = VRAI;
}
