/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996 - 1999, Emmanuel CURIS
   Cration du fichier : 27 mai 1998

   Tout ce qui concerne le calcul du spectre d'absorption  partir des fichiers exprimentaux

   Calcul de l'absorption  partir de I0 et de I1 se trouvant dj en mmoire
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "interface.h"
#include "extrait.h"
#include "lissage.h" /* Si on veut recaler I0 ou I1 */
#include "serie.h"   /* Pour insertion dans une srie */

#include "extrait_calcul.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GER_recompose    (unsigned int objet, int x, int y);
static void    prepare_recompose(void);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~- Calculs pratiques ~-~-~-~-~-~-~-~- */

/* ------ Calcul du spectre de fluorescence -------- */

Graphe *MU_Fluorescence(Graphe *i0, Graphe *i1)
{
  Graphe *spectre;
  int erreur;

  /* Vrifications prliminaires */
  if ( ( NULL == i0 ) || ( NULL == i1 ) ) return NULL;
  if ( i0->nombre_point != i1->nombre_point ) {
    (void) BI_ERR_alerte( NULL, &ERR_NombreDePointsDifferent, VRAI );
    return NULL;
  }

  /* On vrifie que le I0 est correct */
  if ( i0->y_min <= 0.0 ) {
    if ( 2 == BI_ERR_alerte(NULL, &ERR_I0_Nul, VRAI) ) {
      return NULL;
    }
  }

  /* Quotient s'occupe des barres d'erreur et tout a
     Remarque : si i1 est isol, on peut l'utiliser pour mmoriser le rsultat */
  spectre = GRF_QuotientGraphes(i1, i0, ( i1->drapeaux.isole == VRAI ), VRAI, NULL, &erreur);
  if ( NULL == spectre ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "MU_Fluorescence : ERREUR dans GRF_QuotientGraphes() [code %d]\n", erreur);
#endif
    HIST_Erreur(1, "Calcul annul"e_AIGU, "Impossible de cr"e_AIGU"er le graphe quotient");
    return NULL;
  }
  GRF_ChangerType( spectre, GR_BRUT );

  GRF_titre(spectre, txt_titre_signal_fluo);

  HIST_Texte(1, 0, "Coefficient "MU" : %s\n", spectre->nom);
  HIST_Texte(2, 0, "Technique : fluorescence\n");
  
  return spectre;
}

/* ------ Calcul du spectre de transmission -------- */

Graphe *MU_Transmission(Graphe *i0, Graphe *i1)
{
  Graphe *spectre;
  double *v_i0, *v_i1, *mu;
  unsigned long i;
  boolean erreur = FAUX; /* VRAI s'il y a dj eu une erreur et que l'on a demand de les ignorer */ 

  /* Vrifications prliminaires */
  if ( ( NULL == i0 ) || ( NULL == i1 ) ) return NULL;
  if ( i0->nombre_point != i1->nombre_point ) {
    (void) BI_ERR_alerte( NULL, &ERR_NombreDePointsDifferent, VRAI );
    return NULL;
  }

  /* On cre le spectre : mme X que i0 [et i1...] ; si possible, aprs I1
     Si I1 est isol, alors on cre le graphe isol aussi -> sera chan dans XAN_extrait() */
  spectre = GRF_nv_TableX( i1, GR_BRUT, GRF_FCT_REELLE, txt_titre_signal_trans, (i1->drapeaux.isole == VRAI), i0->x, i0->nombre_point, VRAI );
  if (spectre == NULL) {
    (void) fprintf(stderr, "MU_Transmission : ERREUR dans GRF_nouveau()\n");
    return NULL;
  }

  HIST_Texte(1, 0, "Coefficient "MU" : %s\n", spectre->nom);
  HIST_Texte(2, 0, "Technique : transmission\n");

  /* Les X des deux graphes sont les mmes */
  memcpy( spectre->x, i0->x, sizeof( double ) * spectre->nombre_point );
  spectre->x_min = i0->x_min;
  spectre->x_max = i0->x_max;

  /* Calcul du spectre proprement dit : ln(I0/If)
     Vrifie aussi qu'il n'y a pas annulation de I0 ni de I1 */
  for ( i = 0, v_i0 = i0->y, v_i1 = i1->y, mu = spectre->y ;
	i < spectre->nombre_point ; i++, v_i0++, v_i1++, mu++ ) {
    if ( *v_i0 <= 0.0 ) {
      if ( FAUX == erreur ) { /* Pas encore de message d'erreur */
	int retour;

	retour = BI_ERR_alerte( NULL, &ERR_I0_Nul, VRAI );
	if ( 2 == retour ) {
	  if ( FAUX == GRF_AdapteTaille( spectre, i ) ) {
	    GRF_efface( spectre, VRAI, VRAI );
	    return NULL;
	  }
	  break; /* Pour adapter les erreurs correctement ensuite ! */
	}
	erreur = VRAI;
      }

      *mu = 0.0; /* Arbitraire */
    } else if ( *v_i1 <= 0.0 ) {
      if ( FAUX == erreur ) { /* Si la question n'a pas dj t pose */
	int retour;

	retour = BI_ERR_alerte( NULL, &ERR_Itransmis_Nul, VRAI );
	if ( 2 == retour ) {
	  if ( FAUX == GRF_AdapteTaille( spectre, i ) ) {
	    GRF_efface( spectre, VRAI, VRAI );
	    return NULL;
	  }
	  break; /* Pour adapter les erreurs correctement ensuite ! */
	}
      }
      erreur = VRAI;
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "Intensit transmise nulle : E = %f eV, I0 = %g coups\n", spectre->x[i], *v_i0 );
#endif
      *mu = 0.0; /* Arbitraire... */
    } else {
      *mu = log( *v_i0 / *v_i1 );
    }
  }

  /* Calcul des bornes */
  GRF_bornes(spectre, NE_PAS_FAIRE, DESORDONNE);

  /* Erreurs sur les abscisses, ventuellement */
  if ( i0->erreur_x != NULL ) {
    if ( i1->erreur_x != NULL ) {
      printf("MU_Transmission :  FAIRE --- Erreurs en X sur I0 et sur I1\n");
    } else {
      GRF_AssignerErreurs( spectre, i0->erreur_x, VRAI, AXE_X );
    }
  } else if ( i1->erreur_x != NULL ) {
    GRF_AssignerErreurs( spectre, i1->erreur_x, VRAI, AXE_X );
  }

  /* Erreurs sur les ordonnes, ventuellement */

  /* On essaye maintenant de voir si, par hasard, on ne peut pas s'occuper des erreurs
     Premier cas : on connat les erreurs sur I0 et sur If/I1 */
  if ( i0->erreur_y != NULL ) {
    if ( i1->erreur_y != NULL ) {
      /* Le thorme de propagation des erreurs donne  V(ln) = s1/i0 + s2/i1 */
      Graphe *err_y;
      register double *y_err = NULL, *i0_err, *i1_err;

      err_y = GRF_creer_erreurs(spectre, AXE_Y);
      if (err_y == NULL) {
	return spectre; /* valuation impossible */
      }

      for (i = 0, i0_err = i0->erreur_y->y, i1_err = i1->erreur_y->y, y_err = err_y->y, v_i0 = i0->y, v_i1 = i1->y ;
	   i < spectre->nombre_point ;
	   i++, i0_err++, i1_err++, y_err++, v_i0++, v_i1++) {
	double sigma;

	if ( (*v_i0 != 0.0) && (*v_i1 != 0.0) ) {
	  sigma  = (*i0_err / *v_i0) * (*i0_err / *v_i0);
	  sigma += (*i1_err / *v_i1) * (*i1_err / *v_i1);
	  *y_err = sigma;
	} else {
	  *y_err = 0.0; /* Puisqu'alors, on a mis 0 arbitrairement... */
	}
      }
    } else {
      printf("MU_Transmission :  FAIRE --- Erreurs en Y sur I0 seulement\n");
    }
  } else if ( i1->erreur_x != NULL ) {
    printf("MU_Transmission :  FAIRE --- Erreurs en Y sur I1 seulement\n");
  }
  GRF_bornes( spectre->erreur_y, NE_PAS_FAIRE, DESORDONNE);

  /* On a fini, tout s'est bien pass */
  return spectre;
}

/* ~-~-~-~-~-~-~-~- Interface ~-~-~-~-~-~-~-~- */

#define LGMAX_NOM_GRAPHE_I0 39
#define LGMAX_NOM_GRAPHE_I1 37

struct infos_calcul {
  Graphe *grf_i0;
  GRF_liste *liste_i1;
};

/* ------------- Calcul d'un spectre  partir de I0 et I1 -------------------------------- *
 *******************************************************************************************/

Graphe *XAN_extrait(void)
{
  Graphe *spectre = NULL, *gr_i0, *gr_i1 = NULL;
  Serie_Graphes *serie_resultat = NULL;
  int retour = -1;
  boolean fluo = FAUX; /* VRAI si fluorescence */
  unsigned long i;
  struct infos_calcul infos = { NULL, NULL };

  if (arbres[RECOMP_BRUT].drapeaux.prepare == FAUX)
    prepare_recompose();

  /* Pas encore de I0 choisi */
  met_chaine( RECOMP_BRUT, RB_I0, LGMAX_NOM_GRAPHE_I0, texte_aucun_i0, TXT_COPIER );

  /* Liste des graphes de type I1 (toutes sries confondues) */
  infos.liste_i1 = GRF_par_type(GR_I1, GRF_LISTE_TYPES_MESSAGES | GRF_LISTE_TYPES_MULTISELECTION | GRF_LISTE_TYPES_MULTISERIE, NULL);
  if ( NULL == infos.liste_i1 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la liste des graphes de type I1\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  GRF_affiche( infos.liste_i1, RECOMP_BRUT, RCP_BRT_I1_1, RCP_BRT_I1_N, LGMAX_NOM_GRAPHE_I1,
	       RCP_BRT_I1_AV, RCP_BRT_I1_AP, RCP_BRT_I1_ASC, RCP_BRT_I1_BOUT );

  ARB_MettreComplement( RECOMP_BRUT, &infos );
  retour = ARB_Bloquant( RECOMP_BRUT, txt_titre_calc_signal, principale, 0 );
  ARB_MettreComplement( RECOMP_BRUT, NULL );
  if (retour == RCP_BRT_ANNULE) {
    GRF_LibererListe( infos.liste_i1 );
    return NULL;
  }

  gr_i0 = infos.grf_i0; /* On rcupre les graphes choisis */

  fluo = OBJ_choisit(RECOMP_BRUT, RCP_BRT_FLUO, QUESTION);

  /* Dans quelle srie seront les rsultats ? */
  if ( VRAI == OBJ_choisit( RECOMP_BRUT, RB_NLLE_SERIE, QUESTION ) ) {
    serie_resultat = SERIE_GRF_CreerSerie( serie_courante, "Graphes calcul"e_AIGU"s" );
  } else {
    serie_resultat = serie_courante;
    if ( VRAI == serie_resultat->drapeaux.protegee ) { /* La srie courante tant protge, on doit en crer une ! */
      serie_resultat = SERIE_GRF_CreerSerie( serie_courante, "Graphes calcul"e_AIGU"s" );
    }
  }
  if ( NULL == serie_resultat ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer la srie des rsultats !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    GRF_LibererListe( infos.liste_i1 );
    return NULL;
  }

  for ( i = 0 ; i < infos.liste_i1->nbr_graphe ; i++ ) {
    Graphe *vx_i0 = NULL, *vx_i1 = NULL;

    if ( FAUX == infos.liste_i1->choisi[ i ] ) continue; /* Non choisi -> on passe au suivant */

    /* On veut le calcul pour ce spectre */
    gr_i1 = infos.liste_i1->graphes[ i ];

    /* Log dans l'historique */
    HIST_Titre(1, "Calcul du coefficient d'absorption", "MU");
    HIST_Texte(1, 0, "Intensit"e_AIGU" I0 : %s\n", gr_i0->nom);
    HIST_Texte(1, 0, "Intensit"e_AIGU" I1 : %s\n", gr_i1->nom);

    if ( VRAI == OBJ_choisit( RECOMP_BRUT, RB_X_I1, QUESTION ) ) { /* On veut caler I0 sur I1 */
      vx_i0 = gr_i0;

      /* On cre un graphe isol avec la table des X de I1 */
      gr_i0 = GRF_nv_TableX( NULL, GR_I0, GRF_SERIE, NULL, VRAI, gr_i1->x, gr_i1->nombre_point, FAUX );
      if ( NULL == gr_i0 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de recaler I0 sur I1\n",
		       __FILE__, __FUNCTION__, __LINE__);
	continue;
      }

      /* Et on recale dessus I0 */
      LIS_RecalerGraphe( vx_i0, gr_i0, INTERPOLE_LINEAIRE, INTERPOLE_BORDS_0_0 );
    } else if ( VRAI == OBJ_choisit( RECOMP_BRUT, RB_X_SUR_I0, QUESTION ) ) { /* On veut caler I1 sur I0 */
      vx_i1 = gr_i1;

      /* On cre un graphe isol avec la table des X de I1 */
      gr_i1 = GRF_nv_TableX( NULL, GR_I1, GRF_SERIE, NULL, VRAI, gr_i0->x, gr_i0->nombre_point, FAUX );
      if ( NULL == gr_i1 ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de recaler I1 sur I0\n",
		       __FILE__, __FUNCTION__, __LINE__);
	continue;
      }

      /* Et on recale dessus I1 */
      LIS_RecalerGraphe( vx_i1, gr_i1, INTERPOLE_LINEAIRE, INTERPOLE_BORDS_0_0 );
    }

    if ( VRAI == fluo ) { /* Fluorescence : calcul de If / I0 */
      spectre = MU_Fluorescence( gr_i0, gr_i1 );
    } else { /* On est en transmission */
      spectre = MU_Transmission( gr_i0, gr_i1 );
    }
    if ( NULL == spectre ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR survenue dans le calcul, I0 = %s, I1 = %s\n",
		     __FILE__, __FUNCTION__, __LINE__, gr_i0->nom, gr_i1->nom);
      continue;
    }

    if ( VRAI == OBJ_choisit(RECOMP_BRUT, RCP_BRT_ERREUR, QUESTION) ) {
      (void) fprintf(stderr, "%s,%s l.%u :  FAIRE : estimer les erreurs sans erreurs  l'origine !\n",
		     __FILE__, __FUNCTION__, __LINE__);
    }

    /* On se dbarrasse des copies recales, si ncessaire ; elles sont isoles par construction */
    if ( vx_i0 != NULL ) {
      if ( spectre != gr_i0 ) GRF_eff_isole( gr_i0 );
      gr_i0 = vx_i0;
    }
    if ( vx_i1 != NULL ) {
      if ( spectre != gr_i1 ) GRF_eff_isole( gr_i1 );
      gr_i1 = vx_i1;
    }
    if ( VRAI == spectre->drapeaux.isole ) { /*  cause des recalages... */
      (void) fprintf(stderr, "%s,%s l.%u : Spectre isol [%s]\n",
		     __FILE__, __FUNCTION__, __LINE__, spectre->nom);
      GRF_InsererDansSerie( spectre, NULL , gr_i1 );
      (void) fprintf(stderr, "%s,%s l.%u : ==> appartient maintenant  la srie %p [%s]\n",
		     __FILE__, __FUNCTION__, __LINE__, GRF_QuelleSerie( spectre )->nom, GRF_QuelleSerie( spectre )->nom );
    }

    /* On installe dans la srie des rsultats */
    if ( serie_resultat != serie_courante ) {
      GRF_ChangeSerie( spectre, serie_resultat, NULL );
    }
  }


  GRF_LibererListe( infos.liste_i1 );

  return spectre;
}

/* -------- Gestion de l'interface -------- */

static int GER_recompose(unsigned int objet, int x, int y)
{
  struct infos_calcul *infos;

  infos = ( struct infos_calcul * ) ARB_LireComplement( RECOMP_BRUT );

  if ( (objet >= RCP_BRT_I1_1) && (objet <= RCP_BRT_I1_N) ) {
    GRF_selecteur( infos->liste_i1, RECOMP_BRUT, RCP_BRT_I1_1, RCP_BRT_I1_N, LGMAX_NOM_GRAPHE_I1,
		   RCP_BRT_I1_AV, RCP_BRT_I1_AP, RCP_BRT_I1_ASC, RCP_BRT_I1_BOUT,
		   objet, x, y );
    return VRAI;
  }

  switch (objet)
  {
  case RB_I0: /* Choix de I0 */
    {
      Graphe *nv_i0;

      nv_i0 = GRF_choisir( GR_I0, FAUX, VRAI, NULL /*serie_courante*/ );
      if ( nv_i0 != NULL ) {
	met_chaine( RECOMP_BRUT, RB_I0, LGMAX_NOM_GRAPHE_I0, nv_i0->nom, TXT_COPIER );
	infos->grf_i0 = nv_i0;
      }
    }
    return VRAI;

  case RCP_BRT_I1_AV:
  case RCP_BRT_I1_AP:
  case RCP_BRT_I1_ASC:
  case RCP_BRT_I1_BOUT: /* Dfileur des graphes de type I1 */
    GRF_selecteur( infos->liste_i1, RECOMP_BRUT, RCP_BRT_I1_1, RCP_BRT_I1_N, LGMAX_NOM_GRAPHE_I1,
		   RCP_BRT_I1_AV, RCP_BRT_I1_AP, RCP_BRT_I1_ASC, RCP_BRT_I1_BOUT,
		   objet, x, y);
    return VRAI;
  }

  return FAUX;
}

/* -------- Prparation de l'arbre -------- */

static void prepare_recompose(void)
{
  if ( FAUX == arbres[ RECOMP_BRUT ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( RECOMP_BRUT, "menu_absorption.html#calc_signal", GER_recompose, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
  }

  OBJ_agrandit(RECOMP_BRUT, RB_FOND_METHODE, 0, +3);
  OBJ_deplace(RECOMP_BRUT, RCP_BRT_DETECT, 0, 1, DEPLACE_PIXELS);
  OBJ_deplace(RECOMP_BRUT, RCP_BRT_ABSORBE, 0, 2, DEPLACE_PIXELS);

  ARB_EnjoliverDefileur( RECOMP_BRUT, RCP_BRT_I1_FND, RCP_BRT_I1_1, RCP_BRT_I1_N,
			 RCP_BRT_I1_AV, RCP_BRT_I1_ASC, RCP_BRT_I1_BOUT, RCP_BRT_I1_AP, ARB_ASCENSEUR_DROITE );

  OBJ_deplace(RECOMP_BRUT, RB_X_SUR_I0, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(RECOMP_BRUT, RB_TXT_X_I0, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(RECOMP_BRUT, RB_X_I1    , 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(RECOMP_BRUT, RB_TXT_X_I1, 0, +1, DEPLACE_PIXELS);

#ifdef AVEC_TRADUCTION_
  met_chaine(RECOMP_BRUT, RCP_BRT_DETECT, 0, txt_arbre_detection, TXT_REMPLACE);
  met_chaine(RECOMP_BRUT, RCP_BRT_ERREUR, 0, txt_arbre_erreur, TXT_REMPLACE);
  met_chaine(RECOMP_BRUT, RCP_BRT_OK, 0, txt_arbre_ok, TXT_REMPLACE);
#endif

#ifndef _DEVELOPPE_
  OBJ_active(RECOMP_BRUT, RCP_BRT_DETECT, FORCE_NON);
#endif

  arbres[RECOMP_BRUT].drapeaux.prepare = VRAI;
}
