/* LASE - Logiciel d'analyse des spectres EXAFS
   (c) 1998, Emmanuel CURIS - Cration de ce fichier: 23 juin 1998

   Module de filtrage des spectres (essentiellement par transforme de Fourier)

   Paramtrage de la transforme de Fourier directe
   Source original rcupr de ../FOURIER.c et ../EXF_GEM.c
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"
#include "filtrage.h"
#include "extrait.h" /* Pour les commandes de chanage... */
#include "tracer.h"

#include "fourier.txt"

/* ******** Variables locales ************************ */

static int num_tf_fenetre = FEN_BESSEL;  /* Fentre  l'ouverture du module */

#define TF_PAS_APERCU  0.1 /* Pas pour le calcul de l'aperu */
#define TF_MAX_APERCU 10.0 /* Rmax de calcul de l'aperu     */


#define CHAINE_COMPLETE  2	/*  remplacer par les I_XAS_CHAINE_* en sparant l'interface ! */
#define CHAINE_PARTIELLE 3

/* ******** Fonctions locales (pour la gestion de l'interface) ******** */

static void RDS_tf(Fenetre *, int x, int y, unsigned int larg, unsigned int haut);
static int  MSG_tf(Fenetre *fen, Message *msg);
static int  GER_tf(unsigned int objet, int x, int y);
static int  CLV_tf(Fenetre *fen, unsigned int touche, int etat);
static int  CLC_tf(Fenetre *fen, int bouton, int n_clic, int etat, int x, int y);

static boolean TF_calc_apercu(TF_infos *infos);
static boolean TF_prep_apercu(TF_infos *infos);

static int GER_debTF(unsigned int objet, int x, int y);

static void met_les_k(TF_infos *infos);
static boolean TF_InstallerFenetre(TF_infos *infos, int nv_type);
static int     TF_DemanderLesk    (Fenetre *fenetre, TF_infos *infos);

static int prepare_param_tf(unsigned int);
static void prepare_debut_tf(void);

extern int GER_reproduit_filtre(unsigned int objet, int x, int y); /* Dans Filtrage/reproduire.c */

/* -------- Calcul de la transforme de Fourier : interface avec l'extrieur -------- */

Graphe *TF_CalculDirect(TF_infos *infos, double r_min, double r_max, double pas,
			int methode, boolean erreurs, boolean covariance)
{
  char *titre, *unite;
  const char *unite_X;
  Graphe *resultat, *fenetre, *original;

  if ( NULL == infos ) return NULL;
  if ( NULL == infos->graphe ) return NULL;

  /* Contrle des bornes et du pas */
  if (r_max < r_min) { /* On change les deux */
    double r = r_max; r_max = r_min; r_min = r;
  }
  if (pas < 0.0) pas = -pas;

  /* Contrle pour le calcul des erreurs */
  if ( NULL == infos->graphe->erreur_y ) /* Pas d'erreur sur le graphe original */
    erreurs = FAUX;
  if ( FAUX == erreurs ) /* Si on ne s'intresse pas aux erreurs, on ne peut calculer la matrice de covariance */
    covariance = FAUX;

  /* ---- Calcul du graphe  transformer ---- */

  /* 0. Reprage de la partie intressante */
  original = TF_IsoleSignal( infos->graphe, &(infos->apodisation) );
  if ( NULL == original ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'isoler le signal original\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* 1. Calcul de la fonction d'apodisation
        Les X sont les mmes que l'original pour faire le produit ensuite ! */
  fenetre = GRF_nv_TableX( NULL, GR_FENETRE, GRF_SERIE, "Fentre de TF l.86", VRAI,
			   original->x, original->nombre_point, FAUX );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le graphe de la fentre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    if ( original != infos->graphe ) { /* Si ce n'est pas la source, c'est un extrait : graphe ISOL */
      GRF_eff_isole( original );
      original = NULL;
    }
    return NULL;
  }
  TF_fenetre( fenetre, &(infos->apodisation) );

  /* 2. Produit graphe * fentre => fenetre - s'occupe des erreurs */
  if ( NULL == GRF_ProduitGraphes(fenetre, original, VRAI, VRAI, NULL) ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le produit des graphes\n",
		   __FILE__, __FUNCTION__, __LINE__);
    if ( original != infos->graphe ) {
      GRF_eff_isole( original ); /* Plus besoin du graphe extrait (graphe isol) */
      original = NULL;
    }
    GRF_eff_isole( fenetre );
    fenetre = NULL;
    return NULL;
  }

  /* Log. dans l'historique */
  HIST_Titre(1, "Transformation de Fourier", "TF");
  HIST_Texte(1, 0, "Graphe original : %s\n", infos->graphe->nom);
  HIST_Texte(3, 0, "Calcul pour R compris entre %.10f et %.10f avec un pas de %.10e\n", r_min, r_max, pas);
  if ( infos->apodisation.type != PAS_DE_FENETRE ) {
    HIST_Texte(2, 0, "Fonction d'apodisation : type = %s\n", Types_fenetres[infos->apodisation.type]);
    HIST_Texte(3, 0, "                         [kmin, kmax] = [%.10f, %.10f]\n", infos->apodisation.x_min, infos->apodisation.x_max);
    if ( Autorise_Symetrique( infos->apodisation.type ) ) { /* Fentres dfinies par 4 points */
      HIST_Texte(3, 0, "Fin de la mont"e_AIGU"e     : %.10g\n", infos->apodisation.x_1);
      HIST_Texte(3, 0, "D"e_AIGU"but de la descente : %.10g\n", infos->apodisation.x_2);
    }
    if ( Avec_Parametre( infos->apodisation.type ) ) {
      HIST_Texte(3, 0, "Param"e_GRAVE"tre : %.10f\n", infos->apodisation.parametre);
    }
  } else {
    HIST_Texte(2, 0, "Pas de fonction d'apodisation\n");
  }

  /* Correction d'amplitude ? */
  if ( infos->amplitude != NULL ) {
    HIST_Texte(2, 0, "Correction d'amplitude demand"e_AIGU"e\n");
    HIST_Texte(3, 2, "Amplitude correctrice : %s\n", infos->amplitude->nom);

    if ( FAUX == TF_coramp( fenetre, infos->amplitude ) ) {
      GRF_eff_isole(fenetre);
      HIST_Erreur(1, "calcul annul"e_AIGU, "Impossible de corriger l'amplitude");
      return NULL;
    }
  }

  /* Correction de phase ? */
  if ( infos->phase != NULL ) {
    HIST_Texte(2, 0, "Correction de phases demand"e_AIGU"e\n");
    HIST_Texte(3, 2, "D"e_AIGU"phasage correcteur : %s\n", infos->phase->nom);

    if ( FAUX == TF_corpha( fenetre, infos->phase ) ) {
      GRF_eff_isole(fenetre);
      HIST_Erreur(1, "calcul annul"e_AIGU, "Impossible de corriger la phase");
      return NULL;
    }
  }

  /* L, fenetre contient le spectre prt  tre transform */

  switch ( methode )
  {
  case TF_CALC_RAPIDE: /* Algorithme rapide basique pour 2^n valeurs */
    HIST_Texte(4, 0, "Algorithme : TF rapide pour 2^n points "e_AIGU"galement r"e_AIGU"partis\n");
    resultat = TF_prep_rapide(fenetre, r_min, r_max, pas, TF_SENS_DIRECT);
    if ( NULL == resultat ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans TF_prep_rapide()\n",
		     __FILE__, __FUNCTION__, __LINE__);
      HIST_Erreur(1, "calcul annul"e_AIGU, "Impossible de pr"e_AIGU"parer les graphes");
      return NULL;
    }
    TF_rapide(fenetre, resultat);
    if (r_max < resultat->x_max) { /* On ralise la troncature */
      double *x;
      unsigned long i;

      for (x = resultat->x , i = 0 ; *x <= r_max ; x++, i++);
      resultat->nombre_point = i; /* Troncature _apparente_, la fin est toujours en mmoire... */
    }
    break;

  case TF_CALC_NORMAL: /* Intgration directe, trapzes */
    HIST_Texte(4, 0, "Algorithme : int"e_AIGU"gration par la m"e_AIGU"thode des trap"e_GRAVE"zes\n");
    resultat = TF_prep_normal(r_min, r_max, pas, TF_SENS_DIRECT, erreurs, covariance);
    if ( NULL == resultat ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans le calcul de la TF normale\n",
		     __FILE__, __FUNCTION__, __LINE__);
      HIST_Erreur(1, "calcul annul"e_AIGU, "Impossible de pr"e_AIGU"parer les graphes");
      return NULL;
    }
    TF_simple(fenetre, resultat);
    break;

  default:
    (void) fprintf(stderr, "TF_CalculDirect : mthode de calcul inconnue [%d]\n", methode);
    return NULL;
  }

  /* Plus besoin de l'original adapt */
  GRF_eff_isole(fenetre);
  fenetre = NULL;
  if ( original != infos->graphe ) {
    /* On ne peut effacer l'original copi (graphe isol) que maintenant,
       car ses X sont aussi ceux de fenetre ! */
    GRF_eff_isole( original );
    original = NULL;
  }

  /* Adaptations finales */
  resultat->pointeur.infos_tf = (Infos_FonctionApodisation *) malloc( sizeof(Infos_FonctionApodisation) );
  if ( NULL != resultat->pointeur.infos_tf )
    *(resultat->pointeur.infos_tf) = infos->apodisation; /* On conserve les paramtres de calcul de la TF */
  resultat->drapeaux.protege = FAUX;
  titre = TXT_genere_chaine(nom_graphe_tf, infos->graphe->nom);
  GRF_titre(resultat, titre);
  free( titre ); titre = NULL;
  GRF_bornes(resultat, ORDONNE, DESORDONNE);

  /* Unit en X : l'inverse de l'unit en X du graphe avant TF */
  unite_X = GRF_QuelleUnite( infos->graphe, AXE_X );
  unite = TXT_ProduitUnites( NULL, unite_X, -1 );
  GRF_ChangerUnite( resultat, AXE_X, unite );
  free( unite ); unite = NULL;

  /* Unit en Y : unit en Y avant TF * unit en X avant TF */
  unite = TXT_ProduitUnites( GRF_QuelleUnite( infos->graphe, AXE_Y ), unite_X, 1 );
  GRF_ChangerUnite( resultat, AXE_Y, unite );
  free( unite ); unite = NULL;

  /* On conserve couleur & style */
  GRF_CopierAffichage( resultat, infos->graphe, GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

  /* Calcul termin */
  HIST_Texte(1, 0, "R"e_AIGU"sultat : %s\n", resultat->nom);
  return resultat;
}

/* ******** Ralisation effective de l'interface ********************** */

/* -------- Appel de la fentre de transforme de Fourier --------------\
 |    Renvoie le graphe qui contient l'aperu, s'il existe, sinon NULL  |
 |    ATTENTION, ce graphe est un graphe isol !                        |
 |                                                                      |
 | graphe = le graphe dont on veut calculer la TF                       |
 | chaine = VRAI si appel lors d'une extraction chane, en une tape   |
 \--------------------------------------------------------------------- */

Graphe *TF_ouvre(Graphe *graphe, boolean chaine)
{
  Fenetre *fenetre;
  TF_infos *infos_tf;
  
  if (graphe == NULL) return NULL; /* Rien  faire l... */

  fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, titre_fenetre_tf, NULL,
				PARAM_FOURIER, BI_FEN_ARBRE_HAUT, prepare_param_tf,
	                        RDS_tf, CLV_tf, CLC_tf, MSG_tf, RCT_pointeur,
				( int (*)(Fenetre *) ) NULL);
  if (fenetre == NULL) return NULL; /* Rien  faire... */

  fenetre->drapeaux.redessin = VRAI;
  fenetre->msg_masque.fermeture = VRAI; /* Messages traits par MSG_tf */

  if (chaine == CHAINE_COMPLETE) {
    FEN_dimensions(fenetre, -250, -250, -500, -500);
  } else if (chaine == CHAINE_PARTIELLE) {
    FEN_dimensions(fenetre, -500, -500, -500, -500);
  }
  FEN_Ouvrir( fenetre );

  infos_tf = (TF_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( infos_tf != NULL ) { /* Dj des infos sur une TF -> on les perd */
    if ( infos_tf->graphe == graphe ) { /* Mme graphe -> pas  se fouler */
      return (infos_tf->apercu);
    } else {
      /* On efface ces infos pour en crer de nouvelles */
      /* On commence donc par les dlier de la fentre... */
      FEN_GRF_LierGraphe( fenetre, NULL );
      FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

      /* Maintenant, on peut effacer les graphes sans risque */
      GRF_efface( infos_tf->fenetre, VRAI, VRAI); /* On force l'effacement */
      infos_tf->fenetre = NULL;
      GRF_eff_isole( infos_tf->apercu );
      infos_tf->apercu = NULL;

      num_tf_fenetre = infos_tf->apodisation.type;

      infos_tf->graphe->drapeaux.protege = VRAI; /* Pas utilis -> on peut l'effacer */

      /* Infos_tf est associ au spectre calcul _aprs_ fermeture,
	 donc normalement s'il est associ  la fentre, c'est qu'il n'est
	 utilis par aucun graphe -> on peut l'effacer sans risque */
      free( infos_tf ); infos_tf = NULL;
    }
  }

  infos_tf = (TF_infos *) malloc( sizeof(TF_infos) ); /* On rserve les infos de TF */
  if ( NULL == infos_tf ) {
    FEN_Detruire( fenetre ); /* On ferme la fentre : sert  rien... */
    SYS_ERREUR(MANQUE_MEMOIRE, "fourier.c", "TF_ouvre", "infos_tf");
    return NULL;
  }

  infos_tf->fenetre = GRF_nouveau(graphe, graphe->nombre_point, GR_FENETRE, GR_NORMAL, nom_graphe_fenetre);
  if (infos_tf->fenetre == NULL) { /* Mmoire insuffisante */
    free( infos_tf ); infos_tf = NULL;
    FEN_Detruire( fenetre ); /* On ferme la fentre : sert  rien... */
    return NULL;
  }
  memcpy( infos_tf->fenetre->x, graphe->x, sizeof(double) * graphe->nombre_point);
  infos_tf->fenetre->x_min = graphe->x_min;
  infos_tf->fenetre->x_max = graphe->x_max;

  GRF_couleur( infos_tf->fenetre, BLEU, -1, -1 ); /* Pour que le module apparaisse en bleu...    */
  infos_tf->fenetre->type    = GR_INTERNE; /* Graphe interne, ineffaable et tout et tout */
  infos_tf->fenetre->drapeaux.protege = VRAI; /* Protg -> ineffaable                   */

  infos_tf->graphe = graphe; /* Le graphe sur lequel on travaille */
  graphe->drapeaux.protege = VRAI;  /* Il est donc ineffaable... */

  /* Paramtres de la TF (fentre en particulier) */
  infos_tf->apodisation.symetrique = OBJ_choisit(PARAM_FOURIER, TF_SYMETRIQUE, QUESTION); /* VRAI si fen. symtrique */

  infos_tf->apodisation.x_min = graphe->x_min + 1.0; /* Intervalle pour lequel on calculera */
  infos_tf->apodisation.x_max = graphe->x_max - 1.0;
  infos_tf->apodisation.x_1 = infos_tf->apodisation.x_min + 1.0;
  infos_tf->apodisation.x_2 = infos_tf->apodisation.x_max - 1.0;
  OBJ_MettreIntervalle( PARAM_FOURIER, FOUR_KMIN, FOUR_KMAX, 0, 0, infos_tf->apodisation.x_min, infos_tf->apodisation.x_max, 0.0, 0 );
  OBJ_MettreIntervalle( PARAM_FOURIER, PARAM_TF_K1, PARAM_TF_K2, 0, 0, infos_tf->apodisation.x_1, infos_tf->apodisation.x_2, 0.0, 0 );

  /* Calcul de la TF : fentre */
  infos_tf->apodisation.type = num_tf_fenetre; /* Fentre de Kaiser-Bessel par dfaut */
  infos_tf->num_depart   = FEN_RECTANGLE;  /* Affichage  partir de la premire */

  ARB_PreparerDefileur_T(PARAM_FOURIER, FOUR_FEN_1, FOUR_FEN_4, 0,
			 FOUR_FEN_AVANT, FOUR_FEN_FOND, FOUR_FEN_BOUTON, FOUR_FEN_APRES, 
			 NOMBRE_FENETRES, Types_fenetres, NULL, num_tf_fenetre, FEN_RECTANGLE);

  OBJ_active(PARAM_FOURIER, PARAM_TF_K1, Autorise_Symetrique(num_tf_fenetre) );
  OBJ_active(PARAM_FOURIER, PARAM_TF_K2, Autorise_Symetrique(num_tf_fenetre)
	                              && (infos_tf->apodisation.symetrique == FAUX) );
  OBJ_active(PARAM_FOURIER, TF_SYMETRIQUE, Autorise_Symetrique(num_tf_fenetre) );

  infos_tf->apodisation.parametre = OBJ_LireNombre( PARAM_FOURIER, FOUR_PARAM, NULL );
  TF_fenetre(infos_tf->fenetre, &(infos_tf->apodisation) ); /* On calcule la fentre */

  infos_tf->amplitude = NULL;
  infos_tf->phase = NULL;
  infos_tf->apercu = NULL;

  /* On associe tout  la fentre, maintenant que tout est prt... */
  FEN_GRF_LierGraphe( fenetre, graphe );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, infos_tf );

  /* On essaye de faire un aperu */
  TF_prep_apercu(infos_tf);
  TF_calc_apercu(infos_tf);

  return infos_tf->apercu;
}

/* ----------- Redessin de cette fentre ------------------------------------------------ */

static void RDS_tf(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Graphe *graphe, *produit;
  InfosRegion2D region;
  TF_infos *infos;
  double echelle_X, echelle_Y;
  int retour;

  /* Prparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );

  /* On rcupre les infos associes  la fentre */
  infos = (TF_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return;	/* Pas encore d'infos associes  la fentre ! */

  graphe = infos->graphe;
  if ( ( NULL == graphe ) || ( graphe->x_max == graphe->x_min ) || ( graphe->y_max == graphe->y_min ) ) {
    TRC_Message( fenetre, txt_err_signal_nul );
    return;
  }

  /* On prpare la rgion 2D : aide pour les dessins aprs en fait */
  retour = TRC_Z2D_Initialiser( &region, fenetre, graphe );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en initialisant la rgion 2D [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* A. Le graphe dont on fait la TF */
  /* On trace le graphe dans la rgion 2D et on rcupre les chelles au passage */
  TRC_GrapheDansRegion2D( fenetre, &region, 0, graphe, NOIR, BI_CONTINU, BI_MARQUE_AUCUNE, 0, &echelle_X, &echelle_Y );

  /* Dessin du graphe multipli par la fentre - en violet */
  /* a. On copie, en isol, la fentre */
  produit = GRF_Copier( infos->fenetre, NULL, VRAI );
  if ( produit != NULL ) {
    /* On calcule le produit */
    GRF_ProduitGraphes( produit, graphe, VRAI, FAUX, NULL );

    /* On trace, avec les mmes chelles */
    TRC_GrapheDansRegion2D( fenetre, &region, 0, produit, VIOLET, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );

    /* On n'a plus besoin de cet isol */
    GRF_eff_isole( produit ); produit = NULL;

    OPTION_A_FAIRE( "Trac du produit graphe * fenetre par un graphe isol  optimiser !" );
  }

  /* B. Dessin de la fentre : forcment comprise entre 0 et 1, occupe toute la fentre */
  region.axe_Y.min = 0.0;
  region.axe_Y.max = 1.0;
  TRC_GrapheDansRegion2D( fenetre, &region, 0, infos->fenetre, BLEU, BI_CONTINU, BI_MARQUE_AUCUNE, 0, NULL, NULL );

  /* C. Mise en place des barres pour k_min (rouge) et k_max (vert)
     [mmes X pour le graphe & la fentre, donc pas de souci] */
  if ( infos->apodisation.type != PAS_DE_FENETRE ) {
    int x_barre;

    if ( ( infos->apodisation.x_min >= graphe->x_min ) && ( infos->apodisation.x_min <= graphe->x_max ) ) {
      x_barre = TRC_ConvertitX_PvE( infos->apodisation.x_min, graphe->x_min, 0, echelle_X );
      TRC_Couleur( ROUGE );
      TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );
    }
	
    if ( ( infos->apodisation.x_max >= graphe->x_min ) && ( infos->apodisation.x_max <= graphe->x_max ) ) {
      x_barre = TRC_ConvertitX_PvE( infos->apodisation.x_max, graphe->x_min, 0, echelle_X );
      TRC_Couleur( VERT );
      TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );
    }

    /* Mise en place des barres pour k_1 et k_2, si la fentre le ncessite */
    if ( Autorise_Symetrique(infos->apodisation.type) ) {
      x_barre = TRC_ConvertitX_PvE( infos->apodisation.x_1, graphe->x_min, 0, echelle_X );
      TRC_Couleur( ROUGE_FONCE );
      TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );
    
      if (infos->apodisation.symetrique) {
	x_barre = TRC_ConvertitX_PvE( infos->apodisation.x_max - (infos->apodisation.x_1 - infos->apodisation.x_min), graphe->x_min, 0, echelle_X );
      } else {
	x_barre = TRC_ConvertitX_PvE( infos->apodisation.x_2, graphe->x_min, 0, echelle_X );
      }
      TRC_Couleur( VERT_FONCE );
      TRC_LigneV( fenetre, x_barre, 0, fenetre->w_haut );
    }
  }

  /* Affichage de l'aperu, si ncessaire */
  if ( infos->apercu != NULL) { /* Il y en a un */
    TRC_GrapheDansRectangle( fenetre, infos->apercu, ROUGE, 1, 0, 0, fenetre->w_larg, fenetre->w_haut );
  }
}

/* -------- Traitement des messages reus par la fentre --------------------------- */

static int MSG_tf(Fenetre *fenetre, Message *message)
{
  Graphe *graphe = NULL, *tf_fenetre = NULL, *fourier = NULL;
  TF_infos *infos; /* Les infos pour la fentre, pour la TF */
  int reponse;

  infos      = (TF_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( infos != NULL ) {
    graphe     = infos->graphe;
    tf_fenetre = infos->fenetre;
  }

  if ( message->type == MSG_GRAPHE_MODIFIE ) { /* Le brut a t modifi */
    Graphe *nv_brut = (Graphe *) (message->divers.pointeur);

    /* On dlie l'ancien graphe de la fentre */
    FEN_GRF_LierGraphe( fenetre, NULL );

    if ( message->divers.drapeau == VRAI) { /* Pas autant de point -> on joue sur realloc
					       NB : on vrifie a parce que le nombre de
					       points a pu changer plusieurs tapes avant,
					       et dans ce cas il semble que
					       la transmission se fasse mal... */
      if ( (GRF_AdapteTaille(tf_fenetre, nv_brut->nombre_point) == FAUX ) ) {
	message->type = WM_CLOSED;
	message->fenetre.haut = -1; /* On force la fermeture de cette fentre */
	return MSG_tf( fenetre, message );
      }

      /* Note : les X ont probablement chang, si le nombre de points a vari
	 En effet, c'est srement que l'on a jou sur E0
	 Or, la suite du calcul ne modifie _pas_ les X -> le faire ici */
      memcpy(tf_fenetre->x, nv_brut->x, sizeof(double) * nv_brut->nombre_point);
    }

    /* On lie le nouveau graphe de la fentre */
    FEN_GRF_LierGraphe( fenetre, nv_brut );
    infos->graphe = nv_brut;

    /* Ici, on a fait ce qu'il fallait pour s'adapter -> on refait les calculs */
    TF_fenetre( infos->fenetre, &(infos->apodisation) );
    TF_calc_apercu( infos );
    FEN_redessin( fenetre, BI_RDS_UTIL ); /* Puisqu'il y a eu modif */

    return BI_EVENEMENT_TRAITE;
  }

  if ( message->type == WM_CLOSED ) { /* On demande la fermeture */
    int methode;

    if ( message->fenetre.haut == -1 ) { /* On force donc la fermeture */
      reponse = 2;
    } else {
      reponse = BI_ERR_alerte(fenetre, &ERR_fermer_tf, VRAI);
      if (reponse == 3) return BI_EVENEMENT_TRAITE; /* On annule */
    }

    /* On n'a plus besoin de l'aperu */
    GRF_eff_isole( infos->apercu );
    infos->apercu = NULL;

    if (reponse == 2) { /* On quitte sans calculer la TF */
      /* On dlie les informations, sinon GRF_efface en-dessous rappelle ici,
	 donc on essaye de fermer 2 fois de suite la mme fentre
	 -> risques d'erreurs X ou de seg. fault, suivant les cas */
      FEN_GRF_LierGraphe( fenetre, NULL );
      FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

      /* On peut maintenant effacer tranquillement */
      GRF_efface( tf_fenetre, VRAI, VRAI ); /* On efface ce graphe */
      if (graphe != NULL) /* A pu tre effac avant ! */
	graphe->drapeaux.protege = FAUX;
      free( infos ); infos = NULL; /* On libre la mmoire */

      FEN_Detruire( fenetre ); /* On ferme la fentre */
      TT_exafs = FAUX;        /* On arrte le traitement automatique */
      return BI_EVENEMENT_FEN_FERMEE;
    }

    /* On ferme la fentre : les choses srieuses commencent... */
    FEN_Fermer( fenetre, FAUX ); /* vite un sale message de redessin */

    /* Comment calculer la TF ? */
    if ( FAUX == TF_init(infos) ) { /* Calcul annul (utilisateur) */
      FEN_Reouvrir( fenetre );
      return BI_EVENEMENT_FEN_FERMEE; /* Car les index X ont pu changer */
    }

    FEN_GRF_LierGraphe( fenetre, NULL );
    FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );
    FEN_Detruire( fenetre ); /* On ferme entirement la fentre */
    GRF_efface( tf_fenetre, VRAI, VRAI ); /* On efface ce graphe */

    if ( OBJ_choisit(DEBUT_TF, TF_RAPIDE, QUESTION) == VRAI ) { /* On veut la mthode rapide */
      methode = TF_CALC_RAPIDE;

      OBJ_MettreNombre( TF_GENERAL, RF_TF_RMIN, 0.0, 0.0 );
      GER_reproduit_filtre(RF_TF_RAPIDE, 0, 0);
    } else {
      methode = TF_CALC_NORMAL;

      OBJ_MettreNombre( TF_GENERAL, RF_TF_RMIN, OBJ_LireNombre( DEBUT_TF, TF_R_MIN, NULL ), 0.0 );
      GER_reproduit_filtre(RF_TF_NORMALE, 0, 0);
    }

    fourier = TF_CalculDirect(infos,
			      OBJ_LireNombre(DEBUT_TF, TF_R_MIN, NULL ), OBJ_LireNombre(DEBUT_TF, TF_R_MAX, NULL ), OBJ_LireNombre(DEBUT_TF, FOUR_PAS, NULL ),
			      methode,
			      OBJ_choisit(DEBUT_TF, CTF_ESTIMER_ERR, QUESTION), OBJ_choisit(DEBUT_TF, CTF_MATRICE_COV, QUESTION) );
    if ( NULL == fourier ) {
      free( infos ); infos = NULL;
      return BI_EVENEMENT_FEN_FERMEE;
    }

    /* On peut supposer que l'on a fini ici... */
    graphe->drapeaux.protege  = FAUX; /* On a fini de calculer sur ces graphes */

    /* On installe les paramtres dans l'arbre de reproduction */
    {
      extern struct infos_reproduit_filtre infos_reproduire; /* Dans Filtrage/reproduire.c */

      OBJ_MettreIntervalle( TF_GENERAL, RF_VAL_KMIN, RF_VAL_KMAX, 0, 0, infos->apodisation.x_min, infos->apodisation.x_max, 0.0, 0 );
      OBJ_MettreNombre( TF_GENERAL, TFG_TF_PARAM, infos->apodisation.parametre, 0.0 );

      infos_reproduire.type_fen_tf = infos->apodisation.type;
      if ( Autorise_Symetrique(infos->apodisation.type) ) {
	OBJ_MettreIntervalle( TF_GENERAL, RF_VAL_K1, RF_VAL_K2, 0, 0, infos->apodisation.x_1, infos->apodisation.x_2, 0.0, 0 );
	infos_reproduire.tf_symetrique = infos->apodisation.symetrique;
      }

      OBJ_MettreIntervalle( TF_GENERAL, 0, RF_TF_RMAX, TF_RF_PAS, 0,
			    0.0, OBJ_LireNombre( DEBUT_TF, TF_R_MAX, NULL ), OBJ_LireNombre( DEBUT_TF, FOUR_PAS, NULL ), 0 );
    }

    GRF_aller(fourier); /* On met ce graphe en graphe courant */
    
    if (TT_exafs == VRAI) { /* On veut faire tout l'E.X.A.F.S. */
      TFI_ouvre(fourier); /* On appelle pour la TF inverse */

      TT_exafs = FAUX; /* On a fini avec la gestion automatique de l'E.X.A.F.S. */
    }

    /* Plus besoin des infos */
    free( infos ); infos = NULL;

    return BI_EVENEMENT_FEN_FERMEE;
  }

  AVERTISSEMENT( "Message non trait [%d]", message->type );
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Gestion du clavier pour la fentre de TF -------- */

static int CLV_tf(Fenetre *fenetre, unsigned int touche, int etat)
{
  int retour;
  TF_infos *infos;
  Message message;
  unsigned long nv_fenetre, prem_visible;

  infos = (TF_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if (infos == NULL) return BI_EVENEMENT_NON_TRAITE;
  if ( etat & K_CTRL ) {
    /* Ctrl-touche : aucune combinaison -> renvoye aux menus */
    return BI_EVENEMENT_NON_TRAITE;
  }

  switch (touche)
  {
  case 'a':
  case 'A':
    retour = ARB_SimulerClic( PARAM_FOURIER, TF_APERCU, 1 );
    if ( retour != BI_EVENEMENT_NON_TRAITE )
      return retour;
    return BI_EVENEMENT_TRAITE;

  case 'k':
  case 'K':
    (void) TF_DemanderLesk( fenetre, infos );
    return BI_EVENEMENT_TRAITE;

  case 'p':
  case 'P':
    if ( Avec_Parametre(infos->apodisation.type) ) {
      retour = ARB_SimulerClic( PARAM_FOURIER, FOUR_PARAM, 1 );
      if ( retour != BI_EVENEMENT_NON_TRAITE )
	return retour; 
    }
    return BI_EVENEMENT_TRAITE; /* Trait, mais sans effet */

  case 's':
  case 'S':
    if ( Autorise_Symetrique(infos->apodisation.type) ) {
      retour = GER_tf(TF_SYMETRIQUE, 0, 0);
      if ( retour != BI_EVENEMENT_NON_TRAITE )
	return retour; 
    }
    return BI_EVENEMENT_TRAITE; /* Trait, mais sans effet */

  case ' ':
    message.type = WM_CLOSED;
    message.fenetre.application = 0;
    message.fenetre.en_plus = 0;
    message.fenetre.x = 0; message.fenetre.y = 0;
    message.fenetre.larg = 0; message.fenetre.haut = 0; /* Dans le cas o != 0, fermeture force... */
    return (fenetre->message)(fenetre, &message);

  case Fleche_Haut:
  case Fleche_Bas:
    /* Prparation de l'appel */
    nv_fenetre   = (unsigned long) infos->apodisation.type;
    prem_visible = (unsigned long) infos->num_depart;

    /* Appel  clavier du dfileur */
    if ( touche == Fleche_Haut ) {
      ARB_GererDefileur_T( PARAM_FOURIER, FOUR_FEN_1, FOUR_FEN_4, 0,
			   FOUR_FEN_AVANT, FOUR_FEN_FOND, FOUR_FEN_BOUTON, FOUR_FEN_APRES,
			   NOMBRE_FENETRES, Types_fenetres, NULL, &nv_fenetre, &prem_visible,
			   ROOT, +1 );
    } else {
      ARB_GererDefileur_T( PARAM_FOURIER, FOUR_FEN_1, FOUR_FEN_4, 0,
			   FOUR_FEN_AVANT, FOUR_FEN_FOND, FOUR_FEN_BOUTON, FOUR_FEN_APRES,
			   NOMBRE_FENETRES, Types_fenetres, NULL, &nv_fenetre, &prem_visible,
			   ROOT, -1 );
    }

    /* Mmorisation des rsultats de l'appel */
    infos->num_depart = prem_visible;

    /* Mise  jour de l'affichage en fonction de la fenetre choisie */
    if ( VRAI == TF_InstallerFenetre( infos, nv_fenetre ) ) {
      /* On redessine tout a */
      FEN_redessin( fenetre, BI_RDS_UTIL );
    }

    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* --------- Gestion d'un clic sur la fentre de TF --------------------------------------- */

static int CLC_tf(Fenetre *fenetre, int bouton, int n_clic __attribute__((__unused__)), int etat __attribute__((__unused__)), int x, int y __attribute__((__unused__)))
{
  TF_infos *infos;
  double x_echelle,x_clic;

  infos = (TF_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return BI_EVENEMENT_TRAITE;

  x_echelle = ( (double) fenetre->w_larg ) / (infos->graphe->x_max - infos->graphe->x_min);
  x_clic  = x - fenetre->w_x;
  x_clic /= x_echelle;
  x_clic += infos->graphe->x_min;

  if (bouton == 1) { /* Bouton gauche */
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) ) { /* On appuie sur shift en mme temps */
      if ( Autorise_Symetrique( infos->apodisation.type ) )
	infos->apodisation.x_1 = x_clic;
      else
	infos->apodisation.x_min = x_clic;
    } else
      infos->apodisation.x_min = x_clic;
  } else { /* Bouton droit */
    if ( (etat & K_LSHIFT) || (etat & K_RSHIFT) ) { /* On appuie sur shift en mme temps */
      if ( Autorise_Symetrique( infos->apodisation.type ) && (infos->apodisation.symetrique == FAUX) )
	infos->apodisation.x_2 = x_clic;
      else
	infos->apodisation.x_max = x_clic;
    } else
      infos->apodisation.x_max = x_clic;
  }

  met_les_k(infos);

  TF_fenetre(infos->fenetre, &(infos->apodisation) );
  TF_calc_apercu(infos);
  FEN_redessin(fenetre, BI_RDS_UTIL); /* On redessine : fentre a chang... */
  return BI_EVENEMENT_TRAITE;
}

/* -------- Prparation de la TF (choix de la fentre principalement) ------------------------- */

static int GER_tf(unsigned int num_objet, int x_souris __attribute__((__unused__)), int y_souris)
{
  Fenetre *fen_cour;
  Graphe *fenetre;
  TF_infos *infos;
  int reponse;
  struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, {0}, {0}, {0} };

  fen_cour = ARB_QuelleFenetre(arbres + PARAM_FOURIER, FAUX); /* Fentre actuelle */
  infos   = (TF_infos *) FEN_LireComplement( fen_cour, FEN_COMPLEMENT_2 );
  if ( NULL == infos ) return BI_EVENEMENT_TRAITE;
  fenetre = infos->fenetre;

  if ( (num_objet >= FOUR_FEN_1) && (num_objet <= FOUR_FEN_4) ) {
    unsigned long nv_fenetre = infos->apodisation.type; /* Nouvelle fentre choisie */
    unsigned long prem_visible = infos->num_depart;

    /* On repre quel objet a t cliqu */
    ARB_GererDefileur_T(PARAM_FOURIER, FOUR_FEN_1, FOUR_FEN_4, 0,
			FOUR_FEN_AVANT, FOUR_FEN_FOND, FOUR_FEN_BOUTON, FOUR_FEN_APRES,
			NOMBRE_FENETRES, Types_fenetres, NULL, &nv_fenetre, &prem_visible, num_objet, y_souris);

    /* Mise  jour de l'affichage en fonction de la fenetre choisie */
    if ( VRAI == TF_InstallerFenetre( infos, nv_fenetre ) ) {
      /* On redessine tout a */
      FEN_redessin( fen_cour, BI_RDS_UTIL );
    }

    return BI_EVENEMENT_TRAITE;
  }

  switch (num_objet)
  {
  case FOUR_FEN_AVANT:      /* Dfilement du slecteur de fentres */
  case FOUR_FEN_APRES:
  case FOUR_FEN_BOUTON - 1:
  case FOUR_FEN_BOUTON:
    {
      unsigned long nv_fenetre = infos->apodisation.type,
	prem_visible = infos->num_depart;

      ARB_GererDefileur_T(PARAM_FOURIER, FOUR_FEN_1, FOUR_FEN_4, 0,
			  FOUR_FEN_AVANT, FOUR_FEN_FOND, FOUR_FEN_BOUTON, FOUR_FEN_APRES,
			  NOMBRE_FENETRES, Types_fenetres, NULL, &nv_fenetre, &prem_visible, num_objet, y_souris);
      infos->num_depart = (int) prem_visible; /* Il n'y aura jamais max_int fentres... */
    }
    return BI_EVENEMENT_TRAITE;

  case FOUR_KMIN: /* Demande des valeurs de position de la fenetre  l'utilisateur */
  case PARAM_TF_K1:
  case PARAM_TF_K2:
  case FOUR_KMAX:
    (void) TF_DemanderLesk( fen_cour, infos );
    return BI_EVENEMENT_TRAITE;

  case FOUR_PARAM:
    nv_valeur.proprietes = 0;
    nv_valeur.texte = "p =";
    OPTION_A_FAIRE( "Mettre les limites en fonction du type de fentre" );
    nv_valeur.valeur_initiale.reel = infos->apodisation.parametre;

    reponse = BI_DemandeValeur( fen_cour, &Q_ValeurParametre, &nv_valeur );
    if ( Q_ValeurParametre.bouton_annuler == (unsigned int) reponse ) return VRAI; /* On annule */

    infos->apodisation.parametre = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( PARAM_FOURIER, FOUR_PARAM, infos->apodisation.parametre, 0.0 );

    TF_fenetre(fenetre, &(infos->apodisation) ); /* On calcule la fentre */
    TF_calc_apercu(infos);
    FEN_redessin(fen_cour, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case TF_SYMETRIQUE:
    infos->apodisation.symetrique = OBJ_choisit(PARAM_FOURIER, TF_SYMETRIQUE, ECHANGE);
    OBJ_active(PARAM_FOURIER, PARAM_TF_K2, !(infos->apodisation.symetrique)); /* Aucun sens si fentre symtrique... */
    TF_fenetre(fenetre, &(infos->apodisation) ); /* On calcule la fentre */
    TF_calc_apercu(infos);
    FEN_redessin(fen_cour, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case TF_APERCU:    
    if (OBJ_choisit(PARAM_FOURIER, TF_APERCU, ECHANGE) == FAUX) { /* Plus d'aperu */
      Graphe *graphe = infos->apercu;
      infos->apercu = NULL;
      GRF_efface(graphe, VRAI, VRAI);
    } else {
      TF_prep_apercu(infos);
    }
    FEN_redessin(fen_cour, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Demander les valeurs dfinissant la fenetre -------- */

static int TF_DemanderLesk(Fenetre *fenetre, TF_infos *infos)
{
  const char *unite_X = NULL;
  int reponse;
  struct infos_demande nv_valeurs[ 4 ] = {
    { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } },
    { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } } };
  unsigned int i, nbr_valeurs_demandees;

  if ( NULL == infos->graphe ) return -1;

  /* Initialisation des valeurs */
  for ( i = 0 ; i < 4 ; i++ ) {
    nv_valeurs[ i ].proprietes = BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeurs[ i ].min = 0.0;
    nv_valeurs[ i ].max = infos->graphe->x_max;
  }

  /* Installation de l'unit, ventuellement */
  unite_X = GRF_QuelleUnite( infos->graphe, AXE_X );
  if ( unite_X != NULL ) {
    for ( i = 0 ; i < 4 ; i++ ) {
      nv_valeurs[ i ].proprietes |= BI_EDITABLE_UNITE;
      nv_valeurs[ i ].unite = unite_X;
    }
#ifdef _DEVELOPPE_
  } else {
    (void) fprintf(stderr, "%s,%s l.%u : Graphe sans unit [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, infos->graphe->nom);
#endif
  }

  /* Borne infrieure */
  nv_valeurs[ 0 ].texte = "kmin =";
  nv_valeurs[ 0 ].valeur_initiale.reel = infos->apodisation.x_min;

  /* Borne infrieure du plateau, s'il existe */
  if ( VRAI == Autorise_Symetrique( infos->apodisation.type ) ) {
    nv_valeurs[ 1 ].texte = "k1 =";
    nv_valeurs[ 1 ].valeur_initiale.reel = infos->apodisation.x_1;

    /* Faut-il une borne suprieure du plateau ? */
    if ( VRAI == infos->apodisation.symetrique ) {
      /* Non : automatique par symtrie */
      nbr_valeurs_demandees = 3;
    } else {
      /* Oui */
      nv_valeurs[ 2 ].texte = "k2 =";
      nv_valeurs[ 2 ].valeur_initiale.reel = infos->apodisation.x_2;

      nbr_valeurs_demandees = 4;
    }
  } else {
    nbr_valeurs_demandees = 2;
  }

  /* Borne suprieure */
  nv_valeurs[ nbr_valeurs_demandees - 1 ].texte = "kmax =";
  nv_valeurs[ nbr_valeurs_demandees - 1 ].valeur_initiale.reel = infos->apodisation.x_max;

  /* On demande les valeurs  l'utilisateur */
  reponse = BI_DemandeMultiple( fenetre, &Q_ValeursK, nbr_valeurs_demandees, nv_valeurs );
  if ( reponse != 1 ) return +1; /* On annule ou il y a eu une erreur */

  /* On mmorise les valeurs */
  infos->apodisation.x_min = nv_valeurs[ 0 ].valeur_donnee.reel;
  if ( VRAI == Autorise_Symetrique( infos->apodisation.type ) ) {
    infos->apodisation.x_1 = nv_valeurs[ 1 ].valeur_donnee.reel;
    if ( VRAI == infos->apodisation.symetrique ) {
      infos->apodisation.x_max = nv_valeurs[ 2 ].valeur_donnee.reel;
    } else {
      infos->apodisation.x_2   = nv_valeurs[ 2 ].valeur_donnee.reel;
      infos->apodisation.x_max = nv_valeurs[ 3 ].valeur_donnee.reel;
    }
  } else {
    infos->apodisation.x_max = nv_valeurs[ 1 ].valeur_donnee.reel;
  }

  /* On controle que les valeurs sont dans le bon ordre */
  met_les_k( infos );

  /* On recalcule et on redessine */
  TF_fenetre( infos->fenetre, &(infos->apodisation) ); /* On calcule la fentre */
  TF_calc_apercu( infos );
  FEN_redessin( fenetre, BI_RDS_UTIL );

  /* Termin sans problme */
  return 0;
}

/* -------- Quand on a choisi une fonction d'apodisation, adapte la fenetre en consquence --------
    Renvoie VRAI s'il faut redessiner
            FAUX si aucun changement n'a t effactu
 */

static boolean TF_InstallerFenetre(TF_infos *infos, int type)
{
  boolean lim_internes, av_parametre;

  /* Est-il ncessaire de changer le type ? */
  if ( type == infos->apodisation.type ) {
    return FAUX;
  }

  /* On installe le nouveau type de fenetre */
  infos->apodisation.type = type;

  if( infos->apodisation.type != PAS_DE_FENETRE ) {
    OBJ_active( PARAM_FOURIER, FOUR_KMIN, FORCE_OUI );
    OBJ_active( PARAM_FOURIER, FOUR_KMAX, FORCE_OUI );
  } else {
    OBJ_active( PARAM_FOURIER, FOUR_KMIN, FORCE_NON );
    OBJ_active( PARAM_FOURIER, FOUR_KMAX, FORCE_NON );
  }

  /* La fenetre a-t-elle des limites internes ? */
  lim_internes = Autorise_Symetrique( infos->apodisation.type ); /* VRAI -> asymtrie possible */
  OBJ_active( PARAM_FOURIER, TF_SYMETRIQUE, lim_internes );
  OBJ_active( PARAM_FOURIER, PARAM_TF_K1  , lim_internes );

  /* Si oui, force-t-on une fenetre symtrique ? */
  if ( VRAI == lim_internes ) {
    if ( OBJ_choisit( PARAM_FOURIER, TF_SYMETRIQUE, QUESTION ) == VRAI ) {
      OBJ_active( PARAM_FOURIER, PARAM_TF_K2, FORCE_NON );
    } else {
      OBJ_active( PARAM_FOURIER, PARAM_TF_K2, FORCE_OUI );
    }
  } else {
    OBJ_active( PARAM_FOURIER, PARAM_TF_K2, FORCE_NON );
  }

  /* La fenetre demande-t-elle un paramtre ? */
  av_parametre = Avec_Parametre( infos->apodisation.type );
  OBJ_active( PARAM_FOURIER, FOUR_PARAM, av_parametre ); /* Seule fentre avec un paramtre... */

  /* On recalcule la fonction d'apodisation */
  TF_fenetre( infos->fenetre, &(infos->apodisation) );

  /* On recalcul l'aperu si ncessaire */
  TF_calc_apercu(infos);

  return VRAI;
}

/* -------- Met les valeurs des k aprs tri (pour la cohrence) -------- */

static void met_les_k(TF_infos *infos)
{
  if (infos->apodisation.x_min > infos->apodisation.x_max) { /* On vrifie la cohrence */
    double k = infos->apodisation.x_min;
    infos->apodisation.x_min = infos->apodisation.x_max;
    infos->apodisation.x_max = k;
  }
  if ( Autorise_Symetrique( infos->apodisation.type ) ) {
    if (infos->apodisation.symetrique) {
      if ( infos->apodisation.x_1 < infos->apodisation.x_min ) {
	double k = infos->apodisation.x_min;
	infos->apodisation.x_min = infos->apodisation.x_1;
	infos->apodisation.x_1 = k;
      }
      if ( infos->apodisation.x_1 > infos->apodisation.x_max ) {
	double k = infos->apodisation.x_max;
	infos->apodisation.x_max = infos->apodisation.x_1;
	infos->apodisation.x_1 = k;
      }
    } else {
      double k[4];
      int i, j;

      k[0] = infos->apodisation.x_min;
      k[1] = infos->apodisation.x_1;
      k[2] = infos->apodisation.x_2;
      k[3] = infos->apodisation.x_max;

      for (i = 0 ; i < 4 ; i++)
	for (j = i ; j < 4 ; j++)
	  if ( k[i] > k[j] ) {
	    double l = k[i];
	    k[i] = k[j]; k[j] = l;
	  }

      infos->apodisation.x_min = k[0];
      infos->apodisation.x_1   = k[1];
      infos->apodisation.x_2   = k[2];
      infos->apodisation.x_max = k[3];
    }
  }

  OBJ_MettreNombre( PARAM_FOURIER, FOUR_KMIN, infos->apodisation.x_min, 0.0 );
  if ( Autorise_Symetrique( infos->apodisation.type ) ) {
    OBJ_MettreNombre( PARAM_FOURIER, PARAM_TF_K1, infos->apodisation.x_1, 0.0 );
    if (infos->apodisation.symetrique == FAUX)
      OBJ_MettreNombre( PARAM_FOURIER, PARAM_TF_K2, infos->apodisation.x_2, 0.0);
  }
  OBJ_MettreNombre( PARAM_FOURIER, FOUR_KMAX, infos->apodisation.x_max, 0.0 );
}

/* ~~~~~~~~ Calcul de l'aperu de la TF ~~~~~~~~
   L'aperu est calcul entre 0 et 10 A, pour un pas de 0,1
   Il est calcul par la mthode brutale.
 */

static boolean TF_calc_apercu(TF_infos *infos_tf)
{
  Graphe *source;

  ENTREE_FONCTION("TF_calc_apercu", "infos_tf=%p", infos_tf);

  if (infos_tf == NULL) {
    SORTIE_FONCTION("TF_calc_apercu", "ERREUR --- %d [infos_tf == NULL]", FAUX);
    return FAUX;
  }
  if (infos_tf->apercu  == NULL) {
    SORTIE_FONCTION("TF_calc_apercu", "ERREUR --- %d [infos_tf->apercu == NULL]", FAUX);
    return FAUX;
  }
  if (infos_tf->graphe  == NULL) {
    SORTIE_FONCTION("TF_calc_apercu", "ERREUR --- %d [infos_tf->graphe == NULL]", FAUX);
    return FAUX;
  }
  if (infos_tf->fenetre == NULL) {
    SORTIE_FONCTION("TF_calc_apercu", "ERREUR --- %d [infos_tf->fenetre == NULL]", FAUX);
    return FAUX;
  }

  /* On cre un graphe [isol] qui contient la fentre que l'on va multiplier par le signal brut */
  source = GRF_Copier( infos_tf->graphe, NULL, VRAI );
  if (source == NULL) {
    (void) fprintf(stderr, "TF_calc_apercu : erreur dans GRF_Copier()\n");
    SORTIE_FONCTION("TF_calc_apercu", "ERREUR --- %d", FAUX);
    return FAUX; /* Impossible d'avoir la source */
  }

  /* Produit par la fonction d'apodisation */
  GRF_ProduitGraphes( source, infos_tf->fenetre, VRAI, FAUX, NULL );

  /* On fait le calcul maintenant */
  TF_simple( source, infos_tf->apercu );

  /* Plus besoin de la source */
  GRF_eff_isole(source);

  OBJ_MettreNombre( PARAM_FOURIER, PARAM_TF_MAXMOD, infos_tf->apercu->y_max, 0.0 );

  SORTIE_FONCTION("TF_calc_apercu", "VRAI [%d]", VRAI);
  return VRAI; /* Tout s'est bien pass */
}

/* -------- Prparation de l'aperu pour la TF -------- */

static boolean TF_prep_apercu(TF_infos *infos_tf)
{
  register double *x;
  unsigned long i, nbr_pts;

  ENTREE_FONCTION("TF_prep_apercu", "infos_tf=%p", infos_tf);

  if (infos_tf == NULL) { 
    SORTIE_FONCTION("TF_prep_apercu", "FAUX=%d [infos_tf = NULL]", FAUX);
    return FAUX;
  }

  if (OBJ_choisit(PARAM_FOURIER, TF_APERCU, QUESTION) == FAUX) { /* On ne veut pas d'aperu */
    Graphe *graphe = infos_tf->apercu;

    DANS_FONCTION("On veut enlever l'aperu", 0);

    GRF_efface(graphe, VRAI, VRAI);
    infos_tf->apercu = NULL;
    OBJ_active(PARAM_FOURIER, PARAM_TF_MAXMOD, FORCE_NON);

    SORTIE_FONCTION("TF_prep_apercu", "VRAI=%d", VRAI);
    return VRAI; /* Pas de problme particulier */
  }

  DANS_FONCTION("On veut mettre l'aperu", 0);

  /* On veut un aperu : de 0  Rmax A par pas de _pas_ donc _pas_ * Rmax + 1 points */
  nbr_pts= (unsigned long) ceil( ( TF_MAX_APERCU / TF_PAS_APERCU ) + 1.0 );
  infos_tf->apercu = GRF_nv_isole( nbr_pts, GR_INTERNE, GRF_FCT_COMPLEXE );
  if (infos_tf->apercu == NULL) {
    (void) fprintf(stderr, "TF_prep_apercu : Impossible de prparer l'aperu\n");
    OBJ_choisit(PARAM_FOURIER, TF_APERCU, FORCE_NON);
    OBJ_active(PARAM_FOURIER, PARAM_TF_MAXMOD, FORCE_NON);

    SORTIE_FONCTION("TF_prep_apercu", "FAUX=%d [erreur dans GRF_creer]", FAUX);
    return FAUX;
  }

  /* On prpare les X */
  for (i = 0, x = infos_tf->apercu->x ; i < nbr_pts ; i++, x++) {
    *x = 0.0 + TF_PAS_APERCU * (double) i;
  }
  infos_tf->apercu->x_min = 0.0;
  infos_tf->apercu->x_max = TF_MAX_APERCU;

  TF_calc_apercu(infos_tf);
  OBJ_active(PARAM_FOURIER, PARAM_TF_MAXMOD, FORCE_OUI);

  SORTIE_FONCTION("TF_prep_apercu", "VRAI=%d", VRAI);
  return VRAI;
}

/* ~~~~~~~~ Paramtrages du calcul de la transforme de Fourier ~~~~~~~~ */

typedef struct {
  GRF_liste *amplitude; /* Choix de la correction d'amplitude */
  GRF_liste *phase;     /* Choix de la correction des phases  */
  boolean err_possible; /* VRAI si le calcul des erreurs est possible */
} Infos_Prepare_TF;

/* -------- Initialisation du graphe qui contiendra la TF ------------------------\
 |   Renvoie FAUX si on annule, VRAI sinon                                        |
 |                                                                                |
 | infos_tf     : informations ncessaires au calcul de la transforme de Fourier |
 | r_min, r_max : bornes de calcul de la transforme de Fourier                   |
 | pas          : pas  utiliser pour ce calcul                                   |
 \------------------------------------------------------------------------------- */

boolean TF_init(TF_infos *infos_tf)
{
  int retour;
  boolean etat;
  Infos_Prepare_TF infos;

  if ( NULL == infos_tf ) return FAUX;
  if ( NULL == infos_tf->graphe ) return FAUX;

  if ( arbres[DEBUT_TF].drapeaux.prepare == FAUX )
    prepare_debut_tf();

  infos.err_possible = (infos_tf->graphe->erreur_y != NULL);

 /* Prparation de la liste des graphes d'amplitudes */
  infos.amplitude = GRF_par_type(GR_EXAFSAMP, GRF_LISTE_TYPES_STRICT | GRF_LISTE_TYPES_MULTISERIE, NULL);
  OBJ_active(DEBUT_TF, TF_COR_AMP, infos.amplitude != NULL);
  if (infos.amplitude == NULL) /* Prparation n'a pu tre faite */
    OBJ_choisit(DEBUT_TF, TF_COR_AMP, FORCE_NON);

  etat = OBJ_choisit(DEBUT_TF, TF_COR_AMP, QUESTION);
  OBJ_cache(DEBUT_TF, TF_COR_AMP_1 - 1, !etat);
  OBJ_cache(DEBUT_TF, TF_COR_AMP_AV, !etat);
  OBJ_cache(DEBUT_TF, TF_COR_AMP_FOND, !etat);
  OBJ_cache(DEBUT_TF, TF_COR_AMP_AP, !etat);

  /* Prparation de la liste des graphes de phase */
  infos.phase = GRF_par_type(GR_EXAFSPHA, GRF_LISTE_TYPES_STRICT | GRF_LISTE_TYPES_MULTISERIE, NULL);
  OBJ_active(DEBUT_TF, TF_COR_PHASE, infos.phase != NULL);
  if (infos.phase == NULL)
    OBJ_choisit(DEBUT_TF, TF_COR_PHASE, FORCE_NON);

  etat = OBJ_choisit(DEBUT_TF, TF_COR_PHASE, QUESTION);
  OBJ_cache(DEBUT_TF, TF_COR_PHA_1 - 1, !etat);
  OBJ_cache(DEBUT_TF, TF_COR_PHA_AV, !etat);
  OBJ_cache(DEBUT_TF, TF_COR_PHA_FOND, !etat);
  OBJ_cache(DEBUT_TF, TF_COR_PHA_AP, !etat);

  GRF_affiche(infos.amplitude, DEBUT_TF, TF_COR_AMP_1, TF_COR_AMP_4, 15,
	      TF_COR_AMP_AV, TF_COR_AMP_AP, TF_COR_AMP_FOND, TF_COR_AMP_BOUT);

  GRF_affiche(infos.phase, DEBUT_TF, TF_COR_PHA_1, TF_COR_PHA_4, 15,
	      TF_COR_PHA_AV, TF_COR_PHA_AP, TF_COR_PHA_FOND, TF_COR_PHA_BOUT);

  etat = OBJ_active(DEBUT_TF, CTF_ESTIMER_ERR, infos.err_possible == VRAI );
  if (etat == FAUX) {
    OBJ_choisit(DEBUT_TF, CTF_ESTIMER_ERR, FORCE_NON);
    OBJ_choisit(DEBUT_TF, CTF_MATRICE_COV, FORCE_NON); /* Pas de sigma -> pas de matrice de corrlation */
  }
  if ( OBJ_choisit(DEBUT_TF, TF_NORMALE, QUESTION) == FAUX )
    etat = OBJ_active(DEBUT_TF, CTF_ESTIMER_ERR, FORCE_NON);
  OBJ_active(DEBUT_TF, CTF_MATRICE_COV, etat);
  /* Estimation des erreurs uniquement s'il y a des erreurs sur le graphe original...
     et si l'on calcule par la mthode des rectangles */

  ARB_MettreComplement( DEBUT_TF, &infos ); /* Infos complmentaires pour l'arbre... */
  retour = ARB_Bloquant( DEBUT_TF, titre_arbre_tf, principale, 0 );
  ARB_MettreComplement( DEBUT_TF, NULL );
  if (retour == FOUR_ANNULE) { /* Finalement, on reste... */
    free(infos.phase);
    free(infos.amplitude);
    return FAUX;
  }

  if ( OBJ_choisit(DEBUT_TF, TF_COR_AMP, QUESTION) == VRAI ) { /* Correction d'amplitude */
    infos_tf->amplitude = infos.amplitude->graphes[ infos.amplitude->num_select ];
  } else {
    infos_tf->amplitude = NULL;
  }

  if ( OBJ_choisit(DEBUT_TF, TF_COR_PHASE, QUESTION) == VRAI ) { /* Correction de phase */
    infos_tf->phase = infos.phase->graphes[ infos.phase->num_select ];
  } else {
    infos_tf->phase = NULL;
  }

  free(infos_tf->phase);     /* On libre : on n'a plus besoin de a... */
  free(infos_tf->amplitude);
  
  return VRAI; /* Prparation termine */	
}

/* -------- Lancement de la TF (choix des graphes de correction, entre autres ----------------------\
 |   Gestion de l'arbre GEM de paramtrage des paramtres de calcul autres que la fentre           |
 \------------------------------------------------------------------------------------------------- */

static int GER_debTF(unsigned int num_objet, int x, int y)
{
  Infos_Prepare_TF *infos;
  boolean etat;

  infos = (Infos_Prepare_TF *) ARB_LireComplement( DEBUT_TF );
  if ( (num_objet >= TF_COR_AMP_1) && (num_objet <= TF_COR_AMP_AP) ) {
    (void) GRF_selecteur( infos->amplitude, DEBUT_TF, TF_COR_AMP_1, TF_COR_AMP_4, 15,
			  TF_COR_AMP_AV, TF_COR_AMP_AP, TF_COR_AMP_FOND, TF_COR_AMP_BOUT,
			  num_objet, x, y );
    return BI_EVENEMENT_TRAITE;
  }

  if ( (num_objet >= TF_COR_PHA_1) && (num_objet <= TF_COR_PHA_AP) ) {
    (void) GRF_selecteur( infos->phase, DEBUT_TF, TF_COR_PHA_1, TF_COR_PHA_4, 15,
			  TF_COR_PHA_AV, TF_COR_PHA_AP, TF_COR_PHA_FOND, TF_COR_PHA_BOUT,
			  num_objet, x, y );
    return BI_EVENEMENT_TRAITE;
  }

  switch (num_objet)
  {
  case TF_COR_AMP: /* Activation ou non de la correction d'amplitude */
    etat = OBJ_choisit(DEBUT_TF, num_objet, ECHANGE);
    OBJ_cache(DEBUT_TF, TF_COR_AMP_1 - 1, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_AMP_AV, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_AMP_FOND, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_AMP_AP, !etat);
    return BI_EVENEMENT_TRAITE;

  case TF_COR_PHASE: /* Activation ou non de la correction de phase */
    etat = OBJ_choisit(DEBUT_TF, num_objet, ECHANGE);
    OBJ_cache(DEBUT_TF, TF_COR_PHA_1 - 1, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_PHA_AV, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_PHA_FOND, !etat);
    OBJ_cache(DEBUT_TF, TF_COR_PHA_AP, !etat);
    return BI_EVENEMENT_TRAITE;

  case TF_RAPIDE:
    etat = OBJ_choisit(DEBUT_TF, num_objet, ECHANGE);
    OBJ_active( DEBUT_TF, TF_R_MIN, FORCE_NON); /* Pas de choix pour Rmin... */
    OBJ_active(DEBUT_TF, CTF_ESTIMER_ERR, FORCE_NON); /* Estimation des erreurs impossible */
    return BI_EVENEMENT_TRAITE;

  case TF_NORMALE:
    etat = OBJ_choisit(DEBUT_TF, num_objet, ECHANGE);
    OBJ_active( DEBUT_TF, TF_R_MIN, FORCE_OUI); /* Choix de Rmin possible */
    OBJ_active(DEBUT_TF, CTF_ESTIMER_ERR, infos->err_possible == VRAI); /* Estimation des erreurs possible, s'il y en a ! */
    return BI_EVENEMENT_TRAITE;

  case CTF_ESTIMER_ERR:
    etat = OBJ_choisit(DEBUT_TF, num_objet, ECHANGE);
    OBJ_active(DEBUT_TF, CTF_MATRICE_COV, etat);
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Prparation des arbres de paramtrage de la TF directe -------- */

static int prepare_param_tf(unsigned int numero_arbre __attribute__((__unused__)))
{
  if ( FAUX == arbres[ PARAM_FOURIER ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( PARAM_FOURIER, "menu_filtre.html#tf", GER_tf, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -1;
    }
  }

  /* On met les boutons en plus joli */
  OBJ_agrandit(PARAM_FOURIER, ROOT, 0, +3);

  ARB_EnjoliverDefileur( PARAM_FOURIER, PTF_FND_DEFILEUR, FOUR_FEN_1, FOUR_FEN_4,
			 FOUR_FEN_AVANT, FOUR_FEN_FOND, FOUR_FEN_BOUTON, FOUR_FEN_APRES, ARB_ASCENSEUR_DROITE );

  OBJ_deplace(PARAM_FOURIER, TF_APERCU , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_FOURIER, TF_SYMETRIQUE, 0, +3, DEPLACE_PIXELS);

  OBJ_deplace(PARAM_FOURIER, PARAM_TF_K1, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_FOURIER, PARAM_TF_K2, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace(PARAM_FOURIER, FOUR_KMAX, 0, +3, DEPLACE_PIXELS);

  /* On met les valeurs par dfaut */
  OBJ_MettreNombre( PARAM_FOURIER, FOUR_PARAM, 1.5, 0.0 ); /* Associ  la fentre de Kaiser-Bessel */

#ifdef AVEC_TRADUCTION_ /* Traductions si ncessaire */
  met_chaine(PARAM_FOURIER, FOUR_PARAM, 0, txt_val_param, TXT_REMPLACE);
  met_chaine(PARAM_FOURIER, TF_SYMETRIQUE, 0, txt_symetrique, TXT_REMPLACE);
  met_chaine(PARAM_FOURIER, TF_APERCU, 0, txt_apercu, TXT_REMPLACE);
#endif

  arbres[PARAM_FOURIER].drapeaux.prepare = VRAI;
  return 0;
}

static void prepare_debut_tf(void)
{
  if ( FAUX == arbres[ DEBUT_TF ].drapeaux.init ) {
    int retour;

    retour = ARB_Initialiser( DEBUT_TF, "menu_filtre.html#tf_2", GER_debTF, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }
  }

  OBJ_deplace ( DEBUT_TF, TF_R_MAX, 0, 1, DEPLACE_PIXELS);

  OBJ_deplace (DEBUT_TF, TF_COR_AMP     ,  0, -1, DEPLACE_PIXELS);
  OBJ_agrandit(DEBUT_TF, TF_COR_AMP     , +1,  0);
  ARB_EnjoliverDefileur( DEBUT_TF, ROOT, TF_COR_AMP_1, TF_COR_AMP_4,
			 TF_COR_AMP_AV, TF_COR_AMP_FOND, TF_COR_AMP_BOUT, TF_COR_AMP_AP, ARB_ASCENSEUR_DROITE );

  OBJ_deplace (DEBUT_TF, TF_COR_PHASE   ,  0, -1, DEPLACE_PIXELS);
  OBJ_agrandit(DEBUT_TF, TF_COR_PHASE   , +1,  0);
  ARB_EnjoliverDefileur( DEBUT_TF, ROOT, TF_COR_PHA_1, TF_COR_PHA_4,
			 TF_COR_PHA_AV, TF_COR_PHA_FOND, TF_COR_PHA_BOUT, TF_COR_PHA_AP, ARB_ASCENSEUR_DROITE );

  OBJ_deplace (DEBUT_TF, TF_MIT, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace (DEBUT_TF, TF_NORMALE, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace (DEBUT_TF, CTF_MATRICE_COV, 0, 1, DEPLACE_PIXELS);

#ifdef AVEC_TRADUCTION_
  met_chaine(DEBUT_TF, TF_R_MIN, 0, txt_Rmin, TXT_REMPLACE);
  met_chaine(DEBUT_TF, TF_R_MAX, 0, txt_Rmax, TXT_REMPLACE);
  met_chaine(DEBUT_TF, FOUR_PAS, 0, txt_pas, TXT_REMPLACE);
  met_chaine(DEBUT_TF, FOUR_OK, 0, txt_calculer, TXT_REMPLACE);
  met_chaine(DEBUT_TF, TF_COR_AMP, 0, txt_amplitude, TXT_REMPLACE);
  met_chaine(DEBUT_TF, TF_COR_PHASE, 0, txt_phase, TXT_REMPLACE);
  met_chaine(DEBUT_TF, TF_RAPIDE, 0, txt_rapide, TXT_REMPLACE);
  met_chaine(DEBUT_TF, TF_NORMALE, 0, txt_normale, TXT_REMPLACE);
  met_chaine(DEBUT_TF, CTF_ESTIMER_ERR, 0, txt_av_erreur, TXT_REMPLACE);
  met_chaine(DEBUT_TF, CTF_MATRICE_COV, 0, txt_covariance, TXT_REMPLACE);
#endif

#ifndef _DEVELOPPE_ /* Ne fonctionne pas encore... */
  OBJ_active(DEBUT_TF, TF_MIT, FORCE_NON);
#endif

  OBJ_MettreIntervalle( DEBUT_TF, TF_R_MIN, TF_R_MAX, FOUR_PAS, ROOT, 0.00, 10.0, 0.05, 0 );

  arbres[DEBUT_TF].drapeaux.prepare = VRAI;
}
