/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2002, Emmanuel CURIS
   Cration : 27 fvrier 2002

   Gestion gnrique des graphes

   Controles sur l'utilisation d'un graphe
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"

#include "grf_controles.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static boolean GRF_LegendesHomogenes(Graphe **graphes, unsigned long nbr_graphes, unsigned int axe);
static boolean GRF_UnitesHomogenes  (Graphe **graphes, unsigned long nbr_graphes, unsigned int axe);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Contrle de validit d'une unit ------------------------------ *
     Renvoie VRAI si l'unit convient
             FAUX si l'unit ne convient pas [ou en cas d'erreur]

  graphe         : le graphe que l'on veut controler [NULL -> FAUX]
  unite_attendue : quelle unit doit-on avoir EXACTEMENT ?
  axe            : sur quel axe doit-on avoir cette unit ?
  convertir      : VRAI -> si multiple, conversion des valeurs du graphe
 * ------------------------------------------------------------------------ */
 
boolean GRF_ControlerUnite(Graphe *graphe, const char *unite_attendue, int axe, boolean convertir)
{
  const char *unite_graphe;
  double norm_attendu, norm_graphe, facteur;
  int code_axe = 0, reponse;
  int type_attendu, type_graphe;

  if ( NULL == graphe ) {
    /* Mthode valide pour controler qu'un graphe existe
     -> ni message, ni arret anticip */
    return FAUX;
  }
  if ( NULL == unite_attendue ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune unit pour contrle !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return FAUX;
  }

  /* Pour affichages : quel axe est en jeu ? */
  if ( AXE_X == axe ) {
    code_axe = 'X';
  } else if ( AXE_Y == axe ) {
    code_axe = 'Y';
  } else if ( AXE_Z == axe ) {
    code_axe = 'Z';
  } else {
    AVERTISSEMENT( "Axe inconnu ou multiple [%d]", axe );
    return FAUX;
  }

  /* Quelle est l'unit du graphe ? */
  unite_graphe = GRF_QuelleUnite( graphe, axe );
  if ( NULL == unite_graphe ) {
    reponse = BI_ERR_alerte( principale, &Q_UniteManquante, VRAI, code_axe, unite_attendue );
    if ( 2 == reponse )		/* Pas d'attribution automatique de l'unit */
      return FAUX;

    /* On attribue l'unit au graphe */
    GRF_ChangerUnite( graphe, axe, unite_attendue );

    /* Et l'unit est donc forcment celle attendue */
    return VRAI;
  }

  /* Les units sont-elles identiques ? */
  if ( 0 == strcmp( unite_attendue, unite_graphe ) ) {
    /* Units identiques : tout va bien */
    return VRAI;
  }

  /* Units diffrentes : peut-on faire une conversion ? */
  norm_attendu = TXT_NormaliserUnite( unite_attendue, FAUX, &type_attendu, NULL );
  norm_graphe  = TXT_NormaliserUnite( unite_graphe  , FAUX, &type_graphe , NULL );

  if ( type_attendu != type_graphe ) {
    /* Gros problme d'homognit */
    (void) BI_ERR_alerte( principale, &ERR_UniteIncorrecte, VRAI, code_axe, unite_attendue, unite_graphe );

    return FAUX;
  }

  /* Facteurs de conversion plausibles ? */
  if ( ( 0.0 == norm_attendu ) || ( 0.0 == norm_graphe ) ) {
    /* Un des facteurs est absurde, conversion impossible */
    (void) BI_ERR_alerte( principale, &ERR_UniteIncorrecte, VRAI, code_axe, unite_attendue, unite_graphe );

    return FAUX;
  }

  /* Elles sont de mme type : on peut esprer faire une conversion */
  if ( norm_attendu == norm_graphe ) {
    /* Les units sont en fait identiques, avec un nom diffrent */
    return VRAI;
  }

  /* Facteur de conversion : graphe -> SI [=*facteur] puis SI -> attendu [=/facteur] */
  facteur = norm_graphe / norm_attendu;

  /* On demande  l'utilisateur ? */
  if ( FAUX == convertir ) {	/* Non */
    (void) BI_ERR_alerte( principale, &ERR_UniteIncorrecte, VRAI, code_axe, unite_attendue, unite_graphe );
    return FAUX;
  }

  reponse = BI_ERR_alerte( principale, &Q_ConvertirUnite, VRAI, code_axe, unite_attendue, unite_graphe, facteur );
  if ( 2 == reponse ) {
    /* Conversion refuse */
    return FAUX;
  }

  /* On effectue la conversion : simple multiplication [on ne s'occupe pas de l'unit...] */
  GRF_FacteurMultiplicatif( graphe, facteur, axe, NULL );

  /* ... puisque l'on peut la mettre tout de suite ! */
  GRF_ChangerUnite( graphe, axe, unite_attendue );

  /* L'unit est donc maintenant correcte */
  return VRAI;
}

/* -------- Contrle de cohrence d'une liste de graphes --------
 |  Renvoie le masque des points qui sont homognes
 |
 | graphes = la liste des graphes  contrler
 | nombre  = le nombre de graphes dans cette liste
 | masque  = le masque des points  vrifier
 | modele  = si non-NULL, contiendra la valeur commune trouve
 */

unsigned int GRF_ControlerCompatibilite(Graphe **graphes, unsigned long nombre, unsigned int masque, Graphe *modele)
{
  boolean compatibles;
  unsigned int masque_retour = 0;

  if ( ( NULL == graphes ) || ( 0 == nombre ) ) {
    /* Aucun graphe : rien d'homogne ! */
    return 0;
  }

  if ( 1 == nombre ) {
    /* Un seul graphe : forcment homogne et valant ce graphe ! */
    if ( modele != NULL ) modele = graphes[ 0 ];
    return masque;
  }

  /* Si l'on est l : 2 graphes ou plus => contrle souhaitable ! */
  if ( masque & GRF_CONTROLE_TYPE ) {
    /* Les graphes ont-ils tous le mme type _exactement_ */
    int type_commun;

    compatibles = GRF_ControleType( graphes, nombre, VRAI, &type_commun );
    if ( VRAI == compatibles ) {
      masque_retour |= GRF_CONTROLE_TYPE;
      if ( modele != NULL ) modele->type = type_commun;
    }
  }

  if ( masque & GRF_CONTROLE_GENRE ) {
    /* Les graphes ont-ils tous le mme genre _exactement_ */
    int genre_commun;

    compatibles = GRF_ControleType( graphes, nombre, VRAI, &genre_commun );
    if ( VRAI == compatibles ) {
      masque_retour |= GRF_CONTROLE_GENRE;
      if ( modele != NULL ) modele->genre_donnees = genre_commun;
    }
  }

  if ( masque & GRF_CONTROLE_LEGENDE_X ) {
    /* Les graphes ont-ils tous la mme lgende en X, exactement ? */
    compatibles = GRF_LegendesHomogenes( graphes, nombre, AXE_X );
    if ( VRAI == compatibles ) {
      masque_retour |= GRF_CONTROLE_LEGENDE_X;
      if ( modele != NULL ) modele->affichage = graphes[ 0 ]->affichage;
    }
  }

  if ( masque & GRF_CONTROLE_UNITE_X ) {
    /* Les graphes ont-ils tous la mme unit en X, exactement ? */
    compatibles = GRF_UnitesHomogenes( graphes, nombre, AXE_X );
    if ( VRAI == compatibles ) {
      masque_retour |= GRF_CONTROLE_UNITE_X;
      if ( modele != NULL ) modele->affichage = graphes[ 0 ]->affichage;
    }
  }

  if ( masque & GRF_CONTROLE_LEGENDE_Y ) {
    /* Les graphes ont-ils tous la mme lgende en X, exactement ? */
    compatibles = GRF_LegendesHomogenes( graphes, nombre, AXE_Y );
    if ( VRAI == compatibles ) {
      masque_retour |= GRF_CONTROLE_LEGENDE_Y;
      if ( modele != NULL ) modele->affichage = graphes[ 0 ]->affichage;
    }
  }

  if ( masque & GRF_CONTROLE_UNITE_Y ) {
    /* Les graphes ont-ils tous la mme unit en Y, exactement ? */
    compatibles = GRF_UnitesHomogenes( graphes, nombre, AXE_Y );
    if ( VRAI == compatibles ) {
      masque_retour |= GRF_CONTROLE_UNITE_Y;
      if ( modele != NULL ) modele->affichage = graphes[ 0 ]->affichage;
    }
  }

  if ( masque & GRF_CONTROLE_NBR_PTS ) {
    /* Les graphes ont-ils exactement le mme nombre de points ? */
    unsigned long nombre_points_commun;

    compatibles = GRF_ControleTaille( graphes, nombre, GRF_CONTROLE_N_STRICT, &nombre_points_commun, FAUX );
    if ( VRAI == compatibles ) {
      masque_retour |= GRF_CONTROLE_NBR_PTS;
      if ( modele != NULL ) modele->nombre_point = nombre_points_commun;
    }
  }

  /* Termin */
  return masque_retour;
}

/* -------- Vrifier que les lgendes sont les mmes dans une srie de graphes -------- */

static boolean GRF_LegendesHomogenes(Graphe **graphes, unsigned long nombre_graphes, unsigned int axe)
{
  unsigned long i;
  const char *leg_reference = NULL, *legende;

  if ( ( NULL == graphes ) || ( 0 == nombre_graphes ) ) {
    return VRAI;
  }

  leg_reference = GRF_QuelleLegende( graphes[ 0 ], axe );
  for ( i = 1 ; i < nombre_graphes ; i++ ) {
    legende = GRF_QuelleLegende( graphes[ i ], axe );

    if ( legende == leg_reference ) {
      /* On a donc forcment la mme lgende ! */
      continue;
    }

    /* L : encore faut-il vrifier que le texte est le mme ou pas ! */
    if ( ( NULL == legende ) || ( NULL == leg_reference ) ) {
      AVERTISSEMENT( "On tombe sur une lgende NULL ! [i=%lu]", i );
      return FAUX;    /* On n'a donc pas la mme lgende partout... */
    }

    if ( strcmp( legende, leg_reference ) != 0 ) {
      return FAUX;		/* Lgendes diffrentes */
    }
  }

  /* Si l'on est l, c'est que les lgendes sont les mmes partout */
  return VRAI;
}

static boolean GRF_UnitesHomogenes(Graphe **graphes, unsigned long nombre_graphes, unsigned int axe)
{
  unsigned long i;
  const char *unite_reference = NULL, *unite;

  if ( ( NULL == graphes ) || ( 0 == nombre_graphes ) ) {
    return VRAI;
  }

  unite_reference = GRF_QuelleUnite( graphes[ 0 ], axe );
  for ( i = 1 ; i < nombre_graphes ; i++ ) {
    unite = GRF_QuelleUnite( graphes[ i ], axe );

    if ( unite == unite_reference ) {
      /* On a donc forcment la mme unit ! */
      continue;
    }

    /* L : encore faut-il vrifier que le texte est le mme ou pas ! */
    if ( ( NULL == unite ) || ( NULL == unite_reference ) ) {
      return FAUX;    /* On n'a donc pas la mme unit partout... */
    }

    if ( strcmp( unite, unite_reference ) != 0 ) {
      return FAUX;		/* Lgendes diffrentes */
    }
  }

  /* Si l'on est l, c'est que les lgendes sont les mmes partout */
  return VRAI;
}
