/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2002, Emmanuel CURIS

   Cration d'un nouveau graphe
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "stats_utils.h"

#include "graphes.h"
#include "tri.h"

#include "grf_doublons.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double DBL_ErreurCommune  (double *valeurs, int *ddl, unsigned long nbr_valeurs, int *ddl_final);
static double DBL_MoyenneDoublons(double *valeurs, unsigned long nbr_valeurs, double *variance);

static unsigned long GRF_GrouperX(Graphe *graphe, unsigned long lg_groupe, boolean moyenner);
static int GrouperX_Simple (Graphe *graphe, unsigned long lg_groupe);
static int GrouperX_R_M    (Graphe *graphe, unsigned long lg_groupe);
static int GrouperX_C_M    (Graphe *graphe, unsigned long lg_groupe);

static unsigned long GRF_QuasiDoublons(Graphe *graphe, double difference, boolean moyenner);
static unsigned long QuasiDoublons_R     (Graphe *graphe, double difference, boolean moyenner, unsigned long *n_doublons);

static unsigned long GRF_VraisDoublons(Graphe *graphe, boolean moyenner);
static unsigned long VraisDoublons_R     (Graphe *graphe, boolean moyenner, unsigned long *n_doublons);

/* ~~~~~~~~ Ralisation ~~~~~~~~~~~~~~ */

/* -------- Ralise la variance commune, compte-tenu des ddl -------- */

static double DBL_ErreurCommune(double *valeurs, int *ddl, unsigned long nbr_valeurs, int *ddl_final)
{
  double somme = 0.0, variance;
  unsigned int ddl_vrai, somme_ddl = 0;
  unsigned long i;

#ifdef _DEVELOPPE_
  if ( nbr_valeurs < 2 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Nombre de valeurs du doublon absurde [%lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, nbr_valeurs);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return *valeurs;
  }
#endif

  *ddl_final = 0;
  for ( i = 0 ; i < nbr_valeurs ; i++ ) {
    variance = valeurs[ i ] * valeurs[ i ]; /* On stocke les carts-types, on somme les variances */
    if ( ddl[ i ] < 0 ) {	/* Incertitude sur la moyenne */
      ddl_vrai = -ddl[ i ];
      variance *= (double) ( ddl_vrai + 1 );
    } else {
      ddl_vrai = ddl[ i ];
    }

    /* On fait la somme des carrs */
    if ( ddl_vrai > 0 ) {
      somme += variance * (double) ddl_vrai;
      somme_ddl += ddl_vrai;
    } else {			/* On considre un ddl... arbitraire ! */
      somme += variance;
      somme_ddl++;
    }
  }

  /* Variance associe */
  variance = somme;
  if ( somme_ddl > 0 ) {
    variance /= (double) somme_ddl;
  }

  /* Calculs termins : on renvoit les ddl et l'erreur  population (cart-type) */
  *ddl_final= (int) somme_ddl;
  return sqrt( variance );
}

/* -------- Ralise la moyenne des valeurs indiques et renvoie l'cart-type -------- */

static double DBL_MoyenneDoublons(double *valeurs, unsigned long nbr_valeurs, double *ecart_type)
{
  double moyenne, variance;
  unsigned long i;

#ifdef _DEVELOPPE_
  if ( nbr_valeurs < 2 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Nombre de valeurs du doublon absurde [%lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, nbr_valeurs);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return *valeurs;
  }
#endif

  /* 1. Calcul de la valeur moyenne */
  (void) STAT_Moyenne_D( valeurs, nbr_valeurs, &moyenne );

  /* 2. Calcul de la variance */
  variance = 0.0;
  for ( i = 0 ; i < nbr_valeurs ; i++ ) {
    double ecart;

    ecart = valeurs[ i ] - moyenne;
    variance += ecart * ecart;
  }
  variance /= (double) ( nbr_valeurs - 1 );

  *ecart_type = sqrt( variance );

  return moyenne;
}

/* -------- liminer les doublons [X] d'un graphe --------
     Renvoie le nombre de doublons traits
             0 en cas d'erreur

   graphe    : le graphe  traiter
   methode   : quelle mthode pour caractriser des doublons ?
   parametre : permettant de prciser la mthode
   moyenner  : VRAI = on moyenne les valeurs en Y [et en X] des doublons
               FAUX = on ne garde que la premire valeur rencontre
*/

unsigned long GRF_TraiterDoublons(Graphe *graphe, int methode, double parametre, boolean moyenner, boolean garder_erreurs)
{
  unsigned long nombre_doublons = 0;
  Graphe *ex_erreurs_Y = NULL;

  /* Controles sur le genre du graphe */
  if ( NULL == graphe ) {
    /* Peut arriver dans l'appel rcursif pour les erreurs, en cas de problme => pas de fin anticipe
       Peut arriver dans le chargement des spectres : pas de I0,... => pas de message d'avertissement */
    return 0;
  }
  if ( ( FAUX == GRF_EstReel( graphe ) ) && ( FAUX == GRF_EstComplexe( graphe ) ) ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Genre de graphe sans possibilit de chercher les doublons [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->genre_donnees );
#endif
    return 0;
  }
  if ( ( GRF_SERIE == graphe->genre_donnees ) && ( graphe->x != NULL ) ) {
    OPTION_A_FAIRE( "Controler que les X ne sont pas en commun pour toute la srie" );
    (void) fprintf(stderr, "%s,%s l.%u : Graphe = srie, X externes : non modifiable !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return +2;
  }

  if ( graphe->erreur_x != NULL ) {
    OPTION_A_FAIRE( "Graphe avec des erreurs en X" );
    return 0;
  }
  if ( graphe->erreur_y != NULL ) {
    int retour;

    /* On isole les erreurs en Y */
    ex_erreurs_Y = graphe->erreur_y;
    graphe->erreur_y= NULL;

    /* On les force  tre de type dviation [pour ne pas perdre ensuite les ddl] */
    GRF_ChangerType( ex_erreurs_Y, GR_DEVIATION );

    /* On leur fait apparaitre des X corrects, autonomes, si ncessaire */
    retour = GRF_ChangerGenre( ex_erreurs_Y, graphe->genre_donnees );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la prparation des incertitudes en Y. Elles seront perdues\n",
		     __FILE__, __FUNCTION__, __LINE__);
      GRF_efface( ex_erreurs_Y, VRAI, VRAI );
      ex_erreurs_Y = NULL;
    }

    /* On leur fait subir le traitement adquat */
    (void) GRF_TraiterDoublons( ex_erreurs_Y, methode, parametre, moyenner, FAUX );
  }

  /* Prparations complmentaires */
  if ( ( VRAI == moyenner ) && ( VRAI == garder_erreurs ) ) {
    Graphe *err_Y;

    /* Si on moyenne, on associe aussi des incertitudes au graphe, en Y */
    err_Y = GRF_creer_erreurs( graphe, AXE_Y );
    if ( NULL == err_Y ) {
      int reponse;

      reponse = BI_ERR_alerte( principale, &MSG_SansIncertitude, VRAI );
      if ( 1 == reponse ) return 0;	/* On abandonne */
    }
  }

  /* Log dans l'historique */
  HIST_Titre(4, E_AIGU"liminer les doublons", "Tri");
  HIST_Texte(4, 0, "Graphe trait"e_AIGU" : %s\n", graphe->nom);

  /* Suivant la mthode, on appelle telle ou telle fonction */
  switch ( methode )
  {
  case GRF_DOUBLONS_SIMPLE:	/* Doublons = points de mme X exactement */
    HIST_Texte(4, 0, "M"e_AIGU"thode : X successifs identiques ");

    nombre_doublons = GRF_VraisDoublons( graphe, moyenner );
    break;

  case GRF_DOUBLONS_GROUPE:	/* Doublons = paquets de k points, X quelconques */
    HIST_Texte(4, 0, "M"e_AIGU"thode : X successifs ");
    HIST_Texte(5, 0, "(par paquets de %lu) ", (unsigned long) parametre );

    nombre_doublons = GRF_GrouperX( graphe, (unsigned long) parametre, moyenner );
    break;

  case GRF_DOUBLONS_SOUPLE:
    HIST_Texte(4, 0, "M"e_AIGU"thode : X trop proches ");
    HIST_Texte(5, 0, "("DELTA_MAJUSCULE"x retenu : %.4g) ", parametre);

    nombre_doublons = GRF_QuasiDoublons( graphe, parametre, moyenner );
    break;

  default:
    (void) fprintf(stderr, "GRF_TraiterDoublons : mthode inconnue [%d]\n", methode);
    return 0;
  }

  if ( FAUX == moyenner ) { /* On ne veut pas moyenner : cas le plus simple... */
    HIST_Texte(4, 0, ", seul le premier est conserv"e_AIGU"\n");
  } else {
    HIST_Texte(4, 0, "puis moyenn"e_AIGU"s\n");
  }
  HIST_Texte(5, 0, "Doublons trouv"e_AIGU"s : %lu\n", nombre_doublons);

  /* Les bornes en X,Y ont peut-etre chang */
  GRF_bornes( graphe, ORDONNE, DESORDONNE );
  GRF_bornes( graphe->erreur_x, ORDONNE, DESORDONNE );
  GRF_bornes( graphe->erreur_y, ORDONNE, DESORDONNE );

  /* Y a-t-il eu des changements ? */
  if ( 0 == nombre_doublons ) {
    if ( ( graphe->erreur_x != NULL ) && ( graphe->erreur_x->y_max == 0.0 ) ) {
      /* Erreurs cres pour l'occasion -> on les efface */
      GRF_eff_isole( graphe->erreur_x );
      graphe->erreur_x = NULL;
    }

    if ( graphe->erreur_y != NULL ) {
      /* Erreurs cres pour l'occasion -> on les efface */
      GRF_eff_isole( graphe->erreur_y );
      graphe->erreur_y = NULL;
    }
  }

  /* Remettre les erreurs originales si ncessaire, maintenant qu'elles ont des X corrects */
  if ( ex_erreurs_Y != NULL ) {
#ifdef _DEVELOPPE_
    if ( FAUX == GRF_ComparerX( graphe, ex_erreurs_Y ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Le graphe et ses erreurs ne conduisent pas au meme rsultat aprs traitement !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    }
#endif

    if ( NULL == graphe->erreur_y ) {
      GRF_AssignerErreurs( graphe, ex_erreurs_Y, VRAI, AXE_Y );
    } else {
      OPTION_A_FAIRE( "Combiner les nouvelles erreurs aux anciennes" );
    }
  }

  /* On a fini : renvoie le nombre de doublons trouvs */
  return nombre_doublons;
}

/* ~-~-~-~-~-~-~-~- Grouper les k X conscutifs, sans contrle des X ~-~-~-~-~-~-~-~- */

/* -------- Appel gnrique -------- */

static unsigned long GRF_GrouperX(Graphe *graphe, unsigned long lg_groupe, boolean moyenner)
{
  int retour;
  unsigned long nombre_doublons;

  /* Controles */
  if ( ( lg_groupe < 2 ) || ( lg_groupe >= graphe->nombre_point ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : valeur de taille de groupe incorrecte [%lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, lg_groupe);
    return 0;
  }
  if ( ( graphe->nombre_point % lg_groupe ) != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : le nombre de points du graphe n'est pas un multiple de la taille de groupe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0;
  }

  HIST_Texte(4, 0, "M"e_AIGU"thode : X successifs group"e_AIGU"s ");
  HIST_Texte(5, 0, "par paquets de %lu ", lg_groupe);

  if ( FAUX == moyenner ) {
    /* Comme on ne fait qu'un filtre, l'adaptation des erreurs ne pose pas de problme ==> elle est automatique */
    retour = GrouperX_Simple( graphe, lg_groupe );
  } else {
    if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
      retour = GrouperX_R_M( graphe, lg_groupe );
    } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
      retour = GrouperX_C_M( graphe, lg_groupe );
    } else {
      (void) fprintf(stderr, "GRF_TraiterDoublons : il n'est pas possible de regrouper par X les graphes de ce type [%d]\n",
		     graphe->genre_donnees);
      retour = -5;
    }
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR survenue [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
  }

  /* On calcule le nombre de doublons */
  nombre_doublons = graphe->nombre_point / lg_groupe;

  /* C'est aussi le nombre de points restant */
  GRF_AdapteTaille( graphe, nombre_doublons );

  return nombre_doublons;
}

/* -------- Cas simple : on garde le 1er point de chaque groupe -------- */

static int GrouperX_Simple(Graphe *graphe, unsigned long lg_groupe)
{
  unsigned long i_lu, i_ecrit;
  unsigned long i;

  if ( ( GRF_FCT_REELLE == graphe->genre_donnees ) || ( GRF_COURBE_PLANE == graphe->genre_donnees ) ) {
    for ( i_lu = 0, i_ecrit = 0 ; i_lu < graphe->nombre_point ; i_lu += lg_groupe, i_ecrit++ ) {
      graphe->x[ i_ecrit ] = graphe->x[ i_lu ];
      graphe->y[ i_ecrit ] = graphe->y[ i_lu ];
    }
  } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    for ( i_lu = 0, i_ecrit = 0 ; i_lu < graphe->nombre_point ; i_lu += lg_groupe, i_ecrit++ ) {
      graphe->x[ i_ecrit ] = graphe->x[ i_lu ];
      ( (Complexe *) graphe->y)[ i_ecrit ] = ( (Complexe *) graphe->y)[ i_lu ];
    }
  } else if ( GRF_SERIE == graphe->genre_donnees ) {
    if ( NULL == graphe->x ) {
      for ( i_lu = 0, i_ecrit = 0 ; i_lu < graphe->nombre_point ; i_lu += lg_groupe, i_ecrit++ ) {
	graphe->y[ i_ecrit ] = graphe->y[ i_lu ];
      }
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : X externes, non-modifiables\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return 0;
    }
  } else {
    (void) fprintf(stderr, "%s,%s l.%u : Genre de graphe non gr [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->genre_donnees);
    return 0;
  }

  if ( GR_DEVIATION == graphe->type ) {
    int *ddl_lu, *ddl_ecrit;

    /* Il faut faire la mme chose sur les ddl */
    for ( ddl_lu = graphe->z.ddl, ddl_ecrit = ddl_lu, i = 0 ;
	  i < graphe->nombre_point ;
	  ddl_lu += lg_groupe, i += lg_groupe, ddl_ecrit++ ) {
      *ddl_ecrit = *ddl_lu;
    }
  }

  /* Pour bien faire, il faudrait aussi adapter la matrice de covariance... */
  if ( graphe->covariance != NULL ) {
    OPTION_A_FAIRE( "Adaptation de la matrice de covariance" );
  }

  /* On a fini */
  return 0;
}

/* -------- liminer en remplaant par la moyenne des valeurs : graphe rel --------
     On suppose qu'il n'y a pas d'incertitude sur le graphe fourni
 */

static int GrouperX_R_M(Graphe *graphe, unsigned long lg_groupe)
{
  unsigned long i_lu, i_ecrit;
  double moyenne, sigma;

  for ( i_lu = 0, i_ecrit = 0 ; i_lu < graphe->nombre_point ; i_lu += lg_groupe, i_ecrit++ ) {
#ifdef _DEVELOPPE_
    if ( i_lu + lg_groupe > graphe->nombre_point ) {
      (void) fprintf(stderr, "%s,%s l.%u : On dborde du graphe !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return -5;
    }
#endif
    moyenne = DBL_MoyenneDoublons( graphe->x + i_lu, lg_groupe, &sigma );
    graphe->x[ i_ecrit ] = moyenne; /* Moyenne des X groups */
    if ( graphe->erreur_x != NULL ) {
      graphe->erreur_x->y[ i_ecrit ] = sigma;
      graphe->erreur_x->z.ddl[ i_ecrit ] = lg_groupe - 1;
    }

    moyenne = DBL_MoyenneDoublons( graphe->y + i_lu, lg_groupe, &sigma );
    graphe->y[ i_ecrit ] = moyenne; /* Moyenne des Y correspondants */
    if ( graphe->erreur_y != NULL ) {
      graphe->erreur_y->y[ i_ecrit ] = sigma;
      graphe->erreur_y->z.ddl[ i_ecrit ] = lg_groupe - 1;
    }
  }

  /* On a fini */
  return 0;
}

/* -------- liminer en remplaant par la moyenne des valeurs : graphe complexe -------- */

static int GrouperX_C_M(Graphe *graphe __attribute__((__unused__)), unsigned long lg_groupe)
{
  unsigned long i_lu, i_ecrit;
  double moyenne_X, sigma_X;
  Complexe moyenne_Y = {0.0, 0.0}, sigma_Y = {0.0, 0.0};

  OPTION_A_FAIRE( "Groupement par paquets avec moyenne sur les complexes" );

  for ( i_lu = 0, i_ecrit = 0 ; i_lu < graphe->nombre_point ; i_lu += lg_groupe, i_ecrit++ ) {
#ifdef _DEVELOPPE_
    if ( i_lu + lg_groupe > graphe->nombre_point ) {
      (void) fprintf(stderr, "%s,%s l.%u : On dborde du graphe !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      return -5;
    }
#endif
    moyenne_X = DBL_MoyenneDoublons( graphe->x + i_lu, lg_groupe, &sigma_X );
    graphe->x[ i_ecrit ] = moyenne_X; /* Moyenne des X groups */
    if ( graphe->erreur_x != NULL ) {
      graphe->erreur_x->y[ i_ecrit ] = sigma_X;
      graphe->erreur_x->z.ddl[ i_ecrit ] = lg_groupe - 1;
    }

/*     moyenne = DBL_MoyenneDoublons( graphe->y + i_lu, lg_groupe, &sigma ); */
    ((Complexe *) graphe->y)[ i_ecrit ] = moyenne_Y; /* Moyenne des Y correspondants */
    if ( graphe->erreur_y != NULL ) {
      ((Complexe *) graphe->erreur_y->y)[ i_ecrit ] = sigma_Y;
      graphe->erreur_y->z.ddl[ i_ecrit ] = lg_groupe - 1;
    }
  }

  /* On a fini */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Grouper les X conscutifs tels que (Xk-X0) <= diff ~-~-~-~-~-~-~-~- */

/* -------------- limination des doublons : avec critre de diffrence ---------------- *\
|    Suppose le graphe tri, donc les "doublons" forcment  la suite                     |
|    Renvoi le nombre de doublons trouvs                                                 |
|                                                                                         |
| difference : si Dx < cette valeur, alors c'est un doublon                               |
| graphe     : le graphe dont on veut liminer les doublons                               |
| moyenner   : VRAI si on veut moyenner les doublons, FAUX si on veut juste les liminer  |
\* ------------------------------------------------------------------------------------- */

static unsigned long GRF_QuasiDoublons(Graphe *graphe, double difference, boolean moyenner)
{
  unsigned long nv_nombre_points = 0, nombre_doublons = 0;

  if ( difference < 0.0 ) {
    AVERTISSEMENT( "Critre de distance ngatif [%g]", difference );
    difference *= -1.0; /* On travaille en valeur absolue */
  }
  
  if ( difference >= ( graphe->x_max - graphe->x_min ) ) {
    (void) BI_ERR_alerte( principale, &ERR_DifferenceTropGrande, VRAI, graphe->x_max - graphe->x_min, difference );
    return 0;
  }

  /* Maintenant que l'on est sur que tout va bien, on appelle la bonne routine...
     Elle doit renvoyer le nombre de points aprs travail & en paramtre le nombre de doublons trouvs */
  if ( ( GRF_FCT_REELLE == graphe->genre_donnees ) || ( GRF_COURBE_PLANE == graphe->genre_donnees ) ) {
    nv_nombre_points = QuasiDoublons_R( graphe, difference, moyenner, &nombre_doublons );
  } else {
    OPTION_A_FAIRE( "Doublons souples, autres cas" );
  }

  /* On adapte la taille du graphe rsultat, si ncessaire */
  if ( nombre_doublons > 0 ) {
    GRF_AdapteTaille( graphe, nv_nombre_points ); /* Nouveau nombre de points */
  }

  /* Termin */
  return nombre_doublons;
}

/* -------- Moyenne faite, graphe  donnes relles --------
     On suppose qu'il n'y avait pas d'erreurs initialement !
 */

static unsigned long QuasiDoublons_R(Graphe *graphe, double difference, boolean moyenner, unsigned long *n_doublons)
{
  unsigned long nombre_points = 0, n_a_grouper;
  unsigned long i_lu = 1, i_ecrit = 0, i_ref = 0;
  double moyenne, sigma;

  /* On n'a encore compt aucun doublon */
  *n_doublons = 0;

  while ( i_lu < graphe->nombre_point ) {
    while ( fabs( graphe->x[ i_lu ] - graphe->x[ i_ref ] ) > difference ) {
      /* On est plus grand que Dx : points distincts */
      /* a. On crit la rfrence, qui ne change pas */
      graphe->x[ i_ecrit ] = graphe->x[ i_ref ];
      graphe->y[ i_ecrit ] = graphe->y[ i_ref ];
      if ( GR_DEVIATION == graphe->type ) {
	graphe->z.ddl[ i_ecrit ] = graphe->z.ddl[ i_ref ];
      }
      i_ecrit++;
      nombre_points++;	/* Un point en plus ! */

      /* b. Le x lu devient la nouvelle rfrence */
      i_ref = i_lu;

      /* c. On se prpare  lire le point suivant */
      i_lu++;

      /* d. On controle que l'on n'a pas fini ! */
      if ( i_lu >= graphe->nombre_point ) {
	graphe->x[ i_ecrit ] = graphe->x[ i_ref ];
	graphe->y[ i_ecrit ] = graphe->y[ i_ref ];
	if ( GR_DEVIATION == graphe->type ) {
	  graphe->z.ddl[ i_ecrit ] = graphe->z.ddl[ i_ref ];
	}
	nombre_points++;	/* Un point en plus ! */

	return nombre_points;
      }
    }

    /* Si on est l, c'est que la diffrence entre deux successifs est infrieure au seuil => doublon */
    (*n_doublons)++;
    n_a_grouper = 1;  /* Il y a au moins un point  grouper : la rfrence */
#ifdef _DEVELOPPE_
    if ( 0 == i_lu ) {
      (void) fprintf(stderr, "%s,%s l.%u : i_lu = 0, dbut de doublon !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    }
#endif

    /* Tous les points qui suivent et qui vrifient le critre sont moyenns pour donner la valeur  crire */
    while ( fabs( graphe->x[ i_lu ] - graphe->x[ i_ref ] ) <= difference ) {
      /* On passe au suivant */
      n_a_grouper++;
      i_lu++;
      if ( i_lu >= graphe->nombre_point) { /* Termin */
	break;
      }
    }
    /* i_lu : index de la premire valeur hors du groupe  */

    if ( FAUX == moyenner ) {
      graphe->x[ i_ecrit ] = graphe->x[ i_ref ];
      graphe->y[ i_ecrit ] = graphe->y[ i_ref ];
    } else {
      /* On moyenne les valeurs en X */
      moyenne = DBL_MoyenneDoublons( graphe->x + i_ref, n_a_grouper, &sigma );
      graphe->x[ i_ecrit ] = moyenne;
      if ( graphe->erreur_x != NULL ) {
	graphe->erreur_x->y[ i_ecrit ] = sigma;
	graphe->erreur_x->z.ddl[ i_ecrit ] = n_a_grouper - 1;
      }

      /* On moyenne les valeurs en Y */
      if ( GR_DEVIATION == graphe->type ) {
	double sigma_commun;
	int ddl;

	sigma_commun = DBL_ErreurCommune( graphe->y + i_ref, graphe->z.ddl + i_ref, n_a_grouper, &ddl );
	graphe->y[ i_ecrit ] = sigma_commun;
	graphe->z.ddl[ i_ecrit ] = ddl;
      } else {
	moyenne = DBL_MoyenneDoublons( graphe->y + i_ref, n_a_grouper, &sigma );
	graphe->y[ i_ecrit ] = moyenne;
	if ( graphe->erreur_y != NULL ) {
	  graphe->erreur_y->y[ i_ecrit ] = sigma;
	  graphe->erreur_y->z.ddl[ i_ecrit ] = n_a_grouper - 1;
	}
      }
    }

    /* Cela fait donc un point en plus d'crit */
    nombre_points++;
    i_ecrit++;

    /* La nouvelle rfrence : le premier point hors-doublons */
    i_ref = i_lu;

    /* Et le suivant est  lire */
    i_lu++;
  }

  if ( i_ref < graphe->nombre_point ) {
    /* Si on est ici, i_ref = dernier point du spectre puisque i_lu est en-dehors
       Ce dernier point ne peut pas etre en double avec celui d'avant, car dj trait...
       Il faut donc le recopier tel quel dans le graphe */
    graphe->x[ i_ecrit ] = graphe->x[ i_ref ];
    graphe->y[ i_ecrit ] = graphe->y[ i_ref ];
    if ( GR_DEVIATION == graphe->type ) {
      graphe->z.ddl[ i_ecrit ] = graphe->z.ddl[ i_ref ];
    }
    nombre_points++;
  }

  /* On a termin ! */
  return nombre_points;
}

/* ~-~-~-~-~-~-~-~- Grouper les X identiques (doublon au sens propre) ~-~-~-~-~-~-~-~- */

/* -------------- L'limination des doublons ------------------------------------------- *\
|    Suppose le graphe tri, donc les doublons forcment  la suite                       |
|    Renvoi le nombre de doublons trouvs                                                 |
|                                                                                         |
| graphe   : le graphe dont on veut liminer les doublons                                 |
| moyenner : VRAI si on veut moyenner les doublons, FAUX si on veut juste les liminer    |
\* ------------------------------------------------------------------------------------- */

static unsigned long GRF_VraisDoublons(Graphe *graphe, boolean moyenner)
{
  unsigned long nv_nombre_points = 0, n_doublons = 0;

  if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
    nv_nombre_points = VraisDoublons_R( graphe, moyenner, &n_doublons );
  } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    OPTION_A_FAIRE( "Elimination des doublons vrais, graphe complexe" );
  } else {
    OPTION_A_FAIRE( "Elimination des doublons vrais, autre genre de graphe" );
  }

  if ( n_doublons > 0 ) {
    GRF_AdapteTaille( graphe, nv_nombre_points ); /* Nouveau nombre de points */
  }

  return n_doublons;
}

/* -------- Cas d'un graphe  valeurs relles -------- */

static unsigned long VraisDoublons_R(Graphe *graphe, boolean moyenner, unsigned long *n_doublons)
{
  unsigned long nombre_points = 0, n_a_grouper;
  unsigned long i_lu = 1, i_ecrit = 0, i_ref = 0;
  double moyenne, sigma;

  /* On n'a encore compt aucun doublon */
  *n_doublons = 0;

  /* On commence la boucle, point par point */
  while ( i_lu < graphe->nombre_point ) {
    while ( graphe->x[ i_lu ] != graphe->x[ i_ref ] ) { /* Tant que le point est diffrent */
      graphe->x[ i_ecrit ] = graphe->x[ i_ref ];
      graphe->y[ i_ecrit ] = graphe->y[ i_ref ];
      if ( GR_DEVIATION == graphe->type ) {
	graphe->z.ddl[ i_ecrit ] = graphe->z.ddl[ i_ref ];
      }
      i_ecrit++;
      nombre_points++;

      /* Le point lu devient la nouvelle rfrence */
      i_ref = i_lu;

      /* On passe au point suivant */
      i_lu++;

      /* Controle : a-t-on fini ? */
      if ( i_lu >= graphe->nombre_point ) {
	/* On place le dernier point de rfrence */
	graphe->x[ i_ecrit ] = graphe->x[ i_ref ];
	graphe->y[ i_ecrit ] = graphe->y[ i_ref ];
	if ( GR_DEVIATION == graphe->type ) {
	  graphe->z.ddl[ i_ecrit ] = graphe->z.ddl[ i_ref ];
	}
	nombre_points++;

	return nombre_points;
      }
    }

    /* Si on est l, c'est que l'on a trouv un doublon */
    (*n_doublons)++;
    n_a_grouper = 1; /* Le point de rfrence */

    while ( graphe->x[ i_lu ] == graphe->x[ i_ref ] ) {
      n_a_grouper++;

      i_lu++;
      if ( i_lu >= graphe->nombre_point ) {
	/* On termine par un doublon */
	break;
      }
    }

    /* Ici : on sort => lot de doublons termin */
    if ( FAUX == moyenner ) {
      graphe->x[ i_ecrit ] = graphe->x[ i_ref ];
      graphe->y[ i_ecrit ] = graphe->y[ i_ref ];
      if ( GR_DEVIATION == graphe->type ) {
	graphe->z.ddl[ i_ecrit ] = graphe->z.ddl[ i_ref ];
      }
    } else {
      /* Moyenne des X */
      moyenne = DBL_MoyenneDoublons( graphe->x + i_ref, n_a_grouper, &sigma );
      graphe->x[ i_ecrit ] = moyenne;
      if ( graphe->erreur_x != NULL ) {
	graphe->erreur_x->y[ i_ecrit ] = sigma;
	graphe->erreur_x->z.ddl[ i_ecrit ] = n_a_grouper - 1;
      }

      /* Moyenne des Y */
      if ( GR_DEVIATION == graphe->type ) {
	double sigma_commun;
	int ddl;

	sigma_commun = DBL_ErreurCommune( graphe->y + i_ref, graphe->z.ddl + i_ref, n_a_grouper, &ddl );
	graphe->y[ i_ecrit ] = sigma_commun;
	graphe->z.ddl[ i_ecrit ] = ddl;
      } else {
	moyenne = DBL_MoyenneDoublons( graphe->y + i_ref, n_a_grouper, &sigma );
	graphe->y[ i_ecrit ] = moyenne;
	if ( graphe->erreur_y != NULL ) {
	  graphe->erreur_y->y[ i_ecrit ] = sigma;
	  graphe->erreur_y->z.ddl[ i_ecrit ] = n_a_grouper - 1;
	}
      }
    }

    i_ecrit++;
    nombre_points++;

    /* La nouvelle rfrence : le premier point hors-doublons */
    i_ref = i_lu;

    /* Et le suivant est  lire */
    i_lu++;
  }

  if ( i_ref < graphe->nombre_point ) {
    /* Si on est ici, i_ref = dernier point du spectre puisque i_lu est en-dehors
       Ce dernier point ne peut pas etre en double avec celui d'avant, car dj trait...
       Il faut donc le recopier tel quel dans le graphe */
    graphe->x[ i_ecrit ] = graphe->x[ i_ref ];
    graphe->y[ i_ecrit ] = graphe->y[ i_ref ];
    if ( GR_DEVIATION == graphe->type ) {
      graphe->z.ddl[ i_ecrit ] = graphe->z.ddl[ i_ref ];
    }
    nombre_points++;
  }

  /* On a termin */
  return nombre_points;
}
