/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2002, Emmanuel CURIS

   Effacement d'un nouveau graphe
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h" /* Pour contrle des arbres */

#include "chemins.h"
#include "couches.h"
#include "experience.h"
#include "feff.h"
#include "fichiers.h" /* Pour sauvegarde si effacement */
#include "filtrage.h"
#include "images.h"
#include "matrices.h" /* Pour effacement matrice de covariance */
#include "modeles.h"
#include "operations.h"
#include "serie.h"
#include "tracer.h"
#include "xanes.h"

#include "grf_effacer.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static boolean libere_graphe   (Graphe *graphe);

static void isole_graphe_experience(Graphe *graphe);
static void isole_graphe_interface(Graphe *graphe);
static void isole_graphe_modele   (Graphe *graphe, struct modele_ajuste *modele, boolean type_chemin);
static void isole_graphe_molecule (Graphe *graphe, Molecule *molecule);

static boolean test_fenetre(Graphe *graphe);

/* ~~~~~~~~ Ralisation ~~~~~~~~~~~~~~ */

/* -------- Libration du contenu (ne s'occupe pas des liens) -------- */

static boolean libere_graphe(Graphe *graphe)
{
  /* On efface les barres d'erreur associes */
  GRF_eff_isole( graphe->erreur_x ); graphe->erreur_x = NULL;
  GRF_eff_isole( graphe->erreur_y ); graphe->erreur_y = NULL;

  /* On efface la matrice de covariance */
  effacer_matrice( graphe->covariance );

  /* On efface les ventuels complments d'affichage */
  (void) GRF_LibererAffichage( graphe );

  if ( ( GRF_SERIE == graphe->genre_donnees ) || ( GRF_ERREUR_R == graphe->genre_donnees ) ||
       ( GRF_ERREUR_C == graphe->genre_donnees ) ) {
    graphe->x = NULL; /* Une srie n'a pas de X, mme si on peut lui en assigner -- ils ne lui appartiennent alors pas ! */
  }

  /* On efface les donnes */
  free( graphe->x ); graphe->x = NULL;
  free( graphe->y ); graphe->y = NULL;
  free( graphe->z.z ); graphe->z.z = NULL;
  free( graphe->pointeur.autre ); /* Y compris les ventuels complments */
  graphe->pointeur.autre = NULL;

  /* On efface le nom [en dernier pour pouvoir l'afficher jusqu'au bout] */
  free( graphe->nom ); graphe->nom = NULL;

  free( graphe ); graphe = NULL;

  return VRAI;
}

/* -------- Enlever tous les liens qui pointent vers le graphe indiqu --------
 | ATTENTION,
 |   par principe, les liens entre _graphes_ ne peuvent se faire qu'au sein d'une _mme_ srie
 */

boolean GRF_DelierGraphe(Serie_Graphes *serie, Graphe *graphe)
{
  int type;

  type = graphe->type; /* Sert pour acclrer les contrles */

  /* On vrifie qu'aucun graphe de la srie n'est li  celui-ci */
  if (serie != NULL) {
    if (serie->premier_graphe) { /* Il reste un graphe au moins en mmoire */
      /* Barres d'erreur [X,Y] : graphes isols -> pas de problme de lien */
    }
  }

  /* On vrifie qu'aucune image ne pointe sur ce graphe */
  if ( IMG_premiere_serie != NULL) {
    Image *image;
    Serie_Images *serie;

    if ( (type < 0) || (type == GR_PROFIL) ) { /* Seuls types admis pour une coupe d'image */
      for ( serie = IMG_premiere_serie ; serie != NULL ; serie = serie->suivante ) {
	for ( image = serie->premiere_image ; image != NULL ; image = image->suivante) {
	  if ( NULL == image->infos_affichage ) continue;

	  if ( image->infos_affichage->coupe == graphe) {
	    image->infos_affichage->coupe = NULL;
	    image->infos_affichage->xp_1 = 0.0; image->infos_affichage->yp_1 = 0.0;
	    image->infos_affichage->xp_2 = 0.0; image->infos_affichage->yp_2 = 0.0;
	    if ( image == image_courante ) { /* On doit refaire le dessin... */
	      FEN_redessin(fen_images, BI_RDS_TOUT);
	    }
	  }
	}
      }
    }

    /* Tous les types sont admis pour une image / graphes, mais le graphe devrait alors tre dans une srie protge
       => on ne devrait jamais arriver ici.
       Version de dveloppement seulement : on le vrifie
    */
#ifdef _DEVELOPPE_
    for ( serie = IMG_premiere_serie ; serie != NULL ; serie = serie->suivante ) {
      for ( image = serie->premiere_image ; image ; image = image->suivante) {
	unsigned long i, nbr_valeurs;
	struct graphe **valeurs;

	if ( IMG_GENRE_REELLE   == image->genre ) continue; /* Pas de risque avec ces types */
	if ( IMG_GENRE_COMPLEXE == image->genre ) continue; /* Pas de risque avec ces types */
	if ( IMG_GENRE_PHOTO    == image->genre ) continue; /* Pas de risque avec ces types */

	nbr_valeurs = image->larg * (unsigned long) image->haut;
	/* VRIFIER QUE LE GRAPHE N'EST PAS UN DES SPECTRES MCA DE L'IMAGE...
	   Normalement, ne devrait pas arriver car srie isole, "protge", efface avec l'image seulement
	*/
	for ( i = 0, valeurs = image->valeurs.G ; i < nbr_valeurs ; i++, valeurs++ ) {
	  if ( *valeurs == graphe ) {
	    *valeurs = NULL; /* Le graphe n'est plus dans l'image */

	    (void) fprintf(stderr, "%s,%s l.%u : ATTENTION, effacement d'un graphe d'une image, ce ne devrait pas arriver !\n",
			   __FILE__, __FUNCTION__, __LINE__);
	    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	  }
	}
      }
    }
#endif
  }

  /* Vrification du modle 3D */
  if ( premiere_molecule != NULL ) { /* Il y a des molcules */
    Molecule *molecule;

    for ( molecule = premiere_molecule ; molecule != NULL ; molecule = molecule->suivante ) {
      isole_graphe_molecule(graphe, molecule);
    }
  }

  /* Vrification des modles d'exprience */
  isole_graphe_experience( graphe );

  /* Vrification de l'utilisation dans les fentres / interface graphique */
  isole_graphe_interface(graphe);

  return VRAI; /* Isolement s'est bien pass */
}

static void isole_graphe_molecule(Graphe *graphe, Molecule *molecule)
{
  Fragment *fragment;
  int type_absorbeur, type_fragment, type_chemin, type;

  type = graphe->type;
  /* Il faudra penser  ajouter les TF quand ajustement des TF fonctionnera */
  type_absorbeur = (type < 0) || (graphe->drapeaux.exafs == VRAI) || /* Sans type, oscillations EXAFS */
    (type == GR_ECART); /* Rsidu d'ajustement */
  type_chemin    = (type < 0) || (graphe->drapeaux.exafs == VRAI) ||
    (type == GR_EXAFSAMP) || (type == GR_EXAFSPHA);
  type_fragment  = (type < 0) || (graphe->drapeaux.exafs == VRAI); /* Il faudra rajouter les types "mouvements" */

  if ( ( FAUX == type_absorbeur ) && (FAUX == type_fragment) ) /* Pas un type rencontr dans une molcule */
    return;

  /* ------ D'abord, on vrifie les absorbeurs ------ */
  if ( molecule->nbr_absorbeurs > 0 ) { /* Elle contient des absorbeurs */
    unsigned int i;
    Absorbeur *absorbeur;

    for (i = 0 ; i < molecule->nbr_absorbeurs ; i++) {
      struct modele_ajuste *modele;

      absorbeur = molecule->absorbeurs + i;

      /* S'il y a des graphes, ils sont dans les modles */
      for ( modele = absorbeur->modeles ; modele != NULL ; modele = modele->suivant ) {
	isole_graphe_modele( graphe, modele, type_chemin );
      }
    }
  }

  /* ------ Ensuite, on vrifie les fragments ------ */
  if (FAUX == type_fragment) return; /* Rien  faire */

  for (fragment = molecule->fragments ; fragment != NULL ; fragment = fragment->suivant) {
    if (fragment->signal == graphe)
      fragment->signal = NULL;
    if (fragment->translation == graphe)
      fragment->translation = NULL;
    if (fragment->rotation == graphe)
      fragment->rotation = NULL;
  }
}

static void isole_graphe_modele(Graphe *graphe, struct modele_ajuste *modele, boolean type_chemin)
{
  /* Graphe utilis comme rfrence, rsidu,... */
  if ( modele->experience == graphe ) {
    modele->experience = NULL;
    if ( modele->reference == modele->experience )
      modele->reference = NULL;
  }
  if ( graphe == modele->modele )
    modele->modele = NULL;
  if ( graphe == modele->residu )
    modele->residu = NULL;

  /* Graphe utilis dans la construction du modle ? */
  if ( NULL == modele->infos_modele.feff ) /* Pas d'infos complmentaires ! */
    return;

  switch ( modele->type )
  {
  case MODELE_MULTI_MODELES:
    {
      struct multi_modeles *multi;
      unsigned int i;

      multi = modele->infos_modele.modeles;
      for (i = 0 ; i < multi->nbr_modeles ; i++) {
	isole_graphe_modele(graphe, multi->modeles[i], type_chemin);
      }
    }
    break;

  case MODELE_CL_MODELES:
    {
      struct cl_modeles *cl;
      unsigned int i;

      cl = modele->infos_modele.cl_modeles;
      for (i = 0 ; i < cl->nbr_termes ; i++) {
	isole_graphe_modele(graphe, cl->modeles[i], type_chemin);
      }
    }
    break;

  case MODELE_EXAFS_CHEMINS:
    {
      struct modele_feff *feff;
      Couche *couche;

      feff = modele->infos_modele.feff;

      /* On vrifie qu'aucune couche ne pointe sur ce graphe */
      if ( VRAI == type_chemin ) { /* Seuls types susceptibles d'tre points */
	for (couche = feff->couches ; couche ; couche = couche->suivante) {
	  if (graphe == couche->amplitude) {
	    couche->amplitude = NULL;
	    free(couche->b_amp); couche->b_amp = NULL;
	    free(couche->c_amp); couche->c_amp = NULL;
	    free(couche->d_amp); couche->d_amp = NULL;
	  }
	  if (graphe == couche->phase) {
	    couche->phase = NULL;
	    free(couche->b_pha); couche->b_pha = NULL;
	    free(couche->c_pha); couche->c_pha = NULL;
	    free(couche->d_pha); couche->d_pha = NULL;
	  }
	  if (graphe == couche->signal)
	    couche->signal = NULL;

	  if ( couche->chemin != NULL ) {
	    Chemin *chemin;

	    chemin = couche->chemin;
	    if (chemin->exf_calcule == graphe) {
	      chemin->exf_calcule = NULL;
	    }
	    if (chemin->amplitude == graphe) {
	      chemin->amplitude = NULL;
	    }
	    if (chemin->phase == graphe) {
	      chemin->phase = NULL;
	    }

	    (void) fprintf(stderr, "isole_graphe_modele :  FAIRE, contrle des chemins quivalents\n");
	  }
	}
      }

      /* Graphes gnraux */
      if (feff->phase_central == graphe)
	feff->phase_central = NULL;
      if (feff->lpm == graphe) {
	feff->lpm = NULL;
	free(feff->b_lpm); feff->b_lpm = NULL;
	free(feff->c_lpm); feff->c_lpm = NULL;
	free(feff->d_lpm); feff->d_lpm = NULL;
      }
    }
    break;

  case MODELE_EXAFS_GNXAS:
  case MODELE_SEUIL_BASIQUE:
  case MODELE_FLUOX:
  case MODELE_DECOMPOSITION:
    (void) fprintf(stderr, "isole_graphe_modele : type de modle pas encore support [%d]\n", modele->type);
    break;

  default:
    (void) fprintf(stderr, "isole_graphe_modele : type de modle inconnu [%d]\n", modele->type);
  }
}

static void isole_graphe_interface(Graphe *graphe)
{
  OPTION_A_FAIRE( "Vrifier que le graphe n'est pas dans une figure !" );

  /* On vrifie maintenant qu'il n'y a pas de fentre ouverte qui utilise ce graphe
     Dans l'tat actuel, seule une fentre de type mixte peut utiliser un graphe ;
     si effectivement elle l'utilise, on force la fermeture

     NB : les fentres GEM pures qui pointent sur une liste de graphes sont des arbres bloquants,
     donc il n'y a pas d'effacement possible pendant leur utilisation
     Seule exception : l'affichage de la liste des graphes,
     qui est de toute faon remis  jour aprs
          la fentre de superposition est un cas particulier dj trait : pas besoin de la fermer
     mme si on efface un graphe qu'elle utilise, mais il faut tout de mme ne plus l'utiliser

     La fermeture d'une fentre pouvant entraner des effacements de graphe qui peuvent entraner
     des fermetures de fentre (...), le chanage des fentres peut tre fortement malmen pendant
     l'opration. En consquence,  chaque effacement de fentre, on annule le parcours de la chane
     et on recommance  zro, jusqu' ce que le graphe ne soit utilis par aucune fentre
  */
  while ( test_fenetre(graphe) == VRAI );
}

/* -------- On vrifie que le graphe n'est pas utilis dans un composant d'une exprience -------- */

static void isole_graphe_experience(Graphe *graphe)
{
  /* ---- Contrle pour les faisceaux ---- */
  /* 1. volution avec l'nergie */
  if ( ( GR_I0 == graphe->type ) || ( GR_SSTYPE == graphe->type ) ) {
    Faisceau *faisceau;

    for ( faisceau = premier_faisceau ; faisceau != NULL ; faisceau = faisceau->suivant ) {
      if ( faisceau->g_energie == graphe ) { /* Le faisceau n'est plus monochromatique ! */
	faisceau->g_energie = NULL;
	faisceau->proprietes &= ~FAISCEAU_DECROIT;
      }
    }
  }

  /* 2. g(E) */
  if ( ( GR_RESOL_NRJ == graphe->type ) || ( GR_SSTYPE == graphe->type ) ) {
    Faisceau *faisceau;

    for ( faisceau = premier_faisceau ; faisceau != NULL ; faisceau = faisceau->suivant ) {
      if ( faisceau->g_E == graphe ) { /* Le faisceau n'est plus monochromatique ! */
	faisceau->g_E = NULL;
	faisceau->proprietes &= ~FAISCEAU_POLYCHROMATIQUE;
      }

      if ( faisceau->nbr_harmoniques > 0 ) {
	unsigned int i;

	for ( i = 0 ; i < faisceau->nbr_harmoniques ; i++ ) {
	  if ( faisceau->g_harmoniques[ i ] == graphe ) {
	    faisceau->g_harmoniques[ i ] = NULL;
	  }
	}
      }
    }
  }

  /* ---- Contrle pour les dtecteurs ---- */

  /* ---- Contrle pour les compositions ---- */
  if ( ( GR_MOINSMAT == graphe->type ) || ( GR_SSTYPE == graphe->type ) ) {
    struct composant *composant;
    unsigned int i;

    for ( composant = premier_composant ; composant != NULL ; composant = composant->suivant ) {
      if ( composant->mu_total == graphe ) {
	composant->mu_total = NULL;
      }

      for ( i = 0 ; i < ZMAX_TABLE_CONSTITUANTS ; i++ ) {
	if ( composant->mu[ i ] == graphe ) {
	  composant->mu[ i ] = NULL;
	}
      }
    }
  }
}

/* -------- ter le graphe du chanage -------- */

int GRF_DefaireChainage(Serie_Graphes *serie, Graphe *graphe)
{
  if ( graphe == NULL ) return FAUX;
  if (VRAI == graphe->drapeaux.isole) return VRAI;

  if ( serie == NULL ) {
    serie = GRF_QuelleSerie(graphe);
    if (serie == NULL) {
      (void) fprintf(stderr, "GRF_DefaireChainage : graphe hors-srie, non marqu comme isol ! [%s]\n", graphe->nom);
      return FAUX; /* Graphe hors-srie : pas de chanage */
    }
  }

  /* On supprime le lien ventuel avec les X communs  la srie */
  if ( GRF_SERIE == graphe->genre_donnees  ) {
    if ( ( graphe->x != NULL ) && ( graphe->x == serie->table_X ) ) {
#ifdef _DEVELOPPE_
      if ( FAUX == serie->drapeaux.table_X ) {
	(void) fprintf(stderr, "%s,%s l.%u : Srie %s marque sans X propres, mais en a utiliss dans un graphe %s !\n",
		       __FILE__, __FUNCTION__, __LINE__, serie->nom, graphe->nom);
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
      }
#endif
      graphe->x = NULL;
      graphe->x_min = 0;
      graphe->x_max = graphe->nombre_point - 1;
    }
  }

  /* On enlve le graphe du chanage */
  if (graphe->precedent == NULL) { /* Premier graphe de la srie */
    if (graphe->suivant == NULL) { /* Dernier graphe de la srie */
      serie->premier_graphe = NULL;
      serie->dernier_graphe = NULL;
    } else {
      serie->premier_graphe = graphe->suivant;
      serie->premier_graphe->precedent = NULL; /* Le suivant devient le premier */
    }
  } else if (graphe->suivant == NULL) { /* Dernier graphe */
    serie->dernier_graphe = graphe->precedent;
    serie->dernier_graphe->suivant = NULL;
  } else {
    (graphe->precedent)->suivant = graphe->suivant;   /* On reforme la chaine */
    (graphe->suivant)->precedent = graphe->precedent;
  }
#ifdef _DEVELOPPE_
  if (serie->nbr_graphes == 0) {
    fprintf(stderr, "GRF_DefaireChainage : problme dans le compte des graphes, srie %s\n", serie->nom);
  } else
#endif
    serie->nbr_graphes--;

  /* On essaye de dterminer le nouveau graphe courant */
  if (serie == serie_courante) { /* Sinon, inutile... */
    if (serie->premier_graphe == NULL) {
      graphe_courant = NULL;
      GRF_aller(NULL);
    } else if (graphe == graphe_courant) {
      Graphe *nv_graphe_courant;

      if ( graphe->suivant != NULL ) {
	nv_graphe_courant = graphe->suivant;
      } else {
	nv_graphe_courant = graphe->precedent;
      }

      /* Le graphe est rellement isol maintenant ! */
      graphe->suivant = NULL;
      graphe->precedent = NULL;
      graphe->drapeaux.isole = VRAI;

      GRF_aller(nv_graphe_courant);

      /* Remarque : on doit procder ainsi, sinon les paramtres enregistrs dans la liste
	 par GRF_aller de l'ancien graphe courant sont attribus au nouveau
	 => p. ex, quand on efface un graphe, son titre passe au suivant !

	 On doit aussi marquer le graphe comme isol, sinon la recherche de la srie associe choue
	 ==> SEG FAULT car incohrence dtecte ! */
    }
  }

  /* Le graphe est marqu comme isol */
  graphe->suivant = NULL;
  graphe->precedent = NULL;
  graphe->drapeaux.isole = VRAI;

  /* Remarque : on n'efface pas les sries vides car
     - l'utilisateur peut vouloir la remplir aprs ;
     - cela facilite l'effacement de tous les graphes d'une srie
  */

  return VRAI; /* Tout s'est bien pass */
}

/* -------------- Effacer un graphe, avec ou sans contrle --------------------------------\
 |    Renvoie VRAI si le graphe a t effac, FAUX sinon                                   |
 |                                                                                         |
 | graphe        : le graphe que l'on veut effacer                                         |
 | force_sauve   : VRAI : on efface mme si le graphe n'a pas t sauv                    |
 | force_protege : VRAI = on efface mme si le graphe est protg                          |
 \---------------------------------------------------------------------------------------- */

boolean GRF_efface(Graphe *graphe, boolean force_sauve, boolean force_protege)
{
  Serie_Graphes *serie;

  if (graphe == NULL) return VRAI; /* Pas de problme */

#ifndef VERSION_SIMPLIFIEE_ /* Pas d'interface dans la version limite */
  /* --- Les vrifications prliminaires --- */
  if ( (graphe->drapeaux.protege == VRAI) && (force_protege == FAUX) ) {
    EXF_ERREUR(GRF_PROTEGE);
    return FAUX;	
  }

  if ( (graphe->drapeaux.sauve == FAUX) && (force_sauve == FAUX) ) {
    int reponse = EXF_ERREUR(NON_SAUVE);

    if (reponse == 3) return FAUX;
    if ( 1 == reponse ) {
      /* On sauve le graphe */
      I_GRF_Sauver( graphe );
    }
  }
#ifdef _DEVELOPPE_
  printf("GRF_effacer : %s [%p]\n", graphe->nom, (void *) graphe);
#endif
  HIST_Texte(4, 0, "On efface le graphe %s\n", graphe->nom);
#endif /* VERSION_SIMPLIFIEE_ */

  /* Quelle srie contient le graphe ? */
  serie = GRF_QuelleSerie(graphe);
  if (NULL == serie) {
    (void) fprintf(stderr, "GRF_efface : appel pour un graphe isol ! [%s]\n", graphe->nom);
    return GRF_eff_isole(graphe);
  }

  if ( (serie->drapeaux.protegee == VRAI) && (force_protege == FAUX) ) {
    (void) fprintf(stderr, "GRF_efface : Srie protge\n");
    return FAUX; /* Graphe non effac : srie protge */
  }

  GRF_DefaireChainage(serie, graphe); /* On enlve le graphe du chanage */
  GRF_DelierGraphe(serie, graphe);    /* On te les liens qui pointent sur lui */
  libere_graphe(graphe);           /* On libre la mmoire */

  return VRAI; /* On a effac le graphe */
}

/* -------- Vrification qu'un graphe effac n'est pas utilis par une fentre ----------\
 |   Renvoie VRAI si le graphe tait utilis par une fentre                             |
 |                dans ce cas, la fentre a t ferme -> il y a eu des changements dans |
 |                le chanage des fentres, donc mieux vaut reprendre  zro             |
 |           FAUX si le graphe n'tait pas utilis par une quelconque fentre            |
 |                                                                                       |
 | graphe : le graphe dont il faut vrifier l'utilisation                                |
 \-------------------------------------------------------------------------------------- */

static boolean test_fenetre(Graphe *graphe)
{
  int numero_arbre;
  Graphe *grf_fenetre;
  Fenetre *fenetre;
  Message message;

  if (graphe == NULL) return FAUX;

  message.type = WM_CLOSED;
  message.fenetre.application = 0;
  message.fenetre.en_plus = 0;
  message.fenetre.x = -1; message.fenetre.y = -1;
  message.fenetre.larg = -1; message.fenetre.haut = -1; /* Fermeture force... */

  for ( fenetre = BI_premiere_fenetre ; fenetre ; fenetre = fenetre->suivante ) {
    /* Y a-t-il un graphe directement li  la fentre ? */
    grf_fenetre = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
    if ( grf_fenetre == graphe ) {
      /* Oui : on ferme donc la fentre correspondante */
      if ( fenetre->message != ( int (*)(Fenetre *, Message *) ) NULL ) {
	/* Normalement, la routine correspondante se charge de dfaire les liens : pas de boucle sans fin (esprons !) */
	(fenetre->message)(fenetre, &message);
      } else {
	/* On dfait le lien avant de fermer la fentre ! */
	FEN_GRF_LierGraphe( fenetre, NULL );
	FEN_Detruire( fenetre );
      }

      return VRAI;
    }

    /* Deuxime possibilit : sur la 2e info ! */
    grf_fenetre = (Graphe *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
    if ( grf_fenetre == graphe ) {
      AVERTISSEMENT( "Graphe mmoris directement dans le 2e complment ! [%s]", fenetre->titre );
      /* Oui : on ferme donc la fentre correspondante */
      if ( fenetre->message != ( int (*)(Fenetre *, Message *) ) NULL ) {
	/* Normalement, la routine correspondante se charge de dfaire les liens : pas de boucle sans fin (esprons !) */
	(fenetre->message)(fenetre, &message);
      } else {
	/* On dfait le lien avant de fermer la fentre ! */
	FEN_GRF_LierGraphe( fenetre, NULL );
	FEN_Detruire( fenetre );
      }

      return VRAI;
    }

    /* Troisime possibilit : inclus dans des infos associes  la fentre */
    /* Cas 1 : il y a un arbre dans la fentre. Aide pour savoir quelles infos sont stockes ! */
    numero_arbre = FEN_QuelArbre( fenetre );
    if ( numero_arbre < 0 ) { /* Pas d'arbre ! */
      if ( fenetre == principale ) {
	/* Aucune info associe  la fentre principale, normalement ! */
      } else {
	void *infos_1, *infos_2;

	/* Infos associes  la fentre */
	infos_1 = FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
	infos_2 = FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

	if ( infos_1 != NULL ) {
	  AVERTISSEMENT( "Infos associes dans le 1er complment (Graphe ?) [%p], fentre %s !", infos_1, fenetre->titre );
	}
	if ( infos_2 != NULL ) {
	  AVERTISSEMENT( "Infos associes dans le 2e complment [%p], fentre %s !", infos_2, fenetre->titre );
	}
      }
    } else {
      void *infos_1, *infos_2;
      int retour;

      /* Infos associes  l'arbre */
      infos_1 = FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
      infos_2 = FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

      switch ( numero_arbre )
      {
      case ELIM_MUMATRICE:
	retour = I_XAN_ControleEffacement( fenetre, graphe );
	if ( VRAI == retour ) {
	  return VRAI;	/* Le contrle s'est traduit par la fermeture de la fentre ! */
	}
	break;

      case ELIM_MU0_1: /* Premire tape de l'limination de 0 */
	if ( infos_2 != NULL ) {
	  MU1_infos *infos;

	  infos = (MU1_infos *) infos_2;
	  if ( infos != NULL ) {
	    if ( graphe == infos->exafs ) {
	      infos->exafs = NULL; /* Graphe en cours d'effacement : n'existe plus
				      vite que la fermeture redemande l'effacement de ce graphe !
				      Sinon, boucle sans fin (-> stack overflow) ou double dcompte
				      de graphe enlev (-> nbr_graphes < 0 -> seg. fault.)
				   */
	      (fenetre->message)(fenetre, &message);
	      return VRAI;
	    }
	    if ( graphe == infos->mu_0 ) {
	      infos->mu_0 = NULL;
	      (fenetre->message)(fenetre, &message);
	      return VRAI;
	    }
	  }
	}
	break;

      case ELIM_MUO_DEUX: /* Deuxime tape de l'limination de 0 */
	if ( infos_2 != NULL ) {
	  MU2_infos *infos;

	  infos = (MU2_infos *) infos_2;
	  if ( graphe == infos->mu_0_ref ) { /*  de l'tape d'avant */
	    infos->mu_0_ref = NULL; /* A priori, sans autre effet -> on continue les essais */
	  }

	  if ( graphe == infos->mu_0 ) { /* Le 0 calcul */
	    infos->mu_0 = NULL;
	    (fenetre->message)(fenetre, &message);
	    return VRAI;
	  }

	  if ( graphe == infos->exafs ) { /* Le signal corrig */
	    infos->exafs = NULL;
	    (fenetre->message)(fenetre, &message);
	    return VRAI;
	  }
	}
	break;

      case PARAM_FOURIER: /* Paramtrage de la fentre de la TF */
	if ( infos_2 != NULL ) {
	  TF_infos *infos;

	  infos = (TF_infos *) infos_2;
	  if ( graphe == infos->graphe ) { /* Normalement, jamais */
	    infos->graphe = NULL;
	    (fenetre->message)(fenetre, &message);
	    return VRAI;
	  }
	  if ( graphe == infos->fenetre ) {
	    infos->fenetre = NULL;
	    (fenetre->message)(fenetre, &message);
	    return VRAI;
	  }
	  if ( graphe == infos->apercu ) {
	    infos->apercu = NULL;
	    OBJ_choisit(PARAM_FOURIER, TF_APERCU, FORCE_NON);
	    /* Pas la peine de fermer la fentre, mais on continue  vrifier au cas o... */
	  }
	}
	break;

      case NORM_XANES: /* Normalisation du XANES */
	if ( infos_2 != NULL ) {
	  MU1_infos *infos;

	  infos = (MU1_infos *) infos_2;
	  if ( infos->mu_0 == graphe ) {
	    infos->mu_0 = NULL; /* Plus de rfrence */
	    FEN_redessin(fenetre, BI_RDS_UTIL);
	  }
	}
	break;

      case OTE_MU0_CALC: /* Soustraction d'un 0 dj calcul */
	if ( infos_2 != NULL ) {
	  MU0_infos *infos;

	  infos = (MU0_infos *) infos_2;
	  if ( infos->mu_0 == graphe ) { /* On efface le 0 utilis */
	    infos->mu_0 = NULL;
	    (fenetre->message)(fenetre, &message);
	    return VRAI;
	  }
	  if ( infos->exafs == graphe ) { /* On efface l'exafs extrait */
	    infos->exafs = NULL;
	    (fenetre->message)(fenetre, &message);
	    return VRAI;
	  }
	}
	break;

      case RESULT_N_COUCHES: /* Rsultats d'ajustement */
	if ( infos_1 != NULL ) {
	  struct modele_ajuste *modele;

	  modele = (struct modele_ajuste *) infos_1;
	  if ( modele->reference == graphe ) { /* On efface le graphe de rfrence */
	    modele->reference = NULL;
	    (fenetre->message)(fenetre, &message);
	    return VRAI;
	  }

	  if ( modele->modele == graphe )
	    modele->modele = NULL;

	  if ( modele->modele == graphe )
	    modele->modele = NULL;
	}
	if ( infos_2 != NULL ) {
	  AVERTISSEMENT( "Infos associes dans le 2e complment [%p], fentre d'ajustements EXAFS !", infos_2 );
	}

      default:
	if ( infos_1 != NULL ) {
	  AVERTISSEMENT( "Infos associes dans le 1er complment (Graphe ?) [%p], fentre avec l'arbre %u !", infos_1, numero_arbre );
	}
	if ( infos_2 != NULL ) {
	  AVERTISSEMENT( "Infos associes dans le 2e complment [%p], fentre avec l'arbre %u !", infos_2, numero_arbre );
	}
      }
    }
  }

  return FAUX; /* Trouv dans aucune fentre -> effacement sr... */
}

/* -------- Effacer un graphe isol. Plus simple, car aucune liaison ni chanage -------- */

boolean GRF_eff_isole(Graphe *graphe)
{
  if (NULL == graphe) return FAUX;
  if ( FAUX == graphe->drapeaux.isole ) {
    (void) fprintf(stderr, "GRF_eff_isole : graphe dans une srie ! [%s]\n", graphe->nom);
    return FAUX;
  }

  /* Si barres d'erreur : graphes isols aussi */
  if (graphe->erreur_x != NULL) {
    libere_graphe(graphe->erreur_x);
    graphe->erreur_x = NULL;
  }
  if (graphe->erreur_y != NULL) {
    libere_graphe(graphe->erreur_y);
    graphe->erreur_y = NULL;
  }

  return libere_graphe(graphe);
}

/* -------- Effacer tous les graphes en mmoire -------- */

/* Fermeture des fentres d'interface, car seront fermes de toute faon => gagne du temps */

static boolean ferme_fenetre_interface(void)
{
  Fenetre *fenetre;

  for (fenetre = BI_premiere_fenetre ; fenetre != NULL ; fenetre = fenetre->suivante) {
    if ( BI_FENETRE_SIMPLE_ARBRE == fenetre->type ) {
      if (DECOUPE_IMAGE == fenetre->numero_arbre)
	continue; /* Contient des images, mais pas de graphes */

      if ( fenetre->message ) {
	Message message;

	message.type = WM_CLOSED;
	message.fenetre.application = 0;
	message.fenetre.en_plus = 0;
	message.fenetre.x    = -1; message.fenetre.y    = -1;
	message.fenetre.larg = -1; message.fenetre.haut = -1; /* Fermeture force... */

	(fenetre->message)(fenetre, &message);
      } else
	FEN_Detruire( fenetre );

      return VRAI; /* On a modifi le chanage des fentres, peut-tre de faon sauvage (cascade...) */
    }
  }

  return FAUX; /* On n'a ferm aucune fentre */
}

/* Effacement effectif. Renvoie FAUX si l'utilisateur n'a finalement pas voulu tout effacer */

boolean GRF_tteffacer(void)
{
  boolean vide = FAUX;
/*   unsigned long num_serie = 0; */ /* Nombre de sries sauves automatiquement... */

  if (premiere_serie == NULL) /* Aucun graphe  effacer */
    return VRAI;

  /* On commence par fermer toutes les fentres de type mixte.
     Efface tous les graphes de travail au passage... */
  while (VRAI == ferme_fenetre_interface());

  /* On devrait mettre tous les liens vers un graphe  0 pour gagner du temps
     et ne pas avoir  isoler les graphes... */

  met_chaine(ARB_QUITTER, FIN_PRG_TITRE, 0, texte_non_sauve, TXT_REMPLACE);
  OBJ_cache(ARB_QUITTER, QUIT_NUM_ATOM, FORCE_OUI); /* Pour les atomes... */
  OBJ_cache(ARB_QUITTER, QUIT_X_ATOME, FORCE_OUI); /* Pour les atomes... */
  OBJ_cache(ARB_QUITTER, QUIT_Y_ATOME, FORCE_OUI); /* Pour les atomes... */
  OBJ_cache(ARB_QUITTER, QUIT_Z_ATOME, FORCE_OUI); /* Pour les atomes... */
#ifndef _DEVELOPPE_
  OBJ_active(ARB_QUITTER, QUIT_TTSAUVER, FORCE_OUI); /* Car dsactiv pour les molcules */
#endif

  /* Premire passe : effacer les sries vides qui peuvent traner... */
  do {
    Serie_Graphes *serie;

    vide = FAUX;
    for ( serie = premiere_serie ; serie ; serie = serie->suivante ) {
      if ( NULL == serie->premier_graphe ) { /* Srie vide ! */
	vide = VRAI;
	SERIE_GRF_EffacerSerie( serie, VRAI );
	break;
      }
    }
  } while ( VRAI == vide );

  /* Deuxime passe : effacer les autres sries */
  while ( premiere_serie != NULL ) {
    SERIE_GRF_EffacerSerie( premiere_serie, VRAI );
  }

#ifdef ancien
  while (premiere_serie != NULL) {
    int reponse;
    unsigned long num_sauve = 1; /* Nombre de graphes sauvs automatiquement / srie */
    boolean tt_sauve = FAUX, tt_perdre = FAUX;
    Serie_Graphes *serie = premiere_serie;

    while ( serie->premier_graphe != NULL ) {
      Graphe *graphe = serie->premier_graphe;

      if (tt_perdre == VRAI) /* L'utilisateur a demand  tout perdre */
	graphe->drapeaux.sauve = VRAI; /* Donc le graphe est considr sauv */
	  
      if (graphe->drapeaux.sauve == FAUX) { /* Le graphe n'est pas sauv */
	char *nom_sauve = NULL;
	if (tt_sauve == VRAI) { /* On veut tous les sauver */
	  num_sauve++;
	  nom_sauve = TXT_genere_chaine("serie%u_graphe%u.%s", num_serie, num_sauve,
					(graphe->type < 0) ? ".sav" : extensions[graphe->type] + 1);
	  sauver(graphe, nom_sauve); /* On sauve le graphe - libre aussi nom_sauve */
	} else {
	  met_chaine(ARB_QUITTER, QUIT_NOM_GRAPHE, 0, graphe->nom, TXT_REMPLACE);
	  reponse = ext_formdo(ARB_QUITTER, titre_tout_effacer);
	  switch (reponse)
	  {
	  case QUIT_RESTER:   /* Finalement, on ne veut plus effacer les graphes... */
	    return FAUX;
		      
	  case QUIT_TTSAUVER: /* On sauve tous les graphes non sauvs               */
	    tt_sauve = VRAI;
	    num_serie++;
	    nom_sauve = TXT_genere_chaine("serie%u_graphe%u.%s", num_serie, num_sauve,
					  (graphe->type < 0) ? ".sav" : extensions[graphe->type] + 1);
	  case QUIT_SAUVER:   /* On sauve ce graphe                                 */
	    sauver(graphe, nom_sauve);
	    break;
	    
	  case QUIT_TTPERDRE: /* On perd tous les graphes                           */
	    tt_perdre = VRAI;
	  }
	}
	graphe->drapeaux.sauve = VRAI; /* De toute faon, on le perd... */
      }

#ifdef _DEVELOPPE_
      /*    printf("GRF_tt_efface : %s (1er passage)\n", graphe->nom); */
#endif
      GRF_efface(graphe, VRAI, FAUX); /* On efface ce graphe */
 
    }
    /* Les sries vides sont effaces automatiquement... */
  }
#endif

  return VRAI; /* On a pu effacer tous les graphes */
}
