/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2003, Emmanuel CURIS

   Fusion de plusieurs graphes
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"

/* #include "grf_fondre.txt" */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int GRF_Fondre_C(unsigned long nbr_graphes, Graphe **graphes, Graphe *resultat);
static int GRF_Fondre_R(unsigned long nbr_graphes, Graphe **graphes, Graphe *resultat);

/* ~~~~~~~~ Réalisation ~~~~~~~~~~~~~~ */

/* -------- Fusion proprement dite. ATTENTION, ne fait ni tri, ni recherche de doublons -------- */

Graphe *GRF_Fondre(unsigned long nbr_graphes, Graphe **graphes, boolean isole)
{
  int genre, type, retour __attribute__((__unused__));
  unsigned long nombre_points;
  boolean correct;
  Graphe *resultat = NULL;

  /* 1. Vérifications préliminaires */
  if ( ( NULL == graphes ) || ( 0 == nbr_graphes ) ) {
    return NULL;
  }
  if ( 1 == nbr_graphes ) {
    return graphes[ 0 ];
  }

  /* Sont-ils de meme genre [meme type de données] ? */
  correct = GRF_ControleGenre( graphes, nbr_graphes, &genre );
  if ( FAUX == correct ) {
    (void) fprintf(stderr, "GRF_Fondre : les graphes n'ont pas tous le meme genre\n");
    return NULL;
  }

  /* Sont-ils de meme type ? */
  correct = GRF_ControleType( graphes, nbr_graphes, FAUX, &type );
  if ( FAUX == correct ) {
    (void) fprintf(stderr, "GRF_Fondre : les graphes n'ont pas tous le meme type\n");
    return NULL;
  }

  /* Combien de points au total ? */
  (void) GRF_ControleTaille( graphes, nbr_graphes, GRF_CONTROLE_N_SOMME, &nombre_points, FAUX );

  /* On crée le graphe contenant le résultat */
  if ( VRAI == isole ) {
    resultat = GRF_nv_isole( nombre_points, type, genre );
  } else {
    char *titre;

    titre = TXT_genere_chaine("Fusion de %d graphes", nbr_graphes);
    resultat = GRF_nouveau( NULL, nombre_points, type, genre, titre );
    free( titre );
  }
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "GRF_Fondre : Impossible de réserver le spectre résultat\n");
    return NULL;
  }

  switch ( genre )
  {
  case GRF_FCT_REELLE:
    retour = GRF_Fondre_R( nbr_graphes, graphes, resultat );
    break;

  case GRF_FCT_COMPLEXE:
    retour = GRF_Fondre_C( nbr_graphes, graphes, resultat );
    break;

  default:
    (void) fprintf(stderr, "GRF_Fondre : Type de graphe pas encore géré [%d]\n", genre);
    break;
  }

  /* On peut maintenant rechercher les bornes... */
  GRF_bornes( resultat, DESORDONNE, DESORDONNE );
  GRF_bornes( resultat->erreur_x, DESORDONNE, DESORDONNE );
  GRF_bornes( resultat->erreur_y, DESORDONNE, DESORDONNE );

  /* La fusion s'est bien passé */
  return resultat;
}

/* -------- Fusion de graphes à valeurs réelles --------
   [Remarque : rien ne prouve que les graphes initiaux aient des X triés]
 */

static int GRF_Fondre_C(unsigned long nbr_graphes, Graphe **graphes, Graphe *resultat)
{
  unsigned long i, idx_depart = 0;
  boolean av_err_X = FAUX, av_err_Y = FAUX;

  /* On recopie toutes les valeurs dans le tableau initial */
  for ( i = 0 ; i < nbr_graphes ; i++ ) {
    memcpy( resultat->x + idx_depart, graphes[ i ]->x, sizeof( double ) * ( graphes[ i ]->nombre_point ) );
    memcpy( resultat->y + idx_depart, graphes[ i ]->y, sizeof( Complexe ) * ( graphes[ i ]->nombre_point ) );

    if ( graphes[ i ]->erreur_x != NULL ) av_err_X = VRAI;
    if ( graphes[ i ]->erreur_y != NULL ) av_err_Y = VRAI;

    idx_depart += graphes[ i ]->nombre_point;
  }

  /* On attaque, si besoin est, les barres d'erreur
     1. Erreurs sur les X  */
  if ( VRAI == av_err_X ) {
    Graphe *err_X;

    err_X = GRF_creer_erreurs( resultat, AXE_X );
    if ( NULL == err_X ) {
      (void) fprintf(stderr, "GRF_Fondre_R : Impossible de créer les erreurs en X\n");
    } else {
      for ( i = 0, idx_depart = 0 ; i < nbr_graphes ; i++ ) {
	if ( graphes[ i ]->erreur_x != NULL ) { /* Rien ne prouve que tous les graphes aient des erreurs ! */
	  memcpy( err_X->y + idx_depart, graphes[ i ]->erreur_x->y, sizeof( double ) * ( graphes[ i ]->nombre_point ) );
	  memcpy( err_X->z.ddl + idx_depart, graphes[ i ]->erreur_x->z.ddl, sizeof( int ) * ( graphes[ i ]->nombre_point ) );
	}	

	idx_depart += graphes[ i ]->nombre_point;
      }
    }
  }

  /* 2. Erreurs sur les Y  */
  if ( VRAI == av_err_Y ) {
    Graphe *err_Y;

    err_Y = GRF_creer_erreurs( resultat, AXE_Y );
    if ( NULL == err_Y ) {
      (void) fprintf(stderr, "GRF_Fondre_R : Impossible de créer les erreurs en Y\n");
    } else {
      for ( i = 0, idx_depart = 0 ; i < nbr_graphes ; i++ ) {
	if ( graphes[ i ]->erreur_y != NULL ) { /* Rien ne prouve que tous les graphes aient des erreurs ! */
	  memcpy( err_Y->y + idx_depart, graphes[ i ]->erreur_y->y, sizeof( Complexe ) * ( graphes[ i ]->nombre_point ) );
	  memcpy( err_Y->z.ddl + idx_depart, graphes[ i ]->erreur_y->z.ddl, sizeof( int ) * ( graphes[ i ]->nombre_point ) );
	}	

	idx_depart += graphes[ i ]->nombre_point;
      }
    }
  }

  /* Tout s'est bien passé */
  return 0;
}

/* -------- Fusion de graphes à valeurs réelles --------
   [Remarque : rien ne prouve que les graphes initiaux aient des X triés]
 */

static int GRF_Fondre_R(unsigned long nbr_graphes, Graphe **graphes, Graphe *resultat)
{
  unsigned long i, idx_depart = 0;
  boolean av_err_X = FAUX, av_err_Y = FAUX;

  /* On recopie toutes les valeurs dans le tableau initial */
  for ( i = 0 ; i < nbr_graphes ; i++ ) {
    memcpy( resultat->x + idx_depart, graphes[ i ]->x, sizeof( double ) * ( graphes[ i ]->nombre_point ) );
    memcpy( resultat->y + idx_depart, graphes[ i ]->y, sizeof( double ) * ( graphes[ i ]->nombre_point ) );

    if ( graphes[ i ]->erreur_x != NULL ) av_err_X = VRAI;
    if ( graphes[ i ]->erreur_y != NULL ) av_err_Y = VRAI;

    idx_depart += graphes[ i ]->nombre_point;
  }

  /* On attaque, si besoin est, les barres d'erreur
     1. Erreurs sur les X  */
  if ( VRAI == av_err_X ) {
    Graphe *err_X;

    err_X = GRF_creer_erreurs( resultat, AXE_X );
    if ( NULL == err_X ) {
      (void) fprintf(stderr, "GRF_Fondre_R : Impossible de créer les erreurs en X\n");
    } else {
      for ( i = 0, idx_depart = 0 ; i < nbr_graphes ; i++ ) {
	if ( graphes[ i ]->erreur_x != NULL ) { /* Rien ne prouve que tous les graphes aient des erreurs ! */
	  memcpy( err_X->y + idx_depart, graphes[ i ]->erreur_x->y, sizeof( double ) * ( graphes[ i ]->nombre_point ) );
	  memcpy( err_X->z.ddl + idx_depart, graphes[ i ]->erreur_x->z.ddl, sizeof( int ) * ( graphes[ i ]->nombre_point ) );
	}	

	idx_depart += graphes[ i ]->nombre_point;
      }
    }
  }

  /* 2. Erreurs sur les Y  */
  if ( VRAI == av_err_Y ) {
    Graphe *err_Y;

    err_Y = GRF_creer_erreurs( resultat, AXE_Y );
    if ( NULL == err_Y ) {
      (void) fprintf(stderr, "GRF_Fondre_R : Impossible de créer les erreurs en Y\n");
    } else {
      for ( i = 0, idx_depart = 0 ; i < nbr_graphes ; i++ ) {
	if ( graphes[ i ]->erreur_y != NULL ) { /* Rien ne prouve que tous les graphes aient des erreurs ! */
	  memcpy( err_Y->y + idx_depart, graphes[ i ]->erreur_y->y, sizeof( double ) * ( graphes[ i ]->nombre_point ) );
	  memcpy( err_Y->z.ddl + idx_depart, graphes[ i ]->erreur_y->z.ddl, sizeof( int ) * ( graphes[ i ]->nombre_point ) );
	}	

	idx_depart += graphes[ i ]->nombre_point;
      }
    }
  }

  /* Tout s'est bien passé */
  return 0;
}
