/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2002, Emmanuel CURIS

   Produit de deux graphes (grande variété de cas...)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"
#include "matrices.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int Produit_CovarGraphe(Matrice *matrice, Graphe *graphe);

static int Produit_CC(Graphe *a, Graphe *b, boolean traite_erreurs);
static int Produit_CR(Graphe *a, Graphe *b, boolean traite_erreurs);
static int Produit_RC(Graphe *a, Graphe *b, boolean traite_erreurs);
static int Produit_RR(Graphe *a, Graphe *b, boolean traite_erreurs);

/* ~~~~~~~~ Réalisation ~~~~~~~~~~~~~~ */

/* -------- Transformation de la matrice de covariance
   Suppose la matrice, existante, assortie au graphe
 -------- */

static int Produit_CovarGraphe(Matrice *matrice, Graphe *graphe)
{
  unsigned long i, j;

  if ( NULL == matrice ) return FAUX;
  if ( NULL == graphe ) return FAUX;

  if ( GRF_FCT_REELLE == graphe->genre_donnees ) {
    double *y;

    y = graphe->y;
    for ( i = 0 ; i < graphe->nombre_point ; i++, y++ ) {
      /* On multiplie la ligne i de la matrice */
      /* Moitié inférieure */
      for (j = 0 ; j <= i ; j++) {
 	matrice->valeurs[ MAT_SYM_element(i, j) ] *= *y;
      }
      /* Moitié supérieure, par symétrie revient à remplir la colonne i */
      for ( ; j < graphe->nombre_point ; j++ ) {
 	matrice->valeurs[ MAT_SYM_element(j, i) ] *= *y;
      }
    }
  } else if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
    Complexe *y;

    y = (Complexe *) graphe->y;
    for ( i = 0 ; i < graphe->nombre_point ; i++, y++ ) {
      /* On multiplie la ligne i de la matrice */
      /* Moitié inférieure */
      for (j = 0 ; j <= i ; j++) {
 	matrice->valeurs[ MAT_SYM_element(i, j) ] *= y->re;
      }
      /* Moitié supérieure, par symétrie revient à remplir la colonne i */
      for (j = i ; j < graphe->nombre_point ; j++ ) { /* Démarre à i pour traiter 2 fois la diagonale ! */
 	matrice->valeurs[ MAT_SYM_element(j, i) ] *= y->re;
      }
    }
  }

  return VRAI;
}

/* -------- Produit de deux graphes complexes -------- */

static int Produit_CC(Graphe *a, Graphe *b, boolean traite_erreurs)
{
  unsigned long i;
  Complexe *y_a, *y_b;

  if ( (NULL == a) || (NULL == b) ) return FAUX;

  /* Produit des deux, dans a */
  y_a = (Complexe *) a->y;
  y_b = (Complexe *) b->y;
  for (i = 0 ; i < a->nombre_point ; i++, y_a++, y_b++) {
    Complexe z;

    z.re = y_a->re * y_b->re - y_a->im * y_b->im;
    z.im = y_a->im * y_b->re + y_a->re * y_b->im;
    *y_a = z;
  }

  GRF_bornes(a, NE_PAS_FAIRE, DESORDONNE);
  if ( FAUX == traite_erreurs ) return VRAI;

  if ( a->erreur_y != NULL ) {
    OPTION_A_FAIRE( "Erreurs sur a, traitement des erreurs" );
  } else if ( b->erreur_y != NULL ) {
    OPTION_A_FAIRE( "Erreurs sur b, traitement des erreurs" );
  }

  return VRAI;
}

/* -------- Produit d'un graphe complexe par un graphe réel -------- */

static int Produit_CR(Graphe *a, Graphe *b, boolean traite_erreurs)
{
  unsigned long i;
  double *y_b;
  Complexe *y_a;

  if ( (NULL == a) || (NULL == b) ) return FAUX;

  /* Produit des deux, dans a */
  y_a = (Complexe *) a->y;
  for (i = 0, y_b = b->y ; i < a->nombre_point ; i++, y_a++, y_b++) {
    y_a->re *= *y_b;
    y_a->im *= *y_b;
  }

  GRF_bornes(a, NE_PAS_FAIRE, DESORDONNE);
  if ( FAUX == traite_erreurs ) return VRAI;

  OPTION_A_FAIRE( "Traitement des erreurs" );
  return VRAI;
}

/* -------- Produit réel * complexe, a réel mais en place... -------- */

static int Produit_RC(Graphe *a, Graphe *b, boolean traite_erreurs)
{
  int retour;
  unsigned long i;
  Complexe *y_a, *y_b;

  if ( (NULL == a) || (NULL == b) ) return FAUX;

  /* On convertit le graphe en graphe à valeurs complexes */
  retour = GRF_ChangerGenre( a, GRF_FCT_COMPLEXE );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de convertir A en complexe\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return FAUX;
  }

  if ( FAUX == traite_erreurs ) {
    /* Produit des deux, dans a - sachant que Im(a) = 0 */
    y_a = (Complexe *) a->y;
    y_b = (Complexe *) b->y;
    for (i = 0 ; i < a->nombre_point ; i++, y_a++, y_b++) {
      Complexe z;

      z.re = y_a->re * y_b->re;
      z.im = y_a->re * y_b->im;
      *y_a = z;
    }

    GRF_bornes(a, NE_PAS_FAIRE, DESORDONNE);
    return VRAI;
  }

  /* On s'occupe D'ABORD des barres d'erreur */

  if ( a->erreur_y != NULL ) {
    OPTION_A_FAIRE( "Erreurs sur a" );
  } else if ( b->erreur_y != NULL ) {
    Graphe *erreurs;
    Complexe *y_err_a;

    erreurs = GRF_creer_erreurs(a, AXE_Y);
    if ( NULL == erreurs ) return FAUX;

    y_a = (Complexe *) a->y;
    y_err_a = (Complexe *) erreurs->y;
    y_b = (Complexe *) b->erreur_y->y;
    for (i = 0 ; i < a->nombre_point ; i++, y_a++, y_err_a++, y_b++) {
      y_err_a->re = fabs(y_a->re) * y_b->re; /* Car Im(a) = 0 */
      y_err_a->im = fabs(y_a->re) * y_b->im;
    }

    GRF_bornes(erreurs, NE_PAS_FAIRE, DESORDONNE);
  }

  /* On s'occupe des matrices de covariance */

  if ( a->covariance != NULL ) {
    OPTION_A_FAIRE( "Covariance sur a, traitement des erreurs" );
  } else if ( b->covariance != NULL ) {
    a->covariance = copier_matrice( b->covariance );

    /* Marche car Im(a->y) = 0 de toute façon... */
    Produit_CovarGraphe(a->covariance, a);
  }

  /* On n'a plus besoin des valeurs dans a : on fait le produit, qui les remplace */

  /* Produit des deux, dans a - sachant que Im(a) = 0 */
  y_a = (Complexe *) a->y;
  y_b = (Complexe *) b->y;
  for (i = 0 ; i < a->nombre_point ; i++, y_a++, y_b++) {
    Complexe z;

    z.re = y_a->re * y_b->re;
    z.im = y_a->re * y_b->im;
    *y_a = z;
  }

  GRF_bornes(a, NE_PAS_FAIRE, DESORDONNE);

  return VRAI;
}

/* -------- Produit de deux graphes à valeurs réelles -------- */

static int Produit_RR(Graphe *a, Graphe *b, boolean traite_erreurs)
{
  unsigned long i;
  double *y_a, *y_b;

  if ( (NULL == a) || (NULL == b) ) return FAUX;

  if ( FAUX == traite_erreurs ) {
    /* Produit des deux, dans a */
    for (i = 0, y_a = a->y, y_b = b->y ; i < a->nombre_point ; i++, y_a++, y_b++) {
      *y_a *= *y_b;
    }

    GRF_bornes(a, NE_PAS_FAIRE, DESORDONNE);
    return VRAI; /* Terminé */
  }

  /* On s'occupe D'ABORD des erreurs, si besoin est */

  /* Le graphe a avait des erreurs en Y */
  if ( a->erreur_y != NULL ) {
    /* Le graphe b a aussi des erreurs en Y */
    if ( b->erreur_y != NULL ) {
      OPTION_A_FAIRE( "Barres d'erreur sur les deux graphes" );
    } else {
      y_a = a->erreur_y->y;
      y_b = b->y;

      for (i = 0 ; i < a->erreur_y->nombre_point ; i++, y_a++, y_b++) {
	*y_a *= fabs( *y_b );
      }
    }

    GRF_bornes( a->erreur_y, NE_PAS_FAIRE, DESORDONNE);
  } else {
    /* Seul le graphe b avait des erreurs en Y */
    if ( b->erreur_y != NULL ) {
      Graphe *erreurs;

      erreurs = GRF_creer_erreurs(a, AXE_Y);
      if ( NULL == erreurs ) return FAUX;

      y_a = erreurs->y;
      y_b = b->erreur_y->y;

      for (i = 0 ; i < erreurs->nombre_point ; i++, y_a++, y_b++) {
	*y_a = *y_b * fabs( (a->y)[i] );
      }

      GRF_bornes(erreurs, NE_PAS_FAIRE, DESORDONNE);
    }
  }

  /* On s'occupe des matrices de covariance, s'il y en a */
  if ( a->covariance != NULL ) {
    if ( b->covariance != NULL ) {
      OPTION_A_FAIRE( "Graphes a et b avec matrice de covariance" );
    } else {
      Produit_CovarGraphe( a->covariance, b);
    }
  } else if ( b->covariance != NULL ) {
    GRF_CreerCovariance(a);
    if ( NULL == a->covariance ) /* Impossible de créer la matrice de covariance */
      return FAUX;

    memcpy(a->covariance->valeurs, b->covariance->valeurs,
	   sizeof(double) * ( ( a->nombre_point * (a->nombre_point + 1) ) >> 1 ) );

    Produit_CovarGraphe( a->covariance, a);
  }

  /* On n'a plus besoin des valeurs dans a : on peut les remplacer par le produit */

  /* Produit des deux, dans a */
  for (i = 0, y_a = a->y, y_b = b->y ; i < a->nombre_point ; i++, y_a++, y_b++) {
    *y_a *= *y_b;
  }

  GRF_bornes(a, NE_PAS_FAIRE, DESORDONNE);

  return VRAI;
}

/* -------- Interface d'appel générale -------- */

Graphe *GRF_ProduitGraphes(Graphe *a, Graphe *b, boolean en_place, boolean traite_erreurs, struct matrice *covariance)
{
  int retour __attribute__((__unused__));

  if ( (NULL == a) || (NULL == b) ) return NULL;

  if ( a->nombre_point != b->nombre_point ) { /* Mauvais nombre de points */
    (void) fprintf(stderr, "GRF_ProduitGraphes : Nombre de points incorrect [A : %lu - B : %lu]\n", a->nombre_point, b->nombre_point);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  if ( FAUX == en_place ) {
    a = GRF_Copier( a, NULL, FAUX );
    if (NULL == a) return NULL;
  }

  /* Appel des différentes routines */
  if ( VRAI == GRF_EstFonctionReelle( a ) ) {
    if ( VRAI == GRF_EstFonctionReelle( b ) ) {
      if ( NULL == covariance )
	retour = Produit_RR(a, b, traite_erreurs);
      else {
	OPTION_A_FAIRE( "Graphes réels, matrice de covariance croisée" );
	a = NULL;
      }
    } else if ( GRF_FCT_COMPLEXE == b->genre_donnees ) {
      if ( NULL == covariance )
	retour = Produit_RC(a, b, traite_erreurs);
      else {
	OPTION_A_FAIRE( "réel × complexe, matrice de covariance croisée" );
	a = NULL;
      }
    }
  } else if ( GRF_FCT_COMPLEXE == a->genre_donnees ) {
    if ( GRF_FCT_REELLE == b->genre_donnees ) {
      retour = Produit_CR(a, b, traite_erreurs);
    } else if ( GRF_FCT_COMPLEXE == b->genre_donnees ) {
      retour = Produit_CC(a, b, traite_erreurs);
    } else {
      return NULL; /* Problème de compatibilité entre les graphes */
    }
  } else {
    (void) fprintf(stderr, "%s,%s l.%u : Graphe a de type inconnu, produit non-effectué\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  return a;
}
