/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 2003, Emmanuel CURIS

   Sauvegarde des graphes, fichiers au format interne
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef VERSION_SIMPLIFIEE_
# define _LIMITE_ 1
#endif

#include "systeme.h"
#include "bin_utils.h"

#include "serie.h"

#include "fichiers.h" /* Pour SV_EnTete() */
#include "filtrage.h" /* Pour chargement de Infos_FonctionApodisation */
#include "matrices.h" /* Pour chargement de la matrice de covariance */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define VERIFIE(n,t,r) if ( n_ecrit < (n) ) {\
                         (void) SYS_ERREUR(ERREUR_ECRITURE, "Fichier en cours", fichier, NULL); \
                         (void) fprintf(stderr, "VERIFIE, sauvegarde, %s : %s\n", __FUNCTION__, t); \
                         return (r); }

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int sv_affichage (FILE *fichier, struct grf_infos_affichage *affichage);
static int sv_covariance(FILE *fichier, Graphe *graphe);
static int sv_table_C   (FILE *fichier, unsigned long nbr_pts, Complexe *valeurs, double min, double max);
static int sv_table_I   (FILE *fichier, unsigned long nbr_pts, int *valeurs);
static int sv_table_R   (FILE *fichier, unsigned long nbr_pts, double *valeurs, double min, double max);

static int sv_graphe_de_serie(FILE *fichier, Serie_Graphes *serie, Graphe *graphe);

/* ~~~~~~~~ Réalisation pratique ~~~~~ */

/* ~-~-~-~-~-~-~-~ Sauvegarde des séries ~-~-~-~-~-~-~-~ */

/* -------- Sauvegarde d'une série, seule dans un fichier -------- */

int Serie_SauverIsolee(Serie_Graphes *serie, const char *nom_fichier, unsigned int options)
{
  FILE *fichier;
  int retour;

  if ( ( NULL == nom_fichier ) || ( NULL == serie ) ) return -1;

  /* On ouvre le fichier */
  fichier = fopen( nom_fichier, "wb" );
  if ( NULL == fichier ) {
    (void) SYS_ERREUR(ERREUR_ECRITURE, nom_fichier, NULL, NULL);
    return -1;
  }

  /* On sauvegarde l'en-tête */
  if ( SV_EnTete( fichier, FICHIER_SERIE ) == FAUX ) {
    return -2;
  }

  /* On sauvegarde la série proprement dite */
  retour = Serie_Sauver( serie, fichier, options );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur en sauvant la série [%s ; code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, serie->nom, retour );
    (void) fclose( fichier ); fichier = NULL;
    return -3;
  }

  /* Terminé ! */
  (void) fclose( fichier ); fichier = NULL;

  /* Terminé sans souci */
  return 0;
}

/* -------- Sauvegarde d'une série, comme bloc dans un fichier
     Renvoie 0 si a pu sauver, !=0 en cas de problème

   Format utilisé :
   << chaine : le nom de la série >>
   flag : 1 si table des X présente
     Si flag = 1
       nbr_pts                   - unsigned long, nombre de points
       <table des X : réels>
       flag : 1 si erreurs en X
         Si flag = 1
           < erreurs en X >      - double * nbr_pts, Y du graphe des erreurs en X
           < ddl en X >          - int * nbr_pts, ddl associées
       <<chaîne : légende en X>>
       <<chaîne : unité en X>>
   nbr_graphes                   - unsigned long, nombre de graphes
   flag : 1 si le graphe 1 utilise les X de la série
   <<graphe 1>>
   flag : 1 si le graphe 2 utilise les X de la série
   <<graphe 2>>
   ...
   flag : 1 si le graphe n utilise les X de la série
   <<graphe n>>
 */

int Serie_Sauver(Serie_Graphes *serie, FILE *fichier, unsigned int options)
{
  int retour;
  short int drapeau;
  int n_ecrit;

  if ( NULL == fichier ) return -1;
  if ( NULL == serie ) return -2;

  /* Le nom de la série : une chaîne de caractères */
  retour = BIN_SV_chaine( fichier, serie->nom );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur en sauvant le nom\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* Y a-t-il une table des X ? */
  drapeau = ( serie->table_X != NULL ) && ( VRAI == serie->drapeaux.table_X );
  n_ecrit = fwrite( &drapeau, sizeof(short int), 1, fichier );
  VERIFIE(1, "table des X, drapeau", -4);
  if ( 1 == drapeau ) {
    /* On sauve donc la table des X */
    n_ecrit = fwrite( &(serie->nbr_points), sizeof(unsigned long), 1, fichier );
    VERIFIE(1, "table des X, nombre de valeurs", -5);

    retour = sv_table_R( fichier, serie->nbr_points, serie->table_X, 0.0, 0.0 );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreur en sauvant les X\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -6;
    }

    drapeau = ( serie->erreurs_X != NULL );
    n_ecrit = fwrite( &drapeau, sizeof(short int), 1, fichier );
    VERIFIE(1, "erreurs en X, drapeau", -7);
    if ( 1 == drapeau ) {
      retour = sv_table_R( fichier, serie->nbr_points, serie->erreurs_X->y, 0.0, 0.0 );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur en sauvant les erreurs sur X\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -8;
      }

      retour = sv_table_I( fichier, serie->nbr_points, serie->erreurs_X->z.ddl );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Erreur en sauvant les ddl des erreurs sur X\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -9;
      }
    }

    retour = BIN_SV_chaine( fichier, serie->legende_X );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreur en sauvant la légende en X\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -10;
    }

    retour = BIN_SV_chaine( fichier, serie->unite_X );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : Erreur en sauvant l'unité en X\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -11;
    }
  }

  if ( 0 == options ) {
    Graphe *graphe;

    /* On sauve tous les graphes */
    n_ecrit = fwrite( &(serie->nbr_graphes), sizeof(unsigned long), 1, fichier );
    VERIFIE(1, "Nombre de graphes [toute la série]", -12);

    for ( graphe = serie->premier_graphe ; graphe != NULL ; graphe = graphe->suivant ) {
      retour = sv_graphe_de_serie( fichier, serie, graphe );
      if ( retour != 0 ) {
	(void) fprintf(stderr, "%s, %s l.%u : Sauvegarde du graphe %s\n",
		       __FILE__, __FUNCTION__, __LINE__, graphe->nom);
	return -13;
      }
    }
  } else {
    OPTION_A_FAIRE( "Sauver les graphes avec un filtre indiqué !" );
    return +255;
  }

  /* Sauvegarde du bloc finie, tout s'est bien passé */
  return 0;
}

/* -------- Sauvegarde d'un graphe, quand elle fait partie de la sauvegarde d'une série
   Isolé car revient souvent dans sv_serie() et il faut éviter de sauver inutilement les X !
   -------- */

static int sv_graphe_de_serie(FILE *fichier, Serie_Graphes *serie, Graphe *graphe)
{
  short int drapeau;
  int retour = 0, n_ecrit;

  drapeau = ( serie->table_X == graphe->x ) && ( serie->table_X != NULL );
  n_ecrit = fwrite( &drapeau, sizeof(short int), 1, fichier );
  VERIFIE(1, "table des X de la série, drapeau", -1);

  if ( 1 == drapeau ) {
    /* X donnés par la série : inutile de les donner ! */
    graphe->x = NULL; /* On les fait donc sauter le temps de la sauvegarde */
    retour = GRF_Sauver( fichier, graphe );
    graphe->x = serie->table_X;
  } else {
    /* X autonomes ou absents : sauvegarde directe */
    retour = GRF_Sauver( fichier, graphe );
  }

  return ( VRAI == retour ) ? ( 0 ) : ( -2 );
}

/* ~-~-~-~-~-~-~-~ Sauvegarde des graphes ~-~-~-~-~-~-~-~ */

/* -------- Sauvegarde d'un graphe, seul dans un fichier -------- */

int GRF_SauverIsole(const char *nom_fichier, Graphe *graphe)
{
  FILE *fichier;

  if ( ( NULL == nom_fichier ) || ( NULL == graphe ) ) return -1;

  /* On ouvre le fichier */
  fichier = fopen( nom_fichier, "wb" );
  if ( NULL == fichier ) {
    (void) SYS_ERREUR(ERREUR_ECRITURE, nom_fichier, NULL, NULL);
    return -1;
  }

  /* On sauvegarde l'en-tête */
  if ( SV_EnTete( fichier, FICHIER_GRAPHE ) == FAUX ) {
    return -2;
  }

  /* On sauvegarde le graphe proprement dit */
  if ( GRF_Sauver(fichier, graphe) == FAUX ) {
    return -3;
  }

  /* Terminé ! */
  (void) fclose( fichier ); fichier = NULL;

  /* On mentionne la sauvegarde dans l'historique, puisqu'elle a réussi */
  HIST_Titre(1, "Sauvegarde", "E/S");
  HIST_Texte(1, 0, "Fichier : %s\n", nom_fichier);
  HIST_Texte(1, 0, "Graphe sauv"e_AIGU" : %s\n", graphe->nom);

  /* Terminé sans souci */
  return 0;
}

/* -------- Sauvegarde d'un graphe
     Renvoie VRAI si a pu sauver, FAUX en cas de problème

  Format de sauvegarde, version 3 :

On a essentiellement 3 informations à la suite, classées par bloc :
 - informations générales : nombre de points,...
 - les données proprement dites
 - les informations pour l'affichage

Cela donne donc :

    nbr_pts                   - unsigned long, nombre de points
    type                      - int, type du graphe
    genre                     - int, la nature des données dans le graphe

    <table des X : réels>
    <table des Y :
       - Si genre_donnees == GRF_FCT_REELLE, GRF_SERIE, GRF_COURBE_PLANE, GRF_COURBE_ESPACE
         <table des Y : réels>
       - Si genre_donnees == GRF_FCT_COMPLEXE
         <table des Y : complexes>
    >
    <table des Z :
       - Si genre_donnees == GRF_COURBE_ESPACE
         <table des Z : réels>
    >
    <Informations complémentaires :
       - Si type == GR_TF
         flag : 1 si infos de fenêtrage associées - short int
         Si flag = 1
          k_min, k_max        - double, limites de la fenêtre
	  paramètre           - double, paramètre de la fenêtre
	  type                - int, type de fenêtre
          k_1, k_2            - double, bornes internes de la fenêtre
	  symétrique          - boolean, caractère symétrique de la fenêtre
       - Si type == GR_TFI ou GR_EXAFS
         flag : 1 si infos de nyquist associées - short int
         Si flag = 1
          r_min, r_max        - double, limites du filtre
    >
    flag : 1 si erreurs en X
      Si flag = 1
        < erreurs en X >      - double * nbr_pts, Y du graphe des erreurs en X
        < ddl en X >          - int * nbr_pts, ddl associées
    flag : 1 si erreurs en Y
      Si flag = 1
        Si genre_donnees == GRF_FCT_COMPLEXE
            <table des erreurs en Y>  - Complexe * nbr_pts, Y du graphe des erreurs en Y
	Sinon
            <table des erreurs en Y>  - double * nbr_pts, Y du graphe des erreurs en Y
            <table des ddl en Y>      - int * nbr_pts, ddl associées
    <matrice de covariance>
    flag : spectre à X triés ?     ** VERSION 3.6
    flag : spectre sans doublons ? ** VERSION 3.6
    flag : spectre théorique ?     ** VERSION 3.6
    flag : spectre normalisé ?     ** VERSION 3.6

    <<chaîne : le nom du graphe>>
    <<affichage>>
*/

int GRF_Sauver(FILE *fichier, Graphe *graphe)
{
  int retour;
  short int drapeau;
  size_t n_ecrit;

  if (fichier == NULL) return FAUX;
  if (graphe == NULL) return FAUX;

  /* "en-tête" : informations sur ce que contient le graphe */
  n_ecrit = fwrite(&(graphe->nombre_point), sizeof(unsigned long), 1, fichier); /* Nombre de points */
  VERIFIE(1, "GRF_Sauver : nombre de points", FAUX);
  n_ecrit = fwrite(&(graphe->type), sizeof(int), 1, fichier);                   /* Type du graphe   */
  VERIFIE(1, "GRF_Sauver : type de graphe", FAUX);
  n_ecrit = fwrite(&(graphe->genre_donnees), sizeof(int), 1, fichier);
  VERIFIE(1, "GRF_Sauver : type de données", FAUX);

  /* Données numériques */
  retour = sv_table_R( fichier, graphe->nombre_point, graphe->x, graphe->x_min, graphe->x_max);
  if ( retour != 0 ) {
    (void) fprintf(stderr, "GRF_Sauver : ERREUR dans sv_table_R [X]\n");
    return FAUX;
  }
  switch ( graphe->genre_donnees )
  {
  case GRF_FCT_REELLE:
  case GRF_SERIE:
  case GRF_COURBE_PLANE:
    retour = sv_table_R( fichier, graphe->nombre_point, graphe->y, graphe->y_min, graphe->y_max);
    break;

  case GRF_FCT_COMPLEXE:
    retour = sv_table_C( fichier, graphe->nombre_point, (Complexe *) graphe->y, graphe->y_min, graphe->y_max);
    break;

  case GRF_COURBE_ESPACE:
  case GRF_SURFACE:
    retour = sv_table_R( fichier, graphe->nombre_point, graphe->y, graphe->y_min, graphe->y_max);
    retour = sv_table_R( fichier, graphe->nombre_point, graphe->z.z, 0.0, 0.0);
    break;

  case GRF_HISTOGRAMME:
    OPTION_A_FAIRE( "Sauver les Y d'un histogramme" );
    break;

  case GRF_FAISCEAU:
    OPTION_A_FAIRE( "Sauver les Y d'un faisceau de graphes" );
    break;

  case GRF_MULTI_SERIE:
    OPTION_A_FAIRE( "Sauver les Y d'un graphe multi-séries" );
    break;

  case GRF_ERREUR_R:
    retour = sv_table_R( fichier, graphe->nombre_point, graphe->y, graphe->y_min, graphe->y_max );
    retour = sv_table_I( fichier, graphe->nombre_point, graphe->z.ddl );
    break;

  case GRF_ERREUR_C:
    retour = sv_table_C( fichier, graphe->nombre_point, (Complexe *) graphe->y, graphe->y_min, graphe->y_max );
    retour = sv_table_I( fichier, graphe->nombre_point, graphe->z.ddl );
    break;

  default:
    (void) fprintf(stderr, "GRF_Sauver : type de données inconnu [%d]\n", graphe->genre_donnees );
    retour = -5;
  }
  if ( retour != 0 ) {
    (void) fprintf(stderr, "GRF_Sauver : ERREUR dans la sauvegarde des Y\n");
    return FAUX;
  }

  /* Informations complémentaires, en fonction du type de grandeur */
  switch ( graphe->type )
  {
  case GR_DEVIATION:
    retour = sv_table_I( fichier, graphe->nombre_point, graphe->z.ddl );

  case GR_TF: /* Informations sur la fenêtre */
    drapeau = (graphe->pointeur.infos_tf != NULL);
    n_ecrit = fwrite(&drapeau, sizeof( short int ), 1, fichier);
    VERIFIE(1, "GRF_Sauver : GR_TF, drapeau", FAUX);
    if ( drapeau ) {
      Infos_FonctionApodisation *infos = graphe->pointeur.infos_tf;

      n_ecrit  = fwrite( &(infos->x_min)     , sizeof(double) , 1, fichier);
      n_ecrit += fwrite( &(infos->x_max)     , sizeof(double) , 1, fichier);
      n_ecrit += fwrite( &(infos->parametre) , sizeof(double) , 1, fichier);
      n_ecrit += fwrite( &(infos->type)      , sizeof(int) , 1, fichier);
      n_ecrit += fwrite( &(infos->x_1)       , sizeof(double) , 1, fichier);
      n_ecrit += fwrite( &(infos->x_2)       , sizeof(double) , 1, fichier);
      n_ecrit += fwrite( &(infos->symetrique), sizeof(boolean), 1, fichier);
      VERIFIE(7, "GRF_Sauver : GR_TF, infos de fenêtre", FAUX);
    }
    break;

  case GR_TFI: /* Informations sur le filtre */
  case GR_EXAFS:
    drapeau = ( graphe->pointeur.nyquist != NULL ) ;
    n_ecrit = fwrite(&drapeau, sizeof( short int ), 1, fichier);
    VERIFIE(1, "GRF_Sauver : GR_TFI / GR_EXAFS, drapeau", FAUX);
    if ( drapeau ) {
      Infos_Nyquist *infos = graphe->pointeur.nyquist;

      n_ecrit  = fwrite( &(infos->r_min), sizeof(double) , 1, fichier);
      n_ecrit += fwrite( &(infos->r_max), sizeof(double) , 1, fichier);
      VERIFIE(2, "GRF_Sauver : GR_TFI / GR_EXAFS, infos de filtrage", FAUX);
    }
    break;
  }

  /* On s'occupe des barres d'erreur, si nécessaire
     Principe : il y a _toujours_ autant de points dans les erreurs que dans le graphe
              et les X sont toujours les même
     Il suffit donc de sauver les Y des spectres de barre d'erreur
  */

  drapeau = (graphe->erreur_x != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(short int), 1, fichier); /* Il y a des erreurs en X */
  VERIFIE(1, "GRF_Sauver : erreurs en X, drapeau", FAUX);
  if ( drapeau == 1 ) { /* Il y a des erreurs en X */
    Graphe *erreur = graphe->erreur_x;

    retour = sv_table_R( fichier, erreur->nombre_point, erreur->y, erreur->y_min, erreur->y_max );
    retour = sv_table_I( fichier, erreur->nombre_point, erreur->z.ddl );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "GRF_Sauver : erreur / erreurs sur les X\n");
      return FAUX;
    }

    graphe->erreur_x->drapeaux.sauve = VRAI;
  }

  drapeau = (graphe->erreur_y != NULL);
  n_ecrit = fwrite(&drapeau, sizeof(short int), 1, fichier); /* Il y a des erreurs en X */
  VERIFIE(1, "GRF_Sauver : erreurs en Y, drapeau", FAUX);
  if ( drapeau == 1 ) { /* Il y a des erreurs en Y */
    Graphe *erreur = graphe->erreur_y;

    if ( GRF_FCT_COMPLEXE == graphe->genre_donnees ) {
      retour = sv_table_C( fichier, graphe->nombre_point, (Complexe *) erreur->y, erreur->y_min, erreur->y_max);
    } else {
      retour = sv_table_R( fichier, graphe->nombre_point, erreur->y, erreur->y_min, erreur->y_max);
      retour = sv_table_I( fichier, erreur->nombre_point, erreur->z.ddl );
    }

    if ( retour != 0 ) {
      (void) fprintf(stderr, "GRF_Sauver : erreur / erreurs sur les Y\n");
      return FAUX;
    }

    graphe->erreur_y->drapeaux.sauve = VRAI;
  }

  /* Matrice de covariance du graphe, si elle existe */
  sv_covariance(fichier, graphe); /* S'occupe du flag pour indiquer la présence... */

  /* Drapeaux des propriétés du graphe [VERSION 3.6] */
  drapeau = graphe->drapeaux.X_tries;
  n_ecrit = fwrite( &drapeau, sizeof(short int), 1, fichier );

  drapeau = graphe->drapeaux.X_uniques;
  n_ecrit += fwrite( &drapeau, sizeof(short int), 1, fichier );

  drapeau = graphe->drapeaux.theorique;
  n_ecrit += fwrite( &drapeau, sizeof(short int), 1, fichier );

  drapeau = graphe->drapeaux.normalise;
  n_ecrit += fwrite( &drapeau, sizeof(short int), 1, fichier );

  VERIFIE(4, "GRF_Sauver : drapeaux du graphe", FAUX);

  /* Titre du graphe */
  BIN_SV_chaine( fichier, graphe->nom );

  /* Paramètres d'affichage */
  retour = sv_affichage( fichier, graphe->affichage );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "GRF_Sauver : erreur / paramètres d'affichage\n");
    return FAUX;
  }

  /* Si on est là, tout s'est bien passé, le graphe est sauvé */
  graphe->drapeaux.sauve = VRAI;
  HIST_Texte(4, 2, " ... graphe : %s\n", graphe->nom);

  return VRAI;
}

/* -------- Sauvegarde des informations d'affichage du graphe
    Renvoie 0 si tout s'est bien passé, != 0 en cas de problème

  Format de sauvegarde :
    drapeau : 1 si les paramètres sont présents --- short int
     Si drapeau == 1
       <<chaîne : légende en X>>
       <<chaîne : légende en Y>>
       couleur  : int, couleur de dessin
       style    : short int, style de dessin  ** VERSION 3.6 (avant : int)
       marque   : short int, marque de dessin ** VERSION 3.6
       <<chaine : unité en X>> ] Depuis la version 3.4 (avant : deux entiers)
       <<chaine : unité en Y>> ]
 */

static int sv_affichage(FILE *fichier, struct grf_infos_affichage *affichage)
{
  int retour __attribute__((__unused__));
  short int drapeau;
  size_t n_ecrit;

  drapeau = ( NULL != affichage );
  n_ecrit = fwrite( &drapeau, sizeof( short int ), 1, fichier );
  VERIFIE(1, "sv_affichage : drapeau de présence", -1);

  if ( NULL == affichage ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "sv_affichage : Attention, graphe sans paramètre d'affichage !\n");
#endif
    return 0;
  }

  retour = BIN_SV_chaine( fichier, affichage->axe_x ); /* Légende en X */
  retour = BIN_SV_chaine( fichier, affichage->axe_y ); /* Légende en Y */
  n_ecrit  = fwrite( &(affichage->couleur), sizeof(int), 1, fichier ); /* Couleur */
  n_ecrit += fwrite( &(affichage->style), sizeof(short int), 1, fichier ); /* Style */
  n_ecrit += fwrite( &(affichage->marque), sizeof(short int), 1, fichier ); /* Marque */
  VERIFIE( 2, "sv_affichage : couleurs, style", -4 );

  retour = BIN_SV_chaine( fichier, affichage->unite_x ); /* Unité en X */
  retour = BIN_SV_chaine( fichier, affichage->unite_y ); /* Unité en Y */

  /* Pas de problème */
  return 0;
}

/* -------- Sauvegarde de la matrice de covariance
     Renvoie 0 si tout s'est bien passé, != 0 en cas de problème

  Format de sauvegarde :
    - short int : VRAI si une matrice de covariance existe
    - (double)  : On utilise la symétrie de la matrice pour le stockage : il y a donc n(n+1)/2 points
                 dans le tableau, que l'on sauve brutalement pour ne pas se compliquer la tâche
*/

static int sv_covariance(FILE *fichier, Graphe *graphe)
{
  short int drap_covar = FAUX;
  unsigned long n_reel;
  size_t n_ecrit;

  if (fichier == NULL) return FAUX;
  if (graphe == NULL) return FAUX;

  if ( NULL == graphe->covariance ) {
/*     printf("sv_covariance / %s : pas de matrice de covariance\n", graphe->nom); */
    n_ecrit = fwrite( &drap_covar, sizeof(short int), 1, fichier);
    VERIFIE(1, "sv_covariance", -1);
    return VRAI;
  }

  drap_covar = VRAI;
  n_ecrit = fwrite( &drap_covar, sizeof(short int), 1, fichier);
  VERIFIE(1, "sv_covariance", -2);

  n_reel = graphe->covariance->n_lignes;
  n_reel = ( n_reel * (n_reel + 1) ) >> 1;
  n_ecrit = fwrite( graphe->covariance->valeurs, sizeof(double), n_reel, fichier);
  VERIFIE(n_reel, "sv_covariance", -3);

  return VRAI;
}

/* -------- Sauvegarde d'une table de valeurs complexes --------
   Renvoie 0 si tout s'est bien passé, != 0 en cas d'erreur

   Format [ref. : v. 3.0] :
   nbr_pts       : unsigned long    --- nombre de valeurs dans la table
   Si nbr_pts > 0
     << valeurs >> : Complexe * nbr_pts --- la suite de valeurs
     min           : double           --- borne inférieure
     max           : double           --- borne supérieure
*/

static int sv_table_C(FILE *fichier, unsigned long nbr_pts, Complexe *valeurs, double min, double max)
{
  size_t n_ecrit;

  if ( NULL == fichier )
    return -1;

  if ( ( NULL == valeurs ) || ( 0 == nbr_pts ) ) {
    nbr_pts = 0;
    n_ecrit = fwrite( &nbr_pts, sizeof( unsigned long ), 1, fichier );
    VERIFIE(1, "sv_table_C : pas de table", -1);

    /* Pas de table : rien d'autre à faire... */
    return 0;
  }

  n_ecrit = fwrite( &nbr_pts, sizeof( unsigned long ), 1, fichier );
  VERIFIE(1, "sv_table_C : nombre de points", -2);

  n_ecrit = fwrite( valeurs, sizeof( Complexe ), nbr_pts, fichier );
  VERIFIE(nbr_pts, "sv_table_C : valeurs", -3);

  n_ecrit = fwrite( &min, sizeof( double ), 1, fichier );
  VERIFIE(1, "sv_table_C : minimum", -4);
  n_ecrit = fwrite( &max, sizeof( double ), 1, fichier );
  VERIFIE(1, "sv_table_C : maximum", -5);

  /* Tout s'est bien passé */ 
  return 0;
}

/* -------- Sauvegarde d'une table de valeurs entières --------
   Renvoie 0 si tout s'est bien passé, != 0 en cas d'erreur

   Format [ref. : v. 3.0] :
   nbr_pts       : unsigned long    --- nombre de valeurs dans la table
   Si nbr_pts > 0
     << valeurs >> : Complexe * nbr_pts --- la suite de valeurs
*/

static int sv_table_I(FILE *fichier, unsigned long nbr_pts, int *valeurs)
{
  size_t n_ecrit;

  if ( NULL == fichier )
    return -1;

  if ( ( NULL == valeurs ) || ( 0 == nbr_pts ) ) {
    nbr_pts = 0;
    n_ecrit = fwrite( &nbr_pts, sizeof( unsigned long ), 1, fichier );
    VERIFIE(1, "sv_table_I : pas de table", -1);

    /* Pas de table : rien d'autre à faire... */
    return 0;
  }

  n_ecrit = fwrite( &nbr_pts, sizeof( unsigned long ), 1, fichier );
  VERIFIE(1, "sv_table_I : nombre de points", -2);

  n_ecrit = fwrite( valeurs, sizeof( int ), nbr_pts, fichier );
  VERIFIE(nbr_pts, "sv_table_I : valeurs", -3);

  /* Tout s'est bien passé */ 
  return 0;
}

/* -------- Sauvegarde d'une table de valeurs réelles --------
   Renvoie 0 si tout s'est bien passé, != 0 en cas d'erreur

   Format [ref. : v. 3.0] :
   nbr_pts       : UNSIGNED LONG    --- nombre de valeurs dans la table
   Si nbr_pts > 0
     << valeurs >> : DOUBLE * nbr_pts --- la suite de valeurs
     min           : DOUBLE           --- borne inférieure
     max           : DOUBLE           --- borne supérieure
*/

static int sv_table_R(FILE *fichier, unsigned long nbr_pts, double *valeurs, double min, double max)
{
  size_t n_ecrit;

  if ( NULL == fichier )
    return -1;

  if ( ( NULL == valeurs ) || ( 0 == nbr_pts ) ) {
    nbr_pts = 0;
    n_ecrit = fwrite( &nbr_pts, sizeof( unsigned long ), 1, fichier );
    VERIFIE(1, "sv_table_R : pas de table", -1);

    /* Pas de table : rien d'autre à faire... */
    return 0;
  }

  n_ecrit = fwrite( &nbr_pts, sizeof( unsigned long ), 1, fichier );
  VERIFIE(1, "sv_table_R : nombre de points", -2);

  n_ecrit = fwrite( valeurs, sizeof( double ), nbr_pts, fichier );
  VERIFIE(nbr_pts, "sv_table_R : valeurs", -3);

  n_ecrit = fwrite( &min, sizeof( double ), 1, fichier );
  VERIFIE(1, "sv_table_R : minimum", -4);
  n_ecrit = fwrite( &max, sizeof( double ), 1, fichier );
  VERIFIE(1, "sv_table_R : maximum", -5);

  /* Tout s'est bien passé */ 
  return 0;
}
