/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2001, Emmanuel CURIS
   Cration : 11 avril 2001

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Recherche des bornes
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "graphes.h"
#include "images.h"
#include "echantillon.h" /* Pour les compositions ! */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int IMG_BornesZ_R(Image *image);
static int IMG_BornesZ_G(Image *image);
static int IMG_BornesZ_C(Image *image);
static int IMG_BornesZ_Compo(Image *image);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Recherche des valeurs extrmes d'une image [axe Z] -------- */

void IMG_bornes(Image *image)
{
  int retour = -4;

  if (image == NULL) return;

  switch ( image->genre )
  {
  case IMG_GENRE_REELLE:
    retour = IMG_BornesZ_R( image );
    break;

  case IMG_GENRE_COMPLEXE:
    retour = IMG_BornesZ_C( image );
    break;

  case IMG_GENRE_GRAPHE:
    retour = IMG_BornesZ_G( image );
    break;

  case IMG_GENRE_DIAGRAMME: /* Arbitraire : de toute faon, non-significatif */
  case IMG_GENRE_PHOTO:
    image->val_min = 0.0;
    image->val_max = 1.0;
    retour = 0;
    break;

  case IMG_GENRE_COMPO:
    retour = IMG_BornesZ_Compo( image );
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Genre inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, image->genre);
  }

  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans la recherche des bornes [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

#ifdef _DEVELOPPE_ /* Quelques contrles par scurit */
  if ( image->val_min == image->val_max ) {
    (void) fprintf(stderr, "%s,%s l.%u : Image sans contraste [\"%s\"]\n",
		   __FILE__, __FUNCTION__, __LINE__, image->nom);
  }

  if ( ( is_inf( image->val_min ) ) || ( is_inf( image->val_max ) ) ) {
    unsigned int i, j;
    unsigned long nbr_cases;
    double *valeurs;

    nbr_cases = image->larg * (unsigned long) image->haut;
    (void) fprintf(stderr, "%s,%s l.%u : VALEURS INFINIES dans l'image \"%s\" [min=%g max=%g pour %lu cases]\n",
		   __FILE__, __FUNCTION__, __LINE__, image->nom, image->val_min, image->val_max, nbr_cases);

    valeurs = image->valeurs.R;
    for( i = 0 ; i < image->larg ; i++ ) {
      for ( j = 0 ; j < image->haut ; j++ ) {
	IMG_IJ_v_N( image, i, j, &nbr_cases );
	printf("%g=%g ", image->valeurs.R[ nbr_cases ], *valeurs);
	valeurs++;
      }
      printf("\n");
    }

    abort();
  }
#endif
}

/* -------- Axe Z : image relle -------- */

static int IMG_BornesZ_R(Image *image)
{
  unsigned long nbr_cases, i;
  double *valeurs;

  nbr_cases = image->larg * (unsigned long) image->haut;
  valeurs = image->valeurs.R;

  image->val_min = *valeurs;
  image->val_max = *valeurs;

  for (i = 1, valeurs++ ; i < nbr_cases ; i++, valeurs++) {
    if (*valeurs > image->val_max ) image->val_max = *valeurs;
    if (*valeurs < image->val_min ) image->val_min = *valeurs;
  }

  return 0;
}

/* -------- Axe Z : image de graphes -------- */

static int IMG_BornesZ_G(Image *image)
{
  double y_min, y_max;
  struct graphe **valeurs;
  unsigned long nbr_cases, i;

  nbr_cases = image->larg * (unsigned long) image->haut;
  valeurs = image->valeurs.G;

  /* On cherche le premier graphe prsent */
  i = 0;
  while ( *valeurs == NULL ) {
    valeurs++;
    if ( i >= nbr_cases ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : Aucun graphe !\n",
		     __FILE__, __FUNCTION__, __LINE__);
#endif
      image->val_min = 0.0;
      image->val_max = 0.0;

      return +1; /* Erreur, mais mineure */
    }
  }

  y_min = (*valeurs)->y_min; y_max = (*valeurs)->y_max;

  for ( ; i < nbr_cases ; i++, valeurs++ ) {
    if ( *valeurs != NULL ) {
      if ( (*valeurs)->y_min < y_min ) y_min = (*valeurs)->y_min;
      if ( (*valeurs)->y_max > y_max ) y_max = (*valeurs)->y_max;
    }
  }

  image->val_min = y_min;
  image->val_max = y_max;

  return 0;
}

/* -------- Axe Z : image complexe -------- */

static int IMG_BornesZ_C(Image *image)
{
  unsigned long nbr_cases, i;
  double module, re_min, im_min, mod_max;
  Complexe *valeurs;

  nbr_cases = image->larg * (unsigned long) image->haut;
  valeurs = image->valeurs.C;

  re_min = valeurs->re; im_min = valeurs->im;
  mod_max = re_min * re_min + im_min * im_min;

  for (i = 1, valeurs++ ; i < nbr_cases ; i++, valeurs++) {

    if ( valeurs->re < re_min ) re_min = valeurs->re;
    if ( valeurs->im < im_min ) im_min = valeurs->im;

    module = valeurs->re * valeurs->re + valeurs->im * valeurs->im;
    if ( module > mod_max ) mod_max = module;
  }

  image->val_min = ( re_min < im_min ) ? ( re_min ) : ( im_min );
  image->val_max = sqrt( mod_max );

  return 0;
}

/* -------- Axe Z : image de compositions - */

static int IMG_BornesZ_Compo(Image *image)
{
  unsigned long i, nbr_cases, i_compo, nbr_compos;
  struct matiere **valeurs, **table_compos;

  /* Il est toujours possible d'avoir un trou */
  image->val_min = 0.0;

  /* Combien de compositions disponibles au total ? */
  nbr_compos = COMPOSITION_Compter();
  if ( 0 == nbr_compos ) {
    image->val_max = 1.0; /* Une seule composition = TROU */
    return 0;
  }

  /* On prpare la table  remplir au fur et  mesure */
  table_compos = (struct matiere **) calloc( nbr_compos, sizeof(struct matiere *) );
  if ( NULL == table_compos ) {
    ERR_Memoire( "Table des compositions", nbr_compos * sizeof(struct matiere *) );
    image->val_max = 1.0; /* Une seule composition = TROU */
    return -1; /* Erreur ! */
  }

  /* On parcourt l'image et, pour chaque case, on vrifie si la composition existe dj */
  nbr_cases = image->larg * (unsigned long) image->haut;
  i_compo = 0; /* 0 = NULL = trou, toujours compt */
  for ( i = 0, valeurs = image->valeurs.M ; i < nbr_cases ; i++, valeurs++ ) {
    struct matiere *composition;
    unsigned long j;
    boolean comptee = FAUX;

    composition = *valeurs;
    if ( NULL == composition ) continue;

    /* On cherche si la composition est dj dans la table */
    for ( j = 0 ; j < i_compo ; j++ ) {
      if ( table_compos[ j ] == composition ) { /* La composition a dj t compte ! */
#ifdef _DEVELOPPE_
/* 	printf("%s : case %lu, composition %s dj trouve\n", __FUNCTION__, i, composition->nom ); */
#endif
	comptee = VRAI;
	break;
      }
    }

    if ( FAUX == comptee ) {
      /* Si l'on est l : nouvelle composition -> on la met dans la table */
      if ( i_compo >= nbr_compos ) {
	(void) fprintf(stderr, "%s, %s l.%u : Plus de compositions diffrentes que disponible !\n",
		       __FILE__, __FUNCTION__, __LINE__);
	FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	free( table_compos ); table_compos = NULL;
	return -1;
      }

      table_compos[ i_compo ] = composition;
      i_compo++;
    }
  }

  /* Plus besoin de la table des compositions */
  free( table_compos ); table_compos = NULL;

  /* i_compo contient le nombre de compositions diffrentes, sans les trous */
  image->val_max = 1.0 + (double) i_compo;

  /* Termin, on est au point */
  return 0;
}

/* ~-~-~-~-~-~-~-~- Axes X/Y ~-~-~-~-~-~-~-~- */

/* -------- Modifier les valeurs des chelles -------- */

int IMG_ChangerEchelle(Image *image, int axes, double echelle_X, double echelle_Y)
{
  if ( NULL == image ) return -1;

#ifdef _DEVELOPPE_
  if ( ( axes & AXE_X ) && ( echelle_X <= 0.0 ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : chelle absurde en X [%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, echelle_X);
    abort();
  }
  if ( ( axes & AXE_Y ) && ( echelle_Y <= 0.0 ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : chelle absurde en Y [%g]\n",
		   __FILE__, __FUNCTION__, __LINE__, echelle_Y);
    abort();
  }
#endif

  if ( axes & AXE_X ) {
    image->l_reel = echelle_X;
  }
  if ( axes & AXE_Y ) {
    image->h_reel = echelle_Y;
  }

  /* Si les bornes ont chang et s'il s'agit de l'image courante, on adapte et on redessine ! */
  if ( ( image == image_courante ) && ( fen_images != NULL ) ) {
    IMG_AdapterAffichage( NULL ); /* Par dfaut, le fait pour l'image courante... */

    /* On force le redessin */
    FEN_redessin(fen_images, BI_RDS_TOUT);
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Modifier les valeurs du repre chelles -------- */

int IMG_ChangerRepere(Image *image, int axes, double x_min, double y_min)
{
  if ( NULL == image ) return -1;

  if ( axes & AXE_X ) {
    image->x_min = x_min;
  }
  if ( axes & AXE_Y ) {
    image->y_min = y_min;
  }

  /* Si les bornes ont change et s'il s'agit de l'image courante, on adapte et on redessine ! */
  if ( ( image == image_courante ) && ( fen_images != NULL ) ) {
    IMG_AdapterAffichage( NULL ); /* Par dfaut, le fait pour l'image courante */

    /* On force le redessin */
    FEN_redessin(fen_images, BI_RDS_TOUT);
  }

  /* Tout s'est bien pass */
  return 0;
}
