/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2001, Emmanuel CURIS
   Cration : 11 avril 2001

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Gnration d'images,  partir de profils,... par exemple
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"

#include "c_formel.h"
#include "exafs.h"
#include "images.h"
#include "lissage.h"
#include "science.h" /* Pour M_PI, si ncessaire */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_TXT_NOM_PROFIL  44 /* Noms apparaissant dans l'arbre */
#define LGMAX_TXT_NOM_CONTOUR 42

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void prepare_arbre_profil(void);

static Image *genere_radial (Graphe *profil, double pas_x, double pas_y);
static Image *genere_ellipse(Graphe *profil, float pas_x, float pas_y, float rapport, float angle);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

Image *IMG_generer(Graphe *profil)
{
  Image *image = NULL;

  if (profil == NULL) return NULL;
  if (profil->type != GR_PROFIL) return NULL;

  if (arbres[PROFIL_V_IMAGE].drapeaux.prepare == FAUX)
    prepare_arbre_profil();
  met_chaine(PROFIL_V_IMAGE, PVI_PROFIL, LGMAX_TXT_NOM_PROFIL, profil->nom, TXT_COPIER);
  met_chaine(PROFIL_V_IMAGE, PVI_CONTOUR, LGMAX_TXT_NOM_CONTOUR, "Pas de contour", TXT_COPIER);

  if ( ARB_Bloquant( PROFIL_V_IMAGE, "Gnrer une image depuis un profil", principale, 0 ) == PVI_ANNULER)
    return NULL;

  if (OBJ_choisit(PROFIL_V_IMAGE, PVI_RADIAL, QUESTION) == VRAI) {

    if (OBJ_choisit(PROFIL_V_IMAGE, PVI_CERCLE, QUESTION) == VRAI) {
      image = genere_radial(profil, 1.0, 1.0);
    }
    else if (OBJ_choisit(PROFIL_V_IMAGE, PVI_ELLIPSE, QUESTION) == VRAI) {
      image = genere_ellipse(profil, 1.0, 1.0, 
			     OBJ_LireNombre(PROFIL_V_IMAGE, PVI_ELLIPCITE, NULL ), OBJ_LireNombre(PROFIL_V_IMAGE, PVI_ANGLE, NULL ));
    }
  }


  return image;
}

/* -------- Objet de section elliptique, profil radial connu --------
   On suppose que le profil connu correspond  une coupe suivant le demi-grand axe "horizontal" a
   rapport donne le rapport des longueurs des deux axes, angle l'inclinaison de l'ellipse par rapport  l'horizontale
*/

static Image *genere_ellipse(Graphe *profil, float pas_x, float pas_y, float rapport, float angle)
{
  int retour;
  double lg_proj, ht_proj; /* Dimensions de la projection de l'ellipse */
  double *c, *d, *b; /* Lissage du profil, par spline */
  unsigned int larg, haut, i, j;
  double x, y, c_angle, s_angle;
  Image *image = NULL;
  
  if (profil == NULL) return NULL;

  retour = SPL_tt_prep(profil, &b, &c, &d);
  if ( retour != 0) {
    return NULL;
  }

  c_angle = cos(angle * M_PI / 180.0);
  s_angle = sin(angle * M_PI / 180.0);
  lg_proj = profil->x_max * sqrt( c_angle * c_angle + rapport * rapport * s_angle * s_angle);
  ht_proj = profil->x_max * sqrt( s_angle * s_angle + rapport * rapport * c_angle * c_angle);
  larg = (unsigned int) rint( (2.0 * lg_proj + 1.0) / pas_x );
  haut = (unsigned int) rint( (2.0 * ht_proj + 1.0) / pas_y );
  image = IMG_CreerImage( NULL, NULL, IMG_GENRE_REELLE, IMG_TYPE_CARTOGRAPHIE, larg, haut);
  if (image == NULL) {
    free(b); free(c); free(d);
    return NULL;
  }
  image->l_reel = pas_x;
  image->h_reel = pas_y;

  image->x_min = -(image->larg * image->l_reel) / 2.0;
  image->y_min = -(image->haut * image->h_reel) / 2.0;

  /* Calcul des points de l'image */
  for (i = 0 ; i < larg ; i++) {
    x = pas_x * (0.5 + (double) i)/* - image->x_centre*/;
    for (j = 0 ; j < haut ; j++) {
      double dist, X, Y;
      float valeur;
      unsigned long n_case;

      y = pas_y * (0.5 + (double) j)/* - image->y_centre*/; /* C'est le centre de la case qui correspond  la valeur */
      /* On passe dans le repre des axes de l'ellipse */
      X =  x * c_angle + y * s_angle;
      Y = -x * s_angle + y * c_angle;

      /* On transforme l'ellipse en cercle : contraction de l'axe Y */
      Y /= rapport;

      dist = X * X + Y * Y;
      dist = sqrt(dist);

      if ( (dist < profil->x_min) || (dist > profil->x_max) )
	valeur = 0.0;
      else
	valeur = (float) SPL_tt_calc(profil, dist, b, c, d);

      IMG_IJ_v_N( image, i, j, &n_case );
      image->valeurs.R[ n_case ] = valeur;
    }
  }

  IMG_AssocierCoupe( image, profil, image->x_min + image->l_reel * (image->larg / 2.0), image->y_min + image->h_reel * (image->haut / 2.0),
		     image->x_min + image->l_reel * image->larg, image->y_min + image->h_reel * (image->haut / 2.0) );
  IMG_bornes(image);
  free(b); free(c); free(d);
  return image;
}

/* -------- Objet de section circulaire, profil radial connu -------- */

static Image *genere_radial(Graphe *profil, double pas_x, double pas_y)
{
  int retour;
  double *c, *d, *b; /* Lissage du profil, par spline */
  unsigned int larg, haut, i, j;
  double x, y, lg_profil;
  Image *image = NULL;

  if (profil == NULL) return NULL;

  retour = SPL_tt_prep(profil, &b, &c, &d);
  if ( retour != 0) {
    (void) fprintf(stderr, "genere_radial : impossible de crer la table des coefficients\n");
    return NULL;
  }

  /* Longueur du profil : donne le rayon maximal */
  lg_profil = profil->x_max - profil->x_min;

  larg = (unsigned int) rint( (2.0 * lg_profil + 1.0 ) / pas_x );
  haut = (unsigned int) rint( (2.0 * lg_profil + 1.0 ) / pas_y );
  image = IMG_CreerImage( NULL, NULL, IMG_GENRE_REELLE, IMG_TYPE_CARTOGRAPHIE, larg, haut);
  if (image == NULL) {
    (void) fprintf(stderr, "genere_radial : impossible de crer l'image\n");
    free( b ); free( c ); free( d );
    return NULL;
  }

  image->l_reel = pas_x;
  image->h_reel = pas_y;

  image->x_min = -(image->larg * image->l_reel) / 2.0;
  image->y_min = -(image->haut * image->h_reel) / 2.0;

  /* Calcul des points de l'image */
  for (i = 0 ; i < larg ; i++) {
    x = pas_x * ((double) i) + image->x_min;
    for (j = 0 ; j < haut ; j++) {
      double dist;
      float valeur;
      unsigned long n_case;

      y = pas_y * ((double) j) + image->y_min; /* C'est le centre de la case qui correspond  la valeur */
      dist = x * x + y * y;
      dist = sqrt(dist);
      dist += profil->x_min; /* Par supposition, le profil commence en 0 */

      if ( dist > profil->x_max )
	valeur = 0.0;
      else
	valeur = (float) SPL_tt_calc(profil, dist, b, c, d);

      IMG_IJ_v_N( image, i, j, &n_case );
      image->valeurs.R[ n_case ] = valeur;
    }
  }

  IMG_AssocierCoupe( image, profil, image->x_min + image->l_reel * (image->larg / 2.0), image->y_min + image->h_reel * (image->haut / 2.0),
		     image->x_min + image->l_reel * image->larg, image->y_min + image->h_reel * (image->haut / 2.0) );
  IMG_bornes(image);
  free(b); free(c); free(d);
  return image;
}

/* ~~~~~~~~ Prparation de l'interface ~~~~~~~~ */

static void prepare_arbre_profil(void)
{
  OBJ_deplace(PROFIL_V_IMAGE, 17, 0, +1, DEPLACE_PIXELS);
  OBJ_agrandit(PROFIL_V_IMAGE, PVI_FOND_FORME, +3, +1);
  OBJ_deplace(PROFIL_V_IMAGE, PVI_ELLIPSE, +1, 0, DEPLACE_PIXELS);
  OBJ_deplace(PROFIL_V_IMAGE, PVI_AUTRE, +2, 0, DEPLACE_PIXELS);

  OBJ_deplace(PROFIL_V_IMAGE, PVI_TXT_RADIAL, 0, +500, DEPLACE_RELATIF);
  OBJ_deplace(PROFIL_V_IMAGE, PVI_TXT_TRANSV, 0, +500, DEPLACE_RELATIF);

  arbres[PROFIL_V_IMAGE].drapeaux.prepare = VRAI;
}

/* ~~~~~~~~ Gnration  partir d'quations z = f(x,y) ~~~~~~~~ */

Image *IMG_equation(double x_min, double pas_x, unsigned long nbr_x,
		    double y_min, double pas_y, unsigned long nbr_y,
		    Expression *formule)
{
  unsigned long i, j, n_case;
  double x, y;
  Image *image;

  if (formule == NULL) return NULL;

  image = IMG_CreerImage(NULL, NULL, IMG_GENRE_REELLE, IMG_SANS_TYPE, (unsigned int) nbr_x, (unsigned int) nbr_y);
  if (image == NULL) return NULL;

  image->l_reel = pas_x; /* Dimensions d'une case */
  image->h_reel = pas_y;

  for (i = 0 ; i < image->larg ; i++) {
    x = x_min + i * pas_x;
    Var_X.valeur_initiale = x;
    for (j = 0 ; j < image->haut ; j++) {
      double z = 0.0;
      y = y_min + j * pas_y;
      Var_Y.valeur_initiale = y;

      if (calcule_arbre(formule, &z) == VRAI) {
	IMG_IJ_v_N( image, i, j, &n_case );
	image->valeurs.R[ n_case ] = z;
      }
    }
  }

  image->x_min = x_min;
  image->y_min = y_min;
  IMG_bornes(image);

  return image;
}
