/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2001, Emmanuel CURIS
   Cration : 11 avril 2001

   Gestion des cartographies obtenues par fluorescence X (y compris micro-XANES)

   Lecture des valeurs de l'image : directe, par interpolation,...
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "images.h"
#include "lissage.h" /* Pour les interpolations */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define PRECISION_BORD_IMAGE (1.0e-7) /* Marge d'incertitude pour la sortie hors de l'image */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static boolean Noeuds_Autour(double x, double y,
			     int *x_BG, int *y_BG, int *x_HG, int *y_HG, int *x_BD, int *y_BD, int *x_HD, int *y_HD);

static float interpole_bilineaire(Image *image, double x, double y);
static float interpole_2t        (Image *image, double x, double y);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Renvoie les coordonnes des noeuds qui entourent la valeur choisie --------\
 |    Renvoie VRAI si tout s'est bien pass, FAUX en cas de problme                   |
 |                                                                                     |
 |  x, y : coordonnes fractionnaires dans un rseau, de maille rectangulaire unitaire |
 | x_, y_ BG/BD/HG/HD : coordonnes (entires) des noeuds du rseau les plus proche du |
 |   point (x,y) indiqu                                                               |
 \------------------------------------------------------------------------------------ */

static boolean Noeuds_Autour(double x, double y,
			     int *x_BG, int *y_BG, int *x_HG, int *y_HG, int *x_BD, int *y_BD, int *x_HD, int *y_HD)
{
  if ( (NULL == x_BG) || (NULL == y_BG) ) return FAUX;
  if ( (NULL == x_BD) || (NULL == y_BD) ) return FAUX;
  if ( (NULL == x_HD) || (NULL == y_HD) ) return FAUX;
  if ( (NULL == x_HG) || (NULL == y_HG) ) return FAUX;

  *x_BG = (int) floor(x); /* Donne toujours l'entier immdiatement infrieur  la valeur indique */
  *y_BG = (int) floor(y); /* Ce sont donc bien les coordonnes du coin infrieur droit */

  *x_BD = *x_BG + 1;
  *y_BD = *y_BG;

  *x_HG = *x_BG;
  *y_HG = *y_BG + 1;

  *x_HD = *x_BD;
  *y_HD = *y_HG;

  return VRAI;
}

/* -------- Interface gnrale pour obtenir la valeur en un point ---------\
 |  Renvoie la valeur estime, 0 en cas d'erreur                           |
 |          ou si le point n'est pas dans l'image                          |
 |                                                                         |
 | image  = l'image que l'on veut lire                                     |
 | x, y   = coordonnes du point dans le repre micromtrique               |
 |          relatif au centre indiqu pour l'image                         |
 | mode   = comment estimer la valeur (voir images.h pour les dfinitions) |
 | dehors = contient VRAI si le point indiqu tait en-dehors de l'image   |
 \------------------------------------------------------------------------ */

double IMG_lit_valeur(Image *image, double x, double y, int mode, boolean *dehors)
{
  double valeur = 0.0;

  if (image == NULL) return 0.0;
  if ( image->genre != IMG_GENRE_REELLE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Image pas de type rel\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0.0;
  }

  ENTREE_FONCTION("IMG_lit_valeur", "image=%p x=%g y=%g mode=%d dehors=%p", image, x, y, mode, dehors);

  if ( (image->l_reel <= 0.0) || (image->h_reel <= 0.0) ) {
    (void) fprintf(stderr, "IMG_lit_valeur : image avec une taille de case nulle !\n");
    if (dehors != NULL) *dehors = VRAI;
    SORTIE_FONCTION("IMG_lit_valeur", "%g [Cases nulles]", 0.0);
    return 0.0;
  }
  x -= image->x_min;  /* On se place dans le repre du coin BG de l'image */
  y -= image->y_min;
  x /= image->l_reel; /* On se place dans le repre dfini par la grille */
  y /= image->h_reel;

  if ( IMG_LIT_SIMPLE == mode ) {
    int x_case, y_case;
    unsigned long numero_case;

    /* Image compose de cases juxtaposes : on cherche les coordonnes de la case la plus proche */
    x_case = (int) rint( x );
    y_case = (int) rint( y );

    if ( ( x_case < 0 ) || ( x_case >= (int) image->larg ) || ( y_case < 0 ) || ( y_case >= (int) image->haut ) ) {
      if (dehors != NULL) *dehors = VRAI;
      SORTIE_FONCTION( __FUNCTION__, "%g [Dehors / cases juxtaposes]", 0.0 );
      return 0.0;
    }

    /* Valeur en cette case */
    IMG_IJ_v_N( image, x_case, y_case, &numero_case );
    valeur = image->valeurs.R[ numero_case ];
  } else {
    /* Interpolation : on est soit
      - dedans purement : interpolation par les quatre valeurs autour ;
      - dans la bande latrale d'une demi-case, soit sur un bord : interpolation par les deux valeurs autour ;
      - dans les coins : valeur gale  celle du coin ;
      - compltement dehors
    */

    if ( ( x < -0.5 ) || ( x > ( image->larg - 0.5 ) ) || ( y < -0.5 ) || ( y > ( image->haut - 0.5 ) ) ) {
      if (dehors != NULL) *dehors = VRAI;
      SORTIE_FONCTION("IMG_lit_valeur", "%g [Hors cadre]", 0.0);
#ifdef _DEVELOPPE_
/*       AVERTISSEMENT( "Hors image : (x,y)=(%g,%g), image de (0,0)  (%u,%u)", x, y, image->larg - 1, image->haut - 1 ); */
#endif
      return 0.0;
    }
    if (dehors != NULL) *dehors = FAUX; /* On est bien dans l'image [ou sa bande latrale] */

    if ( ( x <= 0.0 ) && ( y <= 0.0 ) ) { /* Coin BG */
      valeur = image->valeurs.R[ 0 ];
    } else if ( ( x <= 0.0 ) && ( y >= ( image->haut - 1.0 ) ) ) { /* Coin HG */
      unsigned long n_case;

      IMG_IJ_v_N( image, 0, image->haut - 1, &n_case );
      valeur = image->valeurs.R[ n_case ];
    } else if ( ( x >= ( image->larg - 1.0 ) ) && ( y <= 0.0 ) ) { /* Coin BD */
      unsigned long n_case;

      IMG_IJ_v_N( image, image->larg - 1, 0, &n_case );
      valeur = image->valeurs.R[ n_case ];
    } else if ( ( x >= ( image->larg - 1.0 ) ) && ( y >= ( image->haut - 1.0 ) ) ) { /* Coin HD */
      unsigned long n_case;

      IMG_IJ_v_N( image, image->larg - 1, image->haut - 1, &n_case );
      valeur = image->valeurs.R[ n_case ];
    } else {
      switch (mode)
      {
      case IMG_INTERPOLE_2T:
	valeur = interpole_2t(image, x, y);
	break;

      case IMG_INTERPOLE_BILINEAIRE:
	valeur = interpole_bilineaire(image, x, y);
	break;

      default:
	(void) fprintf(stderr, "%s,%s l.%u : Mode de lecture inconnu [%d]\n",
		       __FILE__, __FUNCTION__, __LINE__, mode);
	valeur = 0.0;
	break;
      }
    }
  }
  
  SORTIE_FONCTION("IMG_lit_valeur", "%g", valeur);
#ifdef _DEVELOPPE_
  if ( is_inf( valeur ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : VALEUR INFINIE [x=%g, y=%g \"%s\"]\n",
		   __FILE__, __FUNCTION__, __LINE__, x, y, image->nom);
    abort();
  }
#endif
  return valeur;
}

/* -------- Interpolation pour une surface, bilinaire.

   Principe : X---------X   X = points de la grille (index du tableau)
              |         |        ils sont de valeur connue, utilise
              |  ?      |   ? = point pour lequel on cherche
              |         |
	      X---------X (Adapt des Numerical Recipes in C) 

   (x, y) : coordonnes fractionnaires dans le rseau dfini par la grille de l'image
*/

static float interpole_bilineaire(Image *image, double x, double y)
{
  int x_HG, y_HG, x_BD, y_BD, x_BG, y_BG, x_HD, y_HD; /* Noeuds autour du point cherch... */
  double val_BG = 0.0, val_BD = 0.0, val_HD = 0.0, val_HG = 0.0; /* Valeurs en ces noeuds */
  double t, u;
  double resultat;
  unsigned long n_case;

  ENTREE_FONCTION("interpole_bilineaire", "image=%p x=%f y=%f", image, x, y);
  Noeuds_Autour(x, y, &x_BG, &y_BG, &x_HG, &y_HG, &x_BD, &y_BD, &x_HD, &y_HD);

  t = ((double) x - (double) x_BG); u = ((double) y - (double) y_BG);

  /* Contrles de dbordement */
#ifdef _DEVELOPPE_
  if ( ( x_BD < 0 ) || ( x_HD < 0 ) || ( y_HD < 0 ) || ( y_HG < 0 ) ||
       ( x_BG >= (int) image->larg) || ( x_HG >= (signed) image->larg ) || ( y_BD >= (signed) image->haut ) || ( y_BG >= (signed) image->haut ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Point compltement hors de l'image [%s : (%g, %g)] : anormal\n",
		   __FILE__, __FUNCTION__, __LINE__, image->nom, x, y);
    abort();
  }
#endif
  /* Si on est dans la marge, on utilise les mmes valeurs pour le ct inconnu et le ct connu
     => alors, par l'interpolation linaire, donne le gradient souhait ! */
  if ( x_BG < 0 ) x_BG = 0; if ( y_BG < 0 ) y_BG = 0;
  if ( x_BD >= (int) image->larg ) x_BD = image->larg - 1; if ( y_BD < 0 ) y_BD = 0;
  if ( x_HD >= (int) image->larg ) x_HD = image->larg - 1; if ( y_HD >= (int) image->haut ) y_HD = image->haut - 1;
  if ( x_HG < 0 ) x_HG = 0; if ( y_HG >= (int) image->haut ) y_HG = image->haut - 1;

#ifdef _DEVELOPPE_
/*   printf("%s : BG (%d, %d) %g, HG (%d, %d) %g, BD (%d, %d) %g, HD : (%d, %d) %g\n", __FUNCTION__, */
/* 	 x_BG, y_BG, val_BG, x_HG, y_HG, val_HG, x_BD, y_BD, val_BD, x_HD, y_HD, val_HD); */
#endif

  IMG_IJ_v_N( image, x_BG, y_BG, &n_case );
  val_BG = (double) image->valeurs.R[ n_case ];

  IMG_IJ_v_N( image, x_BD, y_BD, &n_case );
  val_BD = (double) image->valeurs.R[ n_case ];

  IMG_IJ_v_N( image, x_HD, y_HD, &n_case );
  val_HD = (double) image->valeurs.R[ n_case ];

  IMG_IJ_v_N( image, x_HG, y_HG, &n_case );
  val_HG = (double) image->valeurs.R[ n_case ];
  
  DANS_FONCTION("BG (%d, %d) %g, HG (%d, %d) %g, BD (%d, %d) %g, HD : (%d, %d) %g",
		x_BG, y_BG, val_BG, x_HG, y_HG, val_HG, x_BD, y_BD, val_BD, x_HD, y_HD, val_HD);
  resultat = t * u * val_HD + t * (1.0 - u) * val_BD + (1.0 - t) * u * val_HG + (1.0 - t) * (1.0 - u) * val_BG;

#ifdef _DEVELOPPE_
  if ( is_inf( resultat ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : VALEUR INFINIE [x=%g, y=%g \"%s\" ==>\n %g \t %g\n %g \t %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, x, y, image->nom, val_HG, val_HD, val_BG, val_BD);
    abort();
  }
#endif

  SORTIE_FONCTION("interpole_bilineaire", "%g", resultat);
  return (float) resultat;
}

/* -------- Interpolation pour une surface, pseudolinaire, par triangulation simple.

   Principe : X---------X   X = points de la grille (index du tableau)
              | ?       |        (points de valeur connue)
              |         |   ? = point pour lequel on cherche
              |         |
	      X---------X

    Les trois X les plus proches du ? dfinissent un plan, ? est suppos appartenir  ce plan.
*/

static float interpole_2t(Image *image, double x, double y)
{
  int x_HG, y_HG, x_BD, y_BD, x_BG, y_BG, x_HD, y_HD; /* Noeuds autour du point cherch... */
  double val_BG, val_BD, val_HD, val_HG; /* Valeurs en ces noeuds */
  double resultat;
  unsigned long n_case;

  if (image == NULL) return 0.0;

  Noeuds_Autour(x, y, &x_BG, &y_BG, &x_HG, &y_HG, &x_BD, &y_BD, &x_HD, &y_HD);

  IMG_IJ_v_N( image, x_BG, y_BG, &n_case );
  val_BG = (double) image->valeurs.R[ n_case ];

  IMG_IJ_v_N( image, x_BD, y_BD, &n_case );
  val_BD = (double) image->valeurs.R[ n_case ];

  IMG_IJ_v_N( image, x_HG, y_HG, &n_case );
  val_HG = (double) image->valeurs.R[ n_case ];

  IMG_IJ_v_N( image, x_HD, y_HD, &n_case );
  val_HD = (double) image->valeurs.R[ n_case ];

  x -= (float) x_BG; y -= (float) y_BG; /* Coordonnes par rapport au noeud inf. gauche */

  if ( (x <= 0.5) && (y <= 0.5) ) { /* Coin bas-gauche de la case */
    resultat = LIS_planaire(0.0, 0.0, val_BG, 1.0, 0.0, val_BD, 0.0, 1.0, val_HG, x, y);
  } else if ( (x <= 0.5) && (y > 0.5) ) { /* Coin haut-gauche de la case */
    resultat = LIS_planaire(0.0, 0.0, val_BG, 0.0, 1.0, val_HG, 1.0, 1.0, val_HD, x, y);
  } else if ( (x > 0.5)  && (y <= 0.5) ) { /* Coin bas-droit de la case */
    resultat = LIS_planaire(0.0, 0.0, val_BG, 1.0, 0.0, val_BD, 1.0, 1.0, val_HD, x, y);
  } else { /* Coin haut-droit de la case */
    resultat = LIS_planaire(0.0, 1.0, val_HG, 1.0, 1.0, val_HD, 1.0, 0.0, val_BD, x, y);
  }

  return (float) resultat;
}
