/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) Emmanuel CURIS, 2003

   Modlisation d'un faisceau lumineux
*/

#ifndef FAISCEAU_H_
# define FAISCEAU_H_ /* Protection contre l'inclusion multiple */

# include "utils_3d.h" /* Description d'objets dans l'espace */

struct detecteur;
struct echantillon;
struct graphe;
struct image;

/* -------- Proprits du faisceau -------- */

#define FAISCEAU_FOCALISE        (1U << 0) /* bit mis : le faisceau diverge depuis un point   */
#define FAISCEAU_POLYCHROMATIQUE (1U << 1) /* bit mis : le faisceau n'est pas monochromatique */
#define FAISCEAU_INHOMOGENE      (1U << 2) /* bit mis : le faisceau a une structure spatiale  */
#define FAISCEAU_POLARISE        (1U << 3) /* bit mis : le faisceau est polaris */
#define FAISCEAU_HARMONIQUES     (1U << 4) /* bit mis : on tient compte des harmoniques */
#define FAISCEAU_DECROIT         (1U << 5) /* bit mis : l'intensit varie avec l'nergie */
#define FAISCEAU_PONCTUEL        (1U << 6) /* bit mis : le faisceau est ponctuel */

/* La structure du faisceau, suppos de section rectangulaire avec des cts compatibles avec les axes.
   En fait, c'est une image (mais qu'il faudrait ne pas faire apparatre dans la liste des images).
   On sait qu'il s'agit d'un faisceau homogne si z_min = z_max */ 

/* -------- Dfinition d'un faisceau -------- */

typedef struct faisceau {
  struct faisceau *suivant;
  char *nom;

  double largeur, hauteur; /* en microns. Suppose un faisceau rectangulaire.
			      Si faisceau focalis, correspond aux dimensions sur la face de sortie
			      du dispositif de focalisation */
  double intensite;        /* Intensit nominale.
			      Les termes correctifs correspondent  des pourcentages de cette intensit */
  unsigned int proprietes; /* Masque de bits : proprits du faisceau */
  Orientation3D orientation; /* Orientation du faisceau :
			       - la normale correspond  la direction de propagation [axe des Y donc]
			       - la position correspond au centre du faisceau */
  Vecteur3D point_focal;   /* Point vers lequel converge le faisceau */
  struct graphe *g_energie;/* Variation de l'intensit avec l'nergie nominale */
  struct graphe *g_E;      /* Allure de la courbe g(E) [pour le fondamental]
			      Donne g(E) = f(DE/E) en relatif par rapport  l'nergie relle du faisceau
			   */
  unsigned int nbr_harmoniques;  /* Combien d'harmoniques faut-il prendre en compte ? */
  float *i_harmoniques;          /* Intensits relatives des harmoniques, par rapport au fondamental */
  struct graphe **g_harmoniques; /* Les g(E) pour ces harmoniques ; mmes conventions que pour le fondamental */
  struct image *g_espace;        /* Inhomognit spatiale : image qui donne les intensits relatives
				    Si NULL : faisceau homogne ou ponctuel */
} Faisceau;

#ifndef VARIABLES_GENERALES_
extern struct faisceau *premier_faisceau;
#endif

/* Manipulation & utilisation des faisceau */

/* -------- Cration, effacement, copie,... -------- */

Faisceau *FAISCEAU_Creer  (double l, double h, unsigned int proprietes, unsigned int harmoniques, double I);
Faisceau *FAISCEAU_Copier (Faisceau *faisceau);
int       FAISCEAU_Effacer(Faisceau *faisceau);

int FAISCEAU_ChainerFaisceau(Faisceau *avant, Faisceau *faisceau);

int       FAISCEAU_CreerHarmoniques(Faisceau *faisceau, unsigned int nbr_harmoniques);
int       FAISCEAU_Nommer(Faisceau *faisceau, const char *nom);
int       FAISCEAU_Parametrer(Faisceau *faisceau);
int       FAISCEAU_AssocierCoupe  (Faisceau *faisceau, struct image *coupe, boolean copie, unsigned int nbr_x, unsigned int nbr_y);
int       FAISCEAU_AssocierSpectre(Faisceau *faisceau, struct graphe *g_E, boolean copie);
struct image *FAISCEAU_ReserverCoupe(double larg, double haut, double pas_x, double pas_z );

int       FAISCEAU_ControlerOptions(Faisceau *faisceau, double energie);

/* Version pour le slecteur universel */
int FAISCEAU_Charger_I (unsigned long n_faisceau);
int FAISCEAU_Copier_I  (unsigned long original);
int FAISCEAU_Effacer_I (unsigned long original);
int FAISCEAU_Modifier_I(unsigned long original);
int FAISCEAU_Nouveau_I (unsigned long avant);
int FAISCEAU_Sauver_I  (unsigned long n_faisceau);

unsigned long FAISCEAU_Compter   (void);
unsigned long FAISCEAU_F_Vers_N  (Faisceau *premier, Faisceau *faisceau);
Faisceau     *FAISCEAU_N_Vers_F  (Faisceau *premier, unsigned long n);
const char   *FAISCEAU_TrouverNom(unsigned int arbre, unsigned long objet);

Faisceau *FAISCEAU_Charger(FILE *fichier, int version, const char *nom_fichier);
int       FAISCEAU_Sauver (FILE *fichier, Faisceau *faisceau);

/* int       FAISCEAU_GenererSpectre(Faisceau *faisceau, struct spectre_local spectre); */

Faisceau *FAISCEAU_GeometrieApresEchantillon(Faisceau *faisceau, struct echantillon *echantillon);
Faisceau *FAISCEAU_GeometrieFluorescence    (struct detecteur *detecteur, double *table_E, unsigned long nbr_E);
Faisceau *FAISCEAU_ProjeterSurEchantillon   (Faisceau *faisceau, struct echantillon *echantillon);

int FAISCEAU_ProjeterCentre (Faisceau *faisceau, double profondeur, double *x, double *y);
int FAISCEAU_ProjeterLimites(Faisceau *faisceau, double profondeur, double *x_coins, double *y_coins); /* Rq. Taille 5 au moins */

double FAISCEAU_LireIntensite(Faisceau *faisceau, double x, double y, double energie);

int FAISCEAU_TraverserEchantillon(Faisceau *faisceau, struct echantillon *echantillon, double energie, Faisceau *f_transmis);

struct image *CNV_FaisceauVersImage(Faisceau *faisceau);

#endif /* FAISCEAU_H_ */
