/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2006, Emmanuel CURIS
   Création : mars 2006

   Interface graphique avec l'utilisateur

   Ajuster un modèle de type EXAFS sur un spectre EXAFS expérimental
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "atomes.h"
#include "graphes.h"
#include "interface.h"
#include "modeles.h"

#include "int_ajuster_exafs.txt"

/* -------- Variables locales -------- */

/* -------- Fonctions locales -------- */

static struct modele_ajuste *I_EXAFS_Ajuste_ChercherModele(struct graphe *reference);

/* -~-~-~-~-~-~-~-~- Appel pour réaliser un ajustement -~-~-~-~-~-~-~-~- */

/* -------- Ajuster un modèle pour un spectre EXAFS (en k ou en R : TF) -------- */

int I_EXAFS_AjusterModele(struct graphe *reference)
{
  Fenetre *fenetre;
  ModeleAjustement *modele __attribute__((__unused__)) = NULL, *modele_precedent = NULL;
  int reponse __attribute__((__unused__)) = -1;

  if ( NULL == reference ) {
    return -1;
  }

  /* Le spectre doit être du type convenable !
     Actuellement, modèle FEFF ==> exafs (ou TF) */
  if ( ( reference->type != GR_TF ) && ( FAUX == reference->drapeaux.exafs ) &&
       ( reference->type != GR_SSTYPE ) ) { /* Graphes sans type peuvent servir à tout... */
    (void) BI_ERR_alerte( principale, &ERR_TypeIncorrect, VRAI );
    (void) fprintf( stderr, "%s,%s l.%u : graphe de type incorrect [%d, EXAFS = %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, reference->type, reference->drapeaux.exafs );
    return -2;
  }

  /* De toute façon, on enregistre le modèle actif ! */
  LMOD_EnregistrerModele( NULL );

  /* On recherche le modèle « naturel » pour ce spectre */
  modele = I_EXAFS_Ajuste_ChercherModele( reference );

  /* On cherche le dernier modèle qui a été ajusté : stocké dans la fenêtre */
  fenetre = ARB_QuelleFenetre( arbres + RESULT_N_COUCHES, FAUX );
  if ( fenetre != NULL ) {
    modele_precedent = (struct modele_ajuste *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
    if ( modele_precedent != NULL ) {
      /* Il y avait bien un modèle en cours d'ajustement */
      if ( VRAI == modele_precedent->drapeaux.isole ) {
	/* Ce modèle était un modèle temporaire */
	if ( modele_precedent->experience == reference ) {
	  /* Déjà pour ce spectre : on continue donc... */
	  reponse = I_EXAFS_AJUSTE_TEMPORAIRE;
	}
	modele_precedent = NULL;
      } else {
	int rep_temp __attribute__((__unused__));

	/* On ajustait sur un modèle déjà existant */
	rep_temp = BI_ERR_alerte( principale, &Q_QuelModeleChoisir, VRAI );
      }
    }
  }

  /* Terminé, sans encombre */
  return 0;
}

/* -------- Chercher le modèle adapté pour le spectre -------- */

static struct modele_ajuste *I_EXAFS_Ajuste_ChercherModele(struct graphe *reference)
{
  Absorbeur *absorbeur;
  Molecule *molecule;
  struct modele_ajuste *modele = NULL/* , *modele_actif */;
  boolean modele_actif_OK __attribute__((__unused__)) = FAUX;
/*   unsigned int nombre_modeles_absorbeur = 0, nombre_modeles_OK = 0; */

  /* On vérifie qu'il peut exister un modèle : il y a forcément une molécule */
  if ( NULL == mol_courante ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "%s,%s l.%u : Aucune molécule ==> aucun modèle FEFF disponible\n",
		   __FILE__, __FUNCTION__, __LINE__);
#endif
    (void) BI_ERR_alerte( principale, &ERR_AucuneMolecule, VRAI );
    return NULL;
  }

  /* ---- Passe 1 : on cherche des modèles qui auraient déjà ce spectre en référence ---- */

  /* 1. On cherche dans le modèle actif, s'il y en a un */
  modele = (struct modele_ajuste *) ARB_LireComplement( PARAMS_MODELE );
  if ( ( modele != NULL ) && ( MODELE_EXAFS_CHEMINS == modele->type ) ) {
    if ( reference == modele->experience ) {
      /* Modèle de type convenable : on l'utilise donc */
      return modele;
    }
    if ( NULL == modele->experience ) {
      modele_actif_OK = VRAI;
    }
  }

  /* 2. On cherche dans les modèles de la molécule courante */
  if ( mol_courante->absorbeurs != NULL ) {

    /* a. Modèle actif de l'absorbeur courant */
    absorbeur = mol_courante->absorbeurs + mol_courante->absorbeur_courant;
    if ( absorbeur->modele_actif != NULL ) {
      if ( MODELE_EXAFS_CHEMINS == absorbeur->modele_actif->type ) {
	/* Modèle de type convenable : on l'utilise donc */
	if ( reference == absorbeur->modele_actif->experience ) {
	  return absorbeur->modele_actif;
	}
      }
    }

    /* b. Autres modèles de l'absorbeur courant */
    for ( modele = absorbeur->modeles ; modele != NULL ; modele = modele->suivant ) {
      if ( MODELE_EXAFS_CHEMINS == modele->type ) {
	/* Modèle de type convenable : on l'utilise donc */
	if ( reference == modele->experience ) {
	  return modele;
	}
      }
    }

    /* c. Autres absorbeurs */
    OPTION_A_FAIRE( "Chercher dans les autres absorbeurs de la molécule" );

    /* d. Modèles au niveau de la molécule */
    OPTION_A_FAIRE( "Chercher dans les autres modèles de la molécule" );
  }

  /* 3. On cherche dans toutes les molécules*/
  for ( molecule = premiere_molecule ; molecule != NULL ; molecule = molecule->suivante ) {
  }

  /* On renvoie le modèle trouvé */
  return modele;
}
