/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 2005, Emmanuel CURIS
   Cration : 15 juin 2005

   Interface graphique

   Crer "ex-nihilo" un graphe ou une image
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "ascii_utils.h"
#include "bin_utils.h"		/* Fichiers d'un dossier */

#include "systeme.h"
#include "exafs.h"

#include "c_formel.h"
#include "gen_graphe.h"
#include "interface.h"
#include "modeles.h"
#include "operations.h"
#include "serie.h"

/* -------- Variables locales -------- */

#define NOMBRE_VARIABLES_UTILISABLES 6 /* Combien de variables sont potentiellement utilisables, au maximum */

struct infos_graphe {
  char *nom;
  int type;
  struct info_variable X[ NOMBRE_VARIABLES_UTILISABLES ]; /* Les variables utilisables */
};

#define VARIABLE_AFFICHEE_X 0	/* Quelle variable est affiche dans la bote de dialogue ? */
#define VARIABLE_AFFICHEE_Y 1
#define VARIABLE_AFFICHEE_Z 2
#define VARIABLE_AFFICHEE_T 3
#define VARIABLE_AFFICHEE_U 4
#define VARIABLE_AFFICHEE_I 5

struct infos_arbre {
  unsigned long nombre_graphes;	/* Que l'on va crer */
  unsigned long numero_graphe_affiche;
  int genre, affichage;		/* Genre des donnes et mode d'affichage */
  struct infos_graphe *graphes;	/* Pour chacun des graphes crs */
  unsigned int variable_affichee; /* Quelle est la variable affiche ? */
  unsigned long nombre_parametres; /* Intervenant dans les quations */
  unsigned long numero_parametre_affiche;
  struct parametre *parametres;	/* Table des paramtres du modle */
  unsigned long nombre_domaines;
  unsigned long domaine_choisi;
  const char **domaines;
  unsigned long nombre_modeles;	/* Le nombre de modles dans le domaine choisi */
  char **modeles;		/* Les noms des modles */
  char **fichiers_modeles; /* Les noms des fichiers associs  ces modles */
};

#define LGMAX_GENRE_GRAPHE     25
#define LGMAX_AFFICHAGE        10
#define LGMAX_TYPE_GRAPHE      30
#define LGMAX_TYPE_VARIABLE    13
#define LGMAX_ORIGINE_VARIABLE 20
#define LGMAX_GRAPHE_ORIGINE   50
#define LGMAX_DOMAINE          20
#define LGMAX_MODELE           20

/* -------- Les textes [isols pour traduction] -------- */

#include "int_creation.txt"

/* -------- Fonctions locales -------- */

Graphe *I_CG_FonctionReelle(struct infos_arbre *infos, Graphe *avant, Serie_Graphes *serie);
Graphe *I_CG_SerieValeurs  (struct infos_arbre *infos, Graphe *avant, Serie_Graphes *serie);

static int I_CG_GererArbre(unsigned int objet, int x, int y);

static void I_CG_LibererGraphe      (struct infos_arbre *infos, unsigned long numero_graphe);
static void I_CG_LibererInfos       (struct infos_arbre *infos);
static void I_CG_LibererModeles     (struct infos_arbre *infos);
static void I_CG_PreparerArbre      (void);
static int  I_CG_ReserverInfosGraphe(struct infos_arbre *infos, unsigned long nombre_graphes);

static void I_CG_EnregistrerGraphe  (struct infos_arbre *infos);
static void I_CG_EnregistrerParametre(struct infos_arbre *infos);
static void I_CG_EnregistrerVariable(struct infos_arbre *infos);

static void I_CG_ActiverVariable   (struct infos_arbre *infos);
static void I_CG_AjouterParametre  (struct infos_arbre *infos, unsigned long nombre);
static void I_CG_EffacerParametre  (struct infos_arbre *infos, unsigned long numero);
static void I_CG_InstallerGraphe   (struct infos_arbre *infos);
static void I_CG_InstallerModele   (struct infos_arbre *infos, struct modele_generique *modele);
static void I_CG_InstallerParametre(struct infos_arbre *infos);
static void I_CG_InstallerVariable (struct infos_arbre *infos);

static const char *I_CG_TexteMethode (unsigned int type_variable);
static const char *I_CG_TexteVariable(unsigned int num_variable);

/* ~-~-~-~-~-~-~-~-~ Appel par l'utilisateur ~-~-~-~-~-~-~-~-~ */

/* -------- Cration d'un graphe -------- */

Graphe *I_GRF_Generer(Graphe *avant)
{
  int retour;
  struct infos_arbre infos = { 0, 0, GRF_FCT_REELLE, GRF_R_CARTESIEN, NULL, VARIABLE_AFFICHEE_X, 0, 0, NULL, 0, 0, NULL, 0, NULL, NULL };
  Graphe *resultat = NULL;
  Serie_Graphes *serie = NULL;

  /* On prpare l'arbre */
  I_CG_PreparerArbre();

  /* On prpare les informations [pour un seul graphe cr, sans paramtre] */
  retour = I_CG_ReserverInfosGraphe( &infos, 1 );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de rserver les infos\n",
		   __FILE__, __FUNCTION__, __LINE__);
    BI_ERR_annuler( ERR_reservation_infos );
    return NULL;
  }

  /* On installe le type de graphes sur lequel on travaille [on peut remplacer car ces fonctions renvoient des textes constants ! */
  OBJ_RemplacerTexte( CREER_GRAPHE, CG_GENRE_GRAPHE, GRF_TXT_Genre( infos.genre ), VRAI );
  OBJ_RemplacerTexte( CREER_GRAPHE, CG_TYPE_AFFICHAG, GRF_TXT_TypeAffichage( infos.genre, infos.affichage ), VRAI );
  I_CG_ActiverVariable( &infos );

  /* On installe le premier graphe */
  I_CG_InstallerGraphe( &infos );

  /* On installe les paramtres (aucun, a priori) */
  I_CG_InstallerParametre( &infos );

  /* On cherche les modles prdfinis [Remplace car un ventuel prcdent aura t libr !] */
  OBJ_RemplacerTexte( CREER_GRAPHE, CG_DOMAINE, txt_choisir_domaine, VRAI );
  OBJ_active( CREER_GRAPHE, CG_MODELE, FORCE_NON );
  if ( MODELE_ListeDomaines( &(infos.nombre_domaines), &(infos.domaines) ) != 0 ) {
    infos.nombre_domaines = 0;
    infos.domaine_choisi = 0;
    infos.domaines = NULL;
    OBJ_active( CREER_GRAPHE, CG_DOMAINE, FORCE_NON );
  } else {
    OBJ_active( CREER_GRAPHE, CG_DOMAINE, FORCE_OUI );
  }

  /* On appelle l'arbre */
  ARB_MettreComplement( CREER_GRAPHE, &infos );
  retour = ARB_Bloquant( CREER_GRAPHE, titre_creer_graphe, principale, 0 );
  ARB_MettreComplement( CREER_GRAPHE, NULL );
  if ( CG_ANNULER == retour ) {	/* Opration annule */
    I_CG_LibererInfos( &infos );
    return NULL;
  }

  /* On mmorise les paramtres de cration qui auraient pu ne pas l'tre */
  I_CG_EnregistrerGraphe( &infos );
  I_CG_EnregistrerParametre( &infos );

  /* On choisit la srie dans laquelle placer le graphe */
  if ( FAUX == OBJ_choisit( CREER_GRAPHE, CG_NLLE_SERIE, QUESTION ) ) {	/* Ce n'est pas une nouvelle que l'on demande */
    if ( avant != NULL ) {
      /* Ce sera donc la srie du graphe indiqu */
      serie = GRF_QuelleSerie( avant );
    } else {
      serie = serie_courante;
      avant = graphe_courant;
    }
  }

  /* Si pas de srie, on la cre */
  if ( NULL == serie ) {
    serie = SERIE_GRF_CreerSerie( serie_courante, txt_nom_serie );
    if ( NULL == serie ) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de crer une nouvelle srie. Opration annule\n",
		     __FILE__, __FUNCTION__, __LINE__);
      BI_ERR_annuler( ERR_creation_serie );
    }
    avant = NULL;
  }

  /* En fonction du genre de donnes, on appelle les pr-routines de cration */
  switch ( infos.genre )
  {
  case GRF_FCT_REELLE:
    resultat = I_CG_FonctionReelle( &infos, avant, serie );
    break;

  case GRF_FCT_COMPLEXE:
    OPTION_A_FAIRE( "Cration de graphe(s) complexe(s)" );
    break;

  case GRF_SERIE:
    resultat = I_CG_SerieValeurs( &infos, avant, serie );
    break;
  }

  /* On libre les informations si ncessaire */
  I_CG_LibererInfos( &infos );

  /* Termin : on renvoie le dernier graphe cr */
  return resultat;
}

/* -------- Cration des graphes de type "fonction relle", Y=f(X) -------- */

Graphe *I_CG_FonctionReelle(struct infos_arbre *infos, Graphe *avant, Serie_Graphes *serie)
{
  Graphe *resultat = NULL;
  unsigned long i;

  /* Graphe par graphe, on fait les crations */
  for ( i = 0 ; i < infos->nombre_graphes ; i++ ) {
    resultat = GRF_GenererFonctionReelle( infos->graphes[ i ].nom, infos->graphes[ i ].type,
					  infos->graphes[ i ].X, infos->graphes[ i ].X + 1, infos->graphes[ i ].X + VARIABLE_AFFICHEE_I,
					  infos->nombre_parametres, infos->parametres, avant, serie);
  }

  /* On renvoie le dernier graphe cr */
  return resultat;
}

/* -------- Cration des graphes de type "srie de valeurs", Y=f(I) -------- */

Graphe *I_CG_SerieValeurs(struct infos_arbre *infos, Graphe *avant, Serie_Graphes *serie)
{
  Graphe *resultat = NULL;
  unsigned long i;

  /* Graphe par graphe, on fait les crations */
  for ( i = 0 ; i < infos->nombre_graphes ; i++ ) {
    resultat = GRF_GenererSerieValeurs( infos->graphes[ i ].nom, infos->graphes[ i ].type,
					infos->graphes[ i ].X + 1, infos->graphes[ i ].X + VARIABLE_AFFICHEE_I,
					infos->nombre_parametres, infos->parametres, avant, serie);
  }

  /* On renvoie le dernier graphe cr */
  return resultat;
}

/* -------- Prparation de l'arbre -------- */

static void I_CG_PreparerArbre(void)
{
  int retour;

  if ( VRAI == arbres[ CREER_GRAPHE ].drapeaux.prepare ) return; /* Dj initialis */

  /* On initialise automatiquement */
  retour = ARB_Initialiser( CREER_GRAPHE, NULL, I_CG_GererArbre, ( int (*)(unsigned int, int, int *) ) NULL );
  if ( retour != 0 ) {
    AVERTISSEMENT( "chec de l'initialisation de l'arbre [code %d]", retour );
    return;
  }

  /* Amlioration des positions */
  OBJ_deplace( CREER_GRAPHE, CG_MULTI_GRAPHES, 0, -500, DEPLACE_RELATIF );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_MULTIPLE , 0, -500, DEPLACE_RELATIF );
  OBJ_deplace( CREER_GRAPHE, CG_NLLE_SERIE   , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_NLLE_SERI, 0, -2, DEPLACE_PIXELS );

  ARB_EnjoliverAscenseur( CREER_GRAPHE, CG_GRF_AVANT, CG_FOND_ASC, CG_BOUTON_GRF, CG_GRF_APRES, ASC_HORIZONTAL, 0 );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_NOM    , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_NOM_GRAPHE , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_TYPE   , 0, +4, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TYPE_GRAPHE, 0, +4, DEPLACE_PIXELS );

  OBJ_deplace( CREER_GRAPHE, CG_VARIABLE_X, 0, +500, DEPLACE_RELATIF );
  OBJ_deplace( CREER_GRAPHE, CG_VARIABLE_Y, 0, +500, DEPLACE_RELATIF );
  OBJ_deplace( CREER_GRAPHE, CG_VARIABLE_Z, 0, +500, DEPLACE_RELATIF );
  OBJ_deplace( CREER_GRAPHE, CG_VARIABLE_T, 0, +500, DEPLACE_RELATIF );
  OBJ_deplace( CREER_GRAPHE, CG_VARIABLE_U, 0, +500, DEPLACE_RELATIF );
  OBJ_deplace( CREER_GRAPHE, CG_VARIABLE_I, 0, +500, DEPLACE_RELATIF );

  OBJ_deplace( CREER_GRAPHE, CG_TXT_LEGENDE, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_LEGENDE    , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_METHODE, 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_METHODE_VAR, 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_NOMBRE , 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_VALEUR_N   , 0, +3, DEPLACE_PIXELS );

  OBJ_deplace( CREER_GRAPHE, CG_TXT_MIN    , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_VAL_MIN    , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_MAX    , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_VALEUR_MAX , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_PAS    , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_VAL_PAS    , 0, -2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_GRAPHE2, 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_GRF_DES_X  , 0, +3, DEPLACE_PIXELS );

  OBJ_deplace( CREER_GRAPHE, CG_TXT_PARAMETRE, 0, +500, DEPLACE_RELATIF );

  OBJ_deplace( CREER_GRAPHE, CG_NV_PARAM     , 0, -1, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_EFFACER_PARAM, 0, -1, DEPLACE_PIXELS );

  ARB_EnjoliverAscenseur( CREER_GRAPHE, CG_PARAM_AVANT, CG_PARAM_FONDASC, CG_PARAM_BOUTON, CG_PARAM_APRES, ASC_VERTICAL, 0 );

  OBJ_deplace( CREER_GRAPHE, CG_TXT_SYMBOLE  , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_SYMBOLE_PARAM, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_VAL_PARAM, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_VALEUR_PARAM , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_TXT_UNIT_PARA, 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( CREER_GRAPHE, CG_UNITE_PARAM  , 0, +3, DEPLACE_PIXELS );

  OBJ_deplace( CREER_GRAPHE, CG_TXT_BIB, 0, +500, DEPLACE_RELATIF );

  /* Traduction ventuelle des objets */
#ifdef AVEC_TRADUCTION_
#endif

  /* Termin */
  arbres[ CREER_GRAPHE ].drapeaux.prepare = VRAI;
}

/* -------- Macros d'affichage -------- */

static void I_CG_ActiverVariable(struct infos_arbre *infos)
{
  /* On dsactive toutes les variables [sauf I] */
  OBJ_active( CREER_GRAPHE, CG_VARIABLE_X, FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_VARIABLE_Y, FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_VARIABLE_Z, FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_VARIABLE_T, FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_VARIABLE_U, FORCE_NON );

  /* On active les variables utilisables */
  switch ( infos->genre )
  {
  case GRF_FCT_REELLE:
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_X, FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_Y, FORCE_OUI );
    break;

  case GRF_FCT_COMPLEXE:
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_X, FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_Y, FORCE_OUI ); /* Partie relle */
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_Z, FORCE_OUI ); /* Partie imaginaire */
    break;

  case GRF_SERIE:
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_Y, FORCE_OUI );
    break;

  case GRF_COURBE_PLANE:	/* donc {x=f(t), y=g(t)} */
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_X, FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_Y, FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VARIABLE_T, FORCE_OUI );
    break;
  }
}

static void I_CG_InstallerGraphe(struct infos_arbre *infos)
{
  struct infos_graphe *graphe_actif;
  if ( ( NULL == infos ) || ( NULL == infos->graphes ) || ( infos->numero_graphe_affiche >= infos->nombre_graphes ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Infos incompltes, absentes ou absurdes [i=%p, i->g=%p, i->nga=%lu >?= i->ng=%lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, (void *) infos, (void *) (infos->graphes), infos->numero_graphe_affiche, infos->nombre_graphes );
    return;
  }

  /* On place l'ascenseur correctement */
  (void) ARB_ascenseur( CREER_GRAPHE, CG_GRF_AVANT, CG_FOND_ASC, CG_BOUTON_GRF, CG_GRF_APRES, ASC_HORIZONTAL,
			infos->nombre_graphes, 1, infos->numero_graphe_affiche, ROOT, 0, 0, METTRE );

  /* On affiche le bon numro de graphe */
  OBJ_MettreEntier_U( CREER_GRAPHE, CG_NUM_GRAPHE, infos->numero_graphe_affiche + 1 );

  /* Commodit : le graphe actif */
  graphe_actif = infos->graphes + infos->numero_graphe_affiche;

  /* On affiche le nom propos pour le graphe */
  OBJ_CopierTexte( CREER_GRAPHE, CG_NOM_GRAPHE, graphe_actif->nom, 0 );

  /* On affiche le type propos pour le graphe */
  (void) OBJ_ChoisirTypeGraphe( CREER_GRAPHE, CG_TYPE_GRAPHE, LGMAX_TYPE_GRAPHE, graphe_actif->type, METTRE, VRAI );

  /* On affiche la variable active pour ce graphe */
  I_CG_InstallerVariable( infos );
}

static void I_CG_InstallerModele(struct infos_arbre *infos, struct modele_generique *modele)
{
  /* Le genre de graphe cr */
  infos->genre = modele->type;

  /* a. On efface les graphes superflus */
  while ( infos->nombre_graphes > 0 ) {
    I_CG_LibererGraphe( infos, 0 );
    infos->nombre_graphes--;	/* Pas modifi dans I_CG_LibererGraphe ! */
  }

  /* b. On installe le nouveau graphe */
  /* Actuellement, seul un seul graphe sera cr = celui qui sera donc affich */
  I_CG_ReserverInfosGraphe( infos, 1 );

  /* Pour ce premier graphe... */
  infos->graphes[ 0 ].nom = modele->nom;
  modele->nom = NULL;

  infos->graphes[ 0 ].type = GR_AUTRETYPE;

  switch ( infos->genre )
  {
  case GRF_FCT_REELLE:
    /* X : dfini par l'utilisateur, on n'y touche pas ! */

    /* Y : dfini par la formule indique */
    infos->graphes[ 0 ].X[ VARIABLE_AFFICHEE_Y ].type = VARIABLE_FORMULE;

    infos->graphes[ 0 ].X[ VARIABLE_AFFICHEE_Y ].unite = modele->Y.unite;
    modele->Y.unite = NULL;

    infos->graphes[ 0 ].X[ VARIABLE_AFFICHEE_Y ].nom = modele->Y.type;
    modele->Y.type = NULL;

    infos->graphes[ 0 ].X[ VARIABLE_AFFICHEE_Y ].valeurs = modele->Y.formule;
    modele->Y.formule = NULL;
    break;

  default:
    OPTION_A_FAIRE( "Grer ce genre de graphes dans les modles !" );
    break;
  }

  /* Pour les paramtres */
  /* a. On efface les anciens paramtres */
  while ( infos->nombre_parametres > 0 ) {
    I_CG_EffacerParametre( infos, 0 );
  }

  /* b. On installe les nouveaux ventuels */
  infos->parametres = modele->parametres;
  infos->nombre_parametres = modele->nombre_parametres;

  /* c. On les enlve du modle charg pour ne pas les effacer trop tt ! */
  modele->nombre_parametres = 0;
  modele->parametres = NULL;

  /* On affiche tout cela */
  if ( infos->nombre_parametres > 0 ) {
    /* Il y a des paramtres : activer les boutons */
    OBJ_active( CREER_GRAPHE, CG_EFFACER_PARAM, FORCE_OUI );

    OBJ_active( CREER_GRAPHE, CG_NOM_PARAM    , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VALEUR_PARAM , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_UNITE_PARAM  , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_SYMBOLE_PARAM, FORCE_OUI );
  }
  I_CG_InstallerGraphe( infos );
  I_CG_InstallerParametre( infos );
}

static void I_CG_InstallerParametre(struct infos_arbre *infos)
{
  struct parametre *parametre_actif;

  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s, %s l.%u : Informations absentes !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* L'ascenseur des paramtres */
  ARB_ascenseur( CREER_GRAPHE, CG_PARAM_AVANT, CG_PARAM_FONDASC, CG_PARAM_BOUTON, CG_PARAM_APRES, ASC_VERTICAL,
		 infos->nombre_parametres, 1, infos->numero_parametre_affiche,
		 ROOT, 0, 0, METTRE );

  if ( ( 0 == infos->nombre_parametres ) || ( NULL == infos->parametres ) ) {
    /* Aucun paramtre cr : on dsactive tout */
    OBJ_active( CREER_GRAPHE, CG_EFFACER_PARAM, FORCE_NON );

    OBJ_active( CREER_GRAPHE, CG_NOM_PARAM    , FORCE_NON );
    OBJ_active( CREER_GRAPHE, CG_VALEUR_PARAM , FORCE_NON );
    OBJ_active( CREER_GRAPHE, CG_UNITE_PARAM  , FORCE_NON );
    OBJ_active( CREER_GRAPHE, CG_SYMBOLE_PARAM, FORCE_NON );

    return;
  }

  /* Quel est le paramtre affich ? */
  if ( infos->numero_parametre_affiche >= infos->nombre_parametres ) {
    (void) fprintf(stderr, "%s, %s l.%u : Index de paramtre affich absurde !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    infos->numero_parametre_affiche = 0;
  }
  parametre_actif = infos->parametres + infos->numero_parametre_affiche;

  /* On affiche ses proprites */
  OBJ_CopierTexte( CREER_GRAPHE, CG_NOM_PARAM, parametre_actif->nom, 0 );
  OBJ_CopierTexte( CREER_GRAPHE, CG_UNITE_PARAM, parametre_actif->unite, 0 );
  OBJ_CopierTexte( CREER_GRAPHE, CG_SYMBOLE_PARAM, parametre_actif->symbole, 0 );

  OBJ_OptionsEditable( CREER_GRAPHE, CG_VALEUR_PARAM, BI_EDITABLE_MIN | BI_EDITABLE_MAX | BI_EDITABLE_UNITE,
		       parametre_actif->variable.min, parametre_actif->variable.max, parametre_actif->unite );
  OBJ_MettreNombre( CREER_GRAPHE, CG_VALEUR_PARAM, VAR_LireValeur( &(parametre_actif->variable), VARIABLE_LIRE_VALEUR_INITIALE ), 0.0 );
}

static void I_CG_InstallerVariable(struct infos_arbre *infos)
{
  struct info_variable *variable;

  if ( ( NULL == infos ) || ( NULL == infos->graphes ) || ( infos->variable_affichee >= NOMBRE_VARIABLES_UTILISABLES ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Infos incompltes, absentes ou absurdes [i=%p, i->g=%p, i->va=%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, (void *) infos, (void *) (infos->graphes), infos->variable_affichee);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }
  /* On se facilite l'accs  la variable... */
  variable = infos->graphes[ infos->numero_graphe_affiche ].X + infos->variable_affichee;

  /* On dslectionne tous les boutons */
  OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_X, FORCE_NON );
  OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_Y, FORCE_NON );
  OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_Z, FORCE_NON );
  OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_T, FORCE_NON );
  OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_U, FORCE_NON );
  OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_I, FORCE_NON );

  /* On choisit le bouton de la variable choisie */
  switch ( infos->variable_affichee )
  {
  case VARIABLE_AFFICHEE_X:
    OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_X, FORCE_OUI );
    break;
  case VARIABLE_AFFICHEE_Y:
    OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_Y, FORCE_OUI );
    break;
  case VARIABLE_AFFICHEE_Z:
    OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_Z, FORCE_OUI );
    break;
  case VARIABLE_AFFICHEE_T:
    OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_T, FORCE_OUI );
    break;
  case VARIABLE_AFFICHEE_U:
    OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_U, FORCE_OUI );
    break;
  case VARIABLE_AFFICHEE_I:
    OBJ_choisit( CREER_GRAPHE, CG_VARIABLE_I, FORCE_OUI );
    break;
  default:
    (void) fprintf(stderr, "%s, %s l.%u : Type de variable inconnu !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* On met le bon texte */
  OBJ_RemplacerTexte( CREER_GRAPHE, CG_TXT_VARIABLE, I_CG_TexteVariable( infos->variable_affichee ), VRAI );

  /* On met l'unit & la lgende */
  OBJ_CopierTexte( CREER_GRAPHE, CG_UNITE, variable->unite, 0 );
  OBJ_CopierTexte( CREER_GRAPHE, CG_LEGENDE, variable->nom, 0 );
  OPTION_A_FAIRE( "Afficher le type de grandeur physique de la variable" );

  OBJ_OptionsEditable( CREER_GRAPHE, CG_VAL_MIN, BI_EDITABLE_UNITE, 0.0, 0.0, variable->unite );
  OBJ_OptionsEditable( CREER_GRAPHE, CG_VALEUR_MAX, BI_EDITABLE_UNITE, 0.0, 0.0, variable->unite );
  OBJ_OptionsEditable( CREER_GRAPHE, CG_VAL_PAS, BI_EDITABLE_MIN | BI_EDITABLE_UNITE, 0.0, 0.0, variable->unite );

  /* On met les informations de calcul */
  OBJ_MettreNombre( CREER_GRAPHE, CG_VAL_MIN, variable->min, 0.0 );
  OBJ_MettreNombre( CREER_GRAPHE, CG_VALEUR_MAX, variable->max, 0.0 );
  OBJ_MettreNombre( CREER_GRAPHE, CG_VAL_PAS, variable->pas, 0.0 );
  OBJ_MettreEntier_U( CREER_GRAPHE, CG_VALEUR_N, variable->nombre );

  OBJ_RemplacerTexte( CREER_GRAPHE, CG_METHODE_VAR, I_CG_TexteMethode( variable->type ), VRAI );

  /* On met les valeurs, si dj connues... */
  OBJ_CopierTexte( CREER_GRAPHE, CG_VALEURS, variable->valeurs, 0 );
  if ( variable->graphe != NULL ) {
    OBJ_CopierTexte( CREER_GRAPHE, CG_GRF_DES_X, variable->graphe->nom, LGMAX_GRAPHE_ORIGINE );
  } else {
    OBJ_CopierTexte( CREER_GRAPHE, CG_GRF_DES_X, txt_pas_de_graphe, LGMAX_GRAPHE_ORIGINE );
  }

  /* On dsactive tous les champs */
  OBJ_active( CREER_GRAPHE, CG_VAL_MIN   , FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_VALEUR_MAX, FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_VAL_PAS   , FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_VALEUR_N  , FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_VALEURS   , FORCE_NON );
  OBJ_active( CREER_GRAPHE, CG_GRF_DES_X , FORCE_NON );

  /* En fonction du type de variable, on active les champs adquats */
  switch ( variable->type )
  {
  case VARIABLE_INUTILISEE:
    if ( VARIABLE_AFFICHEE_I == infos->variable_affichee ) {
      /* Cette variable permet d'indiquer le nombre de points, mais uniquement si pas fait ailleurs ! */
      OBJ_active( CREER_GRAPHE, CG_VALEUR_N  , FORCE_OUI );
    }
    break;

  case VARIABLE_MIN_MAX_PAS:
    OBJ_active( CREER_GRAPHE, CG_VAL_MIN   , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VALEUR_MAX, FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VAL_PAS   , FORCE_OUI );
    break;

  case VARIABLE_MIN_PAS_NOMBRE:
    OBJ_active( CREER_GRAPHE, CG_VAL_MIN   , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VAL_PAS   , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VALEUR_N  , FORCE_OUI );
    break;

  case VARIABLE_MAX_PAS_NOMBRE:
    OBJ_active( CREER_GRAPHE, CG_VALEUR_MAX, FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VAL_PAS   , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VALEUR_N  , FORCE_OUI );
    break;

  case VARIABLE_MIN_MAX_NOMBRE:
    OBJ_active( CREER_GRAPHE, CG_VAL_MIN   , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VALEUR_MAX, FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VALEUR_N  , FORCE_OUI );
    break;

  case VARIABLE_X_GRAPHE:
  case VARIABLE_Y_GRAPHE:
    OBJ_active( CREER_GRAPHE, CG_GRF_DES_X , FORCE_OUI );
    break;

  case VARIABLE_VALEURS:
  case VARIABLE_FORMULE:
    OBJ_active( CREER_GRAPHE, CG_VALEURS   , FORCE_OUI );
    break;
  }
}

/* -------- Manipulation des structures de donnes -------- */

static void I_CG_LibererInfos(struct infos_arbre *infos)
{
  unsigned long i;

  if ( NULL == infos ) return;

  /* On libre les infos de graphes */
  for ( i = 0 ; i < infos->nombre_graphes ; i++ ) {
    I_CG_LibererGraphe( infos, i );
  }
  free( infos->graphes ); infos->graphes = NULL;
  infos->nombre_graphes = 0;

  /* On libre les noms de domaines */
  /* On met au pralable un texte sr dans le bouton ! */
  OBJ_RemplacerTexte( CREER_GRAPHE, CG_DOMAINE, txt_choisir_domaine, VRAI );

  for ( i = 0 ; i < infos->nombre_domaines ; i++ ) {
    free( (char *) infos->domaines[ i ] ); infos->domaines[ i ] = NULL;
  }
  free( infos->domaines ); infos->domaines = NULL;

  /* On libre les infos de modles */
  I_CG_LibererModeles( infos );
}

static void I_CG_LibererGraphe(struct infos_arbre *infos, unsigned long numero_graphe)
{
  struct infos_graphe *graphe;
  unsigned int i;

  if ( numero_graphe >= infos->nombre_graphes ) {
    (void) fprintf(stderr, "%s, %s l.%u : Numro de graphe absurde [%lu, max %lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, numero_graphe, infos->nombre_graphes - 1);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }

  /* On se facilite l'accs  la structure */
  graphe = infos->graphes + numero_graphe;

  /* On libre le nom */
  free( graphe->nom ); graphe->nom = NULL;

  /* On libre les infos des variables */
  for ( i = 0 ; i < NOMBRE_VARIABLES_UTILISABLES ; i++ ) {
    free( graphe->X[ i ].nom ); graphe->X[ i ].nom = NULL;
    free( graphe->X[ i ].unite ); graphe->X[ i ].unite = NULL;
    free( graphe->X[ i ].valeurs ); graphe->X[ i ].valeurs = NULL;

    graphe->X[ i ].graphe = NULL;
  }
}

static void I_CG_LibererModeles(struct infos_arbre *infos)
{
  unsigned long i;

  /* On met au pralable un texte sr dans le bouton ! */
  OBJ_RemplacerTexte( CREER_GRAPHE, CG_MODELE, txt_aucun_modele, VRAI );

  /* Libration des noms de modles */
  for ( i = 0 ; i < infos->nombre_modeles ; i++ ) {
    free( infos->modeles[ i ] ); infos->modeles[ i ] = NULL;
    free( infos->fichiers_modeles[ i ] ); infos->fichiers_modeles[ i ] = NULL;
  }

  /* Libration des tables associes */
  free( infos->modeles ); infos->modeles = NULL;
  free( infos->fichiers_modeles ); infos->fichiers_modeles = NULL;

  /* Il ne reste donc plus aucun modle */
  infos->nombre_modeles = 0;
}

static int I_CG_ReserverInfosGraphe(struct infos_arbre *infos, unsigned long nombre_graphes)
{
  struct infos_graphe *graphes;
  unsigned long i;

  if ( 0 == nombre_graphes ) {
    (void) fprintf(stderr, "%s, %s l.%u : Valeur absurde du nombre de graphes  rserver !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }

  /* Faut-il vraiment rserver ? */
  if ( nombre_graphes <= infos->nombre_graphes ) {
    return 0;
  }

  /* On prpare pour les nouveaux graphes */
  graphes = (struct infos_graphe *) realloc( infos->graphes, nombre_graphes * sizeof(struct infos_graphe) );
  if ( NULL == graphes ) {
    ERR_Memoire( NULL, nombre_graphes * sizeof(struct infos_graphe) );
    return -2;
  }

  /* On initialise les nouveaux graphes  0 */
  for ( i = 0 ; i < (nombre_graphes - infos->nombre_graphes) ; i++ ) {
    unsigned int j;
    graphes[ i ].nom = NULL;
    graphes[ i ].type = GR_SSTYPE;

    for ( j = 0 ; j < NOMBRE_VARIABLES_UTILISABLES ; j++ ) {
      graphes[ i ].X[ j ].nom = NULL; /* Sans lgende initialement... */
      graphes[ i ].X[ j ].type = VARIABLE_MIN_MAX_PAS;
      graphes[ i ].X[ j ].grandeur = 0;
      graphes[ i ].X[ j ].unite = NULL;
      graphes[ i ].X[ j ].min = 0.0; /* Purement arbitraire ! */
      graphes[ i ].X[ j ].max = 1.0;
      graphes[ i ].X[ j ].pas = 0.1;
      graphes[ i ].X[ j ].nombre = 100;
      graphes[ i ].X[ j ].valeurs = NULL;
      graphes[ i ].X[ j ].graphe = NULL;
    }
  }

  /* On indique le bon nouveau nombre de graphes... */
  infos->graphes = graphes;
  infos->nombre_graphes = nombre_graphes;

  /* On affiche le bon nombre total de graphe */
  OBJ_MettreEntier_U( CREER_GRAPHE, CG_TOTAL_GRAPHE, infos->nombre_graphes );

  /* Tout s'est bien pass */
  return 0;
}

static void I_CG_AjouterParametre(struct infos_arbre *infos, unsigned long nombre)
{
  struct parametre *nv_parametres;
  unsigned long i;

  if ( NULL == infos ) return;
  if ( 0 == nombre ) return;	/* Aucun paramtre  ajouter ! */

  /* On rserve pour le nouveau nombre de paramtres */
  nv_parametres = (struct parametre *) realloc( infos->parametres, sizeof( struct parametre ) * ( infos->nombre_parametres + nombre ) );
  if ( NULL == nv_parametres ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de rserver la table des paramtres\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* On installe la nouvelle table  la place de l'ancienne */
  infos->parametres = nv_parametres;

  /* On initialise ces nouveaux paramtres */
  for ( i = 0 ; i < nombre ; i++ ) {
    nv_parametres[ infos->nombre_parametres + i ].symbole = NULL;
    nv_parametres[ infos->nombre_parametres + i ].unite = NULL;
    nv_parametres[ infos->nombre_parametres + i ].nom = NULL;

    nv_parametres[ infos->nombre_parametres + i ].variable.valeur_initiale = 0.0;
    nv_parametres[ infos->nombre_parametres + i ].variable.min = 0.0;
    nv_parametres[ infos->nombre_parametres + i ].variable.max = 0.0;
    nv_parametres[ infos->nombre_parametres + i ].variable.delta = 0.0;
    nv_parametres[ infos->nombre_parametres + i ].variable.type = VAR_CONSTANTE;
    nv_parametres[ infos->nombre_parametres + i ].variable.contenu.valeur = 0.0;
  }

  /* On s'assure que l'on affiche bien le premier cr */
  infos->numero_parametre_affiche = infos->nombre_parametres;

  /* On adapte l'affichage ventuellement */
  if ( 0 == infos->nombre_parametres ) {
    /* On vient de crer le premier paramtre : activer les boutons */
    OBJ_active( CREER_GRAPHE, CG_EFFACER_PARAM, FORCE_OUI );

    OBJ_active( CREER_GRAPHE, CG_NOM_PARAM    , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_VALEUR_PARAM , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_UNITE_PARAM  , FORCE_OUI );
    OBJ_active( CREER_GRAPHE, CG_SYMBOLE_PARAM, FORCE_OUI );
  }

  /* On mmorise le nouveau nombre */
  infos->nombre_parametres += nombre;
}

static void I_CG_EffacerParametre(struct infos_arbre *infos, unsigned long numero)
{
  struct parametre *param_efface;
  unsigned long i;

  if ( NULL == infos ) return;
  if ( numero >= infos->nombre_parametres ) {
    AVERTISSEMENT( "Demande d'effacer le paramtre %lu, seuls 1-%lu prsents", numero + 1, infos->nombre_parametres );
    return;
  }

  /* On repre celui qui est effac */
  param_efface = infos->parametres + numero;

  /* On l'efface */
  free( param_efface->symbole ); param_efface->symbole = NULL;
  free( param_efface->unite ); param_efface->unite = NULL;
  free( param_efface->nom ); param_efface->nom = NULL;

  /* On dcale les suivants */
  for ( i = numero + 1 ; i < infos->nombre_parametres ; i++ ) {
    infos->parametres[ i - 1 ] = infos->parametres[ i ];
  }

  /* On corrige le nombre de paramtres */
  infos->nombre_parametres--;
  if ( infos->numero_parametre_affiche >= infos->nombre_parametres ) {
    /* On tait donc sur le dernier ! */
    if ( infos->nombre_parametres > 0 ) {
      infos->numero_parametre_affiche = infos->nombre_parametres - 1;
    } else {
      infos->numero_parametre_affiche = 0;
    }
  }
}

static const char *I_CG_TexteMethode(unsigned int type_variable)
{
  if ( type_variable >= NOMBRE_METHODES_UTILISABLES ) {
    (void) fprintf(stderr, "%s, %s l.%u : Code de mthode incorrect [%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, type_variable);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  return TXT_NomsMethodes[ type_variable ];
}

static const char *I_CG_TexteVariable(unsigned int num_variable)
{
  if ( num_variable >= NOMBRE_VARIABLES_UTILISABLES ) {
    (void) fprintf(stderr, "%s, %s l.%u : Code de variable incorrect [%d, valide : 0  %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, num_variable, NOMBRE_VARIABLES_UTILISABLES - 1);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  return TXT_NomsVariables[ num_variable ];
}

/* -------- Gestion de l'arbre -------- */

static int I_CG_GererArbre(unsigned int objet, int x, int y)
{
  int retour;
  struct infos_arbre *infos;

  /* On rcupre les infos de l'arbre */
  infos = (struct infos_arbre *) ARB_LireComplement( CREER_GRAPHE );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s, %s l.%u : Aucune information associe  l'arbre !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_TRAITE;
  }

  switch ( objet )
  {
  case CG_GENRE_GRAPHE:		/* Choisir le genre de graphe gnr */
    retour = OBJ_ChoisirGenreGraphe( CREER_GRAPHE, CG_GENRE_GRAPHE, LGMAX_GENRE_GRAPHE, infos->genre, QUESTION );
    if ( retour >= 0 ) {	/* On a bien chang le genre */
      if ( ( GRF_HISTOGRAMME == retour ) || ( GRF_FAISCEAU == retour ) || ( GRF_MULTI_SERIE == retour ) ||
	   ( GRF_ERREUR_R == retour ) || ( GRF_ERREUR_C == retour ) ) {
	BI_ERR_annuler( ERR_genre_incorrect );

	/* On remet le genre prcdent */
	OBJ_RemplacerTexte( CREER_GRAPHE, CG_GENRE_GRAPHE, GRF_TXT_Genre( infos->genre ), VRAI );
      } else {
	infos->genre = retour;

	/* On installe l'affichage associ */
	infos->affichage = GRF_R_CARTESIEN;
	OBJ_RemplacerTexte( CREER_GRAPHE, CG_TYPE_AFFICHAG, GRF_TXT_TypeAffichage( infos->genre, infos->affichage ), VRAI );

	/* On active les variables en fonction de cela */
	I_CG_ActiverVariable( infos );
      }
    }
    return BI_EVENEMENT_TRAITE;

  case CG_TYPE_AFFICHAG:
    retour = OBJ_ChoisirAffichage( CREER_GRAPHE, CG_TYPE_AFFICHAG, infos->genre, infos->affichage, QUESTION );
    if ( retour >= 0 ) {
      infos->affichage = retour;
    }
    return BI_EVENEMENT_TRAITE;

  case CG_MULTI_GRAPHES:
    retour = OBJ_choisit( CREER_GRAPHE, objet, ECHANGE );
    if ( VRAI == retour ) {
      OBJ_active( CREER_GRAPHE, CG_AJOUTER_GRF, FORCE_OUI );
    } else {
      unsigned long i;

      /* On efface les infos des graphes autres que le premier */
      for ( i = 1 ; i < infos->nombre_graphes ; i++ ) {
	I_CG_LibererGraphe( infos, i );
      }

      /* On fait semblant qu'il n'y a qu'un graphe --- de toute faon, le reste sera libr  la fin... */
      infos->nombre_graphes = 1;
      infos->numero_graphe_affiche = 0;

      /* On remet l'ascenseur et les boutons comme il faut... */
      (void) ARB_ascenseur( CREER_GRAPHE, CG_GRF_AVANT, CG_FOND_ASC, CG_BOUTON_GRF, CG_GRF_APRES, ASC_HORIZONTAL,
			    infos->nombre_graphes, 1, infos->numero_graphe_affiche, ROOT, 0, 0, METTRE );
      OBJ_active( CREER_GRAPHE, CG_AJOUTER_GRF, FORCE_NON );
      OBJ_active( CREER_GRAPHE, CG_EFFACER_GRF, FORCE_NON );
    }
    return BI_EVENEMENT_TRAITE;

  case CG_TYPE_GRAPHE:
    retour = OBJ_ChoisirTypeGraphe( CREER_GRAPHE, CG_TYPE_GRAPHE, LGMAX_TYPE_GRAPHE, infos->graphes[ infos->numero_graphe_affiche ].type, QUESTION, FAUX );
    if ( retour != infos->graphes[ infos->numero_graphe_affiche ].type ) {
      infos->graphes[ infos->numero_graphe_affiche ].type = retour; /* Nouveau type de graphe cr... */
    }
    return BI_EVENEMENT_TRAITE;

  case CG_VARIABLE_X:
  case CG_VARIABLE_Y:
  case CG_VARIABLE_Z:
  case CG_VARIABLE_T:
  case CG_VARIABLE_U:
  case CG_VARIABLE_I:
    /* On mmorise la variable actuelle */
    I_CG_EnregistrerVariable( infos );

    /* On enregistre quelle est cette nouvelle variable [Suppose le mme ordre de codage et des boutons, successifs dans l'arbre] */
    infos->variable_affichee = ( objet - CG_VARIABLE_X ) + VARIABLE_AFFICHEE_X;

    /* On installe la nouvelle variable choisie */
    I_CG_InstallerVariable( infos );
    return BI_EVENEMENT_TRAITE;

  case CG_METHODE_VAR:
    retour = SS_menu_texte( TXT_NomsMethodes, NOMBRE_METHODES_UTILISABLES, titre_menu_methodes, CREER_GRAPHE, CG_METHODE_VAR );
    if ( retour >= 0 ) {
      /* On enregistre les valeurs actuelles */
      I_CG_EnregistrerVariable( infos );

      /* On change le type */
      infos->graphes[ infos->numero_graphe_affiche ].X[ infos->variable_affichee ].type = retour;

      /* Et on raffiche la variable */
      I_CG_InstallerVariable( infos );
    }
    return BI_EVENEMENT_TRAITE;

  case CG_NV_PARAM:
    OBJ_choisit( CREER_GRAPHE, CG_NV_PARAM, FORCE_OUI );
    I_CG_EnregistrerParametre( infos );	/* Enregistrer l'actuel, puisqu'il sera remplac par le nouveau  l'affichage ! */
    I_CG_AjouterParametre( infos, 1 );
    I_CG_InstallerParametre( infos );
    OBJ_choisit( CREER_GRAPHE, CG_NV_PARAM, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case CG_EFFACER_PARAM:
    OBJ_choisit( CREER_GRAPHE, objet, FORCE_OUI );
    I_CG_EffacerParametre( infos, infos->numero_parametre_affiche );
    I_CG_InstallerParametre( infos );
    OBJ_choisit( CREER_GRAPHE, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case CG_PARAM_AVANT:
  case CG_PARAM_FONDASC:
  case CG_PARAM_BOUTON:
  case CG_PARAM_APRES:
    /* On enregistre le paramtre actuel */
    I_CG_EnregistrerParametre( infos );

    /* On se dplace */
    infos->numero_parametre_affiche = ARB_ascenseur( CREER_GRAPHE, CG_PARAM_AVANT, CG_PARAM_FONDASC, CG_PARAM_BOUTON, CG_PARAM_APRES, ASC_VERTICAL,
						     infos->nombre_parametres, 1, infos->numero_parametre_affiche,
						     objet, x, y, ECHANGE );

    /* On installe le nouveau paramtre */
    I_CG_InstallerParametre( infos );
    return BI_EVENEMENT_TRAITE;

  case CG_DOMAINE:
    OBJ_choisit( CREER_GRAPHE, objet, FORCE_OUI );
    {
      int retour;

      retour = SS_menu_texte( infos->domaines, infos->nombre_domaines, titre_menu_domaines, CREER_GRAPHE, CG_DOMAINE );
      if ( retour >= 0 ) {
	char *nom_dossier = NULL;

	infos->domaine_choisi = retour;
	OBJ_RemplacerTexte( CREER_GRAPHE, objet, infos->domaines[ infos->domaine_choisi ], VRAI );

	/* On efface les anciens modles */
	I_CG_LibererModeles( infos );

	/* On rcupre le nom du dossier de cette catgorie */
	nom_dossier = MODELE_DossierDomaine( infos->domaines[ infos->domaine_choisi ] );
	if ( NULL == nom_dossier ) {
	  OBJ_choisit( CREER_GRAPHE, CG_MODELE, FORCE_NON );
	  return BI_EVENEMENT_TRAITE;
	}

	/* On lit les modles disponibles dans cette catgorie */
	retour = MODELE_ListeModeles( nom_dossier, &(infos->nombre_modeles), &(infos->modeles), &(infos->fichiers_modeles) );
	if ( ( retour != 0 ) || ( 0 == infos->nombre_modeles ) ) {
	  /* Aucun modle */
	  OBJ_active( CREER_GRAPHE, CG_MODELE, FORCE_NON );
	} else {
	  OBJ_active( CREER_GRAPHE, CG_MODELE, FORCE_OUI );
	  OBJ_RemplacerTexte( CREER_GRAPHE, CG_MODELE, infos->modeles[ 0 ], VRAI );
	}

	/* Termin */
	free( nom_dossier ); nom_dossier = NULL;
      }
    }
    OBJ_choisit( CREER_GRAPHE, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case CG_MODELE:
    OBJ_choisit( CREER_GRAPHE, objet, FORCE_OUI );
    {
      int retour;

      retour = SS_menu_texte( (const char **) infos->modeles, infos->nombre_modeles, titre_menu_modeles, CREER_GRAPHE, CG_MODELE );
      if ( retour >= 0 ) {
	struct modele_generique *modele = NULL;

	/* On charge le modle correspondant ! */
	modele = MODELE_LireModele( infos->fichiers_modeles[ retour ] );
	if ( modele != NULL ) {
	  /* On installe les paramtres de ce modle dans l'arbre */
	  I_CG_InstallerModele( infos, modele );

	  /* On n'a plus besoin du modle */
	  OPTION_A_FAIRE( "Effacer le modle gnrique !" );
	} else {
	  (void) fprintf(stderr, "%s,%s l.%u : ERREUR en voulant lire le modle dans %s\n",
			 __FILE__, __FUNCTION__, __LINE__, infos->fichiers_modeles[ retour ]);
	}
      }
    }
    OBJ_choisit( CREER_GRAPHE, objet, FORCE_NON );
    return BI_EVENEMENT_TRAITE;
  }

  /* Objet non-trait : traitement par dfaut ! */
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Mmorisation des donnes de l'arbre -------- */

static void I_CG_EnregistrerGraphe(struct infos_arbre *infos)
{
  struct infos_graphe *graphe;
  char *texte;

  /* On se facilite l'accs au graphe courant */
  if ( infos->numero_graphe_affiche >= infos->nombre_graphes ) {
    (void) fprintf(stderr, "%s, %s l.%u : Numro de graphe absurde [%lu, devrait tre entre 0 et %lu]\n",
		   __FILE__, __FUNCTION__, __LINE__, infos->numero_graphe_affiche, infos->nombre_graphes - 1);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }
  graphe = infos->graphes + infos->numero_graphe_affiche;

  /* On mmorise le nom du graphe */
  free( graphe->nom ); graphe->nom = NULL;
  texte = lit_chaine( CREER_GRAPHE, CG_NOM_GRAPHE, 0, NULL, TXT_REMPLACE );
  if ( texte != NULL ) {
    graphe->nom = strdup( texte );
  }

  /* Son type a dj t mmoris au vol... */

  /* On mmorise la variable courante */
  I_CG_EnregistrerVariable( infos );
}

static void I_CG_EnregistrerParametre(struct infos_arbre *infos)
{
  char *texte;
  struct parametre *parametre_actif;

  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s, %s l.%u : Informations absentes !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  if ( ( 0 == infos->nombre_parametres ) || ( NULL == infos->parametres ) ) {
    /* Aucun paramtre rien  faire. Peu se produire quand on quitte l'arbre... */
/*     (void) fprintf(stderr, "%s, %s l.%u : Demander d'enregistrer un paramtre, mais aucun paramtre prsent !\n", */
/* 		   __FILE__, __FUNCTION__, __LINE__); */
/*     FIN_ANTICIPEE( FIN_ERREUR_GRAVE ); */
    return;
  }

  /* Quel est le paramtre affich ? */
  if ( infos->numero_parametre_affiche >= infos->nombre_parametres ) {
    (void) fprintf(stderr, "%s, %s l.%u : Index de paramtre affich absurde !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }
  parametre_actif = infos->parametres + infos->numero_parametre_affiche;

  /* Enregistrement proprement dit */
  free( parametre_actif->nom ); parametre_actif->nom = NULL;
  texte = lit_chaine( CREER_GRAPHE, CG_NOM_PARAM, 0, NULL, TXT_REMPLACE );
  if ( texte != NULL ) {
    parametre_actif->nom = strdup( texte );
  }

  free( parametre_actif->unite ); parametre_actif->unite = NULL;
  texte = lit_chaine( CREER_GRAPHE, CG_UNITE_PARAM, 0, NULL, TXT_REMPLACE );
  if ( texte != NULL ) {
    parametre_actif->unite = strdup( texte );
  }

  free( parametre_actif->symbole ); parametre_actif->symbole = NULL;
  texte = lit_chaine( CREER_GRAPHE, CG_SYMBOLE_PARAM, 0, NULL, TXT_REMPLACE );
  if ( texte != NULL ) {
    parametre_actif->symbole = strdup( texte );
  }

  parametre_actif->variable.contenu.valeur  = OBJ_LireNombre( CREER_GRAPHE, CG_VALEUR_PARAM, NULL );
  parametre_actif->variable.valeur_initiale = parametre_actif->variable.contenu.valeur;
}

static void I_CG_EnregistrerVariable(struct infos_arbre *infos)
{
  struct info_variable *variable;
  char *texte;

  /* On se facilite l'accs  la variable courante */
  if ( infos->variable_affichee >= NOMBRE_VARIABLES_UTILISABLES ) {
    (void) fprintf(stderr, "%s, %s l.%u : Code de variable affich absurde [%u, max = %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, infos->variable_affichee, NOMBRE_VARIABLES_UTILISABLES );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }
  variable = infos->graphes[ infos->numero_graphe_affiche ].X + infos->variable_affichee;

  /* Le type, l'origine et le graphe ont t modifis au vol -> inutile de s'en occuper */

  /* La lgende = nom de la variable */
  free( variable->nom ); variable->nom = NULL;
  texte = lit_chaine( CREER_GRAPHE, CG_LEGENDE, 0, NULL, TXT_REMPLACE );
  if ( texte != NULL ) {
    variable->nom = strdup( texte );
  }

  /* L'unit */
  free( variable->unite ); variable->unite = NULL;
  texte = lit_chaine( CREER_GRAPHE, CG_UNITE, 0, NULL, TXT_REMPLACE );
  if ( texte != NULL ) {
    variable->unite = strdup( texte );
  }

  /* Les valeurs numriques [mme si inutiles...] */
  variable->min = OBJ_LireNombre( CREER_GRAPHE, CG_VAL_MIN, NULL );
  variable->max = OBJ_LireNombre( CREER_GRAPHE, CG_VALEUR_MAX, NULL );
  variable->pas = OBJ_LireNombre( CREER_GRAPHE, CG_VAL_PAS, NULL );
  variable->nombre = OBJ_LireEntier_U( CREER_GRAPHE, CG_VALEUR_N );

  /* La liste de valeurs, ou la fonction */
  free( variable->valeurs ); variable->valeurs = NULL;
  texte = lit_chaine( CREER_GRAPHE, CG_VALEURS, 0, NULL, TXT_REMPLACE );
  if ( texte != NULL ) {
    variable->valeurs = strdup( texte );
  }
}
