/* LASE - Logiciel d'analyse des spectres exprimentaux
   (C) 1998-2004, Emmanuel CURIS

   Module d'interface utilisateur

   Cration, aperu et modification d'une figure
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"
#include "exafs.h"

#include "fichiers.h"		/* Pour export PS */
#include "figures.h"
#include "graphes.h"
#include "interface.h"

/* Options pour le sous-menu des options sur une figure */
#define OPT_FIGURE_NOUVEAU_CADRE  0
#define OPT_FIGURE_EFFACER_CADRE  1
#define OPT_FIGURE_COPIER_CADRE   2
#define OPT_FIGURE_MODIFIER_CADRE 3 /* 4 = sparateur */
#define OPT_PREMIER_PLAN       5
#define OPT_DERNIER_PLAN       6 /* 7 = sparateur */
#define OPT_EXPORTER_IMAGE     8
#define OPT_IMPRIMER_IMAGE     9 /* TOS uniquement */

#define NOMBRE_OPTIONS_FIGURES 10

#include "int_figures.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void RDS_Figure(Fenetre *, int, int, unsigned int, unsigned int);
static void RCT_Figure(Fenetre *fen, int x, int y, int vx_x, int vx_y);
static int  MSG_Figure(Fenetre *fen, Message *message);
static int  CLC_Figure(Fenetre *fen, int bouton, int n_clic, int etat, int x, int y);
static void CLC_SousMenu(Fenetre *fenetre, Figure *figure);

static struct bloc_figure *RCT_TrouverBloc(Figure *figure, double x_mm, double y_mmn, unsigned int niveau);

static int  BLOC_GererArbre   (unsigned int objet, int x, int y);
static int  BLOC_InstallerType(Bloc_Figure *bloc);
static int  BLOC_Modifier     (Fenetre *mere, Figure *figure, Bloc_Figure *bloc);
static void BLOC_PreparerArbre(Figure *figure, Bloc_Figure *bloc);

static const char *BLOC_TexteGraphe(unsigned int arbre, unsigned long numero);

/* ~~~~~~~~ Realisation pratique ~~~~~ */

/* -------- Ouvrir une fentre contenant la figure indique -------- */

Fenetre *FIG_OuvrirFenetre(Figure *figure)
{
  Fenetre *fenetre;

  /* Une fentre contenant la figure indique existe-t-elle dj ? */
  if ( figure != NULL ) {
    OPTION_A_FAIRE( "Chercher la fentre contenant cette figure" );
  }

  /* Si on est l, c'est que l'on n'a pas trouv de fentre */
  if ( NULL == figure ) {
    /* Normal, on cre une nouvelle figure ! */
    figure = FIG_CreerFigure();
    if ( NULL == figure ) {
      (void) fprintf(stderr, "%s, %s l.%u : Impossible de crer la figure\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }
  }

  /* On cre donc la fentre */
  fenetre = FEN_CreerFenetreSimple( TOUS_SF_ASC, titre_fenetre_figures, "",
				    RDS_Figure, (void (*)(Fenetre *, int, int, int)) NULL,
				    (int (*)(Fenetre *, unsigned int, int)) NULL,
				    CLC_Figure, MSG_Figure, RCT_Figure,
				    (int (*)(Fenetre *)) NULL );
  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible de crer la fentre\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  /* On prpare pour les messages */
  fenetre->msg_masque.taille = VRAI;
  fenetre->drapeaux.redessin = VRAI;

  /* On associe la figure  la fentre */
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_1, figure );

  /* Et on ouvre la fentre */
  FEN_Ouvrir( fenetre );

  /* Fentre cre */
  return fenetre;
}

/* ~-~-~-~-~-~-~-~-~ Redessin ~-~-~-~-~-~-~-~-~ */

static void RDS_Figure(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  Figure *figure;
  struct bloc_figure *bloc;
  double mm_v_pixels;		/* Conversion pixels <=> mm, en mm/pixel */
  unsigned int larg_figure, haut_figure; /* Dimensions de la figure  l'cran */

  /* Prparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut);

  /* Quelle est la figure de la fentre ? */
  figure = (Figure *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == figure ) {
    TRC_Message( fenetre, texte_erreur_aucune_figure );
    return;
  }

  /* Facteur de conversion mm <-> cran, pour garder le rapport largeur / hauteur */
  mm_v_pixels = TRC_Echelle2D( (double) figure->largeur, (double) figure->hauteur, fenetre->w_larg, fenetre->w_haut );
  if ( mm_v_pixels <= 0.0 ) {
    (void) fprintf( stderr, "%s, %s l.%u : Valeur d'chelle absurde [%g]\n",
		    __FILE__, __FUNCTION__, __LINE__, mm_v_pixels );
    return;
  }

  /* On en dduit les dimensions de la figure  l'cran */
  larg_figure = (unsigned int) ( ( (double) figure->largeur ) / mm_v_pixels );
  haut_figure = (unsigned int) ( ( (double) figure->hauteur ) / mm_v_pixels );

  /* On trace le tour de la page, en rouge */
  TRC_Couleur( ROUGE );
  TRC_Rectangle( fenetre, 0, 0, larg_figure, haut_figure );

  /* On trace les marges, en pointill bleu */
  TRC_Couleur( BLEU );
  TRC_Style( DASHDOTDOT );

  /* a. horizontales, G puis D */
  TRC_LigneV( fenetre, figure->marge_gauche / mm_v_pixels, 0, haut_figure );
  TRC_LigneV( fenetre, (figure->largeur - figure->marge_droite) / mm_v_pixels, 0, haut_figure );

  /* b. verticales, H puis B */
  TRC_LigneH( fenetre, 0, figure->marge_haute / mm_v_pixels, larg_figure );
  TRC_LigneH( fenetre, 0, (figure->hauteur - figure->marge_basse) / mm_v_pixels, larg_figure );

  if ( NULL == figure->premier_bloc ) {
    TRC_Z_Message( fenetre, 0, 0, larg_figure, haut_figure, texte_erreur_aucun_bloc );
    return;
  }

  /* Faut-il calculer la dimension cran de chaque bloc ? */
  if ( figure->proprietes & FIGURE_A_REECHELONNER ) {
    for ( bloc = figure->premier_bloc ; bloc != NULL ; bloc = bloc->suivant ) {
      TRC_RelatifVAbsolu( &(bloc->dessin), 0, haut_figure, larg_figure, haut_figure );

#ifdef _DEVELOPPE_
      if ( bloc->dessin.y_BG < bloc->dessin.hauteur ) {
	/* Cela revient  y_BG - hauteur < 0 ==> coint haut en-dehors de la fentre */
	(void) fprintf(stderr, "%s, %s l.%u : Coin haut du bloc HORS-FENTRE [X=%d(%g%%), Y=%d(%g%%), L=%u(%g%%), H=%u(%g%%) ; fentre : X=[0..%u],Y=[0..%u]\n",
		       __FILE__, __FUNCTION__, __LINE__, bloc->dessin.x_BG, bloc->dessin.x_BG_pc, bloc->dessin.y_BG, bloc->dessin.y_BG_pc,
		       bloc->dessin.largeur, bloc->dessin.larg_pc, bloc->dessin.hauteur, bloc->dessin.haut_pc,
		       fenetre->w_larg - 1, fenetre->w_haut - 1);
	FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      }
#endif
    }

    figure->proprietes &= ~FIGURE_A_REECHELONNER;
  }

  /* On trace chacun des blocs de la figure */
/*   printf("Figure : %u*%u mm = %u * %u pixels, facteur d'chelle %g pixels/mm\n", figure->largeur, figure->hauteur, larg_figure, haut_figure, 1.0/mm_v_pixels ); */
  for ( bloc = figure->premier_bloc ; bloc != NULL ; bloc = bloc->suivant ) {
    FIG_DessinerBloc( fenetre, bloc, 0, 0, larg_figure, haut_figure );
  }

  /* On affiche le bloc actif : tour en vert continu */
  if ( figure->bloc_actif != NULL ) {
    Bloc_Figure *bloc;

    bloc = figure->bloc_actif;

    TRC_Style( BI_CONTINU );

    /* Tour en vert continu */
    TRC_Couleur( VERT );
    TRC_Z_Rectangle( fenetre, 0, 0, larg_figure, haut_figure,
		     bloc->dessin.x_BG, bloc->dessin.y_BG - bloc->dessin.hauteur + 1, bloc->dessin.largeur, bloc->dessin.hauteur );

    /* Poignes pour le dplacement : petits carrs bleus en chaque coin ; 10 pixels de ct */
    TRC_Couleur( BLEU );
    TRC_Z_Rectangle( fenetre, 0, 0, larg_figure, haut_figure,
		     bloc->dessin.x_BG - TRC_TAILLE_POIGNEE, bloc->dessin.y_BG - bloc->dessin.hauteur - TRC_TAILLE_POIGNEE,
		     2 * TRC_TAILLE_POIGNEE, 2 * TRC_TAILLE_POIGNEE );
    TRC_Z_Rectangle( fenetre, 0, 0, larg_figure, haut_figure,
		     bloc->dessin.x_BG - TRC_TAILLE_POIGNEE, bloc->dessin.y_BG - TRC_TAILLE_POIGNEE,
		     2 * TRC_TAILLE_POIGNEE, 2 * TRC_TAILLE_POIGNEE );
    TRC_Z_Rectangle( fenetre, 0, 0, larg_figure, haut_figure,
		     bloc->dessin.x_BG + bloc->dessin.largeur - TRC_TAILLE_POIGNEE, bloc->dessin.y_BG - TRC_TAILLE_POIGNEE,
		     2 * TRC_TAILLE_POIGNEE, 2 * TRC_TAILLE_POIGNEE );
    TRC_Z_Rectangle( fenetre, 0, 0, larg_figure, haut_figure,
		     bloc->dessin.x_BG + bloc->dessin.largeur - TRC_TAILLE_POIGNEE, bloc->dessin.y_BG - bloc->dessin.hauteur - TRC_TAILLE_POIGNEE,
		     2 * TRC_TAILLE_POIGNEE, 2 * TRC_TAILLE_POIGNEE );
  }
}

/* ~-~-~-~-~-~-~-~-~ Gestion souris ~-~-~-~-~-~-~-~-~ */

/* -------- Dplacements du pointeur -------- */

static void RCT_Figure(Fenetre *fenetre, int x, int y, int vx_x, int vx_y)
{
  Figure *figure;
  struct bloc_figure *bloc;
  double mm_v_pixels;		/* Conversion cran, mm en mm/pixel */

  /* Quelle est la figure de la fentre ? */
  figure = (Figure *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == figure ) {
    return;
  }
  if ( NULL == figure->premier_bloc ) {
    return;
  }
  if ( figure->proprietes & ( FIGURE_A_REECHELONNER | FIGURE_A_REDESSINER ) ) {
    /* Inutile de faire quoi que ce soit : il y aura un redessin juste aprs ! */
    return;
  }

  /* Facteur de conversion mm <-> cran, pour garder le rapport largeur / hauteur */
  mm_v_pixels = TRC_Echelle2D( (double) figure->largeur, (double) figure->hauteur, fenetre->w_larg, fenetre->w_haut );
  if ( mm_v_pixels <= 0.0 ) {
    (void) fprintf( stderr, "%s, %s l.%u : Valeur d'chelle absurde [%g]\n",
		    __FILE__, __FUNCTION__, __LINE__, mm_v_pixels );
    return;
  }

  /* On cherche le cadre sous l'ancienne position souris */
  if ( ( vx_x > 0 ) && ( vx_y > 0 ) ) {
    /* Quel est le bloc sous l'ancienne souris ? */
    bloc = RCT_TrouverBloc( figure, vx_x, vx_y, 0 );
    if ( bloc != NULL ) {
      BLOC_TracerReticule( fenetre, bloc, AXE_X | AXE_Y, vx_x, vx_y );
    }
  }

  /* On voit si la nouvelle position demande un pointeur */
  if ( ( x > 0 ) && ( y > 0 ) ) {
    /* Quel est le bloc la souris ? */
    bloc = RCT_TrouverBloc( figure, x, y, 0 );
    if ( bloc != NULL ) {
      BLOC_TracerReticule( fenetre, bloc, AXE_X | AXE_Y, x, y );
    }
  }
}

/* -------- Clic de la souris -------- */

static int CLC_Figure(Fenetre *fenetre, int bouton, int n_clic, int etat, int x, int y)
{
  Figure *figure;
  struct bloc_figure *bloc, *premier_bloc;
  double mm_v_pixels;		/* en mm/pixel */
  int larg_figure, haut_figure;

  /* Quelle est la figure de la fentre ? */
  figure = (Figure *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );
  if ( NULL == figure ) {
    OPTION_A_FAIRE( "Crer une nouvelle figure" );
  }

  /* Facteur de conversion mm <-> cran, pour garder le rapport largeur / hauteur */
  mm_v_pixels = TRC_Echelle2D( (double) figure->largeur, (double) figure->hauteur, fenetre->w_larg, fenetre->w_haut );
  if ( mm_v_pixels <= 0.0 ) {
    (void) fprintf( stderr, "%s, %s l.%u : Valeur d'chelle absurde [%g]\n",
		    __FILE__, __FUNCTION__, __LINE__, mm_v_pixels );
    return BI_EVENEMENT_TRAITE;
  }

  /* Hauteur totale de la figure  l'cran : Y est compris entre fenetre->w_y et fenetre->w_y + haut_figure */
  haut_figure = (int) ( ( (double) figure->hauteur ) / mm_v_pixels );
  larg_figure = (int) ( ( (double) figure->largeur ) / mm_v_pixels );

  if ( 4 == bouton ) {		/* Bouton   droit  */
    CLC_SousMenu( fenetre, figure );
  } else if ( 2 == bouton ) {	/* Bouton du milieu */
    int x_fin, y_fin;

    /* On dfinit une zone  l'cran : nouveau bloc */
    bouton = prepare_zoom( fenetre, x, y, &x_fin, &y_fin );
    if ( 1 == bouton ) { /* On confirme la cration */
      Bloc_Figure *nv_bloc;
      double xBG_mm, yBG_mm, larg_mm, haut_mm;

      /* Contrles : bloc reste dans la figure et non-rduit  un point */
      if ( x_fin > larg_figure ) x_fin = larg_figure;
      if ( x > larg_figure ) x = larg_figure;

      if ( y_fin > haut_figure ) y_fin = haut_figure;
      if ( y > haut_figure ) y = haut_figure;

      if ( ( x_fin == x ) || ( y_fin == y ) ) {
	AVERTISSEMENT( "Cadre de dimension nulle [dX = %d, dY = %d]", abs( x_fin - x ), abs( y_fin - y ) );
	return BI_EVENEMENT_TRAITE;
      }

      /* Conversion des coordonnes en mm */
      if ( x_fin < x ) {
	int tmp_x;
	tmp_x = x; x = x_fin; x_fin = tmp_x;
      }
/*       x     -= fenetre->w_x; */
/*       x_fin -= fenetre->w_x; */

      if ( y_fin < y ) {
	int tmp_y;
	tmp_y = y; y = y_fin; y_fin = tmp_y;
      }
/*       printf("Cration : Y=%d, Y_fin=%d, f->w_y = %d, hauteur = %u\n", y, y_fin, fenetre->w_y, haut_figure ); */
/*       y     -= fenetre->w_y; */
/*       y_fin -= fenetre->w_y; /\* y_fin > y => y_fin correspond  y_BG *\/ */

      xBG_mm  = x * mm_v_pixels;
      yBG_mm  = ( haut_figure - y_fin ) * mm_v_pixels;	/* Attention, axe cran invers par rapport  un axe normal ! */
      larg_mm = ( x_fin - x ) * mm_v_pixels;
      haut_mm = ( y_fin - y ) * mm_v_pixels;

      /* On cre le bloc avec ces dimensions dans la figure, comme premier bloc, avec des graphes */
      nv_bloc = FIG_CreerBloc( figure, NULL, BLOCFIGURE_TYPE_GRAPHES, xBG_mm, yBG_mm, larg_mm, haut_mm );
      if ( nv_bloc != NULL ) {
	FIG_ChangerBlocActif( figure, nv_bloc );
      }
    }
  } else {			/* Bouton gauche */
    unsigned int position;

    /* Position du clic sur le bloc courant */
    position = BLOC_PositionPoint( figure->bloc_actif, x, y, TRC_TAILLE_POIGNEE );

    if ( 2 == n_clic ) {		/* Double-clic = diter le bloc actif */
      if ( NULL == figure->bloc_actif ) {
	return BI_EVENEMENT_TRAITE;
      }

      switch ( figure->bloc_actif->type )
      {
      case BLOCFIGURE_TYPE_GRAPHES:
	I_BLOC_ChangerAxes_Souris( fenetre, figure->bloc_actif, position );
	figure->proprietes |= FIGURE_A_REDESSINER;
	break;

      default:
	BLOC_Modifier( fenetre, figure, figure->bloc_actif );
      }
    } else {
      int x_BG_figure = 0, y_BG_figure, retour;
      boolean traite = FAUX;

      /* Prparation : coordonnes cran du coin BG de la figure */
      y_BG_figure = haut_figure;

      /* 1er essai : les poignes, pour changer la taille  la souris ! */
      retour = I_BLOC_ChangerTaille_Souris( fenetre, &(figure->bloc_actif->dessin), x_BG_figure, y_BG_figure, larg_figure, haut_figure, position );
      if ( retour < 0 ) return VRAI; /* Erreur ou annulation */
      if ( 0 == retour ) {
	figure->proprietes |= FIGURE_A_REECHELONNER | FIGURE_A_REDESSINER; /* Il faut recalculer pour le bloc les dimensions cran ! */
	traite = VRAI;
      }

      /* 2e essai : le dplacement du bloc, si shift est enfonce */
      if ( ( FAUX == traite ) && ( ( etat & K_LSHIFT ) || ( etat & K_RSHIFT ) ) && ( position != TRC_EST_HORS_REGION ) ) {
	retour = I_BLOC_Deplacer_Souris( fenetre, &(figure->bloc_actif->dessin), x_BG_figure, y_BG_figure, larg_figure, haut_figure,
					 x - figure->bloc_actif->dessin.x_BG, figure->bloc_actif->dessin.y_BG - y );
	if ( retour < 0 ) return VRAI; /* Erreur ou annulation */
	if ( 0 == retour ) {
	  figure->proprietes |= FIGURE_A_REECHELONNER | FIGURE_A_REDESSINER; /* Il faut recalculer pour le bloc les dimensions cran ! */
	  traite = VRAI;
	}
      }

      /* 3e essai : zooms dans le bloc, si le bloc s'y prte */
      if ( ( FAUX == traite ) && ( position != TRC_EST_HORS_REGION ) ) {
	OPTION_A_FAIRE( "Essayer de faire un zoom dans le bloc " );
	if ( retour < 0 ) return VRAI; /* Erreur ou annulation */
	if ( 0 == retour ) {
	  figure->proprietes |= FIGURE_A_REDESSINER;
	  traite = VRAI;
	}
      }

      /* 4e essai : changer le bloc actif */
      if ( FAUX == traite ) {
	/* Quel est le bloc sous l'ancienne souris ? */
	premier_bloc = RCT_TrouverBloc( figure, x, y, 0 );
	if ( premier_bloc != NULL ) {
	  unsigned int niveau_max = 0;
	  int niveau_actuel = -1;

	  if ( premier_bloc == figure->bloc_actif ) niveau_actuel = 0;

	  /* Il y a bien un ou des blocs sous la souris
	     ==> on cherche sur combien de niveaux */
	  do {
	    niveau_max++;

	    bloc = RCT_TrouverBloc( figure, x, y, niveau_max );
	    if ( bloc == figure->bloc_actif ) {
	      niveau_actuel = (int) niveau_max;
	    }
	  } while ( bloc != NULL );

	  if ( niveau_actuel < 0 ) {
	    /* Point cliqu hors du cadre actif
	       ==> on active le cadre de dessus (niveau 0) */
	    bloc = premier_bloc;
	  } else if ( niveau_actuel < (int) (niveau_max - 1) ) {
	    /* Point cliqu dans le cadre de niveau k, il y a au moins un cadre encore dessous
	       ==> on active le cadre en-dessous (niveau k + 1) */
	    bloc = RCT_TrouverBloc( figure, x, y, niveau_actuel + 1 );
	  } else {
	    /* Point cliqu dans le cadre le plus au fond
	       ==> on active le cadre de dessus (niveau 0) */
	    bloc = premier_bloc;
	  }

	  if ( bloc != figure->bloc_actif ) {
	    FIG_ChangerBlocActif( figure, bloc );
	  }
	}
      }
    }
  }

  /* Doit-on redessiner ? */
  if ( figure->proprietes & FIGURE_A_REDESSINER ) {
    FEN_redessin( fenetre, BI_RDS_UTIL );
    figure->proprietes &= ~FIGURE_A_REDESSINER; 
  }

  return BI_EVENEMENT_TRAITE;
}

/* -------- Le menu contextuel -------- */

static void CLC_SousMenu(Fenetre *fenetre, Figure *figure)
{
  int option;

  if ( NULL == figure ) return;

  /* Appel du menu contextuel */
  option = SS_menu_texte( txt_ssmenu_figures, NOMBRE_OPTIONS_FIGURES, titre_ssmenu_figures, -1, 0);

  /* En fonction de l'option choisie... */
  switch ( option )
  {
  case OPT_FIGURE_NOUVEAU_CADRE: /* Crer un nouveau cadre : graphes par dfaut, align sur les marges */
    {
      Bloc_Figure *nouveau;

      nouveau = FIG_CreerBloc( figure, NULL, BLOCFIGURE_TYPE_GRAPHES, figure->marge_gauche, figure->marge_basse,
			       figure->largeur - (figure->marge_gauche + figure->marge_basse), figure->hauteur - (figure->marge_basse + figure->marge_haute) );
      BLOC_Modifier( fenetre, figure, nouveau );
    }
    break;

  case OPT_FIGURE_EFFACER_CADRE: /* Effacer le cadre actif */
    {
      int retour;
    
      retour = FIG_EffacerBloc( figure, figure->bloc_actif );
      if ( 0 == retour ) {	/* Effacement ralis -> on doit redessiner la figure ! */
	figure->proprietes |= FIGURE_A_REDESSINER;
#ifdef _DEVELOPPE_
      } else {
	(void) fprintf(stderr, "%s, %s l.%u : ERREUR en voulant effacer le bloc actif [code %d]\n",
		       __FILE__, __FUNCTION__, __LINE__, retour);
#endif
      }
    }
    break;

  case OPT_FIGURE_COPIER_CADRE:	/* Copier le cadre actif */
    OPTION_A_FAIRE( "Copier le bloc actif" );
    break;

  case OPT_FIGURE_MODIFIER_CADRE: /* Modifier le cadre actif */
    (void) BLOC_Modifier( fenetre, figure, figure->bloc_actif );
    break;

  case OPT_PREMIER_PLAN:
    {
      Bloc_Figure *avant, *dernier;

      /* Dernier bloc de la liste */
      dernier = figure->premier_bloc;
      while ( dernier->suivant != NULL ) {
	dernier = dernier->suivant;
      }

      if ( dernier == figure->bloc_actif ) {
	return;
      }

      /* Bloc juste avant le bloc actif */
      avant = figure->premier_bloc;
      if ( figure->bloc_actif != avant ) {
	while ( avant->suivant != figure->bloc_actif ) {
	  avant = avant->suivant;
	  if ( NULL == avant ) {
	    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le chanage des blocs, bloc actif hors chanage\n",
			   __FILE__, __FUNCTION__, __LINE__);
	    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	    return;
	  }
	}

	/* On a premier->[...]->avant->actif->apres->[...]->dernier, on veut premier->[...]->avant->apres->[...]->dernier->actif */
	avant->suivant = figure->bloc_actif->suivant;
      } else {
	/* On a actif->apres->[...]->dernier, on veut apres->[...]->dernier->actif */
	figure->premier_bloc = figure->premier_bloc->suivant;
      }
      figure->bloc_actif->suivant = NULL;
      dernier->suivant = figure->bloc_actif;

      /* On doit redessiner l'image */
      figure->proprietes |= FIGURE_A_REDESSINER;
    }
    break;

  case OPT_DERNIER_PLAN:
    if ( figure->bloc_actif != figure->premier_bloc ) {
      Bloc_Figure *avant;

      avant = figure->premier_bloc;
      while ( avant->suivant != figure->bloc_actif ) {
	avant = avant->suivant;
	if ( NULL == avant ) {
	  (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans le chanage des blocs, bloc actif hors chanage\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
	  return;
	}
      }

      /* Ici : premier->[...]->avant->actif->apres->[...] ; on veut actif->premier->[...]->avant->apres->[...] */
      avant->suivant = figure->bloc_actif->suivant;
      figure->bloc_actif->suivant = figure->premier_bloc;
      figure->premier_bloc = figure->bloc_actif;

      /* On doit redessiner l'image */
      figure->proprietes |= FIGURE_A_REDESSINER;
    }
    break;

  case OPT_EXPORTER_IMAGE:
    (void) EXP_PS_Figure( figure );
    break;
  }

  /* Faut-il redessiner ? */
  if ( figure->proprietes & FIGURE_A_REDESSINER ) {
    FEN_redessin( fenetre, BI_RDS_UTIL );
    figure->proprietes &= ~FIGURE_A_REDESSINER; 
  }
}

/* -------- Trouver le bloc  la position indique [= contenant le point indiqu] --------
    Renvoie NULL si aucun bloc ne contient le point indiqu

  figure : la figure  chercher
  x, y   : coordonnes du point ( l'cran)
  niveau : en cas de cadres superposs, lequel doit-on renvoyer ? [0 : premier, 1 : 2e,...]
 */

static struct bloc_figure *RCT_TrouverBloc(Figure *figure, double x, double y, unsigned int niveau)
{
  struct bloc_figure *bloc, *bloc_trouve = NULL;
  unsigned int i = 0;

  if ( NULL == figure ) return NULL;

  /* Coefficients de conversion */

  for ( bloc = figure->premier_bloc ; bloc != NULL ; bloc = bloc->suivant ) {
    /* Contrles en X */
    if ( ( bloc->dessin.x_BG > x ) || ( bloc->dessin.x_BG + bloc->dessin.largeur < x ) ) continue; /* En-dehors */

    /* Contrles en Y */
    if ( ( bloc->dessin.y_BG < y ) || ( bloc->dessin.y_BG - bloc->dessin.hauteur > y ) ) continue; /* En-dehors */

    /* Si on est l : on est dans le bloc */
    bloc_trouve = bloc;
    if ( i == niveau ) return bloc_trouve;

    /* On passe au niveau suivant */
    i++;
  }

  /* On est l : soit on n'a rien trouv du tout, soit on n'a rien trouv pour le niveau indiqu */
  return NULL;
}

/* ~-~-~-~-~-~-~-~-~ Gestion des messages ~-~-~-~-~-~-~-~-~ */

static int MSG_Figure(Fenetre *fenetre, Message *message)
{
  Figure *figure;

  figure = (Figure *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_1 );

  switch ( message->type )
  {
  case WM_SIZED:
    if ( figure != NULL ) {
      figure->proprietes |= FIGURE_A_REECHELONNER;
    }
    break;			/* On garde le traitement classique mis  part a */

  case WM_CLOSED:
    /* On dissocie la figure de la fentre */
    FEN_StockerComplement( fenetre, FEN_COMPLEMENT_1,NULL );

    /* On efface la figure associe  la fentre */
    FIG_EffacerFigure( figure ); figure = NULL;

    /* On ferme la fentre */
    FEN_Detruire( fenetre );

    return BI_EVENEMENT_FEN_FERMEE;

  default:
    (void) fprintf(stderr, "%s, %s l.%u : Message de type inconnu [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, message->type );
    break;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* ~-~-~-~-~-~-~-~-~ Modification des paramtres ~-~-~-~-~-~-~-~-~ */

#define LGMAX_TEXTE_TYPE_CADRE 22
#define LGMAX_TEXTE_SOUS_TYPE  13
#define LGMAX_TEXTE_NOM_GRAPHE 40

struct infos_arbre {
  Bloc_Figure *bloc;
  unsigned long p_affiche;	/* Dans le dfileur, 1er affich */
  unsigned long n_choisi;	/* Dans le dfileur, slectionn */
};

/* -------- Appel de l'arbre -------- */

static int BLOC_Modifier(Fenetre *mere, Figure *figure, Bloc_Figure *bloc)
{
  int reponse;
  struct infos_arbre infos = { NULL, 0UL, 0UL };

  if ( NULL == bloc ) return -1;

  /* On associe les informations  l'arbre */
  infos.bloc = bloc;

  /* On met les paramtres dans la bote de dialogue */
  BLOC_PreparerArbre( figure, bloc );

  /* On appelle la bote de dialogue */
  ARB_MettreComplement( PARAM_RECT_TRACE, &infos );
  reponse = ARB_Bloquant( PARAM_RECT_TRACE, titre_modifier_bloc, mere, PARAM_RECT_TRACE );
  ARB_MettreComplement( PARAM_RECT_TRACE, NULL );

  /* On dissocie les infos de l'arbre */
  infos.bloc = NULL;

  /* On annule ou on continue ? */
  if ( PRT_ANNULER == reponse ) return 0;

  /* On met  jour la taille, si ncessaire */
  bloc->dessin.x_BG_pc = OBJ_LireNombre( PARAM_RECT_TRACE, PRT_X      , NULL ) * 100.0 / (double) figure->largeur;
  bloc->dessin.y_BG_pc = OBJ_LireNombre( PARAM_RECT_TRACE, PRT_Y      , NULL ) * 100.0 / (double) figure->hauteur;
  bloc->dessin.larg_pc = OBJ_LireNombre( PARAM_RECT_TRACE, PRT_LARGEUR, NULL ) * 100.0 / (double) figure->largeur;
  bloc->dessin.haut_pc = OBJ_LireNombre( PARAM_RECT_TRACE, PRT_HAUTEUR, NULL ) * 100.0 / (double) figure->hauteur;

  /* Et on contrle le tout, pour ne pas dborder de la page ! */
  if ( bloc->dessin.x_BG_pc < 0.0 ) bloc->dessin.x_BG_pc = 0.0;
  if ( bloc->dessin.y_BG_pc < 0.0 ) bloc->dessin.y_BG_pc = 0.0;
  if ( bloc->dessin.larg_pc > 100.0 ) bloc->dessin.larg_pc = 100.0;
  if ( bloc->dessin.haut_pc > 100.0 ) bloc->dessin.haut_pc = 100.0;

  /* Il faut reconstruire et redessiner la figure */
  figure->proprietes |= FIGURE_A_REECHELONNER | FIGURE_A_REDESSINER;

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Gestion de l'arbre -------- */

static int BLOC_GererArbre(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  struct infos_arbre *infos;

  /* Rcupration des infos */
  infos = (struct infos_arbre *) ARB_LireComplement( PARAM_RECT_TRACE );
  if ( NULL == infos ) return BI_EVENEMENT_NON_TRAITE;

  switch ( objet )
  {
  case PRT_CHOIX_TYPE:		/* Changer le type du cadre */
    OBJ_choisit( PARAM_RECT_TRACE, PRT_CHOIX_TYPE, FORCE_OUI );
    OPTION_A_FAIRE( "Changer le type du cadre" );
    OBJ_choisit( PARAM_RECT_TRACE, PRT_CHOIX_TYPE, FORCE_NON );
    return BI_EVENEMENT_TRAITE;

  case PRT_AJOUTER:
    OBJ_choisit( PARAM_RECT_TRACE, PRT_AJOUTER, FORCE_OUI );
    if ( BLOCFIGURE_TYPE_GRAPHES == infos->bloc->type ) {
      Graphe **graphes;
      unsigned long nombre_graphes;

      /* On choisit le(s) graphe(s)  ajouter */
      graphes = GRF_mchoisir( GR_TOUS, 1, 0, &nombre_graphes );

      /* On les ajoute */
      BLOC_AjouterGraphes( infos->bloc, graphes, nombre_graphes );

      /* Plus besoin de la liste des graphes */
      free( graphes ); graphes = NULL;

      /* On redessine la liste des graphes */
      ARB_PreparerDefileur_O( PARAM_RECT_TRACE, PRT_PREM_GRAPHE, PRT_DERN_GRAPHE, LGMAX_TEXTE_NOM_GRAPHE,
			      PRT_LST_AVANT, PRT_FOND_ASC, PRT_BOUTON_ASC, PRT_LST_APRES,
			      infos->bloc->donnees.graphes.nbr_graphes, NULL, infos->n_choisi, infos->p_affiche,
			      BLOC_TexteGraphe );
    }
    OBJ_choisit( PARAM_RECT_TRACE, PRT_AJOUTER, FORCE_NON );
    return BI_EVENEMENT_TRAITE;
  }

  /* Objet non gr : gestion par dfaut */
  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Prparation de l'arbre -------- */

static void BLOC_PreparerArbre(Figure *figure, Bloc_Figure *bloc)
{
  if ( FAUX == arbres[ PARAM_RECT_TRACE ].drapeaux.prepare ) {
    int retour;

    /* Objets utilisateurs */
    OBJ_ChangerType( PARAM_RECT_TRACE, PRT_CH_STYLE, G_STYLELIGNE );

    /* Initialisation de l'arbre */
    OPTION_A_FAIRE( "Documentation, options de mise en forme des rsultats : menu_affichage.html#mod_cadre" );
    retour = ARB_Initialiser( PARAM_RECT_TRACE, "menu_affichage.html#mod_cadre", BLOC_GererArbre, ( int (*)() ) NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s, %s l.%u : ERREUR en voulant initialiser l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }

    /* Positions plus correctes des objets */
    OBJ_deplace( PARAM_RECT_TRACE, PRT_Y      , 0, 1, DEPLACE_PIXELS );
    OBJ_deplace( PARAM_RECT_TRACE, PRT_LARGEUR, 0, 2, DEPLACE_PIXELS );
    OBJ_deplace( PARAM_RECT_TRACE, PRT_HAUTEUR, 0, 3, DEPLACE_PIXELS );

    OBJ_deplace( PARAM_RECT_TRACE, PRT_CH_STYLE, 0, 1, DEPLACE_PIXELS );
    OBJ_deplace( PARAM_RECT_TRACE, PRT_CHX_MOTIF, 0, 2, DEPLACE_PIXELS );
    OBJ_deplace( PARAM_RECT_TRACE, PRT_CHX_FONTE, 0, 3, DEPLACE_PIXELS );

    OBJ_deplace( PARAM_RECT_TRACE, PRT_EFFACER   , 0, 1, DEPLACE_PIXELS );
    OBJ_deplace( PARAM_RECT_TRACE, PRT_TT_EFFACER, 0, 1, DEPLACE_PIXELS );

    (void) ARB_EnjoliverDefileur( PARAM_RECT_TRACE, PRT_FOND_LST_GRF, PRT_PREM_GRAPHE, PRT_DERN_GRAPHE,
				  PRT_LST_AVANT, PRT_FOND_ASC, PRT_BOUTON_ASC, PRT_LST_APRES, ARB_ASCENSEUR_GAUCHE );
    /* Traduction ventuelle */
#ifdef AVEC_TRADUCTION_
#endif

    arbres[ PARAM_RECT_TRACE ].drapeaux.prepare = VRAI;
  }

  /* On place les paramtres */
  OBJ_MettreNombre( PARAM_RECT_TRACE, PRT_X      , bloc->dessin.x_BG_pc * (double) figure->largeur / 100.0, 0.0 );
  OBJ_MettreNombre( PARAM_RECT_TRACE, PRT_Y      , bloc->dessin.y_BG_pc * (double) figure->hauteur / 100.0, 0.0 );
  OBJ_MettreNombre( PARAM_RECT_TRACE, PRT_LARGEUR, bloc->dessin.larg_pc * (double) figure->largeur / 100.0, 0.0 );
  OBJ_MettreNombre( PARAM_RECT_TRACE, PRT_HAUTEUR, bloc->dessin.haut_pc * (double) figure->hauteur / 100.0, 0.0 );

  /* En particulier ceux dpendants du type de cadre */
  (void) BLOC_InstallerType( bloc );
}

/* -------- Initialiser en fonction du type de cadre -------- */

static int BLOC_InstallerType(Bloc_Figure *bloc)
{
  if ( bloc->type >= NOMBRE_TYPES_BLOC_FIGURE ) {
    (void) fprintf(stderr, "%s, %s l.%u : Bloc de type inconnu [code %u]\n",
		   __FILE__, __FUNCTION__, __LINE__, bloc->type );
    return +1;
  }

  /* On met le type dans le bouton */
  OBJ_RemplacerTexte( PARAM_RECT_TRACE, PRT_CHOIX_TYPE, txt_types_blocs[ bloc->type ], VRAI );

  /* Prparation gnrique */
  OBJ_active( PARAM_RECT_TRACE, PRT_EFFACER, FORCE_NON );
  OBJ_active( PARAM_RECT_TRACE, PRT_TT_AJOUTER, FORCE_NON );
  OBJ_active( PARAM_RECT_TRACE, PRT_TT_EFFACER, FORCE_NON );

  /* En fonction du type, initialisation complmentaire */
  switch ( bloc->type )
  {
  case BLOCFIGURE_TYPE_GRAPHES:
    OBJ_RemplacerTexte( PARAM_RECT_TRACE, PRT_AJOUTER, txt_ajouter, VRAI );
    OBJ_active( PARAM_RECT_TRACE, PRT_AJOUTER, FORCE_OUI );

    ARB_PreparerDefileur_O( PARAM_RECT_TRACE, PRT_PREM_GRAPHE, PRT_DERN_GRAPHE, LGMAX_TEXTE_NOM_GRAPHE,
			    PRT_LST_AVANT, PRT_FOND_ASC, PRT_BOUTON_ASC, PRT_LST_APRES,
			    bloc->donnees.graphes.nbr_graphes, NULL, 0, 0,
			    BLOC_TexteGraphe );
    break;

  case BLOCFIGURE_TYPE_TEXTE:
    OBJ_active( PARAM_RECT_TRACE, PRT_AJOUTER, FORCE_NON );
    break;

  case BLOCFIGURE_TYPE_LEGENDE:
    OBJ_RemplacerTexte( PARAM_RECT_TRACE, PRT_AJOUTER, txt_choisir, VRAI );
    OBJ_active( PARAM_RECT_TRACE, PRT_AJOUTER, FORCE_OUI );
    break;

  case BLOCFIGURE_TYPE_IMAGE:
    OBJ_RemplacerTexte( PARAM_RECT_TRACE, PRT_AJOUTER, txt_choisir, VRAI );
    OBJ_active( PARAM_RECT_TRACE, PRT_AJOUTER, FORCE_OUI );
    break;

  case BLOCFIGURE_TYPE_SURFACE:
    OBJ_active( PARAM_RECT_TRACE, PRT_AJOUTER, FORCE_NON );
    break;
  }

  /* Termin correctement */
  return 0;
}

/* -------- Textes pour le dfileur -------- */

static const char *BLOC_TexteGraphe(unsigned int arbre, unsigned long numero)
{
  struct infos_arbre *infos;

  /* Rcupration des infos */
  infos = (struct infos_arbre *) ARB_LireComplement( arbre );
  if ( NULL == infos ) return NULL;

  if ( BLOCFIGURE_TYPE_GRAPHES == infos->bloc->type ) {
    Graphe *graphe;

    if ( numero >= infos->bloc->donnees.graphes.nbr_graphes ) return NULL;

    /* Quel graphe est concern ? */
    graphe = infos->bloc->donnees.graphes.infos_graphe[ numero ].graphes;
    if ( graphe != NULL ) return graphe->nom;

    return NULL;
  }

  /* Autres types : rien de spcial  afficher ! */
  return NULL;
}
