/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) Emmanuel CURIS 1996 - 2008

   Interface pour la gestion de base des images (copie, effacement,...)
*/

#include <stdio.h>
#include <stdlib.h>

#include "exafs.h"
#include "systeme.h"
#include "txt_utils.h"

#include "echantillon.h"	/* Pour la conversion -> image de composition */
#include "images.h"
#include "interface.h"
#include "statistiques.h"

#include "int_images.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int I_Deplacement(Image *image, Serie_Images *serie_arrivee, Image *avant);

/* -------- Effacer une image, avec contrôles et confirmations -------- */

int I_IMG_Effacer(Image *image, boolean demander)
{
  int retour;
  boolean effacer_graphes = VRAI;

  if ( NULL == image ) return 0;

  if ( VRAI == demander ) {
    retour = BI_ERR_alerte( fen_images, &Q_EffacerImage, VRAI, image->nom );
    if ( 2 == retour ) {
      return +1; /* Annulation, donc erreur mineure */
    }

    if ( IMG_GENRE_GRAPHE == image->genre ) {
      retour = BI_ERR_alerte( fen_images, &Q_EffacerGraphes, VRAI, image->nom );
      if ( 3 == retour )
	return +2; /* Annulation tardive */
      else if ( 1 == retour )
	effacer_graphes = VRAI;
      else
	effacer_graphes = FAUX;
    }
  }

  /* On détermine la nouvelle image courante */
  if ( image == image_courante ) {
    Image *nv_courante;

    /* Premier essai : la suivante */
    nv_courante = image->suivante;
    if ( NULL == nv_courante ) {
      /* 2e essai : la précédente */
      nv_courante = IMG_serie_courante->premiere_image;
      if ( ( nv_courante != NULL ) && ( nv_courante != image ) ) {
	while ( nv_courante->suivante != image ) {
	  nv_courante = nv_courante->suivante;
	}
      }
    }

    /* On va à cette nouvelle image courante */
    IMG_AllerImage( nv_courante );
  }

  /* On fait l'effacement effectif */
  retour = IMG_EffacerImage( image, effacer_graphes );
  if ( retour != 0 ) {
    OPTION_A_FAIRE( "Avertir : problème à l'effacement !" );
  }

  /* On met à jour l'interface, si nécessaire */

  /* Terminé */
  return 0;
}

/* -------- Créer une image qui contienne la copie des incertitudes de l'image indiquée -------- */

Image *I_IMG_SortirErreurs(Image *image, unsigned int axe)
{
  Image *erreurs = NULL;
  char *titre;

  if ( NULL == image ) return NULL;

  if ( axe != AXE_Z ) {
    /* Seul l'axe Z peut avoir des incertitudes, actuellement */
    (void) BI_ERR_alerte( fen_images, &ERR_AxeSansErreurs, VRAI, ( AXE_X == axe ) ? ('X') : ('Y') );
    return NULL;
  }

  /* Y a-t-il effectivement des écarts-types indiqués ? */
  if ( NULL == image->erreurs ) {
    (void) BI_ERR_alerte( fen_images, &ERR_AxeSansErreurs, VRAI, 'Z' );
    return NULL;
  }


  /* On sort une copie de l'image des erreurs, placée dans la série juste après l'image traitée */
  erreurs = IMG_CopierImage( image->erreurs, NULL, VRAI );
  if ( NULL == erreurs ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur en voulant copier l'image des erreurs\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }
  IMG_InsererDansSerie( erreurs, NULL, image );

  /* On donne le titre */
  titre = TXT_genere_chaine( titre_image_des_erreurs, image->nom );
  IMG_Titre( erreurs, titre );
  free( titre ); titre = NULL;

  /* Les légendes et unités des axes en X et Y sont inchangées */
  IMG_ChangerLegende( erreurs, AXE_X, IMG_QuelleLegende( image, AXE_X ) );
  IMG_ChangerUnite  ( erreurs, AXE_X, IMG_QuelleUnite  ( image, AXE_X ) );

  IMG_ChangerLegende( erreurs, AXE_Y, IMG_QuelleLegende( image, AXE_Y ) );
  IMG_ChangerUnite  ( erreurs, AXE_Y, IMG_QuelleUnite  ( image, AXE_Y ) );

  /* On adapte la légende pour Z, unité inchangée */
  titre = TXT_genere_chaine( "Erreurs sur %s", IMG_QuelleLegende( image, AXE_Z ) );
  IMG_ChangerLegende( erreurs, AXE_Z, titre );
  free( titre ); titre = NULL;

  IMG_ChangerUnite( erreurs, AXE_Z, IMG_QuelleUnite( image, AXE_Z ) );

  /* Terminé */
  return erreurs;
}

/* -------- Copier une image, éventuellement dans une nouvelle série -------- */

int I_IMG_Copier(struct image *image, boolean multiple)
{
  Image **images = NULL;
  Serie_Images *serie_arrivee;
  unsigned long nbr_images = 0;

  if ( VRAI == multiple ) {
    /* Choix des images à copier */
    if ( NULL == image ) {
      images = I_IMG_ChoisirMultiples( GR_TOUS, 1, 0, &nbr_images );
    } else {
      images = I_IMG_ChoisirMultiples( image->type, 1, 0, &nbr_images );
    }
    if ( NULL == images ) return -1;
  } else if ( NULL == image ) {
    return -20;
  }

  /* Choix de la série d'arrivée */
  serie_arrivee = I_SERIE_IMG_Choisir( -1, ROOT, 0 );
  if ( NULL == serie_arrivee ) {
    free( images ); images = NULL;
    return -2;
  }

  if ( VRAI == serie_arrivee->drapeaux.protegee ) {
    (void) BI_ERR_alerte( fen_images, &ERR_SerieImageProtegee, VRAI );
    free( images ); images = NULL;
    return -3;
  }

  if ( VRAI == multiple ) {
    unsigned long i;

    for ( i = 0 ; i < nbr_images ; i++ ) {
      IMG_CopierImage( images[ i ], serie_arrivee, FAUX );
    }
  } else {
    IMG_CopierImage( image, serie_arrivee, FAUX );
  }

  /* On a fini... */
  free( images ); images = NULL;
  return 0;
}

/* -------- Déplacer un image dans une autre série -------- */

int I_IMG_Deplacer(struct image *image, boolean multiple)
{
  Image **images = NULL;
  Serie_Images *serie_arrivee;
  unsigned long nbr_images = 0;

  if ( VRAI == multiple ) {
    /* Choix des images à copier */
    if ( NULL == image ) {
      images = I_IMG_ChoisirMultiples( GR_TOUS, 1, 0, &nbr_images );
    } else {
      images = I_IMG_ChoisirMultiples( image->type, 1, 0, &nbr_images );
    }
    if ( NULL == images ) return -1;
  } else if ( NULL == image ) {
    return -20;
  }

  /* Choix de la série d'arrivée */
  serie_arrivee = I_SERIE_IMG_Choisir( -1, ROOT, 0 );
  if (NULL == serie_arrivee) {
    free( images ); images = NULL;
    return -2;
  }

  if ( VRAI == serie_arrivee->drapeaux.protegee ) {
    (void) BI_ERR_alerte( fen_images, &ERR_SerieImageProtegee, VRAI );
    free( images ); images = NULL;
    return -3;
  }

  if ( serie_arrivee == IMG_serie_courante ) {
    (void) BI_ERR_alerte( fen_images, &MSG_MemeSerieImage, VRAI );
    return 0;
  }

  /* Déplacement */
  if ( VRAI == multiple ) {
    unsigned long i;

    for ( i = 0 ; i < nbr_images ; i++ ) {
      I_Deplacement( images[ i ], serie_arrivee, (i > 0) ? (images[ i - 1 ]) : (NULL) );
    }
  } else {
    I_Deplacement( image, serie_arrivee, NULL );
  }

  /* On a fini... */
  free( images ); images = NULL;
  return 0;
}

/* -------- Isoler un image, en créant une nouvelle série -------- */

int I_IMG_Isoler(struct image *image, boolean multiple)
{
  char *titre;
  Image **images = NULL;
  Serie_Images *serie_arrivee;
  unsigned long nbr_images = 1;

  if ( VRAI == multiple ) {
    /* Choix des images à copier */
     if ( NULL == image ) {
      images = I_IMG_ChoisirMultiples( GR_TOUS, 1, 0, &nbr_images );
    } else {
      images = I_IMG_ChoisirMultiples( image->type, 1, 0, &nbr_images );
    }
    if ( NULL == images ) return -1;

    if ( nbr_images == IMG_serie_courante->nbr_images ) {
      (void) BI_ERR_alerte( principale, &MSG_ToutesImagesChoisies, VRAI );
      free( images ); images = NULL;
      return -19;
    }
  } else if ( NULL == image ) {
    return -20;
  }

  /* Création de la série d'arrivée */
  titre = BI_Demande_Texte( principale, &Q_NomNouvelleSerieImage, IMG_serie_courante->nom, (unsigned int) (nbr_images) );
  serie_arrivee = IMG_SERIE_CreerSerie( IMG_serie_courante, titre );
  free( titre ); titre = NULL;
  if (NULL == serie_arrivee) {
    (void) fprintf(stderr, "%s,%s l.%u : impossible de créer une nouvelle série\n",
		   __FILE__, __FUNCTION__, __LINE__);
    free( images ); images = NULL;
    return -21;
  }

  if ( VRAI == multiple ) {
    unsigned long i;

    for ( i = 0 ; i < nbr_images ; i++ ) {
      I_Deplacement( images[ i ], serie_arrivee, (i > 0) ? (images[ i - 1 ]) : (NULL) );
    }
  } else {
    I_Deplacement( image, serie_arrivee, NULL );
  }

  /* On a fini... */
  free( images ); images = NULL;
  return 0;
}

/* -------- Déplacement d'une seule image avec contrôle -------- */

static int I_Deplacement(Image *image, Serie_Images *serie_arrivee, Image *avant)
{
  IMG_ChangerSerie( image, serie_arrivee, avant );
  return 0;
}

/* ~-~-~-~-~-~-~-~ Interfaces pour les informations annexes d'une image ~-~-~-~-~-~-~-~ */

int I_IMG_ChangerGenre(struct image *image)
{
  int reponse, retour = -1975;
  int nv_genre;

  if ( NULL == image ) return -1;

  /* Message d'avertissement */
  reponse = BI_ERR_alerte( NULL, &MSG_ChangerGenreImage, VRAI );
  if ( reponse != 1 ) return +1;	/* Annulation : erreur mineure */

  /* On demande le nouveau genre */
  nv_genre = OBJ_ChoisirGenreImage( CHOIX_IMAGE, LI_GENRE_IMAGE, image->genre, ECHANGE );
  if ( nv_genre == image->genre ) {
    /* Pas de changement en fin de compte */
    return 0;
  }

  /* En fonction du type initial... */
  switch ( image->genre )
  {
  case IMG_GENRE_REELLE:
    switch ( nv_genre )
    {
    case IMG_GENRE_COMPLEXE:
      /* Conversion directe possible */
      retour = IMG_ChangerGenre( image, nv_genre, NULL, NULL );
      break;

    case IMG_GENRE_GRAPHE:
    case IMG_GENRE_DIAGRAMME:
    case IMG_GENRE_PHOTO:
      /* Conversions interdites */
      (void) BI_ERR_alerte( NULL, &ERR_ConversionGenreImpossible, VRAI );
      retour = -3;
      break;

    case IMG_GENRE_COMPO:
      {
	struct composant *matrice = NULL, *compose = NULL;

	/* Il faut choisir les deux compositions qui serviront en CL */
	/* 1. Composition du côté 0 */
	matrice = ChoisirComposant( fen_images );

	/* 2. Composition du côté 1 */
	compose = ChoisirComposant( fen_images );

	/* 3. On peut réaliser la conversion */
	retour = IMG_ChangerGenre( image, IMG_GENRE_COMPO, matrice, compose );
      }
      break;

    case IMG_GENRE_ERREUR:
      OPTION_A_FAIRE( "Conversion Réelle -> erreurs" );
      break;

    case IMG_GENRE_CONTINGENCE:
      OPTION_A_FAIRE( "Conversion Réelle -> table de contigence" );
      break;
    }
    break;

  case IMG_GENRE_COMPLEXE:
    OPTION_A_FAIRE( "Conversion Complexe -> ??" );
    break;

  case IMG_GENRE_GRAPHE:
    OPTION_A_FAIRE( "Conversion Graphe -> ??" );
    break;

  case IMG_GENRE_DIAGRAMME:
    OPTION_A_FAIRE( "Conversion Diagramme -> ??" );
    break;

  case IMG_GENRE_PHOTO:
    OPTION_A_FAIRE( "Conversion Photo -> ??" );
    break;

  case IMG_GENRE_COMPO:
    OPTION_A_FAIRE( "Conversion Composition -> ??" );
    break;

  case IMG_GENRE_ERREUR:
    OPTION_A_FAIRE( "Conversion Erreur -> ??" );
    break;

  case IMG_GENRE_CONTINGENCE:
    OPTION_A_FAIRE( "Conversion Contingence -> ??" );
    break;

  default:
    AVERTISSEMENT( "Genre d'image inconnu [%d]", image->type );
  }

  /* On renvoie le code de retour ainsi obtenu */
  return retour;
}

/* ~-~-~-~-~-~-~-~ Interfaces pour les créations simples d'images ~-~-~-~-~-~-~-~ */

#define I_IMG_DEMANDE_NOM     0	/* Toujours le nom demandé en premier ! */

#define I_IMG_DEMANDE_DERNIERE 1  /* Les réponses sont toujours dans cet ordre, par homogénéité */
#define I_IMG_DEMANDE_ENCORE   2
#define I_IMG_DEMANDE_ANNULE   3

/* -------- Création d'un image contenant une série de valeurs aléatoires -------- */

Image *I_IMG_CreerAleatoire(Image *avant)
{
  Image *nouveau = NULL;
  int reponse;
  struct processus_aleatoire conditions __attribute__((__unused__)) = { PA_BRUIT_BLANC, { { 0 } }, { DP_GAUSSIENNE, 0, { { 0, 0 } } } };
  struct serie_images *serie_cible = NULL;
  unsigned long nombre_valeurs __attribute__((__unused__));

  OPTION_A_FAIRE( "Adapter l'arbre pour une image !" );

  /* On appelle l'arbre considéré */
  reponse = ARB_Bloquant( TIRAGE_ALEATOIRE, "Paramètres des tirages", fenetre_onglets, 0 );
  if ( TA_ANNULER == reponse ) return NULL;

  /* On mémorise les paramètres pour la génération */
  nombre_valeurs = OBJ_LireEntier_U( TIRAGE_ALEATOIRE, TA_NOMBRE );
  conditions.loi.parametres.Gauss.mu    = OBJ_LireNombre( TIRAGE_ALEATOIRE, TA_PARAMETRE_1, NULL );
  conditions.loi.parametres.Gauss.sigma = OBJ_LireNombre( TIRAGE_ALEATOIRE, TA_PARAMETRE_2, NULL );

  /* On génère les valeurs */
/*   nouveau = PA_GenererStationnaire( &conditions, nombre_valeurs ); */

  /* On insère le résultat après le image indiqué */
  serie_cible = I_IMG_InsererDansSerie( IMG_serie_courante, avant, nouveau );
  if ( NULL == serie_cible ) {
    (void) fprintf(stderr, "%s, %s l.%u : Impossible d'insérer le image dans une quelconque série !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    IMG_EffacerImage( nouveau, FAUX ); nouveau = NULL;
    return NULL;
  }

  /* On met le nom */
/*   GRF_titre( nouveau,  */

  /* On a terminé, sans encombre */
  return nouveau;
}
