/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2006, Emmanuel CURIS
   Création : mars 2006

   Interface graphique avec l'utilisateur

   Interpolation (graphe, image)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "serie.h"
#include "interface.h"
#include "lissage.h"

/* -------- Variables locales -------- */

static int methode_interpolation = INTERPOLE_SPLINE;

#define LGMAX_NOM_METHODE   (30)
#define LGMAX_NOM_REFERENCE (25)

#include "int_interpoler.txt"

/* -------- Fonctions locales -------- */

static int  I_GRF_GererArbre   (unsigned int numero_objet, int x, int y);
static void I_GRF_PreparerArbre(Graphe *original);

static int  I_INT_LireInformations(Graphe *original, struct infos_lissage *infos);

/* -------- Correspondance numéro/nom pour les méthodes d'interpolation -------- */

const char *INT_NomMethode(unsigned int numero_methode)
{
  if ( numero_methode >= NBR_METHODES_INTERPOLATION ) {
    AVERTISSEMENT( "Numéro de méthode absurde [max : %u, donné : %u]", NBR_METHODES_INTERPOLATION - 1, numero_methode );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return NULL;
  }

  return noms_methodes_interpolation[ numero_methode ];
}

/* ~-~-~-~-~-~-~-~-~ Appel de l'arbre pour interpolation d'un graphe ~-~-~-~-~-~-~-~-~ */

Graphe *I_GRF_Interpolation(Graphe *original)
{
  Graphe *interpole;
  boolean nv_graphe __attribute__((__unused__));
  extern const char *txt_aucun_graphe; /* Dans variables.c */
  int reponse;
  struct infos_lissage infos = { INTERPOLE_LINEAIRE, FAUX, 0.0, 0.0, 0.0, 0, NULL, 0 };
  char *nom;

  if ( FAUX == GRF_EstFonction( original ) ) {
    if ( NULL != original ) {
      (void) BI_ERR_alerte( principale, &ERR_GrapheNonReel, VRAI );
    }
    return NULL;
  }

  if ( 1 == original->nombre_point ) {
    /* Graphe réduit à un point : aucune interpolation n'est possible ! */
    (void) BI_ERR_alerte( principale, &ERR_1Point, VRAI );
    return NULL;
  }

  /* On prépare l'arbre */
  I_GRF_PreparerArbre( original );

  /* Préparation de l'arbre */
  OBJ_MettreEntier_U( LISSAGE, LG_NPTS_AVANT, original->nombre_point );
  OBJ_CopierTexte( LISSAGE, LG_REFERENCE, txt_aucun_graphe, LGMAX_NOM_REFERENCE );

  /* On fait choisir l'utilisateur... */
  ARB_MettreComplement( LISSAGE, NULL ); /* Aucun graphe pour la référence */
  reponse = ARB_Bloquant( LISSAGE, titre_arbre_interpolation, principale, 0 );
  if ( LG_ANNULER == reponse ) {
    ARB_MettreComplement( LISSAGE, NULL );
    return NULL;
  }

  HIST_Titre(1, "Interpolation d'un graphe", "LIS");
  HIST_Texte(1, 0, "Graphe original : %s\n", original->nom);

  /* Analyse des paramètres */
  reponse = I_INT_LireInformations ( original, &infos );
  ARB_MettreComplement( LISSAGE, NULL );
  if ( reponse != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : ERREUR dans la lecture des paramètres [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, reponse);
    return NULL;
  }
  nv_graphe = ( OBJ_choisit(LISSAGE, LG_EFFACER, QUESTION) == FAUX );

  /* Contrôles */
  if ( infos.nbr_points <= original->nombre_point ) {
    reponse = BI_ERR_alerte( principale, &ERR_MoinsDePoints, VRAI, infos.nbr_points, original->nombre_point );
    if ( 2 == reponse ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s, %s l.%u : trop peu de points [%lu / %lu]\n",
		     __FILE__, __FUNCTION__, __LINE__, infos.nbr_points, original->nombre_point );
#endif
      HIST_Erreur( 1, "Calcul annulé", "Moins de points que l'original" );
      return NULL;
    }
  }

  if ( GRF_SERIE == original->genre_donnees ) { /* Série -> a-t-on des X en commun avec la série ? */
    Serie_Graphes *serie;

    serie = GRF_QuelleSerie( original );
    if ( ( serie != NULL ) && ( serie->table_X == original->x ) ) {
      int reponse;

      /* X communs -> l'interpolation se fera-t-elle sur tous ? */
      reponse = BI_ERR_alerte( principale, &Q_InterpolerSerie, VRAI );
      if ( 3 == reponse ) {	/* Annulation */
	return NULL;
      } else if ( 1 == reponse ) { /* Toute la série est traitée */
	reponse = LIS_InterpolerSerie( serie, &infos );
	if ( reponse != 0 ) {
	  (void) fprintf(stderr, "%s, %s l.%u : ERREUR survenue dans l'interpolation\n",
			 __FILE__, __FUNCTION__, __LINE__);
	}

	return NULL;
      }
    }
  }

  /* On appelle la fonction d'interpolation générique pour un graphe */
  interpole = GRF_Interpoler( original, &infos );
  if ( infos.code_erreur != 0 ) {
    OPTION_A_FAIRE( "Afficher le message d'erreur approprié" );
  }

  /* On place le résultat dans la série d'origine, juste après l'original */
  I_GRF_InsererDansSerie( NULL, original, interpole );

  /* On copie les légendes et unités, inchangées */
  GRF_CopierAffichage( interpole, original, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y );

  /* On met le nom */
  nom = TXT_genere_chaine( "%s [interpolé]", original->nom );
  GRF_titre( interpole, nom );
  free( nom ); nom = NULL;

  return interpole;
}

/* -------- Gestion effective de la boîte de dialogue -------- */

static int I_GRF_GererArbre(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  boolean etat;

  switch (objet)
  {
  case LG_NV_NBR_PTS:
    OBJ_choisit(LISSAGE, objet, FORCE_OUI);
    OBJ_active(LISSAGE, LG_NPTS_APRES, FORCE_OUI);
    OBJ_active(LISSAGE, LG_NV_INTERVALLE, FORCE_OUI);
    etat = OBJ_choisit(LISSAGE, LG_NV_INTERVALLE, QUESTION);
    OBJ_active(LISSAGE, LG_XMIN, etat);
    OBJ_active(LISSAGE, LG_XMAX, etat);
    OBJ_active(LISSAGE, LG_PAS, FORCE_NON);
    OBJ_active(LISSAGE, LG_REFERENCE, FORCE_NON);
    return BI_EVENEMENT_TRAITE;

  case LG_NV_INTERVALLE:
    etat = OBJ_choisit(LISSAGE, LG_NV_INTERVALLE, ECHANGE);
    OBJ_active(LISSAGE, LG_XMIN, etat);
    OBJ_active(LISSAGE, LG_XMAX, etat);
    return BI_EVENEMENT_TRAITE;

  case LG_NV_PAS:
    OBJ_choisit(LISSAGE, objet, FORCE_OUI);
    OBJ_active(LISSAGE, LG_NPTS_APRES, FORCE_NON);
    OBJ_active(LISSAGE, LG_NV_INTERVALLE, FORCE_OUI);
    etat = OBJ_choisit(LISSAGE, LG_NV_INTERVALLE, QUESTION);
    OBJ_active(LISSAGE, LG_XMIN, etat);
    OBJ_active(LISSAGE, LG_XMAX, etat);
    OBJ_active(LISSAGE, LG_PAS, FORCE_OUI);
    OBJ_active(LISSAGE, LG_REFERENCE, FORCE_NON);
    return BI_EVENEMENT_TRAITE;

  case LG_NV_REFERENCE:
    if ( NULL == ARB_LireComplement( LISSAGE ) ) {
      ARB_SimulerClic( LISSAGE, LG_REFERENCE, 1 );
      if ( NULL == ARB_LireComplement( LISSAGE ) ) {
	return BI_EVENEMENT_TRAITE;
      }
    }
    OBJ_choisit(LISSAGE, objet, FORCE_OUI);
    OBJ_active(LISSAGE, LG_NPTS_APRES, FORCE_NON);
    OBJ_active(LISSAGE, LG_NV_INTERVALLE, FORCE_NON);
    OBJ_active(LISSAGE, LG_XMIN, FORCE_NON);
    OBJ_active(LISSAGE, LG_XMAX, FORCE_NON);
    OBJ_active(LISSAGE, LG_PAS, FORCE_NON);
    OBJ_active(LISSAGE, LG_REFERENCE, FORCE_OUI);
    return BI_EVENEMENT_TRAITE;

  case LG_REFERENCE:
    {
      Graphe *reference;

      reference = (Graphe *) ARB_LireComplement( LISSAGE );
      reference = GRF_choisir( GR_TOUS, FAUX, VRAI, GRF_QuelleSerie( reference ) );
      if ( reference != NULL ) {
	ARB_MettreComplement( LISSAGE, reference );
	OBJ_CopierTexte( LISSAGE, LG_REFERENCE, reference->nom, LGMAX_NOM_REFERENCE );
      }
    }
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Lire les infos dans l'arbre -------- */

static int I_INT_LireInformations(Graphe *original, struct infos_lissage *infos)
{
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s, %s l.%u : Pas d'adresse pour mémoriser les infos\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  /* Quelle méthode de lissage ? */
  infos->methode = OBJ_LireEntier_U( LISSAGE, LG_METHODE );
  HIST_OptionChoisie(2, "M"e_AIGU"thode  : \n",
		     NBR_METHODES_INTERPOLATION, infos->methode, noms_methodes_interpolation); 

  /* Utilise-t-on une table ou des valeurs régulières ? */
  if ( VRAI == OBJ_choisit( LISSAGE, LG_NV_REFERENCE, QUESTION ) ) {
    struct graphe *reference;

    reference = (struct graphe *) ARB_LireComplement( LISSAGE );
    if ( NULL == reference ) {
      HIST_Erreur( 1, "Calcul annulé", "Pas de référence indiquée" );
      (void) fprintf(stderr, "%s, %s l.%u : Pas de référence choisie\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -4;
    }
    if ( NULL == reference->x ) {
      (void) fprintf(stderr, "%s, %s l.%u : Pas de X dans la référence choisie !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -5;
    }

    infos->nbr_points = reference->nombre_point;
    infos->table_x    = reference->x;
    infos->table_fournie = VRAI;
    HIST_Texte(3, 0, "Recalage sur les X de %s\n", reference->nom);

    /* Plus besoin du reste */
    return 0;
  }

  /* Ici : table des X régulière, fournie par l'utilisateur */
  if ( OBJ_choisit( LISSAGE, LG_NV_INTERVALLE, QUESTION ) == VRAI ) {
    /* Bornes données par l'utilisateur */
    OBJ_LireIntervalle( LISSAGE, LG_XMIN, LG_XMAX, ROOT, ROOT, &(infos->x_min), &(infos->x_max), NULL, NULL );
  } else {
    /* Bornes données par le graphe */
    infos->x_min = original->x_min;
    infos->x_max = original->x_max;

    if ( infos->x_max <= infos->x_min ) {
      (void) fprintf(stderr, "%s, %s l.%u : Graphe non-trié par X croissants -> interpolation impossible !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -10;
    }
  }
  HIST_Texte(5, 0, "Intervalle       : [%g, %g]\n", infos->x_min, infos->x_max);

  /* Choix & contrôle du nombre de points */
  if ( OBJ_choisit( LISSAGE, LG_NV_NBR_PTS, QUESTION ) == VRAI ) {
    unsigned long nombre_points;

    nombre_points = OBJ_LireEntier_U( LISSAGE, LG_NPTS_APRES );
    if ( 1 > nombre_points ) {
      (void) fprintf(stderr, "%s, %s l.%u : Nombre de points absurde [0 ou 1]\n",
		     __FILE__, __FUNCTION__, __LINE__);
      HIST_Erreur( 1, "Calcul annulé", "0 ou 1 point demandé" );
      return -13;
    }
    HIST_Texte(5, 0, "Nombre de points : %lu avant, %u apr"e_GRAVE"s\n", original->nombre_point, nombre_points);

    infos->pas = ( infos->x_max - infos->x_min ) / (double) ( nombre_points - 1 );
  } else if ( OBJ_choisit( LISSAGE, LG_NV_PAS, QUESTION ) == VRAI ) {
    infos->pas = OBJ_LireNombre( LISSAGE, LG_PAS, NULL );
    if ( ( infos->pas <= 0.0 ) || ( infos->pas > ( infos->x_max - infos->x_min ) ) ) {
      (void) fprintf(stderr, "%s, %s l.%u : pas illogique [%g] !\n",
		     __FILE__, __FUNCTION__, __LINE__, infos->pas);
      HIST_Erreur( 1, "Calcul annulé", "Pas absurde" );
      return -11;
    }
  } else {
    (void) fprintf(stderr, "%s, %s l.%u : méthode inconnue, ON NE DEVRAIT PAS ÊTRE LÀ\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_AVERTISSEMENT );
    return -12;
  }

  /* Calcul du nombre reel de points */
  infos->nbr_points = 1 + (unsigned long) rint(  ( infos->x_max - infos->x_min ) / infos->pas );

  /* Analyse de la boîte de dialogue terminée */
  return 0;
}

/* -------- Préparation de l'arbre -------- */

static void I_GRF_PreparerArbre(Graphe *source)
{
  int retour;

  if ( VRAI == arbres[ LISSAGE ].drapeaux.prepare ) return;

  /* Conversion en un sous-menu de la liste des méthodes */
  OBJ_ChangerType( LISSAGE, LG_METHODE, G_SSMENU );
  OBJ_MettreMenu ( LISSAGE, LG_METHODE, NBR_METHODES_INTERPOLATION, noms_methodes_interpolation, titre_ssmenu_methodes, NULL );

  /* Aspect visuel */
  retour = ARB_Initialiser( LISSAGE, "menu_graphes.html#interpoler", I_GRF_GererArbre, (int (*)(unsigned int, int, int *)) NULL );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR en voulant initialiser l'arbre [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return;
  }

  /* On enjolive l'arbre */
  OBJ_deplace( LISSAGE, LG_NV_INTERVALLE, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( LISSAGE, LG_TXT_XMIN     , 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( LISSAGE, LG_XMIN         , 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( LISSAGE, LG_TXT_XMAX, 0, +3, DEPLACE_PIXELS );
  OBJ_deplace( LISSAGE, LG_XMAX    , 0, +3, DEPLACE_PIXELS );

  OBJ_deplace( LISSAGE, LG_NV_PAS , 0, +4, DEPLACE_PIXELS );
  OBJ_deplace( LISSAGE, LG_TXT_PAS, 0, +4, DEPLACE_PIXELS );
  OBJ_deplace( LISSAGE, LG_PAS    , 0, +4, DEPLACE_PIXELS );

  OBJ_deplace( LISSAGE, LG_NV_REFERENCE , 0, +6, DEPLACE_PIXELS );
  OBJ_deplace( LISSAGE, LG_TXT_REFERENCE, 0, +6, DEPLACE_PIXELS );

  /* Traduction éventuelle */
#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte( LISSAGE, LG_TXT_METHODE, texte_lissage_methode );

  OBJ_TraduireTexte( LISSAGE, LG_TXT_AVANT    , texte_lissage_avant  );
  OBJ_TraduireTexte( LISSAGE, LG_NPTS_APRES   , texte_lissage_apres  );
  OBJ_TraduireTexte( LISSAGE, LG_TXT_POINTS   , texte_lissage_points );
  OBJ_TraduireTexte( LISSAGE, LG_TXT_PAS      , texte_lissage_pas    );
  OBJ_TraduireTexte( LISSAGE, LG_TXT_REFERENCE, texte_lissage_caler  );

  OBJ_TraduireTexte( LISSAGE, LG_EFFACER, texte_lissage_effacer );

  OBJ_TraduireTexte( LISSAGE, LG_OK     , texte_lissage_ok      );
  OBJ_TraduireTexte( LISSAGE, LG_ANNULER, texte_lissage_annuler );
#endif

  /* Valeurs initiales */
  OBJ_MettreIntervalle( LISSAGE, LG_XMIN, LG_XMAX, LG_PAS, LG_NPTS_APRES,
			source->x_min, source->x_max, (source->x_max - source->x_min) / ( source->nombre_point - 1.0 ), source->nombre_point );
  OBJ_CopierTexte( LISSAGE, LG_METHODE, noms_methodes_interpolation[ methode_interpolation ], LGMAX_NOM_METHODE );

  /* Fin de la préparation */
  arbres[ LISSAGE ].drapeaux.prepare = VRAI;
}
