/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2006, Emmanuel CURIS
   Cration : mars 2006

   Interface graphique avec l'utilisateur

   Normalisation (graphe : rel)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "serie.h"
#include "interface.h"

/* -------- Variables locales -------- */

/* Les mthodes sont dfinies dans graphes.h : GRF_NORM_* ; 4 mthodes actuellement */
#define LG_METHODES 36

#include "int_normaliser.txt"

/* -------- Fonctions locales -------- */

static int  I_GRF_ParametresNormalisation   (unsigned int *methode, double *val_0, double *val_1, boolean *copie, const char *unite);
static void I_GRF_PreparerArbreNormalisation(unsigned int  methode, double  val_0, double  val_1, boolean  copie, const char *unite);
static void I_GRF_Normaliser_Methodes       (unsigned int arbre, unsigned int objet, unsigned long option_choisie);

/* ~-~-~-~-~-~-~-~-~ Normalisation d'un graphe [valeurs relles ou complexes] ~-~-~-~-~-~-~-~-~ */

Graphe *I_GRF_Normaliser(Graphe *graphe)
{
  Graphe *copie = NULL;
  boolean nouveau = FAUX;	   /* Rsultat dans un nouveau graphe */
  double val_0 = 0.0, val_1 = 1.0; /* Avec ces choix, on ne change rien en fait ! */
  int reponse;
  unsigned int methode = GRF_NORM_MIN_MAX;

  /* Contrles prliminaires */
  if ( NULL == graphe ) return NULL;

  if ( VRAI == graphe->drapeaux.normalise ) {
    int reponse;

    reponse = BI_ERR_alerte( principale, &ERR_GrapheDejaNormalise, VRAI );
    if ( reponse != 1 ) return NULL; /* On annule la normalisation */
  }

  /* Choix de la mthode de normalisation */
  reponse = I_GRF_ParametresNormalisation( &methode, &val_0, &val_1, &nouveau, GRF_QuelleUnite( graphe, AXE_Y ) );
  if ( reponse != 0 ) return NULL; /* Annulation ou problme */

  if ( VRAI == nouveau ) { /* On veut normaliser une copie, sans toucher  l'original */
    copie = GRF_Copier( graphe, NULL, VRAI );
    if ( NULL == copie ) {
      (void) BI_ERR_alerte( principale, &ERR_CopieImpossible, VRAI );
      return NULL;
    }
  } else {
    copie = graphe;
  }

  /* On ralise la normalisation _de la copie_ */
  reponse = GRF_Normaliser( copie, methode, val_0, val_1 );
  if ( reponse != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans la normalisation [code %d]\n",
		   __FILE__, __FUNCTION__, __LINE__, reponse);
    if ( copie != graphe ) {
      /* Aucune raison de garder la copie ! */
      GRF_eff_isole( copie );
    }
    return NULL;
  }

  /* Si la normalisation a russi, on insre la copie aprs l'original [si copie != original !] */
  if ( copie != graphe ) {
    reponse = GRF_InsererDansSerie( copie, NULL, graphe );
    if ( reponse != 0 ) {
      OPTION_A_FAIRE( "Insertion dans une autre srie en cas d'chec : demander la srie ou en crer une nouvelle" );
    }

    /* On adapte les paramtres d'affichage */
    GRF_CopierAffichage( copie, graphe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_DESSIN );

    /* En Y, cependant, on est devenu sans unit */
    GRF_ChangerUnite( copie, AXE_Y, "" );
  }

  /* On indique, dans l'historique, ce que l'on a fait */
  HIST_Titre(1, "Normalisation", "NORM");
  HIST_Texte(1, 0, "Graphe normalis"e_AIGU" : %s\n", graphe->nom);
/*   HIST_Texte(5, 0, "Facteur de normalisation utilis"e_AIGU" : %.20g\n", facteur); */
  switch ( methode )
  {
  case GRF_NORM_MIN_MAX: /* minimum  0, maximum  1 */
    HIST_Texte(2, 0, "M"e_AIGU"thode : minimum "a_GRAVE" 0, maximum "a_GRAVE" 1");
    break;

  case GRF_NORM_MIN_UTIL: /* minimum  0, valeur indique devient 1 */
    HIST_Texte(2, 0, "M"e_AIGU"thode : minimum "a_GRAVE" 0, maximum en y = %g\n", val_1);
    break;

  case GRF_NORM_UTIL_MAX: /* valeur indique devient 0, maximum  1 */
    HIST_Texte(2, 0, "M"e_AIGU"thode : minimum en y = %g, maximum "a_GRAVE" 1\n", val_0);
    break;

  case GRF_NORM_UTIL_UTIL:/* valeurs indiques deviennent 0 et 1 */
    HIST_Texte(2, 0, "M"e_AIGU"thode : minimum en y = %g, maximum en y = %g\n", val_0, val_1);
    break;
  }

  /* On a termin : on renvoie le graphe rsultat */
  return copie;
}

/* -------- Appel et gestion de l'arbre de paramtrage de la mthode -------- */

static int I_GRF_ParametresNormalisation(unsigned int *methode, double *val_0, double *val_1, boolean *copie, const char *unite)
{
  int reponse;

  /* On prpare l'arbre */
  I_GRF_PreparerArbreNormalisation( *methode, *val_0, *val_1, *copie, unite );

  /* On l'appelle */
  reponse = ARB_Bloquant( NORMALISE_GRAPHE, titre_arbre_normaliser_graphe, principale, 0 );
  if ( reponse != NG_OK ) {
    /* Annulation ou problme */
    return -1;
  }

  /* On lit les valeurs indiques par l'utilisateur */
  *methode = OBJ_OptionMenu( NORMALISE_GRAPHE, NG_METHODE, 0, QUESTION );
  *val_0   = OBJ_LireNombre( NORMALISE_GRAPHE, NG_VALEUR_0, NULL  );
  *val_1   = OBJ_LireNombre( NORMALISE_GRAPHE, NG_VALEUR_1, NULL  );
  *copie   = OBJ_choisit   ( NORMALISE_GRAPHE, NG_COPIE, QUESTION );

  /* On a termin sans souci */
  return 0;
}

static void I_GRF_PreparerArbreNormalisation(unsigned int methode, double val_0, double val_1, boolean copie, const char *unite)
{
  if ( FAUX == arbres[ NORMALISE_GRAPHE ].drapeaux.prepare ) {
    int retour;

    /* On installe le slecteur de mthodes */
    OBJ_ChangerType( NORMALISE_GRAPHE, NG_METHODE, G_SSMENU );
    OBJ_MettreMenu ( NORMALISE_GRAPHE, NG_METHODE, 4, methodes_normalisation, txt_methode_de_normalisation, I_GRF_Normaliser_Methodes );

    /* Initialisation classique */
    retour = ARB_Initialiser( NORMALISE_GRAPHE, NULL, NULL, NULL );
    if ( retour != 0 ) {
      (void) fprintf(stderr, "%s,%s l.%u : Erreur en voulant initialiser l'arbre\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return;
    }

    /* On amliore le placement des objets */
    OBJ_deplace( NORMALISE_GRAPHE, NG_TXT_VALEUR_1, 0, +1, DEPLACE_PIXELS );
    OBJ_deplace( NORMALISE_GRAPHE, NG_VALEUR_1, 0, +1, DEPLACE_PIXELS );

    /* On traduit si ncessaire */
#ifdef AVEC_TRADUCTION_
#endif

    /* Adaptation termine */
    arbres[ NORMALISE_GRAPHE ].drapeaux.prepare = VRAI;
  }

  /* On installe les valeurs initiales */
  OBJ_OptionMenu( NORMALISE_GRAPHE, NG_METHODE, methode, METTRE );

  OBJ_OptionsEditable( NORMALISE_GRAPHE, NG_VALEUR_0, BI_EDITABLE_UNITE, 0.0, 0.0, unite );
  OBJ_OptionsEditable( NORMALISE_GRAPHE, NG_VALEUR_1, BI_EDITABLE_UNITE, 0.0, 0.0, unite );

  OBJ_MettreNombre( NORMALISE_GRAPHE, NG_VALEUR_0, val_0, 0.0 );
  OBJ_MettreNombre( NORMALISE_GRAPHE, NG_VALEUR_1, val_1, 0.0 );

  OBJ_choisit( NORMALISE_GRAPHE, NG_COPIE, VRAI == copie );
}

/* -------- Choisir et installer une mthode -------- */

static void I_GRF_Normaliser_Methodes(unsigned int arbre __attribute__((__unused__)), unsigned int objet __attribute__((__unused__)), unsigned long option_choisie)
{
  switch ( option_choisie )
  {
  case GRF_NORM_MIN_MAX:
    OBJ_active( NORMALISE_GRAPHE, NG_VALEUR_0, FORCE_NON );
    OBJ_active( NORMALISE_GRAPHE, NG_VALEUR_1, FORCE_NON );
    break;

  case GRF_NORM_MIN_UTIL:
    OBJ_active( NORMALISE_GRAPHE, NG_VALEUR_0, FORCE_NON );
    OBJ_active( NORMALISE_GRAPHE, NG_VALEUR_1, FORCE_OUI );
    break;

  case GRF_NORM_UTIL_MAX:
    OBJ_active( NORMALISE_GRAPHE, NG_VALEUR_0, FORCE_OUI );
    OBJ_active( NORMALISE_GRAPHE, NG_VALEUR_1, FORCE_NON );
    break;

  case GRF_NORM_UTIL_UTIL:
    OBJ_active( NORMALISE_GRAPHE, NG_VALEUR_0, FORCE_OUI );
    OBJ_active( NORMALISE_GRAPHE, NG_VALEUR_1, FORCE_OUI );
    break;
  }
}
