/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 2006, Emmanuel CURIS
   Création : 21 septembre 2010

   Interface graphique

   Tronquer un graphe, quelle que soit la méthode
    (quoi qu'il arrive, n'étend JAMAIS le graphe)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "interface.h"
#include "operations.h"
#include "serie.h"
#include "tracer.h"

/* -------- Variables locales -------- */

static const char *menu_ChoixAxes[] = {
  "X",
  "Y",
  "X & Y"
};

/* -------- Les textes [isolés pour traduction] -------- */

#include "int_tronquer.txt"

/* -------- Fonctions locales -------- */

static boolean I_GRF_TRQ_GrapheConvient(Graphe *graphe, BI_Boite_Alerte **message_erreur);

static int  I_GRF_TRQ_GererArbre       (unsigned int num_objet, int x, int y);
static int  I_GRF_TRQ_GererClicFenetre (Fenetre *fenetre, int bouton, int n_clic, int etat, int x, int y);
static int  I_GRF_TRQ_GererClavier     (Fenetre *fenetre, unsigned int touche, int etat);
static int  I_GRF_TRQ_GererMessage     (Fenetre *fenetre, Message *message);
static int  I_GRF_TRQ_LibererFenetre   (Fenetre *fenetre);
static void I_GRF_TRQ_RedessinerFenetre(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut);

static int  I_GRF_TRQ_PreparerArbre(unsigned int);

/* ~-~-~-~-~-~-~-~-~ Appel par l'utilisateur ~-~-~-~-~-~-~-~-~ */

/* -------- Routine d'appel principale : initialisation, calculs & création d'une figure -------- */

/* ----------------- Ouverture de la fenêtre d'interface ------------------------------- */

Graphe *I_GRF_TronquerGraphe(Graphe *graphe)
{
  TRQ_infos *infos = NULL;
  Fenetre *fenetre = NULL;

  if ( NULL == graphe ) return NULL;

  /* Est-ce que l'arbre est déjà ouvert ? */
  if ( VRAI == ARB_EstOuvert( TRONQUE_GRAPHE ) ) {
    int retour;

    retour = I_GRF_RappelerModule( TRONQUE_GRAPHE, graphe, nom_module,
				   I_GRF_TRQ_GrapheConvient, I_GRF_TRQ_LibererFenetre );
    if ( I_GRF_RAPPELMODULE_ANCIEN == retour ) {
      /* On a juste mis au premier plan, rien d'autre à faire */
      return NULL;
    }

    /* Autre cas : soit erreur, soit on veut un nouveau => on continue */
  } else {
    if ( graphe->drapeaux.protege ) {
      EXF_ERREUR(GRF_PROTEGE);
      return NULL;
    }
    
    fenetre = FEN_CreerAvecArbre( TOUS_SF_ASC, titre_fenetre_tronquer, NULL,
				  TRONQUE_GRAPHE, BI_FEN_ARBRE_HAUT, I_GRF_TRQ_PreparerArbre,
				  I_GRF_TRQ_RedessinerFenetre, I_GRF_TRQ_GererClavier, I_GRF_TRQ_GererClicFenetre,
				  I_GRF_TRQ_GererMessage, RCT_pointeur,
				  ( int (*)(Fenetre *) ) NULL);
    if ( fenetre == NULL ) return NULL; /* Rien à faire... */

    FEN_Ouvrir( fenetre );
  }

  if ( NULL == infos ) { /* Aucune information disponible */
    infos = (TRQ_infos *) malloc( sizeof( TRQ_infos ) );
    if ( NULL == infos ) {
      (void) fprintf(stderr, "TRQ_ouvre : mémoire insuffisante pour les informations de troncature\n");
      FEN_Detruire( fenetre );
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "grf_stat.c", "TRQ_ouvre", "infos");
      return NULL;
    }
  }

  /* On lie les informations à la fenêtre */
  FEN_GRF_LierGraphe( fenetre, graphe );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, infos );

  fenetre->drapeaux.redessin = VRAI;
  fenetre->msg_masque.fermeture = VRAI; /* Messages traités par MSG_mu01 */

  infos->axes = AXE_X; /* Troncature sur l'axe des X uniquement, par défaut */
  met_chaine(TRONQUE_GRAPHE, TRQ_AXE, 0, menu_ChoixAxes[ 0 ], TXT_REMPLACE);

  infos->x_min = graphe->x_min; /* Positions pour la troncature */
  infos->x_max = graphe->x_max;
  OBJ_MettreIntervalle( TRONQUE_GRAPHE, TRQ_XMIN, TRQ_XMAX, 0, 0, infos->x_min, infos->x_max, 0.0, 0 );
  if ( OBJ_choisit(TRONQUE_GRAPHE, TRONQUE_EXT, QUESTION) )
    infos->mode_x = TRQ_GARDER_INTERIEUR;
  else
    infos->mode_x = TRQ_GARDER_EXTERIEUR;

  infos->y_min = graphe->y_min;
  infos->y_max = graphe->y_max;
  OBJ_MettreIntervalle( TRONQUE_GRAPHE, TRQ_YMIN, TRQ_YMAX, 0, 0, infos->y_min, infos->y_max, 0.0, 0 );

  if ( OBJ_choisit(TRONQUE_GRAPHE, TRQ_Y_EXT, QUESTION) )
    infos->mode_y = TRQ_GARDER_INTERIEUR;
  else
    infos->mode_y = TRQ_GARDER_EXTERIEUR;
  
  graphe->drapeaux.protege = VRAI;

  return graphe;
}

/* -------- Est-ce que le graphe proposé convient ? -------- */

static boolean I_GRF_TRQ_GrapheConvient(Graphe *graphe, BI_Boite_Alerte **message_erreur) {
  if ( NULL == graphe ) {
    return FAUX;
  }

  /* Tout graphe peut être tronqué, sauf s'il a un seul point ou s'il est protégé ! */
  if ( 1 == graphe->nombre_point ) {
    *message_erreur = &ERR_UnSeulPoint;
    return FAUX;
  }

  if ( graphe->drapeaux.protege ) {
    *message_erreur = &ERR_GrapheProtege;
    return FAUX;
  }

  return VRAI;
}

/* -------- Libération des informations associées à la fenêtre -------- */

static int I_GRF_TRQ_LibererFenetre(Fenetre *fenetre)
{
  Graphe *graphe;
  TRQ_infos *infos = NULL;

  /* On récupère les infos associées */
  graphe = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (TRQ_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  /* On les délie de la fenêtre */
  FEN_GRF_LierGraphe( fenetre, NULL );
  FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

  /* On n'a plus besoin des infos */
  free( infos ); infos = NULL;

  /* Le graphe n'est plus protégé */
  if ( graphe != NULL ) graphe->drapeaux.protege = FAUX;

  /* On a fini... et tout s'est bien passé */
  return 0;
}

/* ----------------- Clic sur la fenêtre de troncature -----------------------------------\
 |    Renvoie VRAI si l'événement a été traité, donc toujours VRAI                        |
 |                                                                                        |
 | bouton  : le bouton avec lequel on a cliqué (1 = gauche, 2 = droit)                    |
 | etat    : état des touches spéciales lors du clic (shift, control,...)                 |
 | fenetre : la fenêtre correspondant à la fenêtre de troncature                          |
 | n_clic  : le nombre de clics effectués                                                 |
 | x, y    : coordonnées du clic souris                                                   |
 \--------------------------------------------------------------------------------------- */

static int I_GRF_TRQ_GererClicFenetre(Fenetre *fenetre, int bouton, int n_clic __attribute__((__unused__)), int etat, int x_clic, int y_clic)
{
  int retour;
  Graphe *graphe;
  TRQ_infos *infos;
  double echelle_X, echelle_Y, x_vrai, y_vrai;

  if ( ( x_clic < 0 ) || ( x_clic > (int) (fenetre->w_larg) ) ||
       ( y_clic < 0 ) || ( y_clic > (int) (fenetre->w_haut) ) ) {
#ifdef _DEVELOPPE_
    printf("CLC_tronque : clic hors de la région de travail de la fenêtre\n");
#endif
    return BI_EVENEMENT_TRAITE; /* Clic en-dehors de la zone de travail */
  }
  
  graphe = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (TRQ_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if ( ( NULL == infos) || ( NULL == graphe ) ) {
    return BI_EVENEMENT_TRAITE;
  }

  /* Calcul des échelles en X et en Y */
  retour = TRC_Echelles( fenetre->w_larg, fenetre->w_haut,
			 graphe->x_min, graphe->x_max, graphe->y_min, graphe->y_max,
			 &echelle_X, &echelle_Y );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Graphe de X ou Y constant ou fenêtre trop petite\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return BI_EVENEMENT_TRAITE;
  }

  /* On en déduit les coordonnées réelles du point cliqué */
  x_vrai = TRC_ConvertitX_EvP( x_clic, graphe->x_min, 0, echelle_X );
  y_vrai = TRC_ConvertitY_EvP( y_clic, graphe->y_min, fenetre->w_haut - 1, echelle_Y );

  if ( etat & ( K_RSHIFT | K_LSHIFT ) ) { /* Shift-clic => positionne directement la borne la plus proche */
    int d_m, d_M;

    if ( ( infos->axes & AXE_X ) == VRAI ) {
      /* Distance du point cliqué à la borne la plus proche */
      d_m = fabs( x_vrai - infos->x_min );
      d_M = fabs( x_vrai - infos->x_max );
      if ( d_m > d_M ) /* Le clic est plus proche de x_max */
	infos->x_max = x_vrai;
      else
	infos->x_min = x_vrai;
    } else {
      d_m = fabs( y_vrai - infos->y_min );
      d_M = fabs( y_vrai - infos->y_max );
      if ( d_m > d_M ) /* Le clic est plus proche de y_max */
	infos->y_max = y_vrai;
      else
	infos->y_min = y_vrai;
    }
  } else if ( etat & K_CTRL ) { /* Ctrl-clic : pour les Y */
    if ( ( infos->axes & AXE_Y ) == FAUX ) {
      /* Ctrl-clic pour les Y n'a de sens que si l'on s'intéresse à l'axe des Y ! */
      return BI_EVENEMENT_TRAITE;
    }

    if ( 1 == bouton )
      infos->y_min = y_vrai;
    else if (bouton == 2)
      infos->y_max = y_vrai;
  } else {/* Clic normal => on positionne directement les bornes, normalement...   */
    if ( ( infos->axes & AXE_X ) == VRAI ) {
      if ( 1 == bouton )      /* Bouton droit : borne inférieure */
	infos->x_min = x_vrai;
      else if ( 2 == bouton ) /* Bouton gauche : borne supérieure */
	infos->x_max = x_vrai;
    } else {
      if ( 1 == bouton ) {
	infos->y_min = y_vrai;
      } else if ( 2 == bouton ) {
	infos->y_max = y_vrai;
      }
    }
  }

  /* Contrôles de validité */
  if ( infos->x_max < infos->x_min ) { /* On s'assure que min < max */
    double x = infos->x_max; infos->x_max = infos->x_min; infos->x_min = x;
  }
  if ( infos->x_max > graphe->x_max )
    infos->x_max = graphe->x_max;
  if ( infos->x_min < graphe->x_min )
    infos->x_min = graphe->x_min;

  if ( infos->y_max < infos->y_min ) { /* On s'assure que min < max */
    double y = infos->y_max; infos->y_max = infos->y_min; infos->y_min = y;
  }
  if ( infos->y_max > graphe->y_max )
    infos->y_max = graphe->y_max;
  if ( infos->y_min < graphe->y_min )
    infos->y_min = graphe->y_min;

  /* Affichage des valeurs choisies */
  OBJ_MettreIntervalle( TRONQUE_GRAPHE, TRQ_XMIN, TRQ_XMAX, 0, 0, infos->x_min, infos->x_max, 0.0, 0 );
  OBJ_MettreIntervalle( TRONQUE_GRAPHE, TRQ_YMIN, TRQ_YMAX, 0, 0, infos->y_min, infos->y_max, 0.0, 0 );

  FEN_redessin(fenetre, BI_RDS_UTIL); /* On force le redessin... en attendant mieux */
  return BI_EVENEMENT_TRAITE;
}

/* ----------------- Gestion des raccourcis claviers dans cette fenêtre ------------------\
 |   Renvoie VRAI si la touche a été traitée, 0 sinon                                     |
 |                                                                                        |
 | espace : fermer la fenêtre                                                             |
 | x, X   : bornes en X                                                                   |
 | y, Y   : bornes en Y                                                                   |
 \--------------------------------------------------------------------------------------- */

static int I_GRF_TRQ_GererClavier(Fenetre *fenetre, unsigned int touche, int etat)
{
  int retour = BI_EVENEMENT_NON_TRAITE;
  TRQ_infos *infos;
  Message message;

  if ( NULL == fenetre ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune fenêtre associée à l'appel !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return BI_EVENEMENT_ERREUR;
  }
  if ( etat & K_CTRL ) {
    /* Aucun raccourci avec Ctrl... */
    return BI_EVENEMENT_NON_TRAITE;
  }

  infos = (TRQ_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  switch (touche)
  {
  case ' ':
    message.type = WM_CLOSED;
    message.fenetre.application = 0;
    message.fenetre.en_plus = 0;
    message.fenetre.x = 0; message.fenetre.y = 0;
    message.fenetre.larg = 0; message.fenetre.haut = 0; /* Dans le cas où != 0, fermeture forcée... */
    if ( NULL == fenetre->message ) {
      (void) fprintf(stderr, "%s,%s l.%u : Pas de fonction de messages associée à la fenêtre [%p, %s] !\n",
		     __FILE__, __FUNCTION__, __LINE__, (void *) fenetre, fenetre->titre);
      FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
      retour = BI_EVENEMENT_ERREUR;
    } else {
      retour = (fenetre->message)( fenetre, &message );
    }
    if ( retour != BI_EVENEMENT_NON_TRAITE )
      return retour;
    return BI_EVENEMENT_TRAITE;

  case 'x':
  case 'X':
    if ( (AXE_X & infos->axes) == VRAI ) {
      OPTION_A_FAIRE( "Demander les bornes en X" );
    }
    return BI_EVENEMENT_TRAITE;

  case 'y':
  case 'Y':
    if ( (AXE_Y & infos->axes) == VRAI ) {
      OPTION_A_FAIRE( "Demander les bornes en Y" );
    }
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE; /* Touche non-traitée */
}

/* ----------------- Message destiné à la fenêtre de troncature --------------------------\
 |   Renvoie VRAI si le message a pu être traité, FAUX sinon                              |
 |                                                                                        |
 | fenetre : la fenêtre correspondant à la fenêtre de troncature                          |
 \--------------------------------------------------------------------------------------- */

static int I_GRF_TRQ_GererMessage(Fenetre *fenetre, Message *message)
{
  Graphe *graphe;
  TRQ_infos *infos;

  graphe = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (TRQ_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  if ( message->type == WM_CLOSED ) {
    Serie_Graphes *serie;
    int reponse;

    if ( message->fenetre.haut == -1 ) /* Forcer la fermeture */
      reponse = 3;
    else {
      reponse = BI_ERR_alerte(fenetre, &Q_QuitterModule, VRAI);
      if (4 == reponse) return BI_EVENEMENT_TRAITE; /* Fermeture annulée */
    }

    /* On délie les informations de la fenêtre */
    FEN_GRF_LierGraphe( fenetre, NULL );
    FEN_StockerComplement( fenetre, FEN_COMPLEMENT_2, NULL );

    if (3 == reponse) {/* On ferme sans modifier le graphe */
      free( infos ); infos = NULL;
      FEN_Detruire( fenetre );
      if (graphe != NULL) /* On peut appeler avec graphe = NULL, quand graphe effacé... */
	graphe->drapeaux.protege = FAUX;
      return BI_EVENEMENT_FEN_FERMEE;
    }

    /* On réalise la troncature */
    serie = GRF_QuelleSerie( graphe );
    if ( VRAI == serie->drapeaux.protegee ) {
      TRQ_Tronquer(graphe, infos, VRAI, VRAI);
    } else {
      TRQ_Tronquer(graphe, infos, FAUX, VRAI);
    }

    if (2 == reponse) { /* On le fait pour tous les graphes de même type de la série */
      if ( VRAI == serie->drapeaux.protegee ) { /* Travail en place : série non modifiée */
	Graphe *g;

	for (g = serie->premier_graphe ; g != NULL ; g = g->suivant) {
	  if (g == graphe) /* Déjà fait ! */
	    continue;
	  if (g->type != graphe->type) /* Types incompatibles */
	    continue;

	  TRQ_Tronquer(g, infos, VRAI, FAUX);
	}
      } else {
	Graphe **graphes;
	unsigned long nbr_graphes;

	/* Table des graphes à traiter */
	graphes = GRF_ChoisirType( serie, graphe->type, VRAI, &nbr_graphes );
	if ( NULL != graphes ) {
	  unsigned long i;

	  for ( i = 0 ; i < nbr_graphes ; i++ ) {
	    TRQ_Tronquer( graphes[ i ], infos, FAUX, FAUX );

	    graphes[ i ] = NULL;
	  }

	  free( graphes ); graphes = NULL;
	}
      }
    }

    free( infos ); infos = NULL;
    FEN_Detruire( fenetre );
    graphe->drapeaux.protege = FAUX;

    return BI_EVENEMENT_FEN_FERMEE;
  }

  AVERTISSEMENT( "Message non traité [%d]", message->type );
  return BI_EVENEMENT_NON_TRAITE;
}

/* ----------------- Redessin de la fenêtre de troncature --------------------------------\
 |    Ne renvoie rien                                                                     |
 |  Suppose le clipping déjà installé avant l'appel, ainsi que la souris et l'AES coupés  |
 |                                                                                        |
 | fenetre : la fenêtre qui correspond à la fenêtre de troncature                         |
 | x, y, larg, haut : coordonnées de la zone de redessin                                  |
 \--------------------------------------------------------------------------------------- */

static void I_GRF_TRQ_RedessinerFenetre(Fenetre *fenetre, int x, int y, unsigned int larg, unsigned int haut)
{
  TRQ_infos *infos;
  Graphe *graphe;
  double x_echelle, y_echelle;

  /* Préparation du dessin */
  TRC_InitDessins( fenetre, x, y, larg, haut );
  
  /* On trace le graphe que l'on traite */
  graphe = (Graphe *) FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  if (graphe == NULL) {
    TRC_Message( fenetre, ERR_AucunGraphe );
    return;
  }

  if ( (graphe->x_max == graphe->x_min) || (graphe->y_max == graphe->y_min) ) {
    TRC_Message( fenetre, ERR_GrapheConstant );
    return;
  }

  x_echelle = ( (double) fenetre->w_larg ) / (graphe->x_max - graphe->x_min);
  y_echelle = ( (double) fenetre->w_haut ) / (graphe->y_max - graphe->y_min);

  TRC_Graphe( fenetre, graphe, NOIR, 1, 0, 0, fenetre->w_x, fenetre->w_y,
	      graphe->x_min, graphe->x_max, graphe->y_min, graphe->y_max, x_echelle, y_echelle );

  /* On hachure la zone qui n'est pas conservée */
  infos = (TRQ_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );
  if (infos == NULL) {
    TRC_Message( fenetre, "Aucune information de troncature" );
    return;
  }

  TRC_Mode( MD_INVERSE );
  TRC_Couleur( GRIS_FONCE );
  TRC_Motif( FIS_HATCH ); /* On hachure */

  if ( infos->axes == AXE_X ) { /* Axe des X seul */
    int x_min, x_max;

    x_min = (int) rint( (infos->x_min - graphe->x_min) * x_echelle );
    x_max = (int) rint( (infos->x_max - graphe->x_min) * x_echelle );

    /* Le hachurage */
    if ( TRQ_GARDER_INTERIEUR == infos->mode_x ) {
      TRC_RectanglePlein( fenetre, 0, 0, x_min, fenetre->w_haut );
      TRC_RectanglePlein( fenetre, x_max, 0, fenetre->w_larg - x_max, fenetre->w_haut );
    } else {
      TRC_RectanglePlein( fenetre, x_min, 0, x_max - x_min, fenetre->w_haut );
    }

    TRC_Mode( MD_REMPLACE );
    /* Limite inférieure */
    TRC_Couleur( ROUGE );
    TRC_LigneV( fenetre, x_min, 0, fenetre->w_haut );

    /* Limite supérieure */
    TRC_Couleur( VERT );
    TRC_LigneV( fenetre, x_max, 0, fenetre->w_haut );
  } else if ( infos->axes == AXE_Y ) { /* Axe des Y seul */
    int y_min, y_max;

    y_min = fenetre->w_y + fenetre->w_haut - (int) rint( ( infos->y_min - graphe->y_min ) * y_echelle );
    y_max = fenetre->w_y + fenetre->w_haut - (int) rint( ( infos->y_max - graphe->y_min ) * y_echelle );

#ifdef _DEVELOPPE_
    vswr_mode(_PRG_station->handle, MD_INVERSE);
#endif
    /* Le hachurage */
    if ( TRQ_GARDER_INTERIEUR == infos->mode_y ) {
      TRC_RectanglePlein( fenetre, 0, 0, fenetre->w_larg, y_min );
      TRC_RectanglePlein( fenetre, 0, y_max, fenetre->w_larg, fenetre->w_haut - y_max );
    } else {
      TRC_RectanglePlein( fenetre, 0, y_min, fenetre->w_larg, y_max - y_min );
    }

    vswr_mode(_PRG_station->handle, MD_REMPLACE);
    /* Limite inférieure */
    TRC_Couleur( ROUGE );
    TRC_LigneH( fenetre, 0, y_min, fenetre->w_larg );

    /* Limite supérieure */
    TRC_Couleur( VERT );
    TRC_LigneH( fenetre, 0, y_max, fenetre->w_larg );
  } else {
    int x_min, x_max, y_min, y_max;

    x_min = (int) ( fenetre->w_x + (infos->x_min - graphe->x_min) * x_echelle );
    x_max = (int) ( fenetre->w_x + (infos->x_max - graphe->x_min) * x_echelle );
    y_min = (int) ( fenetre->w_y + (infos->y_min - graphe->y_min) * y_echelle );
    y_max = (int) ( fenetre->w_y + (infos->y_max - graphe->y_min) * y_echelle );

#ifdef _DEVELOPPE_
    vswr_mode(_PRG_station->handle, MD_INVERSE);
#endif
    if ( TRQ_GARDER_INTERIEUR == infos->mode_x ) {
      if ( TRQ_GARDER_INTERIEUR == infos->mode_y ) {
	/* Puisque l'on est en inversion : on noircit tout, puis juste la partie conservée ! */
	TRC_RectanglePlein( fenetre, 0, 0, fenetre->w_larg, fenetre->w_haut );
	TRC_RectanglePlein( fenetre, x_min, y_min, x_max - x_min, y_max - y_min );
      } else {
	TRC_RectanglePlein( fenetre, 0, 0, x_min, fenetre->w_haut );
	TRC_RectanglePlein( fenetre, x_max, 0, fenetre->w_larg - x_max, fenetre->w_haut );

	TRC_RectanglePlein( fenetre, x_min, y_min, x_max - x_min, y_max - y_min );
      }
    } else {
      if ( TRQ_GARDER_INTERIEUR == infos->mode_y ) {
	TRC_RectanglePlein( fenetre, 0, 0, fenetre->w_larg, y_min );
	TRC_RectanglePlein( fenetre, 0, y_max, fenetre->w_larg, fenetre->w_haut - y_max );

	TRC_RectanglePlein( fenetre, x_min, y_min, x_max - x_min, y_max - y_min );
      } else {
	/* On garde les coins : on enlève les X mauvais + les deux bords en Y... */
	TRC_RectanglePlein( fenetre, x_min, 0, x_max - x_min, fenetre->w_haut );

	TRC_RectanglePlein( fenetre, 0, y_min, x_min, y_max - y_min );
	TRC_RectanglePlein( fenetre, x_max, y_min, fenetre->w_larg - x_max, y_max - y_min );
      }
    }
  }
}

/* -------- Gestion de l'arbre GEM dans la fenêtre -------- */

static int I_GRF_TRQ_GererArbre(unsigned int num_objet,
				int x __attribute__((__unused__)), int y __attribute__((__unused__)) )
{
  TRQ_infos *infos;
  Fenetre *fenetre;
  Graphe *graphe;
  int reponse;
  struct infos_demande nv_valeur = { BI_CHAMP_DEMANDE_REEL, NULL, NULL, 0, 0.0, 0.0, 0.0, 0, NULL, { 0 }, { 0 }, { 0 } };

  fenetre = ARB_QuelleFenetre( &(arbres[ TRONQUE_GRAPHE ]), FAUX );
  graphe = (Graphe *)    FEN_LireComplement( fenetre, FEN_INFOS_GRAPHE );
  infos  = (TRQ_infos *) FEN_LireComplement( fenetre, FEN_COMPLEMENT_2 );

  switch (num_objet)
  {
  case TRQ_XMIN:
    nv_valeur.proprietes |= BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeur.texte = "Xmin =";
    nv_valeur.min = graphe->x_min;
    nv_valeur.max = infos->x_max;
    nv_valeur.valeur_initiale.reel = infos->x_min;

    reponse = BI_DemandeValeur( fenetre, &Q_BorneInferieure, &nv_valeur, 'X' );
    if ( Q_BorneInferieure.bouton_annuler == (unsigned int) reponse ) return BI_EVENEMENT_TRAITE; /* On annule */

    infos->x_min = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( TRONQUE_GRAPHE, TRQ_XMIN, infos->x_min, 0.0 );
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case TRQ_XMAX:
    nv_valeur.proprietes |= BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeur.texte = "Xmax =";
    nv_valeur.min = infos->x_min;
    nv_valeur.max = graphe->x_max;
    nv_valeur.valeur_initiale.reel = infos->x_max;

    reponse = BI_DemandeValeur( fenetre, &Q_BorneSuperieure, &nv_valeur, 'X' );
    if ( Q_BorneSuperieure.bouton_annuler == (unsigned int) reponse ) return BI_EVENEMENT_TRAITE; /* On annule */

    infos->x_max = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( TRONQUE_GRAPHE, TRQ_XMAX, infos->x_max, 0.0 );
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case TRONQUE_EXT:
  case TRONQUE_INT:
    OBJ_choisit(TRONQUE_GRAPHE, num_objet, FORCE_OUI);
    infos->mode_x = (num_objet == TRONQUE_INT) ? (TRQ_GARDER_EXTERIEUR) : (TRQ_GARDER_INTERIEUR);
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case TRQ_YMIN:
    nv_valeur.proprietes |= BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeur.texte = "Ymin =";
    nv_valeur.min = graphe->y_min;
    nv_valeur.max = infos->y_max;
    nv_valeur.valeur_initiale.reel = infos->y_min;

    reponse = BI_DemandeValeur( fenetre, &Q_BorneInferieure, &nv_valeur, 'Y' );
    if ( Q_BorneInferieure.bouton_annuler == (unsigned int) reponse ) return BI_EVENEMENT_TRAITE; /* On annule */

    infos->y_min = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( TRONQUE_GRAPHE, TRQ_YMIN, infos->y_min, 0.0 );
    FEN_redessin( fenetre, BI_RDS_UTIL );
    return BI_EVENEMENT_TRAITE;

  case TRQ_YMAX:
    nv_valeur.proprietes |= BI_EDITABLE_MIN | BI_EDITABLE_MAX;
    nv_valeur.texte = "Ymax =";
    nv_valeur.min = infos->y_min;
    nv_valeur.max = graphe->y_max;
    nv_valeur.valeur_initiale.reel = infos->y_max;

    reponse = BI_DemandeValeur( fenetre, &Q_BorneSuperieure, &nv_valeur, 'Y' );
    if ( Q_BorneSuperieure.bouton_annuler == (unsigned int) reponse ) return BI_EVENEMENT_TRAITE; /* On annule */

    infos->y_max = nv_valeur.valeur_donnee.reel;
    OBJ_MettreNombre( TRONQUE_GRAPHE, TRQ_YMAX, infos->y_max, 0.0 );
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case TRQ_Y_EXT:
  case TRQ_Y_INT:
    OBJ_choisit(TRONQUE_GRAPHE, num_objet, FORCE_OUI);
    infos->mode_y = ( num_objet == TRQ_Y_INT ) ? (TRQ_GARDER_EXTERIEUR) : (TRQ_GARDER_INTERIEUR);
    FEN_redessin(fenetre, BI_RDS_UTIL);
    return BI_EVENEMENT_TRAITE;

  case TRQ_AXE:
    OBJ_choisit(TRONQUE_GRAPHE, num_objet, FORCE_OUI);
    {
      int nv_methode;

      nv_methode = SS_menu_texte(menu_ChoixAxes, 3, "Axes à considérer", TRONQUE_GRAPHE, TRQ_AXE);
      if ( 0 == nv_methode ) {
	infos->axes = AXE_X;
	OBJ_active( TRONQUE_GRAPHE, TRQ_XMIN, FORCE_OUI );
	OBJ_active( TRONQUE_GRAPHE, TRQ_XMAX, FORCE_OUI );
	OBJ_active( TRONQUE_GRAPHE, TRQ_YMIN, FORCE_NON );
	OBJ_active( TRONQUE_GRAPHE, TRQ_YMAX, FORCE_NON );
	met_chaine( TRONQUE_GRAPHE, TRQ_AXE, 0, menu_ChoixAxes[ 0 ], TXT_REMPLACE );
      } else if ( 1 == nv_methode ) {
	infos->axes = AXE_Y;
	OBJ_active( TRONQUE_GRAPHE, TRQ_XMIN, FORCE_NON );
	OBJ_active( TRONQUE_GRAPHE, TRQ_XMAX, FORCE_NON );
	OBJ_active( TRONQUE_GRAPHE, TRQ_YMIN, FORCE_OUI );
	OBJ_active( TRONQUE_GRAPHE, TRQ_YMAX, FORCE_OUI );
	met_chaine( TRONQUE_GRAPHE, TRQ_AXE, 0, menu_ChoixAxes[ 1 ], TXT_REMPLACE );
      } else if ( 2 == nv_methode ) {
	infos->axes = AXE_X | AXE_Y;
	OBJ_active( TRONQUE_GRAPHE, TRQ_XMIN, FORCE_OUI );
	OBJ_active( TRONQUE_GRAPHE, TRQ_XMAX, FORCE_OUI );
	OBJ_active( TRONQUE_GRAPHE, TRQ_YMIN, FORCE_OUI );
	OBJ_active( TRONQUE_GRAPHE, TRQ_YMAX, FORCE_OUI );
	met_chaine( TRONQUE_GRAPHE, TRQ_AXE, 0, menu_ChoixAxes[ 2 ], TXT_REMPLACE );
      }
      FEN_redessin(fenetre, BI_RDS_UTIL);
    }
    OBJ_choisit(TRONQUE_GRAPHE, num_objet, FORCE_NON);
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

static int I_GRF_TRQ_PreparerArbre(unsigned int numero_arbre __attribute__((__unused__)))
{
  OBJ_agrandit( TRONQUE_GRAPHE, ROOT          , 0, +2 );
  OBJ_deplace ( TRONQUE_GRAPHE, TRQ_XMAX      , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace ( TRONQUE_GRAPHE, TRONQUE_INT   , 0, +1, DEPLACE_PIXELS );
  OBJ_agrandit( TRONQUE_GRAPHE, TRQ_FND_EXTINT, 0, +2 );
  OBJ_deplace ( TRONQUE_GRAPHE, TRQ_YMAX      , 0, +1, DEPLACE_PIXELS );
  OBJ_deplace ( TRONQUE_GRAPHE, TRQ_Y_INT     , 0, +1, DEPLACE_PIXELS );

#ifndef _DEVELOPPE_
  OBJ_active(TRONQUE_GRAPHE, TRQ_YMIN, FORCE_NON);
  OBJ_active(TRONQUE_GRAPHE, TRQ_YMAX, FORCE_NON);
  OBJ_active(TRONQUE_GRAPHE, TRQ_Y_EXT, FORCE_NON);
  OBJ_active(TRONQUE_GRAPHE, TRQ_Y_INT, FORCE_NON);
#endif

  arbres[ TRONQUE_GRAPHE ].fonction.arbre = I_GRF_TRQ_GererArbre;
  arbres[ TRONQUE_GRAPHE ].drapeaux.prepare = VRAI;

  return 0;
}
