/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 2006, Emmanuel CURIS
   Création : 31 décembre 2009 [par reprise de EXAFS/normalise.c]

   Interface graphique

   Normalisation d'oscillations EXAFS
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "exafs.h"

#include "absorption.h"		/* Pour les SEUIL_* */
#include "chimie_utils.h"	/* Pour les éléments chimiques */
#include "extrait.h"		/* Pour les NORM_TH_* */
#include "graphes.h"
#include "interface.h"

/* -------- Variables locales -------- */

static unsigned int methode = MODELE_MU0_LENGELER; /* Méthode théorique de normalisation */
struct infos_normalisation {
  unsigned int methode;
  struct graphe *mu0_normalise;
  int seuil;
};

#define LG_MAX_NOM_GRAPHE 50 /* caractères au plus pour le nom du graphe mu_0 dans l'arbre */
#define LGMAX_NOM_METHODE 30 /* nom de la méthode dans l'arbre */
#define LGMAX_NOM_SEUIL   13 /* nom du seuil étudié */

#define NOMBRE_MAXIMAL_COEFFICIENTS 4 /* Pour le polynôme modèle  */

/* -------- Les textes [isolés pour traduction] -------- */

#include "int_xas_normaliser.txt"

/* -------- Fonctions locales -------- */

static void I_XAS_NORM_PreparerArbre(void);
static int  I_XAS_NORM_GererArbre   (unsigned int objet, int x, int y);

static void I_XAS_NORM_InstallerMethode(int methode);
static void I_XAS_NORM_ProposerValeurs (void);

void I_XAS_ErreurDeNormalisationEXAFS(int code_erreur); /* À passer en static quand étape 3 + reproduction EXAFS seront nettoyés & placés dans Interface... */

/* ~-~-~-~-~-~-~-~-~ Appel par l'utilisateur ~-~-~-~-~-~-~-~-~ */

/* -------- Interface de la normalisation ----------\
 | Renvoi le graphe normalisé (NULL si erreur)      |
 |                                                  |
 | source : le graphe à normaliser                  |
 \------------------------------------------------- */

Graphe *I_XAS_NormaliserEXAFS(Graphe *source)
{
  Graphe *vraie_source;
  double e_0 = 0.0; /* Énergie de seuil, pour les modèles théoriques */
  double h_seuil = 0.0; /* Hauteur de seuil expérimentale */
  int retour, Z, degre;
  struct infos_normalisation infos = { methode, NULL, SEUIL_K };
  boolean mu0_present = FAUX, sortie;

  /* 0. Vérifications préliminaires */
  if ( NULL == source ) return NULL; /* Pas de graphe en entrée */
  if ( FAUX == source->drapeaux.exafs ) {
    int reponse;

    reponse = BI_ERR_alerte( principale, &ERR_GrapheNEstPasExafs, VRAI );
    if ( I_GRF_NORM_EXAFS_NEXAFS_ANNULER == reponse ) {
      return NULL;
    }
  }

  if ( FAUX == GRF_EstFonctionReelle( source ) ) {
    (void) BI_ERR_alerte( principale, &ERR_GrapheNEstPasReel, VRAI );
    return NULL;
  }

  if ( VRAI == source->drapeaux.normalise ) {
    int reponse;

    reponse = BI_ERR_alerte( principale, &ERR_GrapheDejaNormalise, VRAI );
    if ( I_GRF_NORM_EXAFS_DEJA_ANNULER == reponse ) {
      return NULL;
    }
  }

  /* Si on est là : il ne devrait pas y avoir de souci... */
  I_XAS_NORM_PreparerArbre( );

  mu0_present = GRF_type_existe( GR_MU0, FAUX, NULL ); /* Existe-t-il des µ0 en mémoire ? */
  OBJ_active( NORM_EXAFS, MU0_CALCULE, mu0_present );
  OBJ_active( NORM_EXAFS, MU0_A_UTILISER, mu0_present );

  if ( FAUX == mu0_present ) {
    OBJ_choisit( NORM_EXAFS, MU0_THEORIQUE, FORCE_OUI );
    OBJ_CopierTexte( NORM_EXAFS, MU0_A_UTILISER, txt_aucun_mu0_disponible, LG_MAX_NOM_GRAPHE );
  } else {
    OBJ_active( NORM_EXAFS, MU0_A_UTILISER, OBJ_choisit(NORM_EXAFS, MU0_CALCULE, QUESTION) );
    OBJ_CopierTexte( NORM_EXAFS, MU0_A_UTILISER, txt_aucun_mu0_choisi, LG_MAX_NOM_GRAPHE );
  }

  /* Méthode de normalisation ? */
  OBJ_ChoisirModeleMu0( NORM_EXAFS, NE_CHOIX_METHODE, LGMAX_NOM_METHODE, methode, METTRE );
  OBJ_ChoisirSeuil( NORM_EXAFS, N_SUIL, LGMAX_NOM_SEUIL, infos.seuil , METTRE );

  do {
    sortie = VRAI;

    ARB_MettreComplement( NORM_EXAFS, &infos );
    retour = ARB_Bloquant( NORM_EXAFS, txt_titre_normalise, principale, 0 );
    ARB_MettreComplement( NORM_EXAFS, NULL );
    if ( NORM_ANNULER == retour ) {
      return NULL;
    }

    /* On mémorise la méthode */
    methode = infos.methode;

    /* On fait des vérifications... */
    if ( VRAI == OBJ_choisit( NORM_EXAFS, MU0_CALCULE, QUESTION ) ) {
      /* Vérifications : le µ0 indiqué est-il utilisable ? */

      if ( NULL == infos.mu0_normalise ) {
	/* Pas de µ0 calculé indiqué ! */

	(void) BI_ERR_alerte( principale, &ERR_GrapheMU0Absent, VRAI );
	sortie = FAUX;
      } else if ( infos.mu0_normalise->y_min <= 0.0 ) {
	int reponse;
	unsigned long i, nombre_y_negatifs = 0;
	double x_max_negatif;

	/* Certaines valeurs de µ0 sont négatives ! */
	/* On précise, en admettant le graphe trié par X croissants */
	x_max_negatif = infos.mu0_normalise->x_min;
	for ( i = 0 ; i < infos.mu0_normalise->nombre_point ; i++ ) {
	  if ( infos.mu0_normalise->y[ i ] < 0.0 ) {
	    nombre_y_negatifs++;
	    if ( infos.mu0_normalise->x[ i ] > x_max_negatif ) {
	      x_max_negatif = infos.mu0_normalise->x[ i ];
	    }
	  }
	}

	reponse = BI_ERR_alerte( principale, &ERR_GrapheMU0Negatif, VRAI,
				 nombre_y_negatifs, infos.mu0_normalise->nombre_point, x_max_negatif, GRF_QuelleUnite( infos.mu0_normalise, AXE_X ) );
	if ( I_GRF_NORM_EXAFS_MUONEGATIF_ANNULER == reponse ) {
	  sortie = FAUX;
	}
      } else {
	/* À faire : compatibilité des X ? */
      }
    } else {
      /* Vérifications : les paramètres sont-ils raisonnables ? */

      /* Hauteur de seuil, h */
      h_seuil = OBJ_LireNombre( NORM_EXAFS, NEXF_H_SEUIL, NULL );
      if ( h_seuil <= 0.0 ) {
	(void) BI_ERR_alerte( principale, &ERR_hIncorrecte, VRAI, h_seuil );
	sortie = FAUX;
      }

      /* Énergie de seuil, E0 */
      e_0 = OBJ_LireNombre( NORM_EXAFS, NORM_CONSTANTE_1, NULL );
      if ( e_0 <= 0.0 ) {
	(void) BI_ERR_alerte( principale, &ERR_E0Incorrect, VRAI, e_0, "eV" );
	sortie = FAUX;
      }

      /* Paramètres : pas de vérification simple... */
    }
  } while ( FAUX == sortie );

  /* Calculs confirmés... */
  if ( VRAI == OBJ_choisit( NORM_EXAFS, NORM_EXF_NOUVEAU, QUESTION ) ) {
    /* On veut travailler sur une copie... */
    vraie_source = GRF_Copier( source, NULL, VRAI );
    if ( NULL == vraie_source ) {
      (void) BI_ERR_alerte( principale, &ERR_CopieImpossible, VRAI );
      return NULL;
    }
  } else {
    vraie_source = source;
  }

  if ( VRAI == OBJ_choisit( NORM_EXAFS, MU0_CALCULE, QUESTION ) ) { /* On utilise un mu_0 calculé par ailleurs */
    retour = EXF_NormaliseExp( vraie_source, infos.mu0_normalise, FAUX );
  } else {
    /* Valeurs utiles */
    double C_avant[ NOMBRE_MAXIMAL_COEFFICIENTS ], C_apres[ NOMBRE_MAXIMAL_COEFFICIENTS ]; /* Coefficients pour le modèle */

    /* Élément étudié */
    Z = OBJ_NumeroAtomique( NORM_EXAFS, NORM_ELEMENT );

    /* Paramètres du modèle */
    C_avant[ 0 ] = OBJ_LireNombre( NORM_EXAFS, NORM_CTE_2, NULL ); /* A0 / C */
    C_avant[ 1 ] = OBJ_LireNombre( NORM_EXAFS, NORM_CTE_3, NULL ); /* A1 / D */
    C_avant[ 2 ] = OBJ_LireNombre( NORM_EXAFS, NORM_CTE_4, NULL ); /* A2 */
    C_avant[ 3 ] = OBJ_LireNombre( NORM_EXAFS, NORM_CTE_5, NULL ); /* A3 */

    C_apres[ 0 ] = OBJ_LireNombre( NORM_EXAFS, NORM_CTE_6, NULL ); /* A0 / C / Cte */
    C_apres[ 1 ] = OBJ_LireNombre( NORM_EXAFS, NORM_CTE_7, NULL ); /* A1 / D */
    C_apres[ 2 ] = OBJ_LireNombre( NORM_EXAFS, NORM_CTE_8, NULL ); /* A2 */
    C_apres[ 3 ] = OBJ_LireNombre( NORM_EXAFS, NORM_CTE_9, NULL ); /* A3 */

    /* Normalisation proprement dite */
    retour = EXF_NormaliseTh( vraie_source, methode, e_0, SEUIL_K, Z, h_seuil, C_avant, C_apres);
  }
  if ( retour != PAS_DE_PROBLEME ) {
    (void) fprintf( stderr, "%s,%s l.%u : erreur en voulant diviser par µ0 ! [code %d]\n",
		    __FILE__, __FUNCTION__, __LINE__, retour);
    I_XAS_ErreurDeNormalisationEXAFS( retour );

    if ( retour < 0 ) {
      /* Erreur grave : le calcul n'a pas eu lieu. On n'a donc rien changé... */
      if ( vraie_source != source ) {
	GRF_eff_isole( vraie_source ); vraie_source = NULL; 
      }
      return NULL;
    }
  }

  /* Traitement post-normalisation */
  if ( source != vraie_source ) {
    /* On insère cette copie après l'original */
    I_GRF_InsererDansSerie( NULL, source, vraie_source );

    /* Et on lui donne les mêmes attributs... */
    GRF_CopierAffichage( vraie_source, source, GRF_CP_AFF_TOUT );
    GRF_titre( vraie_source, source->nom );
  }

  /* On fait une éventuelle correction de pondération par k^n */
  degre = OBJ_LireEntier_S( NORM_EXAFS, NORM_DEGRE ); /* Pondération par k^n : valeur de n */
  if ( degre != 0 ) {
    HIST_Texte( 5, 0, "Pond"e_AIGU"ration par k^n, n = %d\n", degre );
    EXF_k_p_n( vraie_source, -degre, FAUX ); /* Si EXAFS pondéré par k^n, µ0 aussi et comme en pratique on divise par k^n µ0... */

    GRF_modifier( source, vraie_source, FAUX ); /* Nbr de points n'a pas changé, mais spectre si !
						   Rq : déjà appelé par les EXF_Normaliser() donc inutile de le refaire ! */
  }

  /* Tout s'est bien passé : on renvoie le graphe normalisé */
  return vraie_source;
}

/* -------- Affichage des messages d'erreur en cas de problème [retour de la fonction de normalisation] -------- */

void I_XAS_ErreurDeNormalisationEXAFS(int code_erreur)
{
  if ( code_erreur < 0 ) {
    /* Les codes <0 : erreur grave */
    (void) BI_ERR_alerte( principale, &ERR_ErreurInterne, VRAI, code_erreur, ERR_textes_erreurs_graves[ -code_erreur - 1 ] );
    return;
  }
  switch ( code_erreur )
  {
    /* 0 */
  case PAS_DE_PROBLEME:
    return;

    /* Les codes >0 : simple avertissement */
  case NORMEXAFS_DIV_ZERO:
    (void) BI_ERR_alerte( principale, &MSG_Mu0Nul, VRAI );
    break;

  case NORMEXAFS_ARRONDIS:
    (void) BI_ERR_alerte( principale, &MSG_Mu0QuasiNul, VRAI );
    break;

  case NORMEXAFS_ERREURMU0_NEGLIGEES:
    (void) BI_ERR_alerte( principale, &MSG_ErreursMu0Negligees, VRAI );
    break;

  case NORMEXAFS_K_NEGATIF:
    (void) BI_ERR_alerte( principale, &MSG_X_Negatifs, VRAI );
    break;
  }
}

/* -------- Gestion de l'interface de normalisation (procédure habituelle) -------- */

static int I_XAS_NORM_GererArbre(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  struct infos_normalisation *infos;

  infos = ( struct infos_normalisation * ) ARB_LireComplement( NORM_EXAFS );

  switch (objet)
  {
  case MU0_CALCULE:
    OBJ_choisit( NORM_EXAFS, MU0_CALCULE, FORCE_OUI );
    OBJ_active ( NORM_EXAFS, MU0_A_UTILISER, FORCE_OUI );
    OBJ_active ( NORM_EXAFS, NORM_PROPOSER , FORCE_NON );
    OBJ_active ( NORM_EXAFS, NE_CHOIX_METHODE, FORCE_NON );

    I_XAS_NORM_InstallerMethode( -1 );
    return BI_EVENEMENT_TRAITE;

  case MU0_A_UTILISER:
    infos->mu0_normalise = GRF_choisir(GR_MU0, VRAI, VRAI, NULL);
    if ( infos->mu0_normalise != NULL )
      OBJ_CopierTexte( NORM_EXAFS, MU0_A_UTILISER, infos->mu0_normalise->nom, LG_MAX_NOM_GRAPHE );
    else {
      OBJ_CopierTexte( NORM_EXAFS, MU0_A_UTILISER, txt_aucun_mu0_choisi, LG_MAX_NOM_GRAPHE );
    }
    return BI_EVENEMENT_TRAITE;

  case MU0_THEORIQUE:
    OBJ_choisit( NORM_EXAFS, MU0_THEORIQUE , FORCE_OUI );
    OBJ_active ( NORM_EXAFS, NORM_PROPOSER , FORCE_OUI );
    OBJ_active ( NORM_EXAFS, MU0_A_UTILISER, FORCE_NON );
    OBJ_active ( NORM_EXAFS, NE_CHOIX_METHODE, FORCE_OUI );
    OBJ_active ( NORM_EXAFS, NORM_CONSTANTE_1, FORCE_OUI ); /* E0 */
    I_XAS_NORM_InstallerMethode( infos->methode );
    return BI_EVENEMENT_TRAITE;

  case N_SUIL:
    infos->seuil = OBJ_ChoisirSeuil( NORM_EXAFS, N_SUIL, LGMAX_NOM_SEUIL, infos->seuil, QUESTION );
    return BI_EVENEMENT_TRAITE;

  case NE_CHOIX_METHODE:
    infos->methode = OBJ_ChoisirModeleMu0( NORM_EXAFS, NE_CHOIX_METHODE, LGMAX_NOM_METHODE, infos->methode, 0 );
    I_XAS_NORM_InstallerMethode( infos->methode );
    return BI_EVENEMENT_TRAITE;

  case NORM_PROPOSER: /* Propositions de constantes en fonction de Z et de la méthode */
    I_XAS_NORM_ProposerValeurs();
    return BI_EVENEMENT_TRAITE;
  }

  return BI_EVENEMENT_NON_TRAITE;
}

/* -------- Activation des objets en fonction de la méthode -------- */

static void I_XAS_NORM_InstallerMethode(int methode)
{
  switch ( methode )
  {
  case MODELE_MU0_CONSTANT:
    OBJ_active( NORM_EXAFS, NORM_CTE_2, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_3, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_4, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_5, FORCE_NON );

    /* On demande la copie car l'ancien est libéré avec free()... */
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_6, "C  =", VRAI );
    OBJ_active( NORM_EXAFS, NORM_CTE_6, FORCE_OUI ); /* Constante de normalisation */
    OBJ_active( NORM_EXAFS, NORM_CTE_7, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_8, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_9, FORCE_NON );
    break;

  case MODELE_MU0_LENGELER: /* Pas de constante utilisée, autre que E0 */
  case MODELE_MU0_HEITLER:
    OBJ_active( NORM_EXAFS, NORM_CTE_2, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_3, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_4, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_5, FORCE_NON );

    OBJ_active( NORM_EXAFS, NORM_CTE_6, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_7, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_8, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_9, FORCE_NON );
    break;

  case MODELE_MU0_VICTOREEN: /* Deux constantes avant, deux après */
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_2, "C0 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_3, "D0 =", VRAI );

    OBJ_active( NORM_EXAFS, NORM_CTE_2, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_3, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_4, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_5, FORCE_NON );

    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_6, "C1 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_7, "D1 =", VRAI );

    OBJ_active( NORM_EXAFS, NORM_CTE_6, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_7, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_8, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_9, FORCE_NON );
    break;

  case MODELE_MU0_MCMASTER: /* 4 constantes avant et après */
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_2, "A0 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_3, "A1 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_4, "A2 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_5, "A3 =", VRAI );

    OBJ_active( NORM_EXAFS, NORM_CTE_2, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_3, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_4, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_5, FORCE_OUI );

    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_6, "B0 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_7, "B1 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_8, "B2 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_9, "B3 =", VRAI );

    OBJ_active( NORM_EXAFS, NORM_CTE_6, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_7, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_8, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_9, FORCE_OUI );
    break;

  case MODELE_MU0_EBEL:
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_2, "C0 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_3, "C1 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_4, "C2 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_5, "C3 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_6, "C4 =", VRAI );
    OBJ_RemplacerTexte( NORM_EXAFS, NORM_TXT_CTE_7, "C5 =", VRAI );

    OBJ_active( NORM_EXAFS, NORM_CTE_2, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_3, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_4, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_5, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_6, FORCE_OUI );
    OBJ_active( NORM_EXAFS, NORM_CTE_7, FORCE_OUI );

    OBJ_active( NORM_EXAFS, NORM_CTE_8, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_9, FORCE_NON );
    break;

  default:			/* Méthode inconnue => aucune constante */
    /* Constante 1 : ??? */
    OBJ_active( NORM_EXAFS, NORM_CTE_2, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_3, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_4, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_5, FORCE_NON );

    OBJ_active( NORM_EXAFS, NORM_CTE_6, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_7, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_8, FORCE_NON );
    OBJ_active( NORM_EXAFS, NORM_CTE_9, FORCE_NON );
    break;
  }
}

/* -------- Proposition de valeurs de constantes pour les calculs théoriques ------ */

static void I_XAS_NORM_ProposerValeurs(void)
{
  double e_0;
  int element;
  struct infos_normalisation *infos;

  infos = ( struct infos_normalisation * ) ARB_LireComplement( NORM_EXAFS );

  element = OBJ_NumeroAtomique( NORM_EXAFS, NORM_ELEMENT );
  if ( (element < HYDROGENE) || (element > 119) ) {
    ATM_ERREUR(ELEMENT_INCONNU, element);
    return;
  }

  /* Valeur de E0 proposée */
  e_0 = TH_DonneesSeuil( element, infos->seuil, NULL, NULL );    
  if (e_0 == -2.0) { /* Valeur inconnue */
    (void) BI_ERR_alerte( NULL, &MSG_SeuilInconnu, VRAI, TH_TexteSeuil( infos->seuil ), ATM_Z_vers_nom( element ), element );
  } else if (e_0 == -1.0) { /* Valeur n'a pas de sens */
    (void) BI_ERR_alerte( NULL, &ERR_SeuilInexistant, VRAI, TH_TexteSeuil( infos->seuil ), ATM_Z_vers_nom( element ), element );
  }

  OBJ_MettreNombre( NORM_EXAFS, NORM_CONSTANTE_1, e_0, 0.0 );

  if ( infos->methode == MODELE_MU0_VICTOREEN ) {
    double c1, d1, c2, d2;

    /* Seuils classés par ordre croissant... */
    /* a. Avant seuil */
    TH_coef_Victoreen( element, infos->seuil, &c1, &d1 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_2, c1, 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_3, d1, 0.0 );

    /* b. Après seuil */
    TH_coef_Victoreen( element, infos->seuil - 1, &c2, &d2 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_6, c2, 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_7, d2, 0.0 );
  }

  if ( MODELE_MU0_MCMASTER == infos->methode ) { /* McMaster */
    double a1, b1, c1, d1, a2, b2, c2, d2;

    /* Seuils classés par ordre croissant... */
    /* a. Avant seuil */
    TH_coef_McMaster( element, infos->seuil, &a1, &b1, &c1, &d1 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_2, a1, 0.0 ); /* Avant seuil */
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_3, b1, 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_4, c1, 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_5, d1, 0.0 );

    /* b. Après seuil */
    TH_coef_McMaster( element, infos->seuil - 1, &a2, &b2, &c2, &d2 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_6, a2, 0.0 ); /* Après seuil */
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_7, b2, 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_8, c2, 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_9, d2, 0.0 );
  }

  if ( MODELE_MU0_EBEL == infos->methode ) { /* Ebel */
    double c[ 6 ];

    /* Il suffit du modèle de section efficace pour le seuil étudié seulement */
    TH_coef_Ebel( element, infos->seuil, FAUX, c, NULL );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_2, c[ 0 ], 0.0 ); /* Coefficients du polynome */
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_3, c[ 1 ], 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_4, c[ 2 ], 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_5, c[ 3 ], 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_6, c[ 4 ], 0.0 );
    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_7, c[ 5 ], 0.0 );
    /*    OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_8,    0.0, 0.0 ); */
  }
}

/* -------- Préparation de l'arbre de paramétrage de la normalisation -------- */

static void I_XAS_NORM_PreparerArbre(void)
{
  int retour;

  if ( VRAI == arbres[ NORM_EXAFS ].drapeaux.prepare ) return; /* Déjà initialisé ! */

  /* On initialise */
  retour = ARB_Initialiser( NORM_EXAFS, "menu_exafs.html#normaliser",
			    I_XAS_NORM_GererArbre, ( int (*)() ) NULL );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s,%s l.%u : ERREUR dans l'initialisation de l'arbre\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return;
  }

  /* On enjolive un peu... */
  OBJ_deplace( NORM_EXAFS, NEXF_H_SEUIL  , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NE_TXT_HAUTEUR, 0, +2, DEPLACE_PIXELS);

  OBJ_deplace( NORM_EXAFS, MU0_A_UTILISER, 0, +1, DEPLACE_PIXELS);

  OBJ_deplace( NORM_EXAFS, N_TXT_SUIL      , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, N_SUIL          , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NE_CHOIX_METHODE, 0, +4, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NE_TXT_METHODE  , 0, +4, DEPLACE_PIXELS);

  OBJ_deplace( NORM_EXAFS, NORM_CONSTANTE_1, 0, -500, DEPLACE_RELATIF);
  OBJ_deplace( NORM_EXAFS, NORM_TXT_CTE_1  , 0, -500, DEPLACE_RELATIF);

  OBJ_deplace( NORM_EXAFS, NORM_CTE_3    , 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_TXT_CTE_3, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_CTE_4    , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_TXT_CTE_4, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_CTE_5    , 0, +3, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_TXT_CTE_5, 0, +3, DEPLACE_PIXELS);

  OBJ_deplace( NORM_EXAFS, NORM_CTE_7    , 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_TXT_CTE_7, 0, +1, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_CTE_8    , 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_TXT_CTE_8, 0, +2, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_CTE_9    , 0, +3, DEPLACE_PIXELS);
  OBJ_deplace( NORM_EXAFS, NORM_TXT_CTE_9, 0, +3, DEPLACE_PIXELS);

#ifdef AVEC_TRADUCTION_ /* Traduction des textes de l'arbre */
  OBJ_TraduireTexte( NORM_EXAFS, NORM_TXT_EXF_NV, txt_norm_nouveau_graphe);
  OBJ_TraduireTexte( NORM_EXAFS, N_TXT_CALCUL, txt_norm_utiliser_calcule );
  OBJ_TraduireTexte( NORM_EXAFS, N_TXT_SMI_MPIRIK, txt_norm_utiliser_semiempirique );
  OBJ_TraduireTexte( NORM_EXAFS, NORM_PROPOSER, txt_norm_propose );
  OBJ_TraduireTexte( NORM_EXAFS, N_TXT_SUIL, txt_norm_seuil );
  OBJ_TraduireTexte( NORM_EXAFS, NORMALISER, txt_norm_normaliser );
  OBJ_TraduireTexte( NORM_EXAFS, NORM_ANNULER, txt_norm_annuler );
  OBJ_TraduireTexte( NORM_EXAFS, NE_TXT_HAUTEUR, txt_norm_hauteur_seuil );
  OBJ_TraduireTexte( NORM_EXAFS, N_TXT_PONDR_AVNT, txt_norm_travail_sur );
  OBJ_TraduireTexte( NORM_EXAFS, N_TXT_PONDR_APRS, txt_norm_exprime_en );
  OBJ_TraduireTexte( NORM_EXAFS, NE_TXT_METHODE, txt_norm_methode );
  OBJ_TraduireTexte( NORM_EXAFS, NE_TXT_TT_SERIE, txt_norm_toute_serie );
#endif

  /* On met des unités adaptées */
  OBJ_MettreUnite ( NORM_EXAFS, NORM_CONSTANTE_1, "eV" ); /* E0 */

  /* On met des valeurs pas absurdes dans les champs éditables */
  OBJ_MettreNombre( NORM_EXAFS, NORM_CONSTANTE_1, 0.0, 0.0 );
  OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_2, 0.0, 0.0 );
  OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_3, 0.0, 0.0 );
  OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_4, 0.0, 0.0 );
  OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_5, 0.0, 0.0 );
  OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_6, 0.0, 0.0 );
  OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_7, 0.0, 0.0 );
  OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_8, 0.0, 0.0 );
  OBJ_MettreNombre( NORM_EXAFS, NORM_CTE_9, 0.0, 0.0 );

  arbres[ NORM_EXAFS ].drapeaux.prepare = VRAI; /* Arbre a déjà été préparé */
}
