/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2002, Emmanuel CURIS

   Recalage d'un graphe sur une nouvelle grille des X
   Remarque : toutes les interpolations supposent un graphe tri en X (par ordre croissant)
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"
#include "lissage.h"
#include "operations.h"		/* Pour le calcul de la primitive ! */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int LIS_RecalerSpline_AUC(const Graphe *original, Graphe *resultat, int bords);

static int LIS_RecalerSplineC(const Graphe *original, Graphe *resultat, int bords);
static int LIS_RecalerSplineR(const Graphe *original, Graphe *resultat, int bords);

static int LIS_RecalerLineaireC(const Graphe *original, Graphe *resultat, int bords);
static int LIS_RecalerLineaireR(const Graphe *original, Graphe *resultat, int bords);

/* ~~~~~~~~ Ralisation ~~~~~~~~~~~~~~ */

/* -------- Interpolation d'un graphe sur une table de X indique -------- */

Graphe *GRF_CalerSurTable(const Graphe *graphe, unsigned long nbr_pts, double *x, int methode, int bords, int *code_erreur)
{
  int retour;
  Graphe *resultat;
  unsigned long i, nombre_points = 0;
  double *x_lu;

#ifdef _DEVELOPPE_
  if ( NULL == code_erreur ) {
    AVERTISSEMENT( "Pas de mmorisation de code d'erreur... [Graphe %p pour %lu points]", (void *) graphe, nbr_pts );
  }
#endif

  if ( NULL == graphe ) {
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_AUCUN_GRAPHE;
    return NULL;
  }

  /* On regarde combien de points de la table peuvent servir */
  for (i = 0, x_lu = x ; i < nbr_pts ; i++, x_lu++) {
    if ( (*x_lu >= graphe->x_min) && (*x_lu <= graphe->x_max) ) {
      nombre_points++;
    }
  }
  if (nombre_points < 2) {
    (void) fprintf(stderr, "%s,%s l.%u : trop peu de points compatibles avec le graphe %s\n",
		   __FILE__, __FUNCTION__, __LINE__, graphe->nom);
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_TROP_PEU_POINTS;
    return NULL;
  }

  /* On cre le graphe recal, dans la mme srie que l'original */
  OPTION_A_FAIRE( "Cration d'un graphe isol : mise en srie par les routines d'interface !" )
  resultat = GRF_nv_isole( nombre_points, graphe->type, graphe->genre_donnees );
  if (NULL == resultat) {
    (void) fprintf(stderr, "%s,%s l.%u : erreur dans GRF_nouveau()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_AUTRE_ERREUR;
    return NULL;
  }

  /* On remplit les X */
  for (i = 0, x_lu = x, nombre_points = 0 ; i < nbr_pts ; i++, x_lu++) {
    if ( (*x_lu >= graphe->x_min) && (*x_lu <= graphe->x_max) ) {
      *(resultat->x + nombre_points) = *x_lu;
      nombre_points++;
    }
  }
  GRF_bornes(resultat, DESORDONNE, NE_PAS_FAIRE);

  /* On copie les lgendes et units, inchanges */
  GRF_CopierAffichage( resultat, graphe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y );

  retour = LIS_RecalerGraphe(graphe, resultat, methode, bords);
  if ( code_erreur != NULL ) *code_erreur = retour;

  return resultat;
}

/* -------- Interpolation d'un graphe sur les X d'un autre graphe -------- */

Graphe *GRF_CalerSurGraphe(const Graphe *source, int methode, int bords, const Graphe *reference, int *code_erreur)
{
  Graphe *resultat;

#ifdef _DEVELOPPE_
  if ( NULL == code_erreur ) {
    AVERTISSEMENT( "Pas de mmorisation de code d'erreur... [Graphe %p  caler sur %p]", (void *) source, (void *) reference );
  }
#endif

  if ( (NULL == source) || (NULL == reference) ) {
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_AUCUN_GRAPHE;
    return NULL;
  }
  if ( (reference->x_min >= source->x_max) || (reference->x_max <= source->x_min) ) {
    (void) fprintf(stderr, "%s,%s l.%u : graphes disjoints (%s // base %s)\n",
		   __FILE__, __FUNCTION__, __LINE__, source->nom, reference->nom);
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_GRAPHES_DISJOINTS;
    return NULL;
  }

  resultat = GRF_CalerSurTable(source, reference->nombre_point, reference->x, methode, bords, code_erreur);
  if ( NULL == resultat ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur dans GRF_CalerSurTable()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    /* Le code est dj mis dans code_erreur, normalement... */
    return NULL;
  }

  /* Termin */
  return resultat;
}

/* -------- Interpolation d'un graphe sur une grille de X quitablement rpartis -------- */

Graphe *GRF_CalerSurGrille(const Graphe *source, int methode, int bords, double x_min, double x_max, unsigned long n_points, int *code_erreur)
{
  Graphe *resultat;
  double pas, vraie_borne;
  int retour;
  unsigned long i;

#ifdef _DEVELOPPE_
/*   printf( "GRF_CalerSurGrille : source = %p, nouveau = %d, methode = %d, x_min = %g, x_max = %g, n_points = %lu\n", */
/* 	  (void *) source, nouveau, methode, x_min, x_max, n_points ); */
  if ( NULL == code_erreur ) {
    AVERTISSEMENT( "Pas de mmorisation de code d'erreur... [Graphe %p sur grille [%g->%g], %lu points]", (void *) source, x_min, x_max, n_points );
  }
#endif

  if ( NULL == source ) {
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_AUCUN_GRAPHE;
    return NULL;
  }

  if ( x_min > x_max ) {
    double x = x_min ; x_min = x_max ; x_max = x;
  }

  if ( (x_min >= source->x_max) || (x_max <= source->x_min) ) {
    (void) fprintf(stderr, "GRF_CalerSurGrille : le graphe %s est hors de l'intervalle [%g, %g]\n",
		   source->nom, x_min, x_max);
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_GRAPHES_DISJOINTS;
    return NULL;
  }
  if (n_points < 3) {
    (void) fprintf(stderr, "GRF_CalerSurGrille : trop peu de points (%lu) / graphe %s, intervalle [%g, %g]\n",
		   n_points, source->nom, x_min, x_max);
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_TROP_PEU_POINTS;
    return NULL;
  }

  /* 1. On calcule le pas associ au nombre de points indiqu */
  pas = (x_max - x_min) / (n_points - 1);

  /* 2. On en dduit l'intervalle rel inclus dans le graphe interpol */
  vraie_borne = x_min;
  i = 0;
  while (vraie_borne < source->x_min) {
    i++;
    vraie_borne = x_min + pas * i;
  }
  n_points -= i; /* Nombre de points enlevs au dbut */
  x_min = vraie_borne;

  vraie_borne = x_max;
  i = 0;
  while (vraie_borne > source->x_max) {
    i++;
    vraie_borne = x_max - pas * i;
  }
  n_points -= i;
  x_max = vraie_borne;

  /* 3. Contrle qu'il reste des points */
  if (n_points < 2) {
    (void) fprintf(stderr, "GRF_CalerSurGrille : trop peu de points (%lu) aprs contrle / graphe %s, intervalle [%g, %g]\n",
		   n_points, source->nom, x_min, x_max);
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_TROP_PEU_POINTS;
    return NULL;
  }

  /* On cre le graphe recal */
  resultat = GRF_nv_GrilleXP( NULL, source->type, source->genre_donnees, source->nom, VRAI,
			      x_min, x_max, pas );
  if (NULL == resultat) {
    (void) fprintf(stderr, "GRF_CalerSurGrille : erreur dans GRF_nouveau()\n");
    if ( code_erreur != NULL ) *code_erreur = ERR_LIS_AUTRE_ERREUR;
    return NULL;
  }

  /* On copie les lgendes et units, inchanges */
  GRF_CopierAffichage( resultat, source, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_LEGENDE_Y | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y );

  /* Recalage effectif */
  retour = LIS_RecalerGraphe(source, resultat, methode, bords);
  if ( code_erreur != NULL ) *code_erreur = retour;

  return resultat;
}

/* -~-~-~-~-~-~-~-~ Ralisation concrte de l'interpolation -~-~-~-~-~-~-~-~ */

/* -------- Routine globale d'appel --------
    Calcule la valeur interpole des Y de l'original pour chacun des X du rsultat par la mthode indique
 */

int LIS_RecalerGraphe(const Graphe *original, Graphe *resultat, int methode, int bords)
{
  int retour = 0;

  if ( ( NULL == original ) || ( NULL == resultat ) ) {
    return ERR_LIS_AUCUN_GRAPHE;
  }

  /* On s'occupe des Y, en fonction de la mthode choisie */
  switch ( methode )
  {
  case INTERPOLE_LINEAIRE:
    if ( GRF_EstFonctionReelle( original ) ) {
      retour = LIS_RecalerLineaireR( original, resultat, bords );
    } else if ( GRF_FCT_COMPLEXE == original->genre_donnees ) {
      retour = LIS_RecalerLineaireC( original, resultat, bords );
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : interpolation linaire impossible pour ce type de graphes\n",
		     __FILE__, __FUNCTION__, __LINE__);
      retour = ERR_LIS_METHODE_INADAPTEE;
    }
    break;

  case INTERPOLE_PARABOLE:
    OPTION_A_FAIRE( "Interpolation parabolique" );
    break;

  case INTERPOLE_CUBIQUE:
    OPTION_A_FAIRE( "Interpolation cubique" );
    break;

  case INTERPOLE_SPLINE:
    if ( GRF_FCT_REELLE == original->genre_donnees ) {
      retour = LIS_RecalerSplineR( original, resultat, bords );
    } else if ( GRF_FCT_COMPLEXE == original->genre_donnees ) {
      retour = LIS_RecalerSplineC( original, resultat, bords );
    } else {
      (void) fprintf(stderr, "%s,%s l.%u : interpolation par spline impossible pour ce type de graphes\n",
		     __FILE__, __FUNCTION__, __LINE__);
      retour = ERR_LIS_METHODE_INADAPTEE;
    }
    break;

  case INTERPOLE_SPLINE_AUC:
    if ( GRF_EstFonctionReelle( original ) ) {
      retour = LIS_RecalerSpline_AUC( original, resultat, bords );
    } else {
      (void) fprintf(stderr, "%s, %s l.%u : Interpolation  aire constante impossible pour ce type de graphes\n",
		     __FILE__, __FUNCTION__, __LINE__);
      retour = ERR_LIS_METHODE_INADAPTEE;
    }
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : mthode d'interpolation inconnue [%d, pour %s]\n",
		   __FILE__, __FUNCTION__, __LINE__, methode, original->nom);
    retour = -10;
    break;
  }

  GRF_bornes( resultat, NE_PAS_FAIRE, DESORDONNE );
  /* Tout s'est bien pass */
  return retour;
}

/* -------- Lissage linaire, valeurs complexes -------- */

static int LIS_RecalerLineaireC(const Graphe *original, Graphe *resultat, int bords __attribute__((__unused__)))
{
  if ( ( NULL == original ) || ( NULL == resultat ) ) return -1;

  OPTION_A_FAIRE( "interpolation linaire / spectre complexe" );

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Lissage linaire, valeurs relles -------- */

static int LIS_RecalerLineaireR(const Graphe *original, Graphe *resultat, int bords)
{
  unsigned long i, i_depart = 0;
  double *x, *y;

  if ( (NULL == original) || (NULL == resultat) ) return -1;

  /* Le spectre proprement dit */
  for (i = 0, x = resultat->x, y = resultat->y ; i < resultat->nombre_point ; i++, x++, y++) {
    if ( *x < original->x_min ) {
      if ( ( INTERPOLE_BORDS_0_0 == bords ) || ( INTERPOLE_BORDS_0_Y == bords ) ) {
	*y = 0.0;
      } else if ( ( INTERPOLE_BORDS_Y_0 == bords ) || ( INTERPOLE_BORDS_Y_Y == bords ) ) {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Valeurs aux bords / dbut !\n",
		       __FILE__, __FUNCTION__, __LINE__);
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Traitement des bords, mthode inconnue [%d dbut]!\n",
		       __FILE__, __FUNCTION__, __LINE__, bords);
      }
    } else if ( *x > original->x_max ) {
      if ( ( INTERPOLE_BORDS_0_0 == bords ) || ( INTERPOLE_BORDS_Y_0 == bords ) ) {
	*y = 0.0;
      } else if ( ( INTERPOLE_BORDS_0_Y == bords ) || ( INTERPOLE_BORDS_Y_Y == bords ) ) {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Valeurs aux bords / fin !\n",
		       __FILE__, __FUNCTION__, __LINE__);
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Traitement des bords, mthode inconnue [%d fin] !\n",
		       __FILE__, __FUNCTION__, __LINE__, bords);
      }
    } else {
      *y = GRF_InterpoleLineaire(original, *x, &i_depart);
    }
  }

  /* ventuellement les barres d'erreur */
  if ( original->erreur_y != NULL ) {
    Graphe *erreurs;

    erreurs = GRF_creer_erreurs(resultat, AXE_Y);
    if (NULL == erreurs) {
      (void) fprintf(stderr, "LIS_RecalerLineaireR : impossible de crer les erreurs associes\n"); 
      return +1; /* Erreur "mineure" */
    }

    if ( original->erreur_x != NULL ) {
      printf("LIS_RecalerLineaireR :  FAIRE, estimation des erreurs interpoles / erreurs en X & en Y\n"); 
    } else {
      i_depart = 0;
      /* Attention, plus de X dans les erreurs : sont forcement egales a ceux du graphe ==>
         il faut lire les X dans le resultat sinon seg. fault ! [EC, mai 2011...] */
      for (i = 0, x = resultat->x, y = erreurs->y ; i < erreurs->nombre_point ; i++, x++, y++) {
	if ( *x < original->x_min ) {
	  if ( ( INTERPOLE_BORDS_0_0 == bords ) || ( INTERPOLE_BORDS_0_Y == bords ) ) {
	    *y = 0.0;
	  } else if ( ( INTERPOLE_BORDS_Y_0 == bords ) || ( INTERPOLE_BORDS_Y_Y == bords ) ) {
	    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Valeurs aux bords / dbut !\n",
			   __FILE__, __FUNCTION__, __LINE__);
	  } else {
	    (void) fprintf(stderr, "%s,%s l.%u : Traitement des bords, mthode inconnue [%d dbut]!\n",
			   __FILE__, __FUNCTION__, __LINE__, bords);
	  }
	} else if ( *x > original->x_max ) {
	  if ( ( INTERPOLE_BORDS_0_0 == bords ) || ( INTERPOLE_BORDS_Y_0 == bords ) ) {
	    *y = 0.0;
	  } else if ( ( INTERPOLE_BORDS_0_Y == bords ) || ( INTERPOLE_BORDS_Y_Y == bords ) ) {
	    (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Valeurs aux bords / fin !\n",
			   __FILE__, __FUNCTION__, __LINE__);
	  } else {
	    (void) fprintf(stderr, "%s,%s l.%u : Traitement des bords, mthode inconnue [%d fin] !\n",
			   __FILE__, __FUNCTION__, __LINE__, bords);
	  }
	} else {
	  *y = GRF_InterpoleLineaire_Erreur(original, *x, &i_depart);
	}
      }
    }
    GRF_bornes(erreurs, ORDONNE, DESORDONNE);
  } else if ( original->erreur_x != NULL ) {
    printf("LIS_RecalerLineaireR :  FAIRE, estimation des erreurs interpoles / erreurs en X seulement\n"); 
  }

  /* Tout s'est bien pass */
  return 0;
}

/* -------- Cas d'une spline,  valeurs complexes -------- */

static int LIS_RecalerSplineC(const Graphe *original, Graphe *resultat, int bords)
{
  Complexe *b, *c, *d, *y;
  double *x;
  int retour;
  unsigned long i;

  if ( (NULL == original) || (NULL == resultat) ) return -1;

  /* Prparation des coefficients */
  retour = CSPL_tt_prep(original, &b, &c, &d);
  if (retour != 0) {
    (void) LIS_ERREUR(retour, original->nom);
  }

  /* Boucle sur les points du graphe rsultat */
  for (i = 0, x = resultat->x, y = (Complexe *) (resultat->y) ; i < resultat->nombre_point ; i++, x++, y++) {
    if ( *x < original->x_min ) {
      if ( ( INTERPOLE_BORDS_0_0 == bords ) || ( INTERPOLE_BORDS_0_Y == bords ) ) {
	y->re = 0.0;
	y->im = 0.0;
      } else if ( ( INTERPOLE_BORDS_Y_0 == bords ) || ( INTERPOLE_BORDS_Y_Y == bords ) ) {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Valeurs aux bords / dbut !\n",
		       __FILE__, __FUNCTION__, __LINE__);
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Traitement des bords, mthode inconnue [%d dbut]!\n",
		       __FILE__, __FUNCTION__, __LINE__, bords);
      }
    } else if ( *x > original->x_max ) {
      if ( ( INTERPOLE_BORDS_0_0 == bords ) || ( INTERPOLE_BORDS_Y_0 == bords ) ) {
	y->re = 0.0;
	y->im = 0.0;
      } else if ( ( INTERPOLE_BORDS_0_Y == bords ) || ( INTERPOLE_BORDS_Y_Y == bords ) ) {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Valeurs aux bords / fin !\n",
		       __FILE__, __FUNCTION__, __LINE__);
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Traitement des bords, mthode inconnue [%d fin] !\n",
		       __FILE__, __FUNCTION__, __LINE__, bords);
      }
    } else {
      *y = CSPL_tt_calc(original, *x, b, c, d);
    }
  }

  /* Tout s'est bien pass */
  free(b); free(c); free(d);
  return 0;
}

/* -------- Cas d'une spline,  valeurs relles -------- */

static int LIS_RecalerSplineR(const Graphe *original, Graphe *resultat, int bords)
{
  double *b, *c, *d;
  double *x, *y;
  int retour;
  unsigned long i;

  if ( (NULL == original) || (NULL == resultat) ) return -1;

  /* Prparation des coefficients */
  retour = SPL_tt_prep(original, &b, &c, &d);
  if (retour != 0) {
    (void) LIS_ERREUR(retour, original->nom);
  }

  /* Boucle sur les points du graphe rsultat */
  for (i = 0, x = resultat->x, y = resultat->y ; i < resultat->nombre_point ; i++, x++, y++) {
    if ( *x < original->x_min ) {
      if ( ( INTERPOLE_BORDS_0_0 == bords ) || ( INTERPOLE_BORDS_0_Y == bords ) ) {
	*y = 0.0;
      } else if ( ( INTERPOLE_BORDS_Y_0 == bords ) || ( INTERPOLE_BORDS_Y_Y == bords ) ) {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Valeurs aux bords / dbut !\n",
		       __FILE__, __FUNCTION__, __LINE__);
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Traitement des bords, mthode inconnue [%d dbut]!\n",
		       __FILE__, __FUNCTION__, __LINE__, bords);
      }
    } else if ( *x > original->x_max ) {
      if ( ( INTERPOLE_BORDS_0_0 == bords ) || ( INTERPOLE_BORDS_Y_0 == bords ) ) {
	*y = 0.0;
      } else if ( ( INTERPOLE_BORDS_0_Y == bords ) || ( INTERPOLE_BORDS_Y_Y == bords ) ) {
	(void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- Valeurs aux bords / fin !\n",
		       __FILE__, __FUNCTION__, __LINE__);
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Traitement des bords, mthode inconnue [%d fin] !\n",
		       __FILE__, __FUNCTION__, __LINE__, bords);
      }
    } else {
      *y = SPL_tt_calc(original, *x, b, c, d);
    }
  }

  /* Tout s'est bien pass */
  free(b); free(c); free(d);
  return 0;
}

/* -------- Idem, en imposant la contrainte de conserver l'aire sous la courbe.
   On fait donc une interpolation de la primitive, que l'on drive ensuite
   -------- */

static int LIS_RecalerSpline_AUC(const Graphe *original, Graphe *resultat, int bords __attribute__((__unused__)) )
{
  Graphe *primitive;
  double *b, *c, *d, derivee;
  int retour;
  unsigned long i;
  union valeur constante;

  /* On calcule la primitive du graphe original. En graphe isol, car sera perdu ensuite ! */
  constante.reel = 0.0;
  primitive = GRF_Primitive( original, INTG_TRAPEZE, constante, 0.0 );
  if ( NULL == primitive ) {
    AVERTISSEMENT( "Impossible d'obtenir la primitive du graphe %s", original->nom );
    return ERR_LIS_AUCUN_GRAPHE;
  }

  /* On calcule les coefficients de spline pour la primitive */
  retour = SPL_tt_prep( primitive, &b, &c, &d );
  if ( retour != 0 ) {
    AVERTISSEMENT( "Impossible de calculer les coefficients de la spline pour la primitive de %s", original->nom );
    GRF_eff_isole( primitive ); primitive = NULL;
    free( b ); b = NULL;
    free( c ); c = NULL;
    free( d ); d = NULL;
    return retour;
  }

  /* On calcule,  l'aide de cette spline, la drive de la primitive en chaque point voulu */
  for ( i = 0 ; i < resultat->nombre_point ; i++ ) {
    (void) SPL_tt_derive( primitive, resultat->x[ i ], b, c, d, &derivee );
    resultat->y[ i ] = derivee;
  }

  /* Inutile de mettre les bornes : la fonction appelante s'en chargera... */

  /* On n'a plus besoin de la primitive */
  GRF_eff_isole( primitive ); primitive = NULL;

  /* Ni des tables de coefficients */
  free( b ); b = NULL;
  free( c ); c = NULL;
  free( d ); d = NULL;

  /* Termin sans erreur */
  return 0;
}
