/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2002, Emmanuel Curis

   Cration et gestion de modles en vue d'un ajustement

   Cration de modles (partie gnrique)
*/

#include <stdio.h>
#include <stdlib.h>

#include "systeme.h"

#include "feff.h"		/* Pour les modles feff */
#include "graphes.h"
#include "modeles.h"
#include "decomposition.h"	/* Modles de dcomposition linaire sur une base de spectres */

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

/* -------- Cration d'un nouveau modle --------\
 |  Renvoie le modle cr, NULL si erreur       |
 |                                               |
 | avant      = modle prcdent (facultatif)    |
 | nom        = le nom du modle                 |
 | type       = le type de modle                |
 | modele     = le spectre de rfrence          |
 | complement = VRAI, on rserve le complment   |
 \---------------------------------------------- */

ModeleAjustement *MODELE_CreerModele(ModeleAjustement *avant, const char *nom, int type,
				     struct graphe *grf_modele, boolean complement)
{
  ModeleAjustement *modele;

  /* Rservation du modle */
  modele = (ModeleAjustement *) malloc( sizeof(ModeleAjustement) );
  if ( NULL == modele ) {
    ERR_Memoire( "Structude ModeleAjustement", sizeof(ModeleAjustement) );
    return NULL;
  }

  /* Son nom */
  if ( NULL == nom )
    modele->nom_modele = strdup( "Sans nom" );
  else
    modele->nom_modele = strdup( nom );

  /* Initialisation */
  modele->modele = NULL;
  modele->residu = NULL;
  modele->nbr_variables = 0;
  modele->algorithme = ALGO_PSEUDOGRADIENT; /* Le seul actuellement disponible... */
  modele->erreur = -1.0; /* Absurde : indique qu'aucun ajustement n'a encore t fait */
  modele->iteration = 0;
  modele->max_iteration = 100000; /* Arbitraire, correspond  l'absence de limitation...*/
  modele->condition_arret = ARRET_AJUSTEMENT_ERREUR | ARRET_AJUSTEMENT_DELTA_ERREUR;
  modele->err_min = 5;  /* 10^-5 considr comme une erreur ngligeable */
  modele->derr_min = 8; /* 10^-8 considr comme une erreur variant trop peu */

  modele->drapeaux.pondere = VRAI;
  modele->drapeaux.contraintes = FAUX;
  modele->drapeaux.isole = FAUX;
  modele->drapeaux.en_cours = FAUX;
  modele->drapeaux.preparation = FAUX;
  modele->drapeaux.arret_force = FAUX;

  /* Graphe exprimental sert pour les X */
  modele->reference = grf_modele;
  modele->experience = grf_modele;
  if ( NULL == grf_modele ) {
    modele->x_min = 0.0;
    modele->x_max = 0.0;
  } else {
    modele->x_min = grf_modele->x_min;
    modele->x_max = grf_modele->x_max;

    if ( NULL == grf_modele->erreur_y ) {
      /* Pas d'incertitude en Y : ajustement pondr impossible */
      modele->drapeaux.pondere = FAUX;
    }
  }

  /* Les informations complmentaires, typiques du modle */
  modele->type = type;
  if ( VRAI == complement ) {
    switch (type)
    {
    case MODELE_MULTI_MODELES:
      OPTION_A_FAIRE( "Modle multiple" );
      break;

    case MODELE_CL_MODELES:
      OPTION_A_FAIRE( "Combinaison linaire de modles" );
      break;

    case MODELE_EXAFS_CHEMINS: /* Modle de type FEFF */
      modele->infos_modele.feff = FEFF_CreerModele();
      if ( NULL == modele->infos_modele.feff ) {
	free( modele ); modele = NULL;
	return NULL;
      }
      break;

    case MODELE_EXAFS_GNXAS: /* Modle de type GnXAS */
      OPTION_A_FAIRE( "Modle pour GnXAS" );
      break;

    case MODELE_SEUIL_BASIQUE:
      OPTION_A_FAIRE( "Modle simple de seuil" );
      break;

    case MODELE_FLUOX:
      OPTION_A_FAIRE( "Modle de spectre de fluorescence X" );
      break;

    case MODELE_DECOMPOSITION:
      modele->infos_modele.decomposition = MODELE_NouveauDecomposition();
      if ( NULL == modele->infos_modele.decomposition ) {
	(void) fprintf( stderr, "%s,%s l.%u : Impossible de rserver pour le modle de dcomposition !\n",
			__FILE__, __FUNCTION__, __LINE__ );
	free( modele ); modele = NULL;
	return NULL;
      }
      break;

    case MODELE_GENERIQUE:
      modele->infos_modele.generique = MODELE_NouveauGenerique( NULL );
      if ( NULL == modele->infos_modele.generique ) {
	(void) fprintf( stderr, "%s,%s l.%u : Impossible de rserver pour le modle gnrique !\n",
		        __FILE__, __FUNCTION__, __LINE__ );
	free( modele ); modele = NULL;
	return NULL;
      }
      break;
      
    default:
      AVERTISSEMENT( "Type de modle inconnu - %d", type );
    }
  } else { /* C'est la routine appelante qui s'en occupera */
    modele->infos_modele.feff = NULL;
  }

  /* Aucune information sur les variables */
  modele->infos_variables.pseudogradient = NULL;
  modele->covariance = NULL; /* Aucune matrice de covariance ! */

  /* Chanage ventuel - en fin car plus facile d'effacer avant */
  if ( NULL != avant ) {
    modele->suivant = avant->suivant;
    avant->suivant = modele;
    modele->drapeaux.actif = FAUX;
  } else {
    modele->suivant = NULL;
    modele->drapeaux.actif = VRAI;
  }

  return modele;
}

/* -------- Donner un nom  un modle -------- */

int MODELE_Nommer(ModeleAjustement *modele, const char *nom)
{
  if (NULL == modele) return -1;
  if (NULL == nom) return -2;

  free( modele->nom_modele );
  modele->nom_modele = strdup( nom );
  if ( NULL == modele->nom_modele ) {
    ERR_Memoire( "Nom d'un mod"e_GRAVE"le", strlen( nom ) + 1 );
    return -3;
  }

  /* Tout s'est bien pass */
  return 0;
}
