/* LASE - Logiciel d'Analyse des Spectres Expérimentaux
   (C) 1996-2005, Emmanuel Curis

   Création et gestion de modèles en vue d'un ajustement

   Modèles d'ajustement génériques
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "systeme.h"

#include "modeles.h"
#include "graphes.h"		/* Pour les types de modèles = genres de graphes ! */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int MODELE_Generique_Bornes(struct modele_generique *modele);

/* ~~~~~~~~ Réalisation pratique ~~~~~~~~ */

/* -------- Créer un nouveau modèle générique -------- */

struct modele_generique *MODELE_NouveauGenerique(const char *nom_modele)
{
  struct modele_generique *modele = NULL;

  /* On réserve la place nécessaire */
  modele = (struct modele_generique *) malloc( sizeof(struct modele_generique ) );
  if ( NULL == modele ) {
    (void) fprintf(stderr, "%s,%s l.%u : Mémoire insuffisante pour réserver un modèle\n",
		   __FILE__, __FUNCTION__, __LINE__);
  }

  /* Initialisation */
  modele->type = GRF_FCT_REELLE;

  if ( NULL == nom_modele ) {
    modele->nom = NULL;
  } else {
    modele->nom = strdup( nom_modele );
  }

  modele->X.unite          = NULL;
  modele->X.type           = NULL;
  modele->X.formule        = NULL;
  modele->X.formule_erreur = NULL;

  modele->Y.unite          = NULL;
  modele->Y.type           = NULL;
  modele->Y.formule        = NULL;
  modele->Y.formule_erreur = NULL;

  modele->nombre_parametres = 0;
  modele->parametres = NULL;

  modele->arbre_formule = NULL;

  /* Terminé : on renvoie le modèle */
  return modele;
}

/* -------- Ajouter des paramètres à un modèle générique -------- */

int MODELE_Generique_AjouterParametres(struct modele_generique *modele, unsigned long nbr_parametres)
{
  struct parametre *nouveaux_parametres = NULL, *parametre;
  unsigned long i, nombre_total;

  if ( ( NULL == modele ) || ( 0 == nbr_parametres ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Rien à faire...\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return 0;
  }

  /* Combien de paramètres y aura-t-il finalement ? */
  nombre_total = modele->nombre_parametres + nbr_parametres;

  /* On essaye de réserver la nouvelle place mémoire */
  nouveaux_parametres = ( struct parametre * ) realloc( modele->parametres, nombre_total * sizeof( struct parametre ) );
  if ( NULL == nouveaux_parametres ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible d'allouer pour %lu paramètres en plus [soit %lu au total]\n",
		   __FILE__, __FUNCTION__, __LINE__, nbr_parametres, nombre_total );
    /* Aucun changement n'est fait ! */
    return -1;
  }

  /* On initialise les nouveaux paramètres */
  for ( i = 0, parametre = nouveaux_parametres + modele->nombre_parametres ; i < nbr_parametres ; i++, parametre++ ) {
    parametre->symbole = NULL;
    parametre->unite = NULL;
    parametre->nom = NULL;
    parametre->derivee = NULL;
    parametre->erreur = 0.0;

    parametre->variable.valeur_initiale = 0.0;
    parametre->variable.min = 0.0;
    parametre->variable.max = 0.0;
    parametre->variable.delta = 0.0;
    parametre->variable.type = VAR_LIBRE;
    parametre->variable.contenu.valeur = 0.0;

    parametre->arbre_derivee = NULL;
  }

  /* On installe ces nouvelles informations */
  modele->parametres = nouveaux_parametres;
  modele->nombre_parametres = nombre_total;

  /* Terminé sans erreur */
  return 0;
}

/* -------- Libérer les informations associées à un modèle générique -------- */

int MODELE_Generique_Liberer(struct modele_generique *modele)
{
  unsigned long i;

  if ( NULL == modele ) return +1;

  /* On libère les chaînes */
  free( modele->nom ); modele->nom = NULL;

  free( modele->X.unite ); modele->X.unite = NULL;
  free( modele->X.type ); modele->X.type = NULL;
  free( modele->X.formule ); modele->X.formule = NULL;
  free( modele->X.formule_erreur ); modele->X.formule_erreur = NULL;

  free( modele->Y.unite ); modele->Y.unite = NULL;
  free( modele->Y.type ); modele->Y.type = NULL;
  free( modele->Y.formule ); modele->Y.formule = NULL;
  free( modele->Y.formule_erreur ); modele->Y.formule_erreur = NULL;

  /* On libère les paramètres */
  for ( i = 0 ; i < modele->nombre_parametres ; i++ ) {
    free( modele->parametres[ i ].symbole ); modele->parametres[ i ].symbole = NULL;
    free( modele->parametres[ i ].unite ); modele->parametres[ i ].unite = NULL;
    free( modele->parametres[ i ].nom ); modele->parametres[ i ].nom = NULL;
    free( modele->parametres[ i ].derivee ); modele->parametres[ i ].derivee = NULL;
  }
  free( modele->parametres ); modele->parametres = NULL;

  /* On libère le modèle lui-même */
  free( modele );

  /* Tout s'est bien passé */
  return 0;
}


/* -------- Contrôler que le modèle est utilisable -------- */

int MODELE_Generique_Controler(struct modele_generique *modele)
{
  unsigned long i;
  Variable *variable;

  if ( NULL == modele ) return -1;

  for ( i = 0 ; i < modele->nombre_parametres ; i++ ) {
    variable = &( modele->parametres[ i ].variable );

    if ( variable->type != VAR_LIBRE ) continue; /* Variable qui sera ignorée ! */

    if ( variable->min == variable->max ) {
      (void) fprintf(stderr, "%s,%s l.%u : Variable avec des bornes identiques : constante !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      OPTION_A_FAIRE( "Boîte d'alerte propre" );
      variable->type = VAR_CONSTANTE;
      continue;
    }

    if ( ( variable->valeur_initiale < variable->min ) ||
	 ( variable->valeur_initiale > variable->max ) ) {
      (void) fprintf(stderr, "%s,%s l.%u : Variable avec une valeur initiale hors-bornes : corrigé !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      OPTION_A_FAIRE( "Boîte d'alerte propre" );
      variable->valeur_initiale = ( variable->min + variable->max ) / 2.0;
    }
  }

  /* Pas de problème sérieux détecté */
  return 0;
}

/* -------- Routines pour l'ajustement par un modèle générique -------- */

int MODELE_Generique_PreparerAjustement(ModeleAjustement *modele)
{
  unsigned long nombre_parametres = 0, i;
  struct modele_generique *modele_generique;
  struct parametre *parametre;

  if ( NULL == modele ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de modèle\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( modele->type != MODELE_GENERIQUE ) {
    (void) fprintf(stderr, "%s,%s l.%u : Modèle de type incorrect [%d, seul MODELE_GENERIQUE=%d autorisé]\n",
		   __FILE__, __FUNCTION__, __LINE__, modele->type, MODELE_GENERIQUE );
    return -2;
  }

  modele_generique = modele->infos_modele.generique;
  if ( NULL == modele_generique ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de paramètres de modèle générique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -1;
  }
  if ( NULL == modele_generique->parametres ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas de table de paramètres !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -50;
  }

  /* On construit l'arbre de la formule Y=f(X) */
  if ( modele_generique->arbre_formule != NULL ) {
    effacer_arbre( modele_generique->arbre_formule );
    modele_generique->arbre_formule = NULL;
  }
  modele_generique->arbre_formule = CF_ConstruireArbre( modele_generique->Y.formule,
							modele_generique->nombre_parametres, modele_generique->parametres );
  if ( NULL == modele_generique->arbre_formule ) {
    (void) fprintf(stderr, "%s,%s l.%u : Erreur en construisant l'arbre Y=f(X)\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -100;
  }

  /* On compte le nombre de paramètres ajustables */
  for ( i = 0, parametre = modele_generique->parametres ; i < modele_generique->nombre_parametres ; i++, parametre++ ) {
    /* On efface un éventuel ancien arbre restant [ne devrait pas arriver]  */
    if ( parametre->arbre_derivee != NULL ) {
      effacer_arbre( parametre->arbre_derivee );
      parametre->arbre_derivee = NULL;
    }

    if ( VAR_LIBRE == parametre->variable.type ) {
      nombre_parametres++;	/* Paramètre réellement ajustable */

      /* On construit l'arbre pour la dérivée chaque paramètre*/
      parametre->arbre_derivee = CF_ConstruireArbre( parametre->derivee,
						     modele_generique->nombre_parametres, modele_generique->parametres );
      if ( NULL == parametre->arbre_derivee ) {
	(void) fprintf(stderr, "%s,%s l.%u : Impossible de construire l'arbre pour la dérivée, paramètre %s\n",
		       __FILE__, __FUNCTION__, __LINE__, parametre->nom );
	return -150;
      }
    } else {
    }
  }

  /* On en déduit le nombre de variables ajustées */
  modele->nbr_variables = nombre_parametres;
 
  /* Terminé */
  return 0;
}

/* -------- Initialisation pour un algorithme de type pseudo-gradient -------- */

int MODELE_Generique_PGInitVariables(struct modele_ajuste *modele, unsigned int prem_variable)
{
  PG_variable *var;
  struct modele_generique *generique;
  unsigned long i;

  if (modele == NULL) return -1; /* Normalement, ne devrait jamais arriver, mais bon... */
  if (prem_variable >= modele->nbr_variables) {
    (void) fprintf(stderr, "%s,%s l.%u : incohérence pour prem_variable (%u, mais %u variables)\n",
		   __FILE__, __FUNCTION__, __LINE__, prem_variable, modele->nbr_variables);
    return -10;
  }

  var = modele->infos_variables.pseudogradient + prem_variable;
  if ( NULL == var ) {
    (void) fprintf(stderr, "%s,%s l.%u : modèle sans variables\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  generique = modele->infos_modele.generique;
  if ( NULL == generique ) {
    (void) fprintf(stderr, "%s,%s l.%u : modèle sans informations sur le modèle générique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -2;
  }

  /* On initialise les variables */
  for ( i = 0 ; i < generique->nombre_parametres ; i++ ) {
    if ( VAR_LIBRE == generique->parametres[ i ].variable.type ) {
      var->var  = &( generique->parametres[ i ].variable );
      var->type = 0; 		/* Aucun type spécial : pas besoin */
      var->origine.parametre = generique->parametres + i;
      var->modele = modele;
      var++;
    }
  }

  return 0; /* Tout s'est bien passé */
}

/* -------- Calcul du graphe modèle (théorique) -------- */

int MODELE_Generique_CalculerModele(ModeleAjustement *modele)
{
  int retour = 0;
  double y_calcule;
  unsigned long i;

  if ( NULL == modele->infos_modele.generique->arbre_formule ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'arbre créé pour calculer la formule !\n",
		   __FILE__, __FUNCTION__, __LINE__);

    for ( i = 0 ; i < modele->modele->nombre_point ; i++ ) {
      modele->modele->y[ i ] = 0.0;
    }

    return +1;
  }

  for ( i = 0 ; i < modele->modele->nombre_point ; i++ ) {
    /* Préparation de la variable en X [et I si jamais...] */
    Var_I.valeur_initiale = (double) i;
    Var_I.contenu.valeur  = (double) i;

    Var_X.valeur_initiale = modele->modele->x[ i ];
    Var_X.contenu.valeur  = modele->modele->x[ i ];

    /* Calcul de la valeur résultat */
    if ( FAUX == calcule_arbre( modele->infos_modele.generique->arbre_formule, &y_calcule ) ) {
      (void) fprintf( stderr, "%s, %s l.%u : Erreur de calcul, X=%g [i=%lu] ==> Y = 0.0\n",
		      __FILE__, __FUNCTION__, __LINE__, Var_X.valeur_initiale, i );
      y_calcule = 0.0;
      retour = +5;
    }

    /* Mémorisation */
    modele->modele->y[ i ] = y_calcule;
  }

  /* On recalcule les bornes, pour avoir un affichage correct */
  GRF_bornes( modele->modele, NE_PAS_FAIRE, DESORDONNE ); /* Bornes en Y ont changé... pas en X */

  /* Tout s'est bien passé */
  return retour;
}

/* -------- Calcul des incréments, pseudo-gradient -----------------------------\
 |   Renvoie 0 si tout s'est bien passé, <0 sinon                               |
 |                                                                              |
 | incrément I = 0.5 * ( dS/dvariable ) / Somme_sur_k ( dmodèle/dvariable )²    |
 |                                                                              |
 |               Somme_sur_k [ residu(k) * (dmodèle / dvariable)(k) / sigma_i²] |
 |             = -------------------------------------------------------------- |
 |                  Somme_sur_k [ ( (dmodèle / dvariable)(k) )² / sigma_i²]     |
 |                                                                              |
 | Suppose implicitement que var->x_max > var->x_min et                         |
 |                       que toutes les variables soient à faire varier         |
 \----------------------------------------------------------------------------- */

int MODELE_Generique_IncrementsPG(ModeleAjustement *modele)
{
  struct pg_variable *var, *variables;
  struct parametre *parametre;
  unsigned int num_var; /* Index de la variable utilisée */
  unsigned long i; /* Index sur les x */
  register double *x, *y_residu, *y_modele, *y_erreur = NULL;
  int retour;

  /* Préparation */
  variables = modele->infos_variables.pseudogradient;

  if ( VRAI == modele->drapeaux.pondere )
    y_erreur = modele->reference->erreur_y->y;

  /* Réinitialisation */
  for ( num_var = 0, var = variables ; num_var < modele->nbr_variables ; num_var++, var++ ) {
    var->den = 0.0;
    var->num = 0.0;
  }

  /* Ici, boucle sur les points. On utilise une boucle sur les variables à l'intérieur */
  for ( i = 0, x = modele->modele->x, y_modele = modele->modele->y, y_residu = modele->residu->y ;
	i < modele->modele->nombre_point ;
	i++, x++, y_modele++, y_residu++ ) {
    double sigma2 = 0.0;

    /* Correction : pondération par l'erreur */
    if ( y_erreur != NULL ) {
      sigma2 = *y_erreur * *y_erreur;
      y_erreur++;
    }

    /* Préparation de la variable en X [et I si jamais...] */
    Var_I.valeur_initiale = (double) i;
    Var_I.contenu.valeur  = (double) i;

    Var_X.valeur_initiale = *x;
    Var_X.contenu.valeur  = *x;

    for ( num_var = 0, var = variables ; num_var < modele->nbr_variables ; num_var++, var++ ) {
      double derivee = 0.0;

      /* Quel est le paramètre associé à cette variable ? */
      parametre = var->origine.parametre;

      /* On utilise la fonction dérivée pour le calcul */
      retour = calcule_arbre( parametre->arbre_derivee, &derivee );
      if ( FAUX == retour ) {
	(void) fprintf(stderr, "%s,%s l.%u : Erreur en voulant calculer la dérivée [X=%g]\n",
		       __FILE__, __FUNCTION__, __LINE__, *x);
	derivee = 0.0;
      }

      if ( sigma2 > 0.0 ) { /* On utilise les vraies erreurs ? */
	var->num += derivee * *y_residu / sigma2;
	var->den += derivee * derivee / sigma2;
      } else { /* Pas d'erreur ou erreur constante -> ne joue pas sur l'incrément car se simplifie
		  (même en tenant compte de la pondération en k^n) */
	var->num += derivee * *y_residu;
	var->den += derivee * derivee;
      }
    }
  }

  for (num_var = 0, var = variables ; num_var < modele->nbr_variables ; num_var++, var++) {
    if (var->den == 0.0) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "%s,%s l.%u : variable constante, dérivée nulle...\n  type : %d - N = %f, D = %f",
		     __FILE__, __FUNCTION__, __LINE__, var->type, var->num, var->den );
#endif
      return -1;
    } else {
      var->increment = var->num / var->den;
    }
  }

  return 0; /* Tout s'est bien passé */
}

/* -------- Affichage des résultats -------- */

int MODELE_Generique_AfficherResultats(struct modele_generique *modele, boolean historique)
{
  unsigned long i;
  char *texte = NULL;
  struct parametre *parametre;
  BI_Boite_Alerte message = { ERR_ICONE_INFO, "Résultats de l'ajustement", NULL, 1, 1, 1, { "OK", NULL, NULL, NULL, NULL } };

  if ( NULL == modele ) return -1;

  /* On prépare le texte */
  texte = strdup( "Résultats de l'ajustement" );
  for ( i = 0, parametre = modele->parametres ; i < modele->nombre_parametres ; i++, parametre++ ) {
    texte = TXT_genere_chaine( "%s|%s = %*f", texte, parametre->symbole, VAR_LireValeur( &(parametre->variable), VARIABLE_LIRE_VALEUR_AJUSTEE ), parametre->unite );
  }

  /* On prépare la boîte de dialogue */
  message.texte = texte;

  /* On affiche la boîte de dialogue */
  (void) BI_ERR_alerte( principale, &message, VRAI );

  /* On libère le texte */
  free( texte ); texte = NULL; message.texte = NULL;

  if ( VRAI == historique ) {
    OPTION_A_FAIRE( "Sauvegarde dans l'historique" );
  }

  /* On a terminé, sans souci */
  return 0;
}

/* -------- Validation des résultats -------- */

int MODELE_Generique_Verifier(struct modele_generique *modele, unsigned int type_validation)
{
  int retour = 0;

  if ( NULL == modele ) return -1;

  switch ( type_validation )
  {
  case MODELE_VAL_BORNES:
    retour = MODELE_Generique_Bornes( modele );
    break;

  case MODELE_VAL_PHYSIQUE:
    (void) fprintf(stderr, "%s,%s l.%u : Pas de contrôle possible du sens physique sur un modèle générique !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    break;

  case MODELE_VAL_STATS:
  case MODELE_VAL_MONTECARLO:
    AVERTISSEMENT( "Méthode inadéquate [%d], utiliser MODELE_ControlerValidite(..., %d)", type_validation, type_validation );
    return 0;

  default:
    AVERTISSEMENT( "Méthode inconnue [%d]", type_validation );
    return 0;
  }

  return retour;
}

/* -------- Contrôler que les variables n'ont pas atteint leurs bornes ! -------- */

static int MODELE_Generique_Bornes(struct modele_generique *modele)
{
  int retour = 0;
  unsigned int i, nombre_atteints = 0;

  for ( i = 0 ; i < modele->nombre_parametres ; i++ ) {
    retour = VAR_ControleBornes( &(modele->parametres[ i ].variable) );
    if ( retour != 0 ) {
      nombre_atteints++;
      /* Afficher un message ! */
    }
  }

  /* Tout s'est bien passé */
  return 0;
}
