/* LASE - Logiciel d'Analyse des Spectres Exprimentaux
   (C) 1996-2002, Emmanuel Curis

   Cration et gestion de modles en vue d'un ajustement

   Ajustement par la mthode du pseudo-gradient (algorithme inspir d'EDA)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

/* #include "c_formel.h"		/\* Inclus par modeles.h *\/ */
#include "graphes.h"
#include "modeles.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

static double   (*fct_CalculePoint )(struct modele_ajuste *modele, double x);
static Complexe (*fct_CalculePointC)(struct modele_ajuste *modele, double x);

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int CalculeModelePG    (struct modele_ajuste *modele);
static int CalculeIncrementsPG(struct modele_ajuste *modele);

/* ~~~~~~~~ Ralisation concrte ~~~~~ */

/* -------- Interface visiblev de l'extrieur ---------------------------------------------------\
 |  Renvoie la raison pour laquelle l'ajustement est fini                                        |
 |                                                                                               |
 | modele  : le modle  ajuster                                                                 |
 | fenetre : la fentre  redessiner  chaque itration                                          |
 | calcule_point      : fonction  appeler pour calculer le modle en un point                   |
 | calcule_modele     : la fonction  appeler pour calculer le graphe du modle thorique        |
 | calcule_increments : la fonction  appeler pour calculer les incrments de chaque variable    |
 |              (NULL : utilise la fonction interne avec estimation numrique des drives)      |
 | affichage          : routine  appeler aprs chaque itration, pour affichages intermdiaires |
 \---------------------------------------------------------------------------------------------- */

int MODELE_AjustementPG(struct modele_ajuste *modele, Fenetre *fenetre, 
			double (*calcule_point)(struct modele_ajuste *modele, double x),
			int (*calcule_modele)(struct modele_ajuste *modele),
			int (*calcule_increments)(struct modele_ajuste *modele),
			void (*affichage)(struct modele_ajuste *modele))
{
  struct pg_variable *variables, *var;
  unsigned long max_iterations = 100000; /* De toute faon, plus serait trop long... */
  double err_min = 0.0,  delta_err_min = 0.0;
#ifdef _DEVELOPPE_
/*   unsigned int i; */
#endif

  if (NULL == modele) return -1;
  if ( modele->algorithme != ALGO_PSEUDOGRADIENT ) {
    (void) fprintf(stderr, "%s,%s l.%u : le modle n'a pas t prvu pour cet algorithme ! [%u, attendu ALGO_PSEUDOGRADIENT=%u]\n",
		   __FILE__, __FUNCTION__, __LINE__, modele->algorithme, ALGO_PSEUDOGRADIENT );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -1;
  }

  if ( ( int (*)(struct modele_ajuste *) ) NULL == calcule_modele ) {
    (void) fprintf(stderr, "%s,%s l.%u : Modle calcul par appel de la fonction modle en chaque point\n",
		   __FILE__, __FUNCTION__, __LINE__);
    calcule_modele = CalculeModelePG;
    if ( ( double (*)(struct modele_ajuste *, double x) ) NULL == calcule_point ) {
      (void) fprintf(stderr, "%s,%s l.%u : pas de fonction pour calculer le modle !\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return -2;
    }
    fct_CalculePoint = calcule_point;
  }
  if ( ( int (*)(struct modele_ajuste *) ) NULL == calcule_increments ) {
    printf("MODELE_AjustementPG : incrments obtenus par estimation numrique des drives\n");
    calcule_increments = CalculeIncrementsPG;
    if ( ( double (*)(struct modele_ajuste *, double x) ) NULL == calcule_point ) {
      (void) fprintf(stderr, "MODELE_AjustementPG : pas de fonction pour calculer le modle !\n");
      return -2;
    }
    fct_CalculePoint = calcule_point;
  }

  /* Dtermination des conditions d'arrt */
  if ( modele->condition_arret & ARRET_AJUSTEMENT_ERREUR )
    err_min = pow( 10.0, (double) -modele->err_min );
  if ( modele->condition_arret & ARRET_AJUSTEMENT_DELTA_ERREUR )
    delta_err_min = pow( 10.0, (double) -modele->derr_min );
  if ( modele->condition_arret & ARRET_AJUSTEMENT_ITERATIONS )
    max_iterations = modele->max_iteration;

#ifdef _DEVELOPPE_
/*   (void) printf("--- Valeurs initiales des variables ---\n"); */
/*   for (i = 0 ; i < modele->nbr_variables ; i++) { */
/*     (void) printf("V(%u) : %g / %g -- ", i, */
/* 		  modele->infos_variables.pseudogradient[i].var->valeur_initiale, modele->infos_variables.pseudogradient[i].var->contenu.valeur); */
/*   } */
/*   printf("\n"); */
/*   (void) printf("Itration initiale : %u (max : %lu)\n", modele->iteration, max_iterations); */
#endif

  /* Prparation de l'ajustement
     1. Calcul du modle avec les valeurs initiales des variables */
  calcule_modele( modele );

  /* 2. Erreur & rsidu correspondants */
  modele->erreur = MODELE_CalculerResidu( modele );

  FEN_redessin( fenetre, BI_RDS_UTIL ); /* Affichage du modle initial */

  /* Contrle des variables ajustables */
  if ( modele->nbr_variables == 0 )
    return 0; /* Rien de spcial  faire ! */

  variables = modele->infos_variables.pseudogradient;
  if ( NULL == variables ) {
    AVERTISSEMENT( "Variables mal initialises ! [variables = %p]", variables );
    return -4;
  }

  HIST_Titre( 5, "Erreur interm"e_AIGU"diaire", "ERR" );

  /* On dmarre l'ajustement proprement dit */
#ifdef _DEVELOPPE_
  printf( " ==> %s <== Pret a ajuster... [ponderation : %d]\n",
	  __FUNCTION__, modele->drapeaux.pondere );
#endif
  while ( modele->iteration < max_iterations ) { /* On doit continuer les itrations */
    double pas, pas_max = 1.0;
    double ex_erreur, nv_erreur;
    int retour;
    unsigned int i;

    retour = calcule_increments( modele ); /* Calcul des incrments pour chacune des variables */
    if ( retour != 0 ) { /* Erreur survenue dans le calcul des incrments */
      AVERTISSEMENT( "Erreur dans le calcul des incrments [%d]", retour );
      return ARRET_AJUSTEMENT_FORCE;
    }

#ifdef _DEVELOPPE_
    printf("Incrments / itration %u : \n", modele->iteration);
    for ( i = 0, var = variables ; i < modele->nbr_variables ; i++, var++ )
      printf( "var %u : %g ; ", i, var->increment );
    printf( "\n" );
#endif

    /* On enregistre les anciennes valeurs */
    for ( i = 0, var = variables ; i < modele->nbr_variables ; i++, var++ )
      var->ex_valeur = var->var->contenu.valeur;

    /* On en dduit le pas pour chacune des variables */
    do {
      for (i = 0, var = variables ; i < modele->nbr_variables ; i++, var++) {
	double nv_valeur;

	pas = pas_max;
	nv_valeur = var->ex_valeur + pas * var->increment;
	/* On vrifie que l'on est bien dans l'intervalle de variations autoris */
	while ( (nv_valeur < var->var->min) || (nv_valeur > var->var->max) ) {
	  pas /= 2.0;
	  nv_valeur = var->ex_valeur + pas * var->increment;
	}

	var->var->contenu.valeur = nv_valeur;
      }

      /* Le calcul du spectre - suivant le type de modle utilis */
      retour = calcule_modele(modele);
      if (retour != 0) {
	(void) fprintf(stderr, "MODELE_AjustementPG : problme dans le calcul du modle\n");
	return -50;
      }
      nv_erreur = MODELE_CalculerResidu(modele);

#ifdef _DEVELOPPE_
      /*      printf("Itration %d : erreur de dpart %g, erreur trouve %g\n",
	      infos->etape, infos->erreur, nv_erreur); */
#endif
      BI_AttendEvenement( BI_EVENEMENT_TIMER, /* vnements attendus */
			  0, 0, 0, 0, 0,     /* rectangle 2 : aucun */
			  0,
			  NULL, NULL, NULL, NULL, NULL, NULL );
      if (modele->reference == NULL) { /* Au cours du traitement des vnements, graphe effac... */
	(void) fprintf(stderr, "MODELE_AjustementPG : Graphe de rfrence effac - Arrt d'urgence de l'ajustement\n");
	return -50;
      }
      if ( VRAI == modele->drapeaux.arret_force ) { /* Erreur survenue dans le calcul des incrments */
	(void) fprintf(stderr, "MODELE_AjustementPG : ajustement interrompu par l'utilisateur\n");
	return ARRET_AJUSTEMENT_FORCE;
      }

      pas_max /= 2.0;
    } while ( (pas_max > 2.0e-5) && (nv_erreur > modele->erreur) );

    if ( nv_erreur > modele->erreur ) { /* On n'a pas russi  minimiser */
#ifdef _DEVELOPPE_
      /*      printf("Itration %d : impossible de diminuer l'erreur (%f) - ajustement arrt\n",
	      infos->etape, infos->erreur); */
#endif

      /* On remet les anciennes valeurs, qui donnaient un meilleur rsultat */
      for (i = 0, var = variables ; i < modele->nbr_variables ; i++, var++)
	var->var->contenu.valeur = var->ex_valeur;

      /* On recalcule le spectre ajust */
      retour = calcule_modele(modele);
      if (retour != 0) {
	(void) fprintf(stderr, "MODELE_AjustementPG : problme survenu dans le calcul du modle\n");
	return -50;
      }
      modele->erreur = MODELE_CalculerResidu(modele);

      /* On raffiche, si ncessaire, les valeurs correctes */
      if ( affichage != NULL )
	affichage(modele);

      /* On termine l'ajustement */
      return 0; /* Minimum atteint, arrt "normal" */
    }

    /* Si on est l, c'est que l'on a russi  minimiser l'erreur (joie !) */
    /* Le spectre ajust est dj calcul, le rsidu aussi, les valeurs ne sont pas mises  jour */

    /* On calcule ce qu'il faut */
    ex_erreur = modele->erreur;
    modele->erreur = nv_erreur;
/*     modele->erreur = MODELE_CalculerResidu(modele); */

    HIST_Texte(5, 2, "It"e_AIGU"ration %u : absolue  : %.20g --- pond"e_AIGU"r"e_AIGU"e : %.20g\n",
	       modele->iteration, modele->erreur, modele->erreur / (double) ( modele->reference->nombre_point ) );

    /* On affiche les valeurs courantes d'ajustement */
    if ( affichage != NULL ) /* Affichage des rsultats */
      affichage(modele);
    FEN_redessin( fenetre, BI_RDS_UTIL ); /* On redessine tout */

    /* Vrifications de sortie */
    if (modele->erreur < err_min) { /* Erreur plus faible que celle souhaite */
      return ARRET_AJUSTEMENT_ERREUR;
    }

    if ( (ex_erreur - modele->erreur) < delta_err_min ) /* Variation de l'erreur trop faible */
      return ARRET_AJUSTEMENT_DELTA_ERREUR;

    (modele->iteration)++; /* On passe  l'itration suivante */

    /* On vrifie que l'utilisateur ne veut pas faire autre chose pendant ce temps */
    BI_AttendEvenement( BI_EVENEMENT_TIMER, /* vnements attendus */
			0, 0, 0, 0, 0,     /* rectangle 2 : aucun */
			0,              /* Compteur  1, sinon plantage sur TOS 1.0 */
			NULL, NULL, NULL, NULL, NULL, NULL );
    if ( VRAI == modele->drapeaux.arret_force ) { /* Erreur survenue dans le calcul des incrments */
      (void) fprintf(stderr, "MODELE_AjustementPG : ajustement interrompu par l'utilisateur\n");
      return ARRET_AJUSTEMENT_FORCE;
    }
  } /* Et c'est reparti pour un tour... */

  /* Si l'on est l, c'est que l'on a atteint le nombre maximal d'itrations */
  return ARRET_AJUSTEMENT_ITERATIONS;
}

/* -------- Calculer le spectre modle (mthode  brutale) -------- */

static int CalculeModelePG(struct modele_ajuste *modele)
{
  unsigned long i;
  double *x;

  if ( GRF_FCT_REELLE == modele->reference->genre_donnees ) { /* Valeurs relles */
    double *y;

    for (i = 0, x = modele->reference->x, y = modele->modele->y ; i < modele->reference->nombre_point ; i++, x++, y++) {
      double tmp;

      tmp = fct_CalculePoint(modele, *x); /* Calcul pour chaque point exprimental */
      *y = tmp;
    }
  } else if ( GRF_FCT_COMPLEXE == modele->reference->genre_donnees ) { /* Valeurs complexes */
    Complexe *y;

    for (i = 0, x = modele->reference->x, y = (Complexe *) (modele->modele->y) ;
	 i < modele->reference->nombre_point ; i++, x++, y++) {
      Complexe tmp;

      tmp = fct_CalculePointC(modele, *x); /* Calcul pour chaque point exprimental */
      *y = tmp;
    }
  }

  GRF_bornes(modele->modele, NE_PAS_FAIRE, DESORDONNE);
  return 0; /* Tout s'est bien pass */
}

/* -------- Calculer les incrments pour chaque variable du modle -------------\
 |   Renvoie 0 si tout s'est bien pass, <0 sinon                               |
 |                                                                              |
 | incrment I = 0.5 * ( dS/dvariable ) / Somme_sur_k ( dmodle/dvariable )    |
 |                                                                              |
 |               Somme_sur_k [ residu(k) * (dmodle / dvariable)(k) / sigma_i] |
 |             = -------------------------------------------------------------- |
 |                  Somme_sur_k [ ( (dmodle / dvariable)(k) ) / sigma_i]     |
 |                                                                              |
 | Suppose implicitement que var->x_max > var->x_min et                         |
 |                       que toutes les variables soient  faire varier         |
 \----------------------------------------------------------------------------- */

static int CalculeIncrementsPG(struct modele_ajuste *modele)
{
  struct pg_variable *var, *variables;
  unsigned int num_var; /* Index de la variable utilise */

  /* Prparation */
  variables = modele->infos_variables.pseudogradient;

  /* Rinitialisation */
  for (num_var = 0, var = variables ; num_var < modele->nbr_variables ; num_var++, var++) {
    var->den = 0.0;
    var->num = 0.0;
  }

  fprintf(stderr, "CalculeIncrementsPG : ENCORE  FAIRE !\n");

  /* Tout s'est bien pass */
  return 0;
}
