/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-1998, Emmanuel CURIS
   Cration              : 12 novembre 1999
   Dernire modification : 18 novembre 1999

   Module d'ajustement par un modle  n couches

   Vrification de l'ajustement
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "systeme.h"

#include "exafs.h"

#include "couches.h"
#include "feff.h"
#include "modeles.h"  /* Pour les codes de vrification */
#include "theorie.h"  /* Pour les modles de lpm */

#include "nc_verifier.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define DW_MINIMAL_REALISTE (0.00005) /* 5/10000 */
#define DW_MAXIMAL_REALISTE (0.1)

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int NC_Bornes      (struct modele_feff *modele);
static int NC_SensPhysique(struct modele_feff *modele);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Fonction gnrale de vrification -------- */

int NC_Verifier(struct modele_feff *modele, int methode)
{
  int retour;

  if ( NULL == modele ) return -1; /* Aucune information d'ajustement ! */

  switch (methode)
  {
  case MODELE_VAL_BORNES:
    retour = NC_Bornes(modele);
    break;

  case MODELE_VAL_PHYSIQUE:
    retour = NC_SensPhysique(modele);
    break;

  case MODELE_VAL_STATS:
  case MODELE_VAL_MONTECARLO:
    (void) fprintf(stderr, "NC_Verifier : mthode inadquate, utiliser MODELE_ControlerValidite(..., %d)\n", methode);
    return 0;

  default:
    (void) fprintf(stderr, "NC_Verifier : mthode inconnue [%d]\n", methode);
    return 0;
  }

  return retour;
}

/* -------- Contrle des bornes des variables -------- */

static int NC_Bornes(struct modele_feff *modele)
{
  int retour = 0;
  Couche *couche;

  if (NULL == modele) return FAUX;

  /* Variables gnrales */
  /* 1. S0 */
  if (0 != VAR_ControleBornes( &(modele->S0_2) ) ) {
    int reponse;
    
    (void) printf("NC_Bornes : Bornes atteintes pour S0^2\n");
    reponse = BI_ERR_alerte( NULL, &MSG_VariableGenerale, VRAI, "S0"DEUX_EXPOSANT,
			     VAR_LireValeur( &( modele->S0_2 ), VARIABLE_LIRE_VALEUR_AJUSTEE ), modele->S0_2.min, modele->S0_2.max);
    retour++;
    if (2 == reponse)
      return retour;
  }

  /* 2. Delta(E0) -- si ncessaire */
  if ( VRAI == modele->drapeaux.DE0_general ) {
    if (0 != VAR_ControleBornes( &(modele->Delta_E) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour Delta(E0)\n");
      reponse = BI_ERR_alerte(NULL, &MSG_VariableGenerale, VRAI, DELTA_MAJUSCULE"E0",
			      VAR_LireValeur( &( modele->Delta_E ), VARIABLE_LIRE_VALEUR_AJUSTEE ), modele->Delta_E.min, modele->Delta_E.max);
      retour++;
      if (2 == reponse)
	return retour;
    }
  }

  /* 3. Gamma -- si ncessaire */
  if ( (VRAI == modele->drapeaux.GAM_general) &&
       ( (MODELE_LPM_GAMMA == modele->type_lpm) || (MODELE_LPM_GAMMA_ETA == modele->type_lpm) ) ) {
    if (0 != VAR_ControleBornes( &(modele->Gamma) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour Gamma\n");
      reponse = BI_ERR_alerte( NULL, &MSG_VariableGenerale, VRAI, GAMMA_MAJUSCULE,
			       VAR_LireValeur( &( modele->Gamma ), VARIABLE_LIRE_VALEUR_AJUSTEE ), modele->Gamma.min, modele->Gamma.max);
      retour++;
      if (2 == reponse)
	return retour;
    }
  }

  /* Contrle des variables de chaque couche */
  for ( couche = modele->couches ; couche != NULL ; couche = couche->suivante ) {
    if ( FAUX == couche->utilisee ) /* Inutile de vrifier ! */
      continue;

    /* 1. N */
    if (0 != VAR_ControleBornes( &(couche->N) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour N / couche %s\n", couche->nom);
      reponse = BI_ERR_alerte( NULL, &MSG_VariableCouche, VRAI, "N", couche->nom,
			       VAR_LireValeur( &( couche->N ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->N.min, couche->N.max );
      retour++;
      if (2 == reponse)
	return retour;
    }

    /* 2. R */
    if ( 0 != VAR_ControleBornes( &(couche->R) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour R / couche %s\n", couche->nom);
      reponse = BI_ERR_alerte( NULL, &MSG_VariableCouche, VRAI, "R", couche->nom,
			       VAR_LireValeur( &( couche->R ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->R.min, couche->R.max);
      retour++;
      if (2 == reponse)
	return retour;
    }

    /* 3. Debye-Waller */
    if ( 0 != VAR_ControleBornes( &(couche->Debye_Waller) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour Debye-Waller / couche %s\n", couche->nom);
      reponse = BI_ERR_alerte( NULL, &MSG_VariableCouche, VRAI, "Debye-Waller", couche->nom,
			       VAR_LireValeur( &( couche->Debye_Waller ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->Debye_Waller.min, couche->Debye_Waller.max);
      retour++;
      if (2 == reponse)
	return retour;
    }

    /* 4. cumulants */
    if (0 != VAR_ControleBornes( &(couche->Cumul_1) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour C1 / couche %s\n", couche->nom);
      reponse = BI_ERR_alerte( NULL, &MSG_VariableCouche, VRAI, "C1", couche->nom,
			       VAR_LireValeur( &( couche->Cumul_1 ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->Cumul_1.min, couche->Cumul_1.max );
      retour++;
      if (2 == reponse)
	return retour;
    }
    if (0 != VAR_ControleBornes( &(couche->Cumul_2) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour C2 / couche %s\n", couche->nom);
      reponse = BI_ERR_alerte( NULL, &MSG_VariableCouche, VRAI, "C2", couche->nom,
			       VAR_LireValeur( &( couche->Cumul_2 ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->Cumul_2.min, couche->Cumul_2.max);
      retour++;
      if (2 == reponse)
	return retour;
    }
    if (0 != VAR_ControleBornes( &(couche->Cumul_3) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour C3 / couche %s\n", couche->nom);
      reponse = BI_ERR_alerte(NULL, &MSG_VariableCouche, VRAI, "C3", couche->nom,
			      VAR_LireValeur( &( couche->Cumul_3 ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->Cumul_3.min, couche->Cumul_3.max);
      retour++;
      if (2 == reponse)
	return retour;
    }
    if (0 != VAR_ControleBornes( &(couche->Cumul_4) ) ) {
      int reponse;
    
      (void) printf("NC_Bornes : Bornes atteintes pour C4 / couche %s\n", couche->nom);
      reponse = BI_ERR_alerte(NULL, &MSG_VariableCouche, VRAI, "C4", couche->nom,
			      VAR_LireValeur( &( couche->Cumul_4 ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->Cumul_4.min, couche->Cumul_4.max);
      retour++;
      if (2 == reponse)
	return retour;
    }

    /* 5. Delta E0, si ncessaire */
    if ( FAUX == modele->drapeaux.DE0_general ) {
      if (0 != VAR_ControleBornes( &(couche->Delta_E) ) ) {
	int reponse;
    
	(void) printf("NC_Bornes : Bornes atteintes pour Delta(E0) / couche %s\n", couche->nom);
	reponse = BI_ERR_alerte(NULL, &MSG_VariableCouche, VRAI, DELTA_MAJUSCULE"E0", couche->nom,
				VAR_LireValeur( &( couche->Delta_E ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->Delta_E.min, couche->Delta_E.max);
	retour++;
	if (2 == reponse)
	  return retour;
      }
    }

    /* 6. Gamma, si ncessaire */
    if ( (FAUX == modele->drapeaux.GAM_general) &&
	 ( (MODELE_LPM_GAMMA == modele->type_lpm) || (MODELE_LPM_GAMMA_ETA == modele->type_lpm) ) ) {
      if (0 != VAR_ControleBornes( &(couche->Gamma) ) ) {
	int reponse;
    
	(void) printf("NC_Bornes : Bornes atteintes pour Gamma / couche %s\n", couche->nom);
	reponse = BI_ERR_alerte(NULL, &MSG_VariableCouche, VRAI, GAMMA_MAJUSCULE, couche->nom,
				VAR_LireValeur( &( couche->Gamma ), VARIABLE_LIRE_VALEUR_AJUSTEE ), couche->Gamma.min, couche->Gamma.max);
	retour++;
	if (2 == reponse)
	  return retour;
      }
    }
  }

  /* Fini ! */
  return retour;
}

/* -------- Contrle du sens physique d'un modle --------
     Renvoie le nombre de variables douteuses dans le modle,
     <0 si le modle a t modifi au passage
 */

static int NC_SensPhysique(struct modele_feff *modele)
{
  int retour = 0;
  boolean modele_modifie = FAUX;
  Couche *couche;

  if (NULL == modele) return 0;

  /* Variables gnrales */
  /* 1. S0 */
  if (modele->S0_2.type != VAR_CONSTANTE) { /* On a utilis le S0 */
    double val_s02;

    val_s02 = VAR_LireValeur( &(modele->S0_2), VARIABLE_LIRE_VALEUR_AJUSTEE );
    if ( (val_s02 >= 1.1) || (val_s02 <= 0.2) ) { /* S0 trop petit ou trop grand */
      int reponse;

      retour++;
      reponse = BI_ERR_alerte(NULL, &MSG_S02Douteux, VRAI, val_s02);
      if (2 == reponse)
	return retour;
    }
#ifdef _DEVELOPPE_
    else
      printf(" - valeur plausible.\n");
#endif
  }

  /* 2. Delta(E0) -- si ncessaire */
  if ( (VRAI == modele->drapeaux.DE0_general) && (modele->Delta_E.type != VAR_CONSTANTE) ) {
    double val_deltae;

    val_deltae = VAR_LireValeur( &(modele->Delta_E), VARIABLE_LIRE_VALEUR_AJUSTEE );
    if ( fabs(val_deltae) >= 10.0 ) { /* Delta_E0 de + de 10 eV */
      int reponse;

      retour++;
      reponse = BI_ERR_alerte(NULL, &MSG_DeltaEDouteux, VRAI, val_deltae);
      if (2 == reponse)
	return retour;
    }
  }

  /* 3. Gamma -- si ncessaire */
  if ( (VRAI == modele->drapeaux.GAM_general) &&
       ( (MODELE_LPM_GAMMA == modele->type_lpm) || (MODELE_LPM_GAMMA_ETA == modele->type_lpm) ) ) {
    if (modele->Gamma.type != VAR_CONSTANTE) { /* On a utilis le Gamma */
      printf("NC_SensPhysique :  FAIRE --- validit de la valeur de Gamma obtenue\n");
    }
  }

  /* On vrifie les couches une par une, si elles sont utilises */
  for (couche = modele->couches ; couche ; couche = couche->suivante) {
    if (FAUX == couche->utilisee)
      continue;

    /* 1. N */
    if ( couche->N.type != VAR_CONSTANTE ) {
      printf("NC_SensPhysique :  FAIRE --- validit de la valeur de N\n");
    }

    /* 2. R */
    if ( couche->R.type != VAR_CONSTANTE ) {
      printf("NC_SensPhysique :  FAIRE --- validit de la valeur de R\n");
    }

    /* 3. Debye-Waller */
    if ( couche->Debye_Waller.type != VAR_CONSTANTE ) {
      double val_dw;

      val_dw = VAR_LireValeur( &(couche->Debye_Waller), VARIABLE_LIRE_VALEUR_AJUSTEE );
      if ( (val_dw <= DW_MINIMAL_REALISTE) || (val_dw >= DW_MAXIMAL_REALISTE) ) {
	int reponse;

	retour++;
	reponse = BI_ERR_alerte(NULL, &MSG_DWDouteux, VRAI, couche->nom, val_dw);
	if (3 == reponse) {
	  return (VRAI == modele_modifie) ? (retour) : (-retour);
	}
	if (2 == reponse) { /* On enlve la couche incrimine */
	  couche->utilisee = FAUX;
	  if (couche == couche_editee)
	    OBJ_choisit(AJ_EXAFS_CHEMINS, AEC_UTILISER, FORCE_NON);
	  modele_modifie = VRAI;
	}
      }
    }

    /* 4. Cumulants */
    if ( couche->Cumul_1.type != VAR_CONSTANTE ) {
      printf("NC_SensPhysique :  FAIRE --- validit de la valeur du 1er cumulant\n");
    }
    if ( couche->Cumul_2.type != VAR_CONSTANTE ) {
      printf("NC_SensPhysique :  FAIRE --- validit de la valeur du 1er cumulant\n");
    }
    if ( couche->Cumul_3.type != VAR_CONSTANTE ) {
      printf("NC_SensPhysique :  FAIRE --- validit de la valeur du 1er cumulant\n");
    }
    if ( couche->Cumul_4.type != VAR_CONSTANTE ) {
      printf("NC_SensPhysique :  FAIRE --- validit de la valeur du 1er cumulant\n");
    }
  }

  /* Contrle achev */
  return (VRAI == modele_modifie) ? (retour) : (-retour);
}
