/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1997 - 2001, Emmanuel CURIS
   Cration en novembre 1997

   Traitement de base des spectres complexes
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"

#include "filtrage.h" /* Pour Nyquist... */
#include "matrices.h" /* Pour la matrice de covariance */
#include "operations.h"
#include "science.h"

#include "op_complexe.txt"

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

/* ----------------------- Conversion graphe complexe ---> graphes rels ---------------------- *
 *    On peut rcuprer partie relle, partie imaginaire, module ou argument                    *
 ************************************************************************************************/

/* ~~~~~~~~~~~~~~~~~~~~~~~ Oprations proprement dites ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* ----------------------- Extraction de la partie imaginaire ----------------------------------\
 |    Renvoie le graphe contenant la partie imaginaire                                          |
 |                                                                                              |
 | complexe : le graphe complexe (y = contient les valeurs complexes)                           |
 \--------------------------------------------------------------------------------------------- */

Graphe *CMP_imaginaire(Graphe *complexe)
{
  Graphe *imag;
  char *titre, *legende_Y;
  register Complexe *y_comp;
  register double   *y_reel;
  unsigned long i;

  if (complexe == NULL) return NULL;
  if ( complexe->genre_donnees != GRF_FCT_COMPLEXE ) {
    (void) fprintf(stderr, "%s, %s l.%u : Graphe  valeurs non complexes [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, complexe->nom );
    return NULL;
  }

  /* Cration d'un graphe rel avec les memes X */
  titre = TXT_genere_chaine(titre_partie_imaginaire, complexe->nom);
  imag = GRF_nv_TableX( complexe, GR_COMPIMA, GRF_FCT_REELLE, titre, FAUX, complexe->x, complexe->nombre_point, VRAI );
  free( titre ); titre = NULL;
  if (imag == NULL) return NULL;

  /* Extraction de la partie imaginaire */
  y_reel = imag->y;
  y_comp = (Complexe *) complexe->y;
  for ( i = 0 ; i < complexe->nombre_point ; i++, y_reel++, y_comp++ ) {
    *y_reel = y_comp->im;
  }
  GRF_bornes(imag, NE_PAS_FAIRE, DESORDONNE);

  /* Conservation des barres d'erreur, si elles existent */
  imag->erreur_x = GRF_Copier( complexe->erreur_x, NULL, VRAI ); /* X inchangs */
  if (complexe->erreur_y != NULL) {
    GRF_creer_erreurs( imag, AXE_Y );
    if (imag->erreur_y == NULL)
      return imag;

    for (i = 0, y_comp = (Complexe *) (complexe->erreur_y->y), y_reel = imag->erreur_y->y ;
	 i < complexe->nombre_point ; i++, y_comp++, y_reel++)
      *y_reel = y_comp->im;

    /* Conservation des ddl */
    memcpy(imag->erreur_y->z.ddl, complexe->erreur_y->z.ddl, sizeof(int) * complexe->nombre_point);
    GRF_bornes(imag->erreur_y, NE_PAS_FAIRE, DESORDONNE);
  }

  /* Les units sont inchanges, ainsi que la lgende en X ; conservation des couleurs et du style */
  GRF_CopierAffichage( imag, complexe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

  /* Adaptation de la lgende en Y */
  legende_Y = TXT_genere_chaine( "Im(%s)", GRF_QuelleLegende( complexe, AXE_Y ) );
  GRF_ChangerLegende( imag, AXE_Y, legende_Y, FAUX );	/* legende_Y sera libr avec le graphe */

  /* Conservation des termes de covariance, s'ils existent */
  if ( complexe->covariance != NULL ) {
    GRF_CreerCovariance( imag );
    if ( imag->covariance != NULL ) {
      unsigned long j;

      for (i = 0 ; i < imag->nombre_point ; i++) {
	for (j = 0 ; j <= i ; j++) {
	  /* Partie imaginaire = termes impairs de la matrice de covariance complexe */
	  imag->covariance->valeurs[ MAT_SYM_element(i, j) ] =
	    complexe->covariance->valeurs[ MAT_SYM_element( 2*i + 1, 2*j + 1) ];
	}
      }
    } else {
      fprintf(stderr, "CMP_imaginaire : impossible de crer la matrice de covariance\n");
    }
  }

  /* Log. dans l'historique */
  HIST_Titre(4, "Extraction de la partie imaginaire", "CMP");
  HIST_Texte(4, 0, "Graphe original : %s\nPartie imaginaire : %s\n", complexe->nom, imag->nom);

  return imag;
}

/* -------- Extraction du module -------------------------------------\
 |    Renvoie le graphe contenant le module (NULL si erreur)          |
 |                                                                    |
 | complexe : le graphe complexe (y = contient les valeurs complexes) |
 \------------------------------------------------------------------- */

Graphe *CMP_module(Graphe *complexe)
{
  Graphe *module;
  char *titre;
  register Complexe *y_comp;
  register double   *y_reel;
  unsigned long i;
  char *legende_Y;

  if (complexe == NULL) return NULL;
  if ( complexe->genre_donnees != GRF_FCT_COMPLEXE ) {
    (void) fprintf(stderr, "%s, %s l.%u : Graphe  valeurs non complexes [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, complexe->nom );
    return NULL;
  }

  titre = TXT_genere_chaine(titre_partie_module, complexe->nom);
  module = GRF_nv_TableX( complexe, GR_COMPMOD, GRF_FCT_REELLE, titre, FAUX, complexe->x, complexe->nombre_point, VRAI );
  free( titre ); titre = NULL;
  if (NULL == module) return NULL;

  /* Extraction du module */
  y_reel = module->y;
  y_comp = (Complexe *) complexe->y;
  for (i = 0 ; i < complexe->nombre_point ; i++, y_reel++, y_comp++) {
    *y_reel = sqrt(y_comp->re * y_comp->re + y_comp->im * y_comp->im);
  }
  GRF_bornes(module, NE_PAS_FAIRE, DESORDONNE);

  /* Les erreurs en X sont inchanges */
  module->erreur_x = GRF_Copier( complexe->erreur_x, NULL, VRAI );

  /* ---- Remarque pour les erreurs en Y ----
     Il n'y a pas de rsultat gnral donnant l'cart-type...
     Tout au plus, pour une distribution gaussienne X = N(,s),
     on a un rsultat sur le carr du module :
     E(X) = s +  et V(X) = 2*s^4 + *s
     et, si s = 1, X suit un khi-deux dcentr de 0,5  1 ddl
     On peut gnraliser pour une N(,)  2 D = (Re, Im), mais...
     Comment passer ensuite  la racine carre ?
  */

  /* Les units sont inchanges, ainsi que la lgende en X ; conservation des couleurs et du style */
  GRF_CopierAffichage( module, complexe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

  /* Adaptation de la lgende en Y */
  legende_Y = TXT_genere_chaine( "|%s|", GRF_QuelleLegende( complexe, AXE_Y ) );
  GRF_ChangerLegende( module, AXE_Y, legende_Y, FAUX );	/* legende_Y sera libr avec le graphe */

  /* Log. dans l'historique */
  HIST_Titre(4, "Extraction du module", "CMP");
  HIST_Texte(4, 0, "Graphe original : %s\nModule : %s\n", complexe->nom, module->nom);

  return module;
}

/* -------- Extraction de l'argument (phase) -------------------------\
 |    Renvoie le graphe contenant l'argument (NULL en cas d'erreur)   |
 |                                                                    |
 | complexe : le graphe complexe (y = contient les valeurs complexes) |
 \------------------------------------------------------------------- */

Graphe *CMP_argument(Graphe *complexe)
{
  Graphe *phase;
  char *titre, *legende_Y;
  register Complexe *y_comp;
  unsigned long i;
  long k = 0; /* Peut etre aussi bien positif que ngatif ! */

  if (complexe == NULL) return NULL;
  if ( complexe->genre_donnees != GRF_FCT_COMPLEXE ) {
    (void) fprintf(stderr, "%s, %s l.%u : Graphe  valeurs non complexes [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, complexe->nom );
    return NULL;
  }

  titre = TXT_genere_chaine(titre_partie_argument, complexe->nom);
  phase = GRF_nv_TableX( complexe, GR_COMPARG, GRF_FCT_REELLE, titre, FAUX, complexe->x, complexe->nombre_point, VRAI );
  free( titre ); titre = NULL;
  if (NULL == phase) return NULL;

  /* Extraction de la phase */
  y_comp = (Complexe *) complexe->y;

  /* a. Premire valeur : dans [-PI, PI] */
  phase->y[ 0 ] = CMPL_argument( y_comp[ 0 ] );

  /* b. Chaque point suivant : on doit tester si l'on reste dans la meme priode */
  for ( i = 1 ; i < phase->nombre_point ; i++ ) {
    double argument, d_id, d_av, d_ap;

    /* Phase calcule */
    argument = CMPL_argument( y_comp[ i ] ) + k * M_PI;

    /* Distance avec le point prcdent */
    d_id = fabs( argument - phase->y[ i - 1 ] );

    /* Distance avec le prcdent en descendant d'une priode */
    d_av = fabs( argument - phase->y[ i - 1 ] - 2.0 * M_PI );

    /* Distance avec le prcdent en montant d'une priode */
    d_ap = fabs( argument - phase->y[ i - 1 ] + 2.0 * M_PI );

    /* Quel point est le plus proche (pour garder la continuit) ? */
    if ( d_av < d_id ) {
      if ( d_ap < d_av ) {
	AVERTISSEMENT( "Bizarre, d_ap < d_av < d_id [%g < %g < %g]", d_ap, d_av, d_id );
	k += 2;
      } else {
#ifdef _DEVELOPPE_
	if ( d_ap < d_id ) {
	  AVERTISSEMENT( "Bizarre, d_av < d_ap < d_id [%g < %g < %g]", d_av, d_ap, d_id );
	}
#endif
	k -= 2;
      }
    } else if ( d_ap < d_id ) {
      k += 2;
    }

    /* Calcul de la phase dans la bonne priode */
    phase->y[ i ] = CMPL_argument( y_comp[ i ] ) + k * M_PI;
  }
  GRF_bornes(phase, NE_PAS_FAIRE, DESORDONNE);

  /* Erreurs en X inchanges */
  phase->erreur_x = GRF_Copier( complexe->erreur_x, NULL, VRAI );

  /* ---- Remarque pour les erreurs en Y ----
     Aucun rsultat pour leur propagation...
  */

  /* la lgende et l'unit en X sont inchangs ; conservation des couleurs et du style */
  GRF_CopierAffichage( phase, complexe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE );

  /* Adaptation de la lgende en Y */
  legende_Y = TXT_genere_chaine( "Arg(%s)", GRF_QuelleLegende( complexe, AXE_Y ) );
  GRF_ChangerLegende( phase, AXE_Y, legende_Y, FAUX );	/* legende_Y sera libr avec le graphe */
  GRF_ChangerUnite( phase, AXE_Y, "rad" );

  /* Log. dans l'historique */
  HIST_Titre(4, "Extraction de l'argument", "CMP");
  HIST_Texte(4, 0, "Graphe original : %s\nArgument : %s\n", complexe->nom, phase->nom);

  return phase;
}

/* -------- Extraction de la partie relle ----------------------------\
 |    Renvoie le graphe contenant la partie relle (NULL si erreur)    |
 |                                                                     |
 | complexe : le graphe complexe (y = contient les valeurs complexes)  |
 \-------------------------------------------------------------------- */

Graphe *CMP_reelle(Graphe *complexe)
{
  Graphe *reelle;
  char *titre, *legende_Y;
  register Complexe *y_comp;
  register double   *y_reel;
  unsigned long i;

  if (complexe == NULL) return NULL;
  if ( complexe->genre_donnees != GRF_FCT_COMPLEXE ) {
    (void) fprintf(stderr, "%s, %s l.%u : Graphe  valeurs non complexes [%s]\n",
		   __FILE__, __FUNCTION__, __LINE__, complexe->nom );
    return NULL;
  }

  titre = TXT_genere_chaine(titre_partie_reelle, complexe->nom);
  reelle = GRF_nv_TableX( complexe, GR_COMPREEL, GRF_FCT_REELLE, titre, FAUX, complexe->x, complexe->nombre_point, VRAI );
  free(titre);
  if (NULL == reelle) return NULL;

  /* Extraction de la partie imaginaire */
  y_reel = reelle->y;
  y_comp = (Complexe *) complexe->y;
  for (i = 0 ; i < complexe->nombre_point ; i++, y_reel++, y_comp++) {
    *y_reel = y_comp->re;
  }
  GRF_bornes(reelle, NE_PAS_FAIRE, DESORDONNE);

  /* Conservation des barres d'erreur, si elles existent */
  reelle->erreur_x = GRF_Copier( complexe->erreur_x, NULL, VRAI ); /* X inchangs */
  if (complexe->erreur_y != NULL) {
    GRF_creer_erreurs( reelle, AXE_Y );
    if (NULL == reelle->erreur_y)
      return reelle;

    for (i = 0, y_comp = (Complexe *) (complexe->erreur_y->y), y_reel = reelle->erreur_y->y ;
	 i < complexe->nombre_point ; i++, y_comp++, y_reel++)
      *y_reel = y_comp->re;

    /* Conservation des ddl */
    memcpy(reelle->erreur_y->z.ddl, complexe->erreur_y->z.ddl, sizeof(int) * complexe->nombre_point);
    GRF_bornes(reelle->erreur_y, NE_PAS_FAIRE, DESORDONNE);
  }

  /* Les units sont inchanges, ainsi que la lgende en X ; conservation des couleurs et du style */
  GRF_CopierAffichage( reelle, complexe, GRF_CP_AFF_LEGENDE_X | GRF_CP_AFF_UNITE_X | GRF_CP_AFF_UNITE_Y | GRF_CP_AFF_COULEUR | GRF_CP_AFF_STYLE | GRF_CP_AFF_MARQUE );

  /* Adaptation de la lgende en Y */
  legende_Y = TXT_genere_chaine( "Re(%s)", GRF_QuelleLegende( complexe, AXE_Y ) );
  GRF_ChangerLegende( reelle, AXE_Y, legende_Y, FAUX );	/* legende_Y sera libr avec le graphe */

  /* Conservation des infos pour appliquer Shannon, si ncessaire */
  if ( (complexe->type == GR_TFI) && (complexe->pointeur.nyquist) ) {
    Infos_Nyquist *nyquist;

    nyquist = (Infos_Nyquist *) malloc( sizeof(Infos_Nyquist) );
    if (nyquist == NULL) {
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "complexe.c", "CMP_reelle", "nyquist");
      return reelle;
    }

    nyquist->r_min = complexe->pointeur.nyquist->r_min;
    nyquist->r_max = complexe->pointeur.nyquist->r_max;
    reelle->pointeur.nyquist = nyquist;
    reelle->type = GR_EXAFS; /* Partie relle d'une TFI = signal EXAFS, en gnral... */
    reelle->drapeaux.exafs = VRAI;
  }

  /* Conservation des termes de covariance, s'ils existent */
  if ( complexe->covariance != NULL ) {
    GRF_CreerCovariance( reelle );
    if ( reelle->covariance != NULL ) {
      unsigned long j;

      for (i = 0 ; i < reelle->nombre_point ; i++) {
	for (j = 0 ; j <= i ; j++) {
	  /* Partie reelle = termes pairs de la matrice de covariance complexe */
	  reelle->covariance->valeurs[ MAT_SYM_element(i, j) ] =
	    complexe->covariance->valeurs[ MAT_SYM_element( 2*i, 2*j ) ];
	}
      }
    } else {
      (void) fprintf(stderr, "CMP_reelle : impossible de crer la matrice de covariance\n");
    }
  }

  /* Log. dans l'historique */
  HIST_Titre(4, "Extraction de la partie r"e_AIGU"elle", "CMP");
  HIST_Texte(4, 0, "Graphe original : %s\nPartie r"e_AIGU"elle : %s\n", complexe->nom, reelle->nom);

  return reelle;
}
