/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1999-2002, Emmanuel CURIS
   Cration de ce fichier : 4 fvrier 2002

   Module d'analyse statistique

   Combinaison des erreurs en chaque point pour avoir une estimation, commune, si possible
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "systeme.h"

#include "exafs.h"

#include "statistiques.h"

#include "interface.h" /* Pour RCT_pointeur */
#include "tracer.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define HV_METHODE_BARTLETT 0 /* Test de Bartlett, comparaison multiple de variances */

#define HV_NOMBRE_METHODES 1

#include "condenser.txt"

#define LGMAX_NOM_METHODE 20

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static double ERRM_Bevington(Graphe *graphe, Infos_ERRM *infos);
/* static double ERRM_Moyenne  (Graphe *graphe, Infos_ERRM *infos); */
static double ERRM_Commune  (Graphe *graphe, Infos_ERRM *infos);

static boolean ERRM_Remplace(Graphe *graphe, Infos_ERRM *infos, double err_commune);

static double ERRM_ValeurBartlett(Infos_ERRM *infos);

static void    HV_PreparerArbre (void);
static int     HV_GererArbre    (unsigned int num_objet, int x, int y);
static int     HV_ClavierArbre  (unsigned int touche, int etat, int *renvoi);
static int     HV_AfficherApercu(PARMBLK *parametres);
static int     HV_AfficherResultats(Infos_ERRM *infos);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Ralisation du test de comparaison -------- */

int STAT_TesterVariances(Infos_ERRM *infos)
{
  double critere;

  if ( NULL == infos ) return -1;
  if ( NULL == infos->graphe ) return -2;

  infos->test.proprietes = 0; /* Bilatral, asymptotique, paramtrique en gnral */
  infos->test.H0 = "Toutes les variances sont gales";
  infos->test.H1 = "Au moins une variance diffre des autres";

  /* En fonction de la mthode de calcul */
  switch ( infos->methode )
  {
  case HV_METHODE_BARTLETT:
    /* Calcul du critre */
    critere = ERRM_ValeurBartlett( infos );

    /* Remplissage du test */
    infos->test.nom = "Test de Bartlett";
    infos->test.table = NULL; /* Khi-deux : "tables" internes */
    infos->test.v_ideale  = 0.0; /* Idalement : 0 */
    infos->test.v_esperee = (double) (infos->nbr_classes - 1); /* Le critre suit sous H0 un khi-deux  k-1 ddl, k = nombre de variances  comparer */
    infos->test.variance  = 2.0 * (double) (infos->nbr_classes - 1);
    infos->test.nombre_valeurs = infos->nbr_classes;

/*     printf("Valeur bartlett : %g - khi-deux  %lu degrs de libert\n", critere, infos->nbr_classes); */
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Mthode inconnue [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, infos->methode);
    return -21;
  }

  /* On mmorise */
  infos->test.resultat = critere;

  /* Calcul termin, sans problme particulier */
  return 0;
}

/* -------- Remplacement des valeurs --------
   Cre un tableau de z qui contient le nombre de mesures pour chaque valeur estime, car plus forcment le mme partout
 */

static boolean ERRM_Remplace(Graphe *graphe, Infos_ERRM *infos, double err_commune)
{
  double *x, *y;
  unsigned long i;
  int *z;

  if ( (graphe == NULL) || (infos == NULL) ) return FAUX;

  if (graphe->z.ddl == NULL) { /* On doit le crer */
    return FAUX;
  }

  for (i = 0, x = graphe->x, y = graphe->y, z = graphe->z.ddl ; i < graphe->nombre_point ; i++, x++, y++, z++) {
    if ( (*x >= infos->x_min) && (*x <= infos->x_max) ) {
      *y = err_commune;

      /* Pour les ddl : on ne les modifie pas. En effet, le lien avec le nombre de mesures rel est important ;
	 de plus, pas mal de routines recalculent une somme de ddl  partir d'eux. En appelant plusieurs fois la
	 variance commune, on pourrait faire monter indfiniment les ddl !
	     De plus, le lien entre V(Y) et V(Y_) ne dpend pas de la prcision de la dtermination de sigma,
         il apparat plus important de mmoriser le nombre de mesures que les ddl

	 Contrepartie : surestimera un ventuel intervalle de confiance ultrieur sur la moyenne...
*/
      if ( *z < 0 ) { /* La variance commune est _toujours_ calcule dans la population */
	*z = -(*z);
      }
    }
  }

  GRF_bornes(graphe, NE_PAS_FAIRE, DESORDONNE);

  return VRAI;
}

/* ~-~-~-~-~-~-~-~- Combinaison des erreurs  ~-~-~-~-~-~-~-~- */

/* -------- Mthode statistique normale -------- */

static double ERRM_Commune(Graphe *graphe, Infos_ERRM *infos)
{
  double *x, *y;
  double var_comm, somme = 0.0;
  int *ddl;
  unsigned long i;

  if ( (graphe == NULL) || (infos == NULL) ) return -1.0;

  if (infos->somme_ddl < 2) {
    /* Au moins 2 mesures ! */
    return -1.0;
  }

  for (i = 0, x = graphe->x, y = graphe->y, ddl = graphe->z.ddl ; i < graphe->nombre_point ; i++, x++, y++, ddl++) {
    if ( (*x >= infos->x_min) && (*x <= infos->x_max) ) {
      if ( abs(*ddl) > 0) {
	double variance;

	variance = *y * *y;  /* Ce sont les carts-types qui sont en mmoire... */
	if (*ddl < 0)
	  variance *= (double) -( *ddl - 1 ); /* On veut la variance dans la population */
	somme += variance * (double) abs(*ddl);
      }
    }
  }

  var_comm = somme / infos->somme_ddl;

  return var_comm;
}

/* -------- Mthode de Bevington ----------------\
 |  Calcul avec  une moyenne quadratique inverse |
 |                1                              |
 |    <s> = -------------                        |
 |              ---------   (origine ?)          |
 |             /       1    (Intrt limit)     |
 |            / Somme ---                        |
 |          \/        s_k                       |
 \---------------------------------------------- */

static double ERRM_Bevington(Graphe *graphe, Infos_ERRM *infos)
{
  double err_Bevington = 0.0, somme = 0.0;
  unsigned long i, nbr_pts = 0;
  double *x, *y;
  boolean err_DivZero = FAUX;

  for (i = 0, x = graphe->x, y = graphe->y ; i < graphe->nombre_point ; i++, x++, y++) {
    if ( (*x >= infos->x_min) && (*x <= infos->x_max) ) {
      if (*y > 0.0) {
	double variance;

	variance = *y * *y;  /* Ce sont les carts-types qui sont en mmoire... */
	somme += 1.0 / variance;
	nbr_pts++;
      } else {
	err_DivZero = VRAI;
      }
    }
  }

  if (somme > 0.0) {
    err_Bevington = (double) nbr_pts / sqrt( somme );
    if ( VRAI == err_DivZero ) {
    }
  }

  return err_Bevington;
}

/* ~-~-~-~-~-~-~-~- Tests de validit : comparaison de plusieurs variances  ~-~-~-~-~-~-~-~- */

/* -------- Comparaison des variances, test de Bartlett -------- */

static double ERRM_ValeurBartlett( Infos_ERRM *infos )
{
  Graphe *graphe;
  double *x, *y;
  double somme_log = 0.0, log_somme = 0.0, somme_inv = 0.0, valeur, facteur;
  int *ddl;
  unsigned long i, nbr_classes = 0, somme_ddl = 0, nbr_passes = 0;

  if ( ( NULL == infos ) || ( NULL == infos->graphe ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Manque d'informations\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
    return -1.0;
  }
  graphe = infos->graphe;

  for (i = 0, x = graphe->x, y = graphe->y, ddl = graphe->z.ddl ; i < graphe->nombre_point ; i++, x++, y++, ddl++) {
    if ( (*x >= infos->x_min) && (*x <= infos->x_max) ) {
      if ( abs(*ddl) > 0 ) {
	double variance;
	double ddl_vrai;

	nbr_classes++; /* Une classe par valeur de E/k = X */
	variance = *y * *y;  /* Ce sont les carts-types qui sont en mmoire... */
	if( 0.0 == variance ) {
	  (void) fprintf(stderr, "%s,%s l.%u : Variance nulle avec des ddl, impossible !\n",
			 __FILE__, __FUNCTION__, __LINE__);
	  return -2.0;
	}
	if ( *ddl < 0 ) { /* Erreur sur la moyenne */
	  variance *= (double) ( -(*ddl - 1) ); /* On tient compte du terme correctif donnant l'erreur dans la population */
	  ddl_vrai = (double) -*ddl;
	} else {
	  ddl_vrai = (double) *ddl;
	}
	somme_log += log( variance ) * ddl_vrai;
	log_somme += variance * ddl_vrai;
	somme_ddl += abs( *ddl);
	somme_inv += 1.0 / ddl_vrai;
#ifdef _DEVELOPPE_
      } else {
	printf("%s : DDL = 0, point (%g, %g)\n", __FUNCTION__, *x, *y );
#endif
      }
    } else {
      nbr_passes++;
    }
  }
  infos->nbr_classes = nbr_classes;
  infos->somme_ddl = somme_ddl;
  if (infos->nbr_classes < 2) {
    return -2.0;
  }
  if (somme_ddl < nbr_classes) { /* Certaines variances n'ont pas de ddl ! */
    return -3.0;
  }

  log_somme = log(log_somme);

  valeur = somme_ddl * (log_somme - log( (double) somme_ddl )) - somme_log;

  /* Terme correctif pour avoir  peu prs un khi-deux  k-1 ddl */
  facteur = 1.0 + (somme_inv - 1.0/(double) somme_ddl) / (3.0 * nbr_classes);
/*  facteur = 1.0 + (infos->nbr_classes + 1.0) / (3.0 * infos->nbr_classes * (infos->nbr_mesures - 1.0)); *//* Car mme nombre de mesures !*/

/*   (void) BI_ERR_alerte(NULL, &Resultat_Bartlett, VRAI, valeur / facteur, infos->nbr_classes - 1, infos->nbr_classes - 1); */

  printf("bartlett sans facteur correctif : %g\n", valeur);

  return valeur / facteur;
}

/* ~-~-~-~-~-~-~-~- Interface utilisateur ~-~-~-~-~-~-~-~- */

int condenser_erreur(Graphe *original)
{
  Graphe *graphe;
  Infos_ERRM infos;
  int reponse;
  double err_commune = 0.0;

  if ( original == NULL ) return -1;
  if ( original->type != GR_DEVIATION ) {
    /* Veut-on travailler sur les incertitudes du graphe ? */
    if ( NULL == original->erreur_x ) {
      if ( NULL == original->erreur_y ) {
	(void) fprintf(stderr, "%s,%s l.%u : Le graphe ne contient aucune incertitude\n",
		       __FILE__, __FUNCTION__, __LINE__);
	return -2;
      }
      graphe = original->erreur_y;
    } else {
      if ( NULL == original->erreur_y ) {
	graphe = original->erreur_x;
      } else {
	reponse = BI_ERR_alerte( principale, &Q_XouY, VRAI );
	if ( 3 == reponse ) return +1; /* On annule */
	if ( 2 == reponse )
	  graphe = original->erreur_y;
	else
	  graphe = original->erreur_x;
      }
    }
  } else {
    graphe = original;
  }
  if ( VRAI == GRF_EstComplexe( graphe ) ) {
    OPTION_A_FAIRE( "Graphe  valeurs complexes" );
  } else if ( FAUX == GRF_EstReel( graphe ) ) {
    (void) fprintf(stderr, "%s,%s l.%u : Graphe  valeurs bizarres\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return -3;
  }

  /* On s'assure d'avoir une table de X correcte pour le dessins */
  if ( ( GRF_ERREUR_R == graphe->genre_donnees ) || ( GRF_ERREUR_C == graphe->genre_donnees ) ) {
    graphe->x = original->x;
    graphe->x_min = original->x_min;
    graphe->x_max = original->x_max;
  }

  /* Prparation de l'arbre */
  HV_PreparerArbre();

  /* Initialisation */
  infos.graphe = graphe;
  infos.x_min  = graphe->x_min;
  infos.x_max  = graphe->x_max;
  infos.somme_ddl = 0;
  infos.nbr_classes = 0;

  infos.methode = HV_METHODE_BARTLETT;
  infos.test.alpha = 0.05;

  infos.echelle = ( graphe->x_max - graphe->x_min ) / arbres[ HOM_VARIANCES ].adresse[ HV_APERCU ].ob_width;

  /* Valeur du critre de test */
  STAT_TesterVariances( &infos );
  if ( 0 == infos.somme_ddl ) {
    (void) BI_ERR_alerte( principale, &ERR_AucunDDL, VRAI );
    return -19;
  }

  /* Afficher les rsultats du test */
  (void) HV_AfficherResultats( &infos );

  /* Association des informations  l'arbre */
  ARB_MettreComplement( HOM_VARIANCES, &infos );

  /* Appel de l'arbre */
  reponse = ARB_Bloquant( HOM_VARIANCES, HV_titre_arbre, principale, 0 );
  ARB_MettreComplement( HOM_VARIANCES, NULL );
  if ( HV_ANNULER == reponse ) { /* Rien de spcial  faire */
    return VRAI;
  }

  /* Remplacement des valeurs : quelle mthode ? */
  reponse = BI_ERR_alerte( NULL, &Quitter_Condense, VRAI );
  if ( CONDENSE_QUITTER_ANNULER == reponse ) {
    return VRAI;
  }

  /* Historique */
  HIST_Titre(1, "Condenser les erreurs", "CondErr");
  HIST_Texte(1, 0, "Graphe trait"e_AIGU" : %s\n", graphe->nom);
  HIST_Texte(2, 0, "Test de validit"e_AIGU" : ");
  HIST_Texte(2, 0, "Bartlett\n");
  /* Afficher les rsultats du test */
  HIST_Texte(2, 0, "M"e_AIGU"thode de combinaison : ");

    /* On estime l'erreur commune */
  switch ( reponse )
  {
  case CONDENSE_QUITTER_VARIANCE:
    HIST_Texte(2, 0, "variance commune (mod"e_GRAVE"le statistique)\n");
    err_commune = ERRM_Commune( graphe, &infos );
    HIST_Texte(3, 3, "Variance commune estim"e_AIGU"e : %g ["e_AIGU"cart-type : %g]\n", err_commune, sqrt(err_commune));
    err_commune = sqrt(err_commune); /* Car le reste travaille en cart-type */
    break;

  case CONDENSE_QUITTER_MOYENNE:
    fprintf(stderr, "MSG_condense :  FAIRE, calcul de l'erreur commune par moyenne des erreurs [ERRM_Moyenne()]\n");
    HIST_Texte(2, 0, "moyenne des "e_AIGU"cart-types : %g\n", err_commune);
    break;

  case CONDENSE_QUITTER_BEVINGTON:
    err_commune = ERRM_Bevington( graphe, &infos );
    HIST_Texte(2, 0, "formule de Bevington\n");
    HIST_Texte(3, 3, "estimation obtenue : %g\n", err_commune);
    break;
  }

  /* On fait le remplacement */
  ERRM_Remplace( graphe, &infos, err_commune );

  return VRAI;
}

/* -------- Gestion de la bote de dialogue -------- */

static int HV_GererArbre(unsigned int num_objet, int x, int y)
{
  Infos_ERRM *infos;
  double x_clic, y_clic;

  /* On rcupre les infos */
  infos = (Infos_ERRM *) ARB_LireComplement( HOM_VARIANCES );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'information associe !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return VRAI;
  }

  switch ( num_objet )
  {
  case HV_METHODE:
    {
      int nv_methode;

      nv_methode = SS_menu_texte( HV_textes_methodes, HV_NOMBRE_METHODES, HV_titre_methodes, HOM_VARIANCES, HV_METHODE );
      if ( ( nv_methode >= 0 ) && ( nv_methode != infos->methode ) ) {
	OPTION_A_FAIRE("Recalculer le critre de test");
      }
    }
    return VRAI;

  case HV_APERCU:
    {
      OBJECT *objet;
      int bouton;

      objet = arbres[ HOM_VARIANCES ].adresse + HV_APERCU;
      TRC_EcranVersAbsolu( 0, 0, objet->ob_width, objet->ob_height, infos->echelle, infos->graphe->x_min, infos->graphe->y_max, x, y, &x_clic, &y_clic );

      /* Quel bouton a t utilis ? */
      BI_EtatSouris( NULL, NULL, &bouton, NULL, NULL );
      if ( 1 == bouton ) {
	infos->x_min = x_clic;
      } else {
	infos->x_max = x_clic;
      }
      if ( infos->x_max < infos->x_min ) {
	double tmp;

	tmp = infos->x_max; infos->x_max = infos->x_min; infos->x_min = tmp;
      }
      BI_DessinerObjet( HOM_VARIANCES, HV_APERCU );

      /* Recalcul */
      (void) STAT_TesterVariances( infos );

      /* Raffichage */
      (void) HV_AfficherResultats( infos );
    }
    return VRAI;
  }

  return FAUX;
}

/* -------- Dessin de l'aperu -------- */

static int HV_AfficherApercu(PARMBLK *parametres)
{
  int x_barre;
  Infos_ERRM *infos;

  /* On rcupre les infos */
  infos = (Infos_ERRM *) ARB_LireComplement( HOM_VARIANCES );

  /* On prpare le dessin */
  TRC_InitDessins( NULL, 0, 0, parametres->pb_w, parametres->pb_h );

  /* On dessine le graphe */
  TRC_GrapheDansRectangle( NULL, infos->graphe, NOIR, 1, 0, 0, parametres->pb_w, parametres->pb_h );

  /* On place les barres */
  /* a. Xmin en rouge */
  x_barre = TRC_ConvertitX_PvE( infos->x_min, infos->graphe->x_min, 0, infos->echelle );
  TRC_Couleur( ROUGE );
  TRC_LigneV( NULL, x_barre, 0, parametres->pb_h );

  /* b. Xmax en vert */
  x_barre = TRC_ConvertitX_PvE( infos->x_max, infos->graphe->x_min, 0, infos->echelle );
  TRC_Couleur( VERT );
  TRC_LigneV( NULL, x_barre, 0, parametres->pb_h );

  /* Termin */
  return 0;
}

/* -------- Affichage des rsultats -------- */

static int HV_AfficherResultats(Infos_ERRM *infos)
{
  if ( NULL == infos ) return -1;

  /* Intervalle utilis */
  OBJ_MettreNombre( HOM_VARIANCES, HV_XMIN, infos->x_min, 0.0 );
  OBJ_MettreNombre( HOM_VARIANCES, HV_XMAX, infos->x_max, 0.0 );

  /* Mthode utilise */
  met_chaine( HOM_VARIANCES, HV_METHODE, LGMAX_NOM_METHODE, HV_textes_methodes[ infos->methode ], TXT_REMPLACE );

  /* Nombre de valeurs utilises */
  OBJ_MettreEntier_U( HOM_VARIANCES, HV_NBR_VALEURS, infos->test.nombre_valeurs );

  /* Rsultats attendus */
  OBJ_MettreNombre( HOM_VARIANCES, HV_ATTENDU, infos->test.v_ideale, 0.0 );
  OBJ_MettreNombre( HOM_VARIANCES, HV_ESPERE , infos->test.v_esperee, sqrt( infos->test.variance ) );

  /* Rsultat observ */
  OBJ_MettreNombre( HOM_VARIANCES, HV_OBSERVE, infos->test.resultat, 0.0 );

  return 0;
}

/* -------- Gestion du clavier : <return> = mise  jour -------- */

static int HV_ClavierArbre(unsigned int touche, int etat __attribute__((__unused__)), int *renvoi __attribute__((__unused__)))
{
  Infos_ERRM *infos;

  infos = (Infos_ERRM *) ARB_LireComplement( HOM_VARIANCES );
  if ( NULL == infos ) {
    (void) fprintf(stderr, "%s,%s l.%u : Pas d'informations associes !\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return FAUX;
  }

  if ( ( XK_Return == touche ) || ( XK_KP_Enter == touche ) ) {
    infos->x_min = OBJ_LireNombre( HOM_VARIANCES, HV_XMIN, NULL );
    infos->x_max = OBJ_LireNombre( HOM_VARIANCES, HV_XMAX, NULL );

    if ( infos->x_max < infos->x_min ) {
      double tmp;
      tmp = infos->x_max; infos->x_max = infos->x_min; infos->x_min = tmp;
    }

    /* On recalcule */
    (void) STAT_TesterVariances( infos );

    /* On remet  jour la bote */
    BI_DessinerObjet( HOM_VARIANCES, HV_APERCU );
    (void) HV_AfficherResultats( infos );

    return VRAI;
  }

  return FAUX;
}

/* -------- Prparation de la bote de dialogue -------- */

static void HV_PreparerArbre(void)
{
  int retour;

  /* Dfinition de l'objet utilisateur */
  OBJ_utilisateur( HOM_VARIANCES, HV_APERCU, HV_AfficherApercu, HOM_VARIANCES, METTRE );

  /* Initialisation de l'arbre */
  retour = ARB_Initialiser( HOM_VARIANCES, "menu_erreurs.html#hom_variance", HV_GererArbre, HV_ClavierArbre );
  if ( retour != 0 ) return;

  /* Traduction des textes */
#ifdef AVEC_TRADUCTION_
  OBJ_TraduireTexte( HOM_VARIANCES, HV_TXT_METHODE, texte_ethode  );
  OBJ_TraduireTexte( HOM_VARIANCES, HV_TXT_VALEURS, texte_valeurs );
  OBJ_TraduireTexte( HOM_VARIANCES, HV_TXT_ATTENDU, texte_attendu );
  OBJ_TraduireTexte( HOM_VARIANCES, HV_TXT_ESPERE , texte_espere  );
  OBJ_TraduireTexte( HOM_VARIANCES, HV_TXT_OBSERVE, texte_observe );
  OBJ_TraduireTexte( HOM_VARIANCES, HV_OK  , texte_estimer_commun );
  OBJ_TraduireTexte( HOM_VARIANCES, HV_ANNULER    , texte_termine );
#endif

  /* Amlioration de la disposition des objets */
  OBJ_deplace( HOM_VARIANCES, HV_TXT_XMAX, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( HOM_VARIANCES, HV_XMAX    , 0, +1, DEPLACE_PIXELS );

  OBJ_deplace( HOM_VARIANCES, HV_NBR_VALEURS, 0, +2, DEPLACE_PIXELS );
  OBJ_deplace( HOM_VARIANCES, HV_TXT_VALEURS, 0, +2, DEPLACE_PIXELS );

  OBJ_deplace( HOM_VARIANCES, HV_TXT_ESPERE, 0, +1, DEPLACE_PIXELS );
  OBJ_deplace( HOM_VARIANCES, HV_ESPERE    , 0, +1, DEPLACE_PIXELS );

  /* Termin */
  arbres[ HOM_VARIANCES ].drapeaux.prepare = VRAI;
  return;
}
