 /* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS
   Cration : le 12 avril 1999

   Module de statistiques
   Estimation d'une distribution (en histogrammes)  partir de valeurs exprimentales
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "exafs.h"

#include "interface.h"
#include "statistiques.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

double *valeurs = NULL; /* Les valeurs pour lesquelles estimer la distribution */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void prepare_classes(Graphe *graphe);
static void realise_histo  (Graphe *histo, double *valeurs, unsigned long nbr_valeurs);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

Graphe *estime_distribution(Graphe *graphe)
{
  Graphe *histogramme = NULL;
  unsigned long nbr_classes;
  int retour;

  if (graphe == NULL) return NULL;
  if ( graphe->genre_donnees != GRF_FCT_REELLE ) {
    fprintf(stderr, "Estimation de distribution uniquement sur des graphes  valeurs relles\n");
    return NULL;
  }
  
  /* Postulat : on travaille surtout sur les X */
  OBJ_MettreIntervalle( ESTIM_DISTRIB, ED_MIN, ED_MAX, 0, 0 , graphe->y_min, graphe->y_max, 0.0, 0 );

#ifndef _DEVELOPPE_
  OBJ_active(ESTIM_DISTRIB, ED_CONJOINTE, FORCE_NON);
#endif

  do {
    retour = ARB_Bloquant( ESTIM_DISTRIB, "Estimer une distribution", principale, 0 );
    if ( retour == ED_PROPOSER_N )
      OBJ_MettreEntier_U( ESTIM_DISTRIB, ED_NBR_CLASSES, (unsigned int) sqrt(graphe->nombre_point) );

    if ( retour == ED_ANNULER )
      return NULL;
  } while (retour != ED_EVALUER);

  nbr_classes = OBJ_LireEntier_U( ESTIM_DISTRIB, ED_NBR_CLASSES );
  if ( (nbr_classes < 2) || (nbr_classes >= graphe->nombre_point) ) {
    fprintf(stderr, "Nombre de classes incorrect\n");
    return NULL;
  }

  OPTION_A_FAIRE( "Mette au point les graphes de GENRE histogramme" );
  histogramme = GRF_nouveau( NULL, nbr_classes, GR_HISTOGRAMME, GR_NORMAL, "Distribution de valeurs" );
  if (histogramme == NULL) {
    return NULL;
  }

  histogramme->x_min = OBJ_LireNombre( ESTIM_DISTRIB, ED_MIN, NULL ); /* Valeurs limites */
  histogramme->x_max = OBJ_LireNombre( ESTIM_DISTRIB, ED_MAX, NULL );

  if ( OBJ_choisit(ESTIM_DISTRIB, ED_X, QUESTION) == VRAI ) { /* Distribution des X */
    valeurs = graphe->x;
  } else if ( OBJ_choisit(ESTIM_DISTRIB, ED_Y, QUESTION) == VRAI ) { /* Distribution des Y */
    valeurs = graphe->y;
  }

  prepare_classes(histogramme);
  realise_histo(histogramme, valeurs, graphe->nombre_point);

  if ( OBJ_choisit(ESTIM_DISTRIB, ED_FREQUENCE, QUESTION) == VRAI ) { /* On travail en %, pas en effectif */
    unsigned long i;
    register double *y;

    for(i = 0, y = histogramme->y ; i < nbr_classes ; i++, y++) {
      *y /= (double) (graphe->nombre_point);
    }
    histogramme->y_min /= (double) (graphe->nombre_point);
    histogramme->y_max /= (double) (graphe->nombre_point);
  }

  return histogramme; /* L'histogramme estim */
}

/* -------- Prparation des classes dans l'histogramme --------------------\
 |   Le graphe pass est tel que : nombre de points = nombre de classes    |
 |                                 min = x_min, max = x_max                |
 |   Les classes sont telles que yi = nombre de points appartenant        |
 |                               [xi, xi+1[                                |
 \------------------------------------------------------------------------ */

static void prepare_classes(Graphe *graphe)
{
  unsigned long i;
  double largeur;
  register double *x;

  if (graphe == NULL) return;

  largeur = (graphe->x_max - graphe->x_min) / (double) (graphe->nombre_point);
  for (i = 0, x = graphe->x ; i < graphe->nombre_point ; i++, x++) {
    *x = graphe->x_min + largeur * (double) i; /* Devrait limiter des erreurs d'arrondi */
  }
}

/* -------- Ralisation effective de l'histogramme -----------------------\
 |   Elle suppose implicitement que les classes sont de mme largeur      |
 \----------------------------------------------------------------------- */

static void realise_histo(Graphe *histo, double *valeurs, unsigned long nbr_valeurs)
{
  unsigned long i, nbr_ignores = 0, place;
  double largeur;
  register double *y;

  largeur = (histo->x_max - histo->x_min) / (double) (histo->nombre_point);

  for (i = 0 ; i < nbr_valeurs ; i++, valeurs++) {
    double y_corrige;

    y_corrige = *valeurs - histo->x_min;
    if (y_corrige < 0) {
      nbr_ignores++;
      continue;
    }

    y_corrige /= largeur;
    place = (unsigned long) floor(y_corrige);
    if (place >= histo->nombre_point)
      nbr_ignores++;
    else
      (histo->y)[place]++; /* Un point de plus */
  }

  for (i = 0, y = histo->y ; i < histo->nombre_point ; i++, y++)
    *y /= largeur; /* Normalisation : on reprsente f, estim  partir de dP = n/N, or dP = fdx */

  if (nbr_ignores > 0) {
    fprintf(stdout, "Attention : %lu point ignors\n", nbr_ignores);
  }

  GRF_bornes(histo, NE_PAS_FAIRE, DESORDONNE);
}
