/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) Emmanuel CURIS 1996 - 2003

   Toutes les fonctions de traitement statistique de l'E.X.A.F.S.

   Tests pour vrifier l'homognit d'une srie de graphes
   H0 : les graphes ne diffrent que par le bruit, les expriences sont indpendantes, les mesures aussi.
   H1 : les graphes ne sont pas homognes
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "graphes.h"
#include "lissage.h"      /* Pour GRF_LireValeur */
#include "science.h"      /* Pour gamma_q, si n'existe pas */
#include "statistiques.h"

#include "stats_homogeneite.txt"

/* ~~~~~~~~ Variables locales  ce module ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales  ce module ~~~~~~~~ */

static Graphe *Test_ConstruireSerie(Graphe **graphes, unsigned long nbr_graphes);

static int Test_LoiUniforme(Graphe **graphes, unsigned long nbr_graphes,
			    struct test_stats *resultat, struct graphe **repartition);

/* ~-~-~-~-~-~-~-~-~ Appel depuis l'extrieur ~-~-~-~-~-~-~-~-~ */

int STAT_TesterHomogeneite(struct graphe **graphes, unsigned long nbr_graphes, int methode,
			   struct test_stats *resultat, struct graphe **repartition)
{
  int retour = -1;

  if ( ( NULL == graphes ) || ( NULL == resultat ) ) {
    /* Pas de graphes ou rien pour stocker le rsultat du test ! */
    return -1;
  }
  if ( 1 == nbr_graphes ) {
    return 0; /* Un seul graphe : on est forcment homogne ! */
  }

  switch ( methode )
  {
  case TEST_HOMO_LOIUNIFORME:
    retour = Test_LoiUniforme( graphes, nbr_graphes, resultat, repartition );
    break;

  default:
    (void) fprintf(stderr, "%s,%s l.%u : Mthode de contrle pas encore faite [%d]\n",
		   __FILE__, __FUNCTION__, __LINE__, methode);
  }

  return retour;
}

/* -------- Construire la srie des numros des graphes maximaux -------- */

static Graphe *Test_ConstruireSerie(Graphe **graphes, unsigned long nbr_graphes)
{
  Graphe *maxima;
  boolean meme_taille;
  double max_ponctuel;
  unsigned long nbr_points, i, i_graphe;

  meme_taille = GRF_ControleTaille( graphes, nbr_graphes, GRF_CONTROLE_N_STRICT, &nbr_points, FAUX );
  if ( FAUX == meme_taille ) {
    boolean x_commun;
    double x_min, x_max, x, pas;

    /* On cherche la partie commune */
    x_commun = GRF_ControleX( graphes, nbr_graphes, GRF_CONTROLE_X_INTER, &x_min, &x_max );
    if ( FAUX == x_commun ) {
      (void) fprintf(stderr, "%s,%s l.%u : Pas d'intervalle en X commun\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* On compte le nombre de points dans cette partie commune, dans le premier
       On suppose que l'utilisateur ne cherche quand mme pas  moyenner n'importe-quoi !
       On peut donc supposer que ce sera voisin d'un spectre  l'autre ! */
    nbr_points = GRF_CompterPoints( graphes[ 0 ], AXE_X, x_min, x_max, FAUX );
    if ( nbr_points < 5 * nbr_graphes ) { /* car pour appliquer khi-deux, il faut au moins 5 dans chaque classe ! */
      (void) fprintf(stderr, "%s,%s l.%u : Trop peu de points en commun [%lu, il en faudrait au moins %lu]\n",
		     __FILE__, __FUNCTION__, __LINE__, nbr_points, 5 * nbr_graphes);
      return NULL;
    }

    /* On conserve le graphe, car un coup d'oeil peut suffire  rejeter l'homognit ! */
    maxima = GRF_nv_isole( nbr_points, GR_AUTRETYPE, GRF_SERIE );
#ifdef _DEVELOPPE_
    GRF_titre( maxima, "Graphes maximaux [interpolation linaire]" );
#endif
    if ( NULL == maxima ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le graphe des maxima\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* Pour chacun des points ainsi dfini */
    pas = (x_max - x_min) / (double) ( nbr_points - 1 );
    for ( i = 0 ; i < nbr_points ; i++ ) {
      x = x_min + pas * (double) i;

      maxima->y[ i ] = 0;
      max_ponctuel = GRF_LireValeur( graphes[ 0 ], x, INTERPOLE_LINEAIRE );

      for ( i_graphe = 1 ; i_graphe < nbr_graphes ; i_graphe++ ) {
	double y;
	
	y = GRF_LireValeur( graphes[ i_graphe ], x, INTERPOLE_LINEAIRE );
	if ( y > max_ponctuel ) {
	  maxima->y[ i ] = i_graphe;
	  max_ponctuel = y;
	}
      }
    }
  } else {
    /* On conserve le graphe, car un coup d'oeil peut suffire  rejeter l'homognit ! */
    maxima = GRF_nv_isole( nbr_points, GR_AUTRETYPE, GRF_SERIE );
#ifdef _DEVELOPPE_
    GRF_titre( maxima, "Graphes maximaux" );
#endif
    if ( NULL == maxima ) {
      (void) fprintf(stderr, "%s,%s l.%u : Impossible de crer le graphe des maxima\n",
		     __FILE__, __FUNCTION__, __LINE__);
      return NULL;
    }

    /* Pour chaque point du graphe de rfrence, on cherche quel graphe a la plus grande valeur */
    for ( i = 0 ; i < nbr_points ; i++ ) {
      maxima->y[ i ] = 0;
      max_ponctuel = graphes[ 0 ]->y[ i ];

      for ( i_graphe = 1 ; i_graphe < nbr_graphes ; i_graphe++ ) {
	double y;
	
	y = graphes[ i_graphe ]->y[ i ];
	if ( y > max_ponctuel ) {
	  maxima->y[ i ] = i_graphe;
	  max_ponctuel = y;
	}
      }
    }
  }

  maxima->y_min = 0.0;
  maxima->y_max = (double) (nbr_graphes - 1);

  /* Construction termine */
  return maxima;
}

/* ~-~-~-~-~-~-~-~-~ Test  loi uniforme  ~-~-~-~-~-~-~-~-~ */

/* Principe du test :  chaque graphe est associ un numro, de 0  N.
   On cre une srie, qui contient pour chaque point le numro du graphe maximal.

   Si H0 est vraie, la srie ci-dessus doit provenir d'une loi uniforme discrte sur [0, N]
   On le vrifie avec un test du khi-deux.
*/

static int Test_LoiUniforme(Graphe **graphes, unsigned long nbr_graphes,
			    struct test_stats *resultat, struct graphe **repartition)
{
  unsigned long *occurrences, i, ddl;
  double occ_attendues, *occ_observee;
  double critere_khideux = 0.0, validite;
  Graphe *maxima;

  /* On construit la srie pour l'tude */
  maxima = Test_ConstruireSerie( graphes, nbr_graphes );
  if ( NULL == maxima ) return -1;

  /* Idalement, on doit observer autant de fois chaque spectre */
  occ_attendues = maxima->nombre_point / (double) nbr_graphes;
  if ( 0.0 == occ_attendues ) {
    (void) fprintf(stderr, "%s,%s l.%u : Trop de graphes, prcision insuffisante\n",
		   __FILE__, __FUNCTION__, __LINE__);
    GRF_eff_isole( maxima ); maxima = NULL;
    return -5;
  }

  /* On construit la table des occurrences relles */
  occurrences = (unsigned long *) malloc( sizeof( unsigned long ) * nbr_graphes );
  if ( NULL == occurrences ) {
    ERR_Memoire( "Table des occurrences", sizeof( unsigned long ) * nbr_graphes );
    GRF_eff_isole( maxima ); maxima = NULL;
    return -2;
  }

  /* Au dbut, aucune occurrence */
  for ( i = 0 ; i < nbr_graphes ; i++ )
    occurrences[ i ] = 0;

  for ( i = 0, occ_observee = maxima->y ; i < maxima->nombre_point ; i++, occ_observee++ ) {
    occurrences[ (unsigned int) *occ_observee ]++;
  }

  for ( i = 0 ; i < nbr_graphes ; i++ ) {
    HIST_Texte(3, 3, "Occurrence du graphe %s : %lu [%g attendu]\n", graphes[ i ]->nom, occurrences[ i ], occ_attendues );
#ifdef _DEVELOPPE_
    printf("Occurrence du graphe %s : %lu [%g attendu]\n", graphes[ i ]->nom, occurrences[ i ], occ_attendues );
#endif
  }

  /* On calcule le critre du khi-deux */
  for ( i = 0 ; i < nbr_graphes ; i++ ) {
    double tmp;

    tmp = occurrences[ i ] - occ_attendues;
    tmp *= tmp;

    critere_khideux += tmp / occ_attendues;
  }

  /* Ce critre suit une loi du khi-deux  (nbr_graphes - 1) degrs de libert
     (asymptotiquement...) */
  ddl = nbr_graphes - 1;
  OPTION_A_FAIRE( "Remplacer par appel  fonction probabiliste du khi-deux, par commodit !" );
  validite = gamma_q( 0.5 * (double) ddl, critere_khideux * 0.5 );

  /* On mmorise les rsultats du test */
  resultat->proprietes = STAT_TEST_NP | STAT_TEST_UNILATERAL;
  resultat->v_ideale   = 0;
  resultat->v_esperee  = (double) ddl;
  resultat->variance   = 2.0 * ddl;

  resultat->resultat   = critere_khideux;
  resultat->p_critique = validite;

#ifdef _DEVELOPPE_
  printf( "Attendu : 0 [en moyenne %lu  %g] ; observ : %g ; p(observer >= %g) = %g\n", ddl, sqrt( 2.0 * ddl ), critere_khideux, critere_khideux, validite );
#endif

  HIST_ResultatTestStatistique( "Homognit, contrle de la loi uniforme", "Les graphes ne diffrent que par le bruit", "Les graphes diffrent",
				critere_khideux, 0.0, (double) ddl, (double) 2.0 * ddl, -1.0, FAUX, NULL );

  /* Calcul termin, tout s'est bien pass */
  free( occurrences );

  /* Faut-il conserver la table des maxima ? */
  if ( repartition != NULL ) {
    *repartition = maxima;
  } else {
    GRF_eff_isole( maxima ); maxima = NULL;
  }
  return 0;
}
