/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) 1996-1999, Emmanuel CURIS
   Cration : 20 mai 1999

   Module de calculs statistiques

   Comparaison de 2 distributions par le test de Kolmogorov-Smirnov
   (actuellement, on ne fait que tester la normalit d'une distribution)
   Rfrences pour le test : Saporta, p. 336 - 338
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"

#include "exafs.h"

#include "statistiques.h"
#include "science.h" /* Pour erf sur les systmes qui ne l'ont pas */
#include "tri.h"

#include "ks.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define LGMAX_NOM_GRAPHES_REF 40
#define LGMAX_DISTRIBUTION    30

extern const char *txt_distributions[]; /* Dans Stats/aleatoire.c */
extern const char *txt_aucun_choisi;    /* Dans Textes/variable.* */

struct infos_KS
{
  Graphe *reference;
  Graphe *experience;
  int type_distrib;
};

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int KS_gaussienne(Graphe *frep_exp, unsigned long nbr_pts, double m, double s, boolean theorie);

int GER_ks(unsigned int objet, int x, int y); /* Pas statiques car utilises dans */
void prepare_arbre_ks(void);             /* c_vmises.c aussi (mme arbre) */

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- La fonction de rpartition indique est-elle gaussienne ? --------\
 |    Renvoie VRAI si l'hypothse est valable avec un seuil de 0,05           |
 |            FAUX dans le cas contraire                                      |
 |                                                                            |
 | frep_exp : contient la fonction de rpartition exprimentale estime       |
 | nbr_pts  : nombre de points exprimentaux (peut tre > frep_exp->nbr_pts si|
 |            le graphe original contient des valeurs rptes                |
 | m        : moyenne attendue de la gaussienne                               |
 | s        : cart-type attendu de la gaussienne                             |
 | theorie  : VRAI si m et s thorique (KS sp), FAUX si exprimentaux (Lilliefors)
 \--------------------------------------------------------------------------- */

static int KS_gaussienne(Graphe *frep_exp, unsigned long nbr_pts, double m, double s, boolean theorie)
{
  unsigned long i;
  register double *x, *y;
  double Dn = 0.0, limite = 0.0;

  if (frep_exp == NULL) return FAUX;

  for (i = 0, x = frep_exp->x, y = frep_exp->y ;
       i < frep_exp->nombre_point ;
       i++, x++, y++) {
    double x_cr;    /* X en coordonnes centres rduites */
    double y_gauss; /* Valeur de la fonction de rpartition gaussienne attendue en ce point */
    double d;
    x_cr = (*x - m)/s;
    y_gauss = 0.5 * (1 + erf(x_cr / M_SQRT2));

    d = fabs(y_gauss - *y);
    if (d > Dn)
      Dn = d;    
  }

  /* Ici, on a donc bien Dn = sup | Fexp(x) - Fgauss(x) | calcul sur les x exprimentaux */
  limite = sqrt( (double) nbr_pts ) + 0.85 / sqrt( (double) nbr_pts ) - 0.01;
  if (limite == 0.0) {
    EXF_ERREUR(DIV_ZERO, "Correction de valeur limite", txt_err_calc_annule);
    return FAUX;
  }

  (void) BI_ERR_alerte(NULL, &KS_Gauss_resultat, VRAI, Dn, 0.895/limite, 1.035/limite);
  if ( VRAI == theorie ) { /* Comparaison  une distribution thorique */
    HIST_ResultatTestStatistique("Kolmogorov -- Smirnov",
				 "la distribution exp"e_AIGU"rimentale est gaussienne",  "la distribution exp"e_AIGU"rimentale n'est pas gaussienne",
				 Dn, 0.0, -1.0, -1.0, -1.0 /* A remplacer par un calcul de la loi... */, FAUX,
				 "Nombre de valeurs assez grand");
    HIST_Texte(3, 0, "Param"e_GRAVE"tres de la gaussienne mod"e_GRAVE"le : "MU" = %.10g, "SIGMA" = %.10g\n", m, s);
  } else { /* Moyenne, variance estims sur l'exprience */
    HIST_ResultatTestStatistique("Kolmogorov -- Smirnov",
				 "la distribution exp"e_AIGU"rimentale est gaussienne",  "la distribution exp"e_AIGU"rimentale n'est pas gaussienne",
				 Dn, 0.0, -1.0, -1.0, -1.0, FAUX,
				 "Nombre de valeurs assez grand, pour pouvoir utiliser les limites qui suivent");
    HIST_Texte(3, 0, "Valeur "a_GRAVE" ne pas d"e_AIGU"passer pour accepter H0 au seuil de 5%% : %g\n", 0.895 / limite);
    HIST_Texte(3, 0, "Valeur "a_GRAVE" ne pas d"e_AIGU"passer pour accepter H0 au seuil de 1%% : %g\n", 1.035 / limite);
    HIST_Texte(3, 0, "Param"e_GRAVE"tres estim"e_AIGU"s pour la gaussienne mod"e_GRAVE"le : m = %.10g, s* = %.10g\n", m, s);
  }

  if (Dn > 0.895 / limite) {
    return FAUX;
  }

  return VRAI; /* A pass le test */
}

/* -------- Ralisation effective du test de Kolmogorov-Smirnov -------- */

void kolmogorov_smirnov(Graphe *experience)
{
  double m, s; /* Valeur de la gaussienne */
  Graphe *frep;
  struct infos_KS *infos;

  if (experience == NULL)
    return;

  if (arbres[PARAM_KSMIRNOV].drapeaux.prepare == FAUX)
    prepare_arbre_ks();

  infos = (struct infos_KS *) malloc( sizeof(struct infos_KS) );
  if (infos == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "Stats/k_smirnov.c", "kolmogorov_smirnov", "infos");
    return;
  }
  infos->experience = experience;
  infos->reference = NULL;
  infos->type_distrib = DP_GAUSSIENNE;
  ARB_MettreComplement( PARAM_KSMIRNOV, infos );

  GER_ks(PKS_REF_THEORIE, 0, 0);
  met_chaine(PARAM_KSMIRNOV, PKS_TYPE_DISTRIB, LGMAX_DISTRIBUTION, txt_distributions[infos->type_distrib], TXT_COPIER);
  met_chaine(PARAM_KSMIRNOV, PKS_GRF_EXP, LGMAX_NOM_GRAPHES_REF, experience->nom, TXT_COPIER);
  met_chaine(PARAM_KSMIRNOV, PFS_GRF_REF, LGMAX_NOM_GRAPHES_REF, txt_aucun_choisi, TXT_COPIER);

  if ( ARB_Bloquant( PARAM_KSMIRNOV, titre_arbre_ks, principale, 0 ) == PKS_ANNULER ) {
    free( infos ); infos = NULL;
    ARB_MettreComplement( PARAM_KSMIRNOV, NULL );
    return;
  }
  ARB_MettreComplement( PARAM_KSMIRNOV, NULL );

  if ( OBJ_choisit(PARAM_KSMIRNOV, PKS_REF_THEORIE, QUESTION) == VRAI ) { /* On compare avec une distribution thorique */
    if ( OBJ_choisit(PARAM_KSMIRNOV, PKS_MV_EXP, QUESTION) == VRAI ) { /* Valeurs exprimentales */
      m = esperance(experience);
      s = sqrt(variance(experience, m));
    } else {
      m = OBJ_LireNombre( PARAM_KSMIRNOV, KS_MOYENNE, NULL );
      s = OBJ_LireNombre( PARAM_KSMIRNOV, PKS_SIGMA, NULL );
    }

    if (experience->type == GR_REPARTITION) { /* C'est dj une fonction de rpartition */
      frep = experience;
    } else {
      frep = estime_repartition(experience);
      if (frep == NULL) {
	free(infos);
	return;
      }
    }

    /* Application du test K-S */
    switch (infos->type_distrib)
    {
    case DP_GAUSSIENNE:
      KS_gaussienne(frep, experience->nombre_point, m, s, OBJ_choisit(PARAM_KSMIRNOV, PKS_MV_EXP, QUESTION));
      break;
    }

    free(infos);
    return;
  }

  /* On compare avec une autre distribution exprimentale */
}

/* -------- Calcul de la fonction de rpartition exprimentale d'une mesure --------\
 |    Renvoie le graphe contenant la fonction de rpartition, NULL si problme      |
 |                                                                                  |
 | experience = le graphe contenant en Y les diverses ralisations de l'exprience  |
 |                                                                                  |
 | on commence par trier puis on compte la fraction de points sont infrieurs  Xi  |
 \--------------------------------------------------------------------------------- */

Graphe *estime_repartition(Graphe *experience)
{
  char *titre = NULL;
  double *y, *y_exp, total = 0.0;
  register double *x_rep, *y_rep;
  Graphe *frep;
  unsigned long i, j;

  y = (double *) calloc(experience->nombre_point, sizeof(double));
  if (y == NULL) {
    SYS_ERREUR(MANQUE_MEMOIRE, "Stats/k_smirnov.c", "kolmogorov_smirnov", "y");
    return NULL;
  }
  memcpy(y, experience->y, sizeof(double) * experience->nombre_point);

  /* Tri par fusion des donnes exprimentales */
  y_exp = TRI_Tableau(y, experience->nombre_point);
  if (y_exp == NULL) { /* Le tri a chou : manque de mmoire a priori */
    free(y);
    return NULL;
  } else
    y = y_exp;

  /* Construction de la fonction de rpartition */
  titre = TXT_genere_chaine( txt_nom_graphe_frep, experience->nom );
  frep = GRF_nouveau(experience, experience->nombre_point, GR_REPARTITION, GR_NORMAL, titre);
  free( titre ); titre = NULL;
  if (frep == NULL) {
    free(y);
    return NULL;
  }

  i = 0; j = 0;
  y_exp = y;
  x_rep = frep->x; y_rep = frep->y;
  while (i < experience->nombre_point) {
    unsigned long n_i = 0;

    *x_rep = *y_exp; /* On met le Y */
    do {
      y_exp++; /* Point suivant */
      i++;
      n_i++;
      if (i >= experience->nombre_point)
	break;
    }
    while (*y_exp == *x_rep); /* Tant qu'il a la mme valeur, on continue */

    /* On est sur le premier point diffrent */
    total += ( (double) n_i) / (double) experience->nombre_point;
    *y_rep = total;
    j++; /* Un point de plus dans le spectre calcul */
    x_rep++; y_rep++;
  }

  free(y);
  frep->nombre_point = j;
  GRF_bornes(frep, DESORDONNE, DESORDONNE);
  return frep;
}

/* -------- Prparation et utilisation de l'interface -------- */

int GER_ks(unsigned int objet, int x __attribute__((__unused__)), int y __attribute__((__unused__)))
{
  Graphe *nv_graphe;
  boolean etat;
  struct infos_KS *infos;

  infos = (struct infos_KS *) ARB_LireComplement( PARAM_KSMIRNOV );

  switch (objet)
  {
  case PKS_GRF_EXP:
    nv_graphe = GRF_choisir(GR_EPREUVE, FAUX, VRAI, NULL);
    if (nv_graphe != NULL) {
      infos->experience = nv_graphe;
      met_chaine(PARAM_KSMIRNOV, PKS_GRF_EXP, LGMAX_NOM_GRAPHES_REF,
		 infos->experience->nom, TXT_COPIER);
    }
    return VRAI;

  case PKS_REF_THEORIE: /* Rfrence = distribution thorique */
    OBJ_choisit(PARAM_KSMIRNOV, PKS_REF_THEORIE, FORCE_OUI);
    OBJ_active(PARAM_KSMIRNOV, PFS_GRF_REF, FORCE_NON);
    OBJ_active(PARAM_KSMIRNOV, PKS_TYPE_DISTRIB, FORCE_OUI);
    OBJ_active(PARAM_KSMIRNOV, PKS_MV_EXP, FORCE_OUI);
    etat = OBJ_choisit(PARAM_KSMIRNOV, PKS_MV_EXP, QUESTION);
    OBJ_active(PARAM_KSMIRNOV, KS_MOYENNE, !etat);
    OBJ_active(PARAM_KSMIRNOV, PKS_SIGMA, !etat);
    return VRAI;

  case PKS_REF_GRAPHE: /* Comparaison de deux distributions exprimentales */
    OBJ_choisit(PARAM_KSMIRNOV, PKS_REF_GRAPHE, FORCE_OUI);
    OBJ_active(PARAM_KSMIRNOV, PFS_GRF_REF, FORCE_OUI);
    OBJ_active(PARAM_KSMIRNOV, PKS_TYPE_DISTRIB, FORCE_NON);
    OBJ_active(PARAM_KSMIRNOV, KS_MOYENNE, FORCE_NON);
    OBJ_active(PARAM_KSMIRNOV, PKS_SIGMA, FORCE_NON);
    OBJ_active(PARAM_KSMIRNOV, PKS_MV_EXP, FORCE_NON);
    OBJ_choisit(PARAM_KSMIRNOV, PKS_MV_EXP, FORCE_OUI);
    return VRAI;

  case PKS_TYPE_DISTRIB: /* Quel type de distribution exprimentale ? */
    {
      int distribution;

      distribution = SS_menu_texte(txt_distributions, 1, titre_ssmenu_loi, PARAM_KSMIRNOV, PKS_TYPE_DISTRIB);
      if ( distribution >= 0 ) /* Sinon, on a annul ! */
	infos->type_distrib = distribution;
    }
    return VRAI;

  case PKS_MV_EXP: /* Paramtres exprimentaux ou postuls ? */
    etat = OBJ_choisit(PARAM_KSMIRNOV, PKS_MV_EXP, ECHANGE);
    OBJ_active(PARAM_KSMIRNOV, KS_MOYENNE, !etat);
    OBJ_active(PARAM_KSMIRNOV, PKS_SIGMA, !etat);
    return VRAI;
  }

  return FAUX;
}

void prepare_arbre_ks(void)
{
#ifndef _DEVELOPPE_
  OBJ_active(PARAM_KSMIRNOV, PKS_REF_GRAPHE, FORCE_NON);
  OBJ_active(PARAM_KSMIRNOV, PKS_TYPE_DISTRIB, FORCE_NON);
  OBJ_active(PARAM_KSMIRNOV, PKS_MV_EXP, FORCE_NON);
#endif

  arbres[PARAM_KSMIRNOV].fonction.arbre = GER_ks;
  arbres[PARAM_KSMIRNOV].drapeaux.prepare = VRAI;
}
