/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2002, Emmanuel CURIS

   Calcul thorique de l'absorption des lments
   Utilisation de la mthode de Victoreen
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "systeme.h"

#include "chimie_utils.h" /* Donnes de base sur les atomes */
#include "absorption.h"
#include "science.h" /* Pour la conversion E <-> longueur d'onde */

/* ~~~~~~~~ Variables locales : table des coefficients de Victoreen ~~~~~~~~ */

#define MAX_Z_VICTOREEN POLONIUM /* Donnes inconnues au-del */

/*  Coefficients C de Victoreen aprs le seuil K
    Origine : International Tables for X-Ray + Teo, EXAFS : basic principle and data analysis
              1982
	      Valeurs en kX / 1,00202 ~ angstrom,  un pouillime prs */

static float CVict_K[POLONIUM] = 
{
  0.0127, 0.0514,                                     /* H                            He */
  0.150, 0.365, 0.609, 1.22, 2.05, 3.18, 4.60, 6.51,  /* Li Be         B  C  N  O  F  Ne */
  8.67, 11.3, 14.4, 18.2, 22.6, 27.6, 33.4, 40,       /* Na Mg         Al Si P  S  Cl Ar */
  47.4, 55.8,                                         /* K  Ca                           */
  65.2, 75.6, 86.9, 99, 112, 126, 141, 158, 176, 195, /*       Sc  Zn                   */
  216, 238, 262, 287, 314, 343,                       /*               Ga Ge As Se Br Kr */
  374, 406,                                           /* Rb Sr                           */
  441, 477, 515, 555, 597, 641, 686, 734, 784, 835,   /* 2 srie de transition */
  889, 944, 1000, 1060, 1120, 1180,                   /*               In Sn Sb Te I  Xe */
  1250, 1310,                                         /* Cs Ba                           */
  1380, 1450, 1418, 1489, 1569, 1645, 1744, 1804, 1910, 1996, 2103, 2213, 2336, 2429, /* Lanthanides */
  2557, 2666, 2796, 2923, 3064, 3182, 3339, 3485, 3656, 3800, /* Troisime srie de transition */
  3946, 4115, 4312, -1.0 /* Tl, Pb, Bi, Po */
};

/*  Coefficients D de Victoreen aprs le seuil K
    Origine : International Tables for X-Ray + Teo, EXAFS : basic principle and data analysis
              1982
	      Valeurs en kX / 1,00202 ~ angstrom,  un pouillime prs */

static float DVict_K[POLONIUM] = 
{
  0.466e-5, 7.52e-5, /* H, He */
  0.494e-3, 2.13e-3, 0.00451, 0.0142, 0.0317, 0.0654, 0.112, 0.206, /* Li -> Ne */
  0.330, 0.539, 0.803, 1.10, 1.55, 2.18, 3.03, 4.18, /* Na -> Ar */
  5.59, 7.56, /* K, Ca */
  9.81, 12.3, 15.1, 18.2, 22.3, 27.2, 33.2, 40.1, 48.3, 57.5, /* Sc -> Zn */
  68.6, 81.1, 95.4, 112, 130, 151, /* Ga, Ge, As, Se, Br, Kr */
  174, 200, /* Rb, Sr */
  229, 261, 296, 336, 379, 427, 479, 537, 599, 667, /* Y -> Cd */
  741, 821, 908, 1000, 1100, 1210, /* In, Sn, Sb, Te, I, Xe */
  1330, 1460, /* Cs, Ba */
  1590, 1740, 1831, 1993, 2175, 2362, 2591, 2773, 3035, 3278, 3572, 3881, 4229, 4538, /* Lanthanides */
  4928, 5299, 5727, 6170, 6662, 7126, 7697, 8925, 8925, 9542, /* Troisime srie de transition - ?? pour Pt = Au, Teo */
  9542, 10189, 10925, 11766 /* Tl, Pb, Bi, Po */
};

/*  Coefficients de Victoreen entre le seuil L1 et le seuil K
    Origine : International Tables for X-Ray
              1982
	      Valeurs en kX / 1,00202 ~ angstrom,  un pouillime prs */

static float CVict_L1[POLONIUM - SCANDIUM] = /* Coef. C */
{
  5.15, 6.14, 7.24, 8.51, 9.95, 11.6, 13.4, 15.6, 17.8, /*       Ti V  Cr Mn Fe Co Ni Cu Zn */
  20.2, 22.7, 25.3, 28.0, 30.9, 33.9,                   /*                                  Ga Ge As Se Br Kr */
  37.1, 40.5,                                           /* Rb Sr */
  44.1, 17.9, 51.9, 56.2, 60.7, 65.5, 70.5, 75.8, 81.4, 87.4, /* 2e serie de transition */
  93.6, 100, 107, 114, 122, 130,                        /*                                  In Sn Sb Te I  Xe */
  139, 147,                                             /* Cs Ba */
  157, 166, 176, 187, 198, 210, 222, 234, 247, 261, 275, 290, 305, 321, /* Lanthanides */
  337, 354, 372, 390, 409, 429, 449, 470, 492, 514, /* 3e srie de transition */
  540, 570, 605, -1.0 /* Tl Pb Bi Po */
};

static float DVict_L1[POLONIUM - SCANDIUM] = /* Coef. D */
{
  0.153, 0.203, 0.268, 0.344, 0.433, 0.535, 0.651, 0.779, 0.937,
  1.13, 1.37, 1.67, 2.02, 2.43, 2.92, /* Ga Ge As Se Br Kr */
  3.48, 4.14, /* Rb, Sr */
  4.88, 5.72, 6.67, 7.73, 8.91, 10.1, 11.4, 12.8, 14.3, 15.9, 
  17.7, 19.7, 21.8, 24.3, 27.1, 30.2, /* In Sn Sb Te I Xe */
  33.6, 37.3, /* Cs Ba */
  41.3, 45.6, 50.2, 55.1, 60.3, 65.8, 71.6, 77.7, 84.1, 90.8, 98.0, 105, 113, 121, /* Lanthanides */
  130, 140, 150, 161, 173, 187, 202, 219, 240, 265, /* 3e srie de transition */
  295, 333, 382, -1.0 /* Tl Pb Bi Po */
};

/*  Coefficients de Victoreen entre le seuil L2 et le seuil L1
    Origine : Teo, EXAFS : Basic Principles and Data Analysis
              1986
	      Valeurs en kX / 1,00202 ~ angstrom,  un pouillime prs */

static float CVict_L2[POLONIUM - HAFNIUM] = /* Coef. C */
{
  325., 342., 360., 375, 395, 413, 435, 453,/* Ta W  Re Os Ir Pt Au Hg             */
  471., 493., 517., -1.0                    /*                         Tl Pb Bi Po */
};

static float DVict_L2[POLONIUM - HAFNIUM] = /* Coef. D */
{
  128, 138, 150, 161, 175, 188, 204, 219,   /* Ta W  Re Os Ir Pt Au Hg             */
  234, 254, 271, -1.0                       /*                         Tl Pb Bi Po */
};

/*  Coefficients de Victoreen entre le seuil L3 et le seuil L2
    Origine : Teo, EXAFS : Basic Principles and Data Analysis
              1986
	      Valeurs en kX / 1,00202 ~ angstrom,  un pouillime prs */

static float CVict_L3[POLONIUM - HAFNIUM] = /* Coef C. */
{
  188, 199, 211, 222, 235, 248, 262, 275,/*Lu Hf Ta W  Re Os Ir Pt Au Hg                   */
  288., 303., 320., -1.0                /*                                     Tl Pb Bi Po At Rn */
};

static float DVict_L3[POLONIUM - HAFNIUM] = /* Coef D. */
{
  56.6, 62, 68, 73.9, 81, 88.2, 96.5, 104,        /*       Lu Hf Ta W  Re Os Ir Pt Au Hg                   */
  113.0, 123 , 134, -1.0                          /*                                     Tl Pb Bi Po At Rn */
};

/*  Coefficients de Victoreen entre le seuil M1 et le seuil L3
    Origine : International Tables for X-Ray
              1982
	      Valeurs en kX / 1,00202 ~ angstrom,  un pouillime prs */

static float CVict_M1[POLONIUM - TELLURE] = /* Coef. C */
{
  12.4, 14.7,     /*        I  Xe */
  17.0, 19.2,                     /* Cs Ba */
  21.5, 23.8, 26.1, 28.4, 30.7, 33.1, 35.4, 37.8, 40.3, 42.8, 45.4, 48.0, 50.7, 53.4, /* Lanthanides */
  56.2, 59.2, 62.2, 65.3, 68.4, 71.8, 75.2, 78.7, 82.3, 86.1, /* 3e srie de transition */
  90.0, 94.1, 98.3, -1.0   /*       Tl Pb Bi Po */
};

static float DVict_M1[POLONIUM - TELLURE] = /* Coef. D */
{
  0.761, 1.06,    /*       In Sn Sb Te I  Xe */
  1.38, 1.72,                     /* Cs Ba */
  2.07, 2.44, 2.83, 3.24, 3.67, 4.12, 4.61, 5.11, 5.64, 6.20, 6.79, 7.41, 8.06, 8.75, /* Lanthanides */
  9.47, 10.2, 11.0, 11.9, 12.7, 13.6, 14.6, 15.6, 16.6, 17.7, /* 3e srie de transition */
  18.9, 20.0, 21.3, -1.0   /*       Tl Pb Bi Po */
};

/* ~~~~~~~~ Ralisation pratique ~~~~~~~~ */

/* -------- Lecture des coefficients des tableaux -----------------------\
 | Renvoie VRAI si pas de problme, FAUX sinon (coefficient inconnu,...) |
 | Coefficients exprims en barn/atome, pour une nergie donne en keV   |
 |     correspond au coefficient d'absorption par effet photo-lectrique |
 |                                                                       |
 | Z     = lment pour lequel on veut les donnes                       |
 | seuil = le seuil _avant_ lequel on veut les coefficients              |
 | c, d  = les coefficients cherchs                                     |
 |   (-1.0 : donne inconnue ; -2.0 : donne n'a pas de sens)            |
 \---------------------------------------------------------------------- */

int TH_coef_Victoreen(int Z, int seuil, double *c, double *d)
{
  if ( (c == NULL) || (d == NULL) )
    return FAUX;

  if ( (Z < HYDROGENE) || (Z > MAX_Z_VICTOREEN) ) { /* lment non tabul */
    *c = -1.0; *d = -1.0; /* <=> Donnes inconnues */
    return FAUX;
  }

  switch (seuil)
  {
  case APRES_SEUIL_K:
    *c = CVict_K[Z - 1];
    *d = DVict_K[Z - 1];
    break;

  case SEUIL_K:   /* On est entre K et LI [avant K pour les lments n'ayant pas de seuil LI] */
    if ( Z <= HELIUM ) { /* Pas d'autre seuil que le K => on n'absorbe rien avant ! */
      *c = 0.0; *d = 0.0;
    } else if ( Z <= SCANDIUM ) { /* Pas de donnes dans les tables... */
      *c = -1.0; *d = -1.0;
    } else {
      *c = CVict_L1[Z - TITANE];
      *d = DVict_L1[Z - TITANE];
    }
    break;

  case SEUIL_LI:  /* On est entre LI et LII [avant LI pour les lments n'ayant pas de seuil LII] */
    if ( Z <= HELIUM ) { /* Seuil K uniquement : cette rgion n'a pas de sens */
      *c = -2.0; *d = -2.0;
    } else if ( Z <= BERYLLIUM ) { /* Pas de seuil LII => on n'absorbe rien avant le seuil LI */
      *c = 0.0; *c = 0.0;
    } else if ( Z < TANTALE ) { /* Pas de donnes dans les tables... */
      *c = -1.0; *d = -1.0;
    } else {
      *c = CVict_L2[Z - TANTALE];
      *d = DVict_L2[Z - TANTALE];
    }
    break;

  case SEUIL_LII: /* On est entre LII et LIII */
    if ( Z <= BERYLLIUM ) { /* Seuils K & LI uniquement : cette rgion n'a pas de sens */
      *c = -2.0; *d = -2.0;
    } else if ( Z <= CARBONE /* ?? Suppose 2 e- dans 2p1/2 puis 4 dans 2p3/2 */ ) { /* Pas de seuil LIII => on n'absorbe rien avant le seuil LII */
      *c = 0.0; *d = 0.0;
    } else if ( Z <= HAFNIUM) { /* Pas de donnes dans les tables... */
      *c = -1.0; *d = -1.0;
    } else {
      *c = CVict_L3[Z - TANTALE];
      *d = DVict_L3[Z - TANTALE];
    }
    break;

  case SEUIL_LIII: /* On est entre LIII et MI */
    if ( Z <= CARBONE ) { /* Seuils K, LI, LII uniquement : cette rgion n'a pas de sens */
      *c = -2.0; *d = -2.0;
    } else if ( Z <= NEON ) { /* Pas de seuil MI => on n'absorbe rien avant le seuil LIII */
      *c = 0.0; *d = 0.0;
    } else if ( Z <= TELLURE ) { /* Pas de donnes dans les tables... */
      *c = -1.0; *d = -1.0;
    } else {
      *c = CVict_M1[Z - IODE];
      *d = DVict_M1[Z - IODE];
    }
    break;
  }

  if ( (*c < 0.0) || (*d < 0.0) ) {
    return FAUX; /* Une valeur, au moins, est inconnue */
  }

  return VRAI;
}

/* -------- Calcul du coefficient d'absorption, mthode de Victoreen --------\
 |   Renvoie le coefficient d'absorption massique (?) en cm/g (?)           |
 |                                                                           |
 | energie = l'nergie o calculer (en eV)                                   |
 \-------------------------------------------------------------------------- */

double TH_Victoreen(double energie, int Z, double C, double D)
{
  double lg_onde, lg_cube;
  double mu;

  /* On convertit l'nergie en longueur d'onde */
  if (energie <= 0.0) {
    (void) fprintf(stderr, "TH_Victoreen(%g, %d, %g, %g) : paramtres absurdes\n", energie, Z, C, D);
    return 0.0;
  }
  lg_onde  = 1.0 / energie;
  lg_onde *= M_V_ANGSTROM * PLANCK * VITESSE_LUMIERE / EV_V_JOULE; /* Longueur d'onde en angstrms */
  lg_cube  = lg_onde * lg_onde * lg_onde;

  mu  = C - D * lg_onde;
  mu *= lg_cube; /* Modle de Victoreen : /rho = C * l^3 - D * l^4, D > 0 */

  return mu;
}
