/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998-2006, Emmanuel CURIS

   Module d'affichage graphique des rsultats

   Gestion des axes gradus [tout ce qui est isol !]
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "systeme.h"
#include "txt_utils.h"

#include "tracer.h"

/* ~~~~~~~~ Variables locales  ce module ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales  ce module ~~~~~~~~ */

/* ~~~~~~~~ Ralisation pratique des fonctions ~~~ */

/* -------- Initialiser un axe -------- */

void AXE_Initialiser(InfosAxe *axe, boolean vertical)
{
  if ( NULL == axe ) return;

  /* Ni lgende, ni unit */
  axe->legende = NULL;
  axe->unite = NULL;

  /* Affichage complet, pour l'cran, par dfaut */
  axe->proprietes = AXE_AVEC_BORNES | AXE_AVEC_GRADUATIONS | AXE_AVEC_LEGENDE | AXE_AVEC_GRILLE | AXE_VISIBLE | AXE_ECRAN;
  if ( VRAI == vertical ) {
    axe->proprietes |= AXE_VERTICAL;
  }

  /* Graduation par dfaut : dcimale automatique, 10 pixels max, intrieur seulement */
  axe->type_graduation = GRADUATION_DECIMALE;
  axe->finesse = 0;
  axe->min = 0.0;
  axe->max = 0.0;
  axe->pas_graduation = -1.0;
  axe->multiple = 0;
  axe->debord_gauche =  0.0;
  axe->debord_droite = 10.0;

  /* Couleurs par dfaut */
  axe->couleur_legende = NOIR;
  axe->couleur_graduations = NOIR;
  axe->couleur_axe = VIOLET;
  axe->couleur_grille = GRIS_FONCE;
  axe->style_grille = BI_TIRET_COURT;
}

/* -------- Librer COMPLTEMENT un axe --------
   axe          : l'axe  librer
   av_structure :  VRAI -> on libre la structure aussi
                   FAUX -> on n'efface que le contenu
 */

void AXE_Liberer(InfosAxe *axe, boolean av_structure)
{
  if ( NULL == axe ) return;

  /* On efface les textes */
  free( axe->legende ); axe->legende = NULL;
  free( axe->unite ); axe->unite = NULL;

  if ( FAUX == av_structure ) return;

  free( axe ); axe = NULL;
}

/* -------- Lire la lgende associe  l'axe --------
   Renvoie une chane  librer avec free, NULL en cas d'erreur
   La lgende est construite  partir du texte, de l'unit et du facteur multiplicatif
 */

char *AXE_LireLegende(InfosAxe *axe)
{
  char *legende = NULL;

  if ( NULL == axe ) return NULL;

  if ( axe->legende != NULL ) {
    if ( axe->unite != NULL ) {
      legende = TXT_genere_chaine( "%s (%s)", axe->legende, axe->unite );
    } else {
      legende = TXT_genere_chaine( "%s [Sans unit"e_AIGU"]", axe->legende );
    }
  } else {
    if ( axe->unite != NULL ) {
      legende = TXT_genere_chaine( "Unit"e_AIGU" : %s", axe->unite );
    }
  }

  if ( ( axe->multiple != 0 ) && ( axe->multiple != ' ' ) ) {
    OPTION_A_FAIRE( "Prendre en compte le facteur multiplicatif" );
  }

  return legende;
}

/* -------- Assigner une lgende  un axe -------- */

int AXE_MettreLegende(InfosAxe *axe, const char *texte_legende)
{
  if ( NULL == axe ) return -1;
  if ( NULL == texte_legende ) { /* Supprimer la lgende */
    free( axe->legende ); axe->legende = NULL;
    return 0;
  }

  if ( axe->legende != NULL ) {
    if ( strcmp( axe->legende, texte_legende ) == 0 ) { /* Lgende inchange */
      return 0;
    }

    free( axe->legende ); axe->legende = NULL;
  }

  /* On copie ; devra tre libr avec free() */
  axe->legende = strdup( texte_legende );
  return 0;
}

/* -------- Assigner une unit  un axe -------- */

int AXE_MettreUnite(InfosAxe *axe, const char *nv_unite)
{
  if ( NULL == axe ) return -1;
  if ( NULL == nv_unite ) { /* Supprimer l'unit */
    free( axe->unite ); axe->unite = NULL;
    return 0;
  }

  if ( axe->unite != NULL ) {
    if ( strcmp( axe->unite, nv_unite ) == 0 ) { /* Unit inchange */
      return 0;
    }

    free( axe->unite ); axe->unite = NULL;
  }

  /* On copie ; devra tre libr avec free() */
  axe->unite = strdup( nv_unite );
  return 0;
}

/* -------- Contrle de la validit des bornes d'un axe -------- */

int AXE_ValiderBornes(InfosAxe *axe)
{
  if ( NULL == axe ) return -1;

  /* Y a-t-il des valeurs aberrantes (NaN) ? */
  if ( ( is_nan(axe->min) ) || ( is_nan(axe->max) ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Une des bornes est un NaN [%g  %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, axe->min, axe->max );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -3;
  }

  /* Y a-t-il des valeurs infinies ? */
  if ( ( is_inf(axe->min) ) || ( is_inf(axe->max) ) ) {
    (void) fprintf(stderr, "%s, %s l.%u : Une des bornes est infinie [%g  %g]\n",
		   __FILE__, __FUNCTION__, __LINE__, axe->min, axe->max );
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return -2;
  }

  /* Les bornes sont-elles bien diffrentes ? */
  if ( axe->min == axe->max ) {
    AVERTISSEMENT( "Bornes identiques [%g]", axe->min );
    return +1;
  }

  /* Les bornes sont-elles dans le bon sens ? */
  if ( axe->min > axe->max ) {
    double x;

    AVERTISSEMENT( "Bornes inverses [min = %g, max = %g]. On corrige.", axe->min, axe->max );
    x = axe->min; axe->min = axe->max; axe->max = x;
    return +2;
  }

  /* L'analyse n'a pas dtect d'erreur */
  return 0;
}

/* -------- Calculer le facteur d'chelle pour un axe (horizontal ou vertical) --------
    Renvoie 0 en cas d'erreur [+ message sur stderr]
            le facteur d'chelle, pour convertir des vrais coordonnes en coordonnes cran.
	                          il est exprim en vrai / pixel [comme partout dans LASE !]
 */

double AXE_Echelle(InfosAxe *axe, int debut_ecran, int fin_ecran)
{
  double echelle;
  int retour;

  /* On contrle les bornes [et que axe != NULL] */
  retour = AXE_ValiderBornes( axe );
  if ( retour != 0 ) {
    (void) fprintf(stderr, "%s, %s l.%u : Erreur dans les bornes de l'axe, code %d\n",
		   __FILE__, __FUNCTION__, __LINE__, retour);
    return 0.0;
  }

  /* On contrle que les coordonnes cran sont logiques */
  if ( fin_ecran == debut_ecran ) {
    AVERTISSEMENT( "Axe rduit  un point [%d]", debut_ecran );
    return 0.0;
  }

  /* On peut faire le calcul */
  echelle = ( axe->max - axe->min ) / ( (double) ( fin_ecran - debut_ecran ) );

  /* Pour rel -> cran :  utiliser dans TRC_ConvertitX_PvE() ou TRC_ConvertitY_PvE() */
  /* Pour cran -> rel :  utiliser dans TRC_ConvertitX_EvP() ou TRC_ConvertitY_EvP() */
  return echelle;
}
