/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1998, Emmanuel CURIS

   Module d'affichage graphique des rsultats

   Trac des images "classiques"
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "math_utils.h"
#include "systeme.h"

#include "tracer.h"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static void TRC_PolynomeQuelconque(Fenetre *fenetre, InfosRegion2D *region, double echelle_X, double echelle_Y,
				   unsigned int degre, double *coefficients, unsigned long nombre_points, double *X);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

/* -------- Tracer, dans une rgion 2D, un polynome de degr quelconque --------
   Ne renvoie rien

   fenetre       : la fentre dans laquelle se feront les affichages
   region        : la rgion 2D dans laquelle dessiner
   couleur       : couleur de trac
   style_ligne   : le style de ligne  utiliser [0 : pas de ligne]
   style_marque  : le style de marque  utiliser [0 : pas de marque]
   degre         : le degr du polynome
   coefficients  : la table des coefficients ; Y = c[ 0 ] + X * c[ 1 ] + X * c[ 2 ] + ...
   nombre_points : le nombre de points  tracer
   X             : les X pour ces points [classs par ordre croissant]
*/

void TRC_Polynome(Fenetre *fenetre, InfosRegion2D *region, unsigned int couleur, unsigned int style_ligne, unsigned int style_marque,
		  unsigned int degre, double *coefficients, unsigned long nombre_points, double *X)
{
  double echelle_X, echelle_Y;
  int retour;
  int x_ecran, y_ecran;

  if ( NULL == region ) {
    (void) fprintf(stderr, "%s,%s l.%u : Aucune rgion de trac indique\n",
		   __FILE__, __FUNCTION__, __LINE__);
    FIN_ANTICIPEE( FIN_ERREUR_GRAVE );
    return;
  }
  if ( NULL == X ) {
    AVERTISSEMENT( "Aucune table de X indique, n = %lu points", nombre_points );
    return;
  }

  /* On place les paramtres de dessin */
  TRC_Couleur( couleur );

  /* On calcule les chelles [sur les axes principaux] */
  retour = TRC_Z2D_Echelles( region, 0, &echelle_X, &echelle_Y );
  if ( retour != 0 ) {
    AVERTISSEMENT( "Problme dans le calcul des chelles, code %d", retour );
    return;
  }

  /* Trac de la version avec les lignes */
  if ( style_ligne > 0 ) {
    TRC_Style( style_ligne );

    if ( 0 == degre ) {
      /* Polynme constant : ligne horizontale ! */
      double y;

      y = coefficients[ 0 ];
      /* Est-ce dans la zone reprsente ? */
      if( ( y >= region->axe_Y.min ) && ( y <= region->axe_Y.max ) ) {
	int y_ecran;

	y_ecran = TRC_ConvertitY_PvE(  y, region->axe_Y.min, region->y_BG, echelle_Y );

	/* Ligne horizontale  la bonne hauteur */
	TRC_LigneH( fenetre, region->x_BG, y_ecran, region->largeur );
      }
    } else if ( 1 == degre ) {
      /* Droite */
      double x0, x1, y0, y1;

      /* Les deux points extrmes */
      x0 = region->axe_X.min; y0 = coefficients[ 0 ] + x0 * coefficients[ 0 ];
      x1 = region->axe_X.max; y1 = coefficients[ 0 ] + x1 * coefficients[ 0 ];

      /* On restreint  la zone en Y tudie */
      retour = TRC_LimiterSegment( region->axe_X.min, region->axe_X.max, region->axe_Y.min, region->axe_Y.max,
				   &x0, &y0, &x1, &y1 );
      if ( 0 == retour ) {
	int x0_ecran, y0_ecran, x1_ecran, y1_ecran;

	/* On convertit en points cran */
	x0_ecran = TRC_ConvertitX_PvE( x0, region->axe_X.min, region->x_BG, echelle_X );
	y0_ecran = TRC_ConvertitY_PvE( y0, region->axe_Y.min, region->y_BG, echelle_Y );

	x1_ecran = TRC_ConvertitX_PvE( x1, region->axe_X.min, region->x_BG, echelle_X );
	y1_ecran = TRC_ConvertitY_PvE( y1, region->axe_Y.min, region->y_BG, echelle_Y );

	/* On trace le segment */
	TRC_Ligne( fenetre, x0_ecran, y0_ecran, x1_ecran, y1_ecran );
      }
    } else {
      /* Polynme quelconque */
      TRC_PolynomeQuelconque( fenetre, region, echelle_X, echelle_Y, degre, coefficients, nombre_points, X );
    }
  }

  /* Trac de la version avec les marques : "facile" */
  if ( style_marque > 0 ) {
    unsigned long i;
    double *x, y;

    TRC_Style( BI_CONTINU );
    TRC_Epaisseur( 1 );
    for ( i = 0, x = X ; i < nombre_points ; i++, x++ ) {
      if ( ( *x < region->axe_X.min ) || ( *x > region->axe_X.max ) ) {
	/* X en-dehors de la rgion tudie */
	continue;
      }

      /* On calcule la valeur du polynme */
      y = Polynome( *x, degre, coefficients );
      if ( ( y < region->axe_Y.min ) || ( y > region->axe_Y.max ) ) {
	/* Y en-dehors de la rgion tudie */
	continue;
      }

      /* On convertit en coordonnes cran */
      x_ecran = TRC_ConvertitX_PvE( *x, region->axe_X.min, region->x_BG, echelle_X );
      y_ecran = TRC_ConvertitY_PvE(  y, region->axe_Y.min, region->y_BG, echelle_Y );

      /* On place la marque en ce point */
      TRC_Marque( fenetre, x_ecran, y_ecran, style_marque, 6 );
    }
  }
}

/* -------- Le trac proprement dit, en lignes, pour un polynme quelconque
   Pour tracer un polynme en ligne, il faut que les X soient tris par ordre croissant, sinon absurde
   ==> on le suppose ici !
   -------- */

static void TRC_PolynomeQuelconque(Fenetre *fenetre, InfosRegion2D *region, double echelle_X, double echelle_Y,
				   unsigned int degre, double *coefficients, unsigned long nombre_points, double *X)
{
  double *x = X;
  double y;
  unsigned long i = 0, compteur = 0;
  Point_Ecran points[ BI_NOMBRE_MAXIMAL_POINTS ];

  /* On saute tous les points de X trop faible */
  while ( *x < region->axe_X.min ) {
    x++; i++;
    if ( i >= nombre_points ) {
      return; /* Fini => aucun X dans la rgion considre ! */
    }
  }

  /* Si on est ici, c'est que les X commencent  convenir */
  while ( ( *x <= region->axe_X.max ) && ( i < nombre_points ) ) {
    /* Calcul du premier point du segment  afficher */
    y = Polynome( *x, degre, coefficients );

    /* On cherche les points qui ont des Y qui conviennent */
    while ( ( y < region->axe_Y.min ) || ( y > region->axe_Y.max ) ) {
      /* On affiche le segment prcdent, s'il existe */
      if ( compteur > 0 ) {
	TRC_LigneBrisee( fenetre, compteur, points );
	compteur = 0;
      }

      x++; i++;
      if ( i >= nombre_points ) {
	return; /* Fini */
      }
      if ( *x > region->axe_X.max ) {
	AVERTISSEMENT( "X=%g > X_max = %g, bizarre ?", *x, region->axe_X.max );
	return;
      }

      y = Polynome( *x, degre, coefficients );
    }

    /* Ici, Y est dans la zone aussi => on le convertit en coordonn cran */
    points[ compteur ].x = TRC_ConvertitX_PvE( *x, region->axe_X.min, region->x_BG, echelle_X );
    points[ compteur ].y = TRC_ConvertitY_PvE(  y, region->axe_Y.min, region->y_BG, echelle_Y );
    compteur++;

    if ( BI_NOMBRE_MAXIMAL_POINTS == compteur ) {
      /* On a atteint le maximum de points traables en une fois */
      TRC_LigneBrisee( fenetre, compteur, points );

      /* On garde le dernier point, qui devient le premier, pour qu'il n'y ait pas de trou ! */
      points[ 0 ] = points[ BI_NOMBRE_MAXIMAL_POINTS - 1 ];
      compteur = 1;
    }

    /* On passe au point suivant */
    x++; i++;
  }

  /* Ici, x > x_max => on a termin */
  if ( compteur > 0 ) {
    /* On trace le segment final */
    TRC_LigneBrisee( fenetre, compteur, points );
  }
}
