/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (C) 1996 - 1999, Emmanuel CURIS

   Chargement des fichiers au format interne
   Voir sauver.c pour la description des formats
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef VERSION_SIMPLIFIEE_
# define _LIMITE_ 1
#endif

#include "systeme.h"

#include "serie.h"

#include "atomes.h"   /* Pour les molcules et tout a */
#include "c_formel.h"
#include "chemins.h"
#include "couches.h"
#include "cristallo.h"
#include "experience.h"
#include "feff.h"
#include "fichiers.h"
#include "filtrage.h" /* Pour chargement de Infos_FonctionApodisation */
#include "images.h"   /* Pour chargement des images */
#include "matrices.h" /* Pour chargement de la matrice de covariance */
#include "modeles.h"
#include "theorie.h"  /* Pour les types de modles de lpm */

#include "science.h" /* pour M_PI */

#include "charger.txt"

/* ~~~~~~~~ Variables locales ~~~~~~~~ */

#define VERIFIE(x, r, ch) if (n_lu < (x)) \
                          { (void) SYS_ERREUR(ERREUR_LECTURE, txt_fichier_courant, fichier, NULL); \
                            (void) fprintf(stderr, "VERIFIE, chargement : %s\n", ch); \
                            return (r); }

/* ~~~~~~~~ Fonctions locales ~~~~~~~~ */

static int       ch_entete    (FILE *fichier, int *type);

static int       ch_absorbeur (FILE *fichier, int version, const char *nom_fichier, Absorbeur *absorbeur, Serie_Graphes *reference, Molecule *molecule);
static Atome    *ch_atome     (FILE *fichier, int version, const char *nom_fichier, Molecule *molecule);
static Chemin   *ch_chemin    (FILE *fichier, int version, const char *nom_fichier, struct modele_ajuste *modele,
			       Serie_Graphes *serie_amp, Serie_Graphes *serie_contrib, Molecule *molecule);
static Couche   *ch_couche    (FILE *fichier, int version, const char *nom_fichier, Molecule *molecule,
			       Serie_Graphes *reference, Serie_Graphes *serie_contrib, struct modele_feff *modele);
static Fragment *ch_fragment  (FILE *fichier, int version, const char *nom_fichier, Serie_Graphes *reference, Molecule *molecule);
static Liaison  *ch_liaison   (FILE *fichier, int version, const char *nom_fichier, Molecule *molecule);
static struct modele_ajuste *ch_modele     (FILE *fichier, int version, const char *nom_fichier, Serie_Graphes *graphes, void *complement_1);
static struct modele_feff   *ch_modele_feff(FILE *fichier, int version, const char *nom_fichier, Molecule *molecule);
static Molecule *ch_molecule  (FILE *fichier, int version, const char *nom_fichier );
static int       ch_variable  (FILE *fichier, int version, const char *nom_fichier, Variable *variable);

/* ~~~~~~~~ Ralisation pratique ~~~~~ */

#ifndef VERSION_SIMPLIFIEE_
/* -------- Fonction gnrale d'appel du chargement -------- */

Graphe *charger(char *nom) /* On charge un fichier LASE contenant n'importe quoi (prfrentiellement un graphe) */
{
  char **liste_noms;
  Graphe *charge = NULL, *nv_courant = NULL;
  unsigned int nbr_fichier, i;

  ENTREE_FONCTION( __FUNCTION__, "nom=%s", nom );

  if (nom == NULL) { /* L'utilisateur va choisir le spectre */
    liste_noms = CH_selecteur( "*.*", txt_titre_selecteur, NOMBRE_TYPES_GRAPHES, extensions, &nbr_fichier, principale );
    if (liste_noms == NULL) {
      SORTIE_FONCTION( __FUNCTION__, "%p [Aucun fichier choisi]", NULL );
      return NULL; /* Aucun fichier  traiter */
    }
  } else { /* Chargement automatique... */
    nbr_fichier = 1;
    liste_noms = (char **) malloc( sizeof(char *) );
    liste_noms[0] = (char *) malloc( (strlen(nom) + 1) * sizeof(char) ); /* strlen ne donne pas le 0 final */
    strcpy(liste_noms[0], nom);
  }

  for (i = 0 ; i < nbr_fichier ; i++) {
    charge = ch_interne(liste_noms[i]);
    if ( charge != NULL )
      nv_courant = charge;
    free( liste_noms[ i ]); liste_noms[ i ] = NULL;
  }
  free( liste_noms ); liste_noms = NULL;

  /* On fait afficher ce graphe */

  if ( nv_courant != NULL ) /* Il y a au moins un graphe charg => devient le graphe courant */
    GRF_aller( nv_courant );

  /* On refait les ascenseurs pour les graphes : srement il y a eu des changements... */
  I_PrincipaleGRF_RefaireAscenseurs( VRAI, VRAI );

  SORTIE_FONCTION( __FUNCTION__, "%p", charge );
  return charge; /* Adresse du spectre charg */
}
#endif /* ndef VERSION_SIMPLIFIEE_ */

/* -------- Chargement au format interne, compatible v. 1.3 -------- */

Graphe *ch_interne(char *nom)
{
  int version, type;
  FILE *fichier;
  Graphe *grf_charge = NULL;

  fichier = fopen(nom, "rb");
  if ( NULL == fichier ) {
    (void) SYS_ERREUR(ERREUR_LECTURE, nom, NULL, NULL);
    return NULL;
  }

  version = ch_entete( fichier, &type );
#ifdef _DEVELOPPE_
  printf("Chargement d'un fichier de version %d\n", version);
#endif
  if ( 0 == version ) { /* Problme  la lecture */
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de charger le fichier %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
  } else if (version < 10) { /* Versions antdiluviennes */
    (void) BI_ERR_alerte(NULL, &ERR_VersionInterdite, VRAI, version / 10.0);
#ifndef _DEVELOPPE_
  } else if (version > VERSION) {
    (void) BI_ERR_alerte(NULL, &ERR_LaseTropVieux, VRAI, version / 10.0);
#endif
  } else {
    HIST_Titre(4, "Chargement d'un fichier", "E/S");
    HIST_Texte(4, 0, "Fichier : %s\n", nom);
    
    if ( FICHIER_GRAPHE == type ) { /* C'est un seul graphe */
      /* On charge le graphe */
      grf_charge = GRF_Charger(fichier, version, nom, serie_courante);
    } else if ( FICHIER_IMAGE == type ) {
      struct image *image;

      image = IMG_Charger( fichier, version, nom );
      IMG_AllerImage( image );
    } else if ( FICHIER_MOLECULE == type ) {
      MOL_aller( ch_molecule(fichier, version, nom) );
    } else if ( FICHIER_EXPERIENCE == type ) {
      struct experience *experience;

      experience = EXP_ChargerExperience( fichier, version, nom );
      if ( experience != NULL ) {
	if ( NULL != experience_courante ) {
	  (void) fprintf(stderr, "%s,%s l.%u :  FAIRE --- effacer proprement l'exprience prcdente\n",
			 __FILE__, __FUNCTION__, __LINE__);
	}
	experience_courante = experience;
      }
    } else if ( FICHIER_SERIE == type ) {
      Serie_Graphes *serie;

      serie = Serie_Charger( fichier, version, nom );
      /* On va  la srie ainsi charge */
      I_SERIE_GRF_Aller( serie );
    } else {
      (void) fprintf(stderr, "ch_interne : type de donnes inconnu [%d]\n", type);
    }
  }

  /* Plus besoin du fichier */
  (void) fclose( fichier ); fichier = NULL;

  /* On renvoie le graphe qui a finalement t charg */
  return grf_charge;
}

/* -------- Chargement d'un modle structural 3d -------- */

Molecule *MOL_charger(char *nom)
{
  int version, type;
  FILE *fichier;
  Molecule *mol_chargee = NULL;

  if (nom == NULL) return NULL;

  fichier = fopen(nom, "rb");
  if (fichier == NULL) {
#ifdef _DEVELOPPE_
    printf("MOL_charger - %s : fichier = NULL\n", nom);
#endif
    (void) SYS_ERREUR(ERREUR_LECTURE, nom, NULL, NULL);
    return NULL;
  }

  version = ch_entete(fichier, &type);
  if (version == 0) { /* Erreur au chargement -> fichier dj ferm */
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de charger le fichier %s\n",
		   __FILE__, __FUNCTION__, __LINE__, nom);
  } else if ( (version < 15) || ( type != FICHIER_MOLECULE ) ) {
    (void) BI_ERR_alerte(NULL, &ERR_FichierGraphe, VRAI);
#ifndef _DEVELOPPE_
  } else if (version > VERSION) {
    (void) BI_ERR_alerte(NULL, &ERR_LaseTropVieux, VRAI, version / 10.0);
#endif
  } else {
    /* Ici, on est donc prt  charger un modle */
    HIST_Titre(4, "Chargement d'un fichier", "E/S");
    HIST_Texte(4, 0, "Fichier : %s\n", nom);

    mol_chargee = ch_molecule(fichier, version, nom);
    MOL_aller( mol_chargee );
  }

  /* Plus besoin du fichier */
  (void) fclose( fichier ); fichier = NULL;

  return mol_chargee;
}

static Molecule *ch_molecule(FILE *fichier, int version, const char *nom_fichier)
{
  char *nom_molecule, flg_cell;
  unsigned int lg_titre, n_item, i;
  Molecule *molecule = NULL;
  size_t n_lu;
  Serie_Graphes *references;

  n_lu = fread(&lg_titre, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, NULL, "ch_molecule - lg_titre");
  nom_molecule = (char *) calloc(sizeof(char), lg_titre + 1);
  if (nom_molecule == NULL) {
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_molecule", "nom_molecule");
    return NULL;
  }
  n_lu = fread(nom_molecule, sizeof(char), lg_titre, fichier);
  if (n_lu < lg_titre) {
    (void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
    free(nom_molecule);
    return NULL;
  }
  nom_molecule[lg_titre] = 0;
  /* On cre la molcule */
  molecule = MOL_creer( nom_molecule );
  free( nom_molecule ); nom_molecule = NULL;
  if ( NULL == molecule ) {
    return NULL;
  }

  if ( version >= 18 ) { /* Version 1.8 et au-del */
    /* Chargement des paramtres de maille */
    n_lu = fread(&flg_cell, sizeof(char), 1, fichier);
    if (n_lu < 1) {
      (void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
      MOL_effacer(molecule, VRAI, VRAI);
      return NULL;
    }
    if (flg_cell == VRAI) {
      double a, b, c, alpha, beta, gamma;
      int grp_esp;

      n_lu  = fread(&(a), sizeof(double), 1, fichier);
      n_lu += fread(&(b), sizeof(double), 1, fichier);
      n_lu += fread(&(c), sizeof(double), 1, fichier);
      n_lu += fread(&(alpha), sizeof(double), 1, fichier);
      n_lu += fread(&(beta), sizeof(double), 1, fichier);
      n_lu += fread(&(gamma), sizeof(double), 1, fichier);
      if (n_lu < 6) {
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
	MOL_effacer(molecule, VRAI, VRAI);
	return NULL;
      }
      n_lu  = fread(&(grp_esp), sizeof(int), 1, fichier);
      if (n_lu < 1) {
	(void) SYS_ERREUR(ERREUR_LECTURE, nom_fichier, fichier, NULL);
	MOL_effacer(molecule, VRAI, VRAI);
	return NULL;
      }

      /* Actuellement, codage en degrs dans le fichier */
      alpha *= M_PI / 180.0; beta *= M_PI / 180.0; gamma *= M_PI / 180.0;

      molecule->maille = CRIS_CreerMaille();
      if ( molecule->maille != NULL ) {
	molecule->maille->a = a; molecule->maille->b = b; molecule->maille->c = c;
	molecule->maille->alpha = alpha; molecule->maille->beta = beta; molecule->maille->gamma = gamma;
      } else {
	(void) fprintf(stderr, "%s,%s l.%u : Maille non cre, paramtres perdus\n",
		       __FILE__, __FUNCTION__, __LINE__);
      }
    }
  }

  if ( ch_atome(fichier, version, nom_fichier, molecule) == NULL ) {
    return NULL;
  }

  nom_molecule = TXT_genere_chaine( "Spectres de la mol"e_AIGU"cule %s", molecule->nom);
  references = SERIE_GRF_CreerSerie( serie_courante, nom_molecule );
  free( nom_molecule ); nom_molecule = NULL;
  if ( NULL == references ) {
    MOL_cherche_barycentre(molecule, FAUX);
    MOL_initialise_vue(molecule);
    MOL_aller(molecule);
    return molecule; /* On renvoie la molcule telle qu'elle... */
  }

  HIST_Texte(4, 0, "Mol"e_AIGU"cule lue : %s\n", molecule->nom);
  n_lu = fread(&n_item, sizeof(unsigned int), 1, fichier);
#ifdef _DEVELOPPE_
  printf("ch_molecule : %u fragments  charger\n", n_item);
#endif
  for (i = 0 ; i < n_item ; i++) {
    Fragment *fragment;

    fragment = ch_fragment(fichier, version, nom_fichier, references, molecule);
    if ( fragment == NULL ) {
      MOL_cherche_barycentre(molecule, FAUX);
      MOL_initialise_vue(molecule);
      MOL_aller(molecule);
      return molecule; /* On renvoie la molcule telle qu'elle... */
    }
  }

  MOL_cherche_barycentre(molecule, FAUX); /* Comme a, on est bon jusqu' la fin des absorbeurs */
  MOL_initialise_vue(molecule);
  MOL_aller(molecule);

  n_lu = fread(&n_item, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, molecule, "MOL_charger, n_item (absorbeurs)");
  if (n_item > 0) { /* Il y a effectivement des absorbeurs... */
    /* On cre la place pour ces absorbeurs */
    molecule->absorbeurs = (Absorbeur *) calloc(n_item, sizeof(Absorbeur));
    if (molecule->absorbeurs == NULL) {
      (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "MOL_charger", "molecule->absorbeurs");
      return molecule;
    }
#ifdef _DEVELOPPE_
    printf("Chargement de %d absorbeurs rservs en %p\n", n_item, (void *) (molecule->absorbeurs));
#endif

    for (i = 0 ; i < n_item ; i++) {
      if ( ch_absorbeur(fichier, version, nom_fichier, molecule->absorbeurs + i, references, molecule) == FAUX ) {
	MOL_cherche_barycentre(molecule, FAUX);
	MOL_initialise_vue(molecule);
	return molecule; /* On renvoie la molcule telle qu'elle... */
      }
      molecule->nbr_absorbeurs++;
    }
    /* Une fois les absorbeurs chargs, on refait la liste, pour qu'ils soient visibles
       Sert surtout si l'on charge la premire molcule, en fait... */
    MOL_PrepareListe();
  }

  if (version >= 21) { /* Schma de liaisons  partir de la version 2.1 seulement */
    Liaison *liaison;

    n_lu = fread(&n_item, sizeof(unsigned int), 1, fichier);
    VERIFIE(1, molecule, "MOL_charger, n_item (liaisons)");

#ifdef _DEVELOPPE_
    printf("MOL_charger : on va charger %u liaisons\n", n_item);
#endif
    
    for (i = 0 ; i < n_item ; i++) {
      liaison = ch_liaison(fichier, version, nom_fichier, molecule);
      if ( liaison == NULL ) {
	MOL_cherche_barycentre(molecule, FAUX);
	MOL_initialise_vue(molecule);
	return molecule; /* On renvoie la molcule telle qu'elle... */
      }
    }
  }

  MOL_cherche_barycentre(molecule, FAUX);
  MOL_initialise_vue(molecule);
  MOL_aller(molecule);

  /* On efface les sries qui n'ont pas t remplies */
  if ( 0 == references->nbr_graphes ) {
    SERIE_GRF_EffacerSerie( references, VRAI );
  }

  return molecule;
}

/* ~~~~~~~~ Les briques pour le chargement ~~~~~~~~
   Reoivent toutes le numro de version du fichier pour adapter le chargement */

/* -------- Chargement d'un graphe avec flag de prsence -------- */

inline struct graphe *CH_Graphe(FILE *fichier, int version, const char *nom_fichier, struct serie *serie, boolean isole)
{
  Graphe *graphe;
  short int drapeau;
  int n_lu;

  if ( NULL == fichier ) return NULL;

  n_lu = fread( &drapeau, sizeof(short int), 1, fichier );
  if ( n_lu != 1 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de charger le drapeau de prsence\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  if ( FAUX == drapeau ) return NULL; /* Pas de graphe */
  if ( NULL == serie ) serie = serie_courante;
  if ( ( serie != NULL ) && ( VRAI == serie->drapeaux.protegee ) ) { /* On cre alors une nouvelle srie */
    serie = SERIE_GRF_CreerSerie( serie, "Chargement..." );
  }

  graphe = GRF_Charger( fichier, version, nom_fichier, serie );
  if ( NULL == graphe ) {
    (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans GRF_Charger()\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  if ( VRAI == isole ) {
    GRF_IsolerGraphe( serie, graphe );
  }

  return graphe;
}

/* -------- Chargement d'une image avec flag de prsence -------- */

inline struct image *CH_Image(FILE *fichier, int version, const char *nom_fichier, boolean isolee )
{
  struct image *image = NULL;
  short int drapeau;
  int n_lu;

  if ( NULL == fichier ) return NULL;

  n_lu = fread( &drapeau, sizeof(short int), 1, fichier );
  if ( n_lu != 1 ) {
    (void) fprintf(stderr, "%s,%s l.%u : Impossible de charger le drapeau de prsence\n",
		   __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  if ( FAUX == drapeau ) return NULL; /* Pas d'image */

  image = IMG_Charger( fichier, version, nom_fichier );
  if ( NULL == image ) {
     (void) fprintf(stderr, "%s,%s l.%u : ERREUR dans IMG_Charger()\n",
		    __FILE__, __FUNCTION__, __LINE__);
    return NULL;
  }

  if ( VRAI == isolee ) {
    IMG_IsolerImage( image );
  }

  return image;
}

/* -------- Chargement d'un absorbeur, stock dans celui pass en paramtre */

static int ch_absorbeur(FILE *fichier, int version, const char *nom_fichier, Absorbeur *absorbeur, Serie_Graphes *reference, Molecule *molecule)
{
  char *titre;
  unsigned char drapeau;
  unsigned int lg_titre, n_item, i;
  unsigned long idx_atome;
  size_t n_lu;
  Variable S02, DE0;

  if (fichier == NULL) return FAUX;
  if (absorbeur == NULL) return FAUX;

  n_lu = fread(&lg_titre, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, FAUX, "ch_absorbeur - lg_titre\n");
  titre = (char *) calloc(lg_titre + 1, sizeof(char));
  if (titre == NULL) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "ch_absorbeur : chec de calloc/titre, taille demande : %u\n", lg_titre + 1);
#endif
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_absorbeur", "titre");
    return FAUX;
  }
  n_lu = fread(titre, sizeof(char), lg_titre, fichier);
  strncpy(absorbeur->nom, titre, LG_MAX_NOM_ABSORBEUR);
  free(titre);
  VERIFIE(lg_titre, FAUX, "ch_absorbeur - titre\n");
  HIST_Texte(4, 0, " ... absorbeur : %s\n", absorbeur->nom);

  n_lu = fread(&idx_atome, sizeof(unsigned long), 1, fichier);
  VERIFIE(1, FAUX, "ch_absorbeur - index atome\n");
#ifdef _DEVELOPPE_
  printf("ch_absorbeur -> ");
#endif
  absorbeur->absorbeur = ATM_index(molecule, idx_atome);
  absorbeur->absorbeur->marques.absorbeur = VRAI; /* Il est utilis comme absorbeur */

  if ( ch_variable( fichier, version, nom_fichier, &(absorbeur->contribution) ) == FAUX ) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "ch_absorbeur : impossible de charger la contribution\n");
#endif
    return FAUX;
  }

  if ( version < 23 ) { /* v. 2.3 : apparition des modles... */
    struct modele_ajuste *modele; /* Le modle contenant les chemins crs */
    Graphe *experience;

    /* S0 */
    if ( ch_variable( fichier, version, nom_fichier, &S02 ) == FAUX ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "ch_absorbeur : impossible de charger S0\n");
#endif
      return FAUX;
    }

    /* Delta E0 */
    if ( ch_variable( fichier, version, nom_fichier, &DE0 ) == FAUX ) {
#ifdef _DEVELOPPE_
      (void) fprintf(stderr, "ch_absorbeur : impossible de charger Delta E0\n");
#endif
      return FAUX;
    }

    modele = ABS_CreerModele(absorbeur, MODELE_EXAFS_CHEMINS, NULL);
    if (NULL == modele) {
      (void) fprintf(stderr, "ch_absorbeur : erreur dans ABS_CreerModele\n");
      (void) SYS_ERREUR(ERREUR_LECTURE, NULL, fichier, "Impossible de cr"e_AIGU"er le mod"e_GRAVE"le"); 
      return FAUX;
    }

    n_lu = fread(&n_item, sizeof(unsigned int), 1, fichier);
    VERIFIE(1, FAUX, "ch_absorbeur - nombre de chemins\n");
    if ( n_item > 0 ) { /* Il y a des chemins */
      Serie_Graphes *amplitudes;    /* Srie des phases et amplitudes des chemins */
      Serie_Graphes *contributions; /* Srie des contributions des chemins */
      char *titre;

      titre = TXT_genere_chaine("Phases et amplitudes de l'absorbeur %s", absorbeur->nom);
      amplitudes = SERIE_GRF_CreerSerie(reference, titre);
      free(titre);
      if ( NULL == amplitudes ) { /* On n'aura de toute faon pas assez de mmoire pour charger la suite... */
	return FAUX;
      }

      titre = TXT_genere_chaine("Contributions des chemins, absorbeur %s", absorbeur->nom);
      contributions = SERIE_GRF_CreerSerie(amplitudes, titre);
      free(titre);
      if ( NULL == contributions ) { /* On n'aura de toute faon pas assez de mmoire pour charger la suite... */
	SERIE_GRF_EffacerSerie(amplitudes, VRAI);
	return FAUX;
      }

      for (i = 0 ; i < n_item ; i++) {
#ifdef _DEVELOPPE_
	printf("ch_absorbeur : chemin %d / %d\n", i, n_item);
#endif
	if ( ch_chemin( fichier, version, nom_fichier, modele, amplitudes, contributions, molecule ) == FAUX ) {
	  return FAUX;
	}
      }

      /* On efface les sries qui n'ont pas t remplies */
      if ( 0 == amplitudes->nbr_graphes ) {
	SERIE_GRF_EffacerSerie(amplitudes, VRAI);
      }
      if ( 0 == contributions->nbr_graphes ) {
	SERIE_GRF_EffacerSerie(contributions, VRAI);
      }
    }

    n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
    VERIFIE(1, FAUX, "ch_absorbeur, drapeau d'exafs exprimental\n");
    if (drapeau == 1) {
      experience = GRF_Charger(fichier, version, nom_fichier, reference);
      if ( experience == NULL ) return FAUX;
      modele->experience = experience;
      GRF_aller(experience);
    }

    n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
    VERIFIE(1, FAUX, "ch_absorbeur, drapeau d'exafs calcul\n");
    if (drapeau == 1) {
      modele->modele = GRF_Charger(fichier, version, nom_fichier, reference);
      if ( modele->modele == NULL ) return FAUX;
    }

    n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
    VERIFIE(1, FAUX, "ch_absorbeur, drapeau de rsidu\n");
    if (drapeau == 1) {
      modele->residu = GRF_Charger(fichier, version, nom_fichier, reference);
      if ( modele->residu == NULL ) return FAUX;
    }

    return VRAI; /* Rien d'autre pour v. < 2.3 */
  }

  n_lu = fread(&n_item, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, FAUX, "ch_absorbeur, nombre de modles\n");
  for (i = 0 ; i < n_item ; i++) {
    struct modele_ajuste *modele;

    modele = ch_modele( fichier, version, nom_fichier, serie_courante, molecule );
    if (NULL == modele) {
      (void) fprintf(stderr, "ch_absorbeur : erreur dans ch_modele\n");
      return FAUX;
    }
    ABS_AjouterModele(absorbeur, modele);
  }

  return VRAI;
}

/* -------- Chargement de _tous_ les atomes  la suite
   Renvoie NULL si erreur */

static Atome *ch_atome(FILE *fichier, int version, const char *nom_fichier __attribute__((__unused__)), Molecule *molecule)
{
  unsigned char Z;
  unsigned long i = 1;
  float rayon;
  size_t n_lu;
  Atome *atome = NULL;
  Vecteur3D position;

  if (fichier == NULL) return NULL;
  if (NULL == molecule) return NULL;

#ifdef _DEVELOPPE_
  printf("-- Chargement des atomes (ch_atome) --\n");
#endif

  do {
    n_lu = fread(&Z, sizeof(unsigned char), 1, fichier);
    VERIFIE(1, NULL, "ch_atome - Z\n");
#ifdef _DEVELOPPE_
    printf(" Z = %d", (int) Z);
#endif
    if (Z == 255U) {
#ifdef _DEVELOPPE_
    printf(" -- FIN DE LA LISTE\n");
#endif
      return atome; /* Dernier atome lu */
    }

    n_lu  = fread( &(position.x), sizeof(double), 1, fichier );
    n_lu += fread( &(position.y), sizeof(double), 1, fichier );
    n_lu += fread( &(position.z), sizeof(double), 1, fichier );
    VERIFIE(3, NULL, "ch_atome - x, y ou z\n");

    if (version < 20) {
      double r;
      n_lu = fread(&r, sizeof(double), 1, fichier);
      rayon = (float) r;
    } else
      n_lu = fread(&rayon, sizeof(float), 1, fichier);
    VERIFIE(1, NULL, "ch_atome - rayon\n");

    atome = ATM_creer(molecule, atome, Z, &position, 0.0, rayon);
    if (atome != NULL)
      atome->index = i++;

    HIST_Texte(5, 0, " ... atome : %s en (%f, %f, %f)\n", ATM_Z_vers_symbole(Z), position.x, position.y, position.z);
#ifdef _DEVELOPPE_
/*     printf(" -- (%f, %f, %f), numro %lu - Rayon : %f\n", position.x, position.y, position.z, atome->index, rayon); */
#endif
  } while (Z < 255U);

  return NULL; /* On ne devrait _jamais_ arriver l */
}

/* -------- Chargement d'un chemin -------- */

static Chemin *ch_chemin(FILE *fichier, int version, const char *nom_fichier, struct modele_ajuste *modele,
			 Serie_Graphes *serie_amp, Serie_Graphes *serie_contrib, Molecule *molecule)
{
  char *titre;
  unsigned char drapeau;
  unsigned int lg_titre, n_item, i;
  size_t n_lu;
  Chemin *chemin;
  Couche *couche = NULL;

  if (fichier == NULL) return NULL;

  n_lu = fread(&lg_titre, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, NULL, "ch_chemin - lg_titre\n");
  titre = (char *) calloc(lg_titre + 1, sizeof(char));
  if (titre == NULL) {
    (void) fprintf(stderr, "ch_chemin : imposible de rserver le nom du chemin [%u caractres]\n", lg_titre + 1);
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_chemin", "titre");
    return NULL;
  }

  n_lu  = fread(titre, sizeof(char), lg_titre, fichier);
#ifdef _DEVELOPPE_
  printf("ch_chemin : chargement du chemin \"%s\"\n", titre);
#endif
  n_lu += fread(&n_item, sizeof(unsigned int), 1, fichier);
  chemin = creer_chemin(n_item);
  if (chemin == NULL) {
    free(titre);
    return NULL;
  }
  strncpy(chemin->nom, titre, LG_MAX_NOM_CHEMIN);
  free(titre);
  VERIFIE(lg_titre + 1, NULL, "ch_chemin - titre ou nombre d'atomes\n");
  HIST_Texte(5, 0, " ... chemin : %s\n", chemin->nom);
  /*  printf("Nom du chemin : %s - contient %d atomes\n", chemin->nom, n_item); */

  for (i = 0 ; i < n_item ; i++) {
    unsigned long idx_atome;

    n_lu = fread(&idx_atome, sizeof(unsigned long), 1, fichier);
    VERIFIE(1, NULL, "ch_chemin, l'un des atomes\n");
    chemin->atomes[i] = ATM_index( molecule, idx_atome );
  }

  if ( version < 23 ) { /* Stock comme des variables... et sans la couche */
    couche = COUCHE_Creer(modele->infos_modele.feff, NULL);
    if (NULL == couche) {
      (void) fprintf(stderr, "ch_chemin : erreur dans COUCHE_Creer, chemin %s\n", chemin->nom);
      return NULL;
    }

    /* Multiplicit */
    if ( ch_variable(fichier, version, nom_fichier, &(couche->N)) == FAUX ) {
      (void) fprintf(stderr, "ch_chemin : erreur dans ch_variable / multiplicit\n");
      return NULL;
    }
    chemin->multiplicite = VAR_LireValeur( &(couche->N), VARIABLE_LIRE_VALEUR_INITIALE );

    /* Longueur de chemin */
    if ( ch_variable( fichier, version, nom_fichier, &(couche->R) ) == FAUX ) {
      (void) fprintf(stderr, "ch_chemin : erreur dans ch_variable / longueur\n");
      return NULL;
    }
    chemin->pseudo_longueur = VAR_LireValeur( &(couche->R), VARIABLE_LIRE_VALEUR_INITIALE );
    chemin->longueur_reelle = chemin->pseudo_longueur; /* Pas sauv... */

    /* Debye-Waller global */
    if ( ch_variable( fichier, version, nom_fichier, &(couche->Debye_Waller) ) == FAUX ) {
      (void) fprintf(stderr, "ch_chemin : erreur dans ch_variable / Debye-Waller\n");
      return NULL;
    }
    chemin->dw_global = VAR_LireValeur( &(couche->Debye_Waller), VARIABLE_LIRE_VALEUR_INITIALE );

    couche->chemin = chemin;
    strncpy(couche->nom, chemin->nom, LG_MAX_NOM_COUCHE);
  } else {
    n_lu = fread( &(chemin->multiplicite), sizeof(double), 1, fichier);
    VERIFIE(1, NULL, "ch_chemin, multiplicit\n");

    n_lu = fread( &(chemin->longueur_reelle), sizeof(double), 1, fichier);
    VERIFIE(1, NULL, "ch_chemin, longueur totale relle\n");

    n_lu = fread( &(chemin->dw_global), sizeof(double), 1, fichier);
    VERIFIE(1, NULL, "ch_chemin, facteur de Debye-Waller\n");
  }

  n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_chemin, drapeau d'amplitude\n");
  if (drapeau == 1) {
    chemin->amplitude = GRF_Charger(fichier, version, nom_fichier, serie_amp);
    if ( chemin->amplitude == NULL ) {
      (void) fprintf(stderr, "ch_chemin : erreur dans GRF_Charger / amplitude\n");
      return NULL;
    }
    if ( couche != NULL )
      couche->amplitude = chemin->amplitude;
  }
  n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_chemin, drapeau de dphasage\n");
  if (drapeau == 1) {
    chemin->phase = GRF_Charger(fichier, version, nom_fichier, serie_amp);
    if ( chemin->phase == NULL ) {
      (void) fprintf(stderr, "ch_chemin : erreur dans GRF_Charger / phase\n");
      return NULL;
    }
    if ( couche != NULL )
      couche->phase = chemin->phase;
  }

  if (version < 17) /* Pas sauv avant... on a donc fini */
    return chemin;

  if (version < 23) {
    n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
    VERIFIE(1, FAUX, "ch_chemin, drapeau d'utilisation\n");
    couche->utilisee = (boolean) drapeau;
  }

  n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, FAUX, "ch_chemin, drapeau de contribution\n");
  if (drapeau == 1) {
    chemin->exf_calcule = GRF_Charger(fichier, version, nom_fichier, serie_contrib);
    if ( chemin->exf_calcule == NULL ) {
      (void) fprintf(stderr, "ch_chemin : erreur dans GRF_Charger / contribution\n");
      return FAUX;
    }
    if (couche != NULL)
      couche->signal = chemin->exf_calcule;
  }

  if (version < 23) /* Suite pas sauve avant... */
    return chemin;

  n_lu = fread(&n_item, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, FAUX, "ch_chemin, nombre de chemins quivalents\n");

  if (n_item > 0) {
    unsigned int i;

    for (i = 0 ; i < n_item ; i++) {
      Chemin *equivalent;

      equivalent = ch_chemin(fichier, version, nom_fichier, modele, serie_amp, serie_contrib, molecule);
      if ( NULL == equivalent ) {
	(void) fprintf(stderr, "ch_chemin : erreur en chargeant le chemin quivalent n%u / %u\n", i + 1, n_item);
	return NULL; /* fichier dj ferm dans ch_chemin */
      }
      CHEMIN_AjouterEquivalent(chemin, equivalent);
    }
  }

  return chemin;
}

/* -------- Chargement d'une couche -------- */

static Couche *ch_couche(FILE *fichier, int version, const char *nom_fichier, Molecule *molecule,
			 Serie_Graphes *amplitudes, Serie_Graphes *contributions, struct modele_feff *modele)
{
  Couche *couche;
  char *titre;
  int n_lu;
  unsigned char drapeau;
  unsigned long lg_titre;

  printf("ch_couche...\n");
  if (NULL == fichier) return NULL;

  couche = COUCHE_Creer(modele, NULL);
  if (NULL == couche) {
    (void) fprintf(stderr, "ch_couche : erreur dans COUCHE_Creer\n");
    return NULL;
  }

  n_lu = fread( &lg_titre, sizeof(unsigned long), 1, fichier);
/*   printf("Titre : %lu caractres\n", lg_titre); */
  VERIFIE(1, NULL, "ch_couche, longueur du titre\n");
  titre = (char *) calloc(lg_titre + 1, sizeof(char));
  if (titre == NULL) {
    (void) fprintf(stderr, "ch_couche : imposible de rserver le nom de la couche [%lu caractres]\n", lg_titre + 1);
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_couche", "titre");
    return NULL;
  }
  n_lu  = fread(titre, sizeof(char), lg_titre, fichier);
/*   printf("Titre lu : %s\n", titre); */
  VERIFIE( (int) lg_titre, NULL, "ch_couche, titre\n");
  COUCHE_Nommer(couche, titre);
  free(titre);

  n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_couche, drapeau / couche utilise\n");
  couche->utilisee = drapeau;

  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->N) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / N\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->R) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / R\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->Debye_Waller) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / Debye-Waller\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->Cumul_1) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / 1er cumulant\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->Cumul_3) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / 3e cumulant\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->Cumul_2) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / 2e cumulant\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->Cumul_4) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / 4e cumulant\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->Delta_E) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / DE0\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(couche->Gamma) ) ) {
    (void) fprintf(stderr, "ch_couche : erreur dans ch_variable / Gamma\n");
    return NULL;
  }

  n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_couche, drapeau / contribution\n");
  if ( VRAI == drapeau ) {
    couche->signal = GRF_Charger(fichier, version, nom_fichier, contributions);
    if ( NULL == couche->signal ) {
      (void) fprintf(stderr, "ch_couche : erreur dans GRF_Charger / contribution\n");
      return NULL;
    }
  }

  n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_couche, drapeau / amplitude\n");
  if ( VRAI == drapeau ) {
    couche->amplitude = GRF_Charger(fichier, version, nom_fichier, amplitudes);
    if ( NULL == couche->amplitude ) {
      (void) fprintf(stderr, "ch_couche : erreur dans GRF_Charger / amplitude\n");
      return NULL;
    }
  }
  n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_couche, drapeau / phase\n");
  if ( VRAI == drapeau ) {
    couche->phase = GRF_Charger(fichier, version, nom_fichier, amplitudes);
    if ( NULL == couche->phase ) {
      (void) fprintf(stderr, "ch_couche : erreur dans GRF_Charger / phase\n");
      return NULL;
    }
  }

  n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_couche, drapeau / chemin\n");
  if ( VRAI == drapeau ) {
    couche->chemin = ch_chemin(fichier, version, nom_fichier, NULL, amplitudes, contributions, molecule);
    if ( NULL == couche->chemin ) {
      (void) fprintf(stderr, "ch_couche : erreur dans ch_chemin\n");
      return NULL;
    }
    if (NULL == couche->chemin->amplitude)
      couche->chemin->amplitude = couche->amplitude;
    if (NULL == couche->chemin->phase)
      couche->chemin->phase = couche->phase;
    if (NULL == couche->chemin->exf_calcule)
      couche->chemin->exf_calcule = couche->signal;
  }

  return couche;
}

/* -------- Chargement de l'en-tte
   Renvoie le numro de version du fichier (0 si problme)
           dans type, le type de donnes contenues dans le fichier
*/

static int ch_entete(FILE *fichier, int *type)
{
  size_t n_lu;
  unsigned long code;
  unsigned int version;
  char ch_code[5] = {'L', 'A', 'S', 'E', 0};

  if (fichier == NULL) return 0;

  n_lu = fread(&code, sizeof(unsigned long), 1, fichier);
  VERIFIE(1, 0, "ch_entete - code\n");

  if ( code == 'LASE' ) { /* Codage correct, version >= 1.5 */
    double pi;

    n_lu = fread(&version, sizeof(unsigned int), 1, fichier);
    VERIFIE(1, 0, "ch_entete - version, > 1.5\n");

    n_lu = fread(&pi, sizeof(double), 1, fichier);
    VERIFIE(1, 0, "ch_entete - pi, version > 1.5\n");
    if ( pi != 3.1415 ) { /* Format des double incorrect */
      EXF_ERREUR(MAUVAIS_DOUBLE);
      return 0;
    }

    if ( version < 30 ) {
      char mode_3d;

      n_lu = fread(&mode_3d, sizeof(char), 1, fichier);
      VERIFIE(1, 0, "ch_entete - type de fichier, version > 1.5 et < 3.0\n");
      if (mode_3d == 1)
	*type = FICHIER_MOLECULE;
      else
	*type = FICHIER_GRAPHE;
    } else {
      n_lu = fread( type, sizeof(int), 1, fichier);
      VERIFIE(1, 0, "ch_entete - type de fichier, version >= 3.0\n");
    }

#ifdef _DEVELOPPE_
    printf("  Version du fichier : %d\n", version);
#endif

    return (int) version; /* Le numro de version dtect */
  }

  /* Ici, on n'a pas 'LASE' : ancienne version ? */

  if (code == 0x454D439BUL) { /* Une des toutes premires versions  : code 'EMC\233' */
    int version;

    n_lu = fread(&version, sizeof(int), 1, fichier);
    VERIFIE(1, 0, "ch_entete - version, < 1.3\n");
    *type = FICHIER_GRAPHE;

#ifdef _DEVELOPPE_
    printf("  Version du fichier : %d\n", version);
#endif

    return version;
  }

  /* On n'a pas 'EMC\233' -> a-t-on "LASE" ? */

  fseek(fichier, 0L, 0); /* On est au dbut */

  if (fgets(ch_code, 5, fichier) == NULL) {
#ifdef _DEVELOPPE_
    printf("ch_entete - fgets(ch_code) = NULL\n");
#endif
    (void) SYS_ERREUR(ERREUR_LECTURE, "Fichier courant\n", fichier, NULL);
    return 0;
  }

  if ( strcmp(ch_code, "LASE") == 0) /* Version 1.3 ou 1.4 */
  {
    int version;
    short int t_dbl;

    n_lu = fread(&version, sizeof(int), 1, fichier); /* Numro de version */
    VERIFIE(1, 0, "ch_entete - version, 1.3 ou 1.4\n");
    if (version < 13) /* Premire version avec LASE */
    {
      EXF_ERREUR(FORMAT_INCONNU);
      return 0;
    }

    n_lu = fread(&t_dbl, sizeof(short int), 1, fichier); /* Taille d'un double en octets */
    VERIFIE(1, 0, "ch_entete - taille des doubles, version 1.3 ou 1.4\n");
    if ( t_dbl != (short int) sizeof(double) ) /* Je suis prt  parier que jamais un double ne fera plus
						  de max(short int) octets ! */
    {
      EXF_ERREUR(MAUVAIS_DOUBLE);
      return 0;
    }

    *type = FICHIER_GRAPHE; /* Forcment, dans ces versions */
#ifdef _DEVELOPPE_
    printf("  Version du fichier : %d\n", version);
#endif

    return version;
  }

  EXF_ERREUR(FORMAT_INCONNU);
  return 0; /* Format ou numro de version inconnu */
}

/* -------- Chargement d'un fragment ------ */

static Fragment *ch_fragment(FILE *fichier, int version, const char *nom_fichier, Serie_Graphes *reference, Molecule *molecule)
{
  char *nom_fragment, drp_maille = FAUX, drp_origine = FAUX;
  unsigned int lg_titre, n_atms, i;
  size_t n_lu;
  Fragment *fragment;

#ifdef _DEVELOPPE_
  printf("ch_fragment : fichier = %p, version = %d, reference = %p, molecule = %p\n",
	 (void *) fichier, version, (void *) reference, (void *) molecule );
#endif

  if (fichier == NULL) return NULL;

  n_lu = fread(&lg_titre, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, NULL, "ch_fragment - lg_titre\n");
  nom_fragment = (char *) calloc(lg_titre + 1, sizeof(char));
  if (nom_fragment == NULL) {
#ifdef _DEVELOPPE_
    (void) fprintf(stderr, "ch_fragment : titre de %u caractres\n", lg_titre);
#endif
    (void) SYS_ERREUR(MANQUE_MEMOIRE, "charger.c", "ch_fragment", "nom_fragment");
    return NULL;
  }
  n_lu = fread(nom_fragment, sizeof(char), lg_titre, fichier);
  nom_fragment[lg_titre] = 0;

#ifdef _DEVELOPPE_
  printf("  chargement du fragment %s\n", nom_fragment);
#endif
  HIST_Texte(5, 0, " ... fragment : %s\n", nom_fragment);

  if (version >= 19) { /* Ces drapeaux sont apparus avec la version 1.9 */
    n_lu = fread(&drp_maille, sizeof(char), 1, fichier);
    if (drp_maille == VRAI) {
      n_lu = fread(&drp_origine, sizeof(char), 1, fichier);
    }
  }

  n_lu = fread(&n_atms, sizeof(unsigned int), 1, fichier);
#ifdef _DEVELOPPE_
  printf("    Il contient %u atomes\n", n_atms);
#endif
  fragment = FRG_creer(molecule, NULL, nom_fragment, n_atms, NULL, NULL);
  free(nom_fragment);
  if (fragment == NULL) {
    (void) fprintf(stderr, "ch_fragment : erreur dans FRG_creer()\n");
    return NULL;
  }
  fragment->drapeaux.maille = drp_maille;
  fragment->drapeaux.origine = drp_origine;

  for (i = 0 ; i < n_atms ; i++) {
    unsigned long idx;
    Atome *atome;

    n_lu = fread(&idx, sizeof(unsigned long), 1, fichier);
    if (n_lu != 1) {
      FRG_effacer(molecule, fragment, VRAI, FAUX);
    }
    VERIFIE(1, NULL, "ch_fragment - index atome\n"); /* On renvoie le fragment tel quel si problme */
    atome = ATM_index(molecule, idx);
    FRG_ajoute_atome(fragment, 1, &atome);
  }

  if (version >= 20) {/* Schma de liaison :  partir de la version 2.0 */
    unsigned int n_liaisons;

    n_lu = fread(&n_liaisons, sizeof(unsigned int), 1, fichier);
    VERIFIE(1, NULL, "ch_fragment - nombre de liaisons\n");
#ifdef _DEVELOPPE_
  printf("    Il contient %u liaisons\n", n_liaisons);
#endif

    for (i = 0 ; i < n_liaisons ; i++) {
      if (ch_liaison(fichier, version, nom_fichier, molecule) == FAUX) {
	FRG_effacer(molecule, fragment, VRAI, FAUX);
	return NULL;
      }
    }
  }

  return fragment;
}

/* -------- Charger une liaison --------- */

static Liaison *ch_liaison(FILE *fichier, int version, const char *nom_fichier __attribute__((__unused__)), Molecule *molecule)
{
  float multiplicite;
  unsigned short int n_atomes, type;
  unsigned long idx;
  Atome *atome_1, *atome_2;
  Liaison *liaison;
  size_t n_lu;

#ifdef _DEVELOPPE_
  printf("ch_liaison : fichier = %p, version = %d, molecule = %p\n", (void *) fichier, version, (void *) molecule);
#endif

  if (fichier == NULL) return NULL;

  n_lu = fread(&n_atomes, sizeof(unsigned short int), 1, fichier);
  VERIFIE(1, NULL, "ch_liaison - nombre d'atomes");

  n_lu = fread(&idx, sizeof(unsigned long), 1, fichier);
  VERIFIE(1, NULL, "ch_liaison - index atome 1");

#ifdef _DEVELOPPE_
  printf("ch_liaison -> ");
#endif
  atome_1 = ATM_index(molecule, idx);
  n_lu = fread(&idx, sizeof(unsigned long), 1, fichier);
  VERIFIE(1, NULL, "ch_liaison - index atome 2");
  atome_2 = ATM_index(molecule, idx);

  if (n_atomes > 2) {
    unsigned short int i;

    n_lu = 0;
    for (i = 0 ; i < n_atomes - 2 ; i++) {
      n_lu += fread(&idx, sizeof(unsigned long), 1, fichier);
    }
    VERIFIE( (unsigned int) (n_atomes - 2), NULL, "ch_liaison - atomes complmentaires");
  }

  n_lu = fread(&multiplicite, sizeof(float), 1, fichier);
  VERIFIE(1, NULL, "ch_liaison - multiplicit de la liaison");

  n_lu = fread(&type, sizeof(unsigned short int), 1, fichier);
  VERIFIE(1, NULL, "ch_liaison - type de liaison");

  liaison = creer_liaison(NULL, atome_1, atome_2, (int) multiplicite, type);
  if (liaison == NULL) {
    return NULL;
  }

  return liaison;
}

/* -------- Charger un modle ----------- */

static struct modele_ajuste *ch_modele(FILE *fichier, int version, const char *nom_fichier, Serie_Graphes *graphes, void *complement_1)
{
  struct modele_ajuste *modele = NULL;
  int type, n_lu;
  char *nom;
  unsigned long lg_nom;
  unsigned char drapeau;

  n_lu = fread(&type, sizeof(int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele - type de modle");

  n_lu = fread(&lg_nom, sizeof(unsigned long), 1, fichier);
  VERIFIE(1, NULL, "ch_modele - longueur du nom");

  nom = (char *) calloc( lg_nom + 1, sizeof(char) );
  if ( NULL == nom ) {
    /* Normalement, pas trop gnant, mais on n'aura srement pas la place pour la suite, alors... */
    ERR_Memoire( "Nom du modele en cours de chargement", sizeof( char ) * ( lg_nom + 1 ) );
    (void) fprintf(stderr, "ch_modele : impossible de rserver le nom !\n");
    return NULL;
  }
  n_lu = fread(nom, sizeof(char), lg_nom, fichier);
  VERIFIE( (int) lg_nom, NULL, "ch_modele - longueur du nom");
  printf("%s : nom du modle = %s\n", __FUNCTION__, nom);

  modele = MODELE_CreerModele( NULL, nom, type, NULL, VRAI );
  free( nom ); nom = NULL ;
  if (NULL == modele) {
    (void) fprintf(stderr, "ch_modele : impossible de crer le modle !\n");
    return NULL;
  }

  n_lu = fread(&(modele->x_min), sizeof(double), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, x min\n");
  n_lu = fread(&(modele->x_max), sizeof(double), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, x max\n");

  n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, drapeau / graphe exprimental\n");
  if (VRAI == drapeau) {
    modele->experience = GRF_Charger(fichier, version, nom_fichier, graphes);
    if (NULL == modele->experience) {
      (void) fprintf(stderr, "ch_modele : erreur dans GRF_Charger / exprience\n");
      MODELE_EffacerModele( NULL, modele, FAUX );
      return NULL;
    }
  }
  n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, drapeau / graphe modle\n");
  if (VRAI == drapeau) {
    modele->modele = GRF_Charger(fichier, version, nom_fichier, graphes);
    if (NULL == modele->modele) {
      (void) fprintf(stderr, "ch_modele : erreur dans GRF_Charger / modle\n");
      MODELE_EffacerModele( NULL, modele, VRAI );
      return NULL;
    }
  }
  n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, drapeau / graphe rsidu\n");
  if (VRAI == drapeau) {
    modele->residu = GRF_Charger(fichier, version, nom_fichier, graphes);
    if (NULL == modele->residu) {
      (void) fprintf(stderr, "ch_modele : erreur dans GRF_Charger / rsidu\n");
      MODELE_EffacerModele( NULL, modele, VRAI );
      return NULL;
    }
  }

  n_lu = fread(&(modele->algorithme), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, algorithme\n");
  n_lu = fread(&(modele->max_iteration), sizeof(unsigned int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, nombre maximal d'itrations\n");
  n_lu = fread(&(modele->condition_arret), sizeof(int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, conditions d'arrt\n");
  n_lu = fread(&(modele->err_min), sizeof(short int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, erreur minimale\n");
  n_lu = fread(&(modele->derr_min), sizeof(short int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, variation minimale de l'erreur\n");

  n_lu = fread( &drapeau, sizeof(unsigned char), 1, fichier );
  VERIFIE(1, NULL, "ch_modele, drapeau / pondration\n");
  modele->drapeaux.pondere = drapeau;
#ifdef _DEVELOPPE_
  printf( " ==> %s <== [ponderation : %d]\n",
	  __FUNCTION__, modele->drapeaux.pondere );
#endif

  n_lu = fread(&drapeau, sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_modele, drapeau / contraintes\n");
  modele->drapeaux.contraintes = drapeau;

  switch (modele->type)
  {
  case MODELE_EXAFS_CHEMINS:
    modele->infos_modele.feff = ch_modele_feff(fichier, version, nom_fichier, (Molecule *) complement_1);
    if (NULL == modele->infos_modele.feff) {
      (void) fprintf(stderr, "ch_modele : erreur dans ch_modele_feff / rsidu\n");
      MODELE_EffacerModele( NULL, modele, VRAI );
      return NULL;
    }
    break;

  default:
    fprintf(stderr, "ch_modele :  FAIRE, chargement des modles de type %d\n", modele->type);
  }

  return modele;
}

/* -------- Charger un modle FEFF -------- */

static struct modele_feff *ch_modele_feff(FILE *fichier, int version, const char *nom_fichier, Molecule *molecule)
{
  Serie_Graphes *amplitudes;
  unsigned char drapeau;
  unsigned int n_lu, nbr_couches=0, i;
  unsigned long idx_atome;
  struct modele_feff *modele;

  if (NULL == fichier) return NULL;

  modele = FEFF_CreerModele();
  if (NULL == modele) {
    (void) fprintf(stderr, "ch_modele_feff : erreur dans FEFF_CreerModele\n");
  }

  n_lu = fread( &(modele->seuil), sizeof(int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele_feff, seuil\n");
#ifdef _DEVELOPPE_
  printf("ch_modele_feff : seuil = %d\n", modele->seuil);
#endif

  n_lu = fread( &idx_atome, sizeof(unsigned long), 1, fichier);
  VERIFIE(1, NULL, "ch_modele_feff, atome absorbeur\n");
  modele->absorbeur = ATM_index(molecule, idx_atome);

  n_lu = fread( modele->potentiels, sizeof(int), 9, fichier);
  VERIFIE(9, NULL, "ch_modele_feff, table des potentiels\n");

  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(modele->S0_2) ) ) {
    (void) fprintf(stderr, "ch_modele_feff : erreur dans ch_variable / S0\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(modele->Delta_E) ) ) {
    (void) fprintf(stderr, "ch_modele_feff : erreur dans ch_variable / DE0\n");
    return NULL;
  }
  if ( FAUX == ch_variable( fichier, version, nom_fichier, &(modele->Gamma) ) ) {
    (void) fprintf(stderr, "ch_modele_feff : erreur dans ch_variable / Gamma\n");
    return NULL;
  }

  n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_modele_feff, drapeau / DE0 gnral\n");
  modele->drapeaux.DE0_general = drapeau;
  n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_modele_feff, drapeau / Gamma gnral\n");
  modele->drapeaux.GAM_general = drapeau;

  amplitudes = SERIE_GRF_CreerSerie(serie_courante, "Param"e_GRAVE"tres "e_AIGU"lectroniques du mod"e_GRAVE"le FEFF");
  if (NULL == amplitudes) {
    (void) fprintf(stderr, "ch_modele_feff : impossible de crer la srie des amplitudes & dphasages !\n");
    amplitudes = serie_courante;
  }

  n_lu = fread( &nbr_couches, sizeof(unsigned int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele_feff, nombre de couches\n");
  if ( nbr_couches > 0 ) {
    Serie_Graphes *contributions;

    contributions = SERIE_GRF_CreerSerie(amplitudes, "Contributions des chemins du mod"e_GRAVE"le FEFF");
    if (NULL == contributions) {
      (void) fprintf(stderr, "ch_modele_feff : impossible de crer la srie des contributions !\n");
      contributions = serie_courante;
    }

    for (i = 0 ; i < nbr_couches ; i++) {
      if ( NULL == ch_couche(fichier, version, nom_fichier, molecule, amplitudes, contributions, modele) ) {
	(void) fprintf(stderr, "ch_modele_feff : erreur dans ch_couche / couche %u sur %u\n", i + 1, nbr_couches);
	return NULL;
      }
    }

    /* On efface la srie des contributions, si elles n'existent pas */
    if ( 0 == contributions->nbr_graphes ) {
      SERIE_GRF_EffacerSerie(contributions, VRAI);
    }
  }

  n_lu = fread( &(modele->ponderation), sizeof(unsigned short int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele_feff, pondration\n");

  n_lu = fread( &(modele->type_lpm), sizeof(unsigned short int), 1, fichier);
  VERIFIE(1, NULL, "ch_modele_feff, type de lpm\n");
  if ( MODELE_LPM_GRAPHE == modele->type_lpm ) {
    n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
    VERIFIE(1, NULL, "ch_modele_feff, drapeau / graphe de lpm\n");
    if ( VRAI == drapeau ) {
      modele->lpm = GRF_Charger(fichier, version, nom_fichier, amplitudes);
      if ( NULL == modele->lpm ) {
	(void) fprintf(stderr, "ch_modele_feff : erreur dans GRF_Charger / lpm\n");
	return NULL;
      }
    }
  } else if ( MODELE_LPM_GAMMA_ETA == modele->type_lpm ) {
    n_lu = fread(&(modele->lpm_val_eta), sizeof(double), 1, fichier);
    VERIFIE(1, NULL, "ch_modele_feff, lpm : valeur pour ta\n");
  }

  n_lu = fread( &(drapeau), sizeof(unsigned char), 1, fichier);
  VERIFIE(1, NULL, "ch_modele_feff, drapeau / graphe de phase de l'absorbeur\n");
  if ( VRAI == drapeau ) {
    modele->phase_central = GRF_Charger(fichier, version, nom_fichier, amplitudes);
    if ( NULL == modele->phase_central ) {
      (void) fprintf(stderr, "ch_modele_feff : erreur dans GRF_Charger / phase de l'absorbeur\n");
      return NULL;
    }
  }

  return modele;
}

/* -------- Charger une variable -------- */

static int ch_variable(FILE *fichier, int version, const char *nom_fichier, Variable *variable)
{
  size_t n_lu;

  if (fichier == NULL) return FAUX;
  if (variable == NULL) return FAUX;

  n_lu = fread(&(variable->valeur_initiale), sizeof(double), 1, fichier);
  VERIFIE(1, FAUX, "ch_variable - valeur initiale\n");
  n_lu = fread(&(variable->type), sizeof(int), 1, fichier);
  VERIFIE(1, FAUX, "ch_variable - type\n");
  /*  printf("Variable : valeur initiale = %f\n", variable->valeur_initiale); */

  switch (variable->type)
  {
  case VAR_CONSTANTE:
    variable->min = variable->valeur_initiale;
    variable->max = variable->valeur_initiale;
    variable->delta = 1e-5;
    variable->contenu.valeur = variable->valeur_initiale;
    return VRAI;

  case VAR_LIBRE:
    n_lu  = fread(&(variable->min), sizeof(double), 1, fichier);
    n_lu += fread(&(variable->max), sizeof(double), 1, fichier);
    n_lu += fread(&(variable->delta), sizeof(double), 1, fichier);
    variable->contenu.valeur = variable->valeur_initiale;
    VERIFIE(3, FAUX, "ch_variable - min, max ou delta");
    return VRAI;

  case VAR_COPIE:
    return ch_variable(fichier, version, nom_fichier, variable); /* Perte de l'info copie,
								    on remplace par la variable originale */
  }

  (void) fprintf( stderr, "%s,%s l.%u : type de variable inconnu = %d\n",
		  __FILE__, __FUNCTION__, __LINE__, variable->type );
  return FAUX;
}
