/* LASE - Logiciel d'Analyse des Spectres EXAFS
   (c) Emmanuel CURIS 1996 - 2015

   Programme principal : répartition des événements
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <locale.h>

#include "systeme.h"

#include "exafs.h"
#include "serie.h" /* graphes, images, molécules */

#include "absorption.h"		/* Pour libérer les tables */
#include "couches.h"
#include "feff.h"		/* Pour MOL_tt_effacer() ??? */
#include "fichiers.h"		/* Pour charger() */
#include "images.h"   /* Pour les effacer à la fin */
#include "interface.h"
#include "statistiques.h"	/* Pour init_aleatoire() */
#include "tracer.h"

#include "fluorescence.h"

#include "principal.txt"

/* À définir pour avoir une version spéciale manip */
/* #define VERSION_POUR_LA_MANIP */

/* ~~~~~~~~ Fonctions extérieures appelées dans ce module ~~~~~~~~ */

extern int CONF_LireFichierRC( void ); /* Dans configuration.c */

extern void init_ncouches(void); /* Dans N_couches/general.c */

/* ~~~~~~~~ Fonctions locales à cette procédure ~~~~~~~~~~~~~~~~~~ */

/* ----------------- La fonction principale - le coeur du programme, comme toujours ------------------- */

int main(int nbr_param, char *liste_param[])
{
  boolean fin = FAUX;    /* VRAI si on quitte le programme            */
  char *localisation = NULL;
  int retour, i;
  int x, y, n_clic, bouton, etat;
  unsigned int touche;
  Graphe *graphe = NULL;

#ifdef _DEVELOPPE_
  if ( DEBOGUE_init_logs() == FAUX ) {
    (void) fprintf( stderr, "Impossible de démarrer les fichiers log !\n" );
    return -1;
  }
# ifdef AVEC_DB_MALLOC_
  {
    union dbmalloptarg options;

    options.i = M_HANDLE_ABORT;
    dbmallopt( MALLOC_WARN, &options );
  }
# endif
#endif

  /* On définit la locale pour être tranquille */
#ifdef _DEVELOPPE_
  (void) printf( "%s,%s l.%u : Choix de la localisation\n",
    __FILE__, __FUNCTION__, __LINE__ );
#endif
#ifdef CODAGE_LATIN1_
  localisation = setlocale( LC_ALL, "ISO-8859-1" );
  if ( NULL == localisation ) {
    (void) fprintf( stderr, "%s,%s l.%u : Impossible de definir la localisation comme du Latin-1 sans preciser la langue...\n",
		    __FILE__, __FUNCTION__, __LINE__ );
  }
#endif
  if ( NULL == localisation ) {
    localisation = setlocale( LC_ALL, "" );
  }
  if ( NULL == localisation ) {
    (void) fprintf( stderr, "%s,%s l.%u : Impossible de definir la localisation avec les valeurs par defaut...\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    localisation = setlocale( LC_ALL, "C" );
  }
  if ( NULL == localisation ) {
    (void) fprintf( stderr, "%s,%s l.%u : Impossible de définir la localisation\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    FIN_ANTICIPEE( FIN_ERREUR_MINEURE );
#ifdef _DEVELOPPE_
  } else {
    (void) printf( "Locale obtenue : %s\n", localisation );
#endif
  }

  /* Contrôles des localisations */
#ifdef utf8
  if ( strstr( localisation, 'UTF-8' ) == NULL ) {
    AVERTISSEMENT( "Version de LASE UTF-8, mais votre locale n'est pas en UTF-8 [%s]. Des problèmes d'affichage sont à prévoir.", localisation )
  }
#endif

  /* Démarrage de la bibliothèque d'interface */
  if ( BI_Initialiser( "exafs.rsc", 0, &nbr_param, &liste_param ) == FAUX ) {
    (void) fprintf( stderr, txt_ERR_Initialisation );
#ifdef _DEVELOPPE_
    DEBOGUE_arrete_logs();
#endif
    return -1;
  }

  retour = InitialiserInterface();
  if ( retour != 0 ) {
    (void) fprintf( stderr, "main : ERREUR dans InitialiserInterface [code %d]\n", retour );
    BI_Terminer();
#ifdef _DEVELOPPE_
    DEBOGUE_arrete_logs();
#endif
    return -2;
  }

  retour = CONF_LireFichierRC();
  if ( retour != 0 ) {
    AVERTISSEMENT( "Problème de lecture du fichier de configuration [code %d]\n", retour );
  }

  init_aleatoire( -1L ); /* Initialisation du générateur de nombres aléatoires - avec l'horloge système */

  Var_X.type  = VAR_CONSTANTE; /* Variables utilisées pour la construction de formules */
  Var_Y.type  = VAR_CONSTANTE;
  Var_Z.type  = VAR_CONSTANTE;
  Var_T.type  = VAR_CONSTANTE;
  Var_I.type  = VAR_CONSTANTE;
  Var_sX.type = VAR_CONSTANTE;
  Var_sY.type = VAR_CONSTANTE;

  /* Ouverture de la fenêtre de travail */
  retour = I_OuvrirPrincipale_GRF( );
  if ( retour != 0 ) {
    (void) fprintf( stderr, "%s, %s l.%u : ERREUR en voulant ouvrir la fenêtre principale.\n",
		    __FILE__, __FUNCTION__, __LINE__ );
    (void) BI_ERR_alerte( NULL, &ERR_FenetrePrincipale, VRAI );
    BI_Terminer();
#ifdef _DEVELOPPE_
    DEBOGUE_arrete_logs();
#endif
    return -5;
  }

  /* Si on a passé des fichiers en paramètre, on essaye de les charger...              */
  /* NB : avec le TOS 4.04 (Falcon), on ne peut passer qu'un seul fichier en paramètre */
  if ( nbr_param > 1 ) {
    for ( i = 1 ; i < nbr_param ; i++ ) {
      if ( liste_param[ i ] != NULL ) { /* Paramètres non traité auparavant */
	boolean opt_inconnue = FAUX; /* VRAI si l'option est inconnue -> message d'erreur */
	char *param = liste_param[ i ];

	if ( *param == '-' ) { /* C'est une option */
	  param++;
	  switch (*param) {
	  case '-':  /* -- -> nom de fichier commençant par un - */
	    graphe = charger( param );
	    break;

	  case 'H': /* -H -> on démarre l'historique */
	    GER_infos( HISTORIQUE, 0, 0 );
	    HIST_Titre( 0, "Lancement de LASE", "DEBUT" );
	    break;

	  default:
	    opt_inconnue = VRAI;
	  }
	} else {
	  /* Supposé un graphe, mais peut être une molécule, une image,... */
	  graphe = charger( param );
	}

	if ( VRAI == opt_inconnue ) {
	  (void) fprintf( stderr, txt_ERR_OptionInconnue, liste_param[ i ] );
	  opt_inconnue = FAUX; /* Pour la suivante */
	}
      }
    }
  }

#ifdef VERSION_POUR_LA_MANIP /* On démarre avec la boîte d'extraction des spectres */
  if ( serie_courante != NULL ) {
    if ( NULL == serie_courante->premier_graphe ) { /* Aucun graphe chargé */
      MEN_extrait( EXTRAIRE ); /* Comme ça s'occupe aussi des redessins */
      graphe = graphe_courant;
    } else
      graphe = serie_courante->dernier_graphe;
  } else {
      MEN_extrait( EXTRAIRE ); /* Comme ça s'occupe aussi des redessins */
      graphe = graphe_courant;
  }
#else /* Version normale : on démarre avec le sélecteur */
  if ( serie_courante != NULL ) {
    if ( NULL == serie_courante->premier_graphe ) /* Aucun graphe chargé par les options */
      graphe = charger( NULL );
    else
      graphe = serie_courante->dernier_graphe; /* Si le dernier fichier passé en paramètre n'est pas au bon format
						  on a graphe == NULL alors qu'il peut y avoir des graphes chargés
						  auparavant
						    -> menus incorrects et risques d'erreurs accrus !
						    -> on force à aller sur le dernier chargé pour être tranquille */
  } else { /* Pas de série : il faut charger... */
    graphe = charger( NULL );
  }
#endif

  if ( graphe != NULL ) { /* On a finalement chargé un graphe */
    GRF_aller( graphe );
  } else if ( serie_courante != NULL ) {
    GRF_aller( serie_courante->premier_graphe );
  } else if ( premiere_serie != NULL ) {
    I_SERIE_GRF_Aller( premiere_serie );
  } else { /* Aucun graphe en mémoire pour le moment */
    GRF_aller( NULL );
  }

  /* On prépare la liste de toute façon, car si 1re molécule créée, pas forcément correcte ! */
  MOL_PrepareListe( );

  /* Si on a chargé une molécule, on ouvre la fenêtre de vue en 3D */
  I_MOL_Visualisation3D();	/* Ne fais rien si pas de molécule... */

  /* Attente des événements */
  while ( fin == FAUX ) {
    retour = BI_EVENEMENT_TRAITE;
    while ( retour != BI_EVENEMENT_FIN_PRG ) {
      retour = BI_AttendEvenement( 0,
				   0, 0, 0, 0, 0, 0, /* Pas de boîte ni de compteur */
				   &x, &y, &bouton, &etat, &touche, &n_clic);
#ifdef _DEVELOPPE_
      if ( retour != BI_EVENEMENT_FIN_PRG ) {
	(void) fprintf( stderr, "%s,%s l.%u : Un événement n'a pas été traité automatiquement ! [code : %d]\n",
			__FILE__, __FUNCTION__, __LINE__, retour );
      }
#endif
    }

    /* On veut quitter */
    fin = VRAI;

    /* On ferme la fenetre de superposition */
    (void) TRC_FermerSuperposition();

    /* On ferme au maximum les fenêtres pour limiter les tests dans l'effacement des graphes */

    /* On efface les images */
    while ( IMG_premiere_serie != NULL ) {
      retour = IMG_SERIE_EffacerSerie( IMG_premiere_serie, VRAI, VRAI );
      if ( retour != 0 ) { /* Problème à l'effacement de la série => effacement annulé */
	fin = FAUX;
	break;
      }
    }

    /* On commence par effacer les molécules, car le sens contraire peut conduire à des surprises :
       si on veut sauver une molécule avant de l'effacer, mais que les graphes ont été effacés,
       on perd toutes les phases et amplitudes des chemins... */
    if ( VRAI == fin )
      fin = MOL_tteffacer();

    if ( VRAI == fin ) /* On n'a pas annulé la sortie */
      fin = GRF_tteffacer(); /* On efface tous les graphes - fin = VRAI : on a pu => on quitte */
  }

  /* On coupe l'historique */
  HIST_Titre( 0, "Fin de la s"e_AIGU"ance de travail", "FIN" );
  HIST_ArreterHistorique();

  /* On libère les tables de données */
  SEUIL_LibererTables();
  Ebel_EffacerTables();

  /* Autres terminaisons propres */
  STAT_TerminerAleatoire();

  BI_Terminer(); /* On veut effectivement quitter, donc on se prépare à quitter */

#ifdef _DEVELOPPE_
  DEBOGUE_arrete_logs();
#endif

  return 0;        /* Pas d'erreur survenue */
}
